/**************************************************************************

   Fotoxx      edit photos and manage collections

   Copyright 2007 2008 2009 2010 2011 2012 Michael Cornelison
   Source URL: http://kornelix.com/fotoxx
   Contact: kornelix2@gmail.com
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see http://www.gnu.org/licenses/.

***************************************************************************/

#define EX extern                                                          //  enable extern declarations
#include "fotoxx.h"


/**************************************************************************
   Fotoxx image edit - retouch functions
***************************************************************************/


/**************************************************************************/

//  automatic image tuneup without user guidance

float       voodoo_brdist[256];                                            //  pixel count per brightness level
int         voodoo_01, voodoo_99;                                          //  1% and 99% brightness levels (0 - 255)
void        voodoo_distribution();                                         //  compute brightness distribution
void *      voodoo_thread(void *);
editfunc    EFvoodoo;


void m_voodoo(GtkWidget *, cchar *)                                        //  new v.12.06
{
   zfuncs::F1_help_topic = "voodoo";

   EFvoodoo.funcname = "voodoo";
   EFvoodoo.Fprev = 0;                                                     //  no preview
   EFvoodoo.Farea = 1;                                                     //  select area ignored
   EFvoodoo.threadfunc = voodoo_thread;
   if (! edit_setup(EFvoodoo)) return;                                     //  setup edit

   voodoo_distribution();                                                  //  compute brightness distribution
   signal_thread();                                                        //  start update thread
   wait_thread_idle();                                                     //  wait for completion
   edit_done();                                                            //  edit done
   return;
}


//  compute brightness distribution for image or selected area

void voodoo_distribution()
{
   int         px, py, ii;
   int         ww = E1pxm->ww, hh = E1pxm->hh;
   float       bright1;
   float       *pix1;

   for (ii = 0; ii < 256; ii++)                                            //  clear brightness distribution data
      voodoo_brdist[ii] = 0;

   for (py = 0; py < hh; py++)                                             //  compute brightness distribution
   for (px = 0; px < ww; px++)
   {
      pix1 = PXMpix(E1pxm,px,py);
      bright1 = pixbright(pix1);
      voodoo_brdist[int(bright1)]++;
   }
   
   for (ii = 1; ii < 256; ii++)                                            //  cumulative brightness distribution
      voodoo_brdist[ii] += voodoo_brdist[ii-1];                            //   0 ... (ww * hh)
   
   voodoo_01 = 0;
   voodoo_99 = 255;

   for (ii = 0; ii < 256; ii++)                                            //  compute index values (0 - 255)
   {                                                                       //    for darkest 1% and brightest 1%
      if (voodoo_brdist[ii] < 0.01 * ww * hh) voodoo_01 = ii;
      if (voodoo_brdist[ii] < 0.99 * ww * hh) voodoo_99 = ii;
   }
   
   for (ii = 0; ii < 256; ii++)
      voodoo_brdist[ii] = voodoo_brdist[ii]                                //  multiplier per brightness level 
                    / (ww * hh) * 256.0 / (ii + 1);                        //  ( = 1 for a flat distribution)
   return;
}


//  thread function - use multiple working threads

void * voodoo_thread(void *)
{
   void  * voodoo_wthread(void *arg);

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(voodoo_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      CEF->Fmod = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}

void * voodoo_wthread(void *arg)                                           //  worker thread function
{
   int         index = *((int *) (arg));
   int         px, py;
   float       *pix1, *pix3;
   float       bright1, bright2, bright3, cmax;
   float       red1, green1, blue1, red3, green3, blue3;
   float       flat1 = 0.3;                                                //  strength of distribution flatten
   float       flat2 = 1.0 - flat1;
   float       sat1 = 0.4, sat2;                                           //  strength of saturation increase
   float       f1, f2;
   float       expand = 256.0 / (voodoo_99 - voodoo_01 + 1);               //  brightness range expander
   
   for (py = index; py < E1pxm->hh; py += Nwt)                             //  voodoo brightness distribution
   for (px = 0; px < E1pxm->ww; px++)
   {
      pix1 = PXMpix(E1pxm,px,py);                                          //  input pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel

      red1 = pix1[0];
      green1 = pix1[1];
      blue1 = pix1[2];

      bright2 = 0.25 * red1 + 0.65 * green1 + 0.10 * blue1;                //  input brightness, 0 - 256
      bright2 = (bright2 - voodoo_01) * expand;                            //  expand to clip low / high 1% 
      if (bright2 < 0) bright2 = 0;
      if (bright2 > 255) bright2 = 255;

      bright1 = voodoo_brdist[int(bright2)];                               //  factor for flat output brightness
      bright3 = flat1 * bright1 + flat2;                                   //  attenuate

      f1 = (256.0 - bright2) / 256.0;                                      //  bright2 = 0 - 256  >>  f1 = 1 - 0
      f2 = 1.0 - f1;                                                       //  prevent banding in bright areas
      bright3 = f1 * bright3 + f2;                                         //  tends to 1.0 for brighter pixels

      red3 = red1 * bright3;                                               //  blend new and old brightness
      green3 = green1 * bright3;
      blue3 = blue1 * bright3;
      
      bright3 = 0.333 * (red3 + green3 + blue3);                           //  mean color brightness
      sat2 = sat1 * (256.0 - bright3) / 256.0;                             //  bright3 = 0 - 256  >>  sat2 = sat1 - 0
      red3 = red3 + sat2 * (red3 - bright3);                               //  amplified color, max for dark pixels
      green3 = green3 + sat2 * (green3 - bright3);
      blue3 = blue3 + sat2 * (blue3 - bright3);
      
      if (red3 < 0) red3 = 0;                                              //  stop possible underflow
      if (green3 < 0) green3 = 0;
      if (blue3 < 0) blue3 = 0;

      cmax = red3;                                                         //  stop overflow, keep color balance
      if (green3 > cmax) cmax = green3;
      if (blue3 > cmax) cmax = blue3;
      if (cmax > 255.9) {
         cmax = 255.9 / cmax;
         red3 = red3 * cmax;
         green3 = green3 * cmax;
         blue3 = blue3 * cmax;
      }

      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************/

//  brightness / color / contrast adjustment

void  tune_curve_update(int spc);                                          //  curve update callback function
void  tune_update_image();                                                 //  process all pixels
void  tune_update_pixel(int px, int py);                                   //  process one pixel
void  *tune_thread(void *);                                                //  thread process for all pixels

int      tune_spc;                                                         //  current spline curve 1-5
int      tune_spc_moved[6];                                                //  tracks which curves changed

float    tune_pow[10000];                                                  //  pow(2,x) for x = -5.0 to +5.0
int      tune_ftf = 1;

editfunc    EFtune;


void m_tune(GtkWidget *, cchar *)                                          //  menu function
{
   int   tune_dialog_event(zdialog *zd, cchar *event);

   cchar    *title = ZTX("Adjust Brightness and Color");
   zfuncs::F1_help_topic = "tune";
   
   EFtune.funcname = "bright-color";
   EFtune.Fprev = 1;                                                       //  use preview
   EFtune.Farea = 2;                                                       //  select area usable
   EFtune.threadfunc = tune_thread;
   if (! edit_setup(EFtune)) return;                                       //  setup edit
   
   if (tune_ftf) {
      tune_ftf = 0;                                                        //  pre-compute pow(2,x)
      for (int ii = 0; ii < 10000; ii++)                                   //   for x = -5.0 to +4.999
         tune_pow[ii] = pow(2,(ii/1000.0 - 5.0));
   }
   
/***
            Adjust Brightness and Color
       ____________________________________________
      |                                            |
      |                                            |
      |                                            |
      |           curve drawing area               |
      |                                            |
      |                                            |
      |____________________________________________|
       darker areas                   lighter areas

      [+++] [---] [+ -] [- +] [+-+] [-+-]
      [x] small-steps  Curve File: [ Open ] [ Save ]
      (o) brightness       [reset 1]  [reset all]
      (o) saturation       [histogram]
      (o) red
      (o) green
      (o) blue
                                 [done] [cancel]
***/

   zdialog *zd = zdialog_new(title,mWin,Bdone,Bcancel,null);
   EFtune.zd = zd;

   zdialog_add_widget(zd,"frame","fr1","dialog",0,"expand");
   zdialog_add_widget(zd,"hbox","hba","dialog");
   zdialog_add_widget(zd,"label","labda","hba",Bdarker,"space=5");
   zdialog_add_widget(zd,"label","space","hba",0,"expand");
   zdialog_add_widget(zd,"label","labba","hba",Blighter,"space=5");

   zdialog_add_widget(zd,"hbox","hbb","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","b +++","hbb"," +++ ");
   zdialog_add_widget(zd,"button","b ---","hbb","‒ ‒ ‒");
   zdialog_add_widget(zd,"button","b +-", "hbb"," + ‒ ");
   zdialog_add_widget(zd,"button","b -+", "hbb"," ‒ + ");
   zdialog_add_widget(zd,"button","b +-+","hbb","+ ‒ +");
   zdialog_add_widget(zd,"button","b -+-","hbb","‒ + ‒");

   zdialog_add_widget(zd,"hbox","hbcf","dialog",0,"space=5");
   zdialog_add_widget(zd,"check","smallstep","hbcf",ZTX("small-steps"));
   zdialog_add_widget(zd,"label","labcf","hbcf",Bcurvefile,"space=5");
   zdialog_add_widget(zd,"button","load","hbcf",Bopen);
   zdialog_add_widget(zd,"button","save","hbcf",Bsave,"space=5");

   zdialog_add_widget(zd,"hbox","hb2","dialog");
   zdialog_add_widget(zd,"vbox","vb21","hb2");
   zdialog_add_widget(zd,"vbox","vb22","hb2");
   zdialog_add_widget(zd,"radio","radbri","vb21",Bbrightness);
   zdialog_add_widget(zd,"radio","radsat","vb21",ZTX("Saturation"));
   zdialog_add_widget(zd,"radio","radR","vb21",Bred);
   zdialog_add_widget(zd,"radio","radG","vb21",Bgreen);
   zdialog_add_widget(zd,"radio","radB","vb21",Bblue);

   zdialog_add_widget(zd,"hbox","hbrs","vb22",0,"space=5");
   zdialog_add_widget(zd,"label","space","hbrs",0,"space=10");
   zdialog_add_widget(zd,"button","reset1","hbrs",ZTX(" reset 1 "));
   zdialog_add_widget(zd,"button","resetA","hbrs",ZTX("reset all"));
   zdialog_add_widget(zd,"hbox","hbhist","vb22",0,"space=6");
   zdialog_add_widget(zd,"label","space","hbhist",0,"space=10");
   zdialog_add_widget(zd,"button","histo","hbhist",Bhistogram);
   
   GtkWidget *frame = zdialog_widget(zd,"fr1");                            //  setup for curve editing
   spldat *sd = splcurve_init(frame,tune_curve_update);
   EFtune.curves = sd;
   
   sd->Nscale = 3;                                                         //  y-scale lines at 1.0 EV intervals
   sd->xscale[0][0] = 0.00;                                                //  x and y values
   sd->yscale[0][0] = 0.25;
   sd->xscale[1][0] = 1.00;
   sd->yscale[1][0] = 0.25;
   sd->xscale[0][1] = 0.00;
   sd->yscale[0][1] = 0.50;
   sd->xscale[1][1] = 1.00;
   sd->yscale[1][1] = 0.50;
   sd->xscale[0][2] = 0.00;
   sd->yscale[0][2] = 0.75;
   sd->xscale[1][2] = 1.00;
   sd->yscale[1][2] = 0.75;

   for (int spc = 0; spc < 6; spc++)                                       //  setup 6 spline curves
   {                                                                       //  no. 0 is active curve
      sd->vert[spc] = 0;                                                   //  all curves are horizontal
      sd->nap[spc] = 3;                                                    //  curves 1-5 are copied to 0 when active
      sd->apx[spc][0] = 0.01;
      sd->apy[spc][0] = 0.5;
      sd->apx[spc][1] = 0.5;
      sd->apy[spc][1] = 0.5;
      sd->apx[spc][2] = 0.99;
      sd->apy[spc][2] = 0.5;
      splcurve_generate(sd,spc);
      tune_spc_moved[spc] = 0;
   }

   sd->Nspc = 1;                                                           //  only one at a time is active
   tune_spc = 1;                                                           //  default curve = brightness

   zdialog_stuff(zd,"radbri",1);                                           //  stuff default selection
   
   zdialog_resize(zd,0,500);
   zdialog_help(zd,"tune");                                                //  zdialog help topic
   zdialog_run(zd,tune_dialog_event,"save");                               //  run dialog - parallel
   return;
}


//  dialog event and completion callback function

int tune_dialog_event(zdialog *zd, cchar *event)
{
   int         Fupdate = 0;
   int         ii, jj, nn;
   static int  smallstep = 0;
   float       step, step2, step4;
   float       px, py;
   spldat      *sd = EFtune.curves;
   spldat      sdtemp;
   
   if (zd->zstat)
   {
      if (zd->zstat == 1) edit_done();                                     //  done
      else edit_cancel();                                                  //  cancel or destroy
      return 0;
   }

   if (strEqu(event,"reset"))                                              //  reset dialog controls
      event = "resetA";

   if (strEqu(event,"blendwidth")) tune_update_image();
   if (strEqu(event,"histo")) m_histogram(0,0);                            //  popup brightness histogram

   if (strnEqu(event,"rad",3)) {                                           //  new choice of curve
      ii = strcmpv(event,"radbri","radsat","radR","radG","radB",null);
      tune_spc = ii;
      sd->nap[0] = sd->nap[ii];                                            //  copy active curve to curve 0
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[0][jj] = sd->apx[ii][jj];
         sd->apy[0][jj] = sd->apy[ii][jj];
      }
      Fupdate++;
   }
   
   if (strEqu(event,"smallstep"))                                          //  toggle [+++] etc. step size
      zdialog_fetch(zd,"smallstep",smallstep);

   ii = tune_spc;                                                          //  current active curve

   if (strnEqu(event,"b ",2))                                              //  button [+++] etc. - move entire curve
   {
      step = 0.025;                                                        //  1/40
      if (smallstep) step = step * 0.3333;
      step2 = 2 * step;
      step4 = 4 * step;

      for (jj = 0; jj < sd->nap[0]; jj++) {                                //  12.04
         px = sd->apx[0][jj];
         py = sd->apy[0][jj];
         if (strEqu(event,"b +++")) py += step;                            //  range is 40 steps
         if (strEqu(event,"b ---")) py -= step;                            //   = -2.0 to +2.0 EV
         if (strEqu(event,"b +-"))  py += step - step2 * px;               //  normal steps are 0.1 EV or 0.03 OD
         if (strEqu(event,"b -+"))  py -= step - step2 * px;               //  small steps are 1/3 as big
         if (strEqu(event,"b +-+")) py -= step - step4 * fabsf(px-0.5);
         if (strEqu(event,"b -+-")) py += step - step4 * fabsf(px-0.5);
         if (py > 1) py = 1;
         if (py < 0) py = 0;
         sd->apy[0][jj] = py;
      }
      tune_spc_moved[ii] = 1;
      Fupdate++;
   }
   
   if (strEqu(event,"reset1")) {                                           //  reset current curve
      sd->nap[0] = 3;
      sd->apx[0][0] = 0.01;                                                //  3 anchor points, flatline
      sd->apy[0][0] = 0.5;
      sd->apx[0][1] = 0.5;
      sd->apy[0][1] = 0.5;
      sd->apx[0][2] = 0.99;
      sd->apy[0][2] = 0.5;
      Fupdate++;
      tune_spc_moved[ii] = 0;
   }
   
   if (strEqu(event,"resetA")) 
   {
      for (jj = 0; jj < 6; jj++) {                                         //  reset all curves
         sd->nap[jj] = 3;
         sd->apx[jj][0] = 0.01;
         sd->apy[jj][0] = 0.5;
         sd->apx[jj][1] = 0.5;
         sd->apy[jj][1] = 0.5;
         sd->apx[jj][2] = 0.99;
         sd->apy[jj][2] = 0.5;
         splcurve_generate(sd,jj);                                         //  regenerate all
         tune_spc_moved[jj] = 0;
      }
      Fupdate++;
   }

   if (strEqu(event,"load"))                                               //  load 5 saved curves
   {
      sdtemp.Nspc = 5;
      sdtemp.drawarea = 0;
      nn = splcurve_load(&sdtemp);
      if (nn != 1) return 0;

      for (ii = 0; ii < 5; ii++) {
         for (jj = 0; jj < sdtemp.nap[ii]; jj++) {
            sd->apx[ii+1][jj] = sdtemp.apx[ii][jj];
            sd->apy[ii+1][jj] = sdtemp.apy[ii][jj];
         }
         splcurve_generate(sd,ii+1);
         tune_spc_moved[ii+1] = 1;
      }

      ii = tune_spc;
      sd->nap[0] = sd->nap[ii];                                            //  copy active curve to curve 0
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[0][jj] = sd->apx[ii][jj];
         sd->apy[0][jj] = sd->apy[ii][jj];
      }
      Fupdate++;
   }

   if (strEqu(event,"save"))                                               //  save 5 curves to file
   {
      sdtemp.Nspc = 5;
      for (ii = 0; ii < 5; ii++) {
         sdtemp.nap[ii] = sd->nap[ii+1];
         for (jj = 0; jj < sd->nap[ii+1]; jj++) {
            sdtemp.apx[ii][jj] = sd->apx[ii+1][jj];
            sdtemp.apy[ii][jj] = sd->apy[ii+1][jj];
         }
      }
      splcurve_save(&sdtemp);
   }

   if (Fupdate)                                                            //  curve has changed
   {
      splcurve_generate(sd,0);                                             //  regenerate curve 0
      
      ii = tune_spc;                                                       //  active curve
      sd->nap[ii] = sd->nap[0];                                            //  copy curve 0 to active curve
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[ii][jj] = sd->apx[0][jj];
         sd->apy[ii][jj] = sd->apy[0][jj];
      }
      for (jj = 0; jj < 1000; jj++)
         sd->yval[ii][jj] = sd->yval[0][jj];

      gtk_widget_queue_draw(sd->drawarea);                                 //  draw curve
      tune_update_image();                                                 //  trigger image update
   }
   
   return 1;
}


//  this function is called when curve 0 is edited using mouse

void tune_curve_update(int)
{
   int      ii = tune_spc, jj;
   spldat   *sd = EFtune.curves;

   sd->nap[ii] = sd->nap[0];                                               //  copy curve 0 to current curve
   for (jj = 0; jj < sd->nap[0]; jj++) {
      sd->apx[ii][jj] = sd->apx[0][jj];
      sd->apy[ii][jj] = sd->apy[0][jj];
   }
   for (jj = 0; jj < 1000; jj++)
      sd->yval[ii][jj] = sd->yval[0][jj];

   tune_spc_moved[ii] = 1;
   tune_update_image();                                                    //  update image
   return;
}


//  Update image based on latest settings of all dialog controls.

void tune_update_image()
{
   int      px, py, ww, hh;

   if (sa_stat == 3 && sa_Npixel < 200000)                                 //  if relatively small area,
   {                                                                       //    process in main() thread
      for (py = sa_miny; py < sa_maxy; py++)
      for (px = sa_minx; px < sa_maxx; px++)
         tune_update_pixel(px,py);

      px = sa_minx;                                                        //  direct window update
      py = sa_miny;
      ww = sa_maxx - sa_minx;
      hh = sa_maxy - sa_miny;
      mwpaint3(px,py,ww,hh);
      CEF->Fmod = 1;
      return;
   }

   signal_thread();                                                        //  else use thread process
   return;
}


//  Update image based on latest settings of all dialog controls.

void * tune_thread(void *)
{
   void * tune_wthread(void *arg);

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(tune_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      CEF->Fmod = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * tune_wthread(void *arg)                                             //  worker thread function
{
   int      px, py;
   int      index = *((int *) arg);

   for (py = index; py < E1pxm->hh; py += Nwt)
   for (px = 0; px < E1pxm->ww; px++)
      tune_update_pixel(px,py);

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


void tune_update_pixel(int px, int py)                                     //  process one pixel
{
   int         ii, kk, dist = 0;
   float       *pix1, *pix3;
   float       red1, green1, blue1, red3, green3, blue3;
   float       xval, brmean, brout, brightness;
   spldat      *sd = EFtune.curves;

   if (sa_stat == 3) {                                                     //  select area active
      ii = py * E1pxm->ww + px;
      dist = sa_pixmap[ii];                                                //  distance from edge
      if (! dist) return;                                                  //  pixel outside area
   }

   pix1 = PXMpix(E1pxm,px,py);                                             //  input pixel
   pix3 = PXMpix(E3pxm,px,py);                                             //  output pixel
   
   red1 = red3 = pix1[0];                                                  //  input and output RGB values
   green1 = green3 = pix1[1];
   blue1 = blue3 = pix1[2];

   brightness = 0.25 * red3 + 0.65 * green3 + 0.10 * blue3;                //  perceived brightness
   xval = brightness / 256.0;                                              //  curve x-value, 0 to 0.999

/* ------------------------------------------------------------------------

      brightness adjustment

      brightness curve values:
           0 = dark
         0.5 = normal, unchanged
         1.0 = 200% brightness, clipped
*/

   if (tune_spc_moved[1])                                                  //  curve has been edited
   {
      kk = 1000 * xval;                                                    //  x-value 
      if (kk > 999) kk = 999;
      brout = sd->yval[1][kk];                                             //  y-value, 0 to 1.0
      brout = 4 * brout - 2;                                               //  F-stops, -2 to +2
      brout = tune_pow[int(1000*brout+5000.5)];                            //  0.25 to 4.0
      
      red3 = red3 * brout;                                                 //  apply to all colors
      green3 = green3 * brout;                                             //  remove limit check
      blue3 = blue3 * brout;                                               //  (result is better)
   }
      
/* ------------------------------------------------------------------------

      color saturation curve values:
           0 = no color saturation (gray scale)
         0.5 = normal (initial unmodified RGB)
         1.0 = max. color saturation

      0.5 >> 0:  move all RGB values to their mean: (R+G+B)/3
      0.5 >> 1.0:  increase RGB spread to 2x original level
*/

   if (tune_spc_moved[2])                                                  //  curve has been edited
   {
      kk = 1000 * xval;
      if (kk > 999) kk = 999;
      brout = sd->yval[2][kk];                                             //  saturation factor, 0 - 1
      brout = 2.0 * brout - 1.0;                                           //  -1.0 .. +1.0
      brmean = 0.333 * (red3 + green3 + blue3);
      red3 = red3 + brout * (red3 - brmean);                               //  simplified 
      green3 = green3 + brout * (green3 - brmean);
      blue3 = blue3 + brout * (blue3 - brmean);
      if (red3 < 0) red3 = 0;
      if (green3 < 0) green3 = 0;
      if (blue3 < 0) blue3 = 0;
   }

/* ------------------------------------------------------------------------

      color balance curve values:
         0 = 0.5 * original color
         0.5 = unmodified
         1.0 = 1.5 * original color, clipped
*/

   if (tune_spc_moved[3]) {                                                //  curve has been edited
      kk = 1000 * xval;
      if (kk > 999) kk = 999;
      brout = sd->yval[3][kk];                                             //  0 to 1.0
      brout = 4 * brout - 2;                                               //  F-stops, -2 to +2
      brout = tune_pow[int(1000*brout+5000.5)];                            //  0.25 to 4.0
      red3 = red3 * brout;
   }

   if (tune_spc_moved[4]) {
      kk = 1000 * xval;
      if (kk > 999) kk = 999;
      brout = sd->yval[4][kk];
      brout = 4 * brout - 2;
      brout = tune_pow[int(1000*brout+5000.5)];
      green3 = green3 * brout;
   }

   if (tune_spc_moved[5]) {
      kk = 1000 * xval;
      if (kk > 999) kk = 999;
      brout = sd->yval[5][kk];
      brout = 4 * brout - 2;
      brout = tune_pow[int(1000*brout+5000.5)];
      blue3 = blue3 * brout;
   }

//  if working within a select area, blend changes over distance from edge
      
   float       dold, dnew;

   if (sa_stat == 3 && dist < sa_blend) {
      dnew = 1.0 * dist / sa_blend;
      dold = 1.0 - dnew;
      red3 = dnew * red3 + dold * red1;
      green3 = dnew * green3 + dold * green1;
      blue3 = dnew * blue3 + dold * blue1;
   }

//  prevent overflow and set output RGB values

   float       cmax;

   cmax = red3;                                                            //  detect overflow
   if (green3 > cmax) cmax = green3;
   if (blue3 > cmax) cmax = blue3;
   
   if (cmax > 255.9) {                                                     //  stop overflow
      red3 = red3 * 255.9 / cmax;
      green3 = green3 * 255.9 / cmax;
      blue3 = blue3 * 255.9 / cmax;
   }

   pix3[0] = red3;                                                         //  0.0 to 255.9
   pix3[1] = green3;
   pix3[2] = blue3;
   
   return;
}


/**************************************************************************/

//  adjust image gamma using classic gamma curve

editfunc    EFgamma;
int         gamma_spc;                                                     //  current spline curve

void m_gamma(GtkWidget *, cchar *)
{
   int    gamma_dialog_event(zdialog* zd, cchar *event);
   void   gamma_curvedit(int spc);
   void * gamma_thread(void *);

   zfuncs::F1_help_topic = "gamma_curve";

   EFgamma.funcname = "gamma";
   EFgamma.Fprev = 1;                                                      //  use preview
   EFgamma.Farea = 2;                                                      //  select area usable
   EFgamma.threadfunc = gamma_thread;
   if (! edit_setup(EFgamma)) return;                                      //  setup edit
   
/***
          _________________________________________
         |                                         |                       //  5 curves are maintained:
         |                                         |                       //  curve 0: current display curve
         |                                         |                       //        1: curve for all colors
         |           curve edit area               |                       //        2,3,4: red, green, blue
         |                                         |
         |                                         |
         |_________________________________________|

            (o) all  (o) red  (o) green  (o) blue                          //  colors: all, red, green, blue
            [ + ] [ - ] [ + - ] [ - + ]  [reset]
            Curve File: [Open] [Save]   [histogram]
                                   [Done]  [Cancel]    
***/

   EFgamma.zd = zdialog_new(ZTX("adjust image gamma"),mWin,Bdone,Bcancel,null);
   zdialog *zd = EFgamma.zd;

   zdialog_add_widget(zd,"frame","frame","dialog",0,"expand");
   zdialog_add_widget(zd,"hbox","hbrgb","dialog",0,"space=3");
   zdialog_add_widget(zd,"radio","all","hbrgb",Ball,"space=5");
   zdialog_add_widget(zd,"radio","red","hbrgb",Bred,"space=5");
   zdialog_add_widget(zd,"radio","green","hbrgb",Bgreen,"space=5");
   zdialog_add_widget(zd,"radio","blue","hbrgb",Bblue,"space=5");
   zdialog_add_widget(zd,"hbox","hbbu","dialog",0,"space=3");
   zdialog_add_widget(zd,"button","b ++","hbbu","   + +   ","space=4");
   zdialog_add_widget(zd,"button","b --","hbbu"," ‒ ‒ ","space=4");
   zdialog_add_widget(zd,"button","b +-","hbbu","  + ‒ ","space=4");
   zdialog_add_widget(zd,"button","b -+","hbbu"," ‒ +  ","space=4");
   zdialog_add_widget(zd,"button","reset","hbbu",Breset,"space=4");
   zdialog_add_widget(zd,"hbox","hbcf","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labcf","hbcf",Bcurvefile,"space=5");
   zdialog_add_widget(zd,"button","load","hbcf",Bopen);
   zdialog_add_widget(zd,"button","save","hbcf",Bsave,"space=3");
   zdialog_add_widget(zd,"button","hist","hbcf",Bhistogram,"space=15");
   
   GtkWidget *frame = zdialog_widget(zd,"frame");
   spldat *sd = splcurve_init(frame,gamma_curvedit);
   EFgamma.curves = sd;

   sd->Nscale = 1;                                                         //  diagonal fixed line for neutral curve
   sd->xscale[0][0] = 0.00;
   sd->yscale[0][0] = 0.00;
   sd->xscale[1][0] = 1.00;
   sd->yscale[1][0] = 1.00;

   for (int ii = 0; ii < 5; ii++)                                          //  loop curves 0-4
   {
      sd->nap[ii] = 4;                                                     //  initial curves are flat
      sd->vert[ii] = 0;
      sd->apx[ii][0] = sd->apy[ii][0] = 0.01;
      sd->apx[ii][1] = sd->apy[ii][1] = 0.33;
      sd->apx[ii][2] = sd->apy[ii][2] = 0.67;
      sd->apx[ii][3] = sd->apy[ii][3] = 0.99;
      splcurve_generate(sd,ii);
   }

   sd->Nspc = 1;                                                           //  only curve 0 is shown 
   gamma_spc = 1;                                                          //  current active curve: 1 (all)
   
   zdialog_stuff(zd,"all",1);                                              //  stuff default selection, all

   zdialog_resize(zd,260,350);
   zdialog_help(zd,"gamma_curve");                                         //  zdialog help topic 
   zdialog_run(zd,gamma_dialog_event,"save");                              //  run dialog - parallel

   return;
}


//  dialog event and completion callback function

int gamma_dialog_event(zdialog *zd, cchar *event)
{
   spldat      *sd = EFgamma.curves;
   int         ii, jj, nn, Fupdate = 0;
   float       step, step2, px, py;
   spldat      sdtemp;

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 1;
   }

   if (strnEqu(event,"b ",2))                                              //  button [ + ] etc.               v.12.04
   {                                                                       //  move entire curve
      Fupdate++;

      step = 0.025;                                                        //  1/40
      step2 = 2 * step;

      for (jj = 1; jj < sd->nap[0]-1; jj++) {                              //  shift anchor points but         v.12.04
         px = sd->apx[0][jj];                                              //    not end points of curve 0
         py = sd->apy[0][jj];
         if (strEqu(event,"b ++")) py += step;
         if (strEqu(event,"b --")) py -= step;
         if (strEqu(event,"b +-"))  py += step - step2 * px;
         if (strEqu(event,"b -+"))  py -= step - step2 * px;
         if (py > 1) py = 1;
         if (py < 0) py = 0;
         sd->apy[0][jj] = py;
      }
   }

   if (strEqu(event,"reset")) {                                            //  [reset] button         v.12.04
      sd->nap[0] = 4;                                                      //  restore curve 0 to initial state
      sd->apx[0][0] = sd->apy[0][0] = 0.01;
      sd->apx[0][1] = sd->apy[0][1] = 0.33;
      sd->apx[0][2] = sd->apy[0][2] = 0.67;
      sd->apx[0][3] = sd->apy[0][3] = 0.99;
      Fupdate++;
   }

   if (strEqu(event,"blendwidth")) signal_thread();                        //  adjust area edge blending

   if (strEqu(event,"hist")) m_histogram(0,0);                             //  show brightness distribution

   if (strstr("all red green blue",event)) {                               //  new choice of curve
      zdialog_fetch(zd,event,ii);
      if (! ii) return 0;                                                  //  ignore button OFF events

      ii = strcmpv(event,"all","red","green","blue",null);
      gamma_spc = ii;                                                      //  1, 2, 3, 4

      sd->nap[0] = sd->nap[ii];                                            //  copy active curve ii to curve 0
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[0][jj] = sd->apx[ii][jj];
         sd->apy[0][jj] = sd->apy[ii][jj];
      }

      Fupdate++;
   }

   if (strEqu(event,"load"))                                               //  load 4 saved curves from file
   {
      sdtemp.Nspc = 4;
      sdtemp.drawarea = 0;
      nn = splcurve_load(&sdtemp);
      if (nn != 1) return 0;

      for (ii = 0; ii < 4; ii++) {                                         //  load curves 0-3 to curves 1-4
         sd->vert[ii+1] = sdtemp.vert[ii];
         sd->nap[ii+1] = sdtemp.nap[ii];
         for (jj = 0; jj < sdtemp.nap[ii]; jj++) {
            sd->apx[ii+1][jj] = sdtemp.apx[ii][jj];
            sd->apy[ii+1][jj] = sdtemp.apy[ii][jj];
         }
         splcurve_generate(sd,ii+1);
      }

      ii = gamma_spc;                                                      //  current active curve, 1-4

      sd->nap[0] = sd->nap[ii];                                            //  copy active curve to curve 0
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[0][jj] = sd->apx[ii][jj];
         sd->apy[0][jj] = sd->apy[ii][jj];
      }

      Fupdate++;
   }

   if (strEqu(event,"save"))                                               //  save 4 curves to a file
   {
      sdtemp.Nspc = 4;
      for (ii = 0; ii < 4; ii++) {
         sdtemp.vert[ii] = sd->vert[ii+1];
         sdtemp.nap[ii] = sd->nap[ii+1];
         for (jj = 0; jj < sd->nap[ii+1]; jj++) {
            sdtemp.apx[ii][jj] = sd->apx[ii+1][jj];
            sdtemp.apy[ii][jj] = sd->apy[ii+1][jj];
         }
      }

      splcurve_save(&sdtemp);
   }

   if (Fupdate)                                                            //  curve has changed
   {
      splcurve_generate(sd,0);                                             //  regenerate curve 0
      
      ii = gamma_spc;                                                      //  active curve 1-4

      sd->nap[ii] = sd->nap[0];                                            //  copy curve 0 to active curve
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[ii][jj] = sd->apx[0][jj];
         sd->apy[ii][jj] = sd->apy[0][jj];
      }

      for (jj = 0; jj < 1000; jj++) 
         sd->yval[ii][jj] = sd->yval[0][jj];

      gtk_widget_queue_draw(sd->drawarea);                                 //  draw curve         gtk3

      if (gamma_spc == 1) {                                                //  "all" curve (1) was edited
         for (ii = 2; ii < 5; ii++) {                                      //  fix R/G/B curves (2-4) to match
            sd->nap[ii] = sd->nap[1];                                      //  v.12.04
            for (jj = 0; jj < sd->nap[1]; jj++) {
               sd->apx[ii][jj] = sd->apx[1][jj];
               sd->apy[ii][jj] = sd->apy[1][jj];
            }

            for (jj = 0; jj < 1000; jj++)
               sd->yval[ii][jj] = sd->yval[1][jj];
         }
      }

      signal_thread();                                                     //  trigger image update
   }

   return 1;
}


//  this function is called when a curve is edited

void gamma_curvedit(int spc)
{
   int      ii = gamma_spc, jj;
   spldat   *sd = EFgamma.curves;                                          //  active curve, 1-4

   sd->nap[ii] = sd->nap[0];                                               //  copy curve 0 to active curve
   for (jj = 0; jj < sd->nap[0]; jj++) {
      sd->apx[ii][jj] = sd->apx[0][jj];
      sd->apy[ii][jj] = sd->apy[0][jj];
   }

   for (jj = 0; jj < 1000; jj++) 
      sd->yval[ii][jj] = sd->yval[0][jj];

   if (gamma_spc == 1) {                                                   //  "all" curve (1) was edited
      for (ii = 2; ii < 5; ii++) {                                         //  fix R/G/B curves (2-4) to match
         sd->nap[ii] = sd->nap[1];
         for (jj = 0; jj < sd->nap[1]; jj++) {
            sd->apx[ii][jj] = sd->apx[1][jj];
            sd->apy[ii][jj] = sd->apy[1][jj];
         }

         for (jj = 0; jj < 1000; jj++)
            sd->yval[ii][jj] = sd->yval[1][jj];
      }
   }

   signal_thread();
   return;
}


//  gamma thread function

void * gamma_thread(void *arg)
{
   void * gamma_wthread(void *);
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(gamma_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion
      
      mwpaint2();                                                          //  update window
      CEF->Fmod = 1;                                                       //  image3 modified
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * gamma_wthread(void *arg)                                            //  worker thread function
{
   int         index = *((int *) arg);
   int         ii, dist = 0, px3, py3;
   float       *pix1, *pix3;
   float       red1, green1, blue1, maxrgb;
   float       red3, green3, blue3;
   float       coeff = 1000.0 / 256.0;
   float       dold, dnew;
   spldat      *sd = EFgamma.curves;

   for (py3 = index; py3 < E3pxm->hh; py3 += Nwt)                          //  loop output pixels
   for (px3 = 0; px3 < E3pxm->ww; px3++)
   {
      if (sa_stat == 3) {                                                  //  select area active
         ii = py3 * E3pxm->ww + px3;
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel outside area
      }
      
      pix1 = PXMpix(E1pxm,px3,py3);                                        //  input pixel
      pix3 = PXMpix(E3pxm,px3,py3);                                        //  output pixel
      
      red1 = pix1[0];                                                      //  input RGB brightness, 0-255.9
      green1 = pix1[1];
      blue1 = pix1[2];
      
      ii = coeff * red1;                                                   //  range 0-999 for gamma curve index
      red3 = 255.9 * sd->yval[2][ii];                                      //  output RGB brightness, 0-255.9
      ii = coeff * green1;
      green3 = 255.9 * sd->yval[3][ii];
      ii = coeff * blue1;
      blue3 = 255.9 * sd->yval[4][ii];
      
      maxrgb = red3;
      if (green3 > maxrgb) maxrgb = green3;
      if (blue3 > maxrgb) maxrgb = blue3;
      
      if (maxrgb > 255.9) {                                                //  stop overflow
         red3 = red3 * 255.9 / maxrgb;
         green3 = green3 * 255.9 / maxrgb;
         blue3 = blue3 * 255.9 / maxrgb;
      }

      if (sa_stat == 3 && dist < sa_blend) {                               //  blend changes over blendwidth
         dnew = 1.0 * dist / sa_blend;
         dold = 1.0 - dnew;
         red3 = dnew * red3 + dold * red1;
         green3 = dnew * green3 + dold * green1;
         blue3 = dnew * blue3 + dold * blue1;
      }

      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************

   Image Tone Mapping function
   enhance local contrast as opposed to overall contrast

   methodology:
   get brightness gradients for each pixel in 4 directions: SE SW NE NW
   amplify gradients using the edit curve (x-axis range 0-max. gradient)
   integrate 4 new brightness surfaces from the amplified gradients:
     - pixel brightness = prior pixel brightness + amplified gradient
     - the Amplify control varies amplification from zero to overamplified
   new pixel brightness = average from 4 calculated brightness surfaces

***************************************************************************/

float    *Tmap_brmap1;
float    *Tmap_brmap3[4];
int      Tmap_contrast99;
float    Tmap_amplify;

editfunc    EFtonemap;


void m_tonemap(GtkWidget *, cchar *)
{
   int      Tmap_dialog_event(zdialog *zd, cchar *event);
   void     Tmap_curvedit(int);
   void *   Tmap_thread(void *);

   zfuncs::F1_help_topic = "tone_mapping";

   int         ii, cc, px, py;
   float       *pix1;
   cchar       *title = ZTX("Tone Mapping");
   int         jj, sum, limit, condist[100];

   EFtonemap.funcname = "tonemap";
   EFtonemap.Farea = 2;                                                    //  select area usable
   EFtonemap.threadfunc = Tmap_thread;
   if (! edit_setup(EFtonemap)) return;                                    //  setup: no preview, select area OK

/***
            _____________________________
           |                             |
           |                             |
           |    curve drawing area       |
           |                             |
           |_____________________________| 
            low       contrast       high

            Amplify ========[]=========

            Curve File: [ Open ] [ Save ]
***/

   zdialog *zd = zdialog_new(title,mWin,Bdone,Bcancel,null);
   EFtonemap.zd = zd;

   zdialog_add_widget(zd,"frame","frame","dialog",0,"expand");
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labcL","hb1",ZTX("low"),"space=4");
   zdialog_add_widget(zd,"label","labcM","hb1",ZTX("contrast"),"expand");
   zdialog_add_widget(zd,"label","labcH","hb1",ZTX("high"),"space=5");

   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labcon","hb2",ZTX("Amplify"),"space=5");
   zdialog_add_widget(zd,"hscale","amplify","hb2","0|1|0.005|0","expand");
   zdialog_add_widget(zd,"label","ampval","hb2",0,"space=5");

   zdialog_add_widget(zd,"hbox","hbcf","dialog",0,"space=8");
   zdialog_add_widget(zd,"label","labcf","hbcf",Bcurvefile,"space=5");
   zdialog_add_widget(zd,"button","load","hbcf",Bopen,"space=5");
   zdialog_add_widget(zd,"button","save","hbcf",Bsave,"space=5");

   GtkWidget *frame = zdialog_widget(zd,"frame");                          //  set up curve edit
   spldat *sd = splcurve_init(frame,Tmap_curvedit);
   EFtonemap.curves = sd;

   sd->Nspc = 1;
   sd->vert[0] = 0;
   sd->nap[0] = 3;                                                         //  initial curve anchor points
   sd->apx[0][0] = 0.01;
   sd->apy[0][0] = 0.2;
   sd->apx[0][1] = 0.50;
   sd->apy[0][1] = 0.4;
   sd->apx[0][2] = 0.99;
   sd->apy[0][2] = 0.0;

   splcurve_generate(sd,0);                                                //  generate curve data

   cc = E1pxm->ww * E1pxm->hh * sizeof(float);                             //  allocate brightness map memory
   Tmap_brmap1 = (float *) zmalloc(cc,"tonemap");
   for (ii = 0; ii < 4; ii++)
      Tmap_brmap3[ii] = (float *) zmalloc(cc,"tonemap");

   for (py = 0; py < E1pxm->hh; py++)                                      //  map initial image brightness
   for (px = 0; px < E1pxm->ww; px++)
   {
      ii = py * E1pxm->ww + px;
      pix1 = PXMpix(E1pxm,px,py);
      Tmap_brmap1[ii] = pixbright(pix1);                                   //  pixel brightness 0-255.9
   }

   for (ii = 0; ii < 100; ii++) 
      condist[ii] = 0;

   for (py = 1; py < E1pxm->hh; py++)                                      //  map contrast distribution
   for (px = 1; px < E1pxm->ww; px++)
   {
      ii = py * E1pxm->ww + px;
      jj = 0.388 * fabsf(Tmap_brmap1[ii] - Tmap_brmap1[ii-1]);             //  horiz. contrast, ranged 0 - 99
      condist[jj]++;
      jj = 0.388 * fabsf(Tmap_brmap1[ii] - Tmap_brmap1[ii-E1pxm->ww]);     //  vertical
      condist[jj]++;
   }
   
   sum = 0;
   limit = 0.99 * 2 * (E1pxm->ww-1) * (E1pxm->hh-1);                       //  find 99th percentile contrast

   for (ii = 0; ii < 100; ii++) {
      sum += condist[ii];
      if (sum > limit) break;
   }
   
   Tmap_contrast99 = 255.0 * ii / 100.0;                                   //  0 to 255
   if (Tmap_contrast99 < 4) Tmap_contrast99 = 4;                           //  rescale low-contrast image

   zdialog_resize(zd,300,300);
   zdialog_help(zd,"tone_mapping");                                        //  zdialog help topic 
   zdialog_run(zd,Tmap_dialog_event,"save");                               //  run dialog - parallel

   Tmap_amplify = 0;
   return;
}


//  dialog event and completion callback function

int Tmap_dialog_event(zdialog *zd, cchar *event)
{
   spldat   *sd = EFtonemap.curves;
   char     text[8];

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      zfree(Tmap_brmap1);                                                  //  free memory
      for (int ii = 0; ii < 4; ii++)
      zfree(Tmap_brmap3[ii]);
      return 1;
   }

   if (strEqu(event,"reset"))                                              //  reset dialog controls 
      zdialog_stuff(zd,"amplify",0);

   if (strEqu(event,"blendwidth")) signal_thread();

   if (strEqu(event,"amplify")) {                                          //  slider value
      zdialog_fetch(zd,"amplify",Tmap_amplify);
      sprintf(text,"%.2f",Tmap_amplify);                                   //  numeric feedback
      zdialog_stuff(zd,"ampval",text);
      signal_thread();                                                     //  trigger update thread
   }

   if (strEqu(event,"load")) {                                             //  load saved curve 
      splcurve_load(sd);
      signal_thread();
      return 0;
   }

   if (strEqu(event,"save")) {                                             //  save curve to file 
      splcurve_save(sd);
      return 0;
   }
   
   return 0;
}


//  this function is called when the curve is edited

void Tmap_curvedit(int)
{
   signal_thread();
   return;
}


//  thread function

void * Tmap_thread(void *)
{
   void * Tmap_wthread1(void *arg);
   void * Tmap_wthread2(void *arg);

   int      ii;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      for (ii = 0; ii < 4; ii++)                                           //  start working threads1 (must be 4)
         start_wthread(Tmap_wthread1,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion
      
      for (ii = 0; ii < Nwt; ii++)                                         //  start working threads2
         start_wthread(Tmap_wthread2,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion
      
      CEF->Fmod = 1;
      mwpaint2(); 
   }

   return 0;                                                               //  not executed, stop g++ warning
}


//  working threads

void * Tmap_wthread1(void *arg)
{
   int         ii, kk, bii, pii, dist = 0;
   int         px, px1, px2, pxinc;
   int         py, py1, py2, pyinc;
   float       b1, b3, xval, yval, grad;
   float       amplify, contrast99;
   spldat      *sd = EFtonemap.curves;

   contrast99 = Tmap_contrast99;                                           //  99th percentile contrast
   contrast99 = 1.0 / contrast99;                                          //  inverted

   amplify = pow(Tmap_amplify,0.4);                                        //  get amplification, 0 to 1    v.12.06

   bii = *((int *) arg);

   if (bii == 0) {                                                         //  direction SE
      px1 = 1; px2 = E3pxm->ww; pxinc = 1;
      py1 = 1; py2 = E3pxm->hh; pyinc = 1;
      pii = - 1 - E3pxm->ww;
   }
   
   else if (bii == 1) {                                                    //  direction SW
      px1 = E3pxm->ww-2; px2 = 0; pxinc = -1;
      py1 = 1; py2 = E3pxm->hh; pyinc = 1;
      pii = + 1 - E3pxm->ww;
   }
   
   else if (bii == 2) {                                                    //  direction NE
      px1 = 1; px2 = E3pxm->ww; pxinc = 1;
      py1 = E3pxm->hh-2; py2 = 0; pyinc = -1;
      pii = - 1 + E3pxm->ww;
   }
   
   else {   /* bii == 3 */                                                 //  direction NW
      px1 = E3pxm->ww-2; px2 = 0; pxinc = -1;
      py1 = E3pxm->hh-2; py2 = 0; pyinc = -1;
      pii = + 1 + E3pxm->ww;
   }

   for (ii = 0; ii < E3pxm->ww * E3pxm->hh; ii++)                          //  initial brightness map
      Tmap_brmap3[bii][ii] = Tmap_brmap1[ii];

   for (py = py1; py != py2; py += pyinc)                                  //  loop all image pixels
   for (px = px1; px != px2; px += pxinc)
   {
      ii = py * E3pxm->ww + px;

      if (sa_stat == 3) {                                                  //  select area active
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  outside pixel
      }

      b1 = Tmap_brmap1[ii];                                                //  this pixel brightness
      grad = b1 - Tmap_brmap1[ii+pii];                                     //   - prior pixel --> gradient

      xval = fabsf(grad) * contrast99;                                     //  gradient scaled 0 to 1+
      kk = 1000.0 * xval;
      if (kk > 999) kk = 999;
      yval = 1.0 + 5.0 * sd->yval[0][kk];

      grad = grad * yval;                                                  //  magnified gradient

      b3 = Tmap_brmap3[bii][ii+pii] + grad;                                //  pixel brightness = prior + gradient
      b3 = (1.0 - amplify) * b1 + amplify * b3;                            //  constrain: push b3 toward b1

      if (b3 > 255) b3 = 255;                                              //  constrain
      if (b3 < 1) b3 = 1;

      Tmap_brmap3[bii][ii] = b3;                                           //  new pixel brightness
   }
   
   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


void * Tmap_wthread2(void *arg)
{
   float       *pix1, *pix3;
   int         index, ii, px, py, dist = 0;
   float       b1, b3, bf, f1, f2;
   float       red1, green1, blue1, red3, green3, blue3;

   index = *((int *) arg);

   for (py = index; py < E3pxm->hh; py += Nwt)                             //  loop all image pixels
   for (px = 0; px < E3pxm->ww; px++)
   {
      ii = py * E3pxm->ww + px;

      if (sa_stat == 3) {                                                  //  select area active 
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel is outside area
      }

      pix1 = PXMpix(E1pxm,px,py);                                          //  input pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel

      b1 = Tmap_brmap1[ii];                                                //  initial pixel brightness
      b3 = Tmap_brmap3[0][ii] + Tmap_brmap3[1][ii]                         //  new brightness = average of four
         + Tmap_brmap3[2][ii] + Tmap_brmap3[3][ii];                        //    calculated brightness surfaces

      bf = 0.25 * b3 / (b1 + 1);                                           //  brightness ratio
      
      red1 = pix1[0];                                                      //  input RGB
      green1 = pix1[1];
      blue1 = pix1[2];
      
      red3 = bf * red1;                                                    //  output RGB
      if (red3 > 255.9) red3 = 255.9;
      green3 = bf * green1;
      if (green3 > 255.9) green3 = 255.9;
      blue3 = bf * blue1;
      if (blue3 > 255.9) blue3 = 255.9;

      if (sa_stat == 3 && dist < sa_blend) {                               //  select area is active,
         f1 = 1.0 * dist / sa_blend;                                       //    blend changes over sa_blend
         f2 = 1.0 - f1;
         red3 = f1 * red3 + f2 * red1;
         green3 = f1 * green3 + f2 * green1;
         blue3 = f1 * blue3 + f2 * blue1;
      }
      
      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }

   exit_wthread();
   return 0;                                                               //  not executed, stop g++ warning
}


/**************************************************************************/

//  flatten brightness distribution

float    flatten_value = 0;                                                //  flatten value, 0 - 100%
float    flatten_deband = 0;                                               //  flatten deband, 0 - 100%
float    flatten_brdist[256];

void  flatten_update_image();
void  flatten_distribution();                                              //  compute brightness distribution
void  flatten_pixel(int px, int py);                                       //  compute new pixel brightness

editfunc    EFflatten;

void m_flatten(GtkWidget *, cchar *)
{
   int    flatten_dialog_event(zdialog* zd, cchar *event);
   void * flatten_thread(void *);

   cchar  *title = ZTX("Flatten Brightness Distribution");
   zfuncs::F1_help_topic = "flatten";

   EFflatten.funcname = "flatten";
   EFflatten.Fprev = 1;                                                    //  preview
   EFflatten.Farea = 2;                                                    //  select area usable
   EFflatten.threadfunc = flatten_thread;
   if (! edit_setup(EFflatten)) return;                                    //  setup edit

   zdialog *zd = zdialog_new(title,mWin,Bdone,Bcancel,null);
   EFflatten.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");               //  flatten slider
   zdialog_add_widget(zd,"label","labfd","hb1",ZTX("Flatten"),"space=5");
   zdialog_add_widget(zd,"hscale","flatten","hb1","0|100|1|0","expand");
   zdialog_add_widget(zd,"label","fvalue","hb1",0,"space=5");
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");               //  deband slider          v.12.06
   zdialog_add_widget(zd,"label","labdb","hb2",ZTX("Deband"),"space=5");
   zdialog_add_widget(zd,"hscale","deband","hb2","0|100|1|0","expand");
   zdialog_add_widget(zd,"label","dvalue","hb2",0,"space=5");

   zdialog_help(zd,"flatten");
   zdialog_resize(zd,300,0);
   zdialog_run(zd,flatten_dialog_event,"save");                            //  run dialog - parallel
   
   flatten_value = flatten_deband = 0;
   return;
}


//  dialog event and completion callback function

int flatten_dialog_event(zdialog *zd, cchar *event)                        //  flatten dialog event function
{
   char     text[8];

   if (zd->zstat)
   {
      if (zd->zstat == 1) edit_done();                                     //  done
      else edit_cancel();                                                  //  cancel or destroy
      return 0;
   }
   
   if (strEqu(event,"blendwidth")) flatten_update_image();

   if (strEqu(event,"flatten")) {
      zdialog_fetch(zd,"flatten",flatten_value);                           //  get slider value
      sprintf(text,"%.0f",flatten_value);                                  //  numeric feedback 
      zdialog_stuff(zd,"fvalue",text);
      flatten_update_image();                                              //  trigger update thread
   }
   
   if (strEqu(event,"deband")) {
      zdialog_fetch(zd,"deband",flatten_deband);                           //  get slider value          v.12.06
      sprintf(text,"%.0f",flatten_deband);                                 //  numeric feedback 
      zdialog_stuff(zd,"dvalue",text);
      flatten_update_image();                                              //  trigger update thread
   }
   
   return 1;
}


//  Update image based on select area and dialog controls

void flatten_update_image()
{
   int      px, py, ww, hh;

   if (sa_stat == 3 && sa_Npixel < 200000)                                 //  if relatively small area,
   {                                                                       //    process in main() thread
      flatten_distribution();

      for (py = sa_miny; py < sa_maxy; py++)
      for (px = sa_minx; px < sa_maxx; px++)
         flatten_pixel(px,py);

      px = sa_minx;                                                        //  direct window update
      py = sa_miny;
      ww = sa_maxx - sa_minx;
      hh = sa_maxy - sa_miny;
      mwpaint3(px,py,ww,hh); 
      CEF->Fmod = 1;
   }
   
   else signal_thread();                                                   //  use thread process
   return;
}


//  thread function - use multiple working threads

void * flatten_thread(void *)
{
   void  * flatten_wthread(void *arg);

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      flatten_distribution();
      
      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(flatten_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      CEF->Fmod = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * flatten_wthread(void *arg)                                          //  worker thread function
{
   int         index = *((int *) (arg));
   int         px, py;

   for (py = index; py < E1pxm->hh; py += Nwt)                             //  flatten brightness distribution
   for (px = 0; px < E1pxm->ww; px++)
      flatten_pixel(px,py);

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


//  compute brightness distribution for image or selected area

void flatten_distribution()
{
   int         px, py, ii, npix;
   int         ww = E1pxm->ww, hh = E1pxm->hh;
   float       bright1;
   float       *pix1;

   for (ii = 0; ii < 256; ii++)                                            //  clear brightness distribution data
      flatten_brdist[ii] = 0;

   if (sa_stat == 3)                                                       //  process selected area
   {
      for (ii = npix = 0; ii < ww * hh; ii++)
      {
         if (! sa_pixmap[ii]) continue;
         py = ii / ww;
         px = ii - py * ww;
         pix1 = PXMpix(E1pxm,px,py);
         bright1 = pixbright(pix1);
         flatten_brdist[int(bright1)]++;
         npix++;
      }
      
      for (ii = 1; ii < 256; ii++)                                         //  cumulative brightness distribution
         flatten_brdist[ii] += flatten_brdist[ii-1];                       //   0 ... npix

      for (ii = 0; ii < 256; ii++)
         flatten_brdist[ii] = flatten_brdist[ii]                           //  multiplier per brightness level
                            / npix * 256.0 / (ii + 1);
   }

   else                                                                    //  process whole image
   {
      for (py = 0; py < hh; py++)                                          //  compute brightness distribution
      for (px = 0; px < ww; px++)
      {
         pix1 = PXMpix(E1pxm,px,py);
         bright1 = pixbright(pix1);
         flatten_brdist[int(bright1)]++;
      }
      
      for (ii = 1; ii < 256; ii++)                                         //  cumulative brightness distribution
         flatten_brdist[ii] += flatten_brdist[ii-1];                       //   0 ... (ww * hh)

      for (ii = 0; ii < 256; ii++)
         flatten_brdist[ii] = flatten_brdist[ii]                           //  multiplier per brightness level 
                            / (ww * hh) * 256.0 / (ii + 1);
   }
   
   return;
}


//  compute new brightness for one pixel

void flatten_pixel(int px, int py)
{
   int         ii, dist = 0;
   float       *pix1, *pix3;
   float       fold, fnew, dold, dnew, cmax;
   float       red1, green1, blue1, red3, green3, blue3;
   float       bright1, bright2, deband;

   if (sa_stat == 3) {                                                     //  select area active
      ii = py * E1pxm->ww + px;   
      dist = sa_pixmap[ii];                                                //  distance from edge
      if (! dist) return;                                                  //  outside pixel
   }

   pix1 = PXMpix(E1pxm,px,py);                                             //  input pixel
   pix3 = PXMpix(E3pxm,px,py);                                             //  output pixel
      
   red1 = pix1[0];
   green1 = pix1[1];
   blue1 = pix1[2];

   bright1 = 0.25 * red1 + 0.65 * green1 + 0.10 * blue1;                   //  input pixel brightness
   bright2 = flatten_brdist[int(bright1)];                                 //  output adjustment factor

   red3 = bright2 * red1;                                                  //  fully flattened brightness
   green3 = bright2 * green1;
   blue3 = bright2 * blue1;

   deband = (256.0 - bright1) / 256.0;                                     //  bright1 = 0 to 256  >>  deband = 1 to 0
   deband += (1.0 - deband) * (1.0 - 0.01 * flatten_deband);               //  deband = 1 to 1  >>  1 to 0

   fnew = 0.01 * flatten_value;                                            //  how much to flatten, 0 to 1
   fnew = deband * fnew;                                                   //  attenuate flatten of brighter pixels
   fold = 1.0 - fnew;                                                      //  how much to retain, 1 to 0

   red3 = fnew * red3 + fold * red1;                                       //  blend new and old brightness
   green3 = fnew * green3 + fold * green1;
   blue3 = fnew * blue3 + fold * blue1;

   if (sa_stat == 3 && dist < sa_blend) {                                  //  select area is active,
      dnew = 1.0 * dist / sa_blend;                                        //    blend changes over sa_blend
      dold = 1.0 - dnew;
      red3 = dnew * red3 + dold * red1;
      green3 = dnew * green3 + dold * green1;
      blue3 = dnew * blue3 + dold * blue1;
   }

   cmax = red3;                                                            //  stop overflow, keep color balance
   if (green3 > cmax) cmax = green3;
   if (blue3 > cmax) cmax = blue3;
   if (cmax > 255.9) {
      cmax = 255.9 / cmax;
      red3 = red3 * cmax;
      green3 = green3 * cmax;
      blue3 = blue3 * cmax;
   }

   pix3[0] = red3;
   pix3[1] = green3;
   pix3[2] = blue3;

   return;
}


/**************************************************************************/

//  Clip pixels outside the given low/high range and expand the remaining
//  brightness range back to the full range of 0 to 255.

float       xbrangeD, xbrangeB;
editfunc    EFxbrange;

void m_xbrange(GtkWidget *, cchar *) 
{
   int      xbrange_dialog_event(zdialog *zd, cchar *event);
   void *   xbrange_thread(void *);
   
   cchar  *title = ZTX("Expand Brightness Range");
   cchar  *labD = ZTX("dark pixels");
   cchar  *labB = ZTX("bright pixels");

   zfuncs::F1_help_topic = "expand_brightness";
   
   EFxbrange.funcname = "bright-range";
   EFxbrange.Fprev = 1;                                                    //  use preview
   EFxbrange.Farea = 2;                                                    //  select area usable
   EFxbrange.threadfunc = xbrange_thread;                                  //  thread function
   if (! edit_setup(EFxbrange)) return;                                    //  setup edit

   m_histogram(0,0);                                                       //  show brightness distribution 

   zdialog *zd = zdialog_new(title,mWin,Bdone,Bcancel,null);
   EFxbrange.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"expand");
   zdialog_add_widget(zd,"vbox","vb1","hb1",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"homog|expand");
   zdialog_add_widget(zd,"vbox","vb3","hb1",0,"homog|space=8");
   zdialog_add_widget(zd,"label","labD","vb1",labD);
   zdialog_add_widget(zd,"label","labB","vb1",labB);
   zdialog_add_widget(zd,"hscale","clipD","vb2","0|99|0.5|0");
   zdialog_add_widget(zd,"hscale","clipB","vb2","0|99|0.5|0");
   zdialog_add_widget(zd,"label","darkval","vb3");
   zdialog_add_widget(zd,"label","briteval","vb3");

   zdialog_resize(zd,300,0);
   zdialog_help(zd,"expand_brightness");                                   //  zdialog help topic
   zdialog_run(zd,xbrange_dialog_event,"save");                            //  run dialog - parallel
   
   xbrangeD = xbrangeB = 0;                                                //  initial clip = 0
   return;
}


//  dialog event and completion callback function

int xbrange_dialog_event(zdialog *zd, cchar *event)
{
   char     text[8];

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 1;
   }

   if (strEqu(event,"reset")) {                                            //  reset dialog controls 
      zdialog_stuff(zd,"clipD",0);
      zdialog_stuff(zd,"clipB",0);
   }
   
   if (strEqu(event,"blendwidth")) signal_thread();

   if (strEqu(event,"clipD")) {
      zdialog_fetch(zd,"clipD",xbrangeD);
      signal_thread();
   }

   if (strEqu(event,"clipB")) {
      zdialog_fetch(zd,"clipB",xbrangeB);
      signal_thread();
   }
   
   sprintf(text,"%.0f",xbrangeD);                                          //  numeric feedback
   zdialog_stuff(zd,"darkval",text);
   sprintf(text,"%.0f",xbrangeB);
   zdialog_stuff(zd,"briteval",text);

   return 0;
}


//  thread function

void * xbrange_thread(void *)
{
   int         ii, px, py, dist = 0;
   float       dark, bright, b1, b3, bf, f1, f2;
   float       red1, green1, blue1, red3, green3, blue3;
   float       *pix1, *pix3;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      dark = 0.01 * xbrangeD * 256;                                        //  clipping brightness levels
      bright = (1.0 - 0.01 * xbrangeB) * 256;

      for (py = 0; py < E3pxm->hh; py++)                                   //  loop all image pixels
      for (px = 0; px < E3pxm->ww; px++)
      {
         ii = py * E3pxm->ww + px;

         if (sa_stat == 3) {                                               //  select area active
            dist = sa_pixmap[ii];                                          //  distance from edge
            if (! dist) continue;                                          //  pixel is outside area
         }

         pix1 = PXMpix(E1pxm,px,py);                                       //  input pixel
         pix3 = PXMpix(E3pxm,px,py);                                       //  output pixel

         b1 = pixbright(pix1);

         if (b1 < dark)                                                    //  clip dark pixels
            b3 = 0;
         else if (b1 > bright)                                             //  clip bright pixels
            b3 = bright;
         else b3 = b1;
         
         if (b3 > dark) b3 = b3 - dark;                                    //  expand the rest
         b3 = b3 * (255.0 / (bright - dark));

         bf = b3 / (b1 + 1);                                               //  brightness ratio
         
         red1 = pix1[0];                                                   //  input RGB 
         green1 = pix1[1];
         blue1 = pix1[2];
         
         red3 = bf * red1;                                                 //  output RGB
         green3 = bf * green1;
         blue3 = bf * blue1;
         
         if (red3 > 255) red3 = 255;
         if (green3 > 255) green3 = 255;
         if (blue3 > 255) blue3 = 255;
         
         if (sa_stat == 3 && dist < sa_blend) {                            //  select area is active,
            f1 = 1.0 * dist / sa_blend;                                    //    blend changes over sa_blend
            f2 = 1.0 - f1;
            red3 = f1 * red3 + f2 * red1;
            green3 = f1 * green3 + f2 * green1;
            blue3 = f1 * blue3 + f2 * blue1;
         }
         
         pix3[0] = red3;
         pix3[1] = green3;
         pix3[2] = blue3;
      }

      CEF->Fmod = 1;
      mwpaint2();
   }

   return 0;                                                               //  not executed, stop g++ warning
}


/**************************************************************************/

//  adjust white balance

void   whitebal_mousefunc();

float       whitebal_red, whitebal_green, whitebal_blue;
float       whitebal_reduce;
editfunc    EFwhitebal;

void m_whitebal(GtkWidget *, cchar *)
{
   int    whitebal_dialog_event(zdialog *zd, cchar *event);
   void * whitebal_thread(void *);

   cchar  *wbtitle = ZTX("Adjust White Balance");
   cchar  *wbhelp = ZTX("Click white or gray image location");

   zfuncs::F1_help_topic = "white_balance";
   
   EFwhitebal.funcname = "white-balance";
   EFwhitebal.Fprev = 1;                                                   //  use preview
   EFwhitebal.Farea = 2;                                                   //  select area usable
   EFwhitebal.threadfunc = whitebal_thread;                                //  thread function
   EFwhitebal.mousefunc = whitebal_mousefunc;                              //  mouse function

   if (! edit_setup(EFwhitebal)) return;                                   //  setup edit: preview

   zdialog *zd = zdialog_new(wbtitle,mWin,Bdone,Bcancel,null);
   EFwhitebal.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zd,"label","labwbh","hb1",wbhelp,"space=5");
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labpix","hb2","pixel:","space=3");
   zdialog_add_widget(zd,"label","pixel","hb2","0000 0000");
   zdialog_add_widget(zd,"label","labrgb","hb2","   RGB:");
   zdialog_add_widget(zd,"label","rgb","hb2","000 000 000");
   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labdamp","hb3",ZTX("reduce"),"space=3");
   zdialog_add_widget(zd,"hscale","reduce","hb3","0|1|0.01|0","expand");

   zdialog_help(zd,"white_balance");                                       //  zdialog help topic
   zdialog_run(zd,whitebal_dialog_event,"save");                           //  run dialog - parallel

   whitebal_red = whitebal_green = whitebal_blue = 1.0;
   whitebal_reduce = 0.0;

   takeMouse(whitebal_mousefunc,dragcursor);                               //  connect mouse function
   return;
}


//  dialog event and completion callback function

int whitebal_dialog_event(zdialog *zd, cchar *event)                       //  dialog event function
{
   if (zd->zstat)
   {
      if (zd->zstat == 1) edit_done();                                     //  done
      else edit_cancel();                                                  //  cancel or destroy
      return 0;
   }
   
   if (strEqu(event,"focus"))                                              //  toggle mouse capture
      takeMouse(whitebal_mousefunc,dragcursor);                            //  connect mouse function
   
   if (strEqu(event,"reduce"))                                             //  reduce impact       v.12.04
      zdialog_fetch(zd,"reduce",whitebal_reduce);
   
   if (strEqu(event,"blendwidth")) signal_thread();
   
   signal_thread();
   return 1;
}


//  get nominal white color from mouse click position

void whitebal_mousefunc()                                                  //  mouse function
{
   int         px, py, dx, dy;
   float       red, green, blue, rgbmean;
   char        work[40];
   float       *ppix;
   zdialog     *zd = EFwhitebal.zd;
   
   if (! LMclick) return;
   
   LMclick = 0;
   px = Mxclick;                                                           //  mouse click position
   py = Myclick;
   
   if (px < 2) px = 2;                                                     //  pull back from edge
   if (px > E3pxm->ww-3) px = E3pxm->ww-3;
   if (py < 2) py = 2;
   if (py > E3pxm->hh-3) py = E3pxm->hh-3;
   
   red = green = blue = 0;

   for (dy = -2; dy <= 2; dy++)                                            //  5x5 block around mouse position
   for (dx = -2; dx <= 2; dx++)
   {
      ppix = PXMpix(E1pxm,px+dx,py+dy);                                    //  input image
      red += ppix[0];
      green += ppix[1];
      blue += ppix[2];
   }

   red = red / 25.0;                                                       //  mean RGB levels 
   green = green / 25.0;
   blue = blue / 25.0;
   rgbmean = (red + green + blue) / 3.0;
   
   whitebal_red = rgbmean / red;                                           //  = 1.0 if all RGB are equal
   whitebal_green = rgbmean / green;
   whitebal_blue = rgbmean / blue;

   snprintf(work,40,"%d %d",px,py);
   zdialog_stuff(zd,"pixel",work);

   snprintf(work,40,"%7.3f %7.3f %7.3f",red/256,green/256,blue/256);
   zdialog_stuff(zd,"rgb",work);
   
   signal_thread();                                                        //  trigger image update

   return;
}


//  Update image based on neutral pixel that was clicked

void * whitebal_thread(void *)
{
   void * whitebal_wthread(void *arg);

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(whitebal_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      CEF->Fmod = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * whitebal_wthread(void *arg)                                         //  worker thread function
{
   int         index = *((int *) arg);
   int         px, py, ii, dist = 0;
   float       *pix1, *pix3;
   float       red1, green1, blue1;
   float       red3, green3, blue3;
   float       brmax, dold, dnew;
   float       Fred, Fgreen, Fblue;

   Fred = whitebal_red + (1.0 - whitebal_red) * whitebal_reduce;           //  as reduce goes from 0 to 1.0.
   Fgreen = whitebal_green + (1.0 - whitebal_green) * whitebal_reduce;     //    Fcolors go from initial to 1.0
   Fblue = whitebal_blue + (1.0 - whitebal_blue) * whitebal_reduce;        //  12.04

   for (py = index; py < E1pxm->hh; py += Nwt)
   for (px = 0; px < E1pxm->ww; px++)
   {
      if (sa_stat == 3) {                                                  //  select area active
         ii = py * E1pxm->ww + px;
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel outside area
      }

      pix1 = PXMpix(E1pxm,px,py);                                          //  input pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel
      
      red1 = pix1[0];
      green1 = pix1[1];
      blue1 = pix1[2];
      
      red3 = Fred * red1;                                                  //  change color ratios
      green3 = Fgreen * green1;
      blue3 = Fblue * blue1;

      if (sa_stat == 3 && dist < sa_blend) {                               //  select area is active,
         dnew = 1.0 * dist / sa_blend;                                     //    blend changes over sa_blend
         dold = 1.0 - dnew;
         red3 = dnew * red3 + dold * red1;
         green3 = dnew * green3 + dold * green1;
         blue3 = dnew * blue3 + dold * blue1;
      }
      
      brmax = red3;                                                        //  brmax = brightest color
      if (green3 > brmax) brmax = green3;
      if (blue3 > brmax) brmax = blue3;
      
      if (brmax > 255.9) {                                                 //  if overflow, reduce
         brmax = 255.9 / brmax;
         red3 = red3 * brmax;
         green3 = green3 * brmax;
         blue3 = blue3 * brmax;
      }

      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }
   
   exit_wthread();
   return 0;                                                               //  not executed, stop gcc warning
}


/**************************************************************************/

//  red eye removal function

struct sredmem {                                                           //  red-eye struct in memory
   char        type, space[3];
   int         cx, cy, ww, hh, rad, clicks;
   float       thresh, tstep;
};
sredmem  redmem[100];                                                      //  store up to 100 red-eyes

int      Nredmem = 0, maxredmem = 100;
void     redeye_mousefunc();

editfunc    EFredeye;


void m_redeye(GtkWidget *, cchar *)
{
   int      redeye_dialog_event(zdialog *zd, cchar *event);

   cchar    *redeye_message = ZTX(
               "Method 1:\n"
               "  Left-click on red-eye to darken.\n"
               "Method 2:\n"
               "  Drag down and right to enclose red-eye.\n"
               "  Left-click on red-eye to darken.\n"
               "Undo red-eye:\n"
               "  Right-click on red-eye.");

   zfuncs::F1_help_topic = "red_eye";

   EFredeye.funcname = "redeye";
   EFredeye.Farea = 1;                                                     //  select area ignored
   EFredeye.mousefunc = redeye_mousefunc;
   if (! edit_setup(EFredeye)) return;                                     //  setup edit

   zdialog *zd = zdialog_new(ZTX("Red Eye Reduction"),mWin,Bdone,Bcancel,null);
   EFredeye.zd = zd;

   zdialog_add_widget(zd,"label","lab1","dialog",redeye_message);
   zdialog_help(zd,"red_eye");                                             //  zdialog help topic
   zdialog_run(zd,redeye_dialog_event,"save");                             //  run dialog - parallel

   Nredmem = 0;
   takeMouse(redeye_mousefunc,dragcursor);                                 //  connect mouse function
   
   return;
}


//  dialog event and completion callback function

int redeye_dialog_event(zdialog *zd, cchar *event)
{
   if (zd->zstat) {
      if (Nredmem > 0) CEF->Fmod = 1;
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 0;
   }

   if (strEqu(event,"focus"))                                              //  toggle mouse capture
      takeMouse(redeye_mousefunc,dragcursor);                              //  connect mouse function

   return 0;
}


//  mouse functions to define, darken, and undo red-eyes

int      redeye_createF(int px, int py);                                   //  create 1-click red-eye (type F)
int      redeye_createR(int px, int py, int ww, int hh);                   //  create robust red-eye (type R)
void     redeye_darken(int ii);                                            //  darken red-eye
void     redeye_distr(int ii);                                             //  build pixel redness distribution
int      redeye_find(int px, int py);                                      //  find red-eye at mouse position
void     redeye_remove(int ii);                                            //  remove red-eye at mouse position
int      redeye_radlim(int cx, int cy);                                    //  compute red-eye radius limit


void redeye_mousefunc()
{
   int         ii, px, py, ww, hh;

   if (Nredmem == maxredmem) {
      zmessageACK(mWin,"%d red-eye limit reached",maxredmem);              //  too many red-eyes
      return;
   }

   if (LMclick)                                                            //  left mouse click
   {
      px = Mxclick;                                                        //  click position
      py = Myclick;
      if (px < 0 || px > E3pxm->ww-1 || py < 0 || py > E3pxm->hh-1) return;     //  outside image area

      ii = redeye_find(px,py);                                             //  find existing red-eye
      if (ii < 0) ii = redeye_createF(px,py);                              //  or create new type F
      redeye_darken(ii);                                                   //  darken red-eye
      mwpaint2();
   }
   
   if (RMclick)                                                            //  right mouse click
   {
      px = Mxclick;                                                        //  click position
      py = Myclick;
      ii = redeye_find(px,py);                                             //  find red-eye
      if (ii >= 0) redeye_remove(ii);                                      //  if found, remove
      mwpaint2();
   }

   LMclick = RMclick = 0;                                                  //  v.12.11

   if (Mxdrag || Mydrag)                                                   //  mouse drag underway
   {
      px = Mxdown;                                                         //  initial position
      py = Mydown;
      ww = Mxdrag - Mxdown;                                                //  increment
      hh = Mydrag - Mydown;
      Mxdrag = Mydrag = 0;
      if (ww < 2 && hh < 2) return;
      if (ww < 2) ww = 2;
      if (hh < 2) hh = 2;
      if (px < 1) px = 1;                                                  //  keep within image area
      if (py < 1) py = 1;      
      if (px + ww > E3pxm->ww-1) ww = E3pxm->ww-1 - px;
      if (py + hh > E3pxm->hh-1) hh = E3pxm->hh-1 - py;
      ii = redeye_find(px,py);                                             //  find existing red-eye
      if (ii >= 0) redeye_remove(ii);                                      //  remove it
      ii = redeye_createR(px,py,ww,hh);                                    //  create new red-eye type R
   }

   return;
}


//  create type F redeye (1-click automatic)

int redeye_createF(int cx, int cy)
{
   int         cx0, cy0, cx1, cy1, px, py, rad, radlim;
   int         loops, ii;
   int         Tnpix, Rnpix, R2npix;
   float       rd, rcx, rcy, redpart;
   float       Tsum, Rsum, R2sum, Tavg, Ravg, R2avg;
   float       sumx, sumy, sumr;
   float       *ppix;
   
   cx0 = cx;
   cy0 = cy;
   
   for (loops = 0; loops < 8; loops++)
   {
      cx1 = cx;
      cy1 = cy;

      radlim = redeye_radlim(cx,cy);                                       //  radius limit (image edge)
      Tsum = Tavg = Ravg = Tnpix = 0;

      for (rad = 0; rad < radlim-2; rad++)                                 //  find red-eye radius from (cx,cy)
      {
         Rsum = Rnpix = 0;
         R2sum = R2npix = 0;

         for (py = cy-rad-2; py <= cy+rad+2; py++)
         for (px = cx-rad-2; px <= cx+rad+2; px++)
         {
            rd = sqrt((px-cx)*(px-cx) + (py-cy)*(py-cy));
            ppix = PXMpix(E3pxm,px,py);
            redpart = pixred(ppix);

            if (rd <= rad + 0.5 && rd > rad - 0.5) {                       //  accum. redness at rad
               Rsum += redpart;
               Rnpix++;
            }
            else if (rd <= rad + 2.5 && rd > rad + 1.5) {                  //  accum. redness at rad+2
               R2sum += redpart;
               R2npix++;
            }
         }
         
         Tsum += Rsum;
         Tnpix += Rnpix;
         Tavg = Tsum / Tnpix;                                              //  avg. redness over 0-rad
         Ravg = Rsum / Rnpix;                                              //  avg. redness at rad
         R2avg = R2sum / R2npix;                                           //  avg. redness at rad+2
         if (R2avg > Ravg || Ravg > Tavg) continue;
         if ((Ravg - R2avg) < 0.2 * (Tavg - Ravg)) break;                  //  0.1 --> 0.2
      }
      
      sumx = sumy = sumr = 0;
      rad = int(1.2 * rad + 1);
      if (rad > radlim) rad = radlim;
      
      for (py = cy-rad; py <= cy+rad; py++)                                //  compute center of gravity for
      for (px = cx-rad; px <= cx+rad; px++)                                //   pixels within rad of (cx,cy)
      {
         rd = sqrt((px-cx)*(px-cx) + (py-cy)*(py-cy));
         if (rd > rad + 0.5) continue;
         ppix = PXMpix(E3pxm,px,py);
         redpart = pixred(ppix);                                           //  weight by redness
         sumx += redpart * (px - cx);
         sumy += redpart * (py - cy);
         sumr += redpart;
      }

      rcx = cx + 1.0 * sumx / sumr;                                        //  new center of red-eye
      rcy = cy + 1.0 * sumy / sumr;
      if (fabsf(cx0 - rcx) > 0.6 * rad) break;                             //  give up if big movement
      if (fabsf(cy0 - rcy) > 0.6 * rad) break;
      cx = int(rcx + 0.5);
      cy = int(rcy + 0.5);
      if (cx == cx1 && cy == cy1) break;                                   //  done if no change
   }

   radlim = redeye_radlim(cx,cy);
   if (rad > radlim) rad = radlim;

   ii = Nredmem++;                                                         //  add red-eye to memory
   redmem[ii].type = 'F';
   redmem[ii].cx = cx;
   redmem[ii].cy = cy;
   redmem[ii].rad = rad;
   redmem[ii].clicks = 0;
   redmem[ii].thresh = 0;
   return ii;
}


//  create type R red-eye (drag an ellipse over red-eye area)

int redeye_createR(int cx, int cy, int ww, int hh)
{
   int      rad, radlim;

   draw_mousearc(cx,cy,2*ww,2*hh);                                         //  draw ellipse around mouse pointer

   if (ww > hh) rad = ww;
   else rad = hh;
   radlim = redeye_radlim(cx,cy);
   if (rad > radlim) rad = radlim;

   int ii = Nredmem++;                                                     //  add red-eye to memory
   redmem[ii].type = 'R';
   redmem[ii].cx = cx;
   redmem[ii].cy = cy;
   redmem[ii].ww = 2 * ww;
   redmem[ii].hh = 2 * hh;
   redmem[ii].rad = rad;
   redmem[ii].clicks = 0;
   redmem[ii].thresh = 0;
   return ii;
}


//  darken a red-eye and increase click count

void redeye_darken(int ii)
{
   int         cx, cy, ww, hh, px, py, rad, clicks;
   float       rd, thresh, tstep;
   char        type;
   float       *ppix;

   type = redmem[ii].type;
   cx = redmem[ii].cx;
   cy = redmem[ii].cy;
   ww = redmem[ii].ww;
   hh = redmem[ii].hh;
   rad = redmem[ii].rad;
   thresh = redmem[ii].thresh;
   tstep = redmem[ii].tstep;
   clicks = redmem[ii].clicks++;
   
   if (thresh == 0)                                                        //  1st click 
   {
      redeye_distr(ii);                                                    //  get pixel redness distribution
      thresh = redmem[ii].thresh;                                          //  initial redness threshhold
      tstep = redmem[ii].tstep;                                            //  redness step size
      draw_mousearc(-1,0,0,0);                                             //  erase mouse ellipse
   }

   tstep = (thresh - tstep) / thresh;                                      //  convert to reduction factor
   thresh = thresh * pow(tstep,clicks);                                    //  reduce threshhold by total clicks

   for (py = cy-rad; py <= cy+rad; py++)                                   //  darken pixels over threshhold
   for (px = cx-rad; px <= cx+rad; px++)
   {
      if (type == 'R') {
         if (px < cx - ww/2) continue;
         if (px > cx + ww/2) continue;
         if (py < cy - hh/2) continue;
         if (py > cy + hh/2) continue;
      }
      rd = sqrt((px-cx)*(px-cx) + (py-cy)*(py-cy));
      if (rd > rad + 0.5) continue;
      ppix = PXMpix(E3pxm,px,py);                                          //  set redness = threshhold
      if (pixred(ppix) > thresh)
         ppix[0] = int(thresh * (0.65 * ppix[1] + 0.10 * ppix[2] + 1) / (25 - 0.25 * thresh));
   }

   return;
}


//  Build a distribution of redness for a red-eye. Use this information 
//  to set initial threshhold and step size for stepwise darkening.

void redeye_distr(int ii)
{
   int         cx, cy, ww, hh, rad, px, py;
   int         bin, npix, dbins[20], bsum, blim;
   float       rd, maxred, minred, redpart, dbase, dstep;
   char        type;
   float       *ppix;
   
   type = redmem[ii].type;
   cx = redmem[ii].cx;
   cy = redmem[ii].cy;
   ww = redmem[ii].ww;
   hh = redmem[ii].hh;
   rad = redmem[ii].rad;
   
   maxred = 0;
   minred = 100;

   for (py = cy-rad; py <= cy+rad; py++)
   for (px = cx-rad; px <= cx+rad; px++)
   {
      if (type == 'R') {
         if (px < cx - ww/2) continue;
         if (px > cx + ww/2) continue;
         if (py < cy - hh/2) continue;
         if (py > cy + hh/2) continue;
      }
      rd = sqrt((px-cx)*(px-cx) + (py-cy)*(py-cy));
      if (rd > rad + 0.5) continue;
      ppix = PXMpix(E3pxm,px,py);
      redpart = pixred(ppix);
      if (redpart > maxred) maxred = redpart;
      if (redpart < minred) minred = redpart;
   }
   
   dbase = minred;
   dstep = (maxred - minred) / 19.99;

   for (bin = 0; bin < 20; bin++) dbins[bin] = 0;
   npix = 0;

   for (py = cy-rad; py <= cy+rad; py++)
   for (px = cx-rad; px <= cx+rad; px++)
   {
      if (type == 'R') {
         if (px < cx - ww/2) continue;
         if (px > cx + ww/2) continue;
         if (py < cy - hh/2) continue;
         if (py > cy + hh/2) continue;
      }
      rd = sqrt((px-cx)*(px-cx) + (py-cy)*(py-cy));
      if (rd > rad + 0.5) continue;
      ppix = PXMpix(E3pxm,px,py);
      redpart = pixred(ppix);
      bin = int((redpart - dbase) / dstep);
      ++dbins[bin];
      ++npix;
   }
   
   bsum = 0;
   blim = int(0.5 * npix);

   for (bin = 0; bin < 20; bin++)                                          //  find redness level for 50% of
   {                                                                       //    pixels within red-eye radius
      bsum += dbins[bin];
      if (bsum > blim) break;
   }

   redmem[ii].thresh = dbase + dstep * bin;                                //  initial redness threshhold
   redmem[ii].tstep = dstep;                                               //  redness step (5% of range)

   return;
}


//  find a red-eye (nearly) overlapping the mouse click position

int redeye_find(int cx, int cy)
{
   for (int ii = 0; ii < Nredmem; ii++)
   {
      if (cx > redmem[ii].cx - 2 * redmem[ii].rad && 
          cx < redmem[ii].cx + 2 * redmem[ii].rad &&
          cy > redmem[ii].cy - 2 * redmem[ii].rad && 
          cy < redmem[ii].cy + 2 * redmem[ii].rad) 
            return ii;                                                     //  found
   }
   return -1;                                                              //  not found
}


//  remove a red-eye from memory

void redeye_remove(int ii)
{
   int      cx, cy, rad, px, py;
   float    *pix1, *pix3;

   cx = redmem[ii].cx;
   cy = redmem[ii].cy;
   rad = redmem[ii].rad;

   for (px = cx-rad; px <= cx+rad; px++)
   for (py = cy-rad; py <= cy+rad; py++)
   {
      pix1 = PXMpix(E1pxm,px,py);
      pix3 = PXMpix(E3pxm,px,py);
      pix3[0] = pix1[0];
      pix3[1] = pix1[1];
      pix3[2] = pix1[2];
   }
   
   for (ii++; ii < Nredmem; ii++) 
      redmem[ii-1] = redmem[ii];
   Nredmem--;
   
   draw_mousearc(-1,0,0,0);                                                //  erase mouse ellipse
   return;
}


//  compute red-eye radius limit: smaller of 100 and nearest image edge

int redeye_radlim(int cx, int cy)
{
   int radlim = 100;
   if (cx < 100) radlim = cx;
   if (E3pxm->ww-1 - cx < 100) radlim = E3pxm->ww-1 - cx;
   if (cy < 100) radlim = cy;
   if (E3pxm->hh-1 - cy < 100) radlim = E3pxm->hh-1 - cy;
   return radlim;
}


/**************************************************************************/

//  image blur function                                                    //  radius steps of 0.5

float       blur_radius;
float       blur_weight[101][101];                                         //  up to blur radius = 99
int         blur_Npixels, blur_pixdone;
int         blur_cancel;

editfunc    EFblur;


void m_blur(GtkWidget *, cchar *)
{
   int    blur_dialog_event(zdialog *zd, cchar *event);
   void * blur_thread(void *);

   zfuncs::F1_help_topic = "blur";

   EFblur.funcname = "blur";
   EFblur.Farea = 2;                                                       //  select area usable
   EFblur.threadfunc = blur_thread;                                        //  thread function
   if (! edit_setup(EFblur)) return;                                       //  setup edit

   blur_radius = 0.5;
   blur_cancel = 0;

   zdialog *zd = zdialog_new(ZTX("Set Blur Radius"),mWin,Bdone,Bcancel,null);
   EFblur.zd = zd;

   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=10");
   zdialog_add_widget(zd,"label","labrad","hb2",Bradius,"space=5");
   zdialog_add_widget(zd,"spin","radius","hb2","0|99|0.5|0.5","space=5");
   zdialog_add_widget(zd,"button","apply","hb2",Bapply,"space=5");

   zdialog_help(zd,"blur");                                                //  zdialog help topic 
   zdialog_run(zd,blur_dialog_event,"save");                               //  run dialog - parallel

   return;
}


//  dialog event and completion callback function

int blur_dialog_event(zdialog * zd, cchar *event)
{
   if (zd->zstat)
   {
      if (zd->zstat == 1) edit_done();                                     //  done
      else {
         blur_cancel = 1;
         edit_cancel();                                                    //  cancel or destroy
      }
      return 1;
   }

   if (strEqu(event,"blendwidth")) signal_thread();

   if (strEqu(event,"apply")) {
      zdialog_fetch(zd,"radius",blur_radius);                              //  get blur radius
      if (blur_radius == 0) edit_reset();
      else signal_thread();                                                //  trigger working thread
   }

   return 1;
}


//  image blur thread function

void * blur_thread(void *)
{
   void * blur_wthread(void *arg);

   int         dx, dy;
   float       rad, rad2;
   float       m, d, w, sum;
  
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      rad = blur_radius - 0.2;
      rad2 = rad * rad;

      for (dx = 0; dx <= rad+1; dx++)                                      //  clear weights array
      for (dy = 0; dy <= rad+1; dy++)
         blur_weight[dx][dy] = 0;

      for (dx = -rad-1; dx <= rad+1; dx++)                                 //  blur_weight[dx][dy] = no. of pixels
      for (dy = -rad-1; dy <= rad+1; dy++)                                 //    at distance (dx,dy) from center
         ++blur_weight[abs(dx)][abs(dy)];

      m = sqrt(rad2 + rad2);                                               //  corner pixel distance from center
      sum = 0;

      for (dx = 0; dx <= rad+1; dx++)                                      //  compute weight of pixel
      for (dy = 0; dy <= rad+1; dy++)                                      //    at distance dx, dy
      {
         d = sqrt(dx*dx + dy*dy);
         w = (m + 1.2 - d) / m;
         w = w * w;
         sum += blur_weight[dx][dy] * w;
         blur_weight[dx][dy] = w;
      }

      for (dx = 0; dx <= rad+1; dx++)                                      //  make weights add up to 1.0
      for (dy = 0; dy <= rad+1; dy++)
         blur_weight[dx][dy] = blur_weight[dx][dy] / sum;
      
      if (sa_stat == 3) Fbusy_goal = sa_Npixel;
      else  Fbusy_goal = E3pxm->ww * E3pxm->hh;
      Fbusy_done = 0;

      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(blur_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      Fbusy_goal = 0;
      CEF->Fmod = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * blur_wthread(void *arg)                                             //  worker thread function
{
   int      index = *((int *) arg);
   int      px, py, ii, dist = 0;
   int      jj, dx, dy, adx, ady, rad;
   float    red, green, blue;
   float    weight1, weight2, f1, f2;
   float    *pix1, *pix3, *pixN;
   
   for (py = index; py < E3pxm->hh-1; py += Nwt)                           //  loop all image pixels
   for (px = 0; px < E3pxm->ww-1; px++)
   {
      if (blur_cancel) exit_wthread();

      if (sa_stat == 3) {                                                  //  select area active
         ii = py * E1pxm->ww + px;
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  outside pixel
      }

      pix1 = PXMpix(E1pxm,px,py);                                          //  source pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  target pixel
      
      rad = blur_radius;
      red = green = blue = 0;
      weight2 = 0.0;
      
      if (sa_stat == 3)                                                    //  select area active
      {
         for (dy = -rad-1; dy <= rad+1; dy++)                              //  loop neighbor pixels within radius
         for (dx = -rad-1; dx <= rad+1; dx++)
         {
            if (px+dx < 0 || px+dx > E3pxm->ww-1) continue;                //  omit pixels off edge
            if (py+dy < 0 || py+dy > E3pxm->hh-1) continue;
            jj = (py+dy) * E3pxm->ww + (px+dx);
            if (! sa_pixmap[jj]) continue;                                 //  omit pixels outside area
            adx = abs(dx);
            ady = abs(dy);
            pixN = pix1 + (dy * E3pxm->ww + dx) * 3;
            weight1 = blur_weight[adx][ady];                               //  weight at distance (dx,dy)
            weight2 += weight1;
            red += pixN[0] * weight1;                                      //  accumulate contributions
            green += pixN[1] * weight1;
            blue += pixN[2] * weight1;
         }
         
         red = red / weight2;                                              //  weighted average
         green = green / weight2;
         blue = blue / weight2;

         if (dist < sa_blend) {                                            //  select area is active,
            f1 = 1.0 * dist / sa_blend;                                    //    blend changes over sa_blend
            f2 = 1.0 - f1;
            red = f1 * red + f2 * pix1[0];
            green = f1 * green + f2 * pix1[1];
            blue = f1 * blue + f2 * pix1[2];
         }

         pix3[0] = int(red);
         pix3[1] = int(green);
         pix3[2] = int(blue);
      }

      else
      {
         for (dy = -rad-1; dy <= rad+1; dy++)                              //  loop neighbor pixels within radius
         for (dx = -rad-1; dx <= rad+1; dx++)
         {
            if (px+dx < 0 || px+dx > E3pxm->ww-1) continue;                //  omit pixels off edge
            if (py+dy < 0 || py+dy > E3pxm->hh-1) continue;
            adx = abs(dx);
            ady = abs(dy);
            pixN = pix1 + (dy * E3pxm->ww + dx) * 3;
            weight1 = blur_weight[adx][ady];                               //  weight at distance (dx,dy)
            weight2 += weight1;
            red += pixN[0] * weight1;                                      //  accumulate contributions
            green += pixN[1] * weight1;
            blue += pixN[2] * weight1;
         }

         red = red / weight2;                                              //  weighted average
         green = green / weight2;
         blue = blue / weight2;

         pix3[0] = red;
         pix3[1] = green;
         pix3[2] = blue;
      }
      
      Fbusy_done++;                                                        //  track progress
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************/

//  image sharpening function

int      sharp_ED_cycles;
int      sharp_ED_reduce;
int      sharp_ED_thresh;
int      sharp_UM_radius;
int      sharp_UM_amount;
int      sharp_UM_thresh;
int      sharp_GR_amount;
int      sharp_GR_thresh;
char     sharp_function[4];

editfunc    EFsharp;


void m_sharpen(GtkWidget *, cchar *)
{
   int    sharp_dialog_event(zdialog *zd, cchar *event);
   void * sharp_thread(void *);

   zfuncs::F1_help_topic = "sharpen";

   EFsharp.funcname = "sharpen";
   EFsharp.Farea = 2;                                                      //  select area usable
   EFsharp.threadfunc = sharp_thread;                                      //  thread function
   if (! edit_setup(EFsharp)) return;                                      //  setup edit

   zdialog *zd = zdialog_new(ZTX("Sharpen Image"),mWin,Bdone,Bcancel,null); 
   EFsharp.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"vbox","vb11","hb1",0,"space=5");
   zdialog_add_widget(zd,"vbox","vb12","hb1",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb13","hb1",0,"homog|space=5");
   zdialog_add_widget(zd,"button","ED","vb11",ZTX("edge detection"),"space=5");
   zdialog_add_widget(zd,"label","lab11","vb12",ZTX("cycles"));
   zdialog_add_widget(zd,"label","lab12","vb12",ZTX("reduce"));
   zdialog_add_widget(zd,"label","lab13","vb12",Bthresh);
   zdialog_add_widget(zd,"spin","cyclesED","vb13","1|30|1|10");
   zdialog_add_widget(zd,"spin","reduceED","vb13","50|95|1|80");
   zdialog_add_widget(zd,"spin","threshED","vb13","1|99|1|1");

   zdialog_add_widget(zd,"hsep","sep2","dialog");
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zd,"vbox","vb21","hb2",0,"space=5");
   zdialog_add_widget(zd,"vbox","vb22","hb2",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb23","hb2",0,"homog|space=5");
   zdialog_add_widget(zd,"button","UM","vb21",ZTX("unsharp mask"),"space=5");
   zdialog_add_widget(zd,"label","lab21","vb22",Bradius);
   zdialog_add_widget(zd,"label","lab22","vb22",Bamount);
   zdialog_add_widget(zd,"label","lab23","vb22",Bthresh);
   zdialog_add_widget(zd,"spin","radiusUM","vb23","1|20|1|2");
   zdialog_add_widget(zd,"spin","amountUM","vb23","0|200|1|100");
   zdialog_add_widget(zd,"spin","threshUM","vb23","0|100|1|0");

   zdialog_add_widget(zd,"hsep","sep3","dialog");
   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zd,"vbox","vb31","hb3",0,"space=5");
   zdialog_add_widget(zd,"vbox","vb32","hb3",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb33","hb3",0,"homog|space=5");
   zdialog_add_widget(zd,"button","GR","vb31",ZTX("brightness gradient"),"space=5");
   zdialog_add_widget(zd,"label","lab32","vb32",Bamount);
   zdialog_add_widget(zd,"label","lab33","vb32",Bthresh);
   zdialog_add_widget(zd,"spin","amountGR","vb33","0|400|1|100");
   zdialog_add_widget(zd,"spin","threshGR","vb33","0|100|1|0");

   zdialog_help(zd,"sharpen");                                             //  zdialog help topic
   zdialog_run(zd,sharp_dialog_event,"save");                              //  run dialog - parallel

   *sharp_function = 0;
   return;
}


//  dialog event and completion callback function

int sharp_dialog_event(zdialog *zd, cchar *event)
{
   if (zd->zstat)
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 0;
   }

   if (strEqu(event,"blendwidth")) signal_thread();

   if (strcmpv(event,"ED","UM","GR",null))
   {
      edit_reset();                                                        //  restore original image

      zdialog_fetch(zd,"cyclesED",sharp_ED_cycles);                        //  get all input values
      zdialog_fetch(zd,"reduceED",sharp_ED_reduce);
      zdialog_fetch(zd,"threshED",sharp_ED_thresh);
      zdialog_fetch(zd,"radiusUM",sharp_UM_radius);
      zdialog_fetch(zd,"amountUM",sharp_UM_amount);
      zdialog_fetch(zd,"threshUM",sharp_UM_thresh);
      zdialog_fetch(zd,"amountGR",sharp_GR_amount);
      zdialog_fetch(zd,"threshGR",sharp_GR_thresh);

      strcpy(sharp_function,event);                                        //  pass to working thread
      signal_thread();
   }

   return 0;
}


//  sharpen image thread function

void * sharp_thread(void *)
{
   int    sharp_ED();
   int    sharp_UM();
   int    sharp_GR();
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      Ffuncbusy++;

      if (strEqu(sharp_function,"ED")) sharp_ED();                         //  do requested function
      if (strEqu(sharp_function,"UM")) sharp_UM();
      if (strEqu(sharp_function,"GR")) sharp_GR();

      Ffuncbusy--;

      CEF->Fmod = 1;
      mwpaint2(); 
   }

   return 0;                                                               //  not executed, stop g++ warning
}


//  image sharpen function by edge detection and compression

int sharp_ED()
{
   void  sharp_pixel_ED(int px, int py, int thresh);

   int      sharp_thresh1 = 100;                                           //  initial threshold
   float    sharp_thresh2 = 0.01 * sharp_ED_reduce;                        //  decline rate
   int      px, py, thresh, cycles;
   
   thresh = sharp_thresh1;
    
   if (sa_stat == 3) Fbusy_goal = sa_Npixel;
   else  Fbusy_goal = E3pxm->ww * E3pxm->hh;
   Fbusy_goal *= sharp_ED_cycles;
   Fbusy_done = 0;

   for (cycles = 0; cycles < sharp_ED_cycles; cycles++)
   {
      if (cycles > 0) thresh = int(thresh * sharp_thresh2);

      for (py = 2; py < E3pxm->hh-2; py++)
      for (px = 2; px < E3pxm->ww-2; px++)                                 //  loop all pixels
      {
         sharp_pixel_ED(px,py,thresh);
         Fbusy_done++;                                                     //  track progress
      }
   }

   Fbusy_goal = 0;
   return 1;
}


void sharp_pixel_ED(int px, int py, int thresh)
{
   int      ii, dist = 0, dd, rgb;
   int      dx[4] = { -1, 0, 1, 1 };                                       //  4 directions: NW N NE E
   int      dy[4] = { -1, -1, -1, 0 };
   float    *pix1, *pix1u, *pix1d;
   float    *pix3, *pix3u, *pix3d, *pix3uu, *pix3dd;
   float    pv2, pv2u, pv2d, pv2uu, pv2dd, pvdiff;
   float    pthresh, f1, f2;
   
   if (sa_stat == 3) {                                                     //  select area active
      ii = py * E1pxm->ww + px;
      dist = sa_pixmap[ii];                                                //  distance from edge
      if (! dist) return;                                                  //  outside pixel
   }

   pthresh = sharp_ED_thresh;                                              //  pthresh = larger
   if (thresh > pthresh) pthresh = thresh;

   pix1 = PXMpix(E1pxm,px,py);                                             //  input pixel
   pix3 = PXMpix(E3pxm,px,py);                                             //  output pixel

   for (dd = 0; dd < 4; dd++)                                              //  4 directions
   {
      pix3u = pix3 + (dy[dd] * E3pxm->ww + dx[dd]) * 3;                    //  upstream pixel
      pix3d = pix3 - (dy[dd] * E3pxm->ww - dx[dd]) * 3;                    //  downstream pixel

      for (rgb = 0; rgb < 3; rgb++)                                        //  loop 3 RGB colors
      {
         pv2 = pix3[rgb];
         pv2u = pix3u[rgb];                                                //  brightness difference
         pv2d = pix3d[rgb];                                                //    across target pixel

         pvdiff = pv2d - pv2u;
         if (pvdiff < 0) pvdiff = -pvdiff;
         if (pvdiff < pthresh) continue;                                   //  brightness slope < threshold

         if (pv2u < pv2 && pv2 < pv2d)                                     //  slope up, monotone
         {
            pix3uu = pix3u + (dy[dd] * E3pxm->ww + dx[dd]) * 3;            //  upstream of upstream pixel
            pix3dd = pix3d - (dy[dd] * E3pxm->ww - dx[dd]) * 3;            //  downstream of downstream
            pv2uu = pix3uu[rgb];
            pv2dd = pix3dd[rgb];

            if (pv2uu >= pv2u) {                                           //  shift focus of changes to
               pix3u = pix3;                                               //    avoid up/down/up jaggies
               pv2u = pv2;
            }
            
            if (pv2dd <= pv2d) {
               pix3d = pix3;
               pv2d = pv2;
            }
               
            if (pv2u > 1) pv2u--;
            if (pv2d < 255) pv2d++;
         }
         
         else if (pv2u > pv2 && pv2 > pv2d)                                //  slope down, monotone
         {
            pix3uu = pix3u + (dy[dd] * E3pxm->ww + dx[dd]) * 3;
            pix3dd = pix3d - (dy[dd] * E3pxm->ww - dx[dd]) * 3;
            pv2uu = pix3uu[rgb];
            pv2dd = pix3dd[rgb];

            if (pv2uu <= pv2u) {
               pix3u = pix3;
               pv2u = pv2;
            }
            
            if (pv2dd >= pv2d) {
               pix3d = pix3;
               pv2d = pv2;
            }

            if (pv2d > 1) pv2d--;
            if (pv2u < 255) pv2u++;
         }

         else continue;                                                    //  slope too small

         if (sa_stat == 3 && dist < sa_blend) {                            //  select area is active,
            f1 = 1.0 * dist / sa_blend;                                    //    blend changes over sa_blend
            f2 = 1.0 - f1;
            pix1u = pix1 + (dy[dd] * E1pxm->ww + dx[dd]) * 3;              //  upstream input pixel
            pix1d = pix1 - (dy[dd] * E1pxm->ww - dx[dd]) * 3;              //  downstream input pixel
            pv2u = int(f1 * pv2u + f2 * pix1u[rgb]);
            pv2d = int(f1 * pv2d + f2 * pix1d[rgb]);
         }

         pix3u[rgb] = pv2u;                                                //  modified brightness values
         pix3d[rgb] = pv2d;                                                //    >> image3 pixel
      }
   }

   return;
}


//  image sharpen function using unsharp mask

float       *brhood_brightness;                                            //  neighborhood brightness per pixel

int sharp_UM()
{
   void  brhood_calc(int radius, char method);                             //  compute neighborhood brightness
   void * sharp_UM_wthread(void *arg);

   int      ii, cc;
   
   cc = E1pxm->ww * E1pxm->hh * sizeof(float);
   brhood_brightness = (float *) zmalloc(cc,"brhood");

   brhood_calc(sharp_UM_radius,'f');

   if (sa_stat == 3) Fbusy_goal = sa_Npixel;
   else  Fbusy_goal = E3pxm->ww * E3pxm->hh;
   Fbusy_done = 0;

   for (ii = 0; ii < Nwt; ii++)                                            //  start worker threads
      start_wthread(sharp_UM_wthread,&wtnx[ii]);
   wait_wthreads();                                                        //  wait for completion

   Fbusy_goal = 0;
   zfree(brhood_brightness);
   return 1;
}


void * sharp_UM_wthread(void *arg)                                         //  worker thread function
{
   void  sharp_pixel_UM(int px, int py, int index);

   int      index = *((int *) arg);
   int      px, py;
   
   for (py = index; py < E3pxm->hh; py += Nwt)                             //  loop all image3 pixels
   for (px = 0; px < E3pxm->ww; px++)
      sharp_pixel_UM(px,py,index);

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


void sharp_pixel_UM(int px, int py, int index)                             //  process one pixel
{
   int         ii, dist = 0;
   float       amount, thresh, bright;
   float       mean, incr, ratio, f1, f2;
   float       red1, green1, blue1, red3, green3, blue3;
   float       *pix1, *pix3;
   
   if (sa_stat == 3) {                                                     //  select area active
      ii = py * E1pxm->ww + px;
      dist = sa_pixmap[ii];                                                //  distance from edge
      if (! dist) return;                                                  //  outside pixel
   }

   amount = 0.01 * sharp_UM_amount;                                        //  0.0 to 2.0
   thresh = 0.4 * sharp_UM_thresh;                                         //  0 to 40 (256 max. possible)

   pix1 = PXMpix(E1pxm,px,py);                                             //  input pixel
   pix3 = PXMpix(E3pxm,px,py);                                             //  output pixel

   bright = pixbright(pix1);
   if (bright < 1) return;                                                 //  effectively black
   ii = py * E1pxm->ww + px;
   mean = brhood_brightness[ii];

   incr = (bright - mean);
   if (fabsf(incr) < thresh) return;                                       //  omit low-contrast pixels

   incr = incr * amount;                                                   //  0.0 to 2.0
   if (bright + incr > 255) incr = 255 - bright;
   ratio = (bright + incr) / bright;
   if (ratio < 0) ratio = 0;

   red1 = pix1[0];                                                         //  input RGB
   green1 = pix1[1];
   blue1 = pix1[2];
   
   red3 = ratio * red1;                                                    //  output RGB
   if (red3 > 255) red3 = 255;
   green3 = ratio * green1;
   if (green3 > 255) green3 = 255;
   blue3 = ratio * blue1;
   if (blue3 > 255) blue3 = 255;

   if (sa_stat == 3 && dist < sa_blend) {                                  //  select area is active,
      f1 = 1.0 * dist / sa_blend;                                          //    blend changes over sa_blend
      f2 = 1.0 - f1;
      red3 = f1 * red3 + f2 * red1;
      green3 = f1 * green3 + f2 * green1;
      blue3 = f1 * blue3 + f2 * blue1;
   }
   
   pix3[0] = red3;
   pix3[1] = green3;
   pix3[2] = blue3;

   Fbusy_done++;                                                           //  track progress
   return;
}


//  sharpen image by increasing brightness gradient

int sharp_GR()
{
   float       *pix1, *pix3;
   int         ii, px, py, dist = 0;   
   float       amount, thresh;
   float       b1, b1x, b1y, b3x, b3y, b3, bf, f1, f2;
   float       red1, green1, blue1, red3, green3, blue3;

   amount = 1 + 0.01 * sharp_GR_amount;                                    //  1.0 - 5.0
   thresh = sharp_GR_thresh;                                               //  0 - 100

   if (sa_stat == 3) Fbusy_goal = sa_Npixel;
   else  Fbusy_goal = E3pxm->ww * E3pxm->hh;
   Fbusy_done = 0;

   for (py = 1; py < E1pxm->hh; py++)                                      //  loop all image pixels
   for (px = 1; px < E1pxm->ww; px++)
   {
      if (sa_stat == 3) {                                                  //  select area active
         ii = py * E1pxm->ww + px;
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel is outside area
      }

      pix1 = PXMpix(E1pxm,px,py);                                          //  input pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel

      b1 = pixbright(pix1);                                                //  pixel brightness, 0 - 256
      b1x = b1 - pixbright(pix1-3);                                        //  horiz. brightness gradient
      b1y = b1 - pixbright(pix1-3 * E1pxm->ww);                            //  vertical
      f1 = fabsf(b1x + b1y);
      
      if (f1 < thresh)                                                     //  moderate brightness change for
         f1 = f1 / thresh;                                                 //    pixels below threshold gradient
      else  f1 = 1.0;
      f2 = 1.0 - f1;

      b1x = b1x * amount;                                                  //  amplified gradient
      b1y = b1y * amount;

      b3x = pixbright(pix1-3) + b1x;                                       //  + prior pixel brightness
      b3y = pixbright(pix1-3 * E3pxm->ww) + b1y;                           //  = new brightness
      b3 = 0.5 * (b3x + b3y);

      b3 = f1 * b3 + f2 * b1;                                              //  possibly moderated

      bf = b3 / b1;                                                        //  ratio of brightness change
      if (bf < 0) bf = 0;
      if (bf > 4) bf = 4;
      
      red1 = pix1[0];                                                      //  input RGB
      green1 = pix1[1];
      blue1 = pix1[2];
      
      red3 = bf * red1;                                                    //  output RGB
      if (red3 > 255.9) red3 = 255.9;
      green3 = bf * green1;
      if (green3 > 255.9) green3 = 255.9;
      blue3 = bf * blue1;
      if (blue3 > 255.9) blue3 = 255.9;

      if (sa_stat == 3 && dist < sa_blend) {                               //  select area is active,
         f1 = 1.0 * dist / sa_blend;                                       //    blend changes over sa_blend
         f2 = 1.0 - f1;
         red3 = f1 * red3 + f2 * red1;
         green3 = f1 * green3 + f2 * green1;
         blue3 = f1 * blue3 + f2 * blue1;
      }
      
      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;

      Fbusy_done++;                                                        //  track progress
   }

   Fbusy_goal = 0;
   return 1;
}


/**************************************************************************/

//  Compute the mean brightness of all pixel neighborhoods,
//  using a Guassian or a flat distribution for the weightings. 
//  If a select area is active, only inside pixels are calculated.
//  The flat method is 10-100x faster.

int         brhood_radius;
float       brhood_kernel[200][200];                                       //  up to radius = 99
char        brhood_method;                                                 //  g = gaussian, f = flat distribution

void brhood_calc(int radius, char method)
{
   void * brhood_wthread(void *arg);

   int      rad, rad2, dx, dy, ii;
   float    kern;
   
   brhood_radius = radius;
   brhood_method = method;

   if (brhood_method == 'g')                                               //  compute Gaussian kernel
   {                                                                       //  (not currently used)
      rad = brhood_radius;
      rad2 = rad * rad;

      for (dy = -rad; dy <= rad; dy++)
      for (dx = -rad; dx <= rad; dx++)
      {
         if (dx*dx + dy*dy <= rad2)                                        //  cells within radius
            kern = exp( - (dx*dx + dy*dy) / rad2);
         else kern = 0;                                                    //  outside radius
         brhood_kernel[dy+rad][dx+rad] = kern;
      }
   }

   if (sa_stat == 3) Fbusy_goal = sa_Npixel;                               //  set up progress tracking
   else Fbusy_goal = E1pxm->ww * E1pxm->hh;
   Fbusy_done = 0;

   for (ii = 0; ii < Nwt; ii++)                                            //  start worker threads
      start_wthread(brhood_wthread,&wtnx[ii]);
   wait_wthreads();                                                        //  wait for completion

   Fbusy_goal = 0;
   return;
}


//  worker thread function

void * brhood_wthread(void *arg)
{
   int      index = *((int *) arg);
   int      rad = brhood_radius;
   int      ii, px, py, qx, qy, Fstart;
   float    kern, bsum, bsamp, bmean;
   float    *pixel;

   if (brhood_method == 'g')                                               //  use round gaussian distribution
   {
      for (py = index; py < E1pxm->hh; py += Nwt)
      for (px = 0; px < E1pxm->ww; px++)
      {
         if (sa_stat == 3 && sa_mode != mode_image) {                      //  select area, not whole image
            ii = py * E1pxm->ww + px;                                      //    use only inside pixels
            if (! sa_pixmap[ii]) continue;
         }
         
         bsum = bsamp = 0;

         for (qy = py-rad; qy <= py+rad; qy++)                             //  computed weighted sum of brightness
         for (qx = px-rad; qx <= px+rad; qx++)                             //    for pixels in neighborhood
         {
            if (qy < 0 || qy > E1pxm->hh-1) continue;
            if (qx < 0 || qx > E1pxm->ww-1) continue;
            kern = brhood_kernel[qy+rad-py][qx+rad-px];
            pixel = PXMpix(E1pxm,qx,qy);
            bsum += pixbright(pixel) * kern;                               //  sum brightness * weight
            bsamp += kern;                                                 //  sum weights
         }

         bmean = bsum / bsamp;                                             //  mean brightness
         ii = py * E1pxm->ww + px;
         brhood_brightness[ii] = bmean;                                    //  pixel value

         Fbusy_done++;                                                     //  track progress
         if (drandz() < 0.0001) zsleep(0.001);                             //  trigger sorry kernel scheduler
      }
   }

   if (brhood_method == 'f')                                               //  use square flat distribution
   {
      Fstart = 1;
      bsum = bsamp = 0;

      for (py = index; py < E1pxm->hh; py += Nwt)
      for (px = 0; px < E1pxm->ww; px++)
      {
         if (sa_stat == 3 && sa_mode != mode_image) {                      //  select area, not whole image
            ii = py * E1pxm->ww + px;                                      //     compute only inside pixels
            if (! sa_pixmap[ii]) {
               Fstart = 1;
               continue;
            }
         }
         
         if (px == 0) Fstart = 1;
         
         if (Fstart) 
         {
            Fstart = 0;
            bsum = bsamp = 0;

            for (qy = py-rad; qy <= py+rad; qy++)                          //  add up all columns
            for (qx = px-rad; qx <= px+rad; qx++)
            {
               if (qy < 0 || qy > E1pxm->hh-1) continue;
               if (qx < 0 || qx > E1pxm->ww-1) continue;
               pixel = PXMpix(E1pxm,qx,qy);
               bsum += pixbright(pixel);
               bsamp += 1;
            }
         }
         else
         {
            qx = px-rad-1;                                                 //  subtract first-1 column
            if (qx >= 0) {
               for (qy = py-rad; qy <= py+rad; qy++)
               {
                  if (qy < 0 || qy > E1pxm->hh-1) continue;
                  pixel = PXMpix(E1pxm,qx,qy);
                  bsum -= pixbright(pixel);
                  bsamp -= 1;
               }
            }
            qx = px+rad;                                                   //  add last column
            if (qx < E1pxm->ww) {
               for (qy = py-rad; qy <= py+rad; qy++)
               {
                  if (qy < 0 || qy > E1pxm->hh-1) continue;
                  pixel = PXMpix(E1pxm,qx,qy);
                  bsum += pixbright(pixel);
                  bsamp += 1;
               }
            }
         }

         bmean = bsum / bsamp;                                             //  mean brightness
         ii = py * E1pxm->ww + px;
         brhood_brightness[ii] = bmean;

         Fbusy_done++;                                                     //  track progress
         if (drandz() < 0.0001) zsleep(0.001);                             //  trigger sorry kernel scheduler
      }
   }
         
   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************/

//  image noise reduction

int      denoise_method = 5;                                               //  default algorithm
int      denoise_radius = 4;
float    denoise_thresh = 0;
float    denoise_soft = 0;

editfunc    EFdenoise;


void m_denoise(GtkWidget *, cchar *)
{
   int    denoise_dialog_event(zdialog *zd, cchar *event);                 //  dialog event function
   void * denoise_thread(void *);

   cchar  *denoise_message = ZTX("Press the reduce button to reduce noise \n"
                                 "in small steps. Use [Undo] to start over.");

   zfuncs::F1_help_topic = "reduce_noise";

   EFdenoise.funcname = "denoise";
   EFdenoise.Farea = 2;                                                    //  select area usable
   EFdenoise.threadfunc = denoise_thread;                                  //  thread function
   if (! edit_setup(EFdenoise)) return;                                    //  setup edit

   zdialog *zd = zdialog_new(ZTX("Noise Reduction"),mWin,Bdone,Bcancel,null);
   EFdenoise.zd = zd;

   zdialog_add_widget(zd,"label","lab1","dialog",denoise_message,"space=5");
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labalg","hb1","Algorithm","space=5");
   zdialog_add_widget(zd,"combo","method","hb1",0,"space=5");
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labrad","hb2",Bradius,"space=5");
   zdialog_add_widget(zd,"spin","radius","hb2","1|9|1|4","space=5");
   zdialog_add_widget(zd,"button","reduce","hb2",Breduce,"space=5");

   zdialog_add_widget(zd,"hsep","sepw","dialog",0,"space=3");              //  wavelets         v.12.09
   zdialog_add_widget(zd,"hbox","hbw","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labw1","hbw","Wavelets:","space=5");
   zdialog_add_widget(zd,"label","labw2","hbw",Bthresh,"space=3");
   zdialog_add_widget(zd,"spin","wthresh","hbw","0|10|0.2|0");
   zdialog_add_widget(zd,"label","space","hbw",0,"space=3");
   zdialog_add_widget(zd,"label","labw3","hbw",ZTX("Soften"),"space=3");
   zdialog_add_widget(zd,"spin","wsoft","hbw","0|1|0.02|0");

   zdialog_cb_app(zd,"method",ZTX("flatten outliers by color (1)"));
   zdialog_cb_app(zd,"method",ZTX("flatten outliers by color (2)"));
   zdialog_cb_app(zd,"method",ZTX("set median brightness by color"));
   zdialog_cb_app(zd,"method",ZTX("top hat filter by color"));

   zdialog_stuff(zd,"method",ZTX("top hat filter by color"));              //  default method

   zdialog_help(zd,"reduce_noise");                                        //  zdialog help topic
   zdialog_run(zd,denoise_dialog_event,"save");                            //  run dialog - parallel
   return;
}


//  dialog event and completion callback function

int denoise_dialog_event(zdialog * zd, cchar *event)
{
   char     method[40];

   if (zd->zstat)
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 0;
   }

   if (strEqu(event,"blendwidth")) signal_thread();                        //  trigger update thread

   if (strEqu(event,"radius")) 
      zdialog_fetch(zd,"radius",denoise_radius);

   if (strEqu(event,"method")) 
   {
      zdialog_fetch(zd,"method",method,39);

      if (strEqu(method,"flatten outliers by color (1)")) {
         denoise_method = 1;
         denoise_radius = 1;
      }

      if (strEqu(method,"flatten outliers by color (2)")) {
         denoise_method = 2;
         denoise_radius = 3;
      }

      if (strEqu(method,"set median brightness by color")) {
         denoise_method = 4;
         denoise_radius = 2;
      }

      if (strEqu(method,"top hat filter by color")) {
         denoise_method = 5;
         denoise_radius = 4;
      }
      
      zdialog_stuff(zd,"radius",denoise_radius);
   }
   
   if (strEqu(event,"reduce")) signal_thread();                            //  trigger update thread for methods 1-5

   if (strstr("wthresh wsoft",event)) {                                    //  method 6, wavelets     v.12.09
      zdialog_fetch(zd,"wthresh",denoise_thresh);
      zdialog_fetch(zd,"wsoft",denoise_soft);
      denoise_method = 6;
      signal_thread();
   }
      
   return 1;
}


//  image noise reduction thread

void * denoise_thread(void *)
{
   void * denoise_wthread(void *arg);
   void * denoise_wavelet_wthread(void *arg);

   int      ii;

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      if (denoise_method == 6)                                             //  wavelet       v.12.09
      {
         for (ii = 0; ii < 3; ii++)                                        //  start worker threads for wavelet denoise
            start_wthread(denoise_wavelet_wthread,&wtnx[ii]);              //  (one thread per RGB color)
         wait_wthreads();                                                  //  wait for completion
         
         CEF->Fmod = 1;
         mwpaint2();                                                       //  update window
         continue;
      }

      E9pxm = PXM_copy(E3pxm);                                             //  image3 is reference source
                                                                           //  image9 will be modified
      if (sa_stat == 3) Fbusy_goal = sa_Npixel;
      else  Fbusy_goal = E3pxm->ww * E3pxm->hh;
      Fbusy_done = 0;

      for (ii = 0; ii < Nwt; ii++)                                         //  start worker threads for methods 1-5
         start_wthread(denoise_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      Fbusy_goal = 0;

      mutex_lock(&Fpixmap_lock);
      PXM_free(E3pxm);                                                     //  image9 >> image3
      E3pxm = E9pxm;
      E9pxm = 0;
      mutex_unlock(&Fpixmap_lock);

      CEF->Fmod = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * denoise_wthread(void *arg)                                          //  worker thread for methods 1-5
{
   void  denoise_flatten_minmax(float *pix3, float *pix9);
   void  denoise_flatten_stdev(float *pix3, float *pix9);
   void  denoise_flatten_median(float *pix3, float *pix9);
   void  denoise_flatten_tophat(float *pix3, float *pix9);
   
   int         index = *((int *) arg);
   int         ii, px, py, rad, dist = 0;
   float       f1, f2;
   float       *pix1, *pix3, *pix9;

   rad = denoise_radius;

   for (py = index+rad; py < E3pxm->hh-rad; py += Nwt)                     //  loop all image3 pixels
   for (px = rad; px < E3pxm->ww-rad; px++)
   {
      if (sa_stat == 3) {                                                  //  select area active
         ii = py * E1pxm->ww + px;
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  outside pixel
      }

      pix3 = PXMpix(E3pxm,px,py);                                          //  source pixel
      pix9 = PXMpix(E9pxm,px,py);                                          //  target pixel

      if (denoise_method == 1) denoise_flatten_minmax(pix3,pix9);
      if (denoise_method == 2) denoise_flatten_stdev(pix3,pix9);
      if (denoise_method == 4) denoise_flatten_median(pix3,pix9);
      if (denoise_method == 5) denoise_flatten_tophat(pix3,pix9);

      if (sa_stat == 3 && dist < sa_blend) {                               //  select area is active,
         f1 = 1.0 * dist / sa_blend;                                       //    blend changes over sa_blend
         f2 = 1.0 - f1;
         pix1 = PXMpix(E1pxm,px,py);                                       //  source pixel
         pix9[0] = f1 * pix9[0] + f2 * pix1[0];
         pix9[1] = f1 * pix9[1] + f2 * pix1[1];                            //  remove int(*)       v.12.09
         pix9[2] = f1 * pix9[2] + f2 * pix1[2];
      }

      Fbusy_done++;                                                        //  track progress
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


//  Flatten outliers within radius, by color.
//  An outlier is the max or min RGB value within a radius.

void denoise_flatten_minmax(float *pix3, float *pix9)
{
   int         dy, dx, rad;
   int         min0, min1, min2, max0, max1, max2;
   float       *pixN;

   min0 = min1 = min2 = 255;
   max0 = max1 = max2 = 0;
   rad = denoise_radius;

   for (dy = -rad; dy <= rad; dy++)                                        //  loop surrounding pixels
   for (dx = -rad; dx <= rad; dx++)
   {
      if (dy == 0 && dx == 0) continue;                                    //  skip self

      pixN = pix3 + (dy * E3pxm->ww + dx) * 3;
      if (pixN[0] < min0) min0 = pixN[0];                                  //  find min and max per color
      if (pixN[0] > max0) max0 = pixN[0];
      if (pixN[1] < min1) min1 = pixN[1];
      if (pixN[1] > max1) max1 = pixN[1];
      if (pixN[2] < min2) min2 = pixN[2];
      if (pixN[2] > max2) max2 = pixN[2];
   }
   
   if (pix3[0] <= min0 && min0 < 255) pix9[0] = min0 + 1;                  //  if outlier, flatten a little
   if (pix3[0] >= max0 && max0 > 1) pix9[0] = max0 - 1;
   if (pix3[1] <= min1 && min1 < 255) pix9[1] = min1 + 1;
   if (pix3[1] >= max1 && max1 > 1) pix9[1] = max1 - 1;
   if (pix3[2] <= min2 && min2 < 255) pix9[2] = min2 + 1;
   if (pix3[2] >= max2 && max2 > 1) pix9[2] = max2 - 1;
   
   return;
}


//  Flatten outliers.
//  An outlier pixel has an RGB value outside one sigma of 
//  the mean for all pixels within a given radius of the pixel.

void denoise_flatten_stdev(float *pix3, float *pix9)
{
   int         rgb, dy, dx, rad, nn;
   float       nn1, val, sum, sum2, mean, variance, sigma;
   float       *pixN;

   rad = denoise_radius;
   nn = (rad * 2 + 1);
   nn = nn * nn - 1;
   nn1 = 1.0 / nn;

   for (rgb = 0; rgb < 3; rgb++)                                           //  loop RGB color
   {
      sum = sum2 = 0;

      for (dy = -rad; dy <= rad; dy++)                                     //  loop surrounding pixels
      for (dx = -rad; dx <= rad; dx++)
      {
         if (dy == 0 && dx == 0) continue;                                 //  skip self
         pixN = pix3 + (dy * E3pxm->ww + dx) * 3;
         val = pixN[rgb];
         sum += val;
         sum2 += val * val;
      }
      
      mean = nn1 * sum;
      variance = nn1 * (sum2 - 2.0 * mean * sum) + mean * mean;
      sigma = sqrtf(variance);

      val = pix3[rgb];      
      if (val > mean + sigma) {                                            //  move value to mean +/- sigma
         val = mean + sigma;
         pix9[rgb] = val;
      }
      else if (val < mean - sigma) {
         val = mean - sigma;
         pix9[rgb] = val;
      }
   }
   
   return;
}


//  use median RGB brightness for pixels within radius

void denoise_flatten_median(float *pix3, float *pix9)
{
   int         dy, dx, rad, ns, rgb;
   float       bsortN[400], *pixN;

   rad = denoise_radius;

   for (rgb = 0; rgb < 3; rgb++)                                           //  loop all RGB colors
   {
      ns = 0;

      for (dy = -rad; dy <= rad; dy++)                                     //  loop surrounding pixels
      for (dx = -rad; dx <= rad; dx++)                                     //  get brightness values
      {
         pixN = pix3 + (dy * E3pxm->ww + dx) * 3;
         bsortN[ns] = pixN[rgb];
         ns++;
      }

      HeapSort(bsortN,ns);
      pix9[rgb] = bsortN[ns/2];                                            //  median brightness of ns pixels
   }

   return;
}


//  Modified top hat filter: execute with increasing radius from 1 to limit.
//  Detect outlier by comparing with pixels along outer radius only.

void denoise_flatten_tophat(float *pix3, float *pix9)
{
   int         dy, dx, rad;
   float       min0, min1, min2, max0, max1, max2;
   float       *pixN;

   for (rad = 1; rad <= denoise_radius; rad++)
   for (int loops = 0; loops < 2; loops++)
   {
      min0 = min1 = min2 = 255;
      max0 = max1 = max2 = 0;

      for (dy = -rad; dy <= rad; dy++)                                     //  loop all pixels within rad
      for (dx = -rad; dx <= rad; dx++)
      {
         if (dx > -rad && dx < rad) continue;                              //  skip inner pixels
         if (dy > -rad && dy < rad) continue;

         pixN = pix3 + (dy * E3pxm->ww + dx) * 3;
         if (pixN[0] < min0) min0 = pixN[0];                               //  find min and max per color
         if (pixN[0] > max0) max0 = pixN[0];                               //    among outermost pixels
         if (pixN[1] < min1) min1 = pixN[1];
         if (pixN[1] > max1) max1 = pixN[1];
         if (pixN[2] < min2) min2 = pixN[2];
         if (pixN[2] > max2) max2 = pixN[2];
      }
      
      if (pix3[0] < min0 && pix9[0] < 255) pix9[0]++;                      //  if central pixel is outlier,
      if (pix3[0] > max0 && pix9[0] > 1) pix9[0]--;                        //    moderate its values
      if (pix3[1] < min1 && pix9[1] < 255) pix9[1]++;
      if (pix3[1] > max1 && pix9[1] > 1) pix9[1]--;
      if (pix3[2] < min2 && pix9[2] < 255) pix9[2]++;
      if (pix3[2] > max2 && pix9[2] > 1) pix9[2]--;
   }

   return;
}


//  worker thread for method 6
//  do wavelet denoise for one color in each of 3 threads

void * denoise_wavelet_wthread(void *arg)                                  //  v.12.09
{
   void wavelet_denoise (float *fimg[3], uint ww, uint hh, float, float);

   int      rgb = *((int *) arg);                                          //  rgb color 0/1/2
   int      ii, jj;
   float    *fimg[3];
   float    f256 = 1.0 / 256.0;
   
   if (sa_stat == 3) goto denoise_area;                                    //  select area is active
   
   fimg[0] = (float *) zmalloc(E1pxm->ww * E1pxm->hh * sizeof(float),"denoise");
   fimg[1] = (float *) zmalloc(E1pxm->ww * E1pxm->hh * sizeof(float),"denoise");
   fimg[2] = (float *) zmalloc(E1pxm->ww * E1pxm->hh * sizeof(float),"denoise");

   for (ii = 0; ii < E1pxm->ww * E1pxm->hh; ii++)                          //  extract one noisy color from E1
      fimg[0][ii] = E1pxm->pixels[3*ii+rgb] * f256;
   
   wavelet_denoise(fimg,E1pxm->ww,E1pxm->hh,denoise_thresh,denoise_soft);
   
   for (ii = 0; ii < E1pxm->ww * E1pxm->hh; ii++)                          //  save one denoised color to E3
      E3pxm->pixels[3*ii+rgb] = 256.0 * fimg[0][ii];
   
   zfree(fimg[0]);
   zfree(fimg[1]);
   zfree(fimg[2]);
   
   exit_wthread();
   return 0;

denoise_area:

   int      px, py, pxl, pxh, pyl, pyh, ww, hh, dist;
   float    f1, f2;

   pxl = sa_minx - 16;
   if (pxl < 0) pxl = 0;
   pxh = sa_maxx + 16;
   if (pxh > E1pxm->ww) pxh = E1pxm->ww;

   pyl = sa_miny - 16;
   if (pyl < 0) pyl = 0;
   pyh = sa_maxy + 16;
   if (pyh > E1pxm->hh) pyh = E1pxm->hh;
   
   ww = pxh - pxl;
   hh = pyh - pyl;

   fimg[0] = (float *) zmalloc(ww * hh * sizeof(float),"denoise");
   fimg[1] = (float *) zmalloc(ww * hh * sizeof(float),"denoise");
   fimg[2] = (float *) zmalloc(ww * hh * sizeof(float),"denoise");
   
   for (py = 0; py < hh; py++)
   for (px = 0; px < ww; px++)
   {
      ii = py * ww + px;
      jj = (py + pyl) * E1pxm->ww + (px + pxl);
      fimg[0][ii] = E1pxm->pixels[3*jj+rgb] * f256;
   }

   wavelet_denoise(fimg,ww,hh,denoise_thresh,denoise_soft);
   
   for (py = 0; py < hh; py++)
   for (px = 0; px < ww; px++)
   {
      ii = py * ww + px;
      jj = (py + pyl) * E1pxm->ww + (px + pxl);

      dist = sa_pixmap[jj];
      if (! dist) continue;

      if (dist < sa_blend) {
         f1 = 1.0 * dist / sa_blend;
         f2 = 1.0 - f1;
         E3pxm->pixels[3*jj+rgb] = f1 * 256.0 * fimg[0][ii] + f2 * E1pxm->pixels[3*jj+rgb];
      }
      else E3pxm->pixels[3*jj+rgb] = 256.0 * fimg[0][ii];
   }

   zfree(fimg[0]);
   zfree(fimg[1]);
   zfree(fimg[2]);

   exit_wthread();
   return 0;
}


//  new v.12.09
//  wavelet denoise algorithm
//  Adapted from Gimp wavelet plugin (and ultimately dcraw by Dave Coffin).
//  fimg[0][rows][cols] = one color of image to denoise
//  fimg[1] and [2] = working space
//  thresh (0-10) and soft (0-1) are adjustable parameters

void wavelet_denoise(float *fimg[3], uint ww, uint hh, float thresh, float soft)
{
   void wavelet_avgpix(float *temp, float *fimg, int st, int size, int sc);

   float    *temp, thold, stdev[5];
   uint     ii, lev, lpass, hpass, size, col, row;
   uint     samples[5];

   size = ww * hh;
   temp = (float *) zmalloc ((ww + hh) * sizeof(float),"denoise");
   hpass = 0;

   for (lev = 0; lev < 5; lev++)
   {
      lpass = ((lev & 1) + 1);                                             //  1, 2, 1, 2, 1

      for (row = 0; row < hh; row++)                                       //  average row pixels
	   {
	      wavelet_avgpix(temp, fimg[hpass] + row * ww, 1, ww, 1 << lev);

	      for (col = 0; col < ww; col++)
	         fimg[lpass][row * ww + col] = temp[col];
	   }

      for (col = 0; col < ww; col++)                                       //  average column pixels
	   {
	      wavelet_avgpix(temp, fimg[lpass] + col, ww, hh, 1 << lev);

	      for (row = 0; row < hh; row++)
	         fimg[lpass][row * ww + col] = temp[row];
	   }

      thold = 5.0 / (1 << 6) * exp (-2.6 * sqrt (lev + 1)) * 0.8002 / exp (-2.6);

      stdev[0] = stdev[1] = stdev[2] = stdev[3] = stdev[4] = 0.0;
      samples[0] = samples[1] = samples[2] = samples[3] = samples[4] = 0;

      for (ii = 0; ii < size; ii++)
	   {
	      fimg[hpass][ii] -= fimg[lpass][ii];

	      if (fimg[hpass][ii] < thold && fimg[hpass][ii] > -thold)
	      {
	         if (fimg[lpass][ii] > 0.8) {
	            stdev[4] += fimg[hpass][ii] * fimg[hpass][ii];
	            samples[4]++;
	         } 
	         else if (fimg[lpass][ii] > 0.6) {
	            stdev[3] += fimg[hpass][ii] * fimg[hpass][ii];
	            samples[3]++;
	         }	
	         else if (fimg[lpass][ii] > 0.4) {
	            stdev[2] += fimg[hpass][ii] * fimg[hpass][ii];
	            samples[2]++;
	         }	
	         else if (fimg[lpass][ii] > 0.2) {
	            stdev[1] += fimg[hpass][ii] * fimg[hpass][ii];
	            samples[1]++;
	         } 
	         else {
	            stdev[0] += fimg[hpass][ii] * fimg[hpass][ii];
	            samples[0]++;
	         }
	      }
	   }

      stdev[0] = sqrt (stdev[0] / (samples[0] + 1));
      stdev[1] = sqrt (stdev[1] / (samples[1] + 1));
      stdev[2] = sqrt (stdev[2] / (samples[2] + 1));
      stdev[3] = sqrt (stdev[3] / (samples[3] + 1));
      stdev[4] = sqrt (stdev[4] / (samples[4] + 1));

      for (ii = 0; ii < size; ii++)                                        //  do thresholding
	   {
	      if (fimg[lpass][ii] > 0.8) 
	         thold = thresh * stdev[4];
	      else if (fimg[lpass][ii] > 0.6) 
	         thold = thresh * stdev[3];
	      else if (fimg[lpass][ii] > 0.4) 
	         thold = thresh * stdev[2];
	      else if (fimg[lpass][ii] > 0.2) 
	         thold = thresh * stdev[1];
	      else 
	         thold = thresh * stdev[0];

	      if (fimg[hpass][ii] < -thold)
	         fimg[hpass][ii] += thold - thold * soft;
	      else if (fimg[hpass][ii] > thold)
	         fimg[hpass][ii] -= thold - thold * soft;
	      else
	         fimg[hpass][ii] *= soft;

	      if (hpass) fimg[0][ii] += fimg[hpass][ii];
	   }

      hpass = lpass;
   }

   for (ii = 0; ii < size; ii++)
      fimg[0][ii] = fimg[0][ii] + fimg[lpass][ii];

   zfree(temp);
   return;
}


//  average pixels in one column or row
//  st = row stride (row length) or column stride (1)
//  sc = 1, 2, 4, 8, 16 = pixels +/- from target pixel to average

void wavelet_avgpix(float *temp, float *fimg, int st, int size, int sc)
{
  int ii;

  for (ii = 0; ii < sc; ii++)
    temp[ii] = 0.25*(2*fimg[st*ii] + fimg[st*(sc-ii)] + fimg[st*(ii+sc)]);

  for (; ii < size - sc; ii++)
    temp[ii] = 0.25*(2*fimg[st*ii] + fimg[st*(ii-sc)] + fimg[st*(ii+sc)]);

  for (; ii < size; ii++)
    temp[ii] = 0.25*(2*fimg[st*ii] + fimg[st*(ii-sc)] + fimg[st*(2*size-2-(ii+sc))]);
  
  return;
}



