/**************************************************************************

   Fotoxx      edit photos and manage collections

   Copyright 2007 2008 2009 2010 2011 2012 Michael Cornelison
   Source URL: http://kornelix.com/fotoxx
   Contact: kornelix2@gmail.com
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see http://www.gnu.org/licenses/.

***************************************************************************/

#define EX extern                                                          //  enable extern declarations
#include "fotoxx.h"

/**************************************************************************
   Fotoxx image edit - retouch2 functions
***************************************************************************/

/**************************************************************************/

//  Smart Erase menu function - Replace pixels inside a select area 
//    with a reflection of pixels outside the area.

editfunc    EFerase;

void m_smart_erase(GtkWidget *, const char *)
{
   int smart_erase_dialog_event(zdialog* zd, const char *event);

   int      cc;
   cchar    *erase_message = ZTX("1. Drag mouse to select. \n"
                                 "2. Erase.   3. Repeat. ");

   zfuncs::F1_help_topic = "smart_erase";
   
   EFerase.funcname = "smart-erase";
   EFerase.Farea = 0;                                                      //  select area deleted
   EFerase.mousefunc = sa_mouse_select;                                    //  mouse function (use select area)
   if (! edit_setup(EFerase)) return;                                      //  setup edit
   
/*     _________________________________________
      |                                         |
      |    1. Drag mouse to select.             |
      |    2. Erase.   3. Repeat.               |
      |                                         |
      | Radius [10|v]    Blur [0.5|v]           |
      | [New Area] [Show] [Hide] [Erase] [Undo] |
      |                                         |
      |                                 [Done]  |
      |_________________________________________|
*/

   zdialog *zd = zdialog_new(ZTX("Smart Erase"),mWin,Bdone,null);
   EFerase.zd = zd;

   zdialog_add_widget(zd,"label","lab1","dialog",erase_message,"space=8");
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labr","hb2",ZTX("Radius"),"space=5");
   zdialog_add_widget(zd,"spin","radius","hb2","1|20|1|10");
   zdialog_add_widget(zd,"label","labb","hb2",ZTX("Blur"),"space=10");
   zdialog_add_widget(zd,"spin","blur","hb2","0|9|0.5|0.5");
   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","newarea","hb3",ZTX("New Area"),"space=3");
   zdialog_add_widget(zd,"button","show","hb3",ZTX(Bshow),"space=3");
   zdialog_add_widget(zd,"button","hide","hb3",Bhide,"space=3");
   zdialog_add_widget(zd,"button","erase","hb3",Berase,"space=3");
   zdialog_add_widget(zd,"button","undo1","hb3",Bundo,"space=3");

   zdialog_help(zd,"smart_erase");                                         //  zdialog help topic 
   zdialog_run(zd,smart_erase_dialog_event,"save");                        //  run dialog - parallel

   sa_unselect();                                                          //  unselect area if any
   cc = E1pxm->ww * E1pxm->hh * sizeof(uint16);                            //  create new area
   sa_pixmap = (uint16 *) zmalloc(cc,"smart_erase");
   memset(sa_pixmap,0,cc);
   sa_mode = mode_mouse;                                                   //  mode = select by mouse
   sa_stat = 1;                                                            //  status = active edit
   sa_fww = E1pxm->ww;
   sa_fhh = E1pxm->hh;
   sa_searchrange = 1;                                                     //  search within mouse radius
   sa_mouseradius = 10;                                                    //  initial mouse select radius
   takeMouse(sa_mouse_select,0);                                           //  use select area mouse function
   sa_show(1);
   return;
}


//  dialog event and completion function

int smart_erase_dialog_event(zdialog *zd, const char *event)               //  overhauled
{
   void smart_erase_func(int mode);
   void smart_erase_blur(float radius);
   
   float       radius;
   int         cc;
   
   if (zd->zstat) {
      sa_unselect();                                                       //  delete select area
      freeMouse();                                                         //  disconnect mouse
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 0;
   }
   
   if (strEqu(event,"radius"))
      zdialog_fetch(zd,"radius",sa_mouseradius);
      
   if (strEqu(event,"newarea")) {
      sa_unselect();
      cc = E1pxm->ww * E1pxm->hh * sizeof(uint16);                         //  create new area
      sa_pixmap = (uint16 *) zmalloc(cc,"smart_erase");
      memset(sa_pixmap,0,cc);
      sa_mode = mode_mouse;                                                //  mode = select by mouse
      sa_stat = 1;                                                         //  status = active edit
      sa_fww = E1pxm->ww;
      sa_fhh = E1pxm->hh;
      sa_show(1);
      takeMouse(sa_mouse_select,0);
   }
   
   if (strEqu(event,"show")) {
      sa_stat = 1;                                                         //  status = active edit
      sa_show(1);
      takeMouse(sa_mouse_select,0);
   }

   if (strEqu(event,"hide")) {
      sa_show(0);
      freeMouse();
   }
   
   if (strEqu(event,"erase")) {                                            //  do smart erase
      sa_finish_auto();                                                    //  finish the area
      smart_erase_func(1);
      zdialog_fetch(zd,"blur",radius);                                     //  add optional blur
      if (radius > 0) smart_erase_blur(radius);
      sa_show(0);
   }
   
   if (strEqu(event,"undo1"))                                              //  dialog undo, undo last erase
      smart_erase_func(2);

   return 0;
}


//  erase the area or restore last erased area

void smart_erase_func(int mode)
{
   int         px, py, npx, npy;
   int         qx, qy, sx, sy, tx, ty;
   int         ii, rad, inc, cc;
   int         dist2, mindist2;
   float       slope;
   char        *pmap;
   float       *pix1, *pix3;

   if (sa_stat != 3) return;                                               //  nothing selected
   if (! sa_validate()) return;                                            //  area invalid for curr. image file  v.12.12
   
   for (py = sa_miny; py < sa_maxy; py++)                                  //  loop all pixels in area
   for (px = sa_minx; px < sa_maxx; px++)
   {
      ii = py * E1pxm->ww + px;
      if (! sa_pixmap[ii]) continue;                                       //  pixel not selected

      pix1 = PXMpix(E1pxm,px,py);                                          //  input pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel

      pix3[0] = pix1[0];                                                   //  restore pixels inside area
      pix3[1] = pix1[1];
      pix3[2] = pix1[2];
   }

   mwpaint2();                                                             //  update window
   
   if (mode == 2) return;                                                  //  mode = undo, done

   cc = E1pxm->ww * E1pxm->hh;                                             //  allocate pixel done map
   pmap = (char *) zmalloc(cc,"smart_erase");
   memset(pmap,0,cc);

   for (py = sa_miny; py < sa_maxy; py++)                                  //  loop all pixels in area
   for (px = sa_minx; px < sa_maxx; px++)
   {
      ii = py * E1pxm->ww + px;
      if (! sa_pixmap[ii]) continue;                                       //  pixel not selected
      if (pmap[ii]) continue;                                              //  pixel already done

      mindist2 = 999999;                                                   //  find nearest edge
      npx = npy = 0;

      for (rad = 1; rad < 50; rad++)                                       //  50 pixel limit
      {
         for (qx = px-rad; qx <= px+rad; qx++)                             //  search within rad
         for (qy = py-rad; qy <= py+rad; qy++)
         {
            if (qx < 0 || qx >= E1pxm->ww) continue;                       //  off image edge
            if (qy < 0 || qy >= E1pxm->hh) continue;
            ii = qy * E1pxm->ww + qx;
            if (sa_pixmap[ii]) continue;                                   //  within selected area

            dist2 = (px-qx) * (px-qx) + (py-qy) * (py-qy);                 //  distance**2 to edge pixel
            if (dist2 < mindist2) {
               mindist2 = dist2;
               npx = qx;                                                   //  save nearest edge pixel found
               npy = qy;
            }
         }

         if (rad * rad >= mindist2) break;                                 //  can quit now
      }

      if (! npx && ! npy) continue;                                        //  edge not found, should not happen

      qx = npx;                                                            //  nearest edge pixel from px/py
      qy = npy;
      
      if (abs(qy - py) > abs(qx - px)) {
         slope = 1.0 * (qx - px) / (qy - py);
         if (qy > py) inc = 1;
         else inc = -1;
         for (sy = py; sy != qy; sy += inc)                                //  line from px/py to qx/qy
         {
            sx = px + slope * (sy - py);
            ii = sy * E1pxm->ww + sx;
            if (pmap[ii]) continue;
            pmap[ii] = 1;
            tx = qx + (qx - sx);                                           //  tx/ty = parallel line from qx/qy 
            ty = qy + (qy - sy);
            if (tx < 0) tx = 0;
            if (tx > E1pxm->ww-1) tx = E1pxm->ww-1;
            if (ty < 0) ty = 0;
            if (ty > E1pxm->hh-1) ty = E1pxm->hh-1;
            pix1 = PXMpix(E3pxm,tx,ty);                                    //  copy pixel from tx/ty to sx/sy
            pix3 = PXMpix(E3pxm,sx,sy);
            pix3[0] = pix1[0];
            pix3[1] = pix1[1];
            pix3[2] = pix1[2];
         }
      }

      else {
         slope = 1.0 * (qy - py) / (qx - px);
         if (qx > px) inc = 1;
         else inc = -1;
         for (sx = px; sx != qx; sx += inc)
         {
            sy = py + slope * (sx - px);
            ii = sy * E1pxm->ww + sx;
            if (pmap[ii]) continue;
            pmap[ii] = 1;
            tx = qx + (qx - sx);
            ty = qy + (qy - sy);
            if (tx < 0) tx = 0;
            if (tx > E1pxm->ww-1) tx = E1pxm->ww-1;
            if (ty < 0) ty = 0;
            if (ty > E1pxm->hh-1) ty = E1pxm->hh-1;
            pix1 = PXMpix(E3pxm,tx,ty);
            pix3 = PXMpix(E3pxm,sx,sy);
            pix3[0] = pix1[0];
            pix3[1] = pix1[1];
            pix3[2] = pix1[2];
         }
      }
   }

   zfree(pmap);                                                            //  free memory
   CEF->Fmod = 1;
   mwpaint2();                                                             //  update window
   return;
}


//  add blur to the erased area to help mask the side-effects

int smart_erase_blur(float radius)
{
   int         ii, px, py, dx, dy, adx, ady;
   float       blur_weight[12][12];                                        //  up to blur radius = 10
   float       rad, rad2;
   float       m, d, w, sum, weight;
   float       red, green, blue;
   float       *pix9, *pix3, *pixN;
   
   if (sa_stat != 3) return 0;
  
   rad = radius - 0.2;
   rad2 = rad * rad;

   for (dx = 0; dx < 12; dx++)                                             //  clear weights array
   for (dy = 0; dy < 12; dy++)
      blur_weight[dx][dy] = 0;

   for (dx = -rad-1; dx <= rad+1; dx++)                                    //  blur_weight[dx][dy] = no. of pixels
   for (dy = -rad-1; dy <= rad+1; dy++)                                    //    at distance (dx,dy) from center
      ++blur_weight[abs(dx)][abs(dy)];

   m = sqrt(rad2 + rad2);                                                  //  corner pixel distance from center
   sum = 0;

   for (dx = 0; dx <= rad+1; dx++)                                         //  compute weight of pixel
   for (dy = 0; dy <= rad+1; dy++)                                         //    at distance dx, dy
   {
      d = sqrt(dx*dx + dy*dy);
      w = (m + 1.2 - d) / m;
      w = w * w;
      sum += blur_weight[dx][dy] * w;
      blur_weight[dx][dy] = w;
   }

   for (dx = 0; dx <= rad+1; dx++)                                         //  make weights add up to 1.0
   for (dy = 0; dy <= rad+1; dy++)
      blur_weight[dx][dy] = blur_weight[dx][dy] / sum;

   E9pxm = PXM_copy(E3pxm);                                                //  copy edited image
   
   for (py = sa_miny; py < sa_maxy; py++)                                  //  loop all pixels in area
   for (px = sa_minx; px < sa_maxx; px++)
   {
      ii = py * E1pxm->ww + px;
      if (! sa_pixmap[ii]) continue;                                       //  pixel not in area

      pix9 = PXMpix(E9pxm,px,py);                                          //  source pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  target pixel
      
      rad = radius;
      red = green = blue = 0;
   
      for (dy = -rad-1; dy <= rad+1; dy++)                                 //  loop neighbor pixels within radius
      for (dx = -rad-1; dx <= rad+1; dx++)                                 //  v.12.04
      {
         if (px+dx < 0 || px+dx >= E3pxm->ww) continue;                    //  omit pixels off edge
         if (py+dy < 0 || py+dy >= E3pxm->hh) continue;
         adx = abs(dx);
         ady = abs(dy);
         pixN = pix9 + (dy * E3pxm->ww + dx) * 3;
         weight = blur_weight[adx][ady];                                   //  weight at distance (dx,dy)
         red += pixN[0] * weight;                                          //  accumulate contributions
         green += pixN[1] * weight;
         blue += pixN[2] * weight;
      }
      
      pix3[0] = red;
      pix3[1] = green;
      pix3[2] = blue;
   }

   PXM_free(E9pxm);

   CEF->Fmod = 1;
   mwpaint2();                                                             //  update window
   return 0;
}


/**************************************************************************

   Vignette function
   
   1. Change the brightness from center to edge using a curve.
   2. Change the color from center to edge using a color and a curve.
      (the pixel varies between original RGB and selected color)
   3. Mouse click or drag on image sets a new vignette center.

***************************************************************************/

void  vign_mousefunc();                                                    //  v.12.09

editfunc    EFvignette;
uint8       vignette_RGB[3] = { 0, 0, 255 };
int         vignette_spc;
float       vign_cx, vign_cy;
float       vign_rad;


void m_vignette(GtkWidget *, cchar *)
{
   int      Vign_dialog_event(zdialog *zd, cchar *event);
   void     Vign_curvedit(int);
   void *   Vign_thread(void *);

   zfuncs::F1_help_topic = "vignette";

   cchar       *title = ZTX("Vignette Tool");

   EFvignette.funcname = "vignette";
   EFvignette.Farea = 2;                                                   //  select area usable
   EFvignette.Fprev = 1;                                                   //  use preview image
   EFvignette.threadfunc = Vign_thread;                                    //  thread function
   EFvignette.mousefunc = vign_mousefunc;                                  //  mouse function

   if (! edit_setup(EFvignette)) return;                                   //  setup edit

/***
            _____________________________
           |                             |
           |                             |
           |    curve drawing area       |
           |                             |
           |_____________________________| 
            center                   edge

            (o) Brightness  (o) Color [___]
            Curve File: [ Open ] [ Save ]

***/

   zdialog *zd = zdialog_new(title,mWin,Bdone,Bcancel,null);
   EFvignette.zd = zd;

   zdialog_add_widget(zd,"frame","frame","dialog",0,"expand");
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labcenter","hb1",ZTX("center"),"space=4");
   zdialog_add_widget(zd,"label","space","hb1",0,"expand");
   zdialog_add_widget(zd,"label","labedge","hb1",ZTX("edge"),"space=5");

   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=3");
   zdialog_add_widget(zd,"radio","RBbrite","hb2",Bbrightness,"space=5");
   zdialog_add_widget(zd,"radio","RBcolor","hb2",Bcolor,"space=5");
   zdialog_add_widget(zd,"colorbutt","color","hb2","0|0|255");

   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labcurve","hb3",Bcurvefile,"space=5");
   zdialog_add_widget(zd,"button","load","hb3",Bopen,"space=5");
   zdialog_add_widget(zd,"button","save","hb3",Bsave,"space=5");
   
   vignette_RGB[0] = vignette_RGB[1] = 0;                                  //  initial color = blue
   vignette_RGB[2] = 255;
   
   vign_cx = E3pxm->ww / 2;                                                //  initial vignette center         v.12.10
   vign_cy = E3pxm->hh / 2;
   
   vign_rad = vign_cx * vign_cx + vign_cy * vign_cy;                       //  radius = distance to corners    v.12.10
   vign_rad = sqrtf(vign_rad);
   
   zdialog_stuff(zd,"RBbrite",1);                                          //  default curve = brightness

   GtkWidget *frame = zdialog_widget(zd,"frame");                          //  set up curve edit
   spldat *sd = splcurve_init(frame,Vign_curvedit);
   EFvignette.curves = sd;

   sd->vert[0] = 0;                                                        //  curve 0 = active curve = brightness
   sd->nap[0] = 2;
   sd->apx[0][0] = 0.01;
   sd->apy[0][0] = 0.5;
   sd->apx[0][1] = 0.99;
   sd->apy[0][1] = 0.5;
   splcurve_generate(sd,0);                                                //  generate curve data

   sd->vert[1] = 0;                                                        //  curve 1 = brightness curve
   sd->nap[1] = 2;
   sd->apx[1][0] = 0.01;
   sd->apy[1][0] = 0.5;
   sd->apx[1][1] = 0.99;
   sd->apy[1][1] = 0.5;
   splcurve_generate(sd,1);

   sd->vert[2] = 0;                                                        //  curve 2 = color curve
   sd->nap[2] = 2;
   sd->apx[2][0] = 0.01;
   sd->apy[2][0] = 0.01;
   sd->apx[2][1] = 0.99;
   sd->apy[2][1] = 0.01;
   splcurve_generate(sd,2);

   sd->Nspc = 1;                                                           //  only one active curve
   vignette_spc = 1;                                                       //  initial curve = brightness

   zdialog_resize(zd,300,300);
   zdialog_help(zd,"vignette");                                            //  zdialog help topic 
   zdialog_run(zd,Vign_dialog_event,"save");                               //  run dialog - parallel
   
   takeMouse(vign_mousefunc,dragcursor);                                   //  connect mouse function

   return;
}


//  dialog event and completion callback function

int Vign_dialog_event(zdialog *zd, cchar *event)
{
   void     Vign_curvedit(int);

   spldat   *sd = EFvignette.curves;
   int      ii, jj;
   char     color[20];
   cchar    *pp;

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) {                                                //  done
         float R = 1.0 * E0pxm->ww / E3pxm->ww;
         vign_cx = R * vign_cx;                                            //  scale geometries to full size   v.12.10
         vign_cy = R * vign_cy;
         vign_rad = R * vign_rad;
         edit_fullsize();                                                  //  last edit on full size image
         edit_done();
      }
      else edit_cancel();                                                  //  cancel
      return 1;
   }

   if (strEqu(event,"focus"))                                              //  toggle mouse capture
      takeMouse(vign_mousefunc,dragcursor);                                //  connect mouse function

   if (strnEqu(event,"RB",2)) {                                            //  new choice of curve
      ii = strcmpv(event,"RBbrite","RBcolor",null);
      vignette_spc = ii;
      sd->nap[0] = sd->nap[ii];                                            //  copy new curve to curve 0
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[0][jj] = sd->apx[ii][jj];
         sd->apy[0][jj] = sd->apy[ii][jj];
      }
      splcurve_generate(sd,0);                                             //  regenerate curve 0
      gtk_widget_queue_draw(sd->drawarea);                                 //  draw curve
   }

   if (strEqu(event,"blendwidth")) signal_thread();

   if (strEqu(event,"color")) {                                            //  change color
      zdialog_fetch(zd,"color",color,19);                                  //  get color from color wheel
      pp = strField(color,"|",1);
      if (pp) vignette_RGB[0] = atoi(pp);
      pp = strField(color,"|",2);
      if (pp) vignette_RGB[1] = atoi(pp);
      pp = strField(color,"|",3);
      if (pp) vignette_RGB[2] = atoi(pp);
      signal_thread();                                                     //  trigger update thread
   }

   if (strEqu(event,"load")) {                                             //  load saved curve 
      splcurve_load(sd);
      Vign_curvedit(0);                                                    //  v.12.09
      signal_thread();
      return 0;
   }

   if (strEqu(event,"save")) {                                             //  save curve to file 
      splcurve_save(sd);
      return 0;
   }
   
   return 0;
}


//  get mouse position and set new center for vignette

void vign_mousefunc()                                                      //  mouse function               v.12.09
{
   if (! LMclick && ! Mdrag) return;
   LMclick = 0;
   
   vign_cx = Mxposn;                                                       //  new vignette center = mouse position
   vign_cy = Myposn;

   Mxdrag = Mydrag = 0;                                                    //  v.12.11

   signal_thread();                                                        //  trigger image update
   return;
}


//  this function is called when the curve is edited

void Vign_curvedit(int)
{
   int      ii = vignette_spc, jj;
   spldat   *sd = EFvignette.curves;

   sd->nap[ii] = sd->nap[0];                                               //  copy curve 0 to current curve
   for (jj = 0; jj < sd->nap[0]; jj++) {
      sd->apx[ii][jj] = sd->apx[0][jj];
      sd->apy[ii][jj] = sd->apy[0][jj];
   }
   for (jj = 0; jj < 1000; jj++)
      sd->yval[ii][jj] = sd->yval[0][jj];

   signal_thread();                                                        //  update image
   return;
}


//  thread function

void * Vign_thread(void *)
{
   void * Vign_wthread(void *arg);

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      for (int ii = 0; ii < Nwt; ii++)                                     //  start working threads
         start_wthread(Vign_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion
      
      CEF->Fmod = 1;
      mwpaint2(); 
   }

   return 0;                                                               //  not executed, stop g++ warning
}


//  working thread

void * Vign_wthread(void *arg)
{
   float       *pix1, *pix3;
   int         index, ii, kk, px, py, dist = 0;
   float       cx, cy, rad, radx, rady, f1, f2, xval, yval;
   float       red1, green1, blue1, red3, green3, blue3, cmax;
   spldat      *sd = EFvignette.curves;

   cx = vign_cx;                                                           //  vignette center (mouse)
   cy = vign_cy;

   index = *((int *) arg);

   for (py = index; py < E3pxm->hh; py += Nwt)                             //  loop all image pixels
   for (px = 0; px < E3pxm->ww; px++)
   {
      ii = py * E3pxm->ww + px;

      if (sa_stat == 3) {                                                  //  select area active 
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel is outside area
      }
      
      pix1 = PXMpix(E1pxm,px,py);                                          //  input pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel

      red1 = pix1[0];                                                      //  input RGB
      green1 = pix1[1];
      blue1 = pix1[2];

      radx = px - cx;                                                      //  distance from vignette center
      rady = py - cy;
      rad = sqrtf(radx*radx + rady*rady);                                  //  (px,py) distance from center
     
      xval = rad / vign_rad;                                               //  scale 0 to 1.0                  v.12.09
      kk = 999.0 * xval;                                                   //  scale 0 to 999
      if (kk > 999) kk = 999;                                              //  beyond radius                   v.12.10

      yval = sd->yval[1][kk];                                              //  brightness curve y-value 0 to 1.0
      if (yval > 1.0) yval = 1.0;
      yval = 2.0 * yval;                                                   //  0 to 2.0
      
      red3 = yval * red1;                                                  //  adjust brightness
      green3 = yval * green1;
      blue3 = yval * blue1;

      yval = sd->yval[2][kk];                                              //  color curve y-value 0 to 1.0
      if (yval > 1.0) yval = 1.0;
      f1 = yval;                                                           //  0 to 1.0   new color
      f2 = 1.0 - f1;                                                       //  1.0 to 0   old color

      red3 = f1 * vignette_RGB[0] + f2 * red3;                             //  mix input and vignette color
      green3 = f1 * vignette_RGB[1] + f2 * green3;
      blue3 = f1 * vignette_RGB[2] + f2 * blue3;

      if (sa_stat == 3 && dist < sa_blend) {                               //  select area is active,
         f1 = 1.0 * dist / sa_blend;                                       //    blend changes over sa_blend
         f2 = 1.0 - f1;
         red3 = f1 * red3 + f2 * red1;
         green3 = f1 * green3 + f2 * green1;
         blue3 = f1 * blue3 + f2 * blue1;
      }
      
      cmax = red3;                                                         //  detect overflow
      if (green3 > cmax) cmax = green3;
      if (blue3 > cmax) cmax = blue3;
      
      if (cmax > 255.9) {                                                  //  stop overflow
         red3 = red3 * 255.9 / cmax;
         green3 = green3 * 255.9 / cmax;
         blue3 = blue3 * 255.9 / cmax;
      }

      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }

   exit_wthread();
   return 0;                                                               //  not executed, stop g++ warning
}


/**************************************************************************/

//  match_color edit function
//  Adjust colors of image 2 to match the colors of image 1
//  using small selected areas in each image as the match standard.

void * match_color_thread(void *);
void   match_color_mousefunc();

float    match_color_RGB1[3];                                              //  image 1 base colors to match
float    match_color_RGB2[3];                                              //  image 2 target colors to match
int      match_color_radius = 10;                                          //  mouse radius
int      match_color_mode = 0;

editfunc    EFmatchcolor;


void m_match_color(GtkWidget *, const char *) 
{
   int    match_color_dialog_event(zdialog* zd, const char *event);
   
   cchar    *title = ZTX("Color Match Images");

   zfuncs::F1_help_topic = "match_colors";
   
   if (checkmenulock()) return;                                            //  test menu lock            v.12.09
   
/*
               Color Match Images
         1  [ 10 ]   mouse radius for color sample
         2  [Open]   image for source color
         3  click on image to get source color
         4  [Open]   image for target color
         5  click on image to set target color
                             [done] [cancel]
*/

   zdialog *zd = zdialog_new(title,mWin,Bdone,Bcancel,null);               //  match_color dialog
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=2");
   zdialog_add_widget(zd,"vbox","vb1","hb1",0,"homog|space=3");
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"homog|space=3");
   zdialog_add_widget(zd,"label","labn1","vb1","1");
   zdialog_add_widget(zd,"label","labn2","vb1","2");
   zdialog_add_widget(zd,"label","labn3","vb1","3");
   zdialog_add_widget(zd,"label","labn4","vb1","4");
   zdialog_add_widget(zd,"label","labn5","vb1","5");
   zdialog_add_widget(zd,"hbox","hbrad","vb2");
   zdialog_add_widget(zd,"spin","radius","hbrad","1|20|1|10","space=5");
   zdialog_add_widget(zd,"label","labrad","hbrad",ZTX("mouse radius for color sample"));
   zdialog_add_widget(zd,"hbox","hbop1","vb2");
   zdialog_add_widget(zd,"button","open1","hbop1",ZTX("Open"),"space=5");
   zdialog_add_widget(zd,"label","labop1","hbop1",ZTX("image for source color"));
   zdialog_add_widget(zd,"hbox","hbclik1","vb2");
   zdialog_add_widget(zd,"label","labclik1","hbclik1",ZTX("click on image to get source color"));
   zdialog_add_widget(zd,"hbox","hbop2","vb2");
   zdialog_add_widget(zd,"button","open2","hbop2",ZTX("Open"),"space=5");
   zdialog_add_widget(zd,"label","labop2","hbop2",ZTX("image to set matching color"));
   zdialog_add_widget(zd,"hbox","hbclik2","vb2");
   zdialog_add_widget(zd,"label","labclik2","hbclik2",ZTX("click on image to set matching color"));

   zdialog_stuff(zd,"radius",match_color_radius);                          //  remember last radius
   zdialog_help(zd,"match_colors");                                        //  zdialog help topic
   zdialog_run(zd,match_color_dialog_event);                               //  run dialog - parallel

   EFmatchcolor.funcname = "match-color";
   EFmatchcolor.Farea = 1;                                                 //  select area ignored
   EFmatchcolor.threadfunc = match_color_thread;
   EFmatchcolor.mousefunc = match_color_mousefunc;
   
   match_color_mode = 0;
   if (curr_file) {
      match_color_mode = 1;                                                //  image 1 ready to click
      takeMouse(match_color_mousefunc,0);                                  //  connect mouse function
   }

   return;
}


//  match_color dialog event and completion function

int match_color_dialog_event(zdialog *zd, const char *event)               //  match_color dialog event function
{
   int      err;
   
   if (zd->zstat)                                                          //  end dialog
   {
      if (match_color_mode == 4) {                                         //  edit was started
         if (zd->zstat == 1) edit_done();
         else edit_cancel();
      }
      freeMouse();
      zdialog_free(zd);
      match_color_mode = 0;
      return 1;
   }
   
   if (strEqu(event,"radius"))                                             //  set new mouse radius
      zdialog_fetch(zd,"radius",match_color_radius);
   
   if (strEqu(event,"open1"))                                              //  get image 1 for color source
   {
      match_color_mode = 0;
      err = f_open(null,0,0);
      if (! err) match_color_mode = 1;                                     //  image 1 ready to click
   }

   if (strEqu(event,"open2"))                                              //  get image 2 to set matching color
   {
      if (match_color_mode < 2) {
         zmessageACK(mWin,ZTX("select source image color first"));         //  check that RGB1 has been set
         return 1;
      }
      match_color_mode = 2;
      err = f_open(null,0,0);
      if (err) return 1;
      match_color_mode = 3;                                                //  image 2 ready to click
   }

   takeMouse(match_color_mousefunc,0);                                     //  reconnect mouse function
   return 1;
}


//  mouse function - click on image and get colors to match

void match_color_mousefunc()
{
   void  match_color_getRGB(int px, int py, float rgb[3]);

   int      px, py, mrect;

   if (match_color_mode < 1) return;                                       //  no image available yet
   
   mrect = 2 * match_color_radius;
   draw_mousearc(Mxposn,Myposn,mrect,mrect);
   
   if (LMclick)
   {
      LMclick = 0;
      px = Mxclick;
      py = Myclick;
   
      if (match_color_mode == 1 || match_color_mode == 2)                  //  image 1 ready to click
      {
         match_color_getRGB(px,py,match_color_RGB1);                       //  get RGB1 color
         match_color_mode = 2;
         return;
      }
      
      if (match_color_mode == 3 || match_color_mode == 4)                  //  image 2 ready to click
      {
         match_color_getRGB(px,py,match_color_RGB2);                       //  get RGB2 color

         if (match_color_mode == 4) {
            edit_cancel();                                                 //  if prior edit, cancel it
            takeMouse(match_color_mousefunc,0);                            //  reconnect mouse function
            match_color_mode = 3;
         }

         if (! edit_setup(EFmatchcolor)) return;                           //  setup edit - thread will launch
         match_color_mode = 4;                                             //  edit waiting for cancel or done
         signal_thread();                                                  //  update the target image
         return;
      }
   }

   return;
}


//  get the RGB averages for pixels within mouse radius

void match_color_getRGB(int px, int py, float rgb[3])
{
   int      rad1 = match_color_radius;
   int      rad2 = rad1 * rad1;
   int      rad, npix, qx, qy;
   float    red, green, blue;
   float    *pix1;
   PXM      *pxm;

   pxm = PXM_load(curr_file,1);                                            //  popup ACK if error     v.12.03
   if (! pxm) return;

   npix = 0;   
   red = green = blue = 0;

   for (qy = py-rad1; qy <= py+rad1; qy++)
   for (qx = px-rad1; qx <= px+rad1; qx++)
   {
      if (qx < 0 || qx > pxm->ww-1) continue;
      if (qy < 0 || qy > pxm->hh-1) continue;
      rad = (qx-px) * (qx-px) + (qy-py) * (qy-py);
      if (rad > rad2) continue;
      pix1 = PXMpix(pxm,qx,qy);
      red += pix1[0];
      green += pix1[1];
      blue += pix1[2];
      npix++;
   }
   
   rgb[0] = red / npix;
   rgb[1] = green / npix;
   rgb[2] = blue / npix;
   
   PXM_free(pxm);
   return;
}


//  thread function - start multiple working threads

void * match_color_thread(void *)
{
   void  * match_color_wthread(void *arg);                                 //  worker thread
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(match_color_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      CEF->Fmod = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, avoid warning
}


void * match_color_wthread(void *arg)                                      //  worker thread function
{
   int         index = *((int *) (arg));
   int         px, py;
   float       *pix3;
   float       Rred, Rgreen, Rblue;
   float       red, green, blue, cmax;
   
   Rred = match_color_RGB1[0] / match_color_RGB2[0];                       //  color adjustment ratios
   Rgreen = match_color_RGB1[1] / match_color_RGB2[1];
   Rblue = match_color_RGB1[2] / match_color_RGB2[2];

   for (py = index; py < E3pxm->hh; py += Nwt)                             //  loop all image pixels
   for (px = 0; px < E3pxm->ww; px++)
   {
      pix3 = PXMpix(E3pxm,px,py);
      
      red = pix3[0] * Rred;                                                //  adjust colors
      green = pix3[1] * Rgreen;
      blue = pix3[2] * Rblue;
      
      cmax = red;                                                          //  check for overflow
      if (green > cmax) cmax = green;
      if (blue > cmax) cmax = blue;

      if (cmax > 255.9) {                                                  //  fix overflow
         red = red * 255.9 / cmax;
         green = green * 255.9 / cmax;
         blue = blue * 255.9 / cmax;
      }      

      pix3[0] = red;
      pix3[1] = green;
      pix3[2] = blue;
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************/

//  CMYK menu function
//  Adjust Density (Brightness) and RGB/CMY levels using 0.01 OD units

editfunc    EFcmyk;                                                        //  edit function data
int         CMYKinputs[8];
int         CMYKbias = 0;
char        *CMYK_file = 0;

void m_CMYK(GtkWidget *, const char *)
{
   int    CMYK_dialog_event(zdialog *zd, cchar *event);
   void * CMYK_thread(void *);

   zfuncs::F1_help_topic = "CMYK";
   
   EFcmyk.funcname = "CMYK";                                               //  function name
   EFcmyk.Fprev = 1;                                                       //  use preview
   EFcmyk.Farea = 2;                                                       //  select area usable
   EFcmyk.threadfunc = CMYK_thread;                                        //  thread function
   if (! edit_setup(EFcmyk)) return;                                       //  setup edit
   
/***
    _________________________________________________
   |                                                 |
   |  [reset]   [x] Add standard bias                |
   |  +Brightness -Density [__|v]   Contrast [__|v]  |
   |  +Red -Cyan           [__|v]     Red    [__|v]  |
   |  +Green -Magenta      [__|v]     Green  [__|v]  |
   |  +Blue -Yellow        [__|v]     Blue   [__|x]  |
   |                                                 |
   |                  [open] [save] [done] [cancel]  |
   |_________________________________________________|
   
***/

   zdialog *zd = zdialog_new("CMYK",mWin,Bopen,Bsave,Bdone,Bcancel,null);  //  new dialog
   zdialog_help(zd,"CMYK");                                                //  add help topic to zdialog
   EFcmyk.zd = zd;                                                         //  add dialog to edit function

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","reset","hb1",Breset,"space=10");
   zdialog_add_widget(zd,"check","bias","hb1",ZTX("Add standard bias"),"space=5");
   zdialog_add_widget(zd,"hbox","hb2","dialog");
   zdialog_add_widget(zd,"vbox","vb1","hb2",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb2","hb2",0,"homog|space=5");
   zdialog_add_widget(zd,"label","space","hb2",0,"space=8");
   zdialog_add_widget(zd,"vbox","vb3","hb2",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb4","hb2",0,"homog|space=5");
   zdialog_add_widget(zd,"button","buBriteDens","vb1",ZTX("+Brightness -Density"));
   zdialog_add_widget(zd,"button","buRedDens","vb1",ZTX("+Red -Cyan"));
   zdialog_add_widget(zd,"button","buGreenDens","vb1",ZTX("+Green -Magenta"));
   zdialog_add_widget(zd,"button","buBlueDens","vb1",ZTX("+Blue -Yellow"));
   zdialog_add_widget(zd,"spin","spBriteDens","vb2","-99|+99|1|0");
   zdialog_add_widget(zd,"spin","spRedDens","vb2","-99|+99|1|0");
   zdialog_add_widget(zd,"spin","spGreenDens","vb2","-99|+99|1|0");
   zdialog_add_widget(zd,"spin","spBlueDens","vb2","-99|+99|1|0");
   zdialog_add_widget(zd,"button","buContrast","vb3",ZTX("Contrast"));
   zdialog_add_widget(zd,"button","buRedCon","vb3",ZTX("Red"));
   zdialog_add_widget(zd,"button","buGreenCon","vb3",ZTX("Green"));
   zdialog_add_widget(zd,"button","buBlueCon","vb3",ZTX("Blue"));
   zdialog_add_widget(zd,"spin","spContrast","vb4","-99|+99|1|0");
   zdialog_add_widget(zd,"spin","spRedCon","vb4","-99|+99|1|0");
   zdialog_add_widget(zd,"spin","spGreenCon","vb4","-99|+99|1|0");
   zdialog_add_widget(zd,"spin","spBlueCon","vb4","-99|+99|1|0");

   if (! CMYK_file) {                                                      //  default file
      CMYK_file = zmalloc(200,"CMYK");                                     //  /.../.fotoxx/CMYK_parameters
      snprintf(CMYK_file,200,"%s/CMYK_parameters",get_zuserdir());
   }

   zdialog_run(zd,CMYK_dialog_event,"save");                               //  run dialog - parallel
   return;
}


//  CMYK dialog event and completion function

int CMYK_dialog_event(zdialog *zd, cchar *event)                           //  CMYK dialog event function
{
   void  CMYK_open();
   void  CMYK_save();
   
   int      bias;
   
   if (zd->zstat)
   {
      if (zd->zstat == 1) {                                                //  open file
         zd->zstat = 0;                                                    //  keep dialog active
         CMYK_open();                                                      //  read file

         zdialog_stuff(zd,"bias",CMYKbias);                                //  set all data
         zdialog_stuff(zd,"spBriteDens",CMYKinputs[0]);
         zdialog_stuff(zd,"spRedDens",CMYKinputs[1]);
         zdialog_stuff(zd,"spGreenDens",CMYKinputs[2]);
         zdialog_stuff(zd,"spBlueDens",CMYKinputs[3]);
         zdialog_stuff(zd,"spContrast",CMYKinputs[4]);
         zdialog_stuff(zd,"spRedCon",CMYKinputs[5]);
         zdialog_stuff(zd,"spGreenCon",CMYKinputs[6]);
         zdialog_stuff(zd,"spBlueCon",CMYKinputs[7]);
      }

      else if (zd->zstat == 2) {                                           //  save file
         zd->zstat = 0;                                                    //  keep dialog active
         CMYK_save();                                                      //  write file
      }

      else if (zd->zstat == 3) edit_done();                                //  done

      else edit_cancel();                                                  //  cancel or destroy

      return 0;
   }

   if (strEqu(event,"bias"))                                               //  bias was changed
   {
      zdialog_fetch(zd,"bias",bias);
      if (bias != CMYKbias) 
      {
         if (bias) {                                                       //  off to on
            CMYKinputs[0] += 72;
            CMYKinputs[2] += 32;
            CMYKinputs[3] += 32;
            zdialog_set_limits(zd,"spBriteDens",-27,+171);
            zdialog_set_limits(zd,"spGreenDens",-67,+131);
            zdialog_set_limits(zd,"spBlueDens",-67,+131);
         }
         else {                                                            //  on to off
            CMYKinputs[0] -= 72;
            CMYKinputs[2] -= 32;
            CMYKinputs[3] -= 32;
            zdialog_set_limits(zd,"spBriteDens",-99,+99);
            zdialog_set_limits(zd,"spGreenDens",-99,+99);
            zdialog_set_limits(zd,"spBlueDens",-99,+99);
         }
         zdialog_stuff(zd,"spBriteDens",CMYKinputs[0]);
         zdialog_stuff(zd,"spGreenDens",CMYKinputs[2]);
         zdialog_stuff(zd,"spBlueDens",CMYKinputs[3]);
      }
      CMYKbias = bias;
      return 1;
   }
   
   if (strEqu(event,"reset"))                                              //  reset dialog controls
   {
      CMYKinputs[0] = CMYKbias * 72;
      CMYKinputs[1] = CMYKbias * 0;
      CMYKinputs[2] = CMYKbias * 32;
      CMYKinputs[3] = CMYKbias * 32;
      CMYKinputs[4] = 0;
      CMYKinputs[5] = 0;
      CMYKinputs[6] = 0;
      CMYKinputs[7] = 0;
      zdialog_stuff(zd,"spBriteDens",CMYKinputs[0]);
      zdialog_stuff(zd,"spRedDens",CMYKinputs[1]);
      zdialog_stuff(zd,"spGreenDens",CMYKinputs[2]);
      zdialog_stuff(zd,"spBlueDens",CMYKinputs[3]);
      zdialog_stuff(zd,"spContrast",CMYKinputs[4]);
      zdialog_stuff(zd,"spRedCon",CMYKinputs[5]);
      zdialog_stuff(zd,"spGreenCon",CMYKinputs[6]);
      zdialog_stuff(zd,"spBlueCon",CMYKinputs[7]);
   }

   zdialog_fetch(zd,"spBriteDens",CMYKinputs[0]);                          //  get all inputs
   zdialog_fetch(zd,"spRedDens",CMYKinputs[1]);
   zdialog_fetch(zd,"spGreenDens",CMYKinputs[2]);
   zdialog_fetch(zd,"spBlueDens",CMYKinputs[3]);
   zdialog_fetch(zd,"spContrast",CMYKinputs[4]);
   zdialog_fetch(zd,"spRedCon",CMYKinputs[5]);
   zdialog_fetch(zd,"spGreenCon",CMYKinputs[6]);
   zdialog_fetch(zd,"spBlueCon",CMYKinputs[7]);

   signal_thread();                                                        //  trigger update thread
   wait_thread_idle();                                                     //  wait until done (optional)
   return 1;
}


//  load CMYK parameters from a previously saved file

void CMYK_open()
{
   int  CMYK_open_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;
   
   zd = zdialog_new(ZTX("Load CMYK parameters"),mWin,Bopen,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","lab1","hb1",ZTX("File:"),"space=3");
   zdialog_add_widget(zd,"entry","file","hb1",0,"expand|scc=40");
   zdialog_add_widget(zd,"button","browse","hb1",Bbrowse,"space=5");
   
   zdialog_stuff(zd,"file",CMYK_file);
   
   zdialog_run(zd,CMYK_open_dialog_event);
   zdialog_wait(zd);
   return;
}


int CMYK_open_dialog_event(zdialog *zd, cchar *event)
{
   char     *pp, file[200];
   FILE     *fid = 0;
   int      zstat, nn;
   
   if (strEqu(event,"browse")) {
      pp = zgetfile1(ZTX("CMYK parameters file"),"open",CMYK_file);
      if (! pp) return 0;
      zdialog_stuff(zd,"file",pp);
      zfree(pp);
   }
   
   zstat = zd->zstat;                                                      //  completion button
   
   if (zstat)
   {
      if (zstat == 1)                                                      //  open
      {
         zdialog_fetch(zd,"file",file,200);                                //  get file from dialog
         if (CMYK_file) zfree(CMYK_file);
         CMYK_file = strdupz(file,0,"CMYK");

         fid = fopen(file,"r");                                            //  open file
         if (! fid) {
            zmessageACK(mWin,ZTX("file not found"));
            return 1;
         }
         
         nn = fscanf(fid,"CMYK parameters");                               //  read header
         
         nn = fscanf(fid,"%4d %4d %4d %4d %4d %4d %4d %4d %4d",            //  read data
                           &CMYKbias, 
                           &CMYKinputs[0], &CMYKinputs[1], &CMYKinputs[2],
                           &CMYKinputs[3], &CMYKinputs[4], &CMYKinputs[5], 
                           &CMYKinputs[6], &CMYKinputs[7]);

         if (nn != 9) zmessageACK(mWin,"file format error");

         fclose(fid);
         zdialog_free(zd);
      }
      
      else zdialog_free(zd);                                               //  cancel or [x]
   }

   return 1;
}


//  save CMYK parameters to a file

void CMYK_save()
{
   int  CMYK_save_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;
   
   zd = zdialog_new(ZTX("Save CMYK parameters"),mWin,Bsave,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","lab1","hb1",ZTX("File:"),"space=3");
   zdialog_add_widget(zd,"entry","file","hb1",0,"expand|scc=40");
   zdialog_add_widget(zd,"button","browse","hb1",Bbrowse,"space=5");
   
   zdialog_stuff(zd,"file",CMYK_file);
   
   zdialog_run(zd,CMYK_save_dialog_event);
   zdialog_wait(zd);

   return;
}


int CMYK_save_dialog_event(zdialog *zd, cchar *event)
{
   char     *pp, file[200];
   FILE     *fid = 0;
   int      zstat;
   
   if (strEqu(event,"browse")) {
      pp = zgetfile1(ZTX("CMYK parameters file"),"save",CMYK_file);
      if (! pp) return 0;
      zdialog_stuff(zd,"file",pp);
      zfree(pp);
   }
   
   zstat = zd->zstat;                                                      //  completion button
   
   if (zstat)
   {
      if (zstat == 1)                                                      //  save
      {
         zdialog_fetch(zd,"file",file,200);                                //  get file from dialog
         if (CMYK_file) zfree(CMYK_file);
         CMYK_file = strdupz(file,0,"CMYK");

         fid = fopen(file,"w");                                            //  open file
         if (! fid) {
            zmessageACK(mWin,strerror(errno));
            return 1;
         }
         
         fprintf(fid,"CMYK parameters \n");                                //  write headers

         fprintf(fid,"%4d %4d %4d %4d %4d %4d %4d %4d %4d ",               //  write data
                              CMYKbias, 
                              CMYKinputs[0], CMYKinputs[1], CMYKinputs[2],
                              CMYKinputs[3], CMYKinputs[4], CMYKinputs[5], 
                              CMYKinputs[6], CMYKinputs[7]);
         fprintf(fid,"\n");
         fclose(fid);
         zdialog_free(zd);
      }

      else zdialog_free(zd);                                               //  cancel or [x]
   }

   return 1;
}


//  thread function - multiple working threads to update image

void * CMYK_thread(void *)
{
   void  * CMYK_wthread(void *arg);                                        //  worker thread

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      if (sa_stat == 3) Fbusy_goal = sa_Npixel;                            //  set up progress monitor
      else  Fbusy_goal = E3pxm->ww * E3pxm->hh;
      Fbusy_done = 0;

      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(CMYK_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      Fbusy_goal = Fbusy_done = 0;
      CEF->Fmod = 1;                                                       //  image modified
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop warning
}


void * CMYK_wthread(void *arg)                                             //  worker thread function
{
   float       Du, Ru, Gu, Bu, Rd, Gd, Bd;
   float       con, conR, conG, conB;
   float       R1, G1, B1, R3, G3, B3;

   int         index = *((int *) (arg));
   int         px, py, ii, dist = 0;
   float       *pix1, *pix3;
   float       cmax, F, f1, f2;
   float       c1 = 100.0 / 256.0;
   float       c2 = 2.0 + log10(1.0/c1);
   float       c3 = 1.0 / 256.0;
   
   #define ODfunc(rgb) (2.0 - log10(c1 * rgb))                             //  RGB units (1-256) to OD units
   #define RGBfunc(od) (pow(10,(c2 - od)))                                 //  OD units to RGB units
   
   Du = -0.01 * CMYKinputs[0];                                             //  convert inputs from cc to OD units
   Ru = -0.01 * CMYKinputs[1];                                             //  range is -0.99 to +0.99
   Gu = -0.01 * CMYKinputs[2];
   Bu = -0.01 * CMYKinputs[3];
   
   if (CMYKbias) {                                                         //  remove bias
      Du += 0.72;
      Gu += 0.32;
      Bu += 0.32;
   }
   
   con = -0.01 * CMYKinputs[4];
   conR = -0.01 * CMYKinputs[5];
   conG = -0.01 * CMYKinputs[6];
   conB = -0.01 * CMYKinputs[7];
   
   for (py = index; py < E3pxm->hh; py += Nwt)                             //  loop all image pixels
   for (px = 0; px < E3pxm->ww; px++)
   {
      if (sa_stat == 3) {                                                  //  select area active
         ii = py * E3pxm->ww + px;
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  outside area
      }
      
      pix1 = PXMpix(E1pxm,px,py);                                          //  input pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel
      
      R1 = pix1[0] + 0.1;                                                  //  input RGB values, 0-255
      G1 = pix1[1] + 0.1;                                                  //  (avoid 0)
      B1 = pix1[2] + 0.1;
      
      Rd = ODfunc(R1);                                                     //  convert to OD units
      Gd = ODfunc(G1);
      Bd = ODfunc(B1);
      
      Rd += Du + Ru;                                                       //  add user OD increments
      Gd += Du + Gu;
      Bd += Du + Bu;
      
      if (con) {                                                           //  overall contrast
         Rd += con * c3 * (R1-128);
         Gd += con * c3 * (G1-128);
         Bd += con * c3 * (B1-128);
      }

      if (conR)                                                            //  single color contrast
         Rd += conR * c3 * (R1-128);
      if (conG) 
         Gd += conG * c3 * (G1-128);
      if (conB) 
         Bd += conB * c3 * (B1-128);
      
      R3 = RGBfunc(Rd);                                                    //  convert back to RGB units
      G3 = RGBfunc(Gd);
      B3 = RGBfunc(Bd);
      
      if (R3 < 0) R3 = 0;                                                  //  stop underflow
      if (G3 < 0) G3 = 0;
      if (B3 < 0) B3 = 0;

      if (R3 > 255.9 || G3 > 255.9 || B3 > 255.9) {                        //  stop overflow
         cmax = R3;
         if (G3 > cmax) cmax = G3;
         if (B3 > cmax) cmax = B3;
         F = 255.9 / cmax;
         R3 *= F;
         G3 *= F;
         B3 *= F;
      }

      if (sa_stat == 3 && dist < sa_blend) {                               //  select area is active,
         f1 = 1.0 * dist / sa_blend;                                       //    blend changes over sa_blend
         f2 = 1.0 - f1;
         R3 = f1 * R3 + f2 * R1;
         G3 = f1 * G3 + f2 * G1;
         B3 = f1 * B3 + f2 * B1;
      }
      
      pix3[0] = R3;                                                        //  output RGB values
      pix3[1] = G3;
      pix3[2] = B3;
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************/

//  Show RGB values for 1-9 pixels selected with mouse-clicks.
//  Revise RGB values for the selected points and then revise all other 
//  pixels to match, using a distance-weighted average of the selected pixels.

void  RGBR_dialog();
void  RGBR_mousefunc();
void  RGBR_stuff();
void * RGBR_thread(void *);

zdialog  *RGBRzd; 
int      RGBRpixel[9][2];                                                  //  last 9 pixel locations clicked
float    RGBRval1[9][3];                                                   //  original RGB values, 0 to 255.9
float    RGBRval3[9][3];                                                   //  revised RGB values, 0 to 255.9
int      RGBRmetric = 1;                                                   //  1/2/3 = /RGB/EV/OD
int      RGBRdelta = 0;                                                    //  flag, delta mode
int      RGBRrestart;                                                      //  flag, restart dialog
int      RGBRnpix;                                                         //  no. of selected pixels
double   RGBRtime;                                                         //  last click time
int      RGBRblend;                                                        //  blend factor, 1 to 100
editfunc EFrgbrevise;                                                      //  edit function parameters

#define RGB2EV(rgb) (log2(rgb)-7)                                          //  RGB 0.1 to 255.9 >> EV -10 to +1
#define EV2RGB(ev) (pow(2,ev+7))                                           //  inverse
#define RGB2OD(rgb) (2-log10(100*rgb/256))                                 //  RGB 0.1 to 255.9 >> OD 3.4 to 0.000017
#define OD2RGB(od) (pow(10,2.40824-od))                                    //  inverse


void m_revise_RGB(GtkWidget *, cchar *menu)
{
   int   RGBR_event(zdialog *zd, cchar *event);

   GtkWidget   *widget;
   cchar       *limits;
   zdialog     *zd;
   cchar       *mess = ZTX("Click image to select pixels.");

   PangoFontDescription    *widgetfont;
   widgetfont = pango_font_description_from_string("Monospace 8");         //  small monospace font for widgets

   zfuncs::F1_help_topic = "revise_RGB";
   
   EFrgbrevise.funcname = "revise_RGB";                                    //  setup edit function
   EFrgbrevise.threadfunc = RGBR_thread;
   EFrgbrevise.mousefunc = RGBR_mousefunc;
   EFrgbrevise.Farea = 1;
   if (! edit_setup(EFrgbrevise)) return;

   RGBRblend = 20;                                                         //  default slider value
   RGBRnpix = 0;                                                           //  no pixels selected yet
   
/***
   
      Click image to select pixels.
      [x] delta
      Metric: (o) RGB  (o) EV  (o) OD
      Pixel          Red      Green    Blue
      A xxxx xxxx   [xxxx|v] [xxxx|v] [xxxx|v]                             //  spin buttons for RGB/EV/OD values
      B xxxx xxxx   [xxxx|v] [xxxx|v] [xxxx|v]
      C xxxx xxxx   [xxxx|v] [xxxx|v] [xxxx|v] 
      D xxxx xxxx   [xxxx|v] [xxxx|v] [xxxx|v] 
      E xxxx xxxx   [xxxx|v] [xxxx|v] [xxxx|v] 
      F xxxx xxxx   [xxxx|v] [xxxx|v] [xxxx|v] 
      G xxxx xxxx   [xxxx|v] [xxxx|v] [xxxx|v] 
      H xxxx xxxx   [xxxx|v] [xxxx|v] [xxxx|v] 
      I xxxx xxxx   [xxxx|v] [xxxx|v] [xxxx|v] 
      Blend  ==============[]================

                     [reset] [done] [cancel]
***/
   
restart:

   zd = zdialog_new(ZTX("Revise RGB"),mWin,Breset,Bdone,Bcancel,null);
   EFrgbrevise.zd = zd;
   RGBRzd = zd;

   zdialog_add_widget(zd,"hbox","hbmess","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labmess","hbmess",mess,"space=5");

   zdialog_add_widget(zd,"hbox","hbmym","dialog");
   zdialog_add_widget(zd,"check","delta","hbmym","delta","space=5");
   zdialog_stuff(zd,"delta",RGBRdelta);

   zdialog_add_widget(zd,"hbox","hbmetr","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labmetr","hbmetr",ZTX("Metric:"),"space=5");
   zdialog_add_widget(zd,"radio","radRGB","hbmetr","RGB","space=3");
   zdialog_add_widget(zd,"radio","radEV","hbmetr","EV","space=3");
   zdialog_add_widget(zd,"radio","radOD","hbmetr","OD","space=3");
   
   if (RGBRmetric == 1) zdialog_stuff(zd,"radRGB",1);                      //  set current metric
   if (RGBRmetric == 2) zdialog_stuff(zd,"radEV",1);
   if (RGBRmetric == 3) zdialog_stuff(zd,"radOD",1);

   zdialog_add_widget(zd,"hbox","hbdata","dialog");
   zdialog_add_widget(zd,"vbox","vbpix","hbdata",0,"space=3|homog");       //  vbox for pixel locations
   zdialog_add_widget(zd,"hbox","hbpix","vbpix");
   zdialog_add_widget(zd,"label","labpix","hbpix","Pixel");                //  header
   
   char hbpixx[8] = "hbpixx", pixx[8] = "pixx";

   for (int ii = 1; ii < 10; ii++) {                                       //  add labels for pixel locations
      hbpixx[5] = '0' + ii;
      pixx[3] = '0' + ii;
      zdialog_add_widget(zd,"hbox",hbpixx,"vbpix");                        //  add hbox "hbpix1" to "hbpix9"
      zdialog_add_widget(zd,"label",pixx,hbpixx);                          //  add label "pix1" to "pix9"
      widget = zdialog_widget(zd,pixx);
      gtk_widget_modify_font(widget,widgetfont);                           //  use monofont
   }

   if (RGBRmetric == 1) limits = "-255.9|255.9|0.1|0.0";                   //  metric = RGB
   else if (RGBRmetric == 2) limits = "-8|8|0.01|0.0";                     //           EV
   else limits = "-3|3|0.002|0.0";                                         //           OD

   zdialog_add_widget(zd,"vbox","vbdat","hbdata",0,"space=3|homog");       //  vbox for pixel RGB values
   zdialog_add_widget(zd,"hbox","hbrgb","vbdat",0,"homog");
   zdialog_add_widget(zd,"label","labr","hbrgb",Bred);
   zdialog_add_widget(zd,"label","labg","hbrgb",Bgreen);
   zdialog_add_widget(zd,"label","labb","hbrgb",Bblue);

   char hbdatx[8] = "hbdatx", redx[8] = "redx", greenx[8] = "greenx", bluex[8] = "bluex";

   for (int ii = 1; ii < 10; ii++) {
      hbdatx[5] = '0' + ii;
      redx[3] = '0' + ii;
      greenx[5] = '0' + ii;
      bluex[4] = '0' + ii;
      zdialog_add_widget(zd,"hbox",hbdatx,"vbdat");                        //  add hbox "hbdat1" to "hbdat9"
      zdialog_add_widget(zd,"spin",redx,hbdatx,limits,"space=3");          //  add spin buttons for "red1" to "red9" etc.
      zdialog_add_widget(zd,"spin",greenx,hbdatx,limits,"space=3");
      zdialog_add_widget(zd,"spin",bluex,hbdatx,limits,"space=3");
      widget = zdialog_widget(zd,redx);
      gtk_widget_modify_font(widget,widgetfont);                           //  use monofont
      widget = zdialog_widget(zd,greenx);
      gtk_widget_modify_font(widget,widgetfont);
      widget = zdialog_widget(zd,bluex);
      gtk_widget_modify_font(widget,widgetfont);
   }
   
   zdialog_add_widget(zd,"hbox","hbsoft","dialog","space=5");
   zdialog_add_widget(zd,"label","labsoft","hbsoft",ZTX("Blend"),"space=3");
   zdialog_add_widget(zd,"hscale","blend","hbsoft","0|100|1|20","space=5|expand");
   zdialog_add_widget(zd,"label","softval","hbsoft","20");

   RGBR_stuff();                                                           //  stuff current pixels (if restart)

   takeMouse(RGBR_mousefunc,dragcursor);                                   //  connect mouse function
   zdialog_help(zd,"revise_RGB");                                          //  zdialog help topic 
   zdialog_run(zd,RGBR_event,"save");                                      //  run dialog
   zdialog_wait(zd);                                                       //  wait for completion
   if (RGBRrestart) goto restart;                                          //  restart dialog

   return;
}


//  dialog event function

int RGBR_event(zdialog *zd, cchar *event)
{
   int         button;
   int         ii, jj;
   char        text[8];
   float       val1, val3;
   
   if (zd->zstat) 
   {
      if (zd->zstat == 1) {                                                //  Reset
         zd->zstat = 0;                                                    //  keep dialog active
         RGBRnpix = 0;                                                     //  no pixels selected yet
         RGBR_stuff();                                                     //  clear dialog
         edit_reset();                                                     //  reset edits 
         return 0;
      }
      else if (zd->zstat == 2) edit_done();
      else edit_cancel();                                                  //  Cancel or [x]
      freeMouse();                                                         //  disconnect mouse function
      zdialog_free(zd);                                                    //  kill dialog
      RGBRzd = 0;
      RGBRrestart = 0;
      erase_toptext(101);                                                  //  erase pixel labels from image
      return 1;
   }
   
   if (strEqu(event,"focus"))                                              //  toggle mouse capture
      takeMouse(RGBR_mousefunc,dragcursor);                                //  connect mouse function
   
   if (strEqu(event,"blend")) {                                            //  new blend factor
      zdialog_fetch(zd,"blend",RGBRblend);
      sprintf(text,"%d",RGBRblend);                                        //  numeric feedback
      zdialog_stuff(zd,"softval",text);
      signal_thread();
      return 1;
   }

   if (strnEqu(event,"rad",3))                                             //  metric was changed
   {
      if (strEqu(event,"radRGB")) {                                        //  RGB
         zdialog_fetch(zd,event,button);
         if (button) RGBRmetric = 1;
      }
         
      if (strEqu(event,"radEV")) {                                         //  EV
         zdialog_fetch(zd,event,button);
         if (button) RGBRmetric = 2;
      }
         
      if (strEqu(event,"radOD")) {                                         //  OD
         zdialog_fetch(zd,event,button);
         if (button) RGBRmetric = 3;
      }

      if (button) {      
         freeMouse();                                                      //  restart dialog with new limits
         zdialog_free(zd);
         RGBRzd = 0;
         RGBRrestart = 1;
      }

      return 1;
   }

   if (strEqu(event,"delta")) {                                            //  change absolute/delta mode
      zdialog_fetch(zd,"delta",RGBRdelta);
      freeMouse();                                                         //  restart dialog with new limits
      zdialog_free(zd);
      RGBRzd = 0;
      RGBRrestart = 1;
      return 1;
   }
   
   ii = -1;                                                                //  no RGB change yet

   if (strnEqu(event,"red",3)) {                                           //  red1 - red9 was changed
      ii = event[3] - '1';                                                 //  pixel index 0-8
      jj = 0;                                                              //  color = red
   }

   if (strnEqu(event,"green",5)) {                                         //  green1 - green9
      ii = event[5] - '1';
      jj = 1;
   }

   if (strnEqu(event,"blue",4)) {                                          //  blue1 - blue9
      ii = event[4] - '1';
      jj = 2;
   }
   
   if (ii >= 0 && ii < 9)                                                  //  RGB value was revised
   {
      val1 = RGBRval1[ii][jj];                                             //  original pixel RGB value
      if (RGBRmetric == 2) val1 = RGB2EV(val1);                            //  convert to EV or OD units
      if (RGBRmetric == 3) val1 = RGB2OD(val1);

      zdialog_fetch(zd,event,val3);                                        //  revised RGB/EV/OD value from dialog
      if (RGBRdelta) val3 += val1;                                         //  if delta mode, make absolute

      if (RGBRmetric == 2) val3 = EV2RGB(val3);                            //  convert EV/OD to RGB value
      if (RGBRmetric == 3) val3 = OD2RGB(val3);

      if (fabsf(RGBRval3[ii][jj] - val3) < 0.001) return 1;                //  ignore re-entry after change

      if (val3 < 0.1) val3 = 0.1;                                          //  limit RGB within 0.1 to 255.9
      if (val3 > 255.9) val3 = 255.9;

      RGBRval3[ii][jj] = val3;                                             //  new RGB value for pixel
      
      if (RGBRmetric == 2) val3 = RGB2EV(val3);                            //  convert RGB to EV/OD units
      if (RGBRmetric == 3) val3 = RGB2OD(val3);

      if (RGBRdelta) val3 -= val1;                                         //  absolute back to relative
      zdialog_stuff(zd,event,val3);                                        //  limited value back to dialog
      
      signal_thread();                                                     //  signal thread to update image
   }

   return 1;
}


//  mouse function

void RGBR_mousefunc()                                                      //  mouse function
{
   int         ii;
   float       *pix3;

   if (! LMclick) return;
   LMclick = 0;

   RGBRtime = get_seconds();                                               //  mark time of pixel click
   
   if (RGBRnpix == 9) {                                                    //  if table is full (9 entries)
      for (ii = 1; ii < 9; ii++) {                                         //    move positions 1-8 up 
         RGBRpixel[ii-1][0] = RGBRpixel[ii][0];                            //      to fill positions 0-7
         RGBRpixel[ii-1][1] = RGBRpixel[ii][1];
      }
      RGBRnpix = 8;                                                        //  count is now 8 entries
   }
   
   ii = RGBRnpix;                                                          //  next table position to fill
   RGBRpixel[ii][0] = Mxclick;                                             //  newest pixel
   RGBRpixel[ii][1] = Myclick;
   
   pix3 = PXMpix(E3pxm,Mxclick,Myclick);                                   //  get initial RGB values from
   RGBRval1[ii][0] = RGBRval3[ii][0] = pix3[0];                            //    modified image E3
   RGBRval1[ii][1] = RGBRval3[ii][1] = pix3[1];
   RGBRval1[ii][2] = RGBRval3[ii][2] = pix3[2];

   RGBRnpix++;                                                             //  up pixel count
   RGBR_stuff();                                                           //  stuff pixels and values into dialog
   return;
}


//  stuff dialog with current pixels and their RGB/EV/OD values

void RGBR_stuff()
{
   static char    lab1[9][4] = { "A", "B", "C", "D", "E", "F", "G", "H", "I" };
   static char    lab2[9][4] = { " A ", " B ", " C ", " D ", " E ", " F ", " G ", " H ", " I " };

   int         px, py;
   float       red1, green1, blue1, red3, green3, blue3;
   char        text[100], pixx[8] = "pixx";
   char        redx[8] = "redx", greenx[8] = "greenx", bluex[8] = "bluex";
   zdialog     *zd = RGBRzd;
   
   erase_toptext(101);                                                     //  erase prior labels from image
   
   for (int ii = 0; ii < 9; ii++)                                          //  loop slots 0-8
   {
      pixx[3] = '1' + ii;                                                  //  widget names "pix1" ... "pix9"
      redx[3] = '1' + ii;                                                  //  widget names "red1" ... "red9"
      greenx[5] = '1' + ii;
      bluex[4] = '1' + ii;

      px = RGBRpixel[ii][0];                                               //  next pixel to report
      py = RGBRpixel[ii][1];
      if (ii >= RGBRnpix) {                                                //  > last pixel selected
         zdialog_stuff(zd,pixx,"");
         zdialog_stuff(zd,redx,0);                                         //  blank pixel and zero values
         zdialog_stuff(zd,greenx,0);
         zdialog_stuff(zd,bluex,0);
         continue;
      }

      sprintf(text,"%s %4d %4d",lab1[ii],px,py);                           //  format pixel "A nnnn nnnn"
      zdialog_stuff(zd,pixx,text);                                         //  pixel >> widget
      
      add_toptext(101,px,py,lab2[ii],"Sans 8");                            //  paint label on image at pixel
      
      red1 = RGBRval1[ii][0];                                              //  original RGB values for pixel
      green1 = RGBRval1[ii][1];
      blue1 = RGBRval1[ii][2];
      red3 = RGBRval3[ii][0];                                              //  revised RGB values
      green3 = RGBRval3[ii][1];
      blue3 = RGBRval3[ii][2];
      
      if (RGBRmetric == 2) {                                               //  convert to EV units if needed
         red1 = RGB2EV(red1);
         green1 = RGB2EV(green1);
         blue1 = RGB2EV(blue1);
         red3 = RGB2EV(red3);
         green3 = RGB2EV(green3);
         blue3 = RGB2EV(blue3);
      }
      
      if (RGBRmetric == 3) {                                               //  or OD units
         red1 = RGB2OD(red1);
         green1 = RGB2OD(green1);
         blue1 = RGB2OD(blue1);
         red3 = RGB2OD(red3);
         green3 = RGB2OD(green3);
         blue3 = RGB2OD(blue3);
      }
      
      if (RGBRdelta) {                                                     //  dialog is delta mode
         red3 -= red1;
         green3 -= green1;
         blue3 -= blue1;
      }
      
      zdialog_stuff(zd,redx,red3);
      zdialog_stuff(zd,greenx,green3);
      zdialog_stuff(zd,bluex,blue3);
   }

   mwpaint2();                                                             //  refresh window
   return;
}


//  thread function - multiple working threads to update image

void * RGBR_thread(void *)
{
   void  * RGBR_wthread(void *arg);                                        //  worker thread

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      zsleep(0.3);                                                         //  more time for dialog controls
      
      if (RGBRnpix < 1) continue;                                          //  must have 1+ pixels in table
      
      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(RGBR_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      CEF->Fmod = 1;                                                       //  image modified
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop warning
}


void * RGBR_wthread(void *arg)                                             //  worker thread function
{
   int         index = *((int *) (arg));
   int         px1, py1, px2, py2, ii;
   float       *pix1, *pix3;
   float       dist[9], weight[9], sumdist;
   float       blend, delta, red, green, blue, max;
   
   blend = E1pxm->ww;
   if (E1pxm->hh > E1pxm->ww) blend = E1pxm->hh;
   blend = blend * blend;
   blend = 0.0002 * RGBRblend * RGBRblend * blend + 100;                   //  100 to 2 * E1pxm->ww**2
   
   for (py1 = index; py1 < E3pxm->hh; py1 += Nwt)                          //  loop all image pixels
   for (px1 = 0; px1 < E3pxm->ww; px1++)
   {
      for (sumdist = ii = 0; ii < RGBRnpix; ii++)                          //  compute weight of each revision point
      {
         px2 = RGBRpixel[ii][0];
         py2 = RGBRpixel[ii][1];
         dist[ii] = (px1-px2)*(px1-px2) + (py1-py2)*(py1-py2);             //  distance (px1,py1) to (px2,py2)
         dist[ii] = 1.0 / (dist[ii] + blend);                              //  blend reduces peaks at revision points
         sumdist += dist[ii];                                              //  sum inverse distances
      }
      
      for (ii = 0; ii < RGBRnpix; ii++)                                    //  weight of each point
         weight[ii] = dist[ii] / sumdist;

      pix1 = PXMpix(E1pxm,px1,py1);                                        //  input pixel
      pix3 = PXMpix(E3pxm,px1,py1);                                        //  output pixel
      
      red = pix1[0];
      green = pix1[1];
      blue = pix1[2];
      
      for (ii = 0; ii < RGBRnpix; ii++) {                                  //  apply weighted color changes
         delta = RGBRval3[ii][0] - RGBRval1[ii][0];                        //    to each color
         red += weight[ii] * delta;
         delta = RGBRval3[ii][1] - RGBRval1[ii][1];
         green += weight[ii] * delta;
         delta = RGBRval3[ii][2] - RGBRval1[ii][2];
         blue += weight[ii] * delta;
      }
      
      max = red;
      if (green > max) max = green;
      if (blue > max) max = blue;

      if (max > 255.9) {                                                   //  stop overflow/underflow
         red = red * 255.9 / max;
         green = green * 255.9 / max;
         blue = blue * 255.9 / max;
      }
      
      if (red < 0) red = 0;
      if (green < 0) green = 0;
      if (blue < 0) blue = 0;
      
      pix3[0] = red;
      pix3[1] = green;
      pix3[2] = blue;
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************/

//  ramp brightness across image, vertical and horizontal gradients

editfunc    EFbrramp;
int         brramp_spc;                                                    //  current spline curves (2 at a time)

void m_brightramp(GtkWidget *, cchar *)                                    //  upgrade for 4 colors
{
   int    brramp_dialog_event(zdialog* zd, cchar *event);
   void   brramp_curvedit(int spc);
   void * brramp_thread(void *);
   void   brramp_init();

   zfuncs::F1_help_topic = "brightness_ramp";

   EFbrramp.funcname = "bright-ramp";
   EFbrramp.Fprev = 1;                                                     //  use preview
   EFbrramp.Farea = 2;                                                     //  select area usable
   EFbrramp.threadfunc = brramp_thread;
   if (! edit_setup(EFbrramp)) return;                                     //  setup edit
   
/***
          _________________________________________
         |                                         |
         |                                         |
         |           curve edit area               |
         |                                         |
         |_________________________________________|

            (o) all  (o) red  (o) green  (o) blue                          //  colors: all, red, green, blue
            Curve File: [Open]  [Save]                                     //  each has vert. and horiz. curve
                                   [Done]  [Cancel]    

***/

   zdialog *zd = zdialog_new(ZTX("Ramp brightness across image"),mWin,Bdone,Bcancel,null);
   EFbrramp.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5|expand");
   zdialog_add_widget(zd,"vbox","vb1","hb1");
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"expand");
   zdialog_add_widget(zd,"label","lmax","vb1","+","space=3");
   zdialog_add_widget(zd,"label","lspace","vb1",0,"expand");
   zdialog_add_widget(zd,"label","lmin","vb1","‒","space=3");
   zdialog_add_widget(zd,"label","lspace","vb1");
   zdialog_add_widget(zd,"frame","frame","vb2",0,"expand");

   zdialog_add_widget(zd,"hbox","hb2","vb2");
   zdialog_add_widget(zd,"label","lmin","hb2","‒","space=3");
   zdialog_add_widget(zd,"label","lspace","hb2",0,"expand");
   zdialog_add_widget(zd,"label","lmax","hb2","+","space=3");
   zdialog_add_widget(zd,"hbox","hbrgb","dialog",0,"space=3");
   zdialog_add_widget(zd,"radio","all","hbrgb",Ball,"space=5");
   zdialog_add_widget(zd,"radio","red","hbrgb",Bred,"space=5");
   zdialog_add_widget(zd,"radio","green","hbrgb",Bgreen,"space=5");
   zdialog_add_widget(zd,"radio","blue","hbrgb",Bblue,"space=5");
   zdialog_add_widget(zd,"hbox","hbcf","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labcf","hbcf",Bcurvefile,"space=8");
   zdialog_add_widget(zd,"button","load","hbcf",Bopen,"space=5");
   zdialog_add_widget(zd,"button","save","hbcf",Bsave,"space=5");

   GtkWidget *frame = zdialog_widget(zd,"frame");
   spldat *sd = splcurve_init(frame,brramp_curvedit);
   EFbrramp.curves = sd;
   
   brramp_init();                                                          //  set up initial flat curves

   sd->Nspc = 2;                                                           //  only curves 0-1 are shown
   brramp_spc = 2;                                                         //  current curves: all (2-3)
   
   zdialog_stuff(zd,"all",1);                                              //  stuff default selection, all

   zdialog_resize(zd,260,350);
   zdialog_help(zd,"brightness_ramp");                                     //  zdialog help topic 
   zdialog_run(zd,brramp_dialog_event,"save");                             //  run dialog - parallel
   return;
}


//  initialize all curves to flat

void brramp_init()
{
   spldat *sd = EFbrramp.curves;

   for (int ii = 0; ii < 10; ii++)                                         //  loop curves: current (0-1), all (2-3),
   {                                                                       //    red (4-5), green (6-7), blue (8-9)
      sd->nap[ii] = 2;                                                     //  horizontal curve ii
      sd->vert[ii] = 0;
      sd->apx[ii][0] = 0.01;
      sd->apx[ii][1] = 0.99;
      sd->apy[ii][0] = sd->apy[ii][1] = 0.5;
      splcurve_generate(sd,ii);

      ii++;
      sd->nap[ii] = 2;                                                     //  vertical curve ii+1
      sd->vert[ii] = 1;
      sd->apx[ii][0] = 0.01;
      sd->apx[ii][1] = 0.99;
      sd->apy[ii][0] = sd->apy[ii][1] = 0.5;
      splcurve_generate(sd,ii);
   }
   
   return;
}


//  dialog event and completion callback function

int brramp_dialog_event(zdialog *zd, cchar *event)
{
   spldat      *sd = EFbrramp.curves;
   int         ii, jj, nn, Fupdate = 0;
   spldat      sdtemp;

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 1;
   }
   
   if (strEqu(event,"reset")) {                                            //  reset dialog controls
      brramp_init();
      Fupdate++;
   }

   if (strEqu(event,"blendwidth")) signal_thread();

   if (strstr("all red green blue",event)) {                               //  new choice of curve
      zdialog_fetch(zd,event,ii);
      if (! ii) return 0;                                                  //  ignore button OFF events

      ii = strcmpv(event,"all","red","green","blue",null);
      brramp_spc = ii = 2 * ii;                                            //  2, 4, 6, 8

      sd->nap[0] = sd->nap[ii];                                            //  copy active curves ii/ii+1 to curves 0/1
      sd->nap[1] = sd->nap[ii+1];
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[0][jj] = sd->apx[ii][jj];
         sd->apy[0][jj] = sd->apy[ii][jj];
      }
      for (jj = 0; jj < sd->nap[1]; jj++) {
         sd->apx[1][jj] = sd->apx[ii+1][jj];
         sd->apy[1][jj] = sd->apy[ii+1][jj];
      }

      Fupdate++;
   }

   if (strEqu(event,"load"))                                               //  load 8 saved curves from file
   {
      sdtemp.Nspc = 8;
      sdtemp.drawarea = 0;
      nn = splcurve_load(&sdtemp);
      if (nn != 1) return 0;

      for (ii = 0; ii < 8; ii++) {                                         //  load curves 0-7 to curves 2-9
         sd->vert[ii+2] = sdtemp.vert[ii];
         sd->nap[ii+2] = sdtemp.nap[ii];
         for (jj = 0; jj < sdtemp.nap[ii]; jj++) {
            sd->apx[ii+2][jj] = sdtemp.apx[ii][jj];
            sd->apy[ii+2][jj] = sdtemp.apy[ii][jj];
         }
         splcurve_generate(sd,ii+2);
      }

      ii = brramp_spc;                                                     //  current active curves: 2, 4, 6, 8

      sd->nap[0] = sd->nap[ii];                                            //  copy 2 active curves to curves 0-1
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[0][jj] = sd->apx[ii][jj];
         sd->apy[0][jj] = sd->apy[ii][jj];
      }

      sd->nap[1] = sd->nap[ii+1];
      for (jj = 0; jj < sd->nap[1]; jj++) {
         sd->apx[1][jj] = sd->apx[ii+1][jj];
         sd->apy[1][jj] = sd->apy[ii+1][jj];
      }

      Fupdate++;
   }

   if (strEqu(event,"save"))                                               //  save 8 curves to file
   {
      sdtemp.Nspc = 8;
      for (ii = 0; ii < 8; ii++) {
         sdtemp.vert[ii] = sd->vert[ii+2];
         sdtemp.nap[ii] = sd->nap[ii+2];
         for (jj = 0; jj < sd->nap[ii+2]; jj++) {
            sdtemp.apx[ii][jj] = sd->apx[ii+2][jj];
            sdtemp.apy[ii][jj] = sd->apy[ii+2][jj];
         }
      }

      splcurve_save(&sdtemp);
   }

   if (Fupdate)                                                            //  curve has changed
   {
      splcurve_generate(sd,0);                                             //  regenerate curves 0-1
      splcurve_generate(sd,1);
      
      ii = brramp_spc;                                                     //  active curves

      sd->nap[ii] = sd->nap[0];                                            //  copy curves 0-1 to active curves
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[ii][jj] = sd->apx[0][jj];
         sd->apy[ii][jj] = sd->apy[0][jj];
      }

      sd->nap[ii+1] = sd->nap[1];
      for (jj = 0; jj < sd->nap[1]; jj++) {
         sd->apx[ii+1][jj] = sd->apx[1][jj];
         sd->apy[ii+1][jj] = sd->apy[1][jj];
      }

      for (jj = 0; jj < 1000; jj++) 
      {
         sd->yval[ii][jj] = sd->yval[0][jj];
         sd->yval[ii+1][jj] = sd->yval[1][jj];
      }

      gtk_widget_queue_draw(sd->drawarea);                                 //  draw curve

      signal_thread();                                                     //  trigger image update
   }

   return 1;
}


//  this function is called when a curve is edited

void brramp_curvedit(int spc)
{
   int      ii = brramp_spc, jj;
   spldat   *sd = EFbrramp.curves;                                         //  active curves: 2, 4, 6, 8 (and +1)

   sd->nap[ii] = sd->nap[0];                                               //  copy curves 0-1 to active curves
   for (jj = 0; jj < sd->nap[0]; jj++) {
      sd->apx[ii][jj] = sd->apx[0][jj];
      sd->apy[ii][jj] = sd->apy[0][jj];
   }

   sd->nap[ii+1] = sd->nap[1];
   for (jj = 0; jj < sd->nap[1]; jj++) {
      sd->apx[ii+1][jj] = sd->apx[1][jj];
      sd->apy[ii+1][jj] = sd->apy[1][jj];
   }

   for (jj = 0; jj < 1000; jj++) {
      sd->yval[ii][jj] = sd->yval[0][jj];
      sd->yval[ii+1][jj] = sd->yval[1][jj];
   }

   if (brramp_spc == 2) {                                                  //  "all" curve was edited
      for (ii = 4; ii < 10; ii += 2) {                                     //  fix R/G/B curves to match
         sd->nap[ii] = sd->nap[2];
         for (jj = 0; jj < sd->nap[2]; jj++) {
            sd->apx[ii][jj] = sd->apx[2][jj];
            sd->apy[ii][jj] = sd->apy[2][jj];
         }
         for (jj = 0; jj < 1000; jj++)
            sd->yval[ii][jj] = sd->yval[2][jj];
      }
      for (ii = 5; ii < 10; ii += 2) {
         sd->nap[ii] = sd->nap[3];
         for (jj = 0; jj < sd->nap[3]; jj++) {
            sd->apx[ii][jj] = sd->apx[3][jj];
            sd->apy[ii][jj] = sd->apy[3][jj];
         }
         for (jj = 0; jj < 1000; jj++)
            sd->yval[ii][jj] = sd->yval[3][jj];
      }
   }

   signal_thread();
   return;
}


//  brramp thread function

void * brramp_thread(void *arg)
{
   void * brramp_wthread(void *);
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(brramp_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion
      
      mwpaint2();                                                          //  update window
      CEF->Fmod = 1;                                                       //  image3 modified
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * brramp_wthread(void *arg)                                           //  worker thread function
{
   int         index = *((int *) arg);
   int         ii, jj, kk, dist = 0, px3, py3;
   float       *pix1, *pix3;
   float       red1, green1, blue1, maxrgb;
   float       red3, green3, blue3;
   float       dispx, dispy;
   float       hramp[3], vramp[3], tramp[3];
   float       spanw, spanh, dold, dnew;
   spldat      *sd = EFbrramp.curves;

   if (sa_stat == 3) {                                                     //  if select area active, ramp
      spanw = sa_maxx - sa_minx;                                           //    brightness over enclosing rectangle
      spanh = sa_maxy - sa_miny;
   }
   else {
      spanw = E3pxm->ww;                                                   //  else over entire image
      spanh = E3pxm->hh;
   }   

   for (py3 = index; py3 < E3pxm->hh; py3 += Nwt)                          //  loop output pixels
   for (px3 = 0; px3 < E3pxm->ww; px3++)
   {
      if (sa_stat == 3) {                                                  //  select area active 
         ii = py3 * E3pxm->ww + px3;
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel outside area
         dispx = (px3 - sa_minx) / spanw;
         dispy = (py3 - sa_miny) / spanh;
      }
      else {
         dispx = px3 / spanw;                                              //  left > right = 0 to 1
         dispy = py3 / spanh;                                              //  top > bottom = 0 to 1
      }
      
      kk = 1000 * dispx;                                                   //  conv. dispx 0-1 to 0-1000
      if (kk > 999) kk = 999;

      for (ii = 0, jj = 4; ii < 3; ii++, jj+=2)                            //  horz. curves for dispx and R/G/B
         hramp[ii] = sd->yval[jj][kk] - 0.5;                               //  scale to -0.5 to +0.5
      
      kk = 1000 * dispy;                                                   //  conv. dispy
      if (kk > 999) kk = 999;

      for (ii = 0, jj = 5; ii < 3; ii++, jj+=2)                            //  vert. curves
         vramp[ii] = sd->yval[jj][kk] - 0.5;

      for (ii = 0; ii < 3; ii++)                                           //  total R/G/B change
         tramp[ii] = 1.0 + hramp[ii] + vramp[ii];                          //  scale to 0.0 to 2.0

      pix1 = PXMpix(E1pxm,px3,py3);                                        //  input pixel
      red1 = pix1[0];
      green1 = pix1[1];
      blue1 = pix1[2];
      
      red3 = red1 * tramp[0];                                              //  change R/G/B
      green3 = green1 * tramp[1];
      blue3 = blue1 * tramp[2];

      maxrgb = red3;
      if (green3 > maxrgb) maxrgb = green3;
      if (blue3 > maxrgb) maxrgb = blue3;
      
      if (maxrgb > 255.9) {                                                //  stop overflow
         red3 = red3 * 255.9 / maxrgb;
         green3 = green3 * 255.9 / maxrgb;
         blue3 = blue3 * 255.9 / maxrgb;
      }

      if (sa_stat == 3 && dist < sa_blend) {                               //  blend changes over blendwidth
         dnew = 1.0 * dist / sa_blend;
         dold = 1.0 - dnew;
         red3 = dnew * red3 + dold * red1;
         green3 = dnew * green3 + dold * green1;
         blue3 = dnew * blue3 + dold * blue1;
      }

      pix3 = PXMpix(E3pxm,px3,py3);                                        //  output pixel
      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************/

//  find and remove "dust" from an image (e.g. from a scanned dusty slide)
//  dust is defined as small dark areas surrounded by brighter areas
//  image 1   original with prior edits
//  image 3   accumulated dust removals that have been comitted
//  image 9   comitted dust removals + pending removal (work in process)

namespace dust_names 
{
   editfunc    EFdust;

   int         spotspann;                                                  //  max. dustspot spann, pixels
   int         spotspann2;                                                 //  spotspann **2
   float       brightness;                                                 //  brightness limit, 0 to 1 = white
   float       contrast;                                                   //  min. contrast, 0 to 1 = black/white
   int         *pixgroup;                                                  //  maps (px,py) to pixel group no.
   int         Fred;                                                       //  red pixels are on

   int         Nstack;

   struct spixstack {
      uint16      px, py;                                                  //  pixel group search stack
      uint16      direc;
   }  *pixstack;

   #define maxgroups 1000000
   int         Ngroups;
   int         groupcount[maxgroups];                                      //  count of pixels in each group
   float       groupbright[maxgroups];                                     //  
   int         edgecount[maxgroups];                                       //  group edge pixel count
   float       edgebright[maxgroups];                                      //  group edge pixel brightness sum

   typedef struct {
      uint16      px1, py1, px2, py2;                                      //  pixel group extreme pixels
      int         spann2;                                                  //  spann from px1/py1 to px2/py2
   }  sgroupspann;
   
   sgroupspann    groupspann[maxgroups];
}


void m_dust(GtkWidget *, const char *)
{
   using namespace dust_names;

   int    dust_dialog_event(zdialog *zd, cchar *event);
   void * dust_thread(void *);
   
   zfuncs::F1_help_topic = "remove_dust";

   EFdust.funcname = "dust";
   EFdust.Farea = 2;                                                       //  select area usable
   EFdust.threadfunc = dust_thread;                                        //  thread function
   if (! edit_setup(EFdust)) return;                                       //  setup edit

   E9pxm = PXM_copy(E3pxm);                                                //  image 9 = copy of image3
   Fred = 0;

   int cc = E1pxm->ww * E1pxm->hh * sizeof(int);
   pixgroup = (int *) zmalloc(cc,"erase_dust");                            //  maps pixels to assigned groups

   cc = E1pxm->ww * E1pxm->hh * sizeof(spixstack);
   pixstack = (spixstack *) zmalloc(cc,"erase_dust");                      //  pixel group search stack
   
/***
                     Remove Dust

        spot size limit    =========[]===========
        max. brightness    =============[]=======
        min. contrast      ========[]============
        [erase] [red] [undo last] [apply]

                                  [Done] [Cancel]
***/

   zdialog *zd = zdialog_new(ZTX("Remove Dust"),mWin,Bdone,Bcancel,null);
   EFdust.zd = zd;

   zdialog_add_widget(zd,"hbox","hbssl","dialog",0,"space=1");
   zdialog_add_widget(zd,"label","labssl","hbssl",ZTX("spot size limit"),"space=5");
   zdialog_add_widget(zd,"hscale","spotspann","hbssl","1|50|1|20","space=5|expand");
   zdialog_add_widget(zd,"hbox","hbmb","dialog",0,"space=1");
   zdialog_add_widget(zd,"label","labmb","hbmb",ZTX("max. brightness"),"space=5");
   zdialog_add_widget(zd,"hscale","brightness","hbmb","1|999|1|700","space=5|expand");
   zdialog_add_widget(zd,"hbox","hbmc","dialog",0,"space=1");
   zdialog_add_widget(zd,"label","labmb","hbmc",ZTX("min. contrast"),"space=5");
   zdialog_add_widget(zd,"hscale","contrast","hbmc","1|500|1|50","space=5|expand");
   zdialog_add_widget(zd,"hbox","hbbutts","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","erase","hbbutts",Berase,"space=5");
   zdialog_add_widget(zd,"button","red","hbbutts",Bred,"space=5");
   zdialog_add_widget(zd,"button","undo1","hbbutts",Bundolast,"space=5");
   zdialog_add_widget(zd,"button","apply","hbbutts",Bapply,"space=5");

   zdialog_fetch(zd,"spotspann",spotspann);                                //  max. dustspot spann (pixels)
   spotspann2 = spotspann * spotspann;

   zdialog_fetch(zd,"brightness",brightness);                              //  max. dustspot brightness
   brightness = 0.001 * brightness;                                        //  scale 0 to 1 = white

   zdialog_fetch(zd,"contrast",contrast);                                  //  min. dustspot contrast
   contrast = 0.001 * contrast;                                            //  scale 0 to 1 = black/white

   zdialog_resize(zd,300,0);
   zdialog_help(zd,"remove_dust");                                         //  zdialog help topic 
   zdialog_run(zd,dust_dialog_event,"save");                               //  run dialog - parallel

   signal_thread();
   return;
}


//  dialog event and completion callback function

int dust_dialog_event(zdialog *zd, cchar *event)
{
   using namespace dust_names;
   
   void dust_erase();

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) {                                                //  done, use committed changes
         mutex_lock(&Fpixmap_lock);
         PXM_free(E3pxm);
         E3pxm = E9pxm;                                                    //  image 3 = image 9
         E9pxm = 0;
         mutex_unlock(&Fpixmap_lock);
         edit_done();
      }
      else {                                                               //  cancel, discard changes
         PXM_free(E9pxm);
         edit_cancel();
      }

      zfree(pixgroup);                                                     //  free memory
      zfree(pixstack);
      return 0;
   }
   
   if (strEqu(event,"spotspann") || strEqu(event,"brightness") 
          || strEqu(event,"contrast") || strEqu(event,"red"))
   {
      zdialog_fetch(zd,"spotspann",spotspann);                             //  max. dustspot spann (pixels)
      spotspann2 = spotspann * spotspann;

      zdialog_fetch(zd,"brightness",brightness);                           //  max. dustspot brightness
      brightness = 0.001 * brightness;                                     //  scale 0 to 1 = white
      
      zdialog_fetch(zd,"contrast",contrast);                               //  min. dustspot contrast
      contrast = 0.001 * contrast;                                         //  scale 0 to 1 = black/white

      signal_thread();                                                     //  do the work
   }
   
   if (strEqu(event,"erase")) dust_erase();
   if (strEqu(event,"blendwidth")) dust_erase();
   
   if (strEqu(event,"undo1")) {
      mutex_lock(&Fpixmap_lock);                                           //  image 3 = copy of image 9
      PXM_free(E3pxm);
      E3pxm = PXM_copy(E9pxm);
      mutex_unlock(&Fpixmap_lock);
      Fred = 0;
      mwpaint2();
   }
   
   if (strEqu(event,"apply")) {
      if (Fred) dust_erase();
      PXM_free(E9pxm);                                                     //  image 9 = copy of image 3
      E9pxm = PXM_copy(E3pxm);
      CEF->Fmod = 1;
   }
   
   return 0;
}


//  dust find thread function - find the dust particles and mark them

void * dust_thread(void *)
{
   using namespace dust_names;

   int         xspann, yspann, spann2;
   int         group, cc, ii, kk, Nremoved;
   int         px, py, dx, dy, ppx, ppy, npx, npy;
   float       gbright, pbright, pcontrast;
   float       ff = 1.0 / 256.0;
   uint16      direc;
   float       *pix3;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      mutex_lock(&Fpixmap_lock);                                           //  image 3 = copy of image 9
      PXM_free(E3pxm);
      E3pxm = PXM_copy(E9pxm);
      mutex_unlock(&Fpixmap_lock);
      mwpaint2();

      cc = E1pxm->ww * E1pxm->hh * sizeof(int);                            //  clear group arrays
      memset(pixgroup,0,cc);
      cc = maxgroups * sizeof(int);
      memset(groupcount,0,cc);
      memset(edgecount,0,cc);
      cc = maxgroups * sizeof(float );
      memset(groupbright,0,cc);
      memset(edgebright,0,cc);
      cc = maxgroups * sizeof(sgroupspann);
      memset(groupspann,0,cc);

      group = 0;
      
      for (py = 0; py < E1pxm->hh; py++)                                   //  loop all pixels
      for (px = 0; px < E1pxm->ww; px++)
      {
         ii = py * E1pxm->ww + px;
         if (sa_stat == 3 && ! sa_pixmap[ii]) continue;                    //  not in active area
         if (pixgroup[ii]) continue;                                       //  already assigned to a group

         pix3 = PXMpix(E3pxm,px,py);                                       //  get pixel brightness
         gbright = ff * pixbright(pix3);                                   //  0 to 1.0 = white
         if (gbright > brightness) continue;                               //  ignore bright pixel

         if (group == maxgroups-1) break;                                  //  too many groups, make no more

         pixgroup[ii] = ++group;                                           //  assign next group
         groupcount[group] = 1;
         groupbright[group] = gbright;

         pixstack[0].px = px;                                              //  put pixel into stack with
         pixstack[0].py = py;                                              //    direction = ahead
         pixstack[0].direc = 0;
         Nstack = 1;

         while (Nstack)
         {
            kk = Nstack - 1;                                               //  get last pixel in stack
            px = pixstack[kk].px;
            py = pixstack[kk].py;
            direc = pixstack[kk].direc;                                    //  next search direction
            
            if (direc == 'x') {
               Nstack--;                                                   //  none left
               continue;
            }

            if (Nstack > 1) {
               ii = Nstack - 2;                                            //  get prior pixel in stack
               ppx = pixstack[ii].px;
               ppy = pixstack[ii].py;
            }
            else {
               ppx = px - 1;                                               //  if only one, assume prior = left
               ppy = py;
            }
            
            dx = px - ppx;                                                 //  vector from prior to this pixel
            dy = py - ppy;
            
            switch (direc) 
            {
               case 0:
                  npx = px + dx;
                  npy = py + dy;
                  pixstack[kk].direc = 1;
                  break;

               case 1:
                  npx = px + dy;
                  npy = py + dx;
                  pixstack[kk].direc = 3;
                  break;
            
               case 2:
                  npx = px - dx;                                           //  back to prior pixel
                  npy = py - dy;                                           //  (this path never taken)
                  zappcrash("stack search bug");
                  break;
            
               case 3:
                  npx = px - dy;
                  npy = py - dx;
                  pixstack[kk].direc = 4;
                  break;
            
               case 4:
                  npx = px - dx;
                  npy = py + dy;
                  pixstack[kk].direc = 5;
                  break;
            
               case 5:
                  npx = px - dy;
                  npy = py + dx;
                  pixstack[kk].direc = 6;
                  break;
            
               case 6:
                  npx = px + dx;
                  npy = py - dy;
                  pixstack[kk].direc = 7;
                  break;
            
               case 7:
                  npx = px + dy;
                  npy = py - dx;
                  pixstack[kk].direc = 'x';
                  break;
               
               default:
                  npx = npy = 0;
                  zappcrash("stack search bug");
            }

            if (npx < 0 || npx > E1pxm->ww-1) continue;                     //  pixel off the edge
            if (npy < 0 || npy > E1pxm->hh-1) continue;
            
            ii = npy * E1pxm->ww + npx;
            if (pixgroup[ii]) continue;                                    //  pixel already assigned
            if (sa_stat == 3 && ! sa_pixmap[ii]) continue;                 //  pixel outside area

            pix3 = PXMpix(E3pxm,npx,npy);                                  //  pixel brightness
            pbright = ff * pixbright(pix3);
            if (pbright > brightness) continue;                            //  brighter than limit

            pixgroup[ii] = group;                                          //  assign pixel to group
            ++groupcount[group];                                           //  count pixels in group
            groupbright[group] += pbright;                                 //  sum brightness for group

            kk = Nstack++;                                                 //  put pixel into stack
            pixstack[kk].px = npx;
            pixstack[kk].py = npy;
            pixstack[kk].direc = 0;                                        //  search direction
         }
      }
      
      Ngroups = group;                                                     //  group numbers are 1-Ngroups
      Nremoved = 0;
      
      for (py = 0; py < E1pxm->hh; py++)                                   //  loop all pixels
      for (px = 0; px < E1pxm->ww; px++)
      {
         ii = py * E1pxm->ww + px;
         group = pixgroup[ii];
         if (! group) continue;
         if (groupspann[group].px1 == 0) {                                 //  first pixel found in this group
            groupspann[group].px1 = px;                                    //  group px1/py1 = this pixel
            groupspann[group].py1 = py;
            continue;
         }
         xspann = groupspann[group].px1 - px;                              //  spann from group px1/py1 to this pixel
         yspann = groupspann[group].py1 - py;
         spann2 = xspann * xspann + yspann * yspann;
         if (spann2 > groupspann[group].spann2) {   
            groupspann[group].spann2 = spann2;                             //  if greater, group px2/py2 = this pixel
            groupspann[group].px2 = px;
            groupspann[group].py2 = py;
         }
      }

      for (py = 0; py < E1pxm->hh; py++)                                   //  loop all pixels
      for (px = 0; px < E1pxm->ww; px++)
      {
         ii = py * E1pxm->ww + px;
         group = pixgroup[ii];
         if (! group) continue;
         if (groupspann[group].spann2 > spotspann2) continue;
         xspann = groupspann[group].px2 - px;                              //  spann from this pixel to group px2/py2
         yspann = groupspann[group].py2 - py;
         spann2 = xspann * xspann + yspann * yspann;
         if (spann2 > groupspann[group].spann2) {   
            groupspann[group].spann2 = spann2;                             //  if greater, group px1/py1 = this pixel
            groupspann[group].px1 = px;
            groupspann[group].py1 = py;
         }
      }

      for (py = 0; py < E1pxm->hh; py++)                                   //  loop all pixels
      for (px = 0; px < E1pxm->ww; px++)
      {
         ii = py * E1pxm->ww + px;                                         //  eliminate group if spann > limit
         group = pixgroup[ii];
         if (! group) continue;
         if (! groupcount[group]) pixgroup[ii] = 0;
         else if (groupspann[group].spann2 > spotspann2) {
            pixgroup[ii] = 0;
            groupcount[group] = 0;
            Nremoved++;
         }
      }
      
      for (py = 1; py < E1pxm->hh-1; py++)                                 //  loop all pixels except image edges
      for (px = 1; px < E1pxm->ww-1; px++)
      {
         ii = py * E1pxm->ww + px;
         group = pixgroup[ii];
         if (group) continue;                                              //  find pixels bordering group pixels
         pix3 = PXMpix(E3pxm,px,py);
         pbright = ff * pixbright(pix3);
         
         group = pixgroup[ii-E1pxm->ww-1];
         if (group) {
            ++edgecount[group];                                            //  accumulate pixel count and
            edgebright[group] += pbright;                                  //      bordering the groups
         }

         group = pixgroup[ii-E1pxm->ww];
         if (group) {
            ++edgecount[group];
            edgebright[group] += pbright;
         }

         group = pixgroup[ii-E1pxm->ww+1];
         if (group) {
            ++edgecount[group];
            edgebright[group] += pbright;
         }

         group = pixgroup[ii-1];
         if (group) {
            ++edgecount[group];
            edgebright[group] += pbright;
         }

         group = pixgroup[ii+1];
         if (group) {
            ++edgecount[group];
            edgebright[group] += pbright;
         }

         group = pixgroup[ii+E1pxm->ww-1];
         if (group) {
            ++edgecount[group];
            edgebright[group] += pbright;
         }

         group = pixgroup[ii+E1pxm->ww];
         if (group) {
            ++edgecount[group];
            edgebright[group] += pbright;
         }

         group = pixgroup[ii+E1pxm->ww+1];
         if (group) {
            ++edgecount[group];
            edgebright[group] += pbright;
         }
      }

      for (group = 1; group <= Ngroups; group++)                           //  compute group pixel and edge pixel
      {                                                                    //    mean brightness
         if (groupcount[group] && edgecount[group]) {
            edgebright[group] = edgebright[group] / edgecount[group];
            groupbright[group] = groupbright[group] / groupcount[group];
            pcontrast = edgebright[group] - groupbright[group];            //  edge - group contrast
            if (pcontrast < contrast) {
               groupcount[group] = 0;
               Nremoved++;
            }
         }
      }

      for (py = 0; py < E1pxm->hh; py++)                                   //  loop all pixels
      for (px = 0; px < E1pxm->ww; px++)
      {
         ii = py * E1pxm->ww + px;                                         //  eliminate group if low contrast
         group = pixgroup[ii];
         if (! group) continue;
         if (! groupcount[group]) pixgroup[ii] = 0;
      }

      for (py = 0; py < E1pxm->hh; py++)                                   //  loop all pixels
      for (px = 0; px < E1pxm->ww; px++)
      {
         ii = py * E1pxm->ww + px;
         if (! pixgroup[ii]) continue;                                     //  not a dust pixel
         pix3 = PXMpix(E3pxm,px,py);                                       //  paint it red
         pix3[0] = 255;
         pix3[1] = pix3[2] = 0;
      }

      Fred = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


//  erase the selected dust areas

void dust_erase()
{
   using namespace dust_names;
   
   int         cc, ii, px, py, inc;
   int         qx, qy, npx, npy;
   int         sx, sy, tx, ty;
   int         rad, dist, dist2, mindist2;
   float       slope, f1, f2;
   float       *pix1, *pix3;
   char        *pmap;

   Ffuncbusy++;

   mutex_lock(&Fpixmap_lock);                                              //  image 3 = copy of image 9
   PXM_free(E3pxm);
   E3pxm = PXM_copy(E9pxm);
   mutex_unlock(&Fpixmap_lock);

   cc = E1pxm->ww * E1pxm->hh;                                             //  allocate pixel done map
   pmap = (char *) zmalloc(cc,"erase_dust");
   memset(pmap,0,cc);

   for (py = 0; py < E1pxm->hh; py++)                                      //  loop all pixels
   for (px = 0; px < E1pxm->ww; px++)
   {
      ii = py * E1pxm->ww + px;
      if (! pixgroup[ii]) continue;                                        //  not a dust pixel
      if (pmap[ii]) continue;                                              //  skip pixels already done
      
      mindist2 = 999999;
      npx = npy = 0;

      for (rad = 1; rad < 10; rad++)                                       //  find nearest edge (10 pixel limit)
      {
         for (qx = px-rad; qx <= px+rad; qx++)                             //  search within rad
         for (qy = py-rad; qy <= py+rad; qy++)
         {
            if (qx < 0 || qx >= E1pxm->ww) continue;                       //  off image edge 
            if (qy < 0 || qy >= E1pxm->hh) continue;
            ii = qy * E1pxm->ww + qx;
            if (pixgroup[ii]) continue;                                    //  within dust area

            dist2 = (px-qx) * (px-qx) + (py-qy) * (py-qy);                 //  distance**2 to edge pixel
            if (dist2 < mindist2) {
               mindist2 = dist2;
               npx = qx;                                                   //  save nearest pixel found
               npy = qy;
            }
         }

         if (rad * rad >= mindist2) break;                                 //  can quit now
      }

      if (! npx && ! npy) continue;                                        //  should not happen

      qx = npx;                                                            //  nearest edge pixel
      qy = npy;

      if (abs(qy - py) > abs(qx - px)) {                                   //  qx/qy = near edge from px/py
         slope = 1.0 * (qx - px) / (qy - py);
         if (qy > py) inc = 1;
         else inc = -1;
         for (sy = py; sy != qy+inc; sy += inc)                            //  line from px/py to qx/qy
         {
            sx = px + slope * (sy - py);
            ii = sy * E1pxm->ww + sx;
            if (pmap[ii]) continue;
            pmap[ii] = 1;
            tx = qx + (qx - sx);                                           //  tx/ty = parallel line from qx/qy 
            ty = qy + (qy - sy);
            if (tx < 0) tx = 0;
            if (tx > E1pxm->ww-1) tx = E1pxm->ww-1;
            if (ty < 0) ty = 0;
            if (ty > E1pxm->hh-1) ty = E1pxm->hh-1;
            pix1 = PXMpix(E3pxm,tx,ty);                                    //  copy pixel from tx/ty to sx/sy
            pix3 = PXMpix(E3pxm,sx,sy);
            pix3[0] = pix1[0];
            pix3[1] = pix1[1];
            pix3[2] = pix1[2];
         }
      }

      else {
         slope = 1.0 * (qy - py) / (qx - px);
         if (qx > px) inc = 1;
         else inc = -1;
         for (sx = px; sx != qx+inc; sx += inc)
         {
            sy = py + slope * (sx - px);
            ii = sy * E1pxm->ww + sx;
            if (pmap[ii]) continue;
            pmap[ii] = 1;
            tx = qx + (qx - sx);
            ty = qy + (qy - sy);
            if (tx < 0) tx = 0;
            if (tx > E1pxm->ww-1) tx = E1pxm->ww-1;
            if (ty < 0) ty = 0;
            if (ty > E1pxm->hh-1) ty = E1pxm->hh-1;
            pix1 = PXMpix(E3pxm,tx,ty);
            pix3 = PXMpix(E3pxm,sx,sy);
            pix3[0] = pix1[0];
            pix3[1] = pix1[1];
            pix3[2] = pix1[2];
         }
      }
   }

   zfree(pmap);

   if (sa_stat == 3)                                                       //  area edge blending
   {
      for (ii = 0; ii < E1pxm->ww * E1pxm->hh; ii++)                       //  find pixels in select area
      {
         dist = sa_pixmap[ii];
         if (! dist || dist >= sa_blend) continue;

         py = ii / E1pxm->ww;
         px = ii - py * E1pxm->ww;
         pix1 = PXMpix(E1pxm,px,py);                                       //  input pixel, unchanged image
         pix3 = PXMpix(E3pxm,px,py);                                       //  output pixel, changed image

         f2 = 1.0 * dist / sa_blend;                                       //  changes over distance sa_blend
         f1 = 1.0 - f2;

         pix3[0] = f1 * pix1[0] + f2 * pix3[0];                            //  blend the pixels
         pix3[1] = f1 * pix1[1] + f2 * pix3[1];
         pix3[2] = f1 * pix1[2] + f2 * pix3[2];
      }
   }

   Fred = 0;
   Ffuncbusy--;
   mwpaint2();                                                             //  update window
   return;
}


/**************************************************************************/

//  Find and fix stuck pixels (always bright or dark) from camera sensor defects.

namespace stuckpix 
{
   editfunc    EFstuckpix;
   int         stuckpix_1x1, stuckpix_2x2, stuckpix_3x3;                   //  pixel blocks to search
   int         stuckpix_color;                                             //  1/2/3 = white/black/red circles
   cchar       *stuckpix_mode;                                             //  find or fix
   float       stuckpix_threshcon;                                         //  min. contrast threshold
   char        *stuckpix_file = 0;                                         //  file for saved bad pixels

   struct badpix_t {                                                       //  memory for bad pixels
      int      px, py;                                                     //  location (NW corner of block)
      int      size;                                                       //  size: 1/2/3 = 1x1/2x2/3x3 block
      float    pcon;                                                       //  contrast with surrounding pixels
      int      rgb[3];                                                     //  surrounding pixel mean RGB
   };
   badpix_t    badpix[50];
   int         maxbad = 50, Nbad;
}


void m_stuckpix(GtkWidget *, cchar *menu)
{
   using namespace stuckpix;

   int    stuckpix_dialog_event(zdialog *zd, cchar *event);
   void * stuckpix_thread(void *);

   zdialog     *zd;

   zfuncs::F1_help_topic = "stuck_pixels";

   EFstuckpix.funcname = "stuckpix";                                       //  function name
   EFstuckpix.threadfunc = stuckpix_thread;                                //  thread function

   if (! edit_setup(EFstuckpix)) return;                                   //  setup edit

/***

   pixel group   [x] 1x1  [x] 2x2  [x] 3x3
   circle color  (o) white  (o) black  (o) red
   contrast      =========[]====================

                  [open] [save] [apply] [cancel]
***/

   zd = zdialog_new(ZTX("Fix Stuck Pixels"),mWin,Bopen,Bsave,Bapply,Bcancel,null);
   EFstuckpix.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"vbox","vb1","hb1",0,"space=5|homog");
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"space=5|expand|homog");
   zdialog_add_widget(zd,"label","labgroup","vb1",ZTX("pixel group"));
   zdialog_add_widget(zd,"label","labcolor","vb1",ZTX("circle color"));
   zdialog_add_widget(zd,"label","labcont","vb1",ZTX("contrast"));
   zdialog_add_widget(zd,"hbox","hbgroup","vb2");
   zdialog_add_widget(zd,"check","1x1","hbgroup","1x1","space=3");
   zdialog_add_widget(zd,"check","2x2","hbgroup","2x2","space=3");
   zdialog_add_widget(zd,"check","3x3","hbgroup","3x3","space=3");
   zdialog_add_widget(zd,"hbox","hbcolor","vb2");
   zdialog_add_widget(zd,"radio","white","hbcolor",Bwhite,"space=5");
   zdialog_add_widget(zd,"radio","black","hbcolor",Bblack,"space=5");
   zdialog_add_widget(zd,"radio","red","hbcolor",Bred,"space=5");
   zdialog_add_widget(zd,"hscale","contrast","vb2","0.1|1.0|0.001|0.5","expand");
   
   zdialog_stuff(zd,"1x1",1);                                              //  initz. dialog controls
   zdialog_stuff(zd,"2x2",1);
   zdialog_stuff(zd,"3x3",1);
   zdialog_stuff(zd,"white",0);
   zdialog_stuff(zd,"black",0);
   zdialog_stuff(zd,"red",1);
   
   stuckpix_1x1 = stuckpix_2x2 = stuckpix_3x3 = 1;                         //  corresp. data values
   stuckpix_color = 3;
   stuckpix_threshcon = 0.5;
   
   if (! stuckpix_file) {                                                  //  default file
      stuckpix_file = zmalloc(200,"stuck_pixels");                         //  /.../.fotoxx/stuck-pixels
      snprintf(stuckpix_file,200,"%s/stuck-pixels",get_zuserdir());
   }

   zdialog_help(zd,"stuck_pixels");
   zdialog_run(zd,stuckpix_dialog_event,"save");                           //  run dialog

   stuckpix_mode = "find";
   signal_thread();                                                        //  find and show the bad pixels
   wait_thread_idle();

   return;
}


//  dialog event function

int stuckpix_dialog_event(zdialog *zd, cchar *event)
{
   using namespace stuckpix;
   
   void  stuckpix_open();
   void  stuckpix_save();

   int      color;
   
   if (zd->zstat)                                                          //  completion button
   {
      if (zd->zstat == 1) {                                                //  open file
         zd->zstat = 0;                                                    //  keep dialog active
         stuckpix_open();
      }

      else if (zd->zstat == 2) {                                           //  save file
         zd->zstat = 0;                                                    //  keep dialog active
         stuckpix_save();
      }

      else if (zd->zstat == 3) {                                           //  apply
         stuckpix_mode = "apply";
         signal_thread();                                                  //  fix the bad pixels
         wait_thread_idle();
         edit_done();
         erase_topcircles();
      }

      else {                                                               //  cancel
         edit_cancel();
         erase_topcircles();
      }

      mwpaint2();                                                          //  update main window
      return 0;
   }

   zdialog_fetch(zd,"1x1",stuckpix_1x1);                                   //  get dialog inputs
   zdialog_fetch(zd,"2x2",stuckpix_2x2);
   zdialog_fetch(zd,"3x3",stuckpix_3x3);
   
   zdialog_fetch(zd,"white",color);
   if (color) stuckpix_color = 1;
   zdialog_fetch(zd,"black",color);
   if (color) stuckpix_color = 2;
   zdialog_fetch(zd,"red",color);
   if (color) stuckpix_color = 3;

   zdialog_fetch(zd,"contrast",stuckpix_threshcon);
   
   stuckpix_mode = "find";
   signal_thread();                                                        //  find and show bad pixels
   wait_thread_idle();

   return 0;
}


//  load bad pixel list from a previously saved file

void stuckpix_open()
{
   using namespace stuckpix;

   int  stuckpix_open_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;
   
   zd = zdialog_new(ZTX("Load Stuck Pixels"),mWin,Bopen,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","lab1","hb1",ZTX("File:"),"space=3");
   zdialog_add_widget(zd,"entry","file","hb1",0,"expand|scc=40");
   zdialog_add_widget(zd,"button","browse","hb1",Bbrowse,"space=5");
   
   zdialog_stuff(zd,"file",stuckpix_file);
   
   zdialog_run(zd,stuckpix_open_dialog_event);
   zdialog_wait(zd);

   return;
}

   
int stuckpix_open_dialog_event(zdialog *zd, cchar *event)
{
   using namespace stuckpix;

   char     *pp, file[200];
   FILE     *fid = 0;
   int      zstat, nn, ii, px, py, size;
   
   if (strEqu(event,"browse")) {
      pp = zgetfile1(ZTX("Stuck Pixels file"),"open",stuckpix_file);
      if (! pp) return 0;
      zdialog_stuff(zd,"file",pp);
      zfree(pp);
   }
   
   zstat = zd->zstat;                                                      //  completion button
   
   if (zstat)
   {
      if (zstat == 1)                                                      //  open
      {
         zdialog_fetch(zd,"file",file,200);                                //  get file from dialog
         if (stuckpix_file) zfree(stuckpix_file);
         stuckpix_file = strdupz(file,0,"stuck_pixels");

         fid = fopen(file,"r");                                            //  open file
         if (! fid) {
            zmessageACK(mWin,ZTX("file not found"));
            return 1;
         }
         
         nn = fscanf(fid,"stuck pixels px py size");                       //  read headers
         
         for (ii = 0; ii < maxbad; ii++)
         {
            nn = fscanf(fid," %5d %5d %5d ",&px,&py,&size);                //  read bad pixels data
            if (nn == EOF) break;
            if (nn != 3) break;
            badpix[ii].px = px;
            badpix[ii].py = py;
            badpix[ii].size = size;
            badpix[ii].pcon = 0;
         }
         
         Nbad = ii;
         fclose(fid);
         zdialog_free(zd);

         if (! Nbad || nn != EOF) 
            zmessageACK(mWin,ZTX("file format error"));

         stuckpix_mode = "file";                                           //  process pixels
         signal_thread();
         wait_thread_idle();
      }
      
      else {                                                               //  cancel or [x]
         zdialog_free(zd);
         return 1;
      }
   }

   return 1;
}


//  save bad pixel list to a file or add them to a previous file

void stuckpix_save()
{
   using namespace stuckpix;

   int  stuckpix_save_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;
   
   if (Nbad == 0) {
      zmessageACK(mWin,ZTX("there are zero stuck pixels"));
      return;
   }
   
   zd = zdialog_new(ZTX("Save Stuck Pixels"),mWin,Bsave,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","lab1","hb1",ZTX("File:"),"space=3");
   zdialog_add_widget(zd,"entry","file","hb1",0,"expand|scc=40");
   zdialog_add_widget(zd,"button","browse","hb1",Bbrowse,"space=5");
   
   zdialog_stuff(zd,"file",stuckpix_file);
   
   zdialog_run(zd,stuckpix_save_dialog_event);
   zdialog_wait(zd);

   return;
}


int stuckpix_save_dialog_event(zdialog *zd, cchar *event)
{
   using namespace stuckpix;

   char     *pp, file[200];
   FILE     *fid = 0;
   int      zstat, ii, px, py, size;
   
   if (strEqu(event,"browse")) {
      pp = zgetfile1(ZTX("Stuck Pixels file"),"save",stuckpix_file);
      if (! pp) return 0;
      zdialog_stuff(zd,"file",pp);
      zfree(pp);
   }
   
   zstat = zd->zstat;                                                      //  completion button
   
   if (zstat)
   {
      if (zstat == 1)                                                      //  save
      {
         zdialog_fetch(zd,"file",file,200);                                //  get file from dialog
         if (stuckpix_file) zfree(stuckpix_file);
         stuckpix_file = strdupz(file,0,"stuck_pixels");

         fid = fopen(file,"w");                                            //  open file
         if (! fid) {
            zmessageACK(mWin,strerror(errno));
            return 1;
         }
         
         fprintf(fid,"stuck pixels \n");                                   //  write headers
         fprintf(fid,"    px    py   size \n");
         
         for (ii = 0; ii < Nbad; ii++)                                     //  write bad pixel data
         {
            px = badpix[ii].px;
            py = badpix[ii].py;
            size = badpix[ii].size;
            fprintf(fid," %5d %5d %5d \n",px,py,size);
         }

         fclose(fid);
         zdialog_free(zd);
      }

      else {                                                               //  cancel or [x]
         zdialog_free(zd);
         return 1;
      }
   }

   return 1;
}


//  perform the fix function
//  find pixel groups with contrast exceeding the limit
//  replace these pixels with surrounding ones

void * stuckpix_thread(void *)
{
   using namespace stuckpix;

   void   stuckpix_pixelblock(int px, int py, int size, float rgb[3]);
   void   stuckpix_surroundings(int px, int py, int size, float rgb[3]);
   float  stuckpix_getcon(float rgbA[3], float rgbB[3]);
   void   stuckpix_insert(int px, int py, int size, float pcon, float rgb[3]);
   void   stuckpix_show();

   int         px, py, qx, qy, ii, size;
   float       rgbA[3], rgbB[3];
   float       threshcon, pcon;
   float       *ppix;
   
   while (true)
   {
      thread_idle_loop();
   
      if (strEqu(stuckpix_mode,"find"))                                    //  find bad pixels in image
      {
         threshcon = stuckpix_threshcon;                                   //  threshold contrast, 0.1 to 1.0
         Nbad = 0;                                                         //  count bad pixels found

         if (stuckpix_3x3)                                                 //  find 3x3 pixel groups FIRST
         {
            for (py = 2; py < E3pxm->hh-5; py++)
            for (px = 2; px < E3pxm->ww-5; px++)
            {
               stuckpix_pixelblock(px,py,3,rgbA);                          //  get mean RGB for 3x3 pixel block
               stuckpix_surroundings(px,py,3,rgbB);                        //  get surrounding pixels mean RGB
               pcon = stuckpix_getcon(rgbA,rgbB);                          //  contrast with surrounding
               if (pcon > threshcon) 
                  stuckpix_insert(px,py,3,pcon,rgbB);                      //  if > threshold, add to table
               if (Nbad == maxbad) goto findquit;
            }
         }

         if (stuckpix_2x2)                                                 //  find 2x2 pixel groups
         {
            for (py = 2; py < E3pxm->hh-4; py++)
            for (px = 2; px < E3pxm->ww-4; px++)
            {
               stuckpix_pixelblock(px,py,2,rgbA);                          //  get mean RGB for 2x2 pixel block
               stuckpix_surroundings(px,py,2,rgbB);                        //  get surrounding pixels mean RGB
               pcon = stuckpix_getcon(rgbA,rgbB);                          //  contrast with surrounding
               if (pcon > threshcon) 
                  stuckpix_insert(px,py,2,pcon,rgbB);                      //  if > threshold, add to table
               if (Nbad == maxbad) goto findquit;
            }
         }

         if (stuckpix_1x1)                                                 //  find 1x1 pixel groups LAST
         {
            for (py = 2; py < E3pxm->hh-3; py++)
            for (px = 2; px < E3pxm->ww-3; px++)
            {
               stuckpix_pixelblock(px,py,1,rgbA);                          //  get mean RGB for 1x1 pixel block
               stuckpix_surroundings(px,py,1,rgbB);                        //  get surrounding pixels mean RGB
               pcon = stuckpix_getcon(rgbA,rgbB);                          //  contrast with surrounding
               if (pcon > threshcon) 
                  stuckpix_insert(px,py,1,pcon,rgbB);                      //  if > threshold, add to table
               if (Nbad == maxbad) goto findquit;
            }
         }

         findquit:
         stuckpix_show();                                                  //  show the bad pixels found
      }

      if (strEqu(stuckpix_mode,"file"))                                    //  process bad pixels read from a file
      {
         for (ii = 0; ii < Nbad; ii++)
         {
            px = badpix[ii].px;
            py = badpix[ii].py;
            size = badpix[ii].size;
            
            stuckpix_pixelblock(px,py,size,rgbA);
            stuckpix_surroundings(px,py,size,rgbB);
            pcon = stuckpix_getcon(rgbA,rgbB);
            
            badpix[ii].pcon = pcon;
            badpix[ii].rgb[0] = rgbB[0];
            badpix[ii].rgb[1] = rgbB[1];
            badpix[ii].rgb[2] = rgbB[2];
         }
      
         stuckpix_show();                                                  //  show the bad pixels read
      }

      if (strEqu(stuckpix_mode,"apply"))                                   //  replace the bad pixels 
      {
         for (ii = 0; ii < Nbad; ii++)                                     //  loop pixel groups found
         {
            px = badpix[ii].px;
            py = badpix[ii].py;
            size = badpix[ii].size;
            if (! size) continue;
            
            if (size == 1)                                                 //  1x1 pixel group
            {
               ppix = PXMpix(E3pxm,px,py);                                 //  replace pixel group
               ppix[0] = badpix[ii].rgb[0];
               ppix[1] = badpix[ii].rgb[1];
               ppix[2] = badpix[ii].rgb[2];
            }
            
            if (size == 2)                                                 //  2x2 pixel group
            {
               for (qy = py; qy < py+2; qy++)                              //  replace pixel group
               for (qx = px; qx < px+2; qx++)
               {
                  ppix = PXMpix(E3pxm,qx,qy);
                  ppix[0] = badpix[ii].rgb[0];
                  ppix[1] = badpix[ii].rgb[1];
                  ppix[2] = badpix[ii].rgb[2];
               }
            }

            if (size == 3)                                                 //  3x3 pixel group
            {
               for (qy = py; qy < py+3; qy++)                              //  replace pixel group
               for (qx = px; qx < px+3; qx++)
               {
                  ppix = PXMpix(E3pxm,qx,qy);
                  ppix[0] = badpix[ii].rgb[0];
                  ppix[1] = badpix[ii].rgb[1];
                  ppix[2] = badpix[ii].rgb[2];
               }
            }
         }
         
         EFstuckpix.Fmod++;                                                //  image modified
         mwpaint2();                                                       //  update window
      }
   }

   return 0;
}


//  get the mean RGB values for a block of (defective) pixels

void stuckpix_pixelblock(int px, int py, int size, float rgb[3])
{
   float    *ppix;
   
   ppix = PXMpix(E3pxm,px,py);                                             //  block of pixels

   if (size == 1)                                                          //  1x1 block, 1 pixel
   {
      rgb[0] = ppix[0];
      rgb[1] = ppix[1];
      rgb[2] = ppix[2];
   }

   if (size == 2)                                                          //  2x2 block, 4 pixels
   {
      rgb[0] = (ppix[0] + ppix[3] + ppix[E3pxm->ww*3]   + ppix[E3pxm->ww*3+3]) / 4;
      rgb[1] = (ppix[1] + ppix[4] + ppix[E3pxm->ww*3+1] + ppix[E3pxm->ww*3+4]) / 4;
      rgb[2] = (ppix[2] + ppix[5] + ppix[E3pxm->ww*3+2] + ppix[E3pxm->ww*3+5]) / 4;
   }

   if (size == 3)                                                          //  3x3 block, 9 pixels
   {
      rgb[0] = (ppix[0] + ppix[3] + ppix[6]
              +  ppix[E3pxm->ww*3] + ppix[E3pxm->ww*3+3] + ppix[E3pxm->ww*3+6]
              +  ppix[E3pxm->ww*6] + ppix[E3pxm->ww*6+3] + ppix[E3pxm->ww*6+6]) / 9;

      rgb[1] = (ppix[1] + ppix[4] + ppix[7]
              +  ppix[E3pxm->ww*3+1] + ppix[E3pxm->ww*3+4] + ppix[E3pxm->ww*3+7]
              +  ppix[E3pxm->ww*6+1] + ppix[E3pxm->ww*6+4] + ppix[E3pxm->ww*6+7]) / 9;

      rgb[2] = (ppix[2] + ppix[5] + ppix[8]
              +  ppix[E3pxm->ww*3+2] + ppix[E3pxm->ww*3+5] + ppix[E3pxm->ww*3+8]
              +  ppix[E3pxm->ww*6+2] + ppix[E3pxm->ww*6+5] + ppix[E3pxm->ww*6+8]) / 9;
   }

   return;
}


//  get the mean RGB values for pixels surrounding a given pixel block

void stuckpix_surroundings(int px, int py, int size, float rgb[3])
{
   int      qx, qy, ii;
   float    red, green, blue;
   float    *ppix;

   int   n8x[8] = { -1, 0, 1,-1, 1,-1, 0, 1 };                             //  8 neighbors of 1x1 group at [0,0]
   int   n8y[8] = { -1,-1,-1, 0, 0, 1, 1, 1 };

   int   n12x[12] = { -1, 0, 1, 2,-1, 2,-1, 2,-1, 0, 1, 2 };               //  12 neighbors of 2x2 group at [0,0]
   int   n12y[12] = { -1,-1,-1,-1, 0, 0, 1, 1, 2, 2, 2, 2 };

   int   n16x[16] = { -1, 0, 1, 2, 3,-1, 3,-1, 3,-1, 3,-1, 0, 1, 2, 3 };   //  16 neighbors of 3x3 group at [0,0]
   int   n16y[16] = { -1,-1,-1,-1,-1, 0, 0, 1, 1, 2, 2, 3, 3, 3, 3, 3 };

   red = green = blue = 0;

   if (size == 1)
   {
      for (ii = 0; ii < 8; ii++)                                           //  surrounding 8 pixels
      {
         qx = px + n8x[ii];
         qy = py + n8y[ii];
         ppix = PXMpix(E3pxm,qx,qy);
         red += ppix[0];
         green += ppix[1];
         blue += ppix[2];
      }
      
      red = red / 8;                                                       //  average surrounding pixels                  
      green = green / 8;
      blue = blue / 8;
   }

   if (size == 2)
   {
      for (ii = 0; ii < 12; ii++)                                          //  surrounding 12 pixels
      {
         qx = px + n12x[ii];
         qy = py + n12y[ii];
         ppix = PXMpix(E3pxm,qx,qy);
         red += ppix[0];
         green += ppix[1];
         blue += ppix[2];
      }
      
      red = red / 12;                                                      //  average surrounding pixels                  
      green = green / 12;
      blue = blue / 12;
   }

   if (size == 3)
   {
      for (ii = 0; ii < 16; ii++)                                          //  surrounding 16 pixels
      {
         qx = px + n16x[ii];
         qy = py + n16y[ii];
         ppix = PXMpix(E3pxm,qx,qy);
         red += ppix[0];
         green += ppix[1];
         blue += ppix[2];
      }
      
      red = red / 16;                                                      //  average surrounding pixels                  
      green = green / 16;
      blue = blue / 16;
   }
   
   rgb[0] = red;
   rgb[1] = green;
   rgb[2] = blue;
   
   return;
}


//  draw circles around the bad pixels

void stuckpix_show()
{
   using namespace stuckpix;

   int      ii, px, py, size, rad;

   erase_topcircles();                                                     //  erase prior circles

   for (ii = 0; ii < Nbad; ii++)                                           //  write circles around bad pixels
   {
      px = badpix[ii].px;
      py = badpix[ii].py;
      size = badpix[ii].size;
      if (! size) continue;
      rad = 8 + 2 * size;
      px += size / 2;
      py += size / 2;
      add_topcircle(px,py,rad,stuckpix_color);
   }
   
   mwpaint2();                                                             //  update window
   return;
}


//  compare two pixels and return contrast
//  0 = no contrast, 1 = max. contrast (black:white)

float  stuckpix_getcon(float rgbA[3], float rgbB[3])
{
   float       ff = 1.0 / 256.0;
   float       Rdiff, Gdiff, Bdiff, match;

   Rdiff = ff * fabsf(rgbA[0] - rgbB[0]);                                  //  0 = perfect match
   Gdiff = ff * fabsf(rgbA[1] - rgbB[1]);                                  //  1 = total mismatch
   Bdiff = ff * fabsf(rgbA[2] - rgbB[2]);
   
   match = (1.0 - Rdiff) * (1.0 - Gdiff) * (1.0 - Bdiff);                  //  1 = perfect match
   return (1.0 - match);
}


//  Insert a new bad pixel block into the bad pixel table.
//  If the new entry touches on a previous entry,
//  then remove the entry with lesser contrast.

void stuckpix_insert(int px, int py, int size, float pcon, float rgb[3])
{
   using namespace stuckpix;
   
   int      ii, px1, py1, px2, py2, px3, py3, px4, py4;
   
   px1 = px - 1;                                                           //  "touch" periphery
   py1 = py - 1;
   px2 = px + size + 1;
   py2 = py + size + 1;
   
   for (ii = 0; ii < Nbad; ii++)                                           //  loop badpix table
   {
      if (badpix[ii].size == 0) continue;                                  //  skip deleted entry

      px3 = badpix[ii].px;
      py3 = badpix[ii].py;
      px4 = px3 + badpix[ii].size - 1;
      py4 = py3 + badpix[ii].size - 1;

      if ((px1 >= px3 && px1 <= px4 && py1 >= py3 && py1 <= py4) ||        //  test if old entry touches new
          (px2 >= px3 && px2 <= px4 && py2 >= py3 && py2 <= py4) ||
          (px1 >= px3 && px1 <= px4 && py2 >= py3 && py2 <= py4) ||
          (px2 >= px3 && px2 <= px4 && py1 >= py3 && py1 <= py4) ||
          (px3 >= px1 && px3 <= px2 && py3 >= py1 && py3 <= py2) ||
          (px4 >= px1 && px4 <= px2 && py4 >= py1 && py4 <= py2) ||
          (px3 >= px1 && px3 <= px2 && py4 >= py1 && py4 <= py2) ||
          (px4 >= px1 && px4 <= px2 && py3 >= py1 && py3 <= py2))
      {
         if (size < badpix[ii].size   &&                                   //  if new touches a larger block,
             pcon < 2.0 * badpix[ii].pcon) return;                         //    keep only if contrast much greater

         if (pcon < 1.0 * badpix[ii].pcon) return;                         //  keep only if contrast is greater

         badpix[ii].size = 0;                                              //  delete old entry
      }
   }

   for (ii = 0; ii < Nbad; ii++)                                           //  loop badpix table
      if (badpix[ii].size == 0) break;                                     //  find first empty slot or last + 1
   if (ii == maxbad) return;                                               //  table full
   
   badpix[ii].px = px;                                                     //  replace overlapping entry
   badpix[ii].py = py;                                                     //    or add to end of table
   badpix[ii].size = size;
   badpix[ii].pcon = pcon;
   badpix[ii].rgb[0] = rgb[0];
   badpix[ii].rgb[1] = rgb[1];
   badpix[ii].rgb[2] = rgb[2];

   if (ii == Nbad) Nbad++;                                                 //  incr. count if added to end
   if (Nbad == maxbad) printf("*** exceed %d bad pixels \n",maxbad);       //  report one time only

   return;
}


/**************************************************************************/

//  Correct chromatic abberation (1st order only) by stretching
//  or shrinking a selected RGB color plane.

editfunc    EFchromatic;
float       chromaticRed, chromaticBlue;
int         chromaticFarea;


void m_chromatic(GtkWidget *, cchar *)
{
   int    chromatic_dialog_event(zdialog *zd, cchar *event);
   void * chromatic_thread(void *);

   cchar  *chromatic_message = ZTX(" Adjust each RGB color to minimize \n"
                                   " color fringes at the image extremes. ");
   
   zfuncs::F1_help_topic = "chromatic";

   EFchromatic.funcname = "chromatic";
   EFchromatic.Farea = 2;                                                  //  select area usable
   EFchromatic.threadfunc = chromatic_thread;

   if (! edit_setup(EFchromatic)) return;                                  //  setup edit
   
/***
              Chromatic Abberation
         
         Adjust each RGB color to minimize
         color fringes at the image extremes.
         
         Red   =======[]===========  [-2.1]
         Blue  ============[]======  [+2.3]
         
                           [done] [cancel]
***/

   zdialog *zd = zdialog_new(ZTX("Chromatic Abberation"),mWin,Bdone,Bcancel,null);
   EFchromatic.zd = zd;

   zdialog_add_widget(zd,"label","lab1","dialog",chromatic_message,"space=5");
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"vbox","vb1","hb1",0,"space=5|homog");
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"space=5|homog|expand");
   zdialog_add_widget(zd,"vbox","vb3","hb1",0,"space=5|homog");
   zdialog_add_widget(zd,"label","labred","vb1",Bred);
   zdialog_add_widget(zd,"label","labblue","vb1",Bblue);
   zdialog_add_widget(zd,"hscale","red","vb2","-5.0|5.0|0.1|0.0");
   zdialog_add_widget(zd,"hscale","blue","vb2","-5.0|5.0|0.1|0.10");
   zdialog_add_widget(zd,"label","redval","vb3"," 0.0");
   zdialog_add_widget(zd,"label","blueval","vb3"," 0.0");

   zdialog_help(zd,"chromatic");                                           //  zdialog help topic
   
   chromaticRed = chromaticBlue = 0;
   chromaticFarea = 0;

   zdialog_run(zd,chromatic_dialog_event,"save");                          //  run dialog, parallel

   return;
}


//  dialog event and completion callback function

int chromatic_dialog_event(zdialog *zd, cchar *event)
{
   char  text[8];

   if (strEqu(event,"red")) {
      zdialog_fetch(zd,"red",chromaticRed);
      sprintf(text,"%+.1f",chromaticRed);
      zdialog_stuff(zd,"redval",text);
      signal_thread();
      return 0;
   }

   if (strEqu(event,"blue")) {
      zdialog_fetch(zd,"blue",chromaticBlue);
      sprintf(text,"%+.1f",chromaticBlue);
      zdialog_stuff(zd,"blueval",text);
      signal_thread();
      return 0;
   }

   if (! zd->zstat) return 0;                                              //  wait for completion
   
   if (zd->zstat) 
   {
      if (zd->zstat == 1)                                                  //  edit done
      {
         if (chromaticFarea && sa_stat != 3)                               //  extend to whole image if
            signal_thread();                                               //    former select area now deleted
         edit_done();      
      }

      else edit_cancel();                                                  //  cancel or destroy

      zdialog_free(zd);
      return 0;
   }

   return 0;
}


//  chromatic image and accumulate chromatic memory

void * chromatic_thread(void *)
{
   void  * chromatic_wthread(void *);                                      //  worker thread process

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(chromatic_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      CEF->Fmod = 1;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * chromatic_wthread(void *arg)                                        //  worker thread
{
   int         index = *((int *) (arg));
   int         px3, py3, ii;
   int         xlo, xhi, ylo, yhi;
   float       px1, py1;
   float       vpix1[3], *pix3;
   float       cx, cy, fx, fy;
   
   cx = E3pxm->ww / 2.0;
   cy = E3pxm->hh / 2.0;
   
   if (sa_stat == 3) {
      chromaticFarea = 1;                                                  //  remember if area active
      xlo = sa_minx + index;                                               //  set area limits
      xhi = sa_maxx;
      ylo = sa_miny + index;
      yhi = sa_maxy;
   }
   
   else {
      chromaticFarea = 0;
      xlo = index;                                                         //  set whole image limits
      xhi = E3pxm->ww;
      ylo = index;
      yhi = E3pxm->hh;
   }
   
   for (py3 = ylo; py3 < yhi; py3 += Nwt)                                  //  loop all output pixels
   for (px3 = xlo; px3 < xhi; px3++)
   {
      if (chromaticFarea) {                                                //  area active
         ii = py3 * E3pxm->ww + px3;
         if (! sa_pixmap[ii]) continue;                                    //  pixel not in area
      }

      pix3 = PXMpix(E3pxm,px3,py3);                                        //  output pixel

      fx = (px3 - cx) / cx;                                                //  -1 to 0 to +1
      fy = (py3 - cy) / cy;

      px1 = px3 + fx * chromaticRed;                                       //  red shift
      py1 = py3 + fy * chromaticRed;
      vpixel(E1pxm,px1,py1,vpix1);
      pix3[0] = vpix1[0];
   
      px1 = px3 + fx * chromaticBlue;                                      //  blue shift
      py1 = py3 + fy * chromaticBlue;
      vpixel(E1pxm,px1,py1,vpix1);
      pix3[2] = vpix1[2];
   }
   
   exit_wthread();
   return 0;
}


/**************************************************************************/

//  Pixel paint function - paint individual pixels with the mouse.
//  The mouse circle pixels have a selected color, or they are taken from the 
//  image at a relative position (x,y) from the mouse position (aka "clone").

#define ppix_undomaxmem (1000 * mega)                                      //  max. undo memory
#define ppix_undomaxpix (mega)                                             //  max. pixel blocks

void  ppix_mousefunc();
void  ppix_dopixels(int px, int py);
void  ppix_undo1();
void  ppix_freeundo();

uint8    ppix_RGB[3] = { 255, 0, 0 };
int      ppix_mode;                                                        //  1/2 = paint / erase
int      ppix_paint_mode;                                                  //  1/2 = paint color / paint image
int      ppix_radius;
int      ppix_imagex, ppix_imagey;                                         //  source image location
float    ppix_kernel[200][200];                                            //  radius <= 99

int      ppix_undototpix = 0;                                              //  total undo pixel blocks
int      ppix_undototmem = 0;                                              //  total undo memory allocated
int      ppix_undoseq = 0;                                                 //  undo sequence no.
char     ppix_undomemmessage[100];                                         //  translated undo memory message

typedef struct {                                                           //  pixel block before edit
   int         seq;                                                        //  undo sequence no.
   uint16      npix;                                                       //  no. pixels in this block
   uint16      px, py;                                                     //  center pixel (radius org.)
   uint16      radius;                                                     //  radius of pixel block
   float       pixel[][3];                                                 //  array of pixel[npix][3] 
}  savepix_t;

savepix_t   **ppix_undopixmem = 0;                                         //  array of *savepix_t

editfunc    EFpaintpix;


void m_paint_pixels(GtkWidget *, cchar *)
{
   int   ppix_dialog_event(zdialog* zd, cchar *event);

   cchar    *mess1 = ZTX("shift + left click: pick color or image position \n"
                         "left click or drag: paint color or copy image \n"
                         "right click or drag: remove color or image");

   char     undomemmessage[100];

   zfuncs::F1_help_topic = "paint_pixels";

   EFpaintpix.funcname = "paint_pixels";
   EFpaintpix.Farea = 2;                                                   //  select area OK
   EFpaintpix.mousefunc = ppix_mousefunc;                                  //  mouse function
   if (! edit_setup(EFpaintpix)) return;                                   //  setup edit

   strncpy0(ppix_undomemmessage,ZTX("Undo Memory %d%c"),99);               //  translate undo memory message

   /********
                    _____
   (o) paint color |_____|    (o) copy from image                          //  paint image added      v.12.04

   left click: pick color or location from image
   left drag: paint color or image from location
   right drag: remove color or image

   paintbrush radius    [___|v]     [undo last]
   transparency center  [___|v]     [undo all]
   transparency edge    [___|v]
   Undo Memory  NN%

   ********/

   zdialog *zd = zdialog_new(ZTX("Paint Pixels"),mWin,Bdone,Bcancel,null);
   EFpaintpix.zd = zd;

   zdialog_add_widget(zd,"hbox","hbr","dialog",0,"space=5");
   zdialog_add_widget(zd,"radio","paintcolor","hbr",ZTX("paint color"),"space=5");
   zdialog_add_widget(zd,"colorbutt","colorbutt","hbr","255|0|0");
   zdialog_add_widget(zd,"label","space","hbr",0,"space=10");
   zdialog_add_widget(zd,"radio","paintimage","hbr",ZTX("copy from image"));
   zdialog_add_widget(zd,"label","labm","dialog",mess1,"space=8");
   zdialog_add_widget(zd,"hbox","hbbri","dialog",0,"space=5");
   zdialog_add_widget(zd,"vbox","vbbr1","hbbri",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vbbr2","hbbri",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","space","hbbri",0,"space=10");
   zdialog_add_widget(zd,"vbox","vbbr3","hbbri",0,"space=10");
   zdialog_add_widget(zd,"label","labbr","vbbr1",ZTX("paintbrush radius"));
   zdialog_add_widget(zd,"label","labtc","vbbr1",ZTX("transparency center"));
   zdialog_add_widget(zd,"label","labte","vbbr1",ZTX("transparency edge"));
   zdialog_add_widget(zd,"spin","radius","vbbr2","1|99|1|5");
   zdialog_add_widget(zd,"spin","trcent","vbbr2","0|100|0.1|95");
   zdialog_add_widget(zd,"spin","tredge","vbbr2","0|100|0.1|100");
   zdialog_add_widget(zd,"button","undlast","vbbr3",Bundolast);
   zdialog_add_widget(zd,"button","undall","vbbr3",Bundoall);
   zdialog_add_widget(zd,"hbox","hb4","dialog");   
   zdialog_add_widget(zd,"label","labmem","hb4",0,"space=10");
   
   zdialog_stuff(zd,"paintcolor",1);                                       //  default paint color mode
   zdialog_stuff(zd,"paintimage",0);

   snprintf(undomemmessage,99,ppix_undomemmessage,0,'%');                  //  stuff undo memory status
   zdialog_stuff(zd,"labmem",undomemmessage);

   zdialog_help(zd,"paint_pixels");                                        //  zdialog help topic 
   zdialog_restore_inputs(zd);                                             //  preload prior user inputs    v.12.12
   zdialog_run(zd,ppix_dialog_event,"save");                               //  run dialog, parallel

   zdialog_fetch(zd,"paintcolor",ppix_paint_mode);                         //  1/2 = paint color / image    v.12.12
   if (ppix_paint_mode != 1) ppix_paint_mode = 2;
   
   zdialog_send_event(zd,"radius");                                        //  get kernel initialized

   ppix_mode = 1;                                                          //  start with paint mode
   ppix_undopixmem = 0;                                                    //  no undo data
   ppix_undototpix = 0;
   ppix_undototmem = 0;
   ppix_undoseq = 0;

   takeMouse(ppix_mousefunc,drawcursor);                                   //  connect mouse function
   return;
}


//  dialog event and completion callback function

int ppix_dialog_event(zdialog *zd, cchar *event)                           //  paintpix dialog event function
{
   char        color[20];
   cchar       *pp;
   int         radius, dx, dy;
   float       rad, kern, trcent, tredge;
   
   if (zd->zstat) 
   {
      if (zd->zstat == 1) edit_done();                                     //  done
      else edit_cancel();                                                  //  cancel or destroy
      ppix_freeundo();                                                     //  free undo memory
      return 0;
   }
   
   draw_mousearc(-1,0,0,0);                                                //  erase mouse ellipse

   if (strEqu(event,"focus"))                                              //  toggle mouse capture
      takeMouse(ppix_mousefunc,drawcursor);
   
   if (strstr("paintcolor paintimage",event))
   {
      zdialog_fetch(zd,"paintcolor",ppix_paint_mode);                      //  1/2 = paint color / paint image
      if (ppix_paint_mode != 1) ppix_paint_mode = 2;
   }

   if (strEqu(event,"colorbutt")) 
   {
      zdialog_fetch(zd,"colorbutt",color,19);                              //  get paint color from color wheel
      pp = strField(color,"|",1);
      if (pp) ppix_RGB[0] = atoi(pp);
      pp = strField(color,"|",2);
      if (pp) ppix_RGB[1] = atoi(pp);
      pp = strField(color,"|",3);
      if (pp) ppix_RGB[2] = atoi(pp);
   }
   
   if (strstr("radius trcent tredge",event))                               //  get new brush attributes
   {
      zdialog_fetch(zd,"radius",radius);                                   //  radius
      zdialog_fetch(zd,"trcent",trcent);                                   //  center transparency
      zdialog_fetch(zd,"tredge",tredge);                                   //  edge transparency

      ppix_radius = radius;
      trcent = 0.01 * trcent;                                              //  scale 0 ... 1
      tredge = 0.01 * tredge;
      tredge = (1 - trcent) * (1 - tredge);
      tredge = 1 - tredge;
      trcent = sqrt(trcent);                                               //  speed up the curve
      tredge = sqrt(tredge);

      for (dy = -radius; dy <= radius; dy++)                               //  build kernel
      for (dx = -radius; dx <= radius; dx++)
      {
         rad = sqrt(dx*dx + dy*dy);
         kern = (radius - rad) / radius;                                   //  center = 1 ... edge = 0 
         kern = kern * (trcent - tredge) + tredge;                         //      trcent ... tredge
         if (rad > radius) kern = 1;
         if (kern < 0) kern = 0;
         if (kern > 1) kern = 1;
         ppix_kernel[dx+radius][dy+radius] = kern;
      }
   }
   
   if (strEqu(event,"undlast"))                                            //  undo last edit (click or drag)
      ppix_undo1();

   if (strEqu(event,"undall")) {                                           //  undo all edits
      edit_reset();
      ppix_freeundo();
   }

   return 1;
}


//  pixel paint mouse function

void ppix_mousefunc()                                                      //  revised  v.12.03
{
   static int  pmxdown = 0, pmydown = 0;
   int         px, py, mrect;
   char        color[20];
   float       *ppix3;
   zdialog     *zd = EFpaintpix.zd;

   if (LMclick && KBshiftkey)                                              //  shift + left mouse click   v.12.12
   {
      if (ppix_paint_mode == 1) {                                          //  paint color mode
         px = Mxclick;
         py = Myclick;
         ppix3 = PXMpix(E3pxm,px,py);                                      //  pick new color from image
         ppix_RGB[0] = ppix3[0];
         ppix_RGB[1] = ppix3[1];
         ppix_RGB[2] = ppix3[2];
         snprintf(color,19,"%d|%d|%d",ppix_RGB[0],ppix_RGB[1],ppix_RGB[2]);
         if (zd) zdialog_stuff(zd,"colorbutt",color);
      }
      
      if (ppix_paint_mode == 2) {                                          //  paint image mode
         ppix_imagex = Mxclick;                                            //  new source image location
         ppix_imagey = Myclick;
      }
   }
   
   else if (LMclick || RMclick)                                            //  v.12.12
   {
      if (LMclick) ppix_mode = 1;                                          //  left click, paint
      if (RMclick) ppix_mode = 2;                                          //  right click, erase
   
      px = Mxclick;
      py = Myclick;

      ppix_undoseq++;                                                      //  new undo seq. no.

      draw_mousearc(-1,0,0,0);                                             //  erase mouse ellipse
      ppix_dopixels(px,py);                                                //  do 1 block of pixels
   }
   
   else if (Mxdrag || Mydrag)                                              //  drag in progress
   {
      if (Mbutton == 1) ppix_mode = 1;                                     //  left drag, paint
      if (Mbutton == 3) ppix_mode = 2;                                     //  right drag, erase
   
      px = Mxdrag;
      py = Mydrag;

      if (Mxdown != pmxdown || Mydown != pmydown) {                        //  new drag
         ppix_undoseq++;                                                   //  new undo seq. no.
         pmxdown = Mxdown;
         pmydown = Mydown;
      }

      draw_mousearc(-1,0,0,0);                                             //  erase mouse ellipse
      ppix_dopixels(px,py);                                                //  do 1 block of pixels
   }
   
   mrect = 2 * ppix_radius;

   draw_mousearc(Mxposn,Myposn,mrect,mrect);                               //  draw mouse ellipse

   LMclick = RMclick = Mxdrag = Mydrag = 0;                                //  v.12.11   
   return;
}


//  paint or erase 1 block of pixels within radius of px, py

void ppix_dopixels(int px, int py)
{
   void  ppix_saveundo(int px, int py);

   float       *ppix1, *ppix3;
   int         radius, dx, dy, qx, qy, sx, sy;
   int         ii, ww, dist;
   float       red, green, blue;
   float       kern;

   ppix_saveundo(px,py);                                                   //  save pixels for poss. undo

   red = ppix_RGB[0];
   green = ppix_RGB[1];
   blue = ppix_RGB[2];

   radius = ppix_radius;

   if (ppix_mode == 1) CEF->Fmod = 1;

   for (dy = -radius; dy <= radius; dy++)                                  //  loop surrounding block of pixels
   for (dx = -radius; dx <= radius; dx++)
   {
      qx = px + dx;
      qy = py + dy;
      
      if (qx < 0 || qx > E3pxm->ww-1) continue;
      if (qy < 0 || qy > E3pxm->hh-1) continue;
      
      if (sa_stat == 3) {                                                  //  select area active
         ii = qy * E3pxm->ww + qx;
         dist = sa_pixmap[ii];
         if (! dist) continue;                                             //  pixel is outside area
      }
      
      kern = ppix_kernel[dx+radius][dy+radius];
      ppix3 = PXMpix(E3pxm,qx,qy);                                         //  edited image pixel
      
      if (sa_stat == 3 && dist < sa_blend)                                 //  blend changes over blendwidth   v.12.04
         kern = kern + (1.0 - kern) * (sa_blend - dist) / sa_blend;
      
      if (ppix_mode == 1 && ppix_paint_mode == 1)                          //  color pixels transparently
      {
         ppix3[0] = (1.0 - kern) * red   + kern * ppix3[0];
         ppix3[1] = (1.0 - kern) * green + kern * ppix3[1];
         ppix3[2] = (1.0 - kern) * blue  + kern * ppix3[2];
      }

      if (ppix_mode == 1 && ppix_paint_mode == 2)                          //  paint from image location    v.12.04
      {
         sx = ppix_imagex + qx - Mxdown;                                   //  image location + mouse drag shift
         sy = ppix_imagey + qy - Mydown;
         if (sx < 0) sx = 0;
         if (sx > E3pxm->ww-1) sx = E3pxm->ww-1;
         if (sy < 0) sy = 0;
         if (sy > E3pxm->hh-1) sy = E3pxm->hh-1;
         ppix1 = PXMpix(E1pxm,sx,sy);                                      //  source image pixel
         ppix3[0] = (1.0 - kern) * ppix1[0] + kern * ppix3[0];
         ppix3[1] = (1.0 - kern) * ppix1[1] + kern * ppix3[1];
         ppix3[2] = (1.0 - kern) * ppix1[2] + kern * ppix3[2];
      }

      if (ppix_mode == 2)                                                  //  restore org. pixels transparently
      {
         ppix1 = PXMpix(E1pxm,qx,qy);                                      //  source image pixel
         ppix3[0] = (1.0 - kern) * ppix1[0] + kern * ppix3[0];
         ppix3[1] = (1.0 - kern) * ppix1[1] + kern * ppix3[1];
         ppix3[2] = (1.0 - kern) * ppix1[2] + kern * ppix3[2];
      }
   }

   px = px - radius - 1;                                                   //  repaint modified area
   py = py - radius - 1;
   ww = 2 * radius + 3;
   mwpaint3(px,py,ww,ww); 
   return;
}


//  save 1 block of pixels for possible undo

void ppix_saveundo(int px, int py)
{
   int            cc, npix, radius, dx, dy;
   float          *ppix3;
   savepix_t      *ppixsave1;
   char           undomemmessage[100];
   int            mempercent;
   static int     ppercent = 0;
   zdialog        *zd = EFpaintpix.zd;

   if (! ppix_undopixmem)                                                  //  first call
   {
      ppix_undopixmem = (savepix_t **) zmalloc(ppix_undomaxpix * sizeof(void *),"paintpix");
      ppix_undototpix = 0;
      ppix_undototmem = 0;
   }
   
   if (ppix_undototmem > ppix_undomaxmem) 
   {
      zmessageACK(mWin,ZTX("Undo memory limit has been reached. \n"
                           "Save work with [done], then resume painting."));
      Mdrag = 0;
      return;
   }

   radius = ppix_radius;
   npix = 0;

   for (dy = -radius; dy <= radius; dy++)                                  //  count pixels in block
   for (dx = -radius; dx <= radius; dx++)
   {
      if (px + dx < 0 || px + dx > E3pxm->ww-1) continue;
      if (py + dy < 0 || py + dy > E3pxm->hh-1) continue;
      npix++;
   }
   
   cc = npix * 3 * sizeof(float) + 12;                                     //  (risky code)
   ppixsave1 = (savepix_t *) zmalloc(cc,"paintpix");                       //  allocate memory for block
   ppix_undopixmem[ppix_undototpix] = ppixsave1;
   ppix_undototpix += 1;
   ppix_undototmem += npix * 6 + 12;
   
   ppixsave1->seq = ppix_undoseq;                                          //  save pixel block poop
   ppixsave1->npix = npix;
   ppixsave1->px = px;
   ppixsave1->py = py;
   ppixsave1->radius = radius;

   npix = 0;

   for (dy = -radius; dy <= radius; dy++)                                  //  save pixels in block
   for (dx = -radius; dx <= radius; dx++)
   {
      if (px + dx < 0 || px + dx > E3pxm->ww-1) continue;
      if (py + dy < 0 || py + dy > E3pxm->hh-1) continue;
      ppix3 = PXMpix(E3pxm,(px+dx),(py+dy));                               //  edited image pixel
      ppixsave1->pixel[npix][0] = ppix3[0];
      ppixsave1->pixel[npix][1] = ppix3[1];
      ppixsave1->pixel[npix][2] = ppix3[2];
      npix++;
   }

   mempercent = int(100.0 * ppix_undototmem / ppix_undomaxmem);            //  update undo memory status
   if (mempercent != ppercent) {
      ppercent = mempercent;
      snprintf(undomemmessage,99,ppix_undomemmessage,mempercent,'%');
      zdialog_stuff(zd,"labmem",undomemmessage);
   }

   return;
}


//  undo last undo sequence number

void ppix_undo1()
{
   int            pindex, npix, radius, mempercent;
   int            ww, px, py, dx, dy;
   float          *ppix3;
   char           undomemmessage[100];
   zdialog        *zd = EFpaintpix.zd;
   savepix_t      *ppixsave1;
   
   pindex = ppix_undototpix;
   
   while (pindex > 0)
   {
      --pindex;
      ppixsave1 = ppix_undopixmem[pindex];
      if (ppixsave1->seq != ppix_undoseq) break;
      px = ppixsave1->px;
      py = ppixsave1->py;
      radius = ppixsave1->radius;

      npix = 0;
      for (dy = -radius; dy <= radius; dy++)
      for (dx = -radius; dx <= radius; dx++)
      {
         if (px + dx < 0 || px + dx > E3pxm->ww-1) continue;
         if (py + dy < 0 || py + dy > E3pxm->hh-1) continue;
         ppix3 = PXMpix(E3pxm,(px+dx),(py+dy));
         ppix3[0] = ppixsave1->pixel[npix][0];
         ppix3[1] = ppixsave1->pixel[npix][1];
         ppix3[2] = ppixsave1->pixel[npix][2];
         npix++;
      }

      px = px - radius - 1;
      py = py - radius - 1;
      ww = 2 * radius + 3;
      mwpaint3(px,py,ww,ww);

      npix = ppixsave1->npix;
      zfree(ppixsave1);
      ppix_undopixmem[pindex] = 0;
      ppix_undototmem -= (npix * 6 + 12);
      --ppix_undototpix;
   }
   
   if (ppix_undoseq > 0) --ppix_undoseq;

   mempercent = int(100.0 * ppix_undototmem / ppix_undomaxmem);            //  update undo memory status
   snprintf(undomemmessage,99,ppix_undomemmessage,mempercent,'%');
   zdialog_stuff(zd,"labmem",undomemmessage);

   return;
}


//  free all undo memory

void ppix_freeundo()
{
   int            pindex;
   savepix_t      *ppixsave1;
   char           undomemmessage[100];
   zdialog        *zd = EFpaintpix.zd;

   pindex = ppix_undototpix;
   
   while (pindex > 0)
   {
      --pindex;
      ppixsave1 = ppix_undopixmem[pindex];
      zfree(ppixsave1);
   }
   
   if (ppix_undopixmem) zfree(ppix_undopixmem);
   ppix_undopixmem = 0;
   
   ppix_undoseq = 0;
   ppix_undototpix = 0;
   ppix_undototmem = 0;

   if (zd) {
      snprintf(undomemmessage,99,ppix_undomemmessage,0,'%');               //  undo memory = 0%
      zdialog_stuff(zd,"labmem",undomemmessage);
   }

   return;
}


/**************************************************************************/

//  Use the image brightness or color values to leverage subsequent edits.
//  Select the whole image as an area.
//  Set "edge distance" 1 to 255 from pixel or RGB color brightness.
//  Set "blend width" to 255
//  Edit function coefficient = edge distance / blend width.

spldat   *lever_edits_curve;

void m_lever_edits(GtkWidget *, cchar *)
{
   int    lever_edits_event(zdialog *, cchar *event);                      //  dialog event and completion func
   void   lever_edits_curve_update(int spc);                               //  curve update callback function

   cchar    *title = ZTX("Leverage Edits");
   cchar    *legend = ZTX("Edit Function Amplifier");

   zfuncs::F1_help_topic = "lever_edits";

   if (zdsela) return;                                                     //  dialog already active
   if (! curr_file) return;                                                //  no image
   if (! CEF) {                                                            //  edit func must be active
      zmessageACK(mWin,ZTX("start edit function first"));
      return;
   }
   if (Fpreview) edit_fullsize();                                          //  use full-size image

/***
                  Edit Function Amplifier
             ------------------------------------------
            |                                          |
            |                                          |
            |           curve drawing area             |
            |                                          |
            |                                          |
             ------------------------------------------
             darker areas                 lighter areas

             [+++]  [---]  [+ -]  [- +]  [+-+]  [-+-]
             (o) Brightness  (o) Red  (o) Green  (o) Blue
             Curve File: [ Open ] [ Save ]
                                              [ Done ]
***/

   zdsela = zdialog_new(title,mWin,Bdone,null);

   zdialog_add_widget(zdsela,"label","labt","dialog",legend);
   zdialog_add_widget(zdsela,"frame","fr1","dialog",0,"expand");
   zdialog_add_widget(zdsela,"hbox","hba","dialog");
   zdialog_add_widget(zdsela,"label","labda","hba",Bdarker,"space=5");
   zdialog_add_widget(zdsela,"label","space","hba",0,"expand");
   zdialog_add_widget(zdsela,"label","labba","hba",Blighter,"space=5");
   zdialog_add_widget(zdsela,"hbox","hbb","dialog",0,"space=5");
   zdialog_add_widget(zdsela,"button","b +++","hbb","+++");
   zdialog_add_widget(zdsela,"button","b ---","hbb","‒ ‒ ‒");
   zdialog_add_widget(zdsela,"button","b +-", "hbb"," + ‒ ");
   zdialog_add_widget(zdsela,"button","b -+", "hbb"," ‒ + ");
   zdialog_add_widget(zdsela,"button","b +-+","hbb","+ ‒ +");
   zdialog_add_widget(zdsela,"button","b -+-","hbb","‒ + ‒");

   zdialog_add_widget(zdsela,"hbox","hbbr","dialog",0,"space=5");
   zdialog_add_widget(zdsela,"radio","bright","hbbr",Bbrightness,"space=5");
   zdialog_add_widget(zdsela,"radio","red","hbbr",Bred,"space=5");
   zdialog_add_widget(zdsela,"radio","green","hbbr",Bgreen,"space=5");
   zdialog_add_widget(zdsela,"radio","blue","hbbr",Bblue,"space=5");

   zdialog_add_widget(zdsela,"hbox","hbcf","dialog",0,"space=5");
   zdialog_add_widget(zdsela,"label","labcf","hbcf",Bcurvefile,"space=5");
   zdialog_add_widget(zdsela,"button","load","hbcf",Bopen,"space=5");
   zdialog_add_widget(zdsela,"button","save","hbcf",Bsave,"space=5");

   GtkWidget *frame = zdialog_widget(zdsela,"fr1");                        //  setup for curve editing
   spldat *sd = splcurve_init(frame,lever_edits_curve_update);
   lever_edits_curve = sd;

   sd->Nspc = 1;
   sd->vert[0] = 0;
   sd->nap[0] = 3;                                                         //  initial curve anchor points
   sd->apx[0][0] = 0.01;
   sd->apy[0][0] = 0.5;
   sd->apx[0][1] = 0.50;
   sd->apy[0][1] = 0.5;
   sd->apx[0][2] = 0.99;
   sd->apy[0][2] = 0.5;
   splcurve_generate(sd,0);                                                //  generate curve data

   zdialog_stuff(zdsela,"bright",1);
   zdialog_stuff(zdsela,"red",0);
   zdialog_stuff(zdsela,"green",0);
   zdialog_stuff(zdsela,"blue",0);

   sa_unselect();                                                          //  unselect current area if any

   int cc = E1pxm->ww * E1pxm->hh * sizeof(uint16);                        //  allocate sa_pixmap[] for area
   sa_pixmap = (uint16 *) zmalloc(cc,"lever_edits");                       //  bugfix     v.12.12

   sa_minx = 0;                                                            //  enclosing rectangle
   sa_maxx = Fpxb->ww;
   sa_miny = 0;
   sa_maxy = Fpxb->hh;

   sa_Npixel = Fpxb->ww * Fpxb->hh;
   sa_stat = 3;                                                            //  area status = complete
   sa_mode = mode_image;                                                   //  area mode = whole image
   sa_calced = 1;                                                          //  edge calculation complete
   sa_blend = 255;                                                         //  "blend width" = 255
   sa_fww = Fpxb->ww;                                                      //  valid image dimensions
   sa_fhh = Fpxb->hh;
   areanumber++;                                                           //  next sequential number
   
   zdialog_resize(zdsela,0,360);
   zdialog_help(zdsela,"lever_edits");                                     //  zdialog help topic
   zdialog_run(zdsela,lever_edits_event,"save");                           //  run dialog - parallel
   lever_edits_event(zdsela,"init");                                       //  initialize default params

   return;
}


//  dialog event and completion function

int lever_edits_event(zdialog *zd, cchar *event)
{
   int         ii, kk, base, pixbright, pixdist;
   float       px, py, xval, yval;
   float       *pixel;
   spldat      *sd = lever_edits_curve;
   
   if (zd->zstat) {                                                        //  done, kill dialog
      zdialog_free(zdsela);
      zfree(sd);                                                           //  free curve edit memory
      sa_unselect();                                                       //  delete area                  v.12.11
      return 0;
   }
   
   if (! sa_validate() || sa_stat != 3) {                                  //  select area gone             v.12.12
      zdialog_free(zdsela);
      zfree(sd);                                                           //  free curve edit memory
      return 0;
   }

   if (strEqu(event,"load")) {                                             //  load saved curve 
      splcurve_load(sd);
      if (CEF && CEF->zd) zdialog_send_event(CEF->zd,"blendwidth");        //  notify edit dialog
      return 0;
   }

   if (strEqu(event,"save")) {                                             //  save curve to file
      splcurve_save(sd);
      return 0;
   }

   base = 0;
   zdialog_fetch(zd,"bright",ii);                                          //  set brightness or color to be used
   if (ii) base = 1;
   zdialog_fetch(zd,"red",ii);
   if (ii) base = 2;
   zdialog_fetch(zd,"green",ii);
   if (ii) base = 3;
   zdialog_fetch(zd,"blue",ii);
   if (ii) base = 4;
   if (! base) return 0;

   if (strnEqu(event,"b ",2)) {                                            //  button to move entire curve
      for (ii = 0; ii < sd->nap[0]; ii++) {
         px = sd->apx[0][ii];
         py = sd->apy[0][ii];
         if (strEqu(event,"b +++")) py += 0.1;
         if (strEqu(event,"b ---")) py -= 0.1;
         if (strEqu(event,"b +-"))  py += 0.1 - 0.2 * px;
         if (strEqu(event,"b -+"))  py -= 0.1 - 0.2 * px;
         if (strEqu(event,"b +-+")) py -= 0.05 - 0.2 * fabsf(px-0.5);
         if (strEqu(event,"b -+-")) py += 0.05 - 0.2 * fabsf(px-0.5);
         if (py > 1) py = 1;
         if (py < 0) py = 0;
         sd->apy[0][ii] = py;
      }
   }

   splcurve_generate(sd,0);                                                //  regenerate the curve
   gtk_widget_queue_draw(sd->drawarea);
   
   if (! CEF) return 0;                                                    //  no edit function active
   
   for (int ipy = 0; ipy < Fpxb->hh; ipy++)                                //  v.12.10
   for (int ipx = 0; ipx < Fpxb->ww; ipx++)
   {
      pixel = PXMpix(E1pxm,ipx,ipy);                                       //  bugfix        v.12.11
      pixbright = pixel[0] + pixel[1] + pixel[2];
      if (base == 1) pixbright = pixbright / 3;                            //  brightness, 0 - 255
      else if (base == 2) pixbright = pixel[0];                            //  red, 0 - 255
      else if (base == 3) pixbright = pixel[1];                            //  green 
      else if (base == 4) pixbright = pixel[2];                            //  blue 
      xval = pixbright / 256.0;                                            //  curve x-value, 0 to 0.999
      kk = 1000 * xval;
      if (kk > 999) kk = 999;
      yval = sd->yval[0][kk];                                              //  y-value, 0 to 0.999
      pixdist = 1 + 254 * yval;                                            //  pixel edge distance, 1 to 255
      ii = ipy * Fpxb->ww + ipx;                                           //  put edge distance into sa_pixmap[]
      sa_pixmap[ii] = pixdist;
   }
   
   if (CEF->zd) zdialog_send_event(CEF->zd,"blendwidth");                  //  notify edit dialog

   return 0;
}


//  this function is called when curve is edited using mouse

void  lever_edits_curve_update(int)
{
   lever_edits_event(zdsela,"edit");
   return;
}


/**************************************************************************/

//  Select area and edit in parallel
//  Current edit function is applied to areas painted with the mouse.
//  Mouse can be weak or strong, and edits are applied incrementally.
//  method: 
//    entire image is a select area with all pixel edge distance = 0 (outside area)
//    blendwidth = 10000 (infinite)
//    pixels painted with mouse have increasing edge distance to amplify edits

int   paint_edits_radius;
int   paint_edits_cpower;
int   paint_edits_epower;

void m_paint_edits(GtkWidget *, cchar *)                                   //  menu function
{
   int   paint_edits_dialog_event(zdialog *, cchar *event);                //  dialog event function
   void  paint_edits_mousefunc();                                          //  mouse function

   cchar    *title = ZTX("Paint Edits");
   cchar    *helptext1 = ZTX("Press F1 for help");
   cchar    *helptext2 = ZTX("Edit function must be active");
   int      cc;

   zfuncs::F1_help_topic = "paint_edits";

   if (! curr_file) return;                                                //  no image
   if (zdsela) return;                                                     //  select area already active
   if (Fpreview) edit_fullsize();                                          //  use full-size image
   
/***
    ____________________________________________
   |           Press F1 for help                |
   |       Edit Function must be active         |
   | mouse radius [___|v]                       |
   | power:  center [___|v]  edge [___|v]       |
   | [reset area]                               |
   |                                  [done]    |
   |____________________________________________|
   
***/

   zdsela = zdialog_new(title,mWin,Bdone,null);
   zdialog_add_widget(zdsela,"label","labhelp1","dialog",helptext1,"space=5");
   zdialog_add_widget(zdsela,"label","labhelp2","dialog",helptext2);
   zdialog_add_widget(zdsela,"label","labspace","dialog");
   zdialog_add_widget(zdsela,"hbox","hbr","dialog",0,"space=3");
   zdialog_add_widget(zdsela,"label","labr","hbr",ZTX("mouse radius"),"space=5");
   zdialog_add_widget(zdsela,"spin","radius","hbr","2|500|1|50");
   zdialog_add_widget(zdsela,"hbox","hbt","dialog",0,"space=3");
   zdialog_add_widget(zdsela,"label","labtc","hbt",ZTX("power:  center"),"space=5");
   zdialog_add_widget(zdsela,"spin","center","hbt","0|100|1|50");
   zdialog_add_widget(zdsela,"label","labte","hbt",ZTX("edge"),"space=5");
   zdialog_add_widget(zdsela,"spin","edge","hbt","0|100|1|0");
   zdialog_add_widget(zdsela,"hbox","hbra","dialog",0,"space=5");
   zdialog_add_widget(zdsela,"button","reset","hbra",ZTX("reset area"),"space=5");
   
   paint_edits_radius = 50;
   paint_edits_cpower = 50;
   paint_edits_epower = 0;

   sa_unselect();                                                          //  unselect current area if any
   cc = Fpxb->ww * Fpxb->hh * sizeof(uint16);                              //  allocate sa_pixmap[] for new area
   sa_pixmap = (uint16 *) zmalloc(cc,"paint_edits");
   memset(sa_pixmap,0,cc);                                                 //  edge distance = 0 for all pixels

   sa_minx = 0;                                                            //  enclosing rectangle
   sa_maxx = Fpxb->ww;
   sa_miny = 0;
   sa_maxy = Fpxb->hh;

   sa_Npixel = Fpxb->ww * Fpxb->hh;
   sa_stat = 3;                                                            //  area status = complete
   sa_mode = mode_image;                                                   //  area mode = whole image
   sa_calced = 1;                                                          //  edge calculation complete
   sa_blend = 10000;                                                       //  "blend width"
   sa_fww = Fpxb->ww;                                                      //  valid image dimensions
   sa_fhh = Fpxb->hh;
   areanumber++;                                                           //  next sequential number

   zdialog_help(zdsela,"paint_edits");                                     //  zdialog help topic
   zdialog_run(zdsela,paint_edits_dialog_event,"save");                    //  run dialog - parallel
   return;
}


//  tailor whole image area to increase edit power for pixels within the mouse radius
//  sa_pixmap[*]  = 0 = never touched by mouse
//                = 1 = minimum edit power (barely painted)
//                = sa_blend = maximum edit power (edit fully applied)

int paint_edits_dialog_event(zdialog *zd, cchar *event)
{
   void  paint_edits_mousefunc();                                          //  mouse function
   int      cc;
   
   if (sa_stat != 3) return 1;                                             //  area gone
   if (! sa_validate()) return 1;                                          //  area invalid for curr. image file  v.12.12

   if (zd->zstat)                                                          //  done or cancel
   {
      freeMouse();                                                         //  disconnect mouse function
      zdialog_free(zdsela);                                                //  kill dialog
      sa_unselect();                                                       //  unselect area
      return 0;
   }

   if (strEqu(event,"focus")) {                                            //  toggle mouse capture
      if (CEF) takeMouse(paint_edits_mousefunc,0);
      else freeMouse();                                                    //  disconnect mouse
   }

   if (strEqu(event,"radius"))
      zdialog_fetch(zd,"radius",paint_edits_radius);                       //  set mouse radius

   if (strEqu(event,"center"))
      zdialog_fetch(zd,"center",paint_edits_cpower);                       //  set mouse center power

   if (strEqu(event,"edge"))
      zdialog_fetch(zd,"edge",paint_edits_epower);                         //  set mouse edge power
   
   if (strEqu(event,"reset")) {
      sa_unselect();                                                       //  unselect current area if any
      cc = Fpxb->ww * Fpxb->hh * sizeof(uint16);                           //  allocate sa_pixmap[] for new area
      sa_pixmap = (uint16 *) zmalloc(cc,"paint_edits");
      memset(sa_pixmap,0,cc);                                              //  edge distance = 0 for all pixels

      sa_minx = 0;                                                         //  enclosing rectangle
      sa_maxx = Fpxb->ww;
      sa_miny = 0;
      sa_maxy = Fpxb->hh;

      sa_Npixel = Fpxb->ww * Fpxb->hh;
      sa_stat = 3;                                                         //  area status = complete
      sa_mode = mode_image;                                                //  area mode = whole image
      sa_calced = 1;                                                       //  edge calculation complete
      sa_blend = 10000;                                                    //  "blend width"
      sa_fww = Fpxb->ww;                                                   //  valid image dimensions
      sa_fhh = Fpxb->hh;
      areanumber++;                                                        //  next sequential number
   }
   
   return 1;
}


//  mouse function - adjust edit strength for areas within mouse radius
//  "edge distance" is increased for more strength, decreased for less

void paint_edits_mousefunc()
{
   int      ii, px, py, rx, ry, mrect;
   int      radius, radius2, cpower, epower;
   float    rad, rad2, power;

   if (! CEF) return;                                                      //  no active edit
   
   if (sa_stat != 3) return;                                               //  area gone?

   radius = paint_edits_radius;                                            //  pixel selection radius
   radius2 = radius * radius;
   cpower = paint_edits_cpower;
   epower = paint_edits_epower;

   mrect = 2 * radius;
   draw_mousearc(Mxposn,Myposn,mrect,mrect);                               //  show mouse selection circle

   if (LMclick || RMclick)                                                 //  mouse click, process normally
      return;

   if (Mbutton != 1 && Mbutton != 3)                                       //  button released
      return;
   
   Mxdrag = Mydrag = 0;                                                    //  neutralize drag                 v.12.11

   for (rx = -radius; rx <= radius; rx++)                                  //  loop every pixel in radius
   for (ry = -radius; ry <= radius; ry++)
   {
      rad2 = rx * rx + ry * ry;
      if (rad2 > radius2) continue;                                        //  outside radius
      px = Mxposn + rx;
      py = Myposn + ry;
      if (px < 0 || px > E1pxm->ww-1) continue;                            //  off the image edge
      if (py < 0 || py > E1pxm->hh-1) continue;

      ii = E1pxm->ww * py + px;
      rad = sqrt(rad2);
      power = cpower + rad / radius * (epower - cpower);                   //  power at pixel radius
      
      if (Mbutton == 1)                                                    //  left mouse button
      {                                                                    //  increase edit power
         sa_pixmap[ii] += power;
         if (sa_pixmap[ii] > sa_blend) sa_pixmap[ii] = sa_blend;
      }

      if (Mbutton == 3)                                                    //  right mouse button
      {                                                                    //  weaken edit power
         if (sa_pixmap[ii] <= power) sa_pixmap[ii] = 0;
         else sa_pixmap[ii] -= power;
      }
   }

   sa_minx = Mxposn - radius;                                              //  set temp. smaller area around mouse
   if (sa_minx < 0) sa_minx = 0;
   sa_maxx = Mxposn + radius;
   if (sa_maxx > E1pxm->ww) sa_maxx = E1pxm->ww;
   sa_miny = Myposn - radius;
   if (sa_miny < 0) sa_miny = 0;
   sa_maxy = Myposn + radius;
   if (sa_maxy > E1pxm->hh) sa_maxy = E1pxm->hh;
   sa_Npixel = (sa_maxx - sa_minx) * (sa_maxy - sa_miny);

   zdialog_send_event(CEF->zd,"blendwidth");                               //  notify edit dialog

   draw_mousearc(Mxposn,Myposn,mrect,mrect);                               //  show mouse selection circle

   sa_minx = 0;                                                            //  restore whole image area 
   sa_maxx = E1pxm->ww;
   sa_miny = 0;
   sa_maxy = E1pxm->hh;
   sa_Npixel = E1pxm->ww * E1pxm->hh;
   
   return;
}



