/**************************************************************************

   Fotoxx      edit photos and manage collections

   Copyright 2007 2008 2009 2010 2011 2012 Michael Cornelison
   Source URL: http://kornelix.com/fotoxx
   Contact: kornelix2@gmail.com
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see http://www.gnu.org/licenses/.

***************************************************************************/

#define EX extern                                                          //  enable extern declarations
#include "fotoxx.h"
#include <sys/wait.h>


/**************************************************************************

   Fotoxx image edit - Tools menu functions

***************************************************************************/


//  toggle foreground drawing and mouse cursor color

void  m_fg_color(GtkWidget *, cchar *)                                     //  v.12.03
{
   zfuncs::F1_help_topic = "color";                                        //  v.12.08
   if (fg_color == white) fg_color = black;
   else if (fg_color == black) fg_color = red;
   else if (fg_color == red) fg_color = green;
   else fg_color = white;
   mwpaint2();
   return;
}


/**************************************************************************/

//  menu function - batch rename files

namespace batch_rename 
{
   char     **filelist;
   char     newfname[100];
   int      sequence;
   int      increment;
};


void m_batch_rename(GtkWidget *, cchar *)                                  //  overhauled          v.12.11
{
   using namespace batch_rename;

   int   batch_rename_dialog_event(zdialog *zd, cchar *event);
   
   zdialog        *zd;
   char           *infile, *outfile;
   char           *pseq1, *pseq2;
   char           *pfile1, *pfile2;
   char           *prest1, *prest2;
   char           *pext1;
   char           chseq[20];
   int            nseq1, nseq2;
   int            zstat, filecount;
   int            ii, err, adder;
   struct stat    statb;

   if (checkmenulock()) return;                                            //  check menu lock
   if (mod_keep()) return;                                                 //  unsaved edits

   zfuncs::F1_help_topic = "batch_rename";

   zd = zdialog_new(ZTX("Batch Rename"),mWin,Bproceed,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","files","hb1",Bselectfiles,"space=10");
   zdialog_add_widget(zd,"label","fcount","hb1",ZTX("0 files selected"),"space=10");
   zdialog_add_widget(zd,"hbox","hb2","dialog","space=5");
   zdialog_add_widget(zd,"label","lab2","hb2",ZTX("new base name"),"space=10");
   zdialog_add_widget(zd,"entry","basename","hb2");
   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","lab31","hb3",ZTX("starting sequence"),"space=10");
   zdialog_add_widget(zd,"entry","sequence","hb3","100","scc=5");
   zdialog_add_widget(zd,"label","lab32","hb3",ZTX("increment"),"space=10");
   zdialog_add_widget(zd,"entry","increment","hb3","01","scc=3");
   
   filelist = 0;
   filecount = 0;

   zdialog_help(zd,"batch_rename");                                        //  zdialog help topic
   zdialog_run(zd,batch_rename_dialog_event);                              //  run dialog
   zstat = zdialog_wait(zd);                                               //  wait for completion
   zdialog_free(zd);
   
   if (zstat != 1) goto cleanup;                                           //  canceled
   if (! filelist) goto cleanup;                                           //  no files selected

   for (ii = 0; filelist[ii]; ii++);                                       //  count files selected
   filecount = ii;
   if (! filecount) goto cleanup;                                          //  none

   if (checkmenulock()) goto cleanup;
   
   pseq1 = strrchr(newfname,'#');                                          //  find embedded '#' characters in base name
   if (! pseq1) goto cleanup;
   for (nseq1 = 0; *pseq1-- == '#'; nseq1++);                              //  number of '#' characters found
   while (*pseq1 != '#') pseq1++;
   prest1 = pseq1 + nseq1;                                                 //  rest of base name after '#' chars.
   
   settab("F");
   gallery_monitor("stop");                                                //  stop tracking gallery updates

   write_popup_text("open","Renaming Files",500,200,mWin);                 //  status monitor popup window
   zmainloop();

   for (ii = 0; filelist[ii]; ii++)                                        //  loop selected files
   {
      infile = filelist[ii];
      pfile1 = strrchr(infile,'/');                                        //  old file name
      if (! pfile1) pfile1 = infile;
      else pfile1++;
      pext1 = strrchr(pfile1,'.');                                         //  file type, if any
      
      adder = sequence + ii * increment;                                   //  new file name sequence number
      sprintf(chseq,"%0*d",nseq1,adder);                                   //  character format, length >= '#' chars.
      nseq2 = strlen(chseq);
      
      outfile = strdupz(infile,110,"rename");                              //  /path/.../filename.ext
      pfile2 = outfile + (pfile1 - infile);
      strcpy(pfile2,newfname);                                             //  /path/.../base #### name
      pseq2 = pfile2 + (pseq1 - newfname);
      strcpy(pseq2,chseq);                                                 //  /path/.../base 0123
      prest2 = pseq2 + nseq2;
      strcpy(prest2,prest1);                                               //  /path/.../base 0123 name
      if (pext1) strcat(prest2,pext1);                                     //  /path/.../base 0123 ma,e.ext
      
      err = stat(outfile,&statb);
      if (! err) {
         snprintf(command,ccc,"%s %s",ZTX("new file already exists:"),outfile);
         write_popup_text("write",command);
         zfree(outfile);
         break;
      }

      snprintf(command,ccc,"cp -p \"%s\" \"%s\"",infile,outfile);          //  copy to new file

      write_popup_text("write",outfile);                                   //  log progress

      err = system(command);
      if (err) {
         snprintf(command,ccc,"%s %s",ZTX("Rename failed:"),wstrerror(err));
         write_popup_text("write",command);
         zfree(outfile);
         break;
      }

      err = remove(infile);                                                //  delete old file
      delete_search_index(infile);                                         //  remove from search index

      f_open(outfile,0,0,0,0);                                             //  open new file

      load_filemeta(outfile);                                              //  update search index
      update_search_index(outfile);

      zfree(outfile);
      zmainloop();
   }
   
   write_popup_text("write","COMPLETED");
   write_popup_text("close",0);

   gallery(curr_file,"init");                                              //  update gallery file list
   gallery_monitor("start");                                               //  resume tracking gallery updates

cleanup:

   if (filelist) {                                                         //  free memory
      for (ii = 0; filelist[ii]; ii++) 
         zfree(filelist[ii]);
      zfree(filelist);
      filelist = 0;
   }

   return;
}


//  dialog event and completion callback function

int batch_rename_dialog_event(zdialog *zd, cchar *event)
{
   using namespace batch_rename;

   int         ii;
   char        countmess[50];
   cchar       *errmess = ZTX("base name / sequence / increment not reasonable");
   
   if (strEqu(event,"files"))                                              //  select images to rename
   {
      if (filelist) {                                                      //  free prior list
         for (ii = 0; filelist[ii]; ii++) 
            zfree(filelist[ii]);
         zfree(filelist);
         filelist = 0;
      }

      zdialog_show(zd,0);                                                  //  hide parent dialog              v.12.11
      filelist = gallery_getfiles();                                       //  get list of files to convert
      zdialog_show(zd,1);

      if (filelist)                                                        //  count files selected
         for (ii = 0; filelist[ii]; ii++);
      else ii = 0;

      snprintf(countmess,50,ZTX("%d files selected"),ii);                  //  stuff count into dialog
      zdialog_stuff(zd,"fcount",countmess);
   }
   
   if (zd->zstat != 1) return 0;                                           //  wait for [proceed]
   
   if (! filelist) {                                                       //  no files selected
      zd->zstat = 0;                                                       //  keep dialog active
      return 0;
   }

   zdialog_fetch(zd,"basename",newfname,100);                              //  get base name and sequence params
   zdialog_fetch(zd,"sequence",sequence);
   zdialog_fetch(zd,"increment",increment);

   if (strlen(newfname) < 2 || sequence < 1 || increment < 1) {
      zd->zstat = 0;                                                       //  keep dialog alive
      zmessageACK(mWin,errmess);
   }
   
   if (! strchr(newfname,'#')) {
      zd->zstat = 0;
      zmessageACK(mWin,errmess);
   }
   
   return 0;
}


/**************************************************************************/

//  Batch file convert - change file type, resize, move

namespace batch_convert 
{
   char        **filelist;
   int         Freplace;
   int         maxw, maxh, Fnoexif, Fkeepwh, Fdelete;
   char        location[maxfcc];
   cchar       *pext, *ptype;
};


void m_batch_convert(GtkWidget *, cchar *)                                 //  overhauled                v.12.11
{
   using namespace batch_convert;

   int   batch_convert_dialog_event(zdialog *zd, cchar *event);
   
   zdialog        *zd;
   int            zstat, filecount;
   char           *infile, *outfile, *tempfile;
   char           *pfile, *pp;
   int            ii, lcc, fcc, err;
   int            ww, hh, bpc;
   float          scale, wscale, hscale;
   PXM            *pxmin, *pxmout;
   struct stat    statb;

   if (checkmenulock()) return;                                            //  menu locked               v.12.09
   if (mod_keep()) return;                                                 //  unsaved edits

   zfuncs::F1_help_topic = "batch_convert";

/***
               Batch Convert/Resize/Move

        [select files]  N files selected
        max. width [____]  height [____]  [x] no change
        new file type  (o) same  (o) JPG  (o) PNG  (o) TIF
        (o) replace originals   [x] remove EXIF
        (o) move to location 
        [browse] [_________________________________]
      
                                 [proceed]  [cancel]
***/

   zd = zdialog_new(ZTX("Batch Convert/Resize/Move"),mWin,Bproceed,Bcancel,null);

   zdialog_add_widget(zd,"hbox","hbf","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","files","hbf",Bselectfiles,"space=5");
   zdialog_add_widget(zd,"label","fcount","hbf","0 files selected","space=10");

   zdialog_add_widget(zd,"hbox","hbwh","dialog");
   zdialog_add_widget(zd,"label","labw","hbwh",ZTX("max. width"),"space=5");
   zdialog_add_widget(zd,"entry","maxw","hbwh","1000","scc=5");
   zdialog_add_widget(zd,"label","space","hbwh",0,"space=5");
   zdialog_add_widget(zd,"label","labh","hbwh",ZTX("height"),"space=5");
   zdialog_add_widget(zd,"entry","maxh","hbwh","700","scc=5");
   zdialog_add_widget(zd,"check","keepwh","hbwh",ZTX("no change"),"space=12");  
   
   zdialog_add_widget(zd,"hbox","hbtyp","dialog",0,"space=4");
   zdialog_add_widget(zd,"label","labtyp","hbtyp",ZTX("new file type"),"space=5");
   zdialog_add_widget(zd,"radio","same","hbtyp",ZTX("same"),"space=6");
   zdialog_add_widget(zd,"radio","jpg","hbtyp","JPG","space=6");
   zdialog_add_widget(zd,"radio","png","hbtyp","PNG","space=6");
   zdialog_add_widget(zd,"radio","tif","hbtyp","TIF","space=6");

   zdialog_add_widget(zd,"hbox","hbrep","dialog",0,"space=5");
   zdialog_add_widget(zd,"vbox","vbrep1","hbrep",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vbrep2","hbrep",0,"homog|space=5");
   zdialog_add_widget(zd,"radio","replace","vbrep1",ZTX("replace originals"));
   zdialog_add_widget(zd,"radio","move","vbrep1",ZTX("move to location"));
   zdialog_add_widget(zd,"check","noexif","vbrep2",ZTX("remove EXIF"));
   zdialog_add_widget(zd,"check","delete","vbrep2",ZTX("delete originals"));

   zdialog_add_widget(zd,"hbox","hbloc","dialog",0,"space=3|expand");
   zdialog_add_widget(zd,"button","browse","hbloc",Bbrowse,"space=5");
   zdialog_add_widget(zd,"entry","location","hbloc",0,"scc=30|expand");
                                                                           //  defaults:
   zdialog_stuff(zd,"same",1);                                             //  same file type
   zdialog_stuff(zd,"keepwh",1);                                           //  same size
   zdialog_stuff(zd,"replace",0);                                          //  replace - no
   zdialog_stuff(zd,"move",1);                                             //  move - yes
   zdialog_stuff(zd,"noexif",1);                                           //  keep EXIF data - no
   zdialog_stuff(zd,"delete",0);                                           //  delete original - no
   
   filelist = 0;
   filecount = 0;

   zdialog_help(zd,"batch_convert");                                       //  zdialog help topic
   zdialog_restore_inputs(zd);                                             //  preload prior user inputs    v.12.09
   zdialog_run(zd,batch_convert_dialog_event,"parent");                    //  run dialog
   zstat = zdialog_wait(zd);                                               //  wait for completion
   zdialog_free(zd);

   if (zstat != 1) goto cleanup;                                           //  canceled
   if (! filelist) goto cleanup;                                           //  no files selected

   for (ii = 0; filelist[ii]; ii++);                                       //  count files selected
   filecount = ii;
   if (! filecount) goto cleanup;                                          //  none

   if (checkmenulock()) goto cleanup;   

   settab("F");
   
   gallery_monitor("stop");                                                //  stop tracking gallery updates

   write_popup_text("open","Processing files",500,200,mWin);               //  status monitor popup window

   lcc = strlen(location);
   if (! Freplace && location[lcc-1] == '/') lcc--;                        //  remove trailing '/'

   for (ii = 0; filelist[ii]; ii++)                                        //  loop selected files
   {
      infile = filelist[ii];

      if (Freplace)
         outfile = strdupz(infile,8,"batch_convert");                      //  new file = old file
      else {
         pfile = strrchr(infile,'/');
         if (! pfile) continue;
         fcc = strlen(pfile);
         outfile = strdupz(location,fcc+9,"batch_convert");
         strcpy(outfile+lcc,pfile);                                        //  new file at location
      }
      
      write_popup_text("write",outfile);                                   //  log each new file
      zmainloop();

      pp = strrchr(outfile,'/');                                           //  find existing file .ext
      pp = strrchr(pp,'.');
      if (! pp) pp = pp + strlen(pp);
      if (strlen(pp) > 5) pp = pp + strlen(pp);

      pext = 0;
      if (ptype) pext = ptype;                                             //  new .ext from user
      else {
         if (strstr(".JPG .jpg .JPEG .jpeg",pp)) pext = ".jpg";            //  new .ext from existing .ext
         if (strstr(".PNG .png",pp)) pext = ".png";
         if (strstr(".TIF .TIFF .tif .tiff",pp)) pext = ".tif";
      }
      if (! pext) pext = ".jpg";                                           //  other = .jpg
      
      strcpy(pp,pext);                                                     //  set new file .ext

      if (! Freplace) {
         err = stat(outfile,&statb);                                       //  if move, check if file exists
         if (! err) {
            write_popup_text("write",ZTX("file already exists"));
            zfree(outfile);
            continue;
         }
      }

      pxmin = PXM_load(infile,0);
      if (! pxmin) {
         write_popup_text("write",ZTX("file type not supported"));
         zfree(outfile);
         continue;
      }

      bpc = f_load_bpc;                                                    //  keep file bpc

      ww = pxmin->ww;                                                      //  input file size
      hh = pxmin->hh;

      if (Fkeepwh) pxmout = pxmin;                                         //  no change in size
      else {
         wscale = hscale = 1.0;
         if (ww > maxw) wscale = 1.0 * maxw / ww;                          //  compute new size
         if (hh > maxh) hscale = 1.0 * maxh / hh;
         if (wscale < hscale) scale = wscale;
         else scale = hscale;
         if (scale > 0.999) pxmout = pxmin;                                //  no change
         else {
            ww = ww * scale;
            hh = hh * scale;
            pxmout = PXM_rescale(pxmin,ww,hh);                             //  rescale file
            PXM_free(pxmin);
         }
      }

      tempfile = strdupz(infile,12,"batch_convert");                       //  temp file needed for EXIF/IPTC copy
      pp = strrchr(tempfile,'.');
      strcpy(pp,"-temp");
      strcpy(pp+5,pext);
      
      if (strEqu(".tif",pext))                                             //  write rescaled file to temp file
         err = PXM_TIFF_save(pxmout,tempfile,bpc);                         //  v.12.10
      else if (strEqu(".png",pext))
         err = PXM_PNG_save(pxmout,tempfile,bpc);
      else 
         err = PXM_ANY_save(pxmout,tempfile);
      
      if (err) {
         write_popup_text("write",ZTX("cannot create new file"));
         zfree(outfile);
         zfree(tempfile);
         PXM_free(pxmout);
         continue;
      }

      if (! Fnoexif)                                                       //  copy EXIF/IPTC if requested
         exif_copy(infile,tempfile,0,0,0);
      
      snprintf(command,ccc,"cp -p \"%s\" \"%s\" ",tempfile,outfile);       //  copy tempfile to new file
      err = system(command);
      if (err) write_popup_text("write",wstrerror(err));

      remove(tempfile);                                                    //  remove tempfile

      if (Fdelete && ! err && strNeq(infile,outfile)) {                    //  delete originals (move or change format)
         remove(infile);                                                   //                                  v.12.12
         delete_search_index(infile);                                      //  remove from search index        v.12.10
      }
      
      if (! err) {
         load_filemeta(outfile);                                           //  update new file search index    v.12.10
         update_search_index(outfile);
      }

      zfree(outfile);
      zfree(tempfile);
      PXM_free(pxmout);
   }

   write_popup_text("write","COMPLETED");
   
   gallery(curr_file,"init");                                              //  update gallery file list
   gallery_monitor("start");                                               //  resume tracking gallery updates

cleanup:

   if (filelist) {                                                         //  free memory
      for (ii = 0; filelist[ii]; ii++) 
         zfree(filelist[ii]);
      zfree(filelist);
      filelist = 0;
   }

   return;
}


//  dialog event and completion callback function

int batch_convert_dialog_event(zdialog *zd, cchar *event)
{
   using namespace batch_convert;

   char     warnmess[500], countmess[50];
   char     *ploc;
   int      ii, cc, yn, err;
   struct stat  statb;

   if (strEqu(event,"keepwh")) {                                           //  keep width/height               v.12.12
      zdialog_fetch(zd,"keepwh",Fkeepwh);
      if (Fkeepwh) {
         zdialog_stuff(zd,"maxw","");
         zdialog_stuff(zd,"maxh","");
      }
   }
   
   if (strEqu(event,"files"))                                              //  select images to convert
   {
      if (filelist) {                                                      //  free prior list
         for (ii = 0; filelist[ii]; ii++) 
            zfree(filelist[ii]);
         zfree(filelist);
         filelist = 0;
      }

      zdialog_show(zd,0);                                                  //  hide parent dialog              v.12.11
      filelist = gallery_getfiles();                                       //  get list of files to convert
      zdialog_show(zd,1);

      if (filelist)                                                        //  count files selected
         for (ii = 0; filelist[ii]; ii++);
      else ii = 0;

      snprintf(countmess,50,ZTX("%d files selected"),ii);                  //  update dialog
      zdialog_stuff(zd,"fcount",countmess);
   }
   
   if (strEqu(event,"browse"))
   {
      ploc = zgetfile1(ZTX("Select directory"),"folder",curr_dirk);        //  get directory
      if (! ploc) return 0;
      zdialog_stuff(zd,"location",ploc);
      zfree(ploc);
      zdialog_stuff(zd,"move",1);
   }

   if (zd->zstat != 1) return 0;                                           //  wait for [proceed]
   
   if (! filelist) {                                                       //  no files selected
      zd->zstat = 0;                                                       //  keep dialog active
      return 0;
   }

   zdialog_fetch(zd,"maxw",maxw);                                          //  new max width
   zdialog_fetch(zd,"maxh",maxh);                                          //  new max height
   zdialog_fetch(zd,"keepwh",Fkeepwh);                                     //  keep width/height option
   zdialog_fetch(zd,"replace",Freplace);                                   //  replace originals y/n
   zdialog_fetch(zd,"noexif",Fnoexif);                                     //  copy EXIF/IPTC y/n
   zdialog_fetch(zd,"delete",Fdelete);                                     //  delete originals y/n
   
   zdialog_fetch(zd,"same",ii);                                            //  get new file type
   if (ii) ptype = 0;
   zdialog_fetch(zd,"jpg",ii);
   if (ii) ptype = ".jpg";
   zdialog_fetch(zd,"png",ii);
   if (ii) ptype = ".png";
   zdialog_fetch(zd,"tif",ii);
   if (ii) ptype = ".tif";

   zdialog_fetch(zd,"location",location,maxfcc);                           //  output location (Freplace = 0)
   strTrim2(location,location);                                            //  trim leading and trailing blanks

   if (Freplace) {                                                         //  keep or reduce size             v.12.12
      if (Fkeepwh) yn = zmessageYN(mWin,ZTX("replace original files?"));
      else yn = zmessageYN(mWin,ZTX("replace original files? (max. %d x %d)"),maxw,maxh);
      if (! yn) {
         zd->zstat = 0;                                                    //  if NO keep dialog active
         return 0;
      }
   }
   else {
      if (Fkeepwh)                                                         //  v.12.12
         snprintf(warnmess,500,ZTX("copy files? \n"
                                   " to location %s"),location);
      else
         snprintf(warnmess,500,ZTX("copy files? (max. %d x %d) \n"
                                   " to location %s"),maxw,maxh,location);
      if (Fdelete) {
         cc = 500 - strlen(warnmess) - 1;
         strncat(warnmess,ZTX("\n and delete originals"),cc);
      }
      
      yn = zmessageYN(mWin,warnmess);
      if (! yn) {
         zd->zstat = 0;
         return 0;
      }
   }

   if (! Freplace) {
      err = stat(location,&statb);
      if (err || ! S_ISDIR(statb.st_mode)) {
         zmessageACK(mWin,ZTX("location is not a valid directory"));
         zd->zstat = 0;
         return 0;
      }
   }

   if (! Fkeepwh && (maxw < 20 || maxh < 20)) {
      zmessageACK(mWin,ZTX("max. size %d x %d is not reasonable"),maxw,maxh);
      zd->zstat = 0;
      return 0;
   }
   
   return 0;
}


/**************************************************************************/

//  convert multiple RAW files to tiff, jpeg, or png

namespace batch_conv_raw
{
   char     **filelist;
   char     dcraw_command[100] = "";
   cchar    *filetype;
   int      bpc;
};


void  m_batch_conv_raw(GtkWidget *, cchar *menu)                           //  v.12.10
{
   using namespace batch_conv_raw;

   int  batch_conv_raw_dialog_event(zdialog *zd, cchar *event);

   zdialog  *zd;
   cchar    *title = ZTX("Batch Convert RAW Files");
   char     *tiffile, *outfile, *pp;
   int      zstat, ii, err, filecount, ftype;

   if (checkmenulock()) return;                                            //  check menu lock
   if (mod_keep()) return;                                                 //  unsaved edits

/***
         Batch Convert RAW Files
         
         [Select Files]  N files selected
         output file type  (o) JPEG  (o) PNG-8  (o) PNG-16  (o) TIFF-8  (o) TIFF-16

         white balance  (o) camera  (o) fixed  (o) calculated
         interpolation  (o) bilinear  (o) VNG  (o) PPG  (o) AHD
         color space  (o) raw  (o) sRGB  (o) Adobe  (o) wide  (o) Kodak  (o) XYZ
         gamma curve  (o) default  (o) sRGB  (o) linear
         dcraw command  [ dcraw -T -6 -w -q 0 -o 1 -g 2.222 4.5 ] [defaults]
         
                                                         [proceed] [cancel]
***/

   zfuncs::F1_help_topic = "batch_convert_raw";

   zd = zdialog_new(title,mWin,Bproceed,Bcancel,null);

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","files","hb1",Bselectfiles,"space=5");
   zdialog_add_widget(zd,"label","fcount","hb1",ZTX("0 files selected"),"space=10");

   zdialog_add_widget(zd,"hbox","hb2","dialog");
   zdialog_add_widget(zd,"label","labtype","hb2",ZTX("output file type"),"space=5");
   zdialog_add_widget(zd,"radio","JPEG","hb2","JPEG","space=3");
   zdialog_add_widget(zd,"radio","PNG-8","hb2","PNG-8","space=3");
   zdialog_add_widget(zd,"radio","PNG-16","hb2","PNG-16","space=3");
   zdialog_add_widget(zd,"radio","TIFF-8","hb2","TIFF-8","space=3");
   zdialog_add_widget(zd,"radio","TIFF-16","hb2","TIFF-16","space=3");
   
   zdialog_add_widget(zd,"hsep","hsep1","dialog",0,"space=5");
   zdialog_add_widget(zd,"hbox","hbvb","dialog");
   zdialog_add_widget(zd,"vbox","vb1","hbvb",0,"space=5|homog");
   zdialog_add_widget(zd,"vbox","vb2","hbvb",0,"homog");
   
   zdialog_add_widget(zd,"label","labwbal","vb1",ZTX("white balance"));
   zdialog_add_widget(zd,"label","labinterp","vb1",ZTX("interpolation"));
   zdialog_add_widget(zd,"label","labcolor","vb1",ZTX("color space"));
   zdialog_add_widget(zd,"label","labgamma","vb1",ZTX("gamma curve"));

   zdialog_add_widget(zd,"hbox","hb3","vb2");
   zdialog_add_widget(zd,"radio","-w","hb3",ZTX("camera"),"space=3");
   zdialog_add_widget(zd,"radio","-W","hb3",ZTX("fixed"),"space=3");
   zdialog_add_widget(zd,"radio","-a","hb3",ZTX("calculated"),"space=3");

   zdialog_add_widget(zd,"hbox","hb4","vb2");
   zdialog_add_widget(zd,"radio","-q 0","hb4","bilinear","space=3");
   zdialog_add_widget(zd,"radio","-q 1","hb4","VNG","space=3");
   zdialog_add_widget(zd,"radio","-q 2","hb4","PPG","space=3");
   zdialog_add_widget(zd,"radio","-q 3","hb4","AHD","space=3");
   
   zdialog_add_widget(zd,"hbox","hb5","vb2");
   zdialog_add_widget(zd,"radio","-o 0","hb5","raw","space=3");
   zdialog_add_widget(zd,"radio","-o 1","hb5","sRGB","space=3");
   zdialog_add_widget(zd,"radio","-o 2","hb5","Adobe","space=3");
   zdialog_add_widget(zd,"radio","-o 3","hb5","Wide","space=3");
   zdialog_add_widget(zd,"radio","-o 4","hb5","Kodak","space=3");
   zdialog_add_widget(zd,"radio","-o 5","hb5","XYZ","space=3");
   
   zdialog_add_widget(zd,"hbox","hb6","vb2");
   zdialog_add_widget(zd,"radio","-g 2.222 4.5","hb6","default","space=3");
   zdialog_add_widget(zd,"radio","-g 2.4 12.92","hb6","sRGB","space=3");
   zdialog_add_widget(zd,"radio","-g 1.0 1.0","hb6","linear","space=3");

   zdialog_add_widget(zd,"hbox","hb7","dialog",0,"space=3");
   zdialog_add_widget(zd,"entry","dcraw_command","hb7",0,"space=5|expand");
   zdialog_add_widget(zd,"button","defaults","hb7","defaults","space=5");

   filelist = 0;                                                           //  set no files selected

   zdialog_restore_inputs(zd);                                             //  default to prior inputs   
   batch_conv_raw_dialog_event(zd,"init");                                 //  set defaults if no prior

   zdialog_help(zd,"batch_convert_raw");
   zdialog_run(zd,batch_conv_raw_dialog_event,"save");                     //  run dialog
   zstat = zdialog_wait(zd);                                               //  wait for completion
   zdialog_free(zd);
   if (zstat != 1) goto cleanup;

   if (! filelist) goto cleanup;
   for (ii = 0; filelist[ii]; ii++);                                       //  count selected files
   filecount = ii;
   if (! filecount) goto cleanup;
   if (checkmenulock()) goto cleanup;
   
   printf("dcraw command: %s \n",dcraw_command);

   settab("F");
   gallery_monitor("stop");                                                //  stop tracking gallery updates

   write_popup_text("open","Converting RAW files",500,200,mWin);           //  status monitor popup window
   zmainloop();
   
   rawfile = 0;                                                            //  v.12.11
   
   for (ii = 0; ii < filecount; ii++)                                      //  loop all RAW files
   {
      rawfile = filelist[ii];                                              //  filename.raw

      write_popup_text("write",rawfile);                                   //  write input raw file to log window
      zmainloop();

      ftype = image_file_type(rawfile);
      if (ftype != 3) {
         write_popup_text("write"," *** unknown RAW file type");
         continue;
      }
      
      tiffile = raw_to_tiff(rawfile);

      snprintf(command,ccc,"%s \"%s\" ",dcraw_command,rawfile);            //  convert filename.raw to filename.tiff
      err = system(command);
      if (err) {
         write_popup_text("write"," *** raw conversion failed");           //  conversion failed
         write_popup_text("write",wstrerror(err));
         zmainloop();
         zfree(tiffile);
         continue;
      }

      err = f_open(tiffile,0,0,0,0);
      if (err) {
         write_popup_text("write"," *** no dcraw .tiff output");
         zmainloop();
         zfree(tiffile);
         continue;
      }

      outfile = strdupz(tiffile,0,"batch-raw");

      if (strNeq(filetype,"tiff")) {                                       //  tiff-16 was not wanted
         pp = strrchr(outfile,'.');                                        //  filename.tiff > filename.xxx
         if (pp) strcpy(pp+1,filetype);
         f_save(outfile,filetype,bpc,0);                                   //  save to final file type & bpc
         remove(tiffile);                                                  //  delete tiff-16 file
         err = f_open(outfile,0,0,0,0);
         if (err) 
            write_popup_text("write"," *** type conversion failed");       //  v.12.11
      }

      write_popup_text("write",outfile);                                   //  write output file to log window
      zfree(outfile);
      zfree(tiffile);
      zmainloop();
   }

   write_popup_text("write","COMPLETED");
   write_popup_text("close");

   gallery(curr_file,"init");                                              //  update gallery file list
   gallery_monitor("start");                                               //  resume tracking gallery updates

cleanup:                                                                   //  clean up and return

   if (filelist) {
      for (ii = 0; filelist[ii]; ii++)                                     //  free memory
         zfree(filelist[ii]);
      zfree(filelist);
      filelist = 0;
   }

   rawfile = 0;                                                            //  v.12.11
   return;
}


//  dialog event and completion callback function

int batch_conv_raw_dialog_event(zdialog *zd, cchar *event)                 //  v.12.10
{
   using namespace batch_conv_raw;

   int            ii, Fdcraw = 0;
   char           countmess[50];
   static cchar   *wbal = 0, *intp = 0, *color = 0, *gamma = 0;
   
   if (! wbal) {                                                           //  get current options if not already
      wbal = "-w";
      zdialog_fetch(zd,"-W",ii);
      if (ii) wbal = "-W";
      zdialog_fetch(zd,"-a",ii);
      if (ii) wbal = "-a";
      Fdcraw++;
   }
   
   if (! intp) {
      intp = "-q 0";
      zdialog_fetch(zd,"-q 1",ii);
      if (ii) intp = "-q 1";
      zdialog_fetch(zd,"-q 2",ii);
      if (ii) intp = "-q 2";
      zdialog_fetch(zd,"-q 3",ii);
      if (ii) intp = "-q 3";
      Fdcraw++;
   }
   
   if (! color) {
      color = "-o 1";
      zdialog_fetch(zd,"-o 0",ii);
      if (ii) color = "-o 0";
      zdialog_fetch(zd,"-o 2",ii);
      if (ii) color = "-o 2";
      zdialog_fetch(zd,"-o 3",ii);
      if (ii) color = "-o 3";
      zdialog_fetch(zd,"-o 4",ii);
      if (ii) color = "-o 4";
      zdialog_fetch(zd,"-o 5",ii);
      if (ii) color = "-o 5";
      Fdcraw++;
   }
   
   if (! gamma) {
      gamma = "-g 2.222 4.5";
      zdialog_fetch(zd,"-g 2.4 12.92",ii);
      if (ii) gamma = "-g 2.4 12.92";
      zdialog_fetch(zd,"-g 1.0 1.0",ii);
      if (ii) gamma = "-g 1.0 1.0";
      Fdcraw++;
   }

   if (strEqu(event,"files"))                                              //  select images to rename
   {
      if (filelist) {                                                      //  free prior list
         for (ii = 0; filelist[ii]; ii++) 
            zfree(filelist[ii]);
         zfree(filelist);
         filelist = 0;
      }

      zdialog_show(zd,0);                                                  //  hide parent dialog              v.12.11
      filelist = gallery_getfiles();                                       //  get list of files to convert
      zdialog_show(zd,1);

      if (filelist)                                                        //  count files in list
         for (ii = 0; filelist[ii]; ii++);
      else ii = 0;

      snprintf(countmess,50,ZTX("%d files selected"),ii);                  //  stuff count into dialog
      zdialog_stuff(zd,"fcount",countmess);
   }
   
   if (strEqu(event,"defaults"))                                           //  default all parameters
   {
      zdialog_stuff(zd,"JPEG",0);
      zdialog_stuff(zd,"PNG-8",0);
      zdialog_stuff(zd,"PNG-16",0);
      zdialog_stuff(zd,"TIFF-8",0);
      zdialog_stuff(zd,"TIFF-16",1);                                       //  output = TIFF-16

      zdialog_stuff(zd,"-w",1);                                            //  color balance = camera value
      zdialog_stuff(zd,"-W",0);
      zdialog_stuff(zd,"-a",0);
      wbal = "-w";

      zdialog_stuff(zd,"-q 0",1);                                          //  interpolation = bilinear
      zdialog_stuff(zd,"-q 1",0);
      zdialog_stuff(zd,"-q 2",0);
      zdialog_stuff(zd,"-q 3",0);
      intp = "-q 0";

      zdialog_stuff(zd,"-o 0",0);
      zdialog_stuff(zd,"-o 1",1);                                          //  color space = sRGB
      zdialog_stuff(zd,"-o 2",0);
      zdialog_stuff(zd,"-o 3",0);
      zdialog_stuff(zd,"-o 4",0);
      zdialog_stuff(zd,"-o 5",0);
      color = "-o 1";

      zdialog_stuff(zd,"-g 2.222 4.5",1);                                  //  gamma = default
      zdialog_stuff(zd,"-g 2.4 12.92",0);
      zdialog_stuff(zd,"-g 1.0 1.0",0);
      gamma = "-g 2.222 4.5";

      Fdcraw++;
   }
   
   if (strstr("-w -W -a",event)) {                                         //  event is one of: -w -W -a
      zdialog_fetch(zd,event,ii);
      if (ii) {
         wbal = event;
         Fdcraw++;
      }
   }

   if (strstr(event,"-q")) {                                               //  event is: -q N
      zdialog_fetch(zd,event,ii);
      if (ii) {
         intp = event;
         Fdcraw++;
      }
   }

   if (strstr(event,"-o")) {                                               //  event is: -o N
      zdialog_fetch(zd,event,ii);
      if (ii) {
         color = event;
         Fdcraw++;
      }
   }

   if (strstr(event,"-g")) {                                               //  event is: -g N1 N2
      zdialog_fetch(zd,event,ii);
      if (ii) {
         gamma = event;
         Fdcraw++;
      }
   }

   if (Fdcraw || ! *dcraw_command)                                         //  initialize or update dcraw command
   {
      sprintf(dcraw_command,"dcraw -T -6 %s %s %s %s",wbal,intp,color,gamma);
      zdialog_stuff(zd,"dcraw_command",dcraw_command);
   }

   if (zd->zstat != 1) return 0;                                           //  wait for [proceed]

   if (! filelist) {                                                       //  no files selected
      zd->zstat = 0;                                                       //  keep dialog active
      return 0;
   }

   zdialog_fetch(zd,"JPEG",ii);                                            //  get file type, color depth
   if (ii) { filetype = "jpg"; bpc = 8; }
   zdialog_fetch(zd,"PNG-8",ii);
   if (ii) { filetype = "png"; bpc = 8; }
   zdialog_fetch(zd,"PNG-16",ii);
   if (ii) { filetype = "png"; bpc = 16; }
   zdialog_fetch(zd,"TIFF-8",ii);
   if (ii) { filetype = "tif"; bpc = 8; }
   zdialog_fetch(zd,"TIFF-16",ii);
   if (ii) { filetype = "tiff"; bpc = 16; }

   return 0;
}


/**************************************************************************/

//  Manage Collections - create, view, edit named collections of images

namespace edit_coll {
   char        *coll_file = 0;                                             //  current collection file
   char        *coll_fname = 0;                                            //  collection name = file base name
   char        **sel_files = 0;                                            //  selected image files
   int         Nselect = 0;                                                //  count
   int         Nth_clicked;                                                //  clicked thumbnail in gallery
   zdialog     *zd2;
   int         selectbusy = 0;
};


void m_manage_coll(GtkWidget *, cchar *)                                   //  v.13.01 overhauled
{
   using namespace edit_coll;

   int   manage_coll_dialog_event(zdialog *zd, cchar *event);              //  manage collectiond dialog event func

   zfuncs::F1_help_topic = "manage_collections";

   if (checkmenulock()) return;                                            //  menu locked
   if (mod_keep()) return;                                                 //  unsaved edits
   if (zd_editcoll) return;                                                //  dialog already active

/***
       _______________________________________________
      |                                               |
      |        Manage Collections                     |
      |                                               |
      |  [New]     Start a new collection             |
      |  [Choose]  Choose a collection to view/edit   |
      |  [Select]  Select images to add/move/remove   |
      |  [Delete]  Delete a collection                |
      |                                               |
      |                                      [Done]   |
      |_______________________________________________|

***/

   zd_editcoll = zdialog_new(ZTX("Manage Collections"),mWin,Bdone,null);
   zdialog *zd = zd_editcoll;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"vbox","vb1","hb1",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"homog|space=5");
   zdialog_add_widget(zd,"button","new","vb1",Bnew);
   zdialog_add_widget(zd,"button","choose","vb1",Bchoose);
   zdialog_add_widget(zd,"button","select","vb1",Bselect);
   zdialog_add_widget(zd,"button","delete","vb1",Bdelete);
   zdialog_add_widget(zd,"hbox","hbnew","vb2");
   zdialog_add_widget(zd,"label","labnew","hbnew",ZTX("Start new collection"));
   zdialog_add_widget(zd,"hbox","hbview","vb2");
   zdialog_add_widget(zd,"label","labview","hbview",ZTX("Choose a collection to view/edit"));
   zdialog_add_widget(zd,"hbox","hbsel","vb2");
   zdialog_add_widget(zd,"label","labsel","hbsel",ZTX("Select images to add/move/remove"));
   zdialog_add_widget(zd,"hbox","hbdel","vb2");
   zdialog_add_widget(zd,"label","labdel","hbdel",ZTX("Delete a collection"));

   zdialog_help(zd,"manage_collections");
   zdialog_run(zd,manage_coll_dialog_event,"save");
   zdialog_wait(zd);
   zd_editcoll = 0;

   return;
}


//  manage collections dialog event and completion function

int manage_coll_dialog_event(zdialog *zd, cchar *event)
{
   using namespace edit_coll;

   int edit_coll_action_event(zdialog *zd, cchar *event);
   
   cchar    *help = ZTX("To insert the selected images, press \n"
                        "[Insert] and click gallery position. \n"
                        "To remove the images, press [Delete]");

   cchar       *newcoll = ZTX("New Collection");
   cchar       *choosecoll = ZTX("Choose Collection");
   cchar       *delcoll = ZTX("Delete Collection");
   cchar       *selact = ZTX("Selected Images Action");
   char        *pp, *file, countmess[60];
   FILE        *fid;
   int         ii;

   if (zd->zstat)                                                          //  dialog complete
   {
      zdialog_free(zd);
      if (coll_fname) zfree(coll_fname);                                   //  no current collection
      coll_fname = 0;
      if (coll_file) zfree(coll_file);
      coll_file = 0;
      if (zd2) zdialog_free(zd2);
      return 0;
   }
   
   if (strEqu(event,"new"))                                                //  start a new collection for editing
   {
      file = zgetfile1(newcoll,"save",collections_dirk);                   //  choose file name
      if (! file) return 0;

      fid = fopen(file,"w");                                               //  write empty collection file
      if (! fid) return 0;
      fprintf(fid,"Fotoxx Collection \n");
      fclose(fid);
     
      if (coll_fname) zfree(coll_fname);                                   //  free current collection
      if (coll_file) zfree(coll_file);

      coll_file = file;                                                    //  set new current collection
      pp = strrchr(coll_file,'/');                                         //  file name = collection name
      coll_fname = strdupz(pp+1,0,"manage_coll");

      settab("F");                                                         //  no gallery
      return 0;
   }

   if (strEqu(event,"choose"))                                             //  choose a collection to view or edit
   {
      file = zgetfile1(choosecoll,"open",collections_dirk);                //  choose collection file
      if (! file) return 0;

      if (coll_fname) zfree(coll_fname);                                   //  free current collection
      if (coll_file) zfree(coll_file);

      coll_file = file;                                                    //  set new current collection
      pp = strrchr(coll_file,'/');                                         //  file name = collection name
      coll_fname = strdupz(pp+1,0,"manage_coll");

      navi::gallerytype = 4;
      gallery(coll_file,"initF");                                          //  generate collection gallery
      gallery(0,"paint");
      return 0;
   }

   if (strEqu(event,"select"))                                             //  select files to add/move/remove
   {
      if (selectbusy) return 0;                                            //  block redundant dialog
      selectbusy = 1;
      
      if (Nselect) {
         for (ii = 0; ii < Nselect; ii++)                                  //  clear prior selected files
            zfree(sel_files[ii]);
         zfree(sel_files);
         sel_files = 0;
         Nselect = 0;
      }
      
      zd->disabled = 0;                                                    //  kludge, keep this dialog active
      sel_files = gallery_getfiles();                                      //  select new list of image files
      if (! sel_files) {
         selectbusy = 0;
         return 0;
      }

      for (ii = 0; sel_files[ii]; ii++);                                   //  count files selected
      Nselect = ii;
      
      navi::gallerytype = 4;
      gallery(coll_file,"initF");                                          //  gallery = current collection
      gallery(0,"paint");

      zd2 = zdialog_new(selact,mWin,Bdone,null);
      zdialog_add_widget(zd2,"hbox","hb0","dialog",0,"space=5");
      zdialog_add_widget(zd2,"label","labhelp","hb0",help,"space=5");
      zdialog_add_widget(zd2,"hbox","hb1","dialog",0,"space=3");           //  dialog for selected images action:
      zdialog_add_widget(zd2,"vbox","vb1","hb1",0,"homog|space=5");        //    insert or delete the images
      zdialog_add_widget(zd2,"vbox","vb2","hb1",0,"homog|space=5");
      zdialog_add_widget(zd2,"button","insert","vb1",Binsert);
      zdialog_add_widget(zd2,"button","delete","vb1",Bdelete);
      zdialog_add_widget(zd2,"hbox","hbins","vb2");
      zdialog_add_widget(zd2,"label","labins","hbins",ZTX("insert images at click position"));
      zdialog_add_widget(zd2,"hbox","hbdel","vb2");
      zdialog_add_widget(zd2,"label","labdel","hbdel",ZTX("remove the images"));
      zdialog_add_widget(zd2,"hbox","hbnsel","dialog",0,"space=10");
      zdialog_add_widget(zd2,"label","labnsel","hbnsel",0,"space=10");

      snprintf(countmess,60,ZTX("%d images selected"),Nselect);            //  stuff count of images selected
      zdialog_stuff(zd2,"labnsel",countmess);

      zdialog_help(zd2,"manage_collections");                              //  start dialog
      zdialog_run(zd2,edit_coll_action_event,"save");
      zdialog_wait(zd2);
      zdialog_free(zd2);
      selectbusy = 0;
      return 0;
   }

   if (strEqu(event,"delete"))                                             //  choose collection to delete
   {
      if (coll_fname) zfree(coll_fname);                                   //  no current collection
      coll_fname = 0;
      if (coll_file) zfree(coll_file);
      coll_file = 0;

      file = zgetfile1(delcoll,"open",collections_dirk);                   //  choose collection file
      if (! file) return 0;
      if (! zmessageYN(mWin,ZTX("delete %s ?"),file)) return 0;
      remove(file);
      zfree(file);

      settab("F");                                                         //  no gallery
      return 0;
   }
   
   return 0;
}


//  image action dialog event and completion function

int edit_coll_action_event(zdialog *zd2, cchar *event)
{
   using namespace edit_coll;
   
   FILE     *fid;
   int      ii, jj, Nth;
   char     *file, tempfile[200];

   if (zd2->zstat) {                                                       //  done / cancel
      edit_coll_mousecapture = 0;
      return 0;
   }
   
   if (strEqu(event,"insert"))                                             //  enable mouse click to insert new images
   {
      if (! coll_file || strNeq(coll_file,navi::galleryname)) {            //  insure gallery = collection
         zmessageACK(mWin,ZTX("Choose a collection first"));
         return 0;
      }

      if (navi::nfiles > 0) {
         edit_coll_mousecapture = 1;
         return 0;
      }
      else {
         event = "clicked";                                                //  handle empty case - no click needed
         Nth_clicked = 0;
      }
   }
   
   if (strEqu(event,"clicked"))                                            //  after click ... 
   {
      if (! coll_file || strNeq(coll_file,navi::galleryname))              //  insure gallery = collection
         return 0;

      Nth = Nth_clicked;                                                   //  clicked thumbnail from mouse func

      strcpy(tempfile,collections_dirk);                                   //  temp file for copying
      strcat(tempfile,"tempfile");

      fid = fopen(tempfile,"w");                                           //  open/write temp file
      if (! fid) {
         zmessageACK(mWin,"%s",strerror(errno));                           //  report file error
         zdialog_destroy(zd2);
         return 0;
      }

      for (ii = 0; ; ii++)
      {
         file = gallery(0,"find",ii);
         if (file) {
            fprintf(fid,"%s\n",file);                                      //  write member files
            zfree(file);
         }

         if (ii == Nth) {
            for (jj = 0; jj < Nselect; jj++)                               //  insert new files after Nth file
               fprintf(fid,"%s\n",sel_files[jj]);
         }
         
         if (! file) break;                                                //  EOF
      }
      
      fclose(fid);
      rename(tempfile,coll_file);
      
      navi::gallerytype = 4;
      gallery(coll_file,"initF",-1);                                       //  refresh gallery
      gallery(0,"paint",-1);
      return 0;
   }

   if (strEqu(event,"delete"))
   {
      edit_coll_mousecapture = 0;
      
      if (! coll_file || strNeq(coll_file,navi::galleryname)) {            //  insure gallery = collection
         zmessageACK(mWin,ZTX("Choose a collection first"));
         return 0;
      }

      strcpy(tempfile,collections_dirk);                                   //  temp file for copying
      strcat(tempfile,"tempfile");

      fid = fopen(tempfile,"w");                                           //  open/write temp file
      if (! fid) {
         zmessageACK(mWin,"%s",strerror(errno));
         zdialog_destroy(zd2);
         return 0;
      }

      for (ii = 0; ; ii++)                                                 //  loop member files
      {
         file = gallery(0,"find",ii);
         if (! file) break;
         
         for (jj = 0; jj < Nselect; jj++)                                  //  match with selected files
            if (strEqu(file,sel_files[jj])) break;
         if (jj == Nselect)                                                //  no match, keep member file
            fprintf(fid,"%s\n",file);
                                                                           //  else omit 
         zfree(file);
      }
      
      fclose(fid);
      rename(tempfile,coll_file);

      navi::gallerytype = 4;
      gallery(coll_file,"initF",-1);                                       //  refresh gallery
      gallery(0,"paint",-1);
      return 0;
   }
   
   return 0;
}


//  mouse function for inserting images into collection
//  called from gallery mouse click if edit_coll_mousecapture is true

int edit_coll_mousefunc(int Nth)
{
   using namespace edit_coll;

   Nth_clicked = Nth;                                                      //  mark clicked thumbnail
   zdialog_send_event(zd2,"clicked");                                      //  send event back to dialog
   edit_coll_mousecapture = 0;
   return 0;
}


/**************************************************************************/

//  Change the top directory for an image collection.                      //  v.12.02  one collection, not all
//  This facilitates changing the Fotoxx top image directory 
//  without having to re-create the collection.

void m_move_coll(GtkWidget *, cchar *)
{
   zdialog     *zd;
   char        *pp, collname[100], oldtop[200], newtop[200];
   char        buffr[maxfcc], buffw[maxfcc];
   char        collfile[200], tempfile[200];
   int         ii, zstat, err;
   FILE        *fidr = 0, *fidw = 0;

   zfuncs::F1_help_topic = "move_collection";

   if (mod_keep()) return;                                                 //  unsaved edits

   zd = zdialog_new(ZTX("Move collection"),mWin,Bapply,Bcancel,null);

   zdialog_add_widget(zd,"hbox","hbcoll","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labcoll","hbcoll",ZTX("collection name"),"space=5");
   zdialog_add_widget(zd,"entry","collname","hbcoll",0,"scc=30");
   zdialog_add_widget(zd,"hbox","hbold","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labold","hbold",ZTX("old top directory"),"space=5");
   zdialog_add_widget(zd,"combo","oldtop","hbold",0,"scc=30");
   zdialog_add_widget(zd,"hbox","hbnew","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labnew","hbnew",ZTX("new top directory"),"space=5");
   zdialog_add_widget(zd,"combo","newtop","hbnew",0,"scc=30");
   
   for (ii = 0; ii < Ntopdirk; ii++) {
      zdialog_cb_app(zd,"oldtop",topdirk[ii]);                             //  add top directories
      zdialog_cb_app(zd,"newtop",topdirk[ii]);                             //    to combo box list    v.12.02
   }
   
   if (Ntopdirk) {
      zdialog_stuff(zd,"oldtop",topdirk[0]);                               //  select first one
      zdialog_stuff(zd,"newtop",topdirk[0]);
   }
   
   zdialog_help(zd,"move_collection");
   zdialog_run(zd);

   zstat = zdialog_wait(zd);
   if (zstat != 1) {
      zdialog_free(zd);
      return;
   }
   
   zdialog_fetch(zd,"collname",collname,100);                              //  collection name
   zdialog_fetch(zd,"oldtop",oldtop,200);                                  //  old and new top directory paths
   zdialog_fetch(zd,"newtop",newtop,200);
   zdialog_free(zd);

   snprintf(collfile,200,"%s/%s",collections_dirk,collname);               //  collection file     v.12.02
   snprintf(tempfile,200,"%s/move-collection",get_zuserdir());             //  temp file

   fidr = fopen(collfile,"r");                                             //  open collection and temp files
   if (! fidr) goto error;
   fidw = fopen(tempfile,"w");
   if (! fidw) goto error;
   
   while ((pp = fgets(buffr,maxfcc,fidr)))                                 //  loop collection recs
   {
      repl_1str(buffr,buffw,oldtop,newtop);                                //  replace top directory path
      err = fputs(buffw,fidw);
      if (err < 0) goto error;
   }
   
   err = fclose(fidr);
   err = fclose(fidw);
   if (err) goto error;
   
   err = rename(tempfile,collfile);                                        //  replace collection with temp file
   if (err) goto error;
   else zmessageACK(mWin,ZTX("completed"));
   return;

error:
   err = errno;
   printf("*** %s \n %s \n",collfile,strerror(err));
   zmessageACK(mWin,"%s \n %s",collfile,strerror(err));
   fclose(fidr);
   fclose(fidw);
   return;
}


/**************************************************************************/

//  enter or leave slideshow mode

double      ss_interval = 3;                                               //  slide show interval
double      ss_timer = 0;                                                  //  slide show timer
int         ss_latest = 0;                                                 //  show only latest version files
int         ss_loop = 0;                                                   //  loop from last back to first
char        *ss_oldfile, *ss_newfile;                                      //  image files for transition
GdkPixbuf   *ss_pxbold, *ss_pxbnew;                                        //  pixbuf images: old, new
int         ss_rsold, ss_rsnew;                                            //  old and new pixbuf row stride
int         ss_ww, ss_hh;                                                  //  full screen window size
int         ss_busy = 0;                                                   //  transition underway
int         ss_escape = 0;                                                 //  user pressed Escape key
int         ss_paused = 0;                                                 //  user pressed space bar (pause/resume)
int         ss_nselect = 0;                                                //  selected files count
int         ss_next = 0;                                                   //  next file to show
char        **ss_selfiles = 0;                                             //  selected files list
char        *ss_musicfile = 0;                                             //  opt. music file or playlist

cairo_t     *ss_cr;

void  slideshow_initfile();                                                //  set up slide show collection file
char * ss_basename(char *file);                                            //  isolate base file name
GdkPixbuf * ss_loadpxb(char *file);                                        //  load image as pixbuf

void  ss_instant();                                                        //  transition functions
void  ss_fadein();
void  ss_rollright();
void  ss_rolldown();
void  ss_shiftleft();
void  ss_venetian();
void  ss_grate();
void  ss_rectangle();
void  ss_ellipse();
void  ss_radar();
void  ss_jaws();
void  ss_raindrops();

#define     SSNF 12                                                        //  no. slide show transition types
int         ss_funcs[SSNF];                                                //  user preferred transitions

struct ss_table_t {
   cchar    *name;
   cchar    *vbox;
   void     (*func)();
};

ss_table_t  ss_table[SSNF] =                                               //  image transition types
{
      {  "arrow keys",  "vb41",   null           },                        //  name, zdialog parent, function
      {  "instant",     "vb41",   ss_instant     },
      {  "fade-in",     "vb41",   ss_fadein      },
      {  "roll-right",  "vb41",   ss_rollright   },
      {  "roll-down",   "vb42",   ss_rolldown    },
      {  "venetian",    "vb42",   ss_venetian    },
      {  "grate",       "vb42",   ss_grate       },
///   {  "shift-left",  "vb42",   ss_shiftleft   },                        //  flickers too much
      {  "rectangle",   "vb42",   ss_rectangle   },
      {  "radar",       "vb43",   ss_radar       },
      {  "jaws",        "vb43",   ss_jaws        },
      {  "ellipse",     "vb43",   ss_ellipse     },
      {  "raindrops",   "vb43",   ss_raindrops   }
};


void m_slideshow(GtkWidget *, cchar *)                                     //  overhaul v.12.12
{
   int   slideshow_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;
   int         zstat, err, ii, jj;
   char        countmess[50];
   char        *name1, *name2;
   cchar       *esc_message = ZTX("Press ESC to exit slide show");
   cchar       *latest_message = ZTX("latest versions only");
   cchar       *loop_message = ZTX("continuous loop");
   
   zfuncs::F1_help_topic = "slide_show";
   
   ss_table[0].name = ZTX("arrow keys");                                   //  use translated names
   ss_table[1].name = ZTX("instant");
   ss_table[2].name = ZTX("fade-in");
   ss_table[3].name = ZTX("roll-right");
   ss_table[4].name = ZTX("roll-down");
   ss_table[5].name = ZTX("venetian");
   ss_table[6].name = ZTX("grate");
///ss_table[7].name = ZTX("shift-left");                                   //  too jerky
   ss_table[7].name = ZTX("rectangle");
   ss_table[8].name = ZTX("radar");
   ss_table[9].name = ZTX("jaws");
   ss_table[10].name = ZTX("ellipse");
   ss_table[11].name = ZTX("raindrops");

   if (! Fslideshow)                                                       //  start slide show
   {
      if (checkmenulock()) return;                                         //  menu locked               v.12.09
      if (mod_keep()) return;                                              //  unsaved edits
      
      zd = zdialog_new(ZTX("Slide Show"),mWin,Bproceed,Bcancel,null);      //  user dialog
      zdialog_add_widget(zd,"hbox","hbf","dialog",0,"space=5");
      zdialog_add_widget(zd,"button","selfiles","hbf",Bselectfiles,"space=3");
      zdialog_add_widget(zd,"label","selcount","hbf",ZTX("0 files selected"),"space=6");
      zdialog_add_widget(zd,"label","space","hbf",0,"space=8");
      zdialog_add_widget(zd,"label","labcoll","hbf",ZTX("collection:"));
      zdialog_add_widget(zd,"button","open","hbf",Bopen,"space=5");
      zdialog_add_widget(zd,"button","save","hbf",Bsave,"space=1");
      zdialog_add_widget(zd,"hbox","hbesc","dialog",0,"space=3");
      zdialog_add_widget(zd,"label","labesc","hbesc",esc_message,"space=3");
      zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
      zdialog_add_widget(zd,"label","labsecs","hb1",ZTX("seconds"),"space=3");
      zdialog_add_widget(zd,"entry","secs","hb1","3","scc=3");
      zdialog_add_widget(zd,"check","latest","hb1",latest_message,"space=8");
      zdialog_add_widget(zd,"check","loop","hb1",loop_message,"space=8");
      zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=3");
      zdialog_add_widget(zd,"label","lab21","hb2",ZTX("music file"),"space=3");
      zdialog_add_widget(zd,"entry","musicfile","hb2",0,"expand|space=5");
      zdialog_add_widget(zd,"button","browse","hb2",Bbrowse,"space=5");
      zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=3");
      zdialog_add_widget(zd,"label","lab3","hb3",ZTX("transitions"),"space=3");
      zdialog_add_widget(zd,"hbox","hb4","dialog",0,"space=8");
      zdialog_add_widget(zd,"vbox","vb41","hb4",0,"space=4");
      zdialog_add_widget(zd,"vbox","vb42","hb4",0,"space=4");
      zdialog_add_widget(zd,"vbox","vb43","hb4",0,"space=4");
      
      for (ii = 0; ii < SSNF; ii++)                                        //  add transition check boxes
         zdialog_add_widget(zd,"check",ss_table[ii].name,
                            ss_table[ii].vbox,ss_table[ii].name);

      zdialog_restore_inputs(zd);                                          //  preload prior user inputs    v.12.09
      
      zdialog_stuff(zd,ss_table[0].name,1);                                //  arrow keys always allowed
      
      snprintf(countmess,50,ZTX("%d files selected"),ss_nselect);          //  no. files selected previously
      zdialog_stuff(zd,"selcount",countmess);

      if (SS_musicfile)                                                    //  command line -music parameter
         zdialog_stuff(zd,"musicfile",SS_musicfile);                       //  stuff dialog if present
      SS_musicfile = 0;
      
      zdialog_help(zd,"slide_show");                                       //  zdialog help topic

   ss_dialog:

      zdialog_run(zd,slideshow_dialog_event);                              //  run dialog
      zstat = zdialog_wait(zd);                                            //  wait for completion

      if (zstat != 1) {                                                    //  cancel
         zdialog_free(zd);
         return;
      }

      for (ii = jj = 0; jj < ss_nselect; jj++)                             //  get selected files
      {
         if (image_file_type(ss_selfiles[jj]) == 2) {                      //  keep image file
            if (jj > ii) {
               ss_selfiles[ii] = ss_selfiles[jj];                          //  pack-down list
               ss_selfiles[jj] = 0;
            }
            ii++;
         }
         else {                                                            //  eliminate non-image file
            zfree(ss_selfiles[jj]);                                        //  (directory, RAW file)
            ss_selfiles[jj] = 0;
         }
      }

      ss_nselect = ii;                                                     //  file count

      if (! ss_nselect) {
         zmessageACK(mWin,ZTX("no image files selected"));                 //  zero files, try again
         zd->zstat = 0;
         goto ss_dialog;
      }
      
      zdialog_fetch(zd,"secs",ss_interval);                                //  get interval, seconds
      zdialog_fetch(zd,"latest",ss_latest);                                //  get "latest files only" option
      zdialog_fetch(zd,"loop",ss_loop);                                    //  get loop option

      for (ii = 0; ii < SSNF; ii++)                                        //  get selected transition types
         zdialog_fetch(zd,ss_table[ii].name,ss_funcs[ii]);
      
      ss_funcs[0] = 1;                                                     //  arrow keys always allowed
      
      for (ii = 1; ii < SSNF; ii++)                                        //  if only arrow keys, set huge interval
         if (ss_funcs[ii]) break;
      if (ii == SSNF) ss_interval = 9999;

      char *pp = zmalloc(500,"musicfile");
      zdialog_fetch(zd,"musicfile",pp,499);                                //  get music file
      ss_musicfile = pp;

      zdialog_free(zd);                                                    //  kill dialog
      
      if (ss_latest)
      {                                                                    //  "latest files only" chosen
         for (ii = 0, jj = 1; jj < ss_nselect; jj++)                       //  (works if default sort order)
         {
            name1 = ss_basename(ss_selfiles[ii]);
            name2 = ss_basename(ss_selfiles[jj]);

            if (strEqu(name1,name2)) {                                     //  same file name
               zfree(ss_selfiles[ii]);                                     //  replace with higher version
               ss_selfiles[ii] = ss_selfiles[jj];                          //  pack-down list
               ss_selfiles[jj] = 0;
            }
            else {
               ii++;                                                       //  new file name, keep prior
               if (jj > ii) {
                  ss_selfiles[ii] = ss_selfiles[jj];                       //  pack-down list
                  ss_selfiles[jj] = 0;
               }
            }
            
            zfree(name1);
            zfree(name2);
         }

         ss_nselect = ii + 1;                                              //  new file count
         slideshow_initfile();                                             //  revise slide show collection file
      }
      
      if (ss_musicfile && *ss_musicfile && strNeq(ss_musicfile,"undefined")) 
      {                                                                    //  if music file present, start it up
         sprintf(command,"xdg-open \"%s\" ",ss_musicfile);
         printf("command: %s \n",command);
         err = system(command);
         if (err) printf("*** %s \n",wstrerror(err));
      }

      f_open(ss_selfiles[0],1);                                            //  open first file
      ss_paused = 0;                                                       //  not paused
      win_fullscreen();                                                    //  full screen, no menu/toolbar/status bar
      gdk_window_set_background_rgba(gdkWin,&GDKblack);                    //  black background

      ss_ww = gdk_window_get_width(gdkWin);                                //  drawing window size
      ss_hh = gdk_window_get_height(gdkWin);

      ss_newfile = ss_selfiles[0];                                         //  first image file

      ss_pxbnew = ss_loadpxb(ss_newfile);                                  //  load pixbuf graphic image
      if (! ss_pxbnew) {
         win_unfullscreen();
         return;
      }

      ss_rsnew = gdk_pixbuf_get_rowstride(ss_pxbnew);                      //  row stride

      ss_oldfile = 0;                                                      //  no old (prior) image
      ss_pxbold = 0;
      ss_rsold = 0;
      ss_next = 0;                                                         //  next to show = 1st in list

      ss_timer = get_seconds() + ss_interval;                              //  set timer for first call to
      Fslideshow = 1;                                                      //    ss_slideshow_next()
      Fblowup = 1;
      Fzoom = 0;
      mwpaint2();
      return;
   }

   else                                                                    //  leave slide show mode
   {
      if (ss_busy) {
         ss_escape = 1;                                                    //  wait for transition done
         return;
      }

      if (ss_pxbold) g_object_unref(ss_pxbold);
      if (ss_pxbnew) g_object_unref(ss_pxbnew);
      ss_pxbold = ss_pxbnew = 0;

      win_unfullscreen();                                                  //  restore old window size, menu etc.

      ss_escape = 0;
      Fslideshow = 0;
      Fblowup = 0;
      Fzoom = 0;                                                           //  fit image to window
      mwpaint2();
      m_slideshow(0,0);                                                    //  go back to dialog               v.12.12
      return;
   }
}


//  dialog event function - file chooser for images to show and music file

int slideshow_dialog_event(zdialog *zd, cchar *event) 
{
   FILE        *fid;   
   char        *file;
   int         ii, err;
   char        countmess[50];
   char        buff[maxfcc];
   struct stat statb;
   
   if (strEqu(event,"selfiles"))                                           //  select images to convert        v.12.11
   {
      if (ss_nselect) {                                                    //  free prior list
         for (ii = 0; ii < ss_nselect; ii++) 
            zfree(ss_selfiles[ii]);
         zfree(ss_selfiles);
         ss_selfiles = 0;
      }

      zdialog_show(zd,0);                                                  //  hide parent dialog              v.12.11
      ss_selfiles = gallery_getfiles();                                    //  get list of files to convert
      zdialog_show(zd,1);

      if (ss_selfiles)                                                     //  count files selected
         for (ii = 0; ss_selfiles[ii]; ii++);
      else ii = 0;
      ss_nselect = ii;

      snprintf(countmess,50,ZTX("%d files selected"),ss_nselect);          //  update dialog file count
      zdialog_stuff(zd,"selcount",countmess);
      
      slideshow_initfile();                                                //  make slide show collection file
   }
   
   if (strEqu(event,"open"))                                               //  open an existing collection     v.12.12
   {
      file = zgetfile1(ZTX("open collection"),"open",collections_dirk);
      if (! file) return 0;
      err = stat(file,&statb);
      if (err) return 0;

      if (ss_nselect) {                                                    //  free prior list
         for (ii = 0; ii < ss_nselect; ii++) 
            zfree(ss_selfiles[ii]);
         zfree(ss_selfiles);
         ss_selfiles = 0;
      }

      fid = fopen(file,"r");                                               //  read collection, count image files
      if (! fid) return 0;
      for (ii = 0; ; ii++) 
         if (! fgets_trim(buff,maxfcc,fid)) break;
      fclose(fid);

      if (! ii) {
         zmessageACK(mWin,ZTX("no image files selected"));
         return 0;
      }

      ss_selfiles = (char **) zmalloc(ii * sizeof(char *));
      ss_nselect = ii;

      fid = fopen(file,"r");                                               //  collection files >> selected files
      for (ii = 0; ii < ss_nselect; ii++) 
         ss_selfiles[ii] = strdupz(fgets_trim(buff,maxfcc,fid));
      fclose(fid);
      zfree(file);

      snprintf(countmess,50,ZTX("%d files selected"),ss_nselect);          //  update dialog file count
      zdialog_stuff(zd,"selcount",countmess);

      slideshow_initfile();                                                //  make slide show collection file
   }

   if (strEqu(event,"save"))                                               //  save as a collection         v.12.12
   {
      if (! ss_nselect) {
         zmessageACK(mWin,ZTX("no image files selected"));
         return 0;
      }

      file = zgetfile1(ZTX("save as collection"),"save",collections_dirk);
      if (! file) return 0;

      fid = fopen(file,"w");
      if (! fid) return 0;
      for (ii = 0; ii < ss_nselect; ii++) 
         fprintf(fid,"%s\n",ss_selfiles[ii]);
      fclose(fid);
      return 0;
   }

   if (strEqu(event,"browse"))                                             //  look for music file
   {
      file = zgetfile1(ZTX("Select music file or playlist"),"open",ss_musicfile);
      if (! file) return 0;
      zdialog_stuff(zd,"musicfile",file);
      zfree(file);
   }

   return 0;
}


//  write set of selected image files to slide show collection file "slide_show"
//  initialize gallery from the same set of files and open the first one

void slideshow_initfile()
{
   char     collfile[200];
   char     *file;
   FILE     *fid;

   if (! ss_nselect) return;

   strcpy(collfile,collections_dirk);
   strcat(collfile,"/slide_show");

   fid = fopen(collfile,"w");                                              //  write file list to collection file
   if (! fid) return;
   for (int ii = 0; ii < ss_nselect; ii++)
      fprintf(fid,"%s\n",ss_selfiles[ii]);
   fclose(fid);

   navi::gallerytype = 4;                                                  //  type = collection
   gallery(collfile,"initF");                                              //  generate gallery of selected files
   file = gallery(0,"find");
   f_open(file,1);                                                         //  open first file
   zfree(file);
   return;
}


//  Show next slide if time is up or user navigates with arrow keys.
//  Called by timer function and by keyboard function if Fslideshow is set.

void slideshow_next(cchar *mode)
{
   double         secs;
   int            Fkey = 0, ii;
   static int     last_transition = 0;
   
   if (cTab != tabF) return;                                               //  left my tab, wait 
   
   if (strEqu(mode,"pause"))                                               //  toggle slide show paused status
      ss_paused = 1 - ss_paused;                                           //  (spacebar)

   if (ss_busy) return;                                                    //  come back later

   if (ss_escape) {                                                        //  user pressed escape key
      m_slideshow(0,0);                                                    //  exit slide show
      return;
   }
   
   if (strEqu(mode,"timer")) {                                             //  timer trigger
      if (ss_paused) return;
      secs = get_seconds();
      if (secs < ss_timer) return;
      mode = "next";                                                       //  time for next image
      Fkey = 0;
   }
   else Fkey = 1;                                                          //  keyboard trigger
   
   if (strEqu(mode,"prev"))                                                //  get previous image in list
   {
      if (ss_next == 0 && ! ss_loop) return;
      if (ss_next == 0) ss_next = ss_nselect-1;
      else ss_next--;
   }

   if (strEqu(mode,"next"))                                                //  get next image in list
   {
      if (ss_next == ss_nselect-1 && ! ss_loop) return;
      if (ss_next == ss_nselect-1) ss_next = 0;
      else ss_next++;
   }

   ss_busy++;
   
   ss_oldfile = ss_newfile;                                                //  old file = new
   ss_newfile = ss_selfiles[ss_next];                                      //  new new file

   if (ss_pxbold) g_object_unref(ss_pxbold);
   ss_pxbold = ss_pxbnew;                                                  //  old pixbuf = new
   ss_rsold = ss_rsnew;

   ss_pxbnew = ss_loadpxb(ss_newfile);                                     //  new new pixbuf from new new file
   if (! ss_pxbnew) return;
   ss_rsnew = gdk_pixbuf_get_rowstride(ss_pxbnew);                         //  row stride   

   mutex_lock(&Fpixmap_lock);                                              //  block other window updates

   gdk_window_set_background_rgba(gdkWin,&GDKblack);                       //  black background                v.12.11
   ss_cr = gdk_cairo_create(gdkWin);                                       //  new cairo context

   if (Fkey) ss_instant();                                                 //  KB input, do instant transition
   else                                                                    //  use next selected transition type
   {
      for (ii = last_transition + 1; ii < SSNF; ii++)                      //  find next selected transition type
         if (ss_funcs[ii]) break;
      if (ii >= SSNF) {                                                    //  not found, search from beginning
         for (ii = 1; ii < SSNF; ii++)                                     //  (omit [0] = arrow keys)
            if (ss_funcs[ii]) break;
      }
      if (ii == SSNF) {                                                    //  no transition types are selected
         last_transition = 0;                                              //  use instant transition
         ss_instant();
      }
      else {
         last_transition = ii;                                             //  use next selected transition type
         ss_table[ii].func();
      }
   }

   cairo_destroy(ss_cr);
   mutex_unlock(&Fpixmap_lock);

   f_open(ss_newfile,0,0,0,0);                                             //  set all image data

   if (zdeditcctext) m_edit_cctext(0,0);                                   //  update caption/comments dialog

   secs = get_seconds();                                                   //  set time for next image
   ss_timer = secs + ss_interval;
   ss_busy = 0;
   return;
}


//  extract base name from image file name
//  base name is file name without directory, version, extension

char * ss_basename(char *file) 
{
   char     *pp1, *pp2, *pp3;

   pp1 = strrchr(file,'/');                                                //  skip over directory if present
   if (pp1) file = pp1 + 1;
   pp1 = strdupz(file,0,"slideshow");
   pp2 = strrchr(pp1,'.');                                                 //  look for .ext
   if (! pp2) return pp1;
   *pp2 = 0;                                                               //  strip it off
   pp3 = pp2 - 4;
   if (! strnEqu(pp3,".v",2)) return pp1;                                  //  look for version .v00 to .v99
   if (pp3[2] < '0' || pp3[2] > '9') return pp1;
   if (pp3[3] < '0' || pp3[3] > '9') return pp1;
   *pp3 = 0;                                                               //  strip it off
   return pp1;
}


//  load image and rescale to fit in window size

GdkPixbuf * ss_loadpxb(char *file)
{
   int         cc, fww, fhh, px, py, orgx, orgy;
   float       wscale, hscale, scale;
   GdkPixbuf   *pxbin, *pxbtemp, *pxbout;
   uint8       *pixels1, *pixels2, *pix1, *pix2;
   int         rs1, rs2;
   GError      *error = 0;

   pxbin = gdk_pixbuf_new_from_file(file,&error);                          //  load image file into pixbuf
   if (! pxbin) {
      zmessageACK(mWin,"%s",error->message);
      return 0;
   }

   fww = gdk_pixbuf_get_width(pxbin);
   fhh = gdk_pixbuf_get_height(pxbin);
   
   wscale = 1.0 * ss_ww / fww;                                             //  find scale to fit in window
   hscale = 1.0 * ss_hh / fhh;
   if (wscale < hscale) scale = wscale;                                    //  use greatest ww/hh ratio
   else  scale = hscale;
   fww = fww * scale;
   fhh = fhh * scale;
   
   pxbtemp = gdk_pixbuf_scale_simple(pxbin,fww,fhh,BILINEAR);              //  rescale image to fit window
   pixels1 = gdk_pixbuf_get_pixels(pxbtemp);
   rs1 = gdk_pixbuf_get_rowstride(pxbtemp);
   g_object_unref(pxbin);

   pxbout = gdk_pixbuf_new(GDKRGB,0,8,ss_ww,ss_hh);                        //  create output pixbuf = window size
   rs2 = gdk_pixbuf_get_rowstride(pxbout);
   pixels2 = gdk_pixbuf_get_pixels(pxbout);
   cc = rs2 * ss_hh;
   memset(pixels2,0,cc);                                                   //  clear output pixbuf black

   orgx = 0.5 * (ss_ww - fww);                                             //  image origin in window
   orgy = 0.5 * (ss_hh - fhh);                                             //    for equal margins

   for (py = 0; py < fhh; py++)                                            //  copy image to output pixmap
   for (px = 0; px < fww; px++)                                            //    and center the image
   {
      pix1 = pixels1 + py * rs1 + px * 3;                                  //  v.12.10
      pix2 = pixels2 + (py + orgy) * rs2 + (px + orgx) * 3;
      pix2[0] = pix1[0];
      pix2[1] = pix1[1];
      pix2[2] = pix1[2];
   }

   g_object_unref(pxbtemp);
   return pxbout;
}


//  instant transition (also used for keyboard arrow keys)

void ss_instant()
{
   return;
}


//  fade-out / fade-in transition

void ss_fadein()
{
   GdkPixbuf   *pxbmix;
   int         ii, jj, kk, px, py, rs;
   float       newpart, oldpart;
   uint8       *pixels1, *pixels2, *pixels3;
   uint8       *pix1, *pix2, *pix3;
   
   pxbmix = gdk_pixbuf_copy(ss_pxbold);
   rs = gdk_pixbuf_get_rowstride(pxbmix);

   pixels1 = gdk_pixbuf_get_pixels(ss_pxbold);
   pixels2 = gdk_pixbuf_get_pixels(ss_pxbnew);
   pixels3 = gdk_pixbuf_get_pixels(pxbmix);
   
   for (ii = 0; ii <= 100; ii += 10)
   {
      newpart = 0.01 * ii;
      oldpart = 1.0 - newpart;

      for (jj = 0; jj < 2; jj++)                                           //  four passes, each modifies 25%
      for (kk = 0; kk < 2; kk++)                                           //    of the pixels (visually smoother)
      {
         for (py = jj; py < ss_hh; py += 2)
         for (px = kk; px < ss_ww; px += 2)
         {
            pix1 = pixels1 + py * ss_rsold + px * 3;                       //  v.12.10
            pix2 = pixels2 + py * ss_rsnew + px * 3;
            pix3 = pixels3 + py * rs + px * 3;
            pix3[0] = newpart * pix2[0] + oldpart * pix1[0];
            pix3[1] = newpart * pix2[1] + oldpart * pix1[1];
            pix3[2] = newpart * pix2[2] + oldpart * pix1[2];
         }

         gdk_cairo_set_source_pixbuf(ss_cr,pxbmix,0,0);
         cairo_paint(ss_cr);
      }
   }

   g_object_unref(pxbmix);
   return;
}


//  new image rolls over prior image from left to right

void ss_rollright()
{
   GdkPixbuf   *pixbuf;
   int         px;
   float       delay = 0.5 / ss_ww;
   uint8       *pixels, *pix3;
   
   pixels = gdk_pixbuf_get_pixels(ss_pxbnew);

   for (px = 0; px < ss_ww; px++)
   {
      pix3 = pixels + px * 3;
      pixbuf = gdk_pixbuf_new_from_data(pix3,GDKRGB,0,8,1,ss_hh,ss_rsnew,0,0);
      gdk_cairo_set_source_pixbuf(ss_cr,pixbuf,px,0);
      cairo_paint(ss_cr);
      g_object_unref(pixbuf);
      zsleep(delay);
   }

   return;
}


//  new image rolls over prior image from top down

void ss_rolldown()
{
   GdkPixbuf   *pixbuf;
   int         py;
   float       delay = 0.5 / ss_hh;
   uint8       *pixels, *pix3;
   
   pixels = gdk_pixbuf_get_pixels(ss_pxbnew);

   for (py = 0; py < ss_hh; py++)
   {
      pix3 = pixels + py * ss_rsnew;
      pixbuf = gdk_pixbuf_new_from_data(pix3,GDKRGB,0,8,ss_ww,1,ss_rsnew,0,0);
      gdk_cairo_set_source_pixbuf(ss_cr,pixbuf,0,py);
      cairo_paint(ss_cr);
      g_object_unref(pixbuf);
      zsleep(delay);
   }

   return;
}


//  shift prior image to the left as next image shifts-in from the right
//  (this one flickers annoingly and cannot be fixed)

void ss_shiftleft()
{
   GdkPixbuf   *pixbuf;
   int         px, Nsteps = 40;
   uint8       *pixels1, *pixels2, *pix3;
   
   pixels1 = gdk_pixbuf_get_pixels(ss_pxbold);
   pixels2 = gdk_pixbuf_get_pixels(ss_pxbnew);
   
   for (px = 20; px < ss_ww; px += ss_ww / Nsteps)
   {
      pix3 = pixels1 + px * 3;
      pixbuf = gdk_pixbuf_new_from_data(pix3,GDKRGB,0,8,ss_ww-px,ss_hh,ss_rsold,0,0);
      gdk_cairo_set_source_pixbuf(ss_cr,pixbuf,0,0);
      cairo_paint(ss_cr);
      g_object_unref(pixbuf);

      pix3 = pixels2;
      pixbuf = gdk_pixbuf_new_from_data(pix3,GDKRGB,0,8,px,ss_hh,ss_rsnew,0,0);
      gdk_cairo_set_source_pixbuf(ss_cr,pixbuf,ss_ww-px,0);
      cairo_paint(ss_cr);
      g_object_unref(pixbuf);

      zsleep(0.01);
   }

   return;
}


//  new image opens up in horizontal rows like venetian blinds

void ss_venetian()
{
   GdkPixbuf   *pixbuf;
   int         py1, py2;
   uint8       *pixels, *pix3;
   int         louver, Nlouvers = 20;
   int         louversize = ss_hh / Nlouvers;
   float       delay = 1.0 / louversize;
   
   pixels = gdk_pixbuf_get_pixels(ss_pxbnew);
   
   for (py1 = 0; py1 < louversize; py1++)                                  //  y-row within each louver
   {
      for (louver = 0; louver < Nlouvers; louver++)                        //  louver, first to last
      {
         py2 = py1 + louver * louversize;
         if (py2 >= ss_hh) break;
         pix3 = pixels + py2 * ss_rsnew;
         pixbuf = gdk_pixbuf_new_from_data(pix3,GDKRGB,0,8,ss_ww,1,ss_rsnew,0,0);
         gdk_cairo_set_source_pixbuf(ss_cr,pixbuf,0,py2);
         cairo_paint(ss_cr);
         g_object_unref(pixbuf);
      }

      zsleep(delay);
   }

   return;
}


//  a grate opens up to show new image

void ss_grate()
{
   GdkPixbuf   *pixbuf;
   int         px1, px2, py1, py2;
   uint8       *pixels, *pix3;
   int         row, col, Nrow, Ncol;                                       //  rows and columns
   int         boxww, boxhh;
   float       delay;
   
   pixels = gdk_pixbuf_get_pixels(ss_pxbnew);

   Ncol = 20;                                                              //  20 columns
   boxww = boxhh = ss_ww / Ncol;                                           //  square boxes
   Nrow = ss_hh / boxhh;                                                   //  corresp. rows
   Ncol++;                                                                 //  round up
   Nrow++;
   delay = 1.0 / boxhh;

   for (py1 = 0; py1 < boxhh; py1++)
   {
      for (row = 0; row < Nrow; row++)
      {
         py2 = py1 + row * boxhh;
         if (py2 >= ss_hh) break;
         pix3 = pixels + py2 * ss_rsnew;
         pixbuf = gdk_pixbuf_new_from_data(pix3,GDKRGB,0,8,ss_ww,1,ss_rsnew,0,0);
         gdk_cairo_set_source_pixbuf(ss_cr,pixbuf,0,py2);
         cairo_paint(ss_cr);
         g_object_unref(pixbuf);
      }
      
      px1 = py1;

      for (col = 0; col < Ncol; col++)
      {
         px2 = px1 + col * boxww;
         if (px2 >= ss_ww) break;
         pix3 = pixels + px2 * 3;
         pixbuf = gdk_pixbuf_new_from_data(pix3,GDKRGB,0,8,1,ss_hh,ss_rsnew,0,0);
         gdk_cairo_set_source_pixbuf(ss_cr,pixbuf,px2,0);
         cairo_paint(ss_cr);
         g_object_unref(pixbuf);
      }
      
      zsleep(delay);
   }

   return;
}


//  A hole opens up from the center and expands outward

void ss_rectangle()
{
   GdkPixbuf   *pixbuf;
   int         px1, py1, px2, py2, px3, py3;
   int         ww1, hh1, ww2, hh2;
   uint8       *pixels, *pix3;
   int         step, Nsteps = 100;
   float       delay = 1.0 / Nsteps;

   pixels = gdk_pixbuf_get_pixels(ss_pxbnew);

   for (step = 1; step < Nsteps; step++)
   {
      ww1 = ss_ww * step / Nsteps;
      hh1 = ww1 * ss_hh / ss_ww;
      ww2 = ss_ww / Nsteps / 2;
      hh2 = ss_hh / Nsteps / 2;

      px1 = (ss_ww - ww1) / 2;
      py1 = (ss_hh - hh1) / 2;
      px2 = px1 + ww1 - ww2;
      py2 = py1;
      px3 = px1;
      py3 = py1 + hh1 - hh2;

      pix3 = pixels + py1 * ss_rsnew + px1 * 3;
      pixbuf = gdk_pixbuf_new_from_data(pix3,GDKRGB,0,8,ww1+1,hh2+1,ss_rsnew,0,0);
      gdk_cairo_set_source_pixbuf(ss_cr,pixbuf,px1,py1);
      cairo_paint(ss_cr);
      g_object_unref(pixbuf);

      pix3 = pixels + py2 * ss_rsnew + px2 * 3;
      pixbuf = gdk_pixbuf_new_from_data(pix3,GDKRGB,0,8,ww2+1,hh1+1,ss_rsnew,0,0);
      gdk_cairo_set_source_pixbuf(ss_cr,pixbuf,px2,py2);
      cairo_paint(ss_cr);
      g_object_unref(pixbuf);

      pix3 = pixels + py3 * ss_rsnew + px3 * 3;
      pixbuf = gdk_pixbuf_new_from_data(pix3,GDKRGB,0,8,ww1+1,hh2+1,ss_rsnew,0,0);
      gdk_cairo_set_source_pixbuf(ss_cr,pixbuf,px3,py3);
      cairo_paint(ss_cr);
      g_object_unref(pixbuf);

      pix3 = pixels + py1 * ss_rsnew + px1 * 3;
      pixbuf = gdk_pixbuf_new_from_data(pix3,GDKRGB,0,8,ww2+1,hh1+1,ss_rsnew,0,0);
      gdk_cairo_set_source_pixbuf(ss_cr,pixbuf,px1,py1);
      cairo_paint(ss_cr);
      g_object_unref(pixbuf);

      zsleep(delay);
   }
   
   return;
}


//  An ellipse opens up from the center and expands outward

void ss_ellipse()
{
   GdkPixbuf   *pixbuf;
   uint8       *pixels, *pix3;
   int         step, Nsteps = 100;
   int         px1, py1, ww;
   float       delay = 0.0 / Nsteps;                                       //  zero for now
   float       a, b, a2, b2, px, py, px2, py2;
   float       ww2 = ss_ww / 2, hh2 = ss_hh / 2;

   pixels = gdk_pixbuf_get_pixels(ss_pxbnew);

   for (step = 1; step < 1.3 * Nsteps; step++)
   {
      a = ww2 * step / Nsteps;                                             //  ellipse a and b constants
      b = a * ss_hh / ss_ww;                                               //    from tiny to >> image size
      a2 = a * a;
      b2 = b * b;
      
      for (py = -b; py <= +b; py += 3)                                     //  py from top of ellipse to bottom
      {
         while (py < -(hh2-2)) py += 3;
         if (py > hh2-2) break;
         py2 = py * py;
         px2 = a2 * (1.0 - py2 / b2);                                      //  corresponding px value,
         px = sqrt(px2);                                                   //  (+/- from center of ellipse)
         if (px > ww2) px = ww2;
         ww = 2 * px;                                                      //  length of line thru ellipse
         if (ww < 2) continue;
         px1 = ww2 - px;                                                   //  relocate origin
         py1 = py + hh2;
         if (px1 + ww > ss_ww) px1 = ss_ww - ww;                           //  insurance     v.12.04
         if (py1 + 3 > ss_hh) py1 = ss_hh - 3;
         pix3 = pixels + py1 * ss_rsnew + px1 * 3;
         pixbuf = gdk_pixbuf_new_from_data(pix3,GDKRGB,0,8,ww,3,ss_rsnew,0,0);
         gdk_cairo_set_source_pixbuf(ss_cr,pixbuf,px1,py1);
         cairo_paint(ss_cr);
         g_object_unref(pixbuf);
      }
      
      zsleep(delay);
   }
   
   return;
}


//  New image sweeps into view like a circular radar image

void ss_radar()
{
   GdkPixbuf   *pixbuf;
   int         px = 0, py = 0, Npx, Npy, Np = 12;                          //  smaller values make compiz jerky
   float       R, Rmax, T, Tmax, dT, delay;
   float       cosT, sinT;
   float       ww2 = ss_ww / 2, hh2 = ss_hh / 2;
   uint8       *pixels, *pix3;
   
   pixels = gdk_pixbuf_get_pixels(ss_pxbnew);

   Rmax = sqrt(ww2 * ww2 + hh2 * hh2);
   Tmax = pi;
   dT = 0.6 * asinf(Np / Rmax);
   delay = dT / Tmax;                                                      //  1 sec. + CPU time
   if (delay < 0.001) delay = 0.001;
   delay = 0.001;                                                          //  for now
   
   for (T = 0; T < Tmax; T += dT)
   {
      cosT = cosf(T);
      sinT = sinf(T);

      for (R = -Rmax; R < Rmax; R += Np)
      {
         px = ww2 + R * cosT;
         py = hh2 - R * sinT;
         if (px < -Np) continue;
         if (py < -Np) continue;
         if (px > ss_ww-1) continue;
         if (py > ss_hh-1) continue;
         Npx = Npy = Np;
         if (px < 0) px = 0;
         if (py < 0) py = 0;
         if (px + Npx > ss_ww-1) Npx = ss_ww-1 - px;
         if (py + Npy > ss_hh-1) Npy = ss_hh-1 - py;
         if (Npx < 1 || Npy < 1) continue;
         pix3 = pixels + py * ss_rsnew + px * 3;
         pixbuf = gdk_pixbuf_new_from_data(pix3,GDKRGB,0,8,Npx,Npy,ss_rsnew,0,0);
         gdk_cairo_set_source_pixbuf(ss_cr,pixbuf,px,py);
         cairo_paint(ss_cr);
         g_object_unref(pixbuf);
      }
      
      zsleep(delay);
   }

   return;
}


//  New image closes in from top and bottom with jagged teeth

void ss_jaws()
{
   GdkPixbuf   *pixbuf;
   int         nteeth = 20, Np = 6;
   int         tbase1, tbase2, twidth, tlength, tooth, tpos;
   int         ii, px, py, ww, ww2;
   uint8       *pixels, *pix3;

   pixels = gdk_pixbuf_get_pixels(ss_pxbnew);

   twidth = ss_ww / nteeth;
   tlength = twidth;

   for (ii = 0; ii < ss_hh/2 - tlength/2; ii += Np)
   {
      tbase1 = ii;                                                         //  tooth base from window top to middle
      tbase2 = ss_hh - tbase1 - 1;                                         //  tooth base from window bottom to middle

      for (tooth = 0; tooth <= nteeth; tooth++)                            //  tooth first to last + 1
      {
         for (tpos = 0; tpos < tlength; tpos += Np)                        //  tooth position from base to point
         {
            ww = twidth * (tlength - tpos) / tlength;                      //  tooth width at scan line
            if (ww < 2) break;

            py = tbase1 + tpos;                                            //  top teeth scan line y
            px = twidth / 2 + tooth * twidth - ww / 2;                     //  scan line x to x + ww
            if (px < ss_ww) {
               ww2 = ww;
               if (px + ww2 > ss_ww) ww2 = ss_ww - px;
               pix3 = pixels + py * ss_rsnew + px * 3;
               pixbuf = gdk_pixbuf_new_from_data(pix3,GDKRGB,0,8,ww2,Np,ss_rsnew,0,0);
               gdk_cairo_set_source_pixbuf(ss_cr,pixbuf,px,py);
               cairo_paint(ss_cr);
               g_object_unref(pixbuf);

            }

            py = tbase2 - tpos;                                            //  bottom teeth scan line y
            py = py - Np;
            px = tooth * twidth - ww / 2;                                  //  scan line x to x + ww
            if (tooth == 0) {
               px = 0;                                                     //  leftmost tooth is half
               ww = ww / 2;
            }
            if (px < ss_ww) {
               ww2 = ww;
               if (px + ww2 > ss_ww) ww2 = ss_ww - px;
               pix3 = pixels + py * ss_rsnew + px * 3;
               pixbuf = gdk_pixbuf_new_from_data(pix3,GDKRGB,0,8,ww2,Np,ss_rsnew,0,0);
               gdk_cairo_set_source_pixbuf(ss_cr,pixbuf,px,py);
               cairo_paint(ss_cr);
               g_object_unref(pixbuf);

            }
         }
      }
   }

   return;
}


//  new image splats on to old image one drop at a time

void ss_raindrops()                                                        //  v.12.11
{
   GdkPixbuf   *pxbmix, *pxbdrop;
   int         rsmix;
   int         px, py, px1, py1, px2, py2, cx, cy;
   int         Rmin, Rmax, R, R2, dist2, Ndrops;
   uint8       *pixels2, *pixels3;
   uint8       *pix2, *pix3 = 0;
   float       dtime;
   
   pixels2 = gdk_pixbuf_get_pixels(ss_pxbnew);                             //  source image

   pxbmix = gdk_pixbuf_new(GDKRGB,0,8,ss_ww,ss_hh);                        //  destination image
   pixels3 = gdk_pixbuf_get_pixels(pxbmix);
   rsmix = gdk_pixbuf_get_rowstride(pxbmix);
   memset(pixels3,0,ss_hh * rsmix);                                        //  clear dest. to black

   gdk_cairo_set_source_pixbuf(ss_cr,pxbmix,0,0);                          //  paint black window
   cairo_paint(ss_cr);
   
   Rmin = ss_ww * 0.01;                                                    //  drop size range
   Rmax = ss_ww * 0.03;
   Ndrops = 1000;
   
   for (int ii = 0; ii < Ndrops; ii++)
   {
      cx = drandz() * ss_ww;                                               //  drop location on image
      cy = drandz() * ss_hh;
      R = drandz() * Rmax + Rmin;                                          //  drop size
      R2 = R * R;
      px1 = cx - R;
      if (px1 < 0) px1 = 0;
      py1 = cy - R;
      if (py1 < 0) py1 = 0;
      px2 = cx + R;
      if (px2 >= ss_ww) px2 = ss_ww;
      py2 = cy + R;
      if (py2 > ss_hh) py2 = ss_hh;

      for (py = py1; py < py2; py++)                                       //  copy drop area from new image
      for (px = px1; px < px2; px++)                                       //    to old image
      {
         dist2 = (px-cx) * (px-cx) + (py-cy) * (py-cy);
         if (dist2 > R2) continue;
         pix2 = pixels2 + py * ss_rsnew + px * 3;
         pix3 = pixels3 + py * rsmix + px * 3;
         pix3[0] = pix2[0];
         pix3[1] = pix2[1];
         pix3[2] = pix2[2];
      }

      pxbdrop = gdk_pixbuf_new_subpixbuf(pxbmix,px1,py1,px2-px1,py2-py1);
      gdk_cairo_set_source_pixbuf(ss_cr,pxbdrop,px1,py1);
      cairo_paint(ss_cr);
      g_object_unref(pxbdrop);
      dtime = 0.02 * (1.0 - pow(1.0*ii/Ndrops,0.1));
      if (dtime > 0.001) zsleep(dtime);
   }  

   g_object_unref(pxbmix);
   return;
}


/**************************************************************************/

//  menu function, start synchronize files process manually

void m_syncfiles(GtkWidget *, cchar *) 
{
   if (mod_keep()) return;                                                 //  unsaved edits
   if (! menulock(1)) return;                                              //  test and lock menus
   syncfiles_func(2);                                                      //  do the file sync
   menulock(0);
   return;
}


/**************************************************************************/

//  Synchronize files when fotoxx is installed for the first time or
//  when new image files have been created or moved outside fotoxx.
//  syncmode: 1   auto sync at fotoxx startup
//            2   manual sync via Synchronize Files menu

int syncfiles_func(int syncmode) 
{
   int   syncfiles_dialog_event(zdialog *zd, cchar *event);                //  private functions
   int   syncfiles_index_compare(cchar *rec1, cchar *rec2);

   GtkWidget      *widget;
   cchar          *syncreason;
   int            ii, cc, contx, ftyp;
   int            newfiles, dosync, yn, err;
   int            sync_count, sync_errors;
   char           *pp, *filespec, *imagefile, *thumbfile;
   char           tpdirk[200], tndirk[200];
   char           findcomm[500], search_index_file1[200];
   int            zstat, zdhh, ftf, Finitz = 0;
   int            Fnothumbs, Fhidethumbs, Fdirthumbs;
   zdialog        *zd;
   FILE           *fid;
   struct stat    statb;

   cchar    *sync_message = ZTX("Synchronize Files makes gallery windows fast \n"
                                "and enables metadata functions to work correctly.\n");

   zfuncs::F1_help_topic = "sync_files";
   
   printf("start sync files process, mode=%d \n",syncmode);

   dosync = 0;
   syncreason = "none";
   
   if (! Ntopdirk) {                                                       //  multiple top directories     v.12.02
      syncreason = ZTX("no top image directory is defined");
      dosync = 2;
   }

   if (Ntopdirk) {
      printf("top image directories: \n");
      for (ii = 0; ii < Ntopdirk; ii++) {                                  //  multiple top directories     v.12.02
         printf(" %s \n",topdirk[ii]);
         err = stat(topdirk[ii],&statb);
         if (err || ! S_ISDIR(statb.st_mode)) {
            syncreason = ZTX("top image directory is invalid");
            dosync = 2;
         }
      }
   }
   
   if (! dosync && Fthumbnails) {                                          //  thumbnails are used          v.12.08
      if (! thumbdirk) thumbdirk = strdupz("undefined",0,"thumbdirk");
      printf("thumbnail directory: %s \n",thumbdirk);
      if (strNeq(thumbdirk,"undefined")) {                                 //  if not "undefined", check
         err = stat(thumbdirk,&statb);                                     //    thumbnail directory valid
         if (err || ! S_ISDIR(statb.st_mode)) {
            syncreason = ZTX("thumbnail directory is invalid");
            dosync = 2;
         }
      }
   }

   if (! dosync && FsyncEnabled) {                                         //  file sync enabled            v.12.08
      snprintf(search_index_file1,200,"%s/index_001",search_index_dirk);   //                               v.12.10
      err = stat(search_index_file1,&statb);
      if (err) {                                                           //  check search index file exists
         syncreason = ZTX("no search index file is present");
         dosync = 2;
      }
      else if (strEqu(last_sync_time,"undefined")) {                       //  check last sync time is defined
         syncreason = ZTX("last file sync time is unknown");
         dosync = 2;
      }
   }

   if (FsyncEnabled && ! dosync)                                           //  file sync enabled, no problems found
   {
      newfiles = 0;                                                        //  check if new files since last sync

      for (ii = 0; ii < Ntopdirk; ii++)                                    //  v.12.08
      {
         sprintf(findcomm,"find -L \"%s\" -type d -newermt \"%s\" ",       //  use directory mod time
                                      topdirk[ii],last_sync_time);         //  compare to last file sync time
         printf("%s \n",findcomm);
         contx = 0;
         while ((pp = command_output(contx,findcomm,null))) {
            newfiles++;
            zfree(pp);
         }
      }

      if (newfiles) {
         syncreason = ZTX("new/modified files are present");               //  v.12.02
         dosync = 1;
      }
   }

   if (dosync) printf("%s \n",syncreason);
   else printf("%s \n",ZTX("file sync not needed"));
   
   if (dosync) FsyncNeeded = 1;                                            //  mark file sync needed           v.12.12

   if (syncmode == 1) {                                                    //  startup auto sync
      if (! dosync) return 0;                                              //  sync not needed
      if (dosync < 2) goto start_sync;                                     //  new files detected, do auto sync
   }

sync_dialog:                                                               //  manual sync request or user inputs needed

   //  Manual sync request, or missing information, or new install.
   //  Dialog to supply top image directory, sync and thumbnail options.

   zd = zdialog_new(ZTX("Synchronize Files"),mWin,Bhelp,Bproceed,Bcancel,null);
   zdhh = 400;

   if (strNeq(syncreason,"none")) {                                        //  show reason for file sync
      zdialog_add_widget(zd,"label","labwhy1","dialog",syncreason,"space=3");
      zdialog_add_widget(zd,"label","labwhy2","dialog",sync_message,"space=3");
      zdialog_add_widget(zd,"hsep","space","dialog",0,"space=3");
      zdhh = 500;
   }

   zdialog_add_widget(zd,"hbox","hbtop1","dialog",0,"space=4");
   zdialog_add_widget(zd,"label","labtop","hbtop1",ZTX("Top Image Directories:"),"space=5");
   zdialog_add_widget(zd,"frame","frtop","dialog",0,"expand");
   zdialog_add_widget(zd,"edit","topdirks","frtop");
   zdialog_add_widget(zd,"hbox","hbtop2","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labaddtop","hbtop2",ZTX("add directory"),"space=5");
   zdialog_add_widget(zd,"button","browsetop","hbtop2",Bbrowse,"space=4");
   
   zdialog_add_widget(zd,"hsep","sepnx","dialog",0,"space=3");             //  v.12.08
   zdialog_add_widget(zd,"hbox","hbnx1","dialog",0);
   zdialog_add_widget(zd,"check","dosync","hbnx1",ZTX("Use file sync function (recommended)"),"space=5");
   zdialog_add_widget(zd,"hbox","hbnx2","dialog",0);
   zdialog_add_widget(zd,"label","space","hbnx2",0,"space=10");
   zdialog_add_widget(zd,"label","labnx","hbnx2",ZTX("(if NO, metadata functions are disabled)"),"space=5");
   
   zdialog_add_widget(zd,"hsep","septn","dialog",0,"space=3");
   zdialog_add_widget(zd,"hbox","hbtn0","dialog",0);                       //  v.12.06
   zdialog_add_widget(zd,"hbox","hbtn1","dialog",0);
   zdialog_add_widget(zd,"hbox","hbtn2","dialog",0);
   zdialog_add_widget(zd,"hbox","hbtn3","dialog",0);
   zdialog_add_widget(zd,"label","labtn","hbtn0",ZTX("Thumbnail Options:"),"space=5");
   zdialog_add_widget(zd,"check","nothumbs","hbtn1",ZTX("None (galleries are slow)"),"space=10");
   zdialog_add_widget(zd,"check","hidethumbs","hbtn2",ZTX("Use hidden directories with images"),"space=10");
   zdialog_add_widget(zd,"check","dirthumbs","hbtn3",ZTX("Use Directory"),"space=10");
   zdialog_add_widget(zd,"entry","thumbdirk","hbtn3",0,"expand");
   zdialog_add_widget(zd,"hsep","septn","dialog",0,"space=3");
   
   zdialog_add_widget(zd,"hbox","hbinz","dialog",0,"space=3");
   zdialog_add_widget(zd,"check","initz","hbinz",ZTX("Initialize Everything"),"space=4");
   
   widget = zdialog_widget(zd,"topdirks");
   for (ii = 0; ii < Ntopdirk; ii++) {                                     //  stuff top directories     v.12.02
      wprintf(widget,topdirk[ii]);
      wprintf(widget,"\n");
   }

   zdialog_stuff(zd,"dosync",FsyncEnabled);                                //  stuff actual index status    v.12.08
   
   if (! thumbdirk || *thumbdirk != '/') {                                 //  thumbnail dirk not defined   v.12.09
      if (thumbdirk) zfree(thumbdirk);
      sprintf(tndirk,"%s/thumbnails",get_zuserdir());                      //  /.../.fotoxx/thumbnails
      thumbdirk = strdupz(tndirk,0,"thumbdirk");                           //  v.12.09
   }
   
   zdialog_stuff(zd,"thumbdirk",thumbdirk);                                //  stuff actual thumbnail status

   if (Fthumbnails) {
      zdialog_stuff(zd,"nothumbs",0);
      if (strEqu(thumbdirk,"undefined")) {                                 
         zdialog_stuff(zd,"hidethumbs",1);
         zdialog_stuff(zd,"dirthumbs",0);
      }
      else {
         zdialog_stuff(zd,"hidethumbs",0);
         zdialog_stuff(zd,"dirthumbs",1);
      }
   }
   else {
      zdialog_stuff(zd,"nothumbs",1);                                      //  no thumbnails
      zdialog_stuff(zd,"hidethumbs",0);
      zdialog_stuff(zd,"dirthumbs",0);
   }

   zdialog_help(zd,"sync_files");
   zdialog_resize(zd,500,zdhh);
   zdialog_run(zd,syncfiles_dialog_event);
   zstat = zdialog_wait(zd);

   if (zstat != 2) {                                                       //  dialog canceled
      zdialog_free(zd);
      if (dosync < 2) return 0;
      yn = zmessageYN(0,ZTX("File sync is necessary. Cancel anyway?"));    //  v.12.01
      if (! yn) goto sync_dialog;
      return 0;
   }

   widget = zdialog_widget(zd,"topdirks");
   ftf = 1;

   for (ii = 0; ii < 10; ii++)                                             //  get top directories       v.12.02
   {
      pp = wscanf(widget,ftf);                                             //  get text lines from edit widget
      if (! pp) break;
      strncpy0(tpdirk,pp,200);
      strTrim2(tpdirk);                                                    //  remove surrounding blanks
      cc = strlen(tpdirk);                                                 //  remove trailing '/' if present
      if (tpdirk[cc-1] == '/') tpdirk[cc-1] = 0; 
      if (cc < 2) { ii--; continue; }                                      //  ignore garbage
      if (topdirk[ii]) zfree(topdirk[ii]);
      topdirk[ii] = strdupz(tpdirk,0,"topdirk");
   }
   Ntopdirk = ii;                                                          //  top directory count
   
   zdialog_fetch(zd,"dosync",FsyncEnabled);                                //  get file sync option      v.12.08

   zdialog_fetch(zd,"thumbdirk",tndirk,200);                               //  get thumbnail options     v.12.06
   zfree(thumbdirk);
   thumbdirk = strdupz(tndirk,0,"thumbdirk");
   zdialog_fetch(zd,"nothumbs",Fnothumbs);
   zdialog_fetch(zd,"hidethumbs",Fhidethumbs);
   zdialog_fetch(zd,"dirthumbs",Fdirthumbs);

   zdialog_fetch(zd,"initz",Finitz);                                       //  get initialize flag       v.12.06

   zdialog_free(zd);                                                       //  kill dialog

   if (! Ntopdirk) {
      zmessageACK(mWin,ZTX("no top image directory is defined"));
      goto sync_dialog;
   }

   for (ii = 0; ii < Ntopdirk; ii++)                                       //  get top directories       v.12.02
   {
      err = stat(topdirk[ii],&statb);                                      //  check for validity
      if (err || ! S_ISDIR(statb.st_mode)) {
         zmessageACK(mWin,"%s \n %s",ZTX("directory is invalid"),topdirk[ii]);
         goto sync_dialog;
      }
   }

   fid = fopen(topdirk_file,"w");                                          //  write top directories file
   if (! fid) {
      zmessageACK(mWin,"cannot write top directories file");               //  fatal error
      gtk_main_quit();                                                     //  gone forever
      return 0;
   }

   for (ii = 0; ii < Ntopdirk; ii++)
      fprintf(fid,"%s\n",topdirk[ii]);
   fclose(fid);
   
   if (Fnothumbs + Fhidethumbs + Fdirthumbs != 1) {                        //  check thumbnail options
      zmessageACK(mWin,ZTX("select a thumbnail option"));
      goto sync_dialog;
   }

   if (Fdirthumbs) {                                                       //  use thumbnail directory   v.12.06
      err = mkdir(thumbdirk,0760);                                         //  create if missing         v.12.09
      err = stat(thumbdirk,&statb);                                        //  check validity
      if (err || ! S_ISDIR(statb.st_mode)) {
         zmessageACK(mWin,ZTX("thumbnail directory is invalid"));
         goto sync_dialog;
      }
   }
   else {
      zfree(thumbdirk);                                                    //  do not use
      thumbdirk = strdupz("undefined",0,"thumbdirk");
   }
   
   for (ii = 0; ii < Ntopdirk; ii++)
   {
      cc = strlen(topdirk[ii]);
      if (strnEqu(thumbdirk,topdirk[ii],cc)) break;
   }
   if (ii < Ntopdirk) {
      zmessageACK(mWin,ZTX("thumbnail directory cannot be under top directory"));
      goto sync_dialog;
   }

   Fthumbnails = 1 - Fnothumbs;                                            //  global flag, thumbnails are used

start_sync:

   //  popup window for log output, OK and cancel buttons  ================      v.12.12

   int  sync_popup_dialog_event(zdialog*, cchar *event);

   zdialog     *zdpopup;
   GtkWidget   *poplog;
   PangoFontDescription   *monofont;

   zdpopup = zdialog_new(ZTX("Synchronize Files"),mWin,"OK",Bcancel,null);
   zdialog_add_widget(zdpopup,"frame","frame","dialog",0,"expand");
   zdialog_add_widget(zdpopup,"scrwin","scrwin","frame");
   zdialog_add_widget(zdpopup,"text","text","scrwin");

   poplog = zdialog_widget(zdpopup,"text");
   monofont = pango_font_description_from_string("monospace 9");
   gtk_widget_modify_font(poplog,monofont);

   zdialog_resize(zdpopup,500,300);
   zdialog_run(zdpopup,sync_popup_dialog_event);

   //  ====================================================================

   sxrec_t     *sxrec_old = 0, *sxrec_new = 0;
   char        *Fupdate;
   char        **ppv, filedirk[maxfcc];
   char        *filename, *exifdate, *iptctags, *iptcrating; 
   char        *exifsize, *iptccapt, *exifcomms;
   char        *exifcity, *exifcountry, *exiflat, *exiflong;
   char        city2[100], country2[100], lat2[20], long2[20], gtags2[200];
   int         Nold = 0, Nnew = 0, orec, nrec, comp;
   struct tm   bdt;
   time_t      currtime;

   cchar    *exifkeys[11] = { "FileName", exif_date_key, iptc_keywords_key, 
            iptc_rating_key, exif_size_key, exif_comment_key, iptc_caption_key, 
            exif_city_key, exif_country_key, exif_latitude_key, exif_longitude_key };

   wprintf(poplog,"%s\n",ZTX("Synchronize Files (may need some time)"));

   Fkillfunc = 0;
   Ffuncbusy = 1;
   FsyncNeeded = 1;                                                        //  if started, must complete       v.12.12
   
   //  delete unneeded thumbnails =========================================
   
   if (Finitz || ! Fthumbnails)                                            //  no thumbnails wanted, or initialize all
   {                                                                       //  v.12.06
      wprintf(poplog,"%s\n",ZTX("delete all thumbnails"));

      for (ii = 0; ii < Ntopdirk; ii++)                                    //  delete .thumbnail dirks under top dirk
      {
         sprintf(findcomm,"find -L \"%s\" -depth -name .thumbnails -exec rm -r '{}' ';'",topdirk[ii]);
         err = system(findcomm);
         if (err) printf("*** %s \n",wstrerror(err));
      }
                                                                           //  if thumbnail directory used,
      if (strNeq(thumbdirk,"undefined")) {                                 //    delete contained files
         sprintf(findcomm,"find -L \"%s\"/* -delete",thumbdirk);
         err = system(findcomm);
         if (err) printf("*** %s \n",wstrerror(err));
      }

      goto update_thumbnails;
   }

   wprintf(poplog,"%s\n",ZTX("delete orphan thumbnails"));

   sync_count = 0;
   
   if (strEqu(thumbdirk,"undefined"))                                      //  thumbnail directory not used    v.12.06
   {
      for (ii = 0; ii < Ntopdirk; ii++)                                    //  multiple top directories        v.12.02
      {
         sprintf(findcomm,"find -L \"%s\" -name \"*.jpeg\"",topdirk[ii]);  //  find all *.jpeg files
         printf("%s \n",findcomm);
         contx = 0;

         while ((thumbfile = command_output(contx,findcomm,null)))         //  next /topdirk/.../filename.jpeg
         {
            if (strstr(thumbfile,"/.thumbnails/")) {                       //  this is a thumbnail file
               imagefile = thumb2imagefile(thumbfile);                     //  is image file present?
               if (! imagefile) {                                          //  no
                  remove(thumbfile);                                       //  delete thumbnail
                  sync_count++;                                            //  count thumbnails deleted
               }
               if (imagefile) zfree(imagefile);
            }
            zfree(thumbfile);
            zmainloop(100);                                                //  v.12.04
            if (Fkillfunc) {                                               //  v.12.11
               command_kill(contx);
               goto file_sync_exit;
            }
         }
      }
   }

   else                                                                    //  thumbnail directory is used     v.12.06
   {
      for (ii = 0; ii < Ntopdirk; ii++)                                    //  delete .thumbnail directories
      {                                                                    //    within image top directories
         sprintf(findcomm,"find -L \"%s\" -depth -name .thumbnails"
                          " -exec rm -r '{}' ';'",topdirk[ii]);
         err = system(findcomm);
         if (err) printf("*** %s \n",wstrerror(err));
      }
                                                                           //  if thumbnail directory used,
      sprintf(findcomm,"find -L \"%s\" -iname \"*.jpeg\"",thumbdirk);      //  find all *.jpeg files (thumbnails)
      printf("%s \n",findcomm);
      contx = 0;

      while ((thumbfile = command_output(contx,findcomm,null)))            //  next /thumbdirk/.../filename.jpeg
      {
         imagefile = thumb2imagefile(thumbfile);                           //  is image file present?
         if (! imagefile) {                                                //  no
            remove(thumbfile);                                             //  delete thumbnail
            sync_count++;                                                  //  count thumbnails deleted
         }
         if (imagefile) zfree(imagefile);
         zfree(thumbfile);
         zmainloop(100);
         if (Fkillfunc) {                                                  //  v.12.11
            command_kill(contx);
            goto file_sync_exit;
         }
      }
   }

   wprintf(poplog,ZTX("%d orphan thumbnails deleted \n"),sync_count);

update_thumbnails:

   //  look for image files with missing or stale thumbnails and create the thumbnails
   //  use fork() instead of threads because gtk_pixbuf functions are not thread safe

   if (! Fthumbnails) {                                                    //  v.12.06
      printf("thumbnail files are not used \n");
      goto search_index;
   }

   wprintf(poplog,"%s\n",ZTX("create missing and update stale thumbnails"));

   sync_count = sync_errors = 0;
   navi::genthumbs = 0;

   for (ii = 0; ii < Ntopdirk; ii++)                                       //  multiple top directories        v.12.02
   {
      sprintf(findcomm,"find -L \"%s\" -type f",topdirk[ii]);              //  all files under top directory   v.12.03
      wprintf(poplog,"%s\n",findcomm);
      contx = 0;
      *filedirk = 0;

      while ((filespec = command_output(contx,findcomm,null)))             //  loop each file
      {
         if (strlen(filespec) > maxfcc-1) {
            printf("filespec too big, ignored: %s \n",filespec);           //  v.12.12
            zfree(filespec);
            continue;
         }

         pp = strrchr(filespec,'/');                                       //  log each new directory          v.12.10
         if (! pp) continue;                                               //  should not happen
         *pp = 0;
         if (strNeq(filespec,filedirk)) {
            strcpy(filedirk,filespec);
            wprintf(poplog,"%s\n",filedirk);
         }
         *pp = '/';

         ftyp = image_file_type(filespec);                                 //  file type
         if (ftyp == 2 || ftyp == 3) {                                     //  skip non-image files
            sync_count++;
            thumbfile = image_thumbfile(filespec);                         //  check thumbnail and create if needed
            if (thumbfile) zfree(thumbfile);                               //  OK
            else {
               sync_errors++;                                              //  count errors
               if (sync_errors > 100 &&                                    //  if > 100 and > 20%
                  (1.0 * sync_errors / sync_count > 0.20)) {
                     wprintf(poplog,"%s\n","too many errors, STOP");
                     break;
               }
            }
         }
         zfree(filespec);
         zmainloop(10);
         if (Fkillfunc) {                                                  //  v.12.11
            command_kill(contx);
            goto file_sync_exit;
         }
      }
   }

   sync_count = navi::genthumbs;                                           //  count of created thumbnails

   wprintf(poplog,ZTX("%d thumbnails created, %d errors \n"),sync_count,sync_errors);
   
search_index:

   //  process search index file ==========================================
   //
   //  Rebuild search index file.
   //  Process all image files within given top-level directory.
   //  Works incrementally and is very fast after the first run.
   
   if (! FsyncEnabled) {                                                   //  v.12.08
      wprintf(poplog,"%s\n",ZTX("no search index file is in use"));
      wprintf(poplog,"%s\n",ZTX("COMPLETED"));
      printf("sync files completed \n");
      goto file_sync_exit;
   }

   if (Fnosync) {                                                          //  v.12.08
      wprintf(poplog,"%s\n","file sync suppressed with -nosync");
      wprintf(poplog,"%s\n",ZTX("COMPLETED"));
      printf("sync files completed \n");
      goto file_sync_exit;
   }

   wprintf(poplog,"%s\n",ZTX("rebuild search index"));

   cc = maximages * sizeof(sxrec_t);
   sxrec_old = (sxrec_t *) zmalloc(cc,"file-sync");                        //  curr. search index recs
   sxrec_new = (sxrec_t *) zmalloc(cc,"file-sync");                        //  new search index recs
   Fupdate = zmalloc(maximages,"file-sync");                               //  flag, update needed
   
   //  read current search index file and build "old list" of index recs

   if (! Finitz)                                                           //  re-initialize flag              v.12.06
   {
      ftf = 1;
      while (true) {
         err = read_search_index_seq(sxrec_old[Nold],ftf);                 //  read curr. index recs           v.12.10
         if (err) break;
         if (++Nold == maximages) zappcrash("too many image files");
      }
   }
   
   wprintf(poplog,ZTX("%d current index records found \n"),Nold);

   //  find all image files and create "new list" with empty data fields

   wprintf(poplog,"%s\n",ZTX("find all image files and build index records"));
   
   for (ii = 0; ii < Ntopdirk; ii++)                                       //  multiple top directories        v.12.02
   {
      sprintf(findcomm,"find -L \"%s\" -type f",topdirk[ii]);              //  all files under top directory   v.12.03
      wprintf(poplog,"%s\n",findcomm);
      contx = 0;

      while ((filespec = command_output(contx,findcomm,null)))
      {
         if (strlen(filespec) > maxfcc-1) {
            printf("filespec too big, ignored: %s \n",filespec);           //  v.12.12
            zfree(filespec);
            continue;
         }

         if (strstr(filespec,"/.")) {                                      //  skip all hidden directories     v.12.03
            zfree(filespec); continue; }

         if (image_file_type(filespec) != 2) {                             //  RAW files not indexed
            zfree(filespec); continue; }

         if (Nnew == maximages) {
            zmessageACK(mWin,ZTX("too many image files")); 
            command_kill(contx);
            goto file_sync_exit; 
         }
         nrec = Nnew++;

         sxrec_new[nrec].file = strdupz(filespec,0,"file-sync");           //  image filespec
         stat(filespec,&statb);
         gmtime_r(&statb.st_mtime,&bdt);                                   //  file date = yyyymmddhhmmss
         snprintf(sxrec_new[nrec].fdate,16,"%04d%02d%02d%02d%02d%02d",
                  bdt.tm_year+1900, bdt.tm_mon+1, bdt.tm_mday,
                  bdt.tm_hour, bdt.tm_min, bdt.tm_sec);
         sxrec_new[nrec].pdate[0] = 0;                                     //  image date = empty
         strcpy(sxrec_new[nrec].rating,"0");                               //  stars = "0"
         strcpy(sxrec_new[nrec].size,"0x0");                               //  size = "0x0"
         sxrec_new[nrec].tags = 0;                                         //  tags = empty
         sxrec_new[nrec].capt = 0;                                         //  caption = empty
         sxrec_new[nrec].comms = 0;                                        //  comments = empty
         sxrec_new[nrec].gtags = 0;                                        //  geotags = empty

         zfree(filespec);
         zmainloop(10);

         if (Fkillfunc) {                                                  //  v.12.11
            command_kill(contx);
            goto file_sync_exit;
         }
      }
   }

   wprintf(poplog,ZTX("found %d image files \n"),Nnew);

   //  merge and compare lists
   //  if filespecs match and have the same date, then current record is OK

   wprintf(poplog,"%s\n",ZTX("merging old and new index records"));

   if (Nold)
      HeapSort((char *) sxrec_old,sizeof(sxrec_t),Nold,syncfiles_index_compare);
   if (Nnew)
      HeapSort((char *) sxrec_new,sizeof(sxrec_t),Nnew,syncfiles_index_compare);

   sync_count = 0;

   for (orec = nrec = 0; nrec < Nnew; nrec++)                              //  12.03
   {
      Fupdate[nrec] = 1;                                                   //  assume nrec update is needed

      if (orec == Nold) continue;                                          //  no more old recs

      while (true) 
      {
         if (orec == Nold) break;
         comp = strcasecmp(sxrec_old[orec].file,sxrec_new[nrec].file);     //  compare orec file to nrec file
         if (comp == 0)                                                    //  v.12.11
            comp = strcmp(sxrec_old[orec].file,sxrec_new[nrec].file);
         if (comp >= 0) break;                                             //  orec >= nrec
         orec++;                                                           //  orec < nrec, next orec
         continue;
      }

      if (comp == 0)                                                       //  orec = nrec (same image file)
      {
         if (strEqu(sxrec_new[nrec].fdate,sxrec_old[orec].fdate))          //  file dates match
         {
            strncpy0(sxrec_new[nrec].pdate,sxrec_old[orec].pdate,12);      //  copy data from old to new
            strncpy0(sxrec_new[nrec].rating,sxrec_old[orec].rating,2);
            strncpy0(sxrec_new[nrec].size,sxrec_old[orec].size,15);        //  v.13.01
            sxrec_new[nrec].tags = sxrec_old[orec].tags;
            sxrec_old[orec].tags = 0;
            sxrec_new[nrec].capt = sxrec_old[orec].capt;
            sxrec_old[orec].capt = 0;
            sxrec_new[nrec].comms = sxrec_old[orec].comms;
            sxrec_old[orec].comms = 0;
            sxrec_new[nrec].gtags = sxrec_old[orec].gtags;
            sxrec_old[orec].gtags = 0;
            Fupdate[nrec] = 0;                                             //  nrec update not needed
            sync_count++;
         }
         orec++;                                                           //  next orec
      }

      if (Fkillfunc) goto file_sync_exit;                                  //  v.12.11
   }

   wprintf(poplog,ZTX("files found in current index recs: %d \n"),sync_count);

   for (orec = 0; orec < Nold; orec++)                                     //  release old list memory
   {
      zfree(sxrec_old[orec].file);
      if (sxrec_old[orec].tags) zfree(sxrec_old[orec].tags);
      if (sxrec_old[orec].capt) zfree(sxrec_old[orec].capt);
      if (sxrec_old[orec].comms) zfree(sxrec_old[orec].comms);
      if (sxrec_old[orec].gtags) zfree(sxrec_old[orec].gtags);
   }

   zfree(sxrec_old);
   sxrec_old = 0;

   //  process entries needing update in new index list
   //  (new files or files dated later than search index date)
   //  get updated metadata from image file EXIF/IPTC data
   
   wprintf(poplog,"%s\n",ZTX("updating index records"));

   sync_count = 0;
   *filedirk = 0;

   for (nrec = 0; nrec < Nnew; nrec++)
   {
      if (! Fupdate[nrec]) continue;                                       //  no update needed
      sync_count++;                                                        //  running count of updates

      filespec = sxrec_new[nrec].file;

      pp = strrchr(filespec,'/');                                          //  log each new directory       v.12.10
      *pp = 0;
      if (strNeq(filespec,filedirk)) {
         strcpy(filedirk,filespec);
         wprintf(poplog,"%s\n",filedirk);
      }
      *pp = '/';
      filename = pp + 1;

      ppv = exif_get(filespec,exifkeys,11);                                //  get iptc/exif metadata       v.12.03
      if (! ppv) {
         zmessLogACK(mWin,"exif_get() fatal error");
         goto file_sync_exit;
      }
      
      if (! ppv[0] || strNeq(ppv[0],filename))                             //  check exiftool server working
         printf("exif_get() failure: %s\n",filespec);                      //          v.12.10

      if (ppv[0]) zfree(ppv[0]);

      exifdate = ppv[1];
      iptctags = ppv[2];
      iptcrating = ppv[3];
      exifsize = ppv[4];
      exifcomms = ppv[5];
      iptccapt = ppv[6];
      exifcity = ppv[7];
      exifcountry = ppv[8];
      exiflat = ppv[9];
      exiflong = ppv[10];
      
      if (exifdate && strlen(exifdate) > 3) {                              //  exif date (photo date)
         if (strlen(exifdate) > 9) exifdate[10] = 0;                       //  truncate to yyyy:mm:dd
         strcpy(sxrec_new[nrec].pdate,exifdate);
      }
      else strcpy(sxrec_new[nrec].pdate,"null");                           //  not present
      
      if (iptcrating && strlen(iptcrating)) {                              //  iptc rating
         sxrec_new[nrec].rating[0] = *iptcrating;
         sxrec_new[nrec].rating[1] = 0;                                    //  v.12.10
      }
      else strcpy(sxrec_new[nrec].rating,"0");                             //  not present
      
      if (exifsize && strlen(exifsize))                                    //  exif size           v.13.01
         strncpy0(sxrec_new[nrec].size,exifsize,15);

      if (iptctags && strlen(iptctags)) {                                  //  iptc tags
         sxrec_new[nrec].tags = iptctags;
         iptctags = 0;
      }

      if (iptccapt && strlen(iptccapt)) {                                  //  iptc caption 
         sxrec_new[nrec].capt = iptccapt;
         iptccapt = 0;
      }

      if (exifcomms && strlen(exifcomms)) {                                //  exif comments
         sxrec_new[nrec].comms = exifcomms;
         exifcomms = 0;
      }

      strcpy(city2,"null");                                                //  geotags = empty     v.12.05
      strcpy(country2,"null");
      strcpy(lat2,"null");
      strcpy(long2,"null");

      if (exifcity) strncpy0(city2,exifcity,99);                           //  get from exif if any
      if (exifcountry) strncpy0(country2,exifcountry,99);
      if (exiflat) strncpy0(lat2,exiflat,10);
      if (exiflong) strncpy0(long2,exiflong,10);

      *gtags2 = 0;
      strncatv(gtags2,200,city2,"^ ",country2,"^ ",lat2,"^ ",long2,null);
      sxrec_new[nrec].gtags = strdupz(gtags2,0,"file-sync");

      if (exifdate) zfree(exifdate);                                       //  free EXIF data
      if (iptctags) zfree(iptctags);
      if (iptcrating) zfree(iptcrating);
      if (exifsize) zfree(exifsize);                                       //  v.13.01
      if (exifcomms) zfree(exifcomms);
      if (iptccapt) zfree(iptccapt);
      if (exifcity) zfree(exifcity);
      if (exifcountry) zfree(exifcountry);
      if (exiflat) zfree(exiflat);
      if (exiflong) zfree(exiflong);

      zmainloop(10);                                                       //  keep window alive
      if (Fkillfunc) goto file_sync_exit;                                  //  v.12.11
   }

   ftf = 1;
   for (nrec = 0; nrec < Nnew; nrec++) {                                   //  write updated search index recs.
      err = write_search_index_seq(&sxrec_new[nrec],ftf);                  //  v.12.10
      if (err) {
         zmessLogACK(mWin,"file sync terminated by error");
         goto file_sync_exit;
      }
   }
   write_search_index_seq(null,ftf);                                       //  close output

   for (nrec = 0; nrec < Nnew; nrec++)                                     //  release new list memory
   {
      zfree(sxrec_new[nrec].file);
      if (sxrec_new[nrec].tags) zfree(sxrec_new[nrec].tags);
      if (sxrec_new[nrec].capt) zfree(sxrec_new[nrec].capt);
      if (sxrec_new[nrec].comms) zfree(sxrec_new[nrec].comms);
      if (sxrec_new[nrec].gtags) zfree(sxrec_new[nrec].gtags);
   }

   zfree(sxrec_new);
   sxrec_new = 0;
   zfree(Fupdate);

   wprintf(poplog,ZTX("%d image files updated \n"),sync_count);

   currtime = time(0);                                                     //  v.12.08
   last_sync_time = strdupz(ctime(&currtime) + 4,0,"file-sync");           //  set last file sync time
   if (last_sync_time[4] == ' ') last_sync_time[4] = '0';                  //  format  mmm dd hh:mm:ss yyyy
   pp = strchr(last_sync_time,'\n');
   if (pp) *pp = 0;

   wprintf(poplog,"%s\n",ZTX("COMPLETED"));
   printf("sync files completed \n");
   save_params();                                                          //  update sync time parameter
   mwpaint2();                                                             //  refresh window               v.12.10

file_sync_exit:

   if (sxrec_old) zfree(sxrec_old);                                        //  v.12.12
   if (sxrec_new) zfree(sxrec_new);
   sxrec_old = sxrec_new = 0;

   if (Fkillfunc) {                                                        //  v.12.11
      wprintf(poplog,"%s\n",ZTX("KILLED"));
      printf("*** synchronize files killed \n");
      Ffuncbusy = 0;
      goto sync_dialog;
   }

   Ffuncbusy = 0;
   Fkillfunc = 0;
   FsyncNeeded = 0;                                                        //  file sync done, not needed      v.12.12
   mwpaint2();
   return 0;
}


//  sync dialog event function - file chooser for top image directory

int syncfiles_dialog_event(zdialog *zd, cchar *event)
{
   GtkWidget   *widget;
   char        *pp, newdirk[200];
   int         nn;
   
   if (zd->zstat == 1) {                                                   //  [help]        v.12.05
      zd->zstat = 0;
      showz_userguide("sync_files");
   }

   if (strEqu(event,"browsetop")) {
      pp = zgetfile1(ZTX("Select top image directory"),"folder",newdirk);  //  get new topmost directory
      if (! pp) return 0;
      widget = zdialog_widget(zd,"topdirks");
      wprintf(widget,"%s\n",pp);
      zfree(pp);
   }
   
   if (strEqu(event,"nothumbs")) {                                         //  use no thumbnails
      zdialog_fetch(zd,"nothumbs",nn);
      if (nn) {                                                            //  if on, set other options off
         zdialog_stuff(zd,"hidethumbs",0);
         zdialog_stuff(zd,"dirthumbs",0);
      }
   }
   
   if (strEqu(event,"hidethumbs")) {                                       //  use default thumbnails
      zdialog_fetch(zd,"hidethumbs",nn);
      if (nn) {
         zdialog_stuff(zd,"nothumbs",0);
         zdialog_stuff(zd,"dirthumbs",0);
      }
   }
   
   if (strEqu(event,"dirthumbs")) {                                        //  use thumbnail directory
      zdialog_fetch(zd,"dirthumbs",nn);
      if (nn) {
         zdialog_stuff(zd,"nothumbs",0);
         zdialog_stuff(zd,"hidethumbs",0);
      }
   }

   return 0;
}


//  sync popup dialog response function

int sync_popup_dialog_event(zdialog *zd, cchar *event)
{
   if (! zd->zstat) return 0;
   if (zd->zstat == 2) Fkillfunc = 2;                                      //  [cancel]
   if (! Ffuncbusy) zdialog_free(zd);
   else zd->zstat = 0;                                                     //  keep dialog open
   return 0;
}


//  sort compare function - compare index record filespecs and return
//   <0 | 0 | >0   for   file1 < | == | > file2

int syncfiles_index_compare(cchar *rec1, cchar *rec2)
{
   char * file1 = ((sxrec_t *) rec1)->file;
   char * file2 = ((sxrec_t *) rec2)->file;
   int nn = strcasecmp(file1,file2);                                       //  compare ignoring case     v.12.10
   if (nn) return nn;
   nn = strcmp(file1,file2);                                               //  if equal, use utf8 compare
   return nn;
}


/**************************************************************************/

//  Show RGB values for 1-10 pixels selected with mouse-clicks.

void  show_RGB_mousefunc();
int   show_RGB_timefunc(void *);

zdialog     *RGBSzd;
int         RGBSpixel[10][2];                                              //  last 10 pixels clicked
int         RGBSnpix;                                                      //  count of pixels
int         RGBSmetric = 1;                                                //  1/2/3 = /RGB/EV/OD
int         RGBSdelta = 0;                                                 //  abs/delta mode
int         RGBSlabels = 0;                                                //  pixel labels on/off
int         RGBSupdate = 0;                                                //  window update needed


void m_show_RGB(GtkWidget *, cchar *menu)
{
   int   show_RGB_event(zdialog *zd, cchar *event);

   PangoFontDescription    *pfontdesc;
   GtkWidget   *widget;
   cchar       *mess = ZTX("Click image to select pixels.");
   cchar       *header = " Pixel            Red     Green   Blue";
   char        hbx[8] = "hbx", pixx[8] = "pixx";                           //  last char. is '0' to '9'
   int         ii;

   zfuncs::F1_help_topic = "show_RGB";

   if (! Fpxb) return;                                                     //  no image file
   RGBSnpix = 0;                                                           //  no pixels yet
   pfontdesc = pango_font_description_from_string("Monospace 9");          //  monospace font

/***
   
   Click image to select pixels.
   [x] delta  [x] labels
   Metric: (o) RGB  (o) EV  (o) OD
   Pixel           Red    Green  Blue
   A xxxxx xxxxx   xxxxx  xxxxx  xxxxx
   B xxxxx xxxxx   xxxxx  xxxxx  xxxxx
   C xxxxx xxxxx   xxxxx  xxxxx  xxxxx
   D xxxxx xxxxx   xxxxx  xxxxx  xxxxx
   E xxxxx xxxxx   xxxxx  xxxxx  xxxxx
   F xxxxx xxxxx   xxxxx  xxxxx  xxxxx
   G xxxxx xxxxx   xxxxx  xxxxx  xxxxx
   H xxxxx xxxxx   xxxxx  xxxxx  xxxxx
   I xxxxx xxxxx   xxxxx  xxxxx  xxxxx
   J xxxxx xxxxx   xxxxx  xxxxx  xxxxx

                            [done]
***/

   if (RGBSzd) zdialog_free(RGBSzd);                                       //  delete previous if any
   zdialog *zd = zdialog_new(ZTX("Show RGB"),mWin,Bclear,Bdone,null);
   RGBSzd = zd;

   zdialog_add_widget(zd,"hbox","hbmess","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labmess","hbmess",mess,"space=5");

   zdialog_add_widget(zd,"hbox","hbmym","dialog");
   zdialog_add_widget(zd,"check","delta","hbmym","delta","space=8");
   zdialog_add_widget(zd,"check","labels","hbmym","labels","space=8");

   if (RGBSdelta) zdialog_stuff(zd,"delta",1);

   zdialog_add_widget(zd,"hbox","hbmetr","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labmetr","hbmetr","Metric:","space=5");
   zdialog_add_widget(zd,"radio","radRGB","hbmetr","RGB","space=3");
   zdialog_add_widget(zd,"radio","radEV","hbmetr","EV","space=3");
   zdialog_add_widget(zd,"radio","radOD","hbmetr","OD","space=3");
   
   if (RGBSmetric == 1) zdialog_stuff(zd,"radRGB",1);                      //  get which metric to use
   if (RGBSmetric == 2) zdialog_stuff(zd,"radEV",1);
   if (RGBSmetric == 3) zdialog_stuff(zd,"radOD",1);

   zdialog_add_widget(zd,"vbox","vbdat","dialog");                         //  vbox for current pixel values
   zdialog_add_widget(zd,"hbox","hbpix","vbdat");
   zdialog_add_widget(zd,"label","labheader","hbpix",header);              //  Pixel        Red    Green  Blue
   widget = zdialog_widget(zd,"labheader"); 
   gtk_widget_modify_font(widget,pfontdesc);
   
   for (ii = 0; ii < 10; ii++)
   {                                                                       //  10 labels for pixel data output
      hbx[2] = '0' + ii;
      pixx[3] = '0' + ii;
      zdialog_add_widget(zd,"hbox",hbx,"vbdat");
      zdialog_add_widget(zd,"label",pixx,hbx);
      widget = zdialog_widget(zd,pixx);                                    //  use monospace font
      gtk_widget_modify_font(widget,pfontdesc);
   }

   zdialog_help(zd,"show_RGB");                                            //  zdialog help topic
   zdialog_run(zd,show_RGB_event,"save");                                  //  run dialog
   takeMouse(show_RGB_mousefunc,dragcursor);                               //  connect mouse function
   g_timeout_add(300,show_RGB_timefunc,0);                                 //  start timer function, 300 ms

   return;
}


//  dialog event function

int show_RGB_event(zdialog *zd, cchar *event)
{
   int            button;

   if (zd->zstat) {
      if (zd->zstat == 1) {                                                //  clear
         zd->zstat = 0;                                                    //  keep dialog active
         RGBSnpix = 0;                                                     //  point count = 0
         erase_toptext(102);                                               //  erase labels on image
         RGBSupdate++;                                                     //  update display window        v.12.08
      }
      else {                                                               //  done or kill
         freeMouse();                                                      //  disconnect mouse function
         zdialog_free(RGBSzd);                                             //  kill dialog
         RGBSzd = 0;
         erase_toptext(102);
      }
      mwpaint2();
   }

   if (strEqu(event,"focus"))                                              //  toggle mouse capture
      takeMouse(show_RGB_mousefunc,dragcursor);                            //  connect mouse function
   
   if (strEqu(event,"delta")) {                                            //  set absolute/delta mode
      zdialog_fetch(zd,"delta",RGBSdelta);
      if (RGBSdelta && ! E3pxm) {
         RGBSdelta = 0;                                                    //  block delta mode if no edit underway
         zdialog_stuff(zd,"delta",0);
      }
   }
   
   if (strEqu(event,"labels")) {                                           //  get labels on/off
      zdialog_fetch(zd,"labels",RGBSlabels);
      RGBSupdate++;                                                        //  update window
   }
   
   if (strEqu(event,"radRGB")) {                                           //  metric = RGB
      zdialog_fetch(zd,event,button);
      if (button) RGBSmetric = 1;
   }
      
   if (strEqu(event,"radEV")) {                                            //  metric = EV
      zdialog_fetch(zd,event,button);
      if (button) RGBSmetric = 2;
   }
      
   if (strEqu(event,"radOD")) {                                            //  metric = OD
      zdialog_fetch(zd,event,button);
      if (button) RGBSmetric = 3;
   }
      
   return 0;
}


//  mouse function

void show_RGB_mousefunc()                                                  //  mouse function
{
   int            ii;

   if (! LMclick) return;

   LMclick = 0;

   if (RGBSnpix == 10) {                                                   //  if all 10 positions filled,
      for (ii = 1; ii < 10; ii++) {                                        //    remove first (oldest) and
         RGBSpixel[ii-1][0] = RGBSpixel[ii][0];                            //      push the rest back
         RGBSpixel[ii-1][1] = RGBSpixel[ii][1];
      }
      RGBSnpix = 9;                                                        //  position for newest
   }
   
   ii = RGBSnpix;                                                          //  next position to fill
   RGBSpixel[ii][0] = Mxclick;                                             //  save newest pixel
   RGBSpixel[ii][1] = Myclick;
   RGBSnpix++;

   RGBSupdate++;                                                           //  update window   
   return;
}


//  timer function - continuously display RGB values for selected pixels

int show_RGB_timefunc(void *arg)                                           //  up to 10 pixels, live update
{
   static char    label[10][4] = { " A ", " B ", " C ", " D ", " E ",
                                   " F ", " G ", " H ", " I ", " J " };

   PXM         *pxm = 0;
   int         ii, jj, px, py, delta;
   int         ww, hh;
   float       red1, green1, blue1;
   float       red3, green3, blue3;
   float       *ppixa, *ppixb;
   double      c1 = 100.0 / 256.0;
   char        text[100], pixx[8] = "pixx";
   
   #define ODfunc(rgb) (2.0 - log10(c1 * rgb))                             //  RGB units (1-256) to OD units
   #define EVfunc(rgb) (log2(rgb) - 7)                                     //  RGB units to EV (128 == 0 EV)

   if (! RGBSzd) return 1;                                                 //  user quit

   if (RGBSdelta && E3pxm) delta = 1;                                      //  delta mode only if edit underway
   else delta = 0;

   if (E3pxm) pxm = E3pxm;                                                 //  report image being edited
   else {
      if (! E1pxm) {                                                       //  if image never edited,
         E1pxm = PXM_load(curr_file,1);                                    //    create E1 image for my use
         if (! E1pxm) return 1;
      }
      pxm = E1pxm;
   }
   
   ww = pxm->ww;
   hh = pxm->hh;

   if (RGBSnpix)                                                           //  bugfix                    v.12.08
   {
      for (ii = 0; ii < 10; ii++)
      {
         px = RGBSpixel[ii][0];                                            //  next pixel to report
         py = RGBSpixel[ii][1];
         if (px >= 0 && px < ww && py >= 0 && py < hh) continue;           //  within image limits
         for (jj = ii+1; jj < 10; jj++) {
            RGBSpixel[jj-1][0] = RGBSpixel[jj][0];                         //  remove pixel outside limits
            RGBSpixel[jj-1][1] = RGBSpixel[jj][1];                         //    and pack the remaining down
         }
         RGBSpixel[9][0] = RGBSpixel[9][1] = 0;                            //  last position is empty
         RGBSnpix--;
         ii--;
         RGBSupdate++;                                                     //  update window
      }
   }

   if (RGBSupdate)                                                         //  refresh window labels
   {
      RGBSupdate = 0;
      erase_toptext(102);

      if (RGBSlabels) {
         for (ii = 0; ii < RGBSnpix; ii++) {                               //  show pixel labels on image
            px = RGBSpixel[ii][0];
            py = RGBSpixel[ii][1];
            add_toptext(102,px,py,label[ii],"Sans 8");
         }
      }
      
      mwpaint2();
   }

   red1 = green1 = blue1 = 0;

   for (ii = 0; ii < 10; ii++)                                             //  loop positons 0 to 9
   {
      pixx[3] = '0' + ii;                                                  //  widget names "pix0" ... "pix9"

      px = RGBSpixel[ii][0];                                               //  next pixel to report
      py = RGBSpixel[ii][1];
      if (ii >= RGBSnpix) {                                                //  no pixel there yet
         zdialog_stuff(RGBSzd,pixx,"");                                    //  blank report line
         continue;
      }
      
      ppixa = PXMpix(pxm,px,py);                                           //  get pixel RGB values
      red3 = ppixa[0];
      green3 = ppixa[1];
      blue3 = ppixa[2];
      
      if (delta && px < E1pxm->ww && py < E1pxm->hh) {                     //  delta RGB for ongoing edited image
         ppixb = PXMpix(E1pxm,px,py);                                      //  "before" image E1
         red1 = ppixb[0]; 
         green1 = ppixb[1];
         blue1 = ppixb[2];
      }

      sprintf(text," %c %5d %5d  ",'A'+ii,px,py);                          //  format pixel " A xxxx yyyy"
      
      if (RGBSmetric == 1) {                                               //  output RGB values
         if (delta) {
            red3 -= red1;                                                  //  delta RGB
            green3 -= green1;
            blue3 -= blue1;
         }
         sprintf(text+14,"   %6.2f  %6.2f  %6.2f ",red3,green3,blue3);     //  xxx.xx
      }

      if (RGBSmetric == 2) {                                               //  output EV values
         red3 = EVfunc(red3);
         green3 = EVfunc(green3);
         blue3 = EVfunc(blue3);
         if (delta) {
            red3 -= EVfunc(red1);                                          //  delta EV
            green3 -= EVfunc(green1);
            blue3 -= EVfunc(blue1);
         }
         sprintf(text+14,"   %6.3f  %6.3f  %6.3f ",red3,green3,blue3);     //  x.xxx
      }

      if (RGBSmetric == 3) {                                               //  output OD values
         red3 = ODfunc(red3); 
         green3 = ODfunc(green3);
         blue3 = ODfunc(blue3);
         if (delta) {
            red3 -= ODfunc(red1);                                          //  delta OD
            green3 -= ODfunc(green1);
            blue3 -= ODfunc(blue1);
         }
         sprintf(text+14,"   %6.3f  %6.3f  %6.3f ",red3,green3,blue3);     //  x.xxx
      }

      zdialog_stuff(RGBSzd,pixx,text);                                     //  pixel and RGB values >> dialog
   }

   return 1;
}


/**************************************************************************/

//  setup x and y grid lines - count/spacing, enable/disable, offsets

void m_gridlines(GtkWidget *, cchar *menu)
{
   int gridlines_dialog_event(zdialog *zd, cchar *event);
   
   if (! menu || strEqu(menu,"G")) {                                       //  from KB shortcut       v.13.01
      toggle_grid(2);
      return;
   }

   zdialog     *zd;
   
   zfuncs::F1_help_topic = "grid_lines";

   zd = zdialog_new(ZTX("Grid Lines"),mWin,Bdone,Bcancel,null);
   
   zdialog_add_widget(zd,"hbox","hb0","dialog",0,"space=10");
   zdialog_add_widget(zd,"vbox","vb1","hb0",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb2","hb0",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vbspace","hb0",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb3","hb0",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb4","hb0",0,"homog|space=5");

   zdialog_add_widget(zd,"label","lab1x","vb1",ZTX("x-spacing"));
   zdialog_add_widget(zd,"label","lab2x","vb1",ZTX("x-count"));
   zdialog_add_widget(zd,"label","lab4x","vb1",ZTX("x-enable"));

   zdialog_add_widget(zd,"spin","spacex","vb2","10|200|1|50");
   zdialog_add_widget(zd,"spin","countx","vb2","0|100|1|2");
   zdialog_add_widget(zd,"check","enablex","vb2",0);

   zdialog_add_widget(zd,"label","lab1y","vb3",ZTX("y-spacing"));
   zdialog_add_widget(zd,"label","lab2y","vb3",ZTX("y-count"));
   zdialog_add_widget(zd,"label","lab4y","vb3",ZTX("y-enable"));

   zdialog_add_widget(zd,"spin","spacey","vb4","10|200|1|50");
   zdialog_add_widget(zd,"spin","county","vb4","0|100|1|2");
   zdialog_add_widget(zd,"check","enabley","vb4",0);

   zdialog_add_widget(zd,"hbox","hboffx","dialog");
   zdialog_add_widget(zd,"label","lab3x","hboffx",ZTX("x-offset"),"space=7");
   zdialog_add_widget(zd,"hscale","offsetx","hboffx","-100|100|1|0","expand");

   zdialog_add_widget(zd,"hbox","hboffy","dialog");
   zdialog_add_widget(zd,"label","lab3y","hboffy",ZTX("y-offset"),"space=7");
   zdialog_add_widget(zd,"hscale","offsety","hboffy","-100|100|1|0","expand");

   zdialog_stuff(zd,"spacex",gridspace[0]);
   zdialog_stuff(zd,"spacey",gridspace[1]);
   zdialog_stuff(zd,"countx",gridcount[0]);
   zdialog_stuff(zd,"county",gridcount[1]);
   zdialog_stuff(zd,"offsetx",gridoffset[0]);
   zdialog_stuff(zd,"offsety",gridoffset[1]);
   zdialog_stuff(zd,"enablex",gridon[0]);
   zdialog_stuff(zd,"enabley",gridon[1]);

   if (gridon[0] || gridon[1]) Fgrid = 1;
   mwpaint2();

   zdialog_help(zd,"grid_lines");                                          //  zdialog help topic
   zdialog_run(zd,gridlines_dialog_event);
   zdialog_wait(zd);
   return;
}


//  dialog event function

int gridlines_dialog_event(zdialog *zd, cchar *event)
{
   int      zstat;

   if (zd->zstat) {
      zstat = zd->zstat;
      if (zstat != 1) Fgrid = gridon[0] = gridon[1] = 0;                   //  cancel, grid lines off
      zdialog_free(zd);
      mwpaint2();
      return 0;
   }

   if (strEqu(event,"enablex"))                                            //  x/y grid enable or disable
      zdialog_fetch(zd,"enablex",gridon[0]);
   
   if (strEqu(event,"enabley")) 
      zdialog_fetch(zd,"enabley",gridon[1]);
   
   if (strEqu(event,"spacex"))                                             //  x/y grid spacing (if counts == 0)
      zdialog_fetch(zd,"spacex",gridspace[0]);

   if (strEqu(event,"spacey")) 
      zdialog_fetch(zd,"spacey",gridspace[1]);

   if (strEqu(event,"countx"))                                             //  x/y grid line counts
      zdialog_fetch(zd,"countx",gridcount[0]);

   if (strEqu(event,"county")) 
      zdialog_fetch(zd,"county",gridcount[1]);

   if (strEqu(event,"offsetx"))                                            //  x/y grid starting offsets
      zdialog_fetch(zd,"offsetx",gridoffset[0]);

   if (strEqu(event,"offsety")) 
      zdialog_fetch(zd,"offsety",gridoffset[1]);
   
   if (gridon[0] || gridon[1]) Fgrid = 1;                                  //  if either grid enabled, show grid
   else Fgrid = 0;

   mwpaint2();
   return 0;
}


//  load or save grid settings from or to external int array

void load_grid(int *griddata)
{
   Fgrid = griddata[0];
   gridon[0] = griddata[1];
   gridon[1] = griddata[2];
   gridspace[0] = griddata[3];
   gridspace[1] = griddata[4];
   gridcount[0] = griddata[5];
   gridcount[1] = griddata[6];
   if (! gridcount[0] && ! gridspace[0]) gridcount[0] = 5;                 //  if never set, use 5 lines
   if (! gridcount[1] && ! gridspace[1]) gridcount[1] = 5;
   mwpaint2();
   return;
}

void save_grid(int *griddata)
{
   griddata[0] = Fgrid;
   griddata[1] = gridon[0];
   griddata[2] = gridon[1];
   griddata[3] = gridspace[0];
   griddata[4] = gridspace[1];
   griddata[5] = gridcount[0];
   griddata[6] = gridcount[1];
   return;
}


//  toggle grid lines on or off
//  action: 0 = off, 1 = on, 2 = toggle: on > off, off > on

void toggle_grid(int action)
{
   if (action == 0) Fgrid = 0;                                             //  grid off
   if (action == 1) Fgrid = 1;                                             //  grid on
   if (action == 2) Fgrid = 1 - Fgrid;                                     //  toggle grid
   
   if (Fgrid && ! gridon[0] && ! gridon[1])                                //  if grid on and x/y both off,
      gridon[0] = gridon[1] = 1;                                           //    set both grids on

   mwpaint2();
   return;
}


/**************************************************************************/

//  burn images to CD/DVD  

void  m_burn(GtkWidget *, cchar *)
{
   int      ii, cc, err;
   char     **filelist, *imagefile, *bcommand;

   if (checkmenulock()) return;                                            //  menu locked               v.12.09
   if (mod_keep()) return;                                                 //  unsaved edits
   
   zfuncs::F1_help_topic = "burn";

   filelist = gallery_getfiles();                                          //  get list of files to burn
   if (! filelist) return;                                                 //  nothing selected
   
   cc = 0;
   for (ii = 0; filelist[ii]; ii++)                                        //  get memory for brasero command line
      cc += strlen(filelist[ii]) + 4;

   bcommand = zmalloc(cc+20,"brasero");
   strcpy(bcommand,"brasero");
   cc = strlen(bcommand);

   for (ii = 0; filelist[ii]; ii++)                                        //  copy files to command line
   {
      imagefile = filelist[ii];
      strcpy(bcommand+cc," \"");
      cc += 2;
      strcpy(bcommand+cc,imagefile);
      cc += strlen(imagefile);
      strcpy(bcommand+cc,"\"");
      cc += 1;
      zfree(imagefile);
   }
   
   zfree(filelist);

   strcat(bcommand," &");                                                  //  brasero command in background
   err = system(bcommand);
   if (err) zmessageACK(mWin,"error: %s",wstrerror(err));
   zfree(bcommand);

   return;
}


/**************************************************************************/

//  monitor test function

void m_moncheck(GtkWidget *, cchar *)
{
   int         err;
   char        colorchart[200];
   zdialog     *zd;
   char        *savecurrent = 0;
   
   cchar       *message = ZTX("Brightness should show a gradual ramp \n"
                              "extending all the way to the edges.");
   
   zfuncs::F1_help_topic = "check_monitor";

   if (mod_keep()) return;                                                 //  unsaved edits
   if (! menulock(1)) return;

   if (curr_file) savecurrent = strdupz(curr_file,0,"moncheck");
   
   snprintf(colorchart,200,"%s/images/colorchart.png",get_zdocdir());      //  color chart .png file
   err = f_open(colorchart,0);
   if (err) goto restore;

   Fzoom = 1;                                                              //  scale 100% (required)
   gtk_window_set_title(MWIN,"monitor check");

   zd = zdialog_new(ZTX("Monitor Check"),mWin,Bdone,null);                 //  start user dialog
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","lab1","hb1",message,"space=5");
   
   zdialog_resize(zd,200,0);
   zdialog_help(zd,"check_monitor");                                       //  zdialog help topic
   zdialog_run(zd);
   zdialog_wait(zd);                                                       //  wait for dialog complete
   zdialog_free(zd);

restore:

   menulock(0);

   if (savecurrent) {
      f_open(savecurrent,0);                                               //  back to current file
      zfree(savecurrent);
   }

   return;
}


/**************************************************************************/

//  adjust monitor gamma

void m_mongamma(GtkWidget *, cchar *)
{
   int   mongamma_event(zdialog *zd, cchar *event);
   
   int         err;
   char        gammachart[200];
   zdialog     *zd;
   char        *savecurrent = 0;
   
   cchar       *permit = "Chart image courtesy of Norman Koren";
   cchar       *website1 = "http://www.normankoren.com";
   cchar       *website2 = "http://www.imatest.org";

   zfuncs::F1_help_topic = "monitor_gamma";

   err = system("which xgamma");                                           //  check for xgamma                v.12.10.1
   if (err) {
      zmessageACK(mWin,"xgamma command is not installed");
      return;
   }

   if (mod_keep()) return;                                                 //  unsaved edits
   if (! menulock(1)) return;

   if (curr_file) savecurrent = strdupz(curr_file,0,"mongamma");
   
   snprintf(gammachart,200,"%s/images/gammachart.png",get_zdocdir());      //  gamma chart .png file
   err = f_open(gammachart,0);
   if (err) goto restore;

   Fzoom = 1;                                                              //  scale 100% (required)
   gtk_window_set_title(MWIN,"adjust gamma chart");

   zd = zdialog_new(ZTX("Monitor Gamma"),mWin,Bdone,null);                 //  start user dialog
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=8");
   zdialog_add_widget(zd,"label","labgamma","hb1","gamma","space=5");
   zdialog_add_widget(zd,"hscale","gamma","hb1","0.6|1.4|0.02|1.0","expand");
   zdialog_add_widget(zd,"hbox","hb2","dialog");
   zdialog_add_widget(zd,"label","permit","hb2",permit,"space=5");
   zdialog_add_widget(zd,"hbox","hb3","dialog");
   zdialog_add_widget(zd,"link","web1","hb3",website1);
   zdialog_add_widget(zd,"hbox","hb4","dialog");
   zdialog_add_widget(zd,"link","web2","hb4",website2);
   
   zdialog_resize(zd,200,0);
   zdialog_help(zd,"monitor_gamma");                                       //  zdialog help topic
   zdialog_restore_inputs(zd);                                             //  preload prior user inputs    v.12.09
   zdialog_run(zd,mongamma_event);
   zdialog_wait(zd);                                                       //  wait for dialog complete
   zdialog_free(zd);
   
restore:

   menulock(0);

   if (savecurrent) {
      f_open(savecurrent,0);                                               //  back to current file
      zfree(savecurrent);
   }

   return;
}


//  dialog event function

int mongamma_event(zdialog *zd, cchar *event)
{
   int      err;
   double   gamma;

   if (strEqu(event,"gamma")) {
      zdialog_fetch(zd,"gamma",gamma);
      sprintf(command,"xgamma -quiet -gamma %.2f",gamma);
      err = system(command);
      if (err) zmessageACK(mWin,"error: %s",wstrerror(err));
   }
   
   return 0;
}


/**************************************************************************/

//  create or update brightness histogram graph

void m_histogram(GtkWidget *, cchar *menu)                                 //  menu function
{
   if (menu) zfuncs::F1_help_topic = "brightness_dist";

   if (! Fpxb) return;

   if (winhisto) {                                                         //  window already present
      gtk_widget_queue_draw(winhistoH);
      return;
   }

   winhisto = gtk_window_new(GTK_WINDOW_TOPLEVEL);                         //  create histogram window
   gtk_window_set_title(GTK_WINDOW(winhisto),ZTX("Brightness Distribution"));
   gtk_window_set_transient_for(GTK_WINDOW(winhisto),MWIN);
   gtk_window_set_default_size(GTK_WINDOW(winhisto),300,200);
   GtkWidget * vbox = gtk_box_new(VERTICAL,3);
   gtk_container_add(GTK_CONTAINER(winhisto),vbox);

   winhistoH = gtk_drawing_area_new();                                     //  histogram drawing area
   gtk_box_pack_start(GTK_BOX(vbox),winhistoH,1,1,3);

   winhistoB = gtk_drawing_area_new();                                     //  brightness scale under histogram
   gtk_box_pack_start(GTK_BOX(vbox),winhistoB,0,0,0);                      //  (progresses from black to white)
   gtk_widget_set_size_request(winhistoB,300,12);

   G_SIGNAL(winhistoH,"destroy",histogram_destroy,0);
   G_SIGNAL(winhistoH,"draw",histogram_paint,1);
   G_SIGNAL(winhistoB,"draw",histogram_paint,2);

   gtk_widget_show_all(winhisto);
   return;
}


void histogram_paint(GtkWidget *drawin, cairo_t *cr, int which)            //  paint graph window
{
   int         bin, brdist[256][4], maxdist;
   int         px, py, winww, winhh;
   int         rgb, bright;
   float       *pixf;
   float       fbright;
   uint8       *pixi;
   
   if (! winhisto) return;
   if (! Fpxb) return;
   
   if (which == 1)                                                         //  graph window
   {
      for (bin = 0; bin < 256; bin++)                                      //  clear brightness distribution
      for (rgb = 0; rgb < 4; rgb++)
         brdist[bin][rgb] = 0;

      mutex_lock(&Fpixmap_lock);

      if (sa_stat == 3 && E3pxm)                                           //  compute brightness distribution   
      {                                                                    //    for selected area being edited
         for (int ii = 0; ii < Fpxb->ww * Fpxb->hh; ii++)
         {
            if (! sa_pixmap[ii]) continue;
            py = ii / Fpxb->ww;
            px = ii - Fpxb->ww * py;
            pixf = PXMpix(E3pxm,px,py);
            for (rgb = 0; rgb < 3; rgb++) {
               bright = pixf[rgb];                                         //  0 to 255
               ++brdist[bright][rgb];
            }
            bright = 0.333 * (pixf[0] + pixf[1] + pixf[2]);
            ++brdist[bright][3];
         }
      }

      else
      {
         for (py = 0; py < dhh; py++)                                      //  compute brightness distribution
         for (px = 0; px < dww; px++)                                      //    for image in visible window
         {
            pixi = PXBpix(Mpxb,px,py);                                     //  window pixels
            for (rgb = 0; rgb < 3; rgb++) {
               bright = pixi[rgb];                                         //  0 to 255
               ++brdist[bright][rgb];                                      //  distribution per RGB color
            }
            bright = 0.333 * (pixi[0] + pixi[1] + pixi[2]);
            ++brdist[bright][3];
         }
      }
      
      mutex_unlock(&Fpixmap_lock);

      maxdist = 0;

      for (bin = 0; bin < 256; bin++)                                      //  find max. bin over all RGB
      for (rgb = 0; rgb < 3; rgb++)
         if (brdist[bin][rgb] > maxdist) 
            maxdist = brdist[bin][rgb];

      winww = gtk_widget_get_allocated_width(winhistoH);                   //  drawing window size
      winhh = gtk_widget_get_allocated_height(winhistoH);
      
      for (rgb = 0; rgb < 4; rgb++)                                        //  3 colors + white          v.13.01
      {
         if (rgb == 0) cairo_set_source_rgb(cr,1,0,0);
         if (rgb == 1) cairo_set_source_rgb(cr,0,1,0);
         if (rgb == 2) cairo_set_source_rgb(cr,0,0,1);
         if (rgb == 3) cairo_set_source_rgb(cr,0,0,0);

         cairo_move_to(cr,0,winhh);                                        //  start at (0,0)

         for (px = 0; px < winww; px++)                                    //  x from 0 to window width
         {
            bin = 256 * px / winww;                                        //  bin 0-255 for x 0-width
            py = 0.9 * winhh * brdist[bin][rgb] / maxdist;                 //  height of bin in window
            py = winhh * sqrt(1.0 * py / winhh);
            py = winhh - py - 1;
            cairo_line_to(cr,px,py);                                       //  draw line from bin to bin
         }
         
         cairo_stroke(cr);
      }
   }
   
   if (which == 2)                                                         //  scale window
   {
      winww = gtk_widget_get_allocated_width(winhistoB);                   //  drawing window size
      winhh = gtk_widget_get_allocated_height(winhistoB);

      for (px = 0; px < winww; px++)                                       //  draw brightness scale underneath
      {
         fbright = 1.0 * px / winww;
         cairo_set_source_rgb(cr,fbright,fbright,fbright);
         cairo_move_to(cr,px,0);
         cairo_line_to(cr,px,winhh-1);
         cairo_stroke(cr);
      }
   }

   return;
}


void histogram_destroy()                                                   //  delete window
{
   if (winhisto) gtk_widget_destroy(winhisto);
   winhisto = 0;
   return;
}


/**************************************************************************/

//  set GUI language

void  m_lang(GtkWidget *, cchar *)                                         //  overhauled
{
   zdialog     *zd;
   int         ii, cc, err, val, zstat;
   char        progexe[300], lang1[8], *pp;

   cchar  *langs[13] = { "en English",                                     //  english first
                         "ca Catalan", "de German", "es Spanish",          //  v.12.10.1
                         "fr French", "gl Galacian", "it Italian", 
                         "nl Dutch", "pt Portuguese", "ru_RU Russian", 
                         "sv Swedish", "zh_CN Chinese", null  };
   
   cchar  *title = ZTX("Available Translations");

   zfuncs::F1_help_topic = "language";

   zd = zdialog_new(ZTX("Set Language"),mWin,Bdone,Bcancel,null);
   zdialog_add_widget(zd,"label","title","dialog",title,"space=10");
   zdialog_add_widget(zd,"hbox","hb1","dialog");
   zdialog_add_widget(zd,"vbox","vb1","hb1");

   for (ii = 0; langs[ii]; ii++)                                           //  make radio button per language
      zdialog_add_widget(zd,"radio",langs[ii],"vb1",langs[ii]);

   cc = strlen(zfuncs::zlang);                                             //  current language
   for (ii = 0; langs[ii]; ii++)                                           //  match on lc_RC
      if (strnEqu(zfuncs::zlang,langs[ii],cc)) break;
   if (! langs[ii]) 
      for (ii = 0; langs[ii]; ii++)                                        //  failed, match on lc alone
         if (strnEqu(zfuncs::zlang,langs[ii],2)) break;
   if (! langs[ii]) ii = 0;                                                //  failed, default english
   zdialog_stuff(zd,langs[ii],1);

   zdialog_resize(zd,200,0);
   zdialog_help(zd,"language");                                            //  zdialog help topic
   zdialog_run(zd);                                                        //  run dialog
   zstat = zdialog_wait(zd);

   for (ii = 0; langs[ii]; ii++) {                                         //  get active radio button
      zdialog_fetch(zd,langs[ii],val);
      if (val) break;
   }

   zdialog_free(zd);                                                       //  kill dialog

   if (zstat != 1) return;                                                 //  user cancel
   if (! val) return;                                                      //  no selection
   
   strncpy0(lang1,langs[ii],8);
   pp = strchr(lang1,' ');                                                 //  isolate lc_RC part
   *pp = 0;

   cc = readlink("/proc/self/exe",progexe,300);                            //  get own program path      v.13.01
   if (cc <= 0) {
      zmessageACK(mWin,"cannot get /proc/self/exe");
      return;
   }
   progexe[cc] = 0;

   sprintf(command,"%s -l %s -p &",progexe,lang1);                         //  start new fotoxx with language
   err = system(command);
   if (err) printf("*** %s \n",wstrerror(err));

   m_quit(0,0);                                                            //  exit
   return;
}


/**************************************************************************/

//  edit translations while running the application interactively

void  m_translate(GtkWidget *, cchar *)
{
   zfuncs::F1_help_topic = "translate";
   ZTX_translation_start(mWin);
   return;
}


/**************************************************************************/

//  create desktop icon / launcher

void  m_menu_launcher(GtkWidget *, cchar *)
{
   char  *command;
   
   zfuncs::F1_help_topic = "menu_launcher";
   command = zdialog_text(mWin,ZTX("Make Launcher"),"fotoxx -recent");
   if (! command || ! *command) return;
   zmake_menu_launcher(command,"Graphics","Image Editor");
   zfree(command);                                                         //  v.13.01
   return;
}


/**************************************************************************/

//  user settings dialog

void m_settings(GtkWidget *, cchar *)                                      //  v.12.01
{
   int   settings_dialog_event(zdialog *zd, cchar *event);

   zdialog  *zd;

/***
         Startup Display   
          (o) Recent Files Gallery
          (o) Previous Image Viewed
          (o) Blank Window
          (o) Directory [____________________________] [browse]
          (o) Image File [___________________________] [browse]

         Menu Style      (o) Classic  (o) Vertical
         Toolbar Style   (o) Text  (o) Icons  (o) Both
         Image Pan       (o) Drag Mouse  (o) Magnified Scroll
         Warn Overwrite  [x] 
         Zooms for 2x    [__|v]
         RAW command     [____________________________________]
         RAW file types  [____________________________________]

                                                         [done]
***/

   zfuncs::F1_help_topic = "user_settings";
   
   zd = zdialog_new(ZTX("Settings"),mWin,Bdone,null);
   
   zdialog_add_widget(zd,"hbox","hbstart","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labstart","hbstart",ZTX("Startup Display"));
   
   zdialog_add_widget(zd,"hbox","hbdisp","dialog");
   zdialog_add_widget(zd,"label","space","hbdisp",0,"space=8");
   zdialog_add_widget(zd,"vbox","vbdisp1","hbdisp",0,"homog");
   zdialog_add_widget(zd,"label","space","hbdisp",0,"space=5");
   zdialog_add_widget(zd,"vbox","vbdisp2","hbdisp",0,"homog|expand");
   
   zdialog_add_widget(zd,"radio","recent","vbdisp1");
   zdialog_add_widget(zd,"radio","prev","vbdisp1");
   zdialog_add_widget(zd,"radio","blank","vbdisp1");
   zdialog_add_widget(zd,"radio","dirk","vbdisp1");
   zdialog_add_widget(zd,"radio","file","vbdisp1");
   
   zdialog_add_widget(zd,"hbox","hbrecent","vbdisp2");
   zdialog_add_widget(zd,"hbox","hbprev","vbdisp2");
   zdialog_add_widget(zd,"hbox","hbblank","vbdisp2");
   zdialog_add_widget(zd,"hbox","hbdirk","vbdisp2");
   zdialog_add_widget(zd,"hbox","hbfile","vbdisp2");

   zdialog_add_widget(zd,"label","labrecent","hbrecent",ZTX("Recent Files Gallery"));
   zdialog_add_widget(zd,"label","labprev","hbprev",ZTX("Previous Image Viewed"));
   zdialog_add_widget(zd,"label","labblank","hbblank",ZTX("Blank Window"));
   zdialog_add_widget(zd,"label","labdirk","hbdirk",ZTX("Directory Gallery"));
   zdialog_add_widget(zd,"label","labfile","hbfile",ZTX("Image File"));

   zdialog_add_widget(zd,"entry","stdirk","hbdirk",0,"expand|space=5");
   zdialog_add_widget(zd,"button","stdbrowse","hbdirk",Bbrowse);

   zdialog_add_widget(zd,"entry","stfile","hbfile",0,"expand|space=5");
   zdialog_add_widget(zd,"button","stfbrowse","hbfile",Bbrowse);
   
   zdialog_add_widget(zd,"hsep","sep1","dialog",0,"space=3");
   
   zdialog_add_widget(zd,"hbox","hbopts","dialog");
   zdialog_add_widget(zd,"label","space","hbopts",0,"space=5");
   zdialog_add_widget(zd,"vbox","vbopts1","hbopts",0,"homog");
   zdialog_add_widget(zd,"label","space","hbopts",0,"space=10");
   zdialog_add_widget(zd,"vbox","vbopts2","hbopts",0,"homog");
   
   zdialog_add_widget(zd,"label","labms","vbopts1",ZTX("Menu Style"));
   zdialog_add_widget(zd,"hbox","hbms","vbopts2",0,"space=2");
   zdialog_add_widget(zd,"radio","classic","hbms",ZTX("Classic"),"space=5");
   zdialog_add_widget(zd,"radio","vertical","hbms",ZTX("Vertical"),"space=5");

   zdialog_add_widget(zd,"label","labtbs","vbopts1",ZTX("Toolbar Style"));
   zdialog_add_widget(zd,"hbox","hbtbs","vbopts2",0,"space=2");
   zdialog_add_widget(zd,"radio","text","hbtbs",ZTX("Text"),"space=5");
   zdialog_add_widget(zd,"radio","icons","hbtbs",ZTX("Icons"),"space=5");
   zdialog_add_widget(zd,"radio","both","hbtbs",ZTX("Both"),"space=5");

   zdialog_add_widget(zd,"label","labtbs","vbopts1",ZTX("Image Pan"));
   zdialog_add_widget(zd,"hbox","hbpan","vbopts2",0,"space=2");
   zdialog_add_widget(zd,"radio","m-drag","hbpan",ZTX("Drag Mouse"),"space=5");
   zdialog_add_widget(zd,"radio","m-scroll","hbpan",ZTX("Magnified Scroll"),"space=5");

   zdialog_add_widget(zd,"hbox","hbwarn","dialog",0,"space=2");
   zdialog_add_widget(zd,"label","labwarn","hbwarn",ZTX("Warn Overwrite Original"),"space=5");
   zdialog_add_widget(zd,"check","warn","hbwarn",0);
   
   zdialog_add_widget(zd,"hbox","hbzoom","dialog",0,"space=2");
   zdialog_add_widget(zd,"label","labzoom","hbzoom","Zooms for 200%","space=5");
   zdialog_add_widget(zd,"spin","zoom","hbzoom","1|4|1|2");
   
   zdialog_add_widget(zd,"hbox","hbraw","dialog",0,"space=2");
   zdialog_add_widget(zd,"label","rawlab1","hbraw","RAW command","space=3");
   zdialog_add_widget(zd,"entry","rawcomm","hbraw","dcraw -w -q 0 -T -6","space=3|expand");
   zdialog_add_widget(zd,"hbox","hbraw2","dialog");
   zdialog_add_widget(zd,"label","rawlab2","hbraw2","RAW file types","space=3");
   zdialog_add_widget(zd,"entry","rawtypes","hbraw2",".raw .rw2","space=3|expand");
   
   zdialog_stuff(zd,"stdirk",startdirk);                                   //  stuff present settings into dialog
   zdialog_stuff(zd,"stfile",startfile);

   zdialog_stuff(zd,"recent",0);
   zdialog_stuff(zd,"prev",0);
   zdialog_stuff(zd,"blank",0);
   zdialog_stuff(zd,"dirk",0);
   zdialog_stuff(zd,"file",0);
   zdialog_stuff(zd,startdisplay,1);
   
   zdialog_stuff(zd,"classic",0);                                          //  v.13.01
   zdialog_stuff(zd,"vertical",0);
   zdialog_stuff(zd,menu_style,1); 
   
   zdialog_stuff(zd,"icons",0);
   zdialog_stuff(zd,"text",0);
   zdialog_stuff(zd,"both",0);
   zdialog_stuff(zd,tbar_style,1);
   
   zdialog_stuff(zd,"m-drag",2-Fdragopt);                                  //  Fdragopt = 1/2 = drag/scroll
   zdialog_stuff(zd,"m-scroll",Fdragopt-1);

   zdialog_stuff(zd,"warn",Fwarnoverwrite);
   
   if (zoomcount >= 1 && zoomcount <= 4)   
      zdialog_stuff(zd,"zoom",zoomcount);
   
   zdialog_stuff(zd,"rawcomm",RAWcommand);
   zdialog_stuff(zd,"rawtypes",RAWfiletypes);
   
   zdialog_help(zd,"user_settings");
   zdialog_resize(zd,500,0);

   zdialog_run(zd,settings_dialog_event);                                  //  run dialog and wait for completion
   zdialog_wait(zd);
   zdialog_free(zd);

   return;
}


//  settings dialog event function

int settings_dialog_event(zdialog *zd, cchar *event)
{
   int         nn;
   char        *pp, temp[200];

   if (zd->zstat && zd->zstat != 1) return 0;                              //  cancel
   
   if (zd->zstat == 1)                                                     //  done, check inputs are valid
   {
      zdialog_fetch(zd,"recent",nn);                                       //  startup display options
      if (nn) {
         zfree(startdisplay);
         startdisplay = strdupz("recent",0,"settings");                    //  recent files gallery
      }
      
      zdialog_fetch(zd,"prev",nn);
      if (nn) {
         zfree(startdisplay);
         startdisplay = strdupz("prev",0,"settings");                      //  start with file from last session
      }
      
      zdialog_fetch(zd,"blank",nn);
      if (nn) {
         zfree(startdisplay);
         startdisplay = strdupz("blank",0,"settings");                     //  start with blank window
      }
      
      zdialog_fetch(zd,"dirk",nn);
      if (nn) {
         zfree(startdisplay);
         startdisplay = strdupz("dirk",0,"settings");                      //  start with directory gallery
      }
      
      zdialog_fetch(zd,"file",nn);
      if (nn) {
         zfree(startdisplay);
         startdisplay = strdupz("file",0,"settings");                      //  start with specified file
      }

      zdialog_fetch(zd,"stdirk",temp,200);                                 //  startup directory gallery
      if (startdirk) zfree(startdirk);
      startdirk = strdupz(temp,0,"settings");

      zdialog_fetch(zd,"stfile",temp,200);                                 //  startup image file
      if (startfile) zfree(startfile);
      startfile = strdupz(temp,0,"settings");

      if (strEqu(startdisplay,"dirk")) {                                   //  startup option = directory gallery
         if (image_file_type(startdirk) != 1) {
            zmessageACK(mWin,ZTX("startup directory is invalid"));
            zd->zstat = 0;
            return 0;
         }
      }

      if (strEqu(startdisplay,"file")) {                                   //  startup option = image file
         if (image_file_type(startfile) != 2) {
            zmessageACK(mWin,ZTX("startup file is invalid"));
            zd->zstat = 0;
            return 0;
         }
      }
      
      zdialog_fetch(zd,"classic",nn);                                      //  menu style       v.13.01
      if (nn) {
         zfree(menu_style);
         menu_style = strdupz("classic",0,"settings");
      }
      
      zdialog_fetch(zd,"vertical",nn);
      if (nn) {
         zfree(menu_style);
         menu_style = strdupz("vertical",0,"settings");
      }
      
      zdialog_fetch(zd,"text",nn);                                         //  toolbar style    v.12.07
      if (nn) {
         zfree(tbar_style);
         tbar_style = strdupz("text",0,"settings");
      }

      zdialog_fetch(zd,"icons",nn);
      if (nn) {
         zfree(tbar_style);
         tbar_style = strdupz("icons",0,"settings");
      }

      zdialog_fetch(zd,"both",nn);
      if (nn) {
         zfree(tbar_style);
         tbar_style = strdupz("both",0,"settings");
      }
      
      zdialog_fetch(zd,"m-drag",nn);                                       //  Fdragopt = 1/2 = drag/scroll
      Fdragopt = 2 - nn;

      zdialog_fetch(zd,"warn",Fwarnoverwrite);                             //  warn overwrite option
      
      zdialog_fetch(zd,"zoom",zoomcount);                                  //  zooms for 2x image, 1-4      v.12.03
      zoomratio = pow( 2.0, 1.0 / zoomcount);                              //  2.0, 1.4142, 1.2599, 1.1892
      
      zdialog_fetch(zd,"rawcomm",temp,200);                                //  command line for RAW to tiff    v.12.09
      if (RAWcommand) zfree(RAWcommand);
      RAWcommand = strdupz(temp,0,"settings");
      
      zdialog_fetch(zd,"rawtypes",temp,200);                               //  RAW file types, .raw .rw2 ...   v.12.09
      if (RAWfiletypes) zfree(RAWfiletypes);
      RAWfiletypes = strdupz(temp,0,"settings");

      save_params();
      return 0;
   }

   if (strEqu(event,"stdbrowse")) {                                        //  startup directory browse
      zdialog_fetch(zd,"stdirk",temp,200);
      pp = zgetfile1(ZTX("Select startup directory"),"folder",temp);  
      if (! pp) return 0;
      zdialog_stuff(zd,"stdirk",pp);
      zfree(pp);
   }
   
   if (strEqu(event,"stfbrowse")) {                                        //  startup file browse
      zdialog_fetch(zd,"stfile",temp,200);
      pp = zgetfile1(ZTX("Select startup image file"),"open",temp);  
      if (! pp) return 0;
      zdialog_stuff(zd,"stfile",pp);
      zfree(pp);
   }

   return 0;
}


/**************************************************************************/

//  Dump CPU and memory usage by category

void m_resources(GtkWidget *, cchar *)
{
   static double  ptime = 0;
   zfuncs::F1_help_topic = "resources";
   
   ptime = jobtime() - ptime;
   printf("\n process time: %.2f \n",ptime);
   zmalloc_report();
   ptime = jobtime();
   return;
}




