/**************************************************************************

   Fotoxx      edit photos and manage collections

   Copyright 2007 2008 2009 2010 2011 2012 Michael Cornelison
   Source URL: http://kornelix.com/fotoxx
   Contact: kornelix2@gmail.com
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see http://www.gnu.org/licenses/.

***************************************************************************/

#define EX                                                                 //  disable extern declarations
#include "fotoxx.h"


const char *help_about = 
"Fotoxx v.13.01 \n"                                                        //  v.13.01
"Free software - GNU General Public License v.3 \n"
"http://kornelix.com/fotoxx \n"
"Bug reports: kornelix2@gmail.com \n"
"Software used: GNU, GTK, libtiff, libpng, dcraw, ufraw, exiftool \n"
"Translators: \n"
"   Victor Ryzhikh (Виктор Рыжих), Miguel Anxo Bouzada, \n"
"   Doriano Blengino, Arthur Kalverboer, André Campos Rodovalho, \n"
"   Jie Luo, Peter Landgren, Josep Antoni Miralles Puignau, \n"
"   Jean-Pierre Vaissière ";


//  fotoxx main program

int main(int argc, char *argv[])
{
   char           lang[8] = "";
   int            Fclone=0, cloxx=0, cloyy=0, cloww=0, clohh=0;
   int            Ftranslate = 0;
   int            err;
   struct stat    statb;
   
   if (argc > 1 && strEqu(argv[1],"-v")) {
      printf(fotoxxvers "\n");                                             //  print version and exit
      return 0;
   }

   gtk_init(&argc,&argv);                                                  //  initz. GTK & strip GTK params
   zinitapp("fotoxx",null);                                                //  initz. app directories

   //  initialize externals to default values 
   //  (saved parameters will override some of these)

   Ffirsttime = 1;                                                         //  first time startup
   for (int ii = 0; ii < 8; ii++) wtnx[ii] = ii;                           //  thread numbers 0-7
   mwgeom[0] = mwgeom[1] = 100;                                            //  default main window geometry
   mwgeom[2] = 800; mwgeom[3] = 600;
   Fzoom = 0;
   wwhh_limit = 20000;                                                     //  limits for editing              v.12.10
   trimsize[0] = 1600;
   trimsize[1] = 1000;
   editresize[0] = 1600;
   editresize[1] = 1200;
   gridcount[0] = gridcount[1] = 5;                                        //  default grid line counts
   menu_style = strdupz("vertical",0,"initz");                             //  default menu style              v.13.01
   tbar_style = strdupz("both",0,"initz");                                 //  default toolbar style
   startdisplay = strdupz("blank",0,"initz");                              //  start with blank window
   RAWcommand = strdupz("dcraw -T -6 -w -q 0 -o 1",0,"initz");             //  default RAW to tiff command
   RAWfiletypes = strdupz(".arw .crw .cr2 .dng .erf .iiq .mef .mos "       //  some of the known RAW file types
                          ".mpo .nef .nrw .orf .orf .pef .ptx .raw "       //  (case is not significant here)
                          ".rw2 .rwl .srf .srw .sr2 ",0,"initz");
   Fdragopt = 2;                                                           //  image drag is magnified scroll
   Fwarnoverwrite = 1;                                                     //  warn original file overwrite
   zoomcount = 2;                                                          //  zooms to reach 2x image size    v.12.03
   zoomratio = sqrtf(2);                                                   //  corresp. zoom ratio 
   Fthumbnails = 1;                                                        //  default, use thumbnails with image files
   thumbdirk = strdupz("undefined",0,"initz");                             //  default, no thumbnail directory
   FsyncEnabled = 1;                                                       //  default, file sync enabled
   last_sync_time = strdupz("Jan 01 00:00:00 2000",0,"initz");             //  default, long ago
   lens_mm = 35;                                                           //  pano lens parameters
   lens_bow = 0;
   strcpy(jpeg_quality,def_jpeg_quality);                                  //  default .jpeg quality

   black[0] = black[1] = black[2] = 0;                                     //  define RGB colors
   white[0] = white[1] = white[2] = 255;
   red[0] = 255; red[1] = red[2] = 0;
   green[1] = 255; green[0] = green[2] = 0;
   fg_color = green;                                                       //  initial foreground drawing color

   GDKdgray.red = GDKdgray.green = GDKdgray.blue = 0.4;                    //  GDK colors for dark gray
   GDKblack.red = GDKblack.green = GDKblack.blue = 0;                      //  GDK colors for black
   GDKdgray.alpha = 1;
   GDKblack.alpha = 1;

   arrowcursor = gdk_cursor_new(GDK_TOP_LEFT_ARROW);                       //  cursor for selection
   dragcursor = gdk_cursor_new(GDK_CROSSHAIR);                             //  cursor for dragging
   drawcursor = gdk_cursor_new(GDK_PENCIL);                                //  cursor for drawing lines

   //  process command line parameters
   
   for (int ii = 1; ii < argc; ii++)                                       //  command line parameters
   {
      char *pp = argv[ii];
      if (strcmpv(pp,"-debug","-d",null))                                  //  -d -debug
         Fdebug = 1;
      else if (strcmpv(pp,"-lang","-l",null) && argc > ii+1)               //  -l -lang lc_RC   language/region code
         strncpy0(lang,argv[++ii],7);
      else if (strcmpv(pp,"-clone","-c",null) && argc > ii+4) {            //  -c -clone        clone new instance
         Fclone = 1;
         cloxx = atoi(argv[ii+1]);                                         //  window position and size
         cloyy = atoi(argv[ii+2]);                                         //    passed from parent instance
         cloww = atoi(argv[ii+3]);
         clohh = atoi(argv[ii+4]);
         ii += 4;
      }
      else if (strcmpv(pp,"-slideshow","-ss",null) && argc > ii+1)         //  -ss -slideshow /image/file.jpg
         SS_imagefile = strdupz(argv[++ii],0,"initz");
      else if (strcmpv(pp,"-music","-m",null) && argc > ii+1)              //  -m -music /music/file.ogg
         SS_musicfile = strdupz(argv[++ii],0,"initz");
      else if (strcmpv(pp,"-recent","-r",null))                            //  -r -recent       recent files gallery
         Frecent = 1;
      else if (strcmpv(pp,"-previous","-prev","-p",null))                  //  -p -prev         open previous file
         Fprev = 1;
      else if (strcmpv(pp,"-blank","-b",null))                             //  -b -blank        start with blank window
         Fblank = 1;
      else if (strcmpv(pp,"-translate","-t",null))                         //  -t -translate    start translation
         Ftranslate = 1;
      else if (strcmpv(pp,"-nosync","-nos",null))                          //  -nos -nosync     omit auto sync  v.12.08
         Fnosync = 1;
      else if (strEqu(pp,"-menu") && argc > ii+1)                          //  -menu menufunc   starting menu   v.12.09
         startmenu = strdupz(argv[++ii],0,"initz");
      else  initial_file = strdupz(pp,0,"initz");                          //  must be initial file or directory
   }

   ZTXinit(lang);                                                          //  setup locale and translations
   setlocale(LC_NUMERIC,"en_US.UTF-8");                                    //  stop comma decimal points     v.12.12.2

   //  get files in user directory /home/<user>/.fotoxx/*

   snprintf(exiftool_input,199,"%s/exiftool_input",get_zuserdir());        //  exiftool_server input file
   snprintf(topdirk_file,199,"%s/top_directory",get_zuserdir());           //  top_directories file
   snprintf(search_index_dirk,199,"%s/search_index3",get_zuserdir());      //  search_index directory
   snprintf(tags_defined_file,199,"%s/tags_defined",get_zuserdir());       //  defined tags file
   snprintf(recentfiles_file,199,"%s/recent_files",get_zuserdir());        //  recent files file
   snprintf(saved_areas_dirk,199,"%s/saved_areas",get_zuserdir());         //  saved areas directory
   snprintf(collections_dirk,199,"%s/collections",get_zuserdir());         //  collections directory
   snprintf(saved_curves_dirk,199,"%s/saved_curves",get_zuserdir());       //  saved curves directory
   snprintf(writetext_dirk,199,"%s/write_text",get_zuserdir());            //  write_text directory
   snprintf(geotags_dirk,199,"%s/geotags",get_zuserdir());                 //  geotags directory
   snprintf(favorites_dirk,199,"%s/favorites",get_zuserdir());             //  favorites directory             v.12.10

   err = stat(saved_areas_dirk,&statb);                                    //  create any missing directories
   if (err) mkdir(saved_areas_dirk,0750);
   err = stat(collections_dirk,&statb);
   if (err) mkdir(collections_dirk,0750);
   err = stat(saved_curves_dirk,&statb);
   if (err) mkdir(saved_curves_dirk,0750);
   err = stat(writetext_dirk,&statb);
   if (err) mkdir(writetext_dirk,0750);
   err = stat(geotags_dirk,&statb);
   if (err) mkdir(geotags_dirk,0750);
   err = stat(favorites_dirk,&statb);
   if (err) mkdir(favorites_dirk,0750);

   load_params();                                                          //  restore parameters from last session
   
   //  create main window with image file tab and gallery tab              //  v.12.07
   
   mWin = gtk_window_new(GTK_WINDOW_TOPLEVEL);                             //  create main window
   gtk_window_set_title(MWIN,fotoxxvers);

   tabWin = gtk_notebook_new();                                            //  add tab window to main window 
   gtk_notebook_set_tab_pos(GTK_NOTEBOOK(tabWin),GTK_POS_LEFT);
   gtk_container_add(GTK_CONTAINER(mWin),tabWin);

   PangoFontDescription *tabfont;                                          //  v.13.01
   tabfont = pango_font_description_from_string("mono bold 12");

   fVbox = gtk_box_new(VERTICAL,0);                                        //  add vbox as tab F (image file window)
   GtkWidget * labF = gtk_label_new("F");
   gtk_widget_override_font(labF,tabfont);
   gtk_notebook_append_page(GTK_NOTEBOOK(tabWin),fVbox,labF);

   gVbox = gtk_box_new(VERTICAL,0);                                        //  add vbox as tab G (gallery window)
   GtkWidget * labG = gtk_label_new("G");
   gtk_widget_override_font(labG,tabfont);
   gtk_notebook_append_page(GTK_NOTEBOOK(tabWin),gVbox,labG);

   if (Ftranslate) {                                                       //  start online translation
      zfuncs::F1_help_topic = "translate";                                 //  (to process the menu names)
      ZTX_translation_start(mWin);
   }

   build_menus();                                                          //  build menus in F and G windows   v.13.01

   GtkWidget * wVbox = gtk_box_new(VERTICAL,0);                            //  create tab W (world map window)
   GtkWidget * labW = gtk_label_new("W");
   gtk_widget_override_font(labW,tabfont);
   gtk_notebook_append_page(GTK_NOTEBOOK(tabWin),wVbox,labW);
   wDwin = gtk_drawing_area_new();                                         //  add drawing window
   gtk_box_pack_start(GTK_BOX(wVbox),wDwin,1,1,0);

   G_SIGNAL(mWin,"delete_event",delete_event,0);                           //  connect signals to main window
   G_SIGNAL(mWin,"destroy",destroy_event,0);
   G_SIGNAL(mWin,"window-state-event",state_event,0);
   G_SIGNAL(mWin,"key-press-event",KBpress,0);                             //  connect KB events to main window
   G_SIGNAL(mWin,"key-release-event",KBrelease,0);

   G_SIGNAL(tabWin,"switch-page",tabchanged,0);                            //  connect tab switched event to tab window
   
   gtk_widget_add_events(fDwin,GDK_BUTTON_PRESS_MASK);                     //  connect mouse events to image window
   gtk_widget_add_events(fDwin,GDK_BUTTON_RELEASE_MASK);
   gtk_widget_add_events(fDwin,GDK_BUTTON_MOTION_MASK);                    //  motion with button pressed
   gtk_widget_add_events(fDwin,GDK_POINTER_MOTION_MASK);                   //  pointer motion
   gtk_widget_add_events(fDwin,GDK_SCROLL_MASK);                           //  scroll wheel                 v.12.01
   G_SIGNAL(fDwin,"button-press-event",mouse_event,0);
   G_SIGNAL(fDwin,"button-release-event",mouse_event,0);
   G_SIGNAL(fDwin,"motion-notify-event",mouse_event,0);
   G_SIGNAL(fDwin,"scroll-event",mouse_event,0);
   G_SIGNAL(fDwin,"draw",mwpaint1,0);                                      //  connect draw signal

   gtk_widget_add_events(gDwin,GDK_BUTTON_PRESS_MASK);                     //  connect mouse events to gallery window
   gtk_widget_add_events(gDwin,GDK_POINTER_MOTION_MASK);                   //  v.13.01
   gtk_widget_add_events(gDwin,GDK_LEAVE_NOTIFY_MASK);
   G_SIGNAL(gDwin,"button-press-event",navi::mouse_event,0);
   G_SIGNAL(gDwin,"motion-notify-event",navi::mouse_event,0);
   G_SIGNAL(gDwin,"leave-notify-event",navi::mouse_event,0);
   G_SIGNAL(gDwin,"draw",navi::gallery_paint,0);                           //  connect draw signal

   gtk_widget_add_events(gPath,GDK_BUTTON_PRESS_MASK);                     //  connect mouse to gallery 
   G_SIGNAL(gPath,"button-press-event",navi::changedirk,0);                //    navigation widgets
   G_SIGNAL(gHome,"clicked",navi::rootdirk,0);

   gtk_widget_add_events(wDwin,GDK_BUTTON_PRESS_MASK);                     //  connect mouse events to world map window
   gtk_widget_add_events(wDwin,GDK_BUTTON_RELEASE_MASK);
   gtk_widget_add_events(wDwin,GDK_BUTTON_MOTION_MASK);                    //  motion with button pressed
   gtk_widget_add_events(wDwin,GDK_POINTER_MOTION_MASK);                   //  pointer motion
   gtk_widget_add_events(wDwin,GDK_SCROLL_MASK);                           //  scroll wheel
   G_SIGNAL(wDwin,"button-press-event",mouse_event,0);
   G_SIGNAL(wDwin,"button-release-event",mouse_event,0);
   G_SIGNAL(wDwin,"motion-notify-event",mouse_event,0);
   G_SIGNAL(wDwin,"scroll-event",mouse_event,0);
   G_SIGNAL(wDwin,"draw",mwpaint1,0);

   drag_drop_connect(mWin,m_open_drag);                                    //  connect drag-drop to main window

   // ---------------------------------------------------------------------

   if (Fclone) {                                                           //  clone: open new window 
      gtk_window_move(MWIN,cloxx+20,cloyy+30);                             //    slightly offset from old window
      gtk_window_resize(MWIN,cloww-20,clohh-30);                           //  v.12.07
   }
   else {
      gtk_window_move(MWIN,mwgeom[0],mwgeom[1]);                           //  main window geometry
      gtk_window_set_default_size(MWIN,mwgeom[2],mwgeom[3]);               //  defaults or last session params
   }

   gtk_widget_show_all(mWin);                                              //  show all widgets
   gtk_widget_realize(fDwin);                                              //  compensate GTK bug
   gtk_widget_realize(wDwin);                                              //  compensate GTK bug
   gtk_widget_override_background_color(fDwin,(GtkStateFlags)0,&GDKdgray); //  gdk_window_set_background_rgba() fails

   cTab = tabF;                                                            //  set current tab = F
   cDwin = fDwin;                                                          //  curr. drawing window
   cSbar = fSbar;                                                          //  status bar active
   gdkWin = gtk_widget_get_window(cDwin);

   g_timeout_add(100,initzfunc,0);                                         //  initz. call from gtk_main()  v.13.01

   gtk_main();                                                             //  start processing window events
   return 0;
}


/**************************************************************************/

//  Initial function called from gtk_main() at startup.
//  This function MUST return 0.

int initzfunc(void *)
{
   int            Fexiftool = 0, Fxdgopen = 0, Fdcraw = 0, Fufraw = 0;
   int            ii, err, flag, npid;
   int            ftype, Fdirk = 0;
   char           procfile[20], buff[100];
   char           badnews[200];
   char           *pp1, *pp2;
   double         freememory, cachememory;                                 //  available memory at startup
   float          exifver = 0;
   FILE           *fid;
   struct stat    statb;

   cchar *ftmess = ZTX("First time Fotoxx startup. \n"
                       "Latest version: http://www.kornelix.com/fotoxx \n"
                       "File sync process will begin (may need some time) \n");

   menulock(1);                                                            //  block menus until initz. done

   printf(fotoxxvers "\n");                                                //  print Fotoxx version
   printf("last file sync: %s \n",last_sync_time);                         //  last file sync time       v.12.08

   snprintf(PIDstring,11,"%06d",getpid());                                 //  get process PID 6-digits

   fid = popen("exiftool -ver","r");                                       //  check exiftool version    v.12.06
   if (fid) {
      ii = fscanf(fid,"%s",buff);
      pclose(fid);                                                         //  bugfix: accept period or comma
      convSF(buff,exifver);                                                //             v.12.06.1
      printf("exiftool version: %.2f \n",exifver);
      if (exifver >= 8.60) Fexiftool = 1;
   }

   err = system("which xdg-open");                                         //  check for xdg-open
   if (! err) Fxdgopen = 1;
   err = system("which dcraw");                                            //  check for dcraw                 v.12.09
   if (! err) Fdcraw = 1;
   err = system("which ufraw");                                            //  check for ufraw                 v.12.09
   if (! err) Fufraw = 1;

   if (Fexiftool + Fxdgopen + Fdcraw < 3) {                                //  check mandatory dependencies    v.12.07
      strcpy(badnews,ZTX("Please install missing programs:"));
      if (! Fexiftool) strcat(badnews,"\n exiftool 8.6 or later");         //  refuse to start if any are missing
      if (! Fxdgopen) strcat(badnews,"\n xdg-utils");
      if (! Fdcraw) strcat(badnews,"\n dcraw");
      if (! Fufraw) strcat(badnews,"\n ufraw");
      zmessageACK(mWin,badnews);
      gtk_main_quit();
      exit(0);
   }

   parseprocfile("/proc/meminfo","MemFree:",&freememory,null);             //  get amount of free memory       v.12.04
   parseprocfile("/proc/meminfo","Cached:",&cachememory,null);
   freememory = (freememory + cachememory) / 1024;                         //  megabytes
   printf("free memory: %.0f MB \n",freememory);
   max_view_megapix = (freememory - 200) / 6;                              //  F + preview images, 3 bytes/pixel each  
   max_edit_megapix = (freememory - 200) / 54;                             //  + E0/E1/E3/ER images, 12 bytes/pixel each
   printf("max. view image: %d megapixels \n",max_view_megapix);           //  v.12.10
   printf("max. edit image: %d megapixels \n",max_edit_megapix);

   Nwt = sysconf(_SC_NPROCESSORS_ONLN);                                    //  get SMP CPU count
   if (Nwt <= 0) Nwt = 1;
   if (Nwt > max_threads) Nwt = max_threads;                               //  compile time limit
   printf("using %d threads \n",Nwt);

   undo_files = zmalloc(200,"undo_files");
   *undo_files = 0;                                                        //  look for orphaned undo files
   strncatv(undo_files,199,get_zuserdir(),"/*_undo_*",null);               //  /.../.fotoxx/*_undo_*
   flag = 1;
   while ((pp1 = (char *) SearchWild(undo_files,flag)))
   {
      pp2 = strstr(pp1,".fotoxx/");
      if (! pp2) continue;
      npid = atoi(pp2+8);                                                  //  pid of file owner
      snprintf(procfile,19,"/proc/%d",npid);
      err = stat(procfile,&statb);
      if (! err) continue;                                                 //  pid is active, keep file
      printf("*** orphaned undo file deleted: %s \n",pp1);
      err = remove(pp1);                                                   //  delete file
   }
   
   *undo_files = 0;                                                        //  setup undo filespec template
   strncatv(undo_files,199,get_zuserdir(),"/",PIDstring,"_undo_nn",null);  //  /.../.fotoxx/pppppp_undo_nn

   while (Ffirsttime) {                                                    //  first time startup
      printf("%s",ftmess);
      ii = zdialog_choose(0,mWin,ftmess,Bhelp,Bproceed,null);
      if (ii == 1) {
         showz_userguide("sync_files");                                    //  show sync files help
         continue;
      }
      if (ii == 2) Ffirsttime = 0;                                         //  proceed
      save_params();
   }

   snprintf(KBshortcuts,199,"%s/KB-shortcuts",get_zuserdir());             //  copy default KB shortcuts file
   err = stat(KBshortcuts,&statb);                                         //    if not already there          v.12.06
   if (err) {
      sprintf(command,"cp %s/KB-shortcuts %s",get_zdocdir(),KBshortcuts);
      err = system(command);
      if (err) printf("*** %s \n",wstrerror(err));
   }
   
   ii = 0;
   fid = fopen(KBshortcuts,"r");                                           //  build table of shortcut keys
   if (fid) {                                                              //    and corresp. menu names       v.12.06
      while (true) {
         pp1 = fgets_trim(buff,100,fid,1);
         if (! pp1) break;
         if (*pp1 == '#') continue;
         while (*pp1 && *pp1 == ' ') pp1++;
         if (! *pp1) continue;
         pp2 = strchr(pp1,' ');
         if (! pp2) continue;
         if (pp2 - pp1 > 20) continue;
         *pp2 = 0;
         shortcutkey[ii] = strdupz(pp1,0,"initz");
         pp1 = pp2 + 1;
         while (*pp1 && *pp1 == ' ') pp1++;
         if (! *pp1) continue;
         shortcutmenu[ii] = strdupz(ZTX(pp1),0,"initz");                   //  menu name can be English or translation
         if (++ii == maxshortcuts) break;
      }
   }
   Nshortcuts = ii;

   mutex_init(&Fpixmap_lock,0);                                            //  setup lock for edit pixmaps
   mutex_init(&preload_lock,0);                                            //  setup lock for preload function

   zdialog_inputs("load");                                                 //  load saved dialog inputs         v.12.09
   zdialog_positions("load");                                              //  load saved dialog positions
   exiftool_server(null);                                                  //  kill orphan exiftool process     v.12.03

   g_timeout_add(100,gtimefunc,0);                                         //  start periodic function (100 ms)

   syncfiles_func(1);                                                      //  process newly added image files  v.12.08

   editlog = pvlist_create(maxedits);                                      //  history log of image edits done
   for (ii = 0; ii < maxedits; ii++)                                       //  pre-load all pvlist entries
      pvlist_append(editlog,"nothing");

   //  set current directory/file from command line or saved parameters

   if (Ntopdirk) curr_dirk = strdupz(topdirk[0],0,"initz");                //  use 1st top direk if defined
   else {
      pp1 = getcwd(command,ccc);                                           //  or use current directory
      if (pp1) curr_dirk = strdupz(pp1,0,"initz");
   }
   
   if (initial_file) {                                                     //  from command line or previous session
      pp1 = realpath(initial_file,0);                                      //  prepend directory if needed
      if (pp1) {
         zfree(initial_file);
         initial_file = strdupz(pp1,0,"initz");
         free(pp1);
      }
      else {
         printf("invalid file: %s \n",initial_file);                       //  invalid file path
         zfree(initial_file);
         initial_file = 0;
      }
   }

   if (initial_file) {
      ftype = image_file_type(initial_file);
      if (ftype == 0) {                                                    //  non-existent file
         printf("invalid file: %s \n",initial_file);
         zfree(initial_file);
         initial_file = 0;
      }
      else if (ftype == 1) {                                               //  is a directory
         if (curr_dirk) zfree(curr_dirk);
         curr_dirk = initial_file;
         initial_file = 0;
         Fdirk = 1;
      }
      else {
         if (curr_dirk) zfree(curr_dirk);
         curr_dirk = strdupz(initial_file,0,"initz");                      //  set current directory from file
         pp1 = strrchr(curr_dirk,'/');
         if (pp1) *pp1 = 0;
      }
   }

   if (curr_dirk) err = chdir(curr_dirk);                                  //  set current directory

   menulock(0);                                                            //  enable menus

   //  process command line options

   if (SS_imagefile) {                                                     //  start slide show if wanted
      f_open(SS_imagefile,1);
      if (SS_musicfile) {
         sprintf(command,"xdg-open \"%s\" ",SS_musicfile);                 //  with music if wanted
         printf("command: %s \n",command);
         err = system(command);
      }
      m_slideshow(0,0);
      return 0;
   }

   if (initial_file)                                                       //  open initial file
      f_open(initial_file,1);
   
   else if (Fdirk) {                                                       //  open gallery for initial directory
      gallery(curr_dirk,"init");                                           //  v.13.01
      gallery(0,"paint");
   }
   
   else if (Fprev)                                                         //  start with previous file
      m_previous(0,0);
   
   else if (Frecent)                                                       //  start with recent file
      m_recent(0,0);
   
   else if (Fblank); /** do nothing **/

   //  if no command line option, get startup option from user settings
   
   else if (strEqu(startdisplay,"recent"))
      m_recent(0,0);

   else if (strEqu(startdisplay,"prev"))
      m_previous(0,0);

   else if (strEqu(startdisplay,"dirk")) {
      if (startdirk && *startdirk == '/') {                                //  v.12.09
         if (curr_dirk) zfree(curr_dirk);
         curr_dirk = strdupz(startdirk,0,"initz");
         err = chdir(curr_dirk); 
         navi::thumbsize = 128;
         gallery(curr_dirk,"init",0);                                      //  v.13.01
         gallery(0,"paint",0);
      }
   }

   else if (strEqu(startdisplay,"file"))
      f_open(startfile,1);
   
   zmainloop();
   
   if (startmenu) {                                                        //  startup menu on command line    v.12.09
      printf("startmenu: %s \n",startmenu);
      for (ii = 0; ii < Nmenus; ii++)                                      //  convert menu name to menu function
         if (strcasecmp(startmenu,menutab[ii].text) == 0) break;
      if (ii < Nmenus) menutab[ii].func(0,menutab[ii].arg);                //  call the menu function
   }

   return 0;                                                               //  don't come back
}


/**************************************************************************/

//  Periodic function (100 milliseconds)
//  Avoid any thread usage of gtk/gdk functions.

int gtimefunc(void *arg)
{
   static zdialog    *zdmessage = 0;
   static int        domore = 0;
   char              busymess[20];
   int               percent, px, py;

   if (Wrepaint) {                                                         //  window update needed
      if (mutex_trylock(&Fpixmap_lock) != 0) return 1;                     //  pixmaps locked, come back later
      mutex_unlock(&Fpixmap_lock);                                         //  (mwpaint1 will lock)      v.12.09
      gtk_widget_queue_draw(cDwin);
   }
   
   if (--domore > 0) return 1;                                             //  do rest 1/5 as frequently
   domore = 5;

   if (Ffuncbusy || Fthreadbusy) {                                         //  something is processing         v.12.11
      if (Fbusy_goal) {
         percent = 100 * Fbusy_done / Fbusy_goal;                          //  construct "BUSY nn%"
         sprintf(busymess," BUSY %d%c ",percent,'%');
      }
      else strcpy(busymess," BUSY ");
      
      px = Morgx / Mscale;
      py = Morgy / Mscale;
      draw_text(px,py,busymess,"monospace 11");                            //  write upper left visible image window
   }

   update_statusbar();                                                     //  update status bar information

   if (Fslideshow) slideshow_next("timer");                                //  show next image if time is up

   if (threadmessage && ! zdmessage)                                       //  display message for thread process
      zdmessage = zmessage_post(mWin,0,threadmessage);                     //  (avoid use of GTK in threads)

   if (zdmessage && ! threadmessage)                                       //  message terminated by thread
      zdialog_free(zdmessage);
   
   if (zdmessage && zdmessage->sentinel != zdsentinel) {                   //  popup window closed by user
      zdmessage = 0;
      threadmessage = 0;
   }
   
   return 1;
}


/**************************************************************************/

//  update status bar with image data and status
//  called from timer function

void update_statusbar()
{
   static double  time1 = 0, time2, cpu1 = 0, cpu2, cpuload;
   char           text1[300], text2[100];
   int            ww, hh, scale, bpc;
   double         file_MB = 1.0 / mega * curr_file_size;
   
   if (! cSbar) return;                                                    //  v.12.07

   *text1 = *text2 = 0;
   
   if (! time1) {
      time1 = get_seconds();
      cpu1 = jobtime();
   }

   time2 = get_seconds();                                                  //  compute process cpu load %
   if (time2 - time1 > 1.0) {                                              //    at 1 second intervals
      cpu2 = jobtime();
      cpuload = 100.0 * (cpu2 - cpu1) / (time2 - time1);
      time1 = time2;
      cpu1 = cpu2;
   }

   sprintf(text1,"CPU %03.0f%c",cpuload,'%');                              //  CPU 023%
   
   if (curr_file)
   {
      if (E3pxm) {
         ww = E3pxm->ww;
         hh = E3pxm->hh;
      }
      else {
         ww = Fpxb->ww;
         hh = Fpxb->hh;
      }

      bpc = curr_file_bpc;

      snprintf(text2,99,"  %dx%dx%d",ww,hh,bpc);                           //  2345x1234x16 (preview) 1.56MB 45%
      strcat(text1,text2);
      if (Fpreview) strcat(text1," (preview)");
      sprintf(text2,"  %.2fMB",file_MB);
      strcat(text1,text2);
      scale = Mscale * 100 + 0.5;
      sprintf(text2,"  %d%c",scale,'%');
      strcat(text1,text2);
   
      if (Pundo) {                                                         //  edit undo stack depth 
         snprintf(text2,99,"  edits: %d",Pundo);
         strcat(text1,text2);
      }
   }

   if (Fmenulock) strcat(text1,"  menu locked");
   if (sa_stat == 3) strcat(text1,"  area active");
   if (zfuncs::zdialog_busy) strcat(text1,"  dialog open");
   
   if (*SB_text) {                                                         //  application text
      strcat(text1,"  ");
      strcat(text1,SB_text);
   }

   stbar_message(cSbar,text1);                                             //  write to status bar

   return;
}


/**************************************************************************/

//  process top-level menu entry

void topmenufunc(GtkWidget *, cchar *menu)
{
   topmenu = (char *) menu;                                                //  remember top-level menu name
   return;
}


//  functions for main window event signals

int delete_event()                                                         //  main window closed
{
   printf("main window delete event \n");
   if (mod_keep()) return 1;                                               //  allow user bailout
   quitxx();
   return 0;
}

int destroy_event()                                                        //  main window destroyed
{
   printf("main window destroy event \n");
   quitxx();
   return 0;
}

int state_event(GtkWidget *, GdkEvent *event)                              //  main window state changed
{
   int state = ((GdkEventWindowState *) event)->new_window_state;          //  track window fullscreen status   v.12.03
   if (state & GDK_WINDOW_STATE_FULLSCREEN) Wfullscreen = 1;
   else Wfullscreen = 0;
   return 0;
}


/**************************************************************************/

//  set window visible tab, F/G/W

int settab(cchar *tabname)                                                 //  v.12.09
{
   int      tab;
   
   if (strEqu(tabname,"F")) tab = tabF;
   if (strEqu(tabname,"G")) tab = tabG;
   if (strEqu(tabname,"W")) tab = tabW;
   if (tab == cTab) return 1;

   if (checkmenulock()) return 0;                                          //  block if menu locked      v.12.09
   gtk_notebook_set_current_page(GTK_NOTEBOOK(tabWin),tab);
   return 1;
}


//  current tab has changed, by user or by function settab() above

void tabchanged(GtkNotebook *, GtkWidget *, int tab, void *)               //  overhauled    v.12.10
{
   int            tabrecover(void *);
   int            Nth, reset = 0;
   static PXB     *Ffpxb = 0, *Fmpxb;                                      //  current image/window state
   static int     Fmorgx, Fmorgy, Fdorgx, Fdorgy;
   static float   Ffzoom, Fpscale;
   static PXM     *Fe0pxm, *Fe3pxm;
   static PXB     *Wmpxb = 0;                                              //  world map image/window state
   static int     Wmorgx, Wmorgy, Wdorgx, Wdorgy;
   static float   Wfzoom, Wpscale;
   
   if (tab == cTab) return;                                                //  no change

   if (checkmenulock()) reset = 1;                                         //  message user if menu locked
   if (mod_keep()) reset = 1;                                              //  warn user of lost edits
   if (reset) {
      g_timeout_add(0,tabrecover,0);                                       //  schedule tab recover function
      return;                                                              //  (does not work from here)
   }
   
   if (Fmodified) undo_all();                                              //  edits lost if changing tabs during edit

   if (cTab == tabW) {                                                     //  leaving tab W
      Wmpxb = Mpxb;                                                        //  save world map window status
      Wmorgx = Morgx;
      Wmorgy = Morgy;
      Wdorgx = Dorgx;
      Wdorgy = Dorgy;
      Wfzoom = Fzoom;
      Wpscale = Pscale;
   }

   if (cTab == tabF) {                                                     //  leaving tab F
      Ffpxb = Fpxb;                                                        //  save image file window status
      Fmpxb = Mpxb;
      Fmorgx = Morgx;
      Fmorgy = Morgy;
      Fdorgx = Dorgx;
      Fdorgy = Dorgy;
      Ffzoom = Fzoom;
      Fpscale = Pscale;
      Fe0pxm = E0pxm;
      Fe3pxm = E3pxm;
   }

   if (tab == tabW) {                                                      //  entering tab W
      if (! init_geotags()) {
         g_timeout_add(0,tabrecover,0);                                    //  schedule tab recover function
         return;                                                           //  (does not work from here)
      }
      cTab = tab;
      Fpxb = Wpxb;                                                         //  restore world map window status
      Mpxb = Wmpxb;
      Morgx = Wmorgx;
      Morgy = Wmorgy;
      Dorgx = Wdorgx;
      Dorgy = Wdorgy;
      Fzoom = Wfzoom;
      Pscale = Wpscale;
      E0pxm = E3pxm = 0;
      cDwin = wDwin;                                                       //  set drawing window and no status bar
      cSbar = 0;
      gdkWin = gtk_widget_get_window(cDwin);                               //  GDK drawing window
      gdk_window_set_background_rgba(gdkWin,&GDKdgray);
      gtk_window_set_title(MWIN,ZTX("Location from World Map"));           //  window title
      m_geotag_worldmap(0,0);                                              //  activate clickable world map    v.12.09
      mwpaint2();
   }

   if (tab == tabF) {                                                      //  entering tab F
      cTab = tab;
      Fpxb = Ffpxb;                                                        //  restore image file window status
      Mpxb = Fmpxb;
      Morgx = Fmorgx;
      Morgy = Fmorgy;
      Dorgx = Fdorgx;
      Dorgy = Fdorgy;
      Fzoom = Ffzoom;
      Pscale = Fpscale;
      E0pxm = Fe0pxm;
      E3pxm = Fe3pxm;
      cDwin = fDwin;                                                       //  set drawing window and status bar
      cSbar = fSbar;
      gdkWin = gtk_widget_get_window(cDwin);                               //  GDK drawing window
      gdk_window_set_background_rgba(gdkWin,&GDKdgray);
      set_mwin_title();                                                    //  window title from image file
      mwpaint2();
   }

   if (tab == tabG) {                                                      //  entering tab G
      cTab = tab;
      gdkWin = 0;                                                          //  v.12.10
      zfuncs::F1_help_topic = "navigation";
      if (curr_file) Nth = curr_file_posn;                                 //  curr. file in top row     v.13.01
      else Nth = -1;
      gallery(0,"paint",Nth);
   }

   return;
}


//  reset to prior tab after disallowed user change                        //  v.12.07
//  use: g_timeout_add(10,tabrecover,0)

int tabrecover(void *)
{
   printf("reset tab change %d \n",cTab);
   gtk_notebook_set_current_page(GTK_NOTEBOOK(tabWin),cTab);
   return 0;                                                               //  don't come back
}
                  

/**************************************************************************/

//  GTK3 "draw" function for the main/drawing window.
//  Paint window when created, exposed, resized, or image modified (edited).
//  Update window image if Frefresh flag, scale change or window size change.
//  Draw the image section currently within the visible window. 
//  May NOT be called from threads.

int mwpaint1(GtkWidget *cDwin, cairo_t *cr)                                //  overhauled    v.12.03
{
   GdkPixbuf      *Dpixbuf;
   uint8          *pixels;
   char           busymess[20];
   static int     pdww = 0, pdhh = 0;                                      //  prior window size
   float          wscale, hscale;
   int            fww, fhh;                                                //  current image size at 1x
   int            mww, mhh;                                                //  scaled image size
   int            centerx, centery; 
   int            refresh = 0;                                             //  flag, full image refesh needed
   int            mousex, mousey;                                          //  mouse position after zoom
   int            rs, px, py, percent;
   float          magx, magy;                                              //  mouse drag, magnification ratios
   
   if (Fshutdown) return 0;                                                //  shutdown underway
   if (! Fpxb) return 0;                                                   //  no image
   if (! gdkWin) return 0;

   if (mutex_trylock(&Fpixmap_lock) != 0) return 0;                        //  cannot lock pixmaps    v.12.09
   
   Dww = gdk_window_get_width(gdkWin);                                     //  (new) drawing window size
   Dhh = gdk_window_get_height(gdkWin);
   if (Dww < 20 || Dhh < 20) return 0;                                     //  too small

   Wrepaint = 0;                                                           //  repaint started        v.12.09

   if (Frefresh) {                                                         //  image was updated
      if (E0pxm || E3pxm) {
         PXB *pxb1;
         if (E3pxm) pxb1 = PXM_PXB_copy(E3pxm);                            //  use image now being edited
         else pxb1 = PXM_PXB_copy(E0pxm);                                  //  or use already edited image
         PXB *pxb2 = Fpxb;
         Fpxb = pxb1;                                                      //  don't leave Fpxb undefined
         PXB_free(pxb2);
      }
      refresh++;                                                           //  full image refresh needed
      Frefresh = 0;
   }

   centerx = (Morgx + 0.5 * dww) / Mscale;                                 //  center of window, image space
   centery = (Morgy + 0.5 * dhh) / Mscale;                                 //  (before window or scale change)

   if (Dww != pdww || Dhh != pdhh) {                                       //  window size changed
      refresh++;                                                           //  image refresh needed
      pdww = Dww; 
      pdhh = Dhh;
   }
   
   fww = Fpxb->ww;                                                         //  1x image size
   fhh = Fpxb->hh;

   if (Fzoom == 0) {                                                       //  scale to fit window
      wscale = 1.0 * Dww / fww;
      hscale = 1.0 * Dhh / fhh;
      if (wscale < hscale) Mscale = wscale;                                //  use greatest ww/hh ratio
      else  Mscale = hscale;
      if (fww <= Dww && fhh <= Dhh && ! Fblowup) Mscale = 1.0;             //  small image 1x unless Fblowup
      zoomx = zoomy = 0;                                                   //  no zoom target
   }
   else Mscale = Fzoom;                                                    //  scale to Fzoom level
   
   mww = fww * Mscale;                                                     //  scaled image size for window
   mhh = fhh * Mscale;
   
   dww = Dww;                                                              //  image fitting inside drawing window
   if (dww > mww) dww = mww;                                               //  image < window size
   dhh = Dhh;
   if (dhh > mhh) dhh = mhh;

   if (Mscale != Pscale) {                                                 //  scale changed
      Morgx = Mscale * centerx - 0.5 * dww;                                //  change origin to keep same center
      Morgy = Mscale * centery - 0.5 * dhh;                                //  (subject to later rules)
      Pscale = Mscale;                                                     //  remember scale
      refresh++;                                                           //  full image refresh needed
   }
   
   if (! Mpxb) refresh++;                                                  //  v.13.01

   if (refresh) {                                                          //  image refresh needed
      if (Mpxb) PXB_free(Mpxb);
      if (Mscale == 1) Mpxb = PXB_copy(Fpxb);                              //  fast 1x zoom                    v.12.11
      else Mpxb = PXB_rescale(Fpxb,mww,mhh);
   }

   if ((Mxdrag || Mydrag)) {                                               //  pan/scroll via mouse drag       v.12.05
      zoomx = zoomy = 0;                                                   //  no zoom target
      if (Fdragopt == 1) {
         Morgx -= round(Mwdragx);                                          //  same direction, not magnified   v.12.10
         Morgy -= round(Mwdragy);
      }
      if (Fdragopt == 2) {                                                 //  scroll full image in one sweep
         magx = 1.0 * (mww - dww) / dww;
         magy = 1.0 * (mhh - dhh) / dhh;                                   //  needed magnification of mouse movement
         if (magx < 1) magx = 1;                                           //  retain a minimum speed
         if (magy < 1) magy = 1;                                           //                                  v.12.09
         Morgx += round(Mwdragx * magx);                                   //  opposite direction, magnified 
         Morgy += round(Mwdragy * magy);
      }
   }

   if (dww < Dww) Dorgx = 0.5 * (Dww - dww);                               //  if scaled image width < window width
   else Dorgx = 0;                                                         //    center image in window
   if (dhh < Dhh) Dorgy = 0.5 * (Dhh - dhh);
   else Dorgy = 0;

   if (zoomx || zoomy) {                                                   //  requested zoom center
      Morgx = Mscale * zoomx - 0.5 * dww;                                  //  corresp. window position within image
      Morgy = Mscale * zoomy - 0.5 * dhh;
   }
   
   if (Morgx < 0) Morgx = 0;                                               //  maximize image within window
   if (Morgy < 0) Morgy = 0;                                               //  (no unused margins)
   if (Morgx + dww > mww) Morgx = mww - dww;
   if (Morgy + dhh > mhh) Morgy = mhh - dhh;
   
   if (zoomx || zoomy) {                                                   //  zoom target
      mousex = zoomx * Mscale - Morgx + Dorgx;
      mousey = zoomy * Mscale - Morgy + Dorgy;                             //  mouse pointer follows target
      gdk_window_move_pointer(gdkWin,mousex,mousey);
      zoomx = zoomy = 0;                                                   //  reset zoom target
   }
   
   pixels = PXBpix(Mpxb,Morgx,Morgy);                                      //  Mpxb image section within drawing window
   rs = Mpxb->rs;                                                          //  (Morgx/y, dww/hh, rowstride rs)
   Dpixbuf = gdk_pixbuf_new_from_data(pixels,GDKRGB,0,8,dww,dhh,rs,0,0);   //  make a pixbuf from section

   gdk_cairo_set_source_pixbuf(cr,Dpixbuf,Dorgx,Dorgy);                    //  paint Mpxb image section
   cairo_paint(cr);
   g_object_unref(Dpixbuf);

   mwcr = cr;                                                              //  cairo context for draw funcs below
   cairo_set_line_width(mwcr,1);

   if (Ntoplines) draw_toplines(1);                                        //  draw line overlays
   if (Fgrid) draw_gridlines();                                            //  draw grid lines
   if (Ntoptext) draw_toptext();                                           //  draw text strings
   if (Ntopcircles) draw_topcircles();                                     //  draw circles
   if (Fshowarea) sa_show(1);                                              //  draw select area outline

   if (Ffuncbusy) {                                                        //  something is processing
      if (Fbusy_goal) {
         percent = 100 * Fbusy_done / Fbusy_goal;                          //  construct "BUSY nn%"
         sprintf(busymess," BUSY %d%c ",percent,'%');
      }
      else strcpy(busymess," BUSY ");

      px = Morgx / Mscale;
      py = Morgy / Mscale;
      draw_text(px,py,busymess,"monospace 11");                            //  write upper left visible image window
   }

   mwcr = 0;                                                               //  cairo context invalid
   mutex_unlock(&Fpixmap_lock);                                            //  unlock pixmaps
   Wpainted++;                                                             //  notify edit function of repaint

   if (winhisto) gtk_widget_queue_draw(winhistoH);                         //  update histogram if active
   return 0;
}


/**************************************************************************/

//  Cause (modified) output image to get repainted soon.
//  Same as mwpaint1() but delayed. MAY be called from threads.

void mwpaint2()
{
   Frefresh++;                                                             //  req. full window refresh
   Wrepaint++;                                                             //  req. repaint by periodic function
   return;
}


//  Update a section of Fpxb and Mpxb from an updated section of E3pxm,
//  then update the corresponding section of the drawing window.
//  This avoids a full image refresh, E3pxm > Fpxb > Mpxb > drawing window.
//  px3, py3, ww3, hh3: modified section within E3pxm to be propagated.
//  May NOT be called from threads.

void mwpaint3(int px3, int py3, int ww3, int hh3)                          //  overhauled       v.12.03
{
   PXM_PXB_update(E3pxm,Fpxb,px3,py3,ww3,hh3);                             //  E3pxm > Fpxb, both 1x scale 
   PXB_PXB_update(Fpxb,Mpxb,px3,py3,ww3,hh3);                              //  Fpxb > Mpxb, scaled up or down
   mwpaint4(px3,py3,ww3,hh3);                                              //  update drawing window from Mpxb
   return;
}


//  Repaint a section of the Mpxb image in the visible window.
//  px3, py3, ww3, hh3: area to be repainted (in 1x image space).
//  May NOT be called from threads.

void mwpaint4(int px3, int py3, int ww3, int hh3)                          //  overhauled       v.12.03
{
   GdkPixbuf   *pixbuf;
   uint8       *pixels;
   int         px2, py2, ww2, hh2, rs;
   int         crflag = 0;
   
   if (! gdkWin) return;                                                   //  v.12.11

   px2 = Mscale * px3 - 2;                                                 //  1x image space to Mpxb space
   py2 = Mscale * py3 - 2;                                                 //  (expanded)
   ww2 = Mscale * ww3 + 2 / Mscale + 4;
   hh2 = Mscale * hh3 + 2 / Mscale + 4;

   if (px2 + ww2 > Mpxb->ww) ww2 = Mpxb->ww - px2;                         //  avoid overshoot
   if (py2 + hh2 > Mpxb->hh) hh2 = Mpxb->hh - py2;
   
   if (px2 < Morgx) {                                                      //  reduce to currently visible window
      ww2 = ww2 - (Morgx - px2);
      px2 = Morgx;
   }

   if (py2 < Morgy) {
      hh2 = hh2 - (Morgy - py2);
      py2 = Morgy;
   }

   if (ww2 <= 0 || hh2 <= 0) return;                                       //  completely outside visible window
   
   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      crflag = 1;
   }

   pixels = PXBpix(Mpxb,px2,py2);
   rs = Mpxb->rs;                                                          //  row stride
   pixbuf = gdk_pixbuf_new_from_data(pixels,GDKRGB,0,8,ww2,hh2,rs,0,0);    //  pixbuf for image within window

   px2 = px2 - Morgx + Dorgx;                                              //  section position in window
   py2 = py2 - Morgy + Dorgy;

   gdk_cairo_set_source_pixbuf(mwcr,pixbuf,px2,py2);                       //  draw area to window
   cairo_paint(mwcr);
   g_object_unref(pixbuf);
   
   if (Fshowarea) { 
      px3 = (px2 - Dorgx + Morgx) / Mscale;                                //  back to image scale, expanded
      py3 = (py2 - Dorgy + Morgy) / Mscale;
      ww3 = ww2 / Mscale + 2;
      hh3 = hh2 / Mscale + 2;
      sa_show_rect(px3,py3,ww3,hh3);                                       //  refresh select area outline
   }

   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


/**************************************************************************/

//  mouse event function - capture buttons and drag movements

void mouse_event(GtkWidget *, GdkEventButton *event, void *)
{
   void mouse_convert(int &xpos, int &ypos);

   int            button, time, type, scroll;
   static int     bdtime = 0, butime = 0, mbusy = 0;
   static int     mdragx0, mdragy0;
   
   if (! Fpxb) return;                                                     //  v.12.10

   if (mutex_trylock(&Fpixmap_lock) != 0) return;                          //  block mouse in image transition  v.12.10
   mutex_unlock(&Fpixmap_lock);

   type = event->type;
   button = event->button;                                                 //  button, 1/2/3 = left/center/right
   time = event->time;
   Mxposn = event->x;                                                      //  mouse position in window
   Myposn = event->y;
   scroll = ((GdkEventScroll *) event)->direction;                         //  scroll wheel event
   
   Mwxposn = Mxposn;                                                       //  mouse position, window space     v.12.05
   Mwyposn = Myposn;

   mouse_convert(Mxposn,Myposn);                                           //  convert to image space
   
   if (type == GDK_SCROLL) {                                               //  scroll wheel = zoom          v.12.01
      zoomx = Mxposn;                                                      //  zoom center = mouse position (doriano)
      zoomy = Myposn;
      if (scroll == GDK_SCROLL_UP) m_zoom(null,"in");
      if (scroll == GDK_SCROLL_DOWN) m_zoom(null,"out");
      return;
   }

   if (type == GDK_MOTION_NOTIFY) {
      if (mbusy) return;                                                   //  discard excess motion events
      mbusy++;
      zmainloop();
      mbusy = 0;
   }

   if (type == GDK_BUTTON_PRESS) {                                         //  button down
      Mdrag++;                                                             //  possible drag start
      bdtime = time;                                                       //  time of button down
      Mbutton = button;
      mdragx0 = Mwxposn;                                                   //  window position at button down   v.12.05
      mdragy0 = Mwyposn;
      Mxdown = Mxposn;                                                     //  image position at button down
      Mydown = Myposn;
      Mxdrag = Mydrag = 0;
   }

   if (type == GDK_MOTION_NOTIFY && Mdrag) {                               //  drag underway
      Mwdragx = Mwxposn - mdragx0;                                         //  drag increment, window space     v.12.05
      Mwdragy = Mwyposn - mdragy0;
      mdragx0 = Mwxposn;                                                   //  new drag origin = current position
      mdragy0 = Mwyposn;
      Mxdrag = Mxposn;                                                     //  drag position, image space
      Mydrag = Myposn;
   }

   if (type == GDK_BUTTON_RELEASE) {                                       //  button up
      Mxclick = Myclick  = 0;                                              //  reset click status
      butime = time;                                                       //  time of button up
      if (butime - bdtime < 400) {                                         //  less than 0.4 secs down
         if (Mxposn == Mxdown && Myposn == Mydown) {                       //       and not moving
            if (Mbutton == 1) LMclick++;                                   //  left mouse click
            if (Mbutton == 3) RMclick++;                                   //  right mouse click
            Mxclick = Mxdown;                                              //  click = button down position
            Myclick = Mydown;
            if (button == 2) {                                             //  center button click
               zoomx = Mxposn;                                             //  re-center at mouse (Doriano)
               zoomy = Myposn;
               gtk_widget_queue_draw(cDwin);                               //  v.12.03
            }
         }
      }
      Mxdown = Mydown = Mxdrag = Mydrag = Mdrag = Mbutton = 0;             //  forget buttons and drag
   }

   Fmousemain = 1;                                                         //  mouse acts on main window
   if (Mcapture) Fmousemain = 0;                                           //  curr. function handles mouse 
   if (mouseCBfunc) Fmousemain = 0;                                        //  mouse owned by callback function
   if (KBcontrolkey) Fmousemain = 1;                                       //  mouse acts on main window
   
   if (mouseCBfunc && ! Fmousemain) {                                      //  pass to handler function
      if (mbusy) return;
      mbusy++;                                                             //  stop re-entrance
      (* mouseCBfunc)();
      mbusy = 0;     
      Fmousemain = 1;                                                      //  click/drag params are processed
   }                                                                       //    if not reset by handler func.  v.12.11
   
   if (! Fmousemain) return;                                               //  curr. function handles mouse     v.12.11
   
   if (LMclick) {                                                          //  left click = zoom request
      LMclick = 0;
      zoomx = Mxclick;                                                     //  zoom center = mouse
      zoomy = Myclick;
      m_zoom(null,"in");
   }

   if (RMclick) {                                                          //  right click
      RMclick = 0;
      if (Fzoom) {                                                         //  if zoomed image, reset to fit window
         zoomx = zoomy = 0;
         m_zoom(null,"fit");
      }
      else if (curr_file && cTab == tabF)                                  //  v.12.12
         image_Rclick_popup(curr_file_posn);                               //  image right-click popup menu
   }
   
   if ((Mxdrag || Mydrag))                                                 //  drag = scroll by mouse
      gtk_widget_queue_draw(cDwin);

   return;
}


//  convert mouse position from window space to image space

void mouse_convert(int &xpos, int &ypos)
{
   xpos = (xpos - Dorgx + Morgx) / Mscale + 0.5;                           //  v.12.03
   ypos = (ypos - Dorgy + Morgy) / Mscale + 0.5;

   if (xpos < 0) xpos = 0;                                                 //  if outside image put at edge
   if (ypos < 0) ypos = 0;

   if (E3pxm) { 
      if (xpos >= E3pxm->ww) xpos = E3pxm->ww-1;
      if (ypos >= E3pxm->hh) ypos = E3pxm->hh-1;
   }
   else {
      if (xpos >= Fpxb->ww) xpos = Fpxb->ww-1;
      if (ypos >= Fpxb->hh) ypos = Fpxb->hh-1;
   }
   
   return;
}


/**************************************************************************/

//  Called from zdialog event whenever KB control key or KB shift key 
//  is pressed or released. Mark key status and set focus to main window.

void zdialog_KBcontrol(int KBkey, int state)                               //  v.12.12
{
   if (KBkey == GDK_KEY_Control_L || KBkey == GDK_KEY_Control_R)           //  Ctrl key changed
       KBcontrolkey = state;
   if (KBkey == GDK_KEY_Shift_L || KBkey == GDK_KEY_Shift_R)               //  Shift key changed
       KBshiftkey = state;
   if (KBcontrolkey) gtk_window_present(MWIN);                             //  control key takes focus
   return;
}


//  keyboard event functions
//  GDK key symbols: /usr/include/gtk-3.0/gdk/gdkkeysyms.h

int KBpress(GtkWidget *win, GdkEventKey *event, void *)                    //  prevent propagation of key-press
{                                                                          //    events to toolbar buttons
   KBkey = event->keyval;

   if (KBkey == GDK_KEY_Control_L || KBkey == GDK_KEY_Control_R)           //  Ctrl key is pressed
       KBcontrolkey = 1;

   if (KBkey == GDK_KEY_Shift_L || KBkey == GDK_KEY_Shift_R)               //  Shift key is pressed
       KBshiftkey = 1;
   
   if (KBkey == GDK_KEY_a || KBkey == GDK_KEY_A)                           //  'A' key is pressed
       KB_A_key = 1;                                                       //  (undo/redo 1 --> undo/redo ALL)

   return 1;
}


int KBrelease(GtkWidget *win, GdkEventKey *event, void *)                  //  some key was released
{
   int      ii, jj, cc, angle = 0;
   char     matchkey[20];
   
   KBkey = event->keyval;

   if (KBkey == GDK_KEY_Control_L || KBkey == GDK_KEY_Control_R) {         //  Ctrl key released
      KBcontrolkey = 0;
      KBkey = 0;
      return 1;
   }
   
   if (KBkey == GDK_KEY_Shift_L || KBkey == GDK_KEY_Shift_R) {             //  Shift key released
      KBshiftkey = 0;
      KBkey = 0;
      return 1;
   }

   if (! KBcontrolkey && ! KBshiftkey) {                                   //  F/G/W keys are tab shortcuts    v.12.11
      if (KBkey == GDK_KEY_F || KBkey == GDK_KEY_f)                        //  image file (main) window
      { settab("F"); return 1; }
      if (KBkey == GDK_KEY_G || KBkey == GDK_KEY_g)                        //  gallery window
      { settab("G"); return 1; }
      if (KBkey == GDK_KEY_W || KBkey == GDK_KEY_w)                        //  world map window
      { settab("W"); return 1; }
   }

   if (cTab == tabG) {                                                     //  tab G is active,
      navi::KBxrelease(win,event,0);                                       //    pass to gallery window function
      return 1;
   }

   if (KBcapture) return 1;                                                //  let current function handle it

//  reserved shortcuts (not user configurable)
   
   if (KBkey == GDK_KEY_a || KBkey == GDK_KEY_A) {                         //  'A' key is released
      KB_A_key = 0;                                                        //  (undo/redo ALL --> undo/redo 1)
      KBkey = 0;
      return 1;
   }

   if (KBkey == GDK_KEY_F1)                                                //  F1  >>  user guide
      showz_userguide(zfuncs::F1_help_topic);                              //  show topic if there, or page 1
   
   if (KBkey == GDK_KEY_F11) {                                             //  F11  >>  toggle main window
      if (! Wfullscreen) win_fullscreen();                                 //    to full-screen mode and back
      else win_unfullscreen();
   }

   if (KBkey == GDK_KEY_Escape && Wfullscreen)                             //  ESC key >> exit full screen      v.12.11
      win_unfullscreen();

   if (KBkey == GDK_KEY_z || KBkey == GDK_KEY_Z)                           //  Z key                            v.12.11
      if (! KBcontrolkey) m_zoom(null,"100");                              //  if no CTRL, toggle zoom 1x

   if (cTab == tabW) return 1;                                             //  tab W active, no other KB events

   if (KBkey == GDK_KEY_plus) m_zoom(null,"in");                           //  + key  >>  zoom in
   if (KBkey == GDK_KEY_minus) m_zoom(null,"fit");                         //  - key  >>  fit to window
   if (KBkey == GDK_KEY_equal) m_zoom(null,"in");                          //  = key: same as +
   if (KBkey == GDK_KEY_KP_Add) m_zoom(null,"in");                         //  keypad +
   if (KBkey == GDK_KEY_KP_Subtract) m_zoom(null,"fit");                   //  keypad -

   if (KBkey == GDK_KEY_R || KBkey == GDK_KEY_r) angle = 90;               //  90 deg. rotation with R/L keys
   if (KBkey == GDK_KEY_L || KBkey == GDK_KEY_l) angle = -90;
   if (angle) {
      if (curr_file && ! E0pxm && ! checkmenulockquiet()) {                //  if no edit in progress
         mutex_lock(&Fpixmap_lock);                                        //  block window updates
         E0pxm = PXM_load(curr_file,1);                                    //  load poss. 16-bit image          v.12.10
         if (E0pxm) {
            E3pxm = PXM_rotate(E0pxm,angle);                               //  rotate
            PXM_free(E0pxm);
            E0pxm = E3pxm;
            E3pxm = 0;
         }
         mutex_unlock(&Fpixmap_lock);
         if (E0pxm) m_save(0,"nowarn");                                    //  save uprighted image
         PXM_free(E0pxm);
         f_open(curr_file,0);
      }
   }

   if (Fslideshow) {                                                       //  slide show mode
      if (KBkey == GDK_KEY_Left) slideshow_next("prev");                   //  arrow keys = prev/next image
      if (KBkey == GDK_KEY_Right) slideshow_next("next");
      if (KBkey == GDK_KEY_space) slideshow_next("pause");                 //  spacebar = pause/resume
      if (KBkey == GDK_KEY_F11) m_slideshow(0,0);                          //  F11 = exit slideshow
      if (KBkey == GDK_KEY_Escape) m_slideshow(0,0);                       //  escape = exit slideshow
   }                                                                       //  (escape also exits zdialogs)
   else {
      if (KBkey == GDK_KEY_Left) m_prev(0,0);                              //  arrow keys = prev/next image
      if (KBkey == GDK_KEY_Right) m_next(0,0);
      if (KBkey == GDK_KEY_Page_Up) m_prev(0,0);                           //  page up/down = prev/next image   v.12.12
      if (KBkey == GDK_KEY_Page_Down) m_next(0,0);
   }

   if (KBkey == GDK_KEY_Delete) m_trash(0,0);                              //  delete  >>  trash

//  look for configurable shortcut keys in shortcut table                  //  v.12.06

   if (KBkey >= GDK_KEY_F1 && KBkey <= GDK_KEY_F9) {                       //  input key is F1 to F9
      ii = KBkey - GDK_KEY_F1;
      strcpy(matchkey,"F1");
      matchkey[1] += ii;
   }
   
   else if (KBkey > 255) return 1;                                         //  not a single letter or digit     v.12.07

   else {
      *matchkey = 0;                                                       //  build input key combination
      if (KBcontrolkey) strcat(matchkey,"Ctrl+");                          //  [Ctrl+] [Shift+] key
      if (KBshiftkey) strcat(matchkey,"Shift+");
      cc = strlen(matchkey);
      matchkey[cc] = KBkey;
      matchkey[cc+1] = 0;
   }
   
   for (ii = 0; ii < Nshortcuts; ii++)                                     //  convert key combination to menu name
      if (strcasecmp(matchkey,shortcutkey[ii]) == 0) break;
   
   if (ii < Nshortcuts) {
      for (jj = 0; jj < Nmenus; jj++)                                      //  convert menu name to menu function
         if (strcasecmp(shortcutmenu[ii],menutab[jj].text) == 0) break;
      if (jj < Nmenus) menutab[jj].func(0,menutab[jj].arg);                //  call the menu function
   }

   KBkey = 0;                                                              //  leave control & shift unchanged  v.12.11 
   return 1;
}


/**************************************************************************/

//  set new image zoom level or magnification
//  zoom:   "in"        zoom in in steps
//          "out"       zoom out in steps
//          "fit"       zoom out to fit window
//          "100"       toggle 100% and fit window

void m_zoom(GtkWidget *, cchar *zoom)
{
   int      fww, fhh;
   float    scalew, scaleh, fitscale, fzoom2, zratio = 1;
   float    Rzoom, pixels;
   
   if (! Fpxb) return;                                                     //  bugfix     v.12.12.2
   
   Rzoom = Fzoom;                                                          //  current zoom ratio
   
   if (E3pxm) {
      fww = E3pxm->ww;                                                     //  1x image size
      fhh = E3pxm->hh;
   }
   else  {
      fww = Fpxb->ww;
      fhh = Fpxb->hh;
   }

   if (fww > Dww || fhh > Dhh) {                                           //  get window fit scale
      scalew = 1.0 * Dww / fww;
      scaleh = 1.0 * Dhh / fhh;
      if (scalew < scaleh) fitscale = scalew;
      else fitscale = scaleh;                                              //  window/image, < 1.0
   }
   else fitscale = 1.0;                                                    //  if image < window use 100%
   
   if (strEqu(zoom,"Zoom+")) zoom = "in";                                  //  toolbar button: + = zoom in
   if (strEqu(zoom,"Zoom-")) zoom = "fit";                                 //                  - = fit window

   if (strEqu(zoom,"fit")) Fzoom = 0;                                      //  zoom to fit window

   if (strEqu(zoom,"100")) {
      if (Fzoom != 0) Fzoom = 0;                                           //  toggle 100% and fit window
      else  Fzoom = 1;
   }

   if (strstr("in out",zoom))                                              //  caller: zoom in or out
   {
      if (! Fzoom) Fzoom = fitscale;                                       //  current zoom scale        v.12.03
      for (fzoom2 = 0.125; fzoom2 < 4.0; fzoom2 *= zoomratio)              //  find nearest natural ratio
         if (Fzoom < fzoom2 * sqrt(zoomratio)) break;
      if (strEqu(zoom,"in")) zratio = zoomratio;                           //  zoom in, make image larger
      if (strEqu(zoom,"out")) zratio = 1.0 / zoomratio;                    //  zoom out, make image smaller
      Fzoom = fzoom2 * zratio;
      if (Fzoom > 0.124 && Fzoom < 0.126) Fzoom = 0.125;                   //  hit these ratios exactly
      if (Fzoom > 0.24  && Fzoom < 0.26)  Fzoom = 0.25;
      if (Fzoom > 0.49  && Fzoom < 0.51)  Fzoom = 0.50;
      if (Fzoom > 0.99  && Fzoom < 1.01)  Fzoom = 1.00;
      if (Fzoom > 1.99  && Fzoom < 2.01)  Fzoom = 2.00;
      if (Fzoom > 3.99) Fzoom = 4.0;                                       //  max. allowed zoom
      if (Fzoom < fitscale) Fzoom = 0;                                     //  image < window
   }
   
   if (cTab == tabW) {                                                     //  optimize for world map       v.12.11
      if (strEqu(zoom,"in")) Fzoom = 1.0;                                  //  zoom to 100% directly
      if (Fzoom == 1.0 && Rzoom == 1.0) {
         gtk_widget_queue_draw(cDwin);                                     //  if already, move mouse pointer only
         return;
      }
   }

   if (Fzoom > 1.0) {                                                      //  limit image in window to 1 gigabyte
      pixels = Fzoom * fww * Fzoom * fhh;                                  //  (333 megapixels x 3 bytes/pixel)
      if (pixels > 333 * mega) {                                           //  if size is at maximum,
         Fzoom = Rzoom;                                                    //    move mouse pointer only    v.12.10
         gtk_widget_queue_draw(cDwin);
         return;
      }
   }

   if (! Fzoom) zoomx = zoomy = 0;                                         //  no requested zoom center
   mwpaint2();                                                             //  refresh window               v.12.10
   return;
}


/**************************************************************************/

//  set the main window to fullscreen status 
//  (with no menu, toolbar, statusbar)

void win_fullscreen()                                                      //  v.12.11
{
   if (fMenu) gtk_widget_hide(fMenu);                                      //  remove menu, toolbar, status bar
   if (fTbar) gtk_widget_hide(fTbar);                                      //  v.12.07
   if (fSbar) gtk_widget_hide(fSbar);

   gtk_window_fullscreen(MWIN);

   for (int ii = 0; ii < 20; ii++) {                                       //  wait for window manager to act
      zmainloop();
      if (Wfullscreen) break;
      zsleep(0.1);
   }
   
   if (gdkWin)
      gdk_window_set_background_rgba(gdkWin,&GDKdgray);                    //  restore background
   return;
}


//  restore window to former size and restore menu etc.

void win_unfullscreen()
{
   gtk_window_unfullscreen(MWIN);                                          //  restore old window size
   gtk_widget_show_all(tabWin);                                            //  show all menus and toolbars
   Wfullscreen = 0;
   if (gdkWin)
      gdk_window_set_background_rgba(gdkWin,&GDKdgray);                    //  restore background
   return;
}


/**************************************************************************/

//  draw a pixel using foreground color.
//  px, py are image space.

void draw_pixel(int px, int py)
{
   int               qx, qy, crflag = 0;
   static int        pqx, pqy;
   static uint8      pixel[12];                                            //  2x2 block of pixels
   static GdkPixbuf  *pixbuf1 = 0, *pixbuf4 = 0;
   
   if (! gdkWin) return;                                                   //  v.12.11

   if (! pixbuf1) {
      pixbuf1 = gdk_pixbuf_new_from_data(pixel,GDKRGB,0,8,1,1,3,0,0);      //  1x1 pixels
      pixbuf4 = gdk_pixbuf_new_from_data(pixel,GDKRGB,0,8,2,2,6,0,0);      //  2x2 pixels
   }

   qx = Mscale * px - Morgx;                                               //  image to window space
   qy = Mscale * py - Morgy;

   if (qx == pqx && qy == pqy) return;                                     //  avoid redundant points

   pqx = qx;
   pqy = qy;
   
   if (qx < 0 || qx > dww-2) return;                                       //  keep off image edges
   if (qy < 0 || qy > dhh-2) return;
   
   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      crflag = 1;
   }

   if (Mscale < 1.99) {                                                    //  write 1x1 pixels
      pixel[0] = fg_color[0];
      pixel[1] = fg_color[1];
      pixel[2] = fg_color[2];

      gdk_cairo_set_source_pixbuf(mwcr,pixbuf1,qx+Dorgx,qy+Dorgy);
      cairo_paint(mwcr);
   }
   
   else {                                                                  //  write 2x2 fat pixels
      pixel[0] = pixel[3] = pixel[6] = pixel[9] = fg_color[0];
      pixel[1] = pixel[4] = pixel[7] = pixel[10] = fg_color[1];
      pixel[2] = pixel[5] = pixel[8] = pixel[11] = fg_color[2];

      gdk_cairo_set_source_pixbuf(mwcr,pixbuf4,qx+Dorgx,qy+Dorgy);
      cairo_paint(mwcr);
   }

   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


//  erase one drawn pixel - restore from window image Mpxb.
//  px, py are image space.

void erase_pixel(int px, int py)
{
   static PXB     *pxb2 = 0;
   static uint8   *pix1, *pix2;
   static int     rs2, pqx, pqy;
   int            qx, qy, crflag = 0;

   if (! gdkWin) return;                                                   //  v.12.11
   
   if (! pxb2) {
      pxb2 = PXB_make(2,2);                                                //  2x2 pixbuf
      pix2 = pxb2->pixels;
      rs2 = pxb2->rs;
   }

   qx = Mscale * px;                                                       //  image to window space
   qy = Mscale * py;

   if (qx == pqx && qy == pqy) return;                                     //  avoid same target pixel

   pqx = qx;
   pqy = qy;
   
   if (qx < 0 || qx > Mpxb->ww-2) return;                                  //  pixel outside scaled image
   if (qy < 0 || qy > Mpxb->hh-2) return;
   
   if (qx < Morgx || qx > Morgx + dww-2) return;                           //  pixel outside drawing window
   if (qy < Morgy || qy > Morgy + dhh-2) return;

   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      crflag = 1;
   }

   pix1 = PXBpix(Mpxb,qx,qy);                                              //  source pixels, 2x2 block
   memcpy(pix2,pix1,6);                                                    //  copy 2 upper pixels, 6 bytes
   memcpy(pix2+rs2,pix1+Mpxb->rs,6);                                       //  copy 2 lower pixels

   qx = qx - Morgx + Dorgx;                                                //  target pixel in window
   qy = qy - Morgy + Dorgy;
   gdk_cairo_set_source_pixbuf(mwcr,pxb2->pixbuf,qx,qy);
   cairo_paint(mwcr);

   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


/**************************************************************************/

//  draw line.
//  coordinates are image space.

void draw_line(int x1, int y1, int x2, int y2)
{
   float    px1, py1, px2, py2;
   int      crflag = 0;

   if (! gdkWin) return;                                                   //  v.12.11

   px1 = Mscale * x1 - Morgx + Dorgx;                                      //  image to window space
   py1 = Mscale * y1 - Morgy + Dorgy;
   px2 = Mscale * x2 - Morgx + Dorgx;
   py2 = Mscale * y2 - Morgy + Dorgy;
   
   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      cairo_set_line_width(mwcr,1);
      crflag = 1;
   }

   cairo_set_source_rgb(mwcr,1,1,1);
   cairo_move_to(mwcr,px1,py1);
   cairo_line_to(mwcr,px2,py2);
   cairo_stroke(mwcr);
   
   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


//  draw dotted line.
//  coordinates are image space.

void draw_dotline(int x1, int y1, int x2, int y2)
{
   float    px1, py1, px2, py2;
   double   dashes[2] = { 4, 4 };
   int      crflag = 0;
   
   if (! gdkWin) return;                                                   //  v.12.11

   px1 = Mscale * x1 - Morgx + Dorgx;                                      //  image to window space
   py1 = Mscale * y1 - Morgy + Dorgy;
   px2 = Mscale * x2 - Morgx + Dorgx;
   py2 = Mscale * y2 - Morgy + Dorgy;
   
   if (px1 > Dww-2) px1 = Dww-2;
   if (py1 > Dhh-2) py1 = Dhh-2;
   if (px2 > Dww-2) px2 = Dww-2;
   if (py2 > Dhh-2) py2 = Dhh-2;
   
   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      cairo_set_line_width(mwcr,1);
      crflag = 1;
   }
   else cairo_save(mwcr);

   cairo_set_source_rgb(mwcr,1,1,1);
   cairo_set_dash(mwcr,dashes,2,0);
   cairo_move_to(mwcr,px1,py1);
   cairo_line_to(mwcr,px2,py2);
   cairo_stroke(mwcr);
   
   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }
   else cairo_restore(mwcr);

   return;
}


//  erase line. refresh line path from Mpxb window image.
//  double line width is erased.
//  coordinates are image space.

void erase_line(int x1, int y1, int x2, int y2)
{
   float    pxm, pym, slope;
   int      crflag = 0;

   if (! gdkWin) return;                                                   //  v.12.11

   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      crflag = 1;
   }

   if (abs(y2 - y1) > abs(x2 - x1)) {
      slope = 1.0 * (x2 - x1) / (y2 - y1);
      for (pym = y1; pym <= y2; pym++) {
         pxm = x1 + slope * (pym - y1);
         erase_pixel(pxm,pym);
      }
   }

   else {
      slope = 1.0 * (y2 - y1) / (x2 - x1);
      for (pxm = x1; pxm <= x2; pxm++) {
         pym = y1 + slope * (pxm - x1);
         erase_pixel(pxm,pym);
      }
   }
   
   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


/**************************************************************************/

//  draw pre-set overlay lines on top of image
//  arg = 1:   paint lines only (because window repainted)
//        2:   erase lines and forget them
//        3:   erase old lines, paint new lines, save new in old

void draw_toplines(int arg)
{
   int      ii, crflag = 0;

   if (! gdkWin) return;                                                   //  v.12.11

   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      cairo_set_line_width(mwcr,1);
      crflag = 1;
   }

   if (arg == 2 || arg == 3)                                               //  erase old lines
      for (ii = 0; ii < Nptoplines; ii++)
         erase_line(ptoplinex1[ii],ptopliney1[ii],ptoplinex2[ii],ptopliney2[ii]);
   
   if (arg == 1 || arg == 3)                                               //  draw new lines
      for (ii = 0; ii < Ntoplines; ii++)
         draw_dotline(toplinex1[ii],topliney1[ii],toplinex2[ii],topliney2[ii]);

   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   if (arg == 2) {
      Nptoplines = Ntoplines = 0;                                          //  forget lines
      return;
   }

   for (ii = 0; ii < Ntoplines; ii++)                                      //  save for future erase
   {
      ptoplinex1[ii] = toplinex1[ii];
      ptopliney1[ii] = topliney1[ii];
      ptoplinex2[ii] = toplinex2[ii];
      ptopliney2[ii] = topliney2[ii];
   }

   Nptoplines = Ntoplines;

   return;
}


/**************************************************************************/

//  draw a grid of horizontal and vertical lines.
//  grid line spacings are in window space.

void draw_gridlines()
{
   int      px, py, stepx, stepy, startx, starty;
   int      crflag = 0;

   if (! gdkWin) return;                                                   //  v.12.11

   if (! Fpxb) return;                                                     //  no image
   if (! Fgrid) return;                                                    //  grid lines off

   stepx = gridspace[0];                                                   //  space between grid lines
   stepy = gridspace[1];                                                   //  (window space)

   if (gridcount[0]) stepx = dww / (1 + gridcount[0]);                     //  if line counts specified,
   if (gridcount[1]) stepy = dhh / ( 1 + gridcount[1]);                    //    set spacing accordingly

   startx = stepx * gridoffset[0] / 100;                                   //  variable offsets
   if (startx < 0) startx += stepx;

   starty = stepy * gridoffset[1] / 100;
   if (starty < 0) starty += stepy;

   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      cairo_set_line_width(mwcr,1);
      crflag = 1;
   }

   cairo_set_source_rgb(mwcr,1,1,1);                                       //  white lines

   if (gridon[0] && stepx)
      for (px = Dorgx+startx; px < Dorgx+dww; px += stepx) {
         cairo_move_to(mwcr,px,Dorgy);
         cairo_line_to(mwcr,px,Dorgy+dhh);
      }

   if (gridon[1] && stepy)
      for (py = Dorgy+starty; py < Dorgy+dhh; py += stepy) {
         cairo_move_to(mwcr,Dorgx,py);
         cairo_line_to(mwcr,Dorgx+dww,py);
      }

   cairo_stroke(mwcr);

   cairo_set_source_rgb(mwcr,0,0,0);                                       //  adjacent black lines

   if (gridon[0] && stepx) 
      for (px = Dorgx+startx+1; px < Dorgx+dww; px += stepx) {
         cairo_move_to(mwcr,px,Dorgy);
         cairo_line_to(mwcr,px,Dorgy+dhh);
      }

   if (gridon[1] && stepy)
      for (py = Dorgy+starty+1; py < Dorgy+dhh; py += stepy) {
         cairo_move_to(mwcr,Dorgx,py);
         cairo_line_to(mwcr,Dorgx+dww,py);
      }

   cairo_stroke(mwcr);

   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


/**************************************************************************/

//  maintain a set of text strings written over the image in the window.
//  add a new text string to the list.
//  multiple text strings can be added with the same ID.
//  px and py are image space.

void add_toptext(int ID, int px, int py, cchar *text, cchar *font)
{
   if (Ntoptext == maxtoptext) {
      printf("*** maxtoptext exceeded \n");
      return;
   }
   
   int ii = Ntoptext++;
   toptext[ii].ID = ID;
   toptext[ii].px = px;
   toptext[ii].py = py;
   toptext[ii].text = text;
   toptext[ii].font = font;
   
   return;
}


//  draw current text strings over the image in window.
//  called from mwpaint1().

void draw_toptext() 
{
   int      crflag = 0;

   if (! gdkWin) return;                                                   //  v.12.11

   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      cairo_set_line_width(mwcr,1);
      crflag = 1;
   }

   for (int ii = 0; ii < Ntoptext; ii++)
      draw_text(toptext[ii].px,toptext[ii].py,toptext[ii].text,toptext[ii].font);

   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


//  delete text strings having the given ID from the list

void erase_toptext(int ID)
{
   int      ii, jj;

   for (ii = jj = 0; ii < Ntoptext; ii++)
   {
      if (toptext[ii].ID == ID) continue;
      else toptext[jj++] = toptext[ii];
   }
   
   Ntoptext = jj;
   return;
}


//  draw text on window, black on white background
//  coordinates are image space

void draw_text(int px, int py, cchar *text, cchar *font)
{
   static PangoFontDescription   *pangofont = null;
   static PangoLayout            *pangolayout = null;
   static char                   priorfont[40] = "";
   int                           ww, hh, crflag = 0;

   if (! gdkWin) return;                                                   //  v.12.11

   px = Mscale * px - Morgx + Dorgx;                                       //  image to window space
   py = Mscale * py - Morgy + Dorgy;

   if (strNeq(font,priorfont)) {                                           //  change font
      strncpy0(priorfont,font,40);
      if (pangofont) pango_font_description_free(pangofont);
      if (pangolayout) g_object_unref(pangolayout);
      pangofont = pango_font_description_from_string(font);                //  make pango layout for font
      pangolayout = gtk_widget_create_pango_layout(cDwin,0);
      pango_layout_set_font_description(pangolayout,pangofont);
   }
   
   pango_layout_set_text(pangolayout,text,-1);                             //  add text to layout
   pango_layout_get_pixel_size(pangolayout,&ww,&hh);
   
   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      cairo_set_line_width(mwcr,1);
      crflag = 1;
   }

   cairo_set_source_rgb(mwcr,1,1,1);                                       //  draw white background
   cairo_rectangle(mwcr,px,py,ww,hh);
   cairo_fill(mwcr);
   
   cairo_move_to(mwcr,px,py);                                              //  draw layout with text
   cairo_set_source_rgb(mwcr,0,0,0);
   pango_cairo_show_layout(mwcr,pangolayout);

   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


/**************************************************************************/

//  maintain a set of circles drawn over the image in the window
//  px, py are image space, radius is window space
//  color = 1/2/3/4 = white/black/red/green

void add_topcircle(int px, int py, int radius, int color)
{
   if (Ntopcircles == maxtopcircles) {
      printf("*** maxtopcircles exceeded \n");
      return;
   }
   
   int ii = Ntopcircles++;
   topcircles[ii].px = px;
   topcircles[ii].py = py;
   topcircles[ii].radius = radius;
   topcircles[ii].color = color;
   
   return;
}


//  draw current circles over the image in the window
//  called from mwpaint1()

void draw_topcircles()
{
   int      ii, px, py, rad, color;
   int      crflag = 0;

   if (! gdkWin) return;                                                   //  v.12.11

   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      cairo_set_line_width(mwcr,1);
      crflag = 1;
   }

   for (ii = 0; ii < Ntopcircles; ii++)
   {
      px = topcircles[ii].px * Mscale - Morgx + Dorgx;                     //  image to window space
      py = topcircles[ii].py * Mscale - Morgy + Dorgy;
      rad = topcircles[ii].radius;                                         //  radius is window space
      color = topcircles[ii].color;                                        //  1/2/3/4 = white/black/red/green

      if (color == 1) cairo_set_source_rgb(mwcr,1,1,1);                    //  set foreground color
      else if (color == 2) cairo_set_source_rgb(mwcr,0,0,0);
      else if (color == 3) cairo_set_source_rgb(mwcr,1,0,0);
      else cairo_set_source_rgb(mwcr,0,1,0);

      cairo_arc(mwcr,px,py,rad,0,2*pi);                                    //  draw 360 deg. arc
      cairo_stroke(mwcr);
   }
   
   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


//  erase top circles (next window repaint)

void erase_topcircles()
{
   Ntopcircles = 0;
   return;
}


/**************************************************************************/

//  Draw circle or ellipse around the mouse pointer.
//  Prior circle/ellipse will be erased first.
//  Used for mouse/brush radius in select/paint functions.
//  cx, cy, cww, chh: center and axis sizes in image space.
//  if cx == -1 then erase previous circle/ellipse only.

void draw_mousearc(int cx, int cy, int cww, int chh)                       //  overhauled   v.12.10
{
   int         px3, py3, ww3, hh3;
   static int  ppx3 = -1, ppy3, pww3, phh3;
   int         px, py, ww, hh, cc, rs;
   int         xlo, xhi, ylo, yhi;
   uint8       *pix1, *pix2, *pix1r, *pix2r;
   int         crflag = 0;
   GdkPixbuf   *pixbuf;

   float       fcx, fcy;
   float       a, b, a2, b2;
   float       x, y, x2, y2;
   float       r2 = 1.42;
   
   if (! gdkWin) return;                                                   //  v.12.11

   if (ppx3 >= 0) {                                                        //  erase prior circle
      mwpaint4(ppx3,ppy3,pww3,phh3);                                       //  refresh from Mpxb
      if (Fshowarea)
         sa_show_rect(ppx3,ppy3,pww3,phh3);                                //  refresh select area outline
      ppx3 = -1;
   }
   
   if (cx < 0) return;                                                     //  erase only was wanted

   px3 = cx - (cww + 1) / 2;                                               //  convert center + axis sizes
   py3 = cy - (chh + 1) / 2;                                               //    to block position, width, length
   ww3 = cww + 2;
   hh3 = chh + 2;
   
   if (px3 < 0) {                                                          //  v.12.10
      ww3 += px3;
      px3 = 0;
   }

   if (py3 < 0) {
      hh3 += py3;
      py3 = 0;
   }

   ppx3 = px3;                                                             //  remember block area for
   ppy3 = py3;                                                             //    for possible erase call
   pww3 = ww3;
   phh3 = hh3;

   px = Mscale * px3;                                                      //  1x image space to Mpxb space
   py = Mscale * py3;
   ww = Mscale * ww3;
   hh = Mscale * hh3;

   xlo = px;                                                               //  convert to x/y low/high limits
   ylo = py;
   xhi = xlo + ww;
   yhi = ylo + hh;

   if (xlo < Morgx) xlo = Morgx;                                           //  truncate to visible window
   if (ylo < Morgy) ylo = Morgy;
   if (xhi > Morgx + dww) xhi = Morgx + dww;
   if (yhi > Morgy + dhh) yhi = Morgy + dhh;

   ww = xhi - xlo;
   hh = yhi - ylo;

   if (ww <= 0 || hh <= 0) {
      ppx3 = -1;                                                           //  completely outside visible window
      return;
   }

   cc = ww * hh * 3;                                                       //  size of Mpxb area to copy

   pix1 = PXBpix(Mpxb,xlo,ylo);                                            //  pix1 block within Mpxb - source
   pix2 = (uint8 *) zmalloc(cc,"mousearc");                                //  allocate pix2 block - destination

   pix1r = pix1;                                                           //  row addresses
   pix2r = pix2;   

   for (py = 0; py < hh; py++)                                             //  copy pix1 block to pix2
   {
      memmove(pix2r,pix1r,ww*3);
      pix1r += Mpxb->rs;
      pix2r += ww * 3;
   }

   fcx = cx * Mscale;                                                      //  ellipse center within Mpxb
   fcy = cy * Mscale;
   fcx -= xlo;                                                             //  center within pix2 block
   fcy -= ylo;

   a = 0.5 * cww * Mscale - 1;                                             //  ellipse constants from
   b = 0.5 * chh * Mscale - 1;                                             //    enclosing rectangle
   a2 = a * a;
   b2 = b * b;

   for (y = -b/r2; y < b/r2; y++)                                          //  step through y values
   {
      y2 = y * y;
      x2 = a2 * (1 - y2 / b2);
      x = sqrt(x2);                                                        //  corresp. x values, + and -
      py = y + fcy + 0.5;
      if (py < 0 || py > hh-1) continue;
      px = fcx - x + 0.5;
      if (px >= 0) {
         pix1 = pix2 + (py * ww + px) * 3;
         pix1[0] = fg_color[0]; 
         pix1[1] = fg_color[1]; 
         pix1[2] = fg_color[2];
      }
      px = fcx + x + 0.5;
      if (px < ww) {
         pix1 = pix2 + (py * ww + px) * 3;
         pix1[0] = fg_color[0]; 
         pix1[1] = fg_color[1]; 
         pix1[2] = fg_color[2];
      }
   }

   for (x = -a/r2; x < a/r2; x++)                                          //  step through x values
   {
      x2 = x * x;
      y2 = b2 * (1 - x2 / a2);
      y = sqrt(y2);                                                        //  corresp. y values, + and -
      px = x + fcx + 0.5;
      if (px < 0 || px > ww-1) continue;
      py = fcy - y + 0.5;
      if (py >= 0) {
         pix1 = pix2 + (py * ww + px) * 3;
         pix1[0] = fg_color[0]; 
         pix1[1] = fg_color[1]; 
         pix1[2] = fg_color[2];
      }
      py = fcy + y + 0.5;
      if (py < hh) {
         pix1 = pix2 + (py * ww + px) * 3;
         pix1[0] = fg_color[0]; 
         pix1[1] = fg_color[1]; 
         pix1[2] = fg_color[2];
      }
   }

   if (! mwcr) {
      mwcr = gdk_cairo_create(gdkWin);                                     //  create cairo context if not already
      crflag = 1;
   }

   rs = ww * 3;                                                            //  pix2 row stride
   pixbuf = gdk_pixbuf_new_from_data(pix2,GDKRGB,0,8,ww,hh,rs,0,0);        //  pixbuf for image pix2
   
   px = xlo - Morgx + Dorgx;                                               //  area position in drawing window
   py = ylo - Morgy + Dorgy;

   gdk_cairo_set_source_pixbuf(mwcr,pixbuf,px,py);                         //  draw area to window
   cairo_paint(mwcr);

   g_object_unref(pixbuf);                                                 //  free memory
   zfree(pix2);
   
   if (Fshowarea)
      sa_show_rect(px3,py3,ww3,hh3);                                       //  refresh select area outline

   if (crflag) {
      cairo_destroy(mwcr);
      mwcr = 0;
   }

   return;
}


/**************************************************************************

   spline curve setup and edit functions

   Support multiple frames with curves in parallel
   (edit curve(s) and image mask curve)

   Usage:
   Add frame widget to dialog or zdialog.
   Set up drawing in frame: 
      sd = curve_init(frame, callback_func)
   Initialize no. of curves in frame (1-10): 
      sd->Nspc = n
   Initialize vert/horz flag for curve spc: 
      sd->vert[spc] = hv
   Initialize anchor points for curve spc: 
      sd->nap[spc], sd->apx[spc][xx], sd->apy[spc][yy]
   Generate data for curve spc:  
      splcurve_generate(sd,spc)
   Curves will now be shown and edited inside the frame when window is realized.
   The callback_func(spc) will be called when curve spc is edited.
   Change curve in program: 
      set anchor points, call splcurve_generate(sd,spc)
   Get y-value (0-1) for curve spc and given x-value (0-1):
      yval = splcurve_yval(sd,spc,xval)
   If faster access to curve is needed (no interpolation):
      kk = 1000 * xval;
      if (kk > 999) kk = 999;
      yval = sd->yval[spc][kk];

***/

//  initialize for spline curve editing
//  initial anchor points are pre-loaded into spldat before window is realized

spldat * splcurve_init(GtkWidget *frame, void func(int spc))
{
   int      cc = sizeof(spldat);                                           //  allocate spc curve data area
   spldat * sd = (spldat *) zmalloc(cc,"curvedat");
   memset(sd,0,cc);

   sd->drawarea = gtk_drawing_area_new();                                  //  drawing area for curves
   gtk_container_add(GTK_CONTAINER(frame),sd->drawarea);
   sd->spcfunc = func;                                                     //  user callback function

   gtk_widget_add_events(sd->drawarea,GDK_BUTTON_PRESS_MASK);              //  connect mouse events to drawing area
   gtk_widget_add_events(sd->drawarea,GDK_BUTTON_RELEASE_MASK);
   gtk_widget_add_events(sd->drawarea,GDK_BUTTON1_MOTION_MASK); 
   G_SIGNAL(sd->drawarea,"motion-notify-event",splcurve_adjust,sd);
   G_SIGNAL(sd->drawarea,"button-press-event",splcurve_adjust,sd);
   G_SIGNAL(sd->drawarea,"draw",splcurve_draw,sd);
   
   return sd;
}


//  modify anchor points in curve using mouse

int splcurve_adjust(void *, GdkEventButton *event, spldat *sd)
{
   int            ww, hh, kk;
   int            mx, my, button, evtype;
   static int     spc, ap, mbusy = 0, Fdrag = 0;                           //  drag continuation logic
   int            minspc, minap;
   float          mxval, myval, cxval, cyval;
   float          dist2, mindist2 = 0;
   
   mx = event->x;                                                          //  mouse position in drawing area
   my = event->y;
   evtype = event->type;
   button = event->button;

   if (evtype == GDK_MOTION_NOTIFY) {
      if (mbusy) return 0;                                                 //  discard excess motion events
      mbusy++;
      zmainloop();
      mbusy = 0;
   }

   if (evtype == GDK_BUTTON_RELEASE) {
      Fdrag = 0;
      return 0;
   }

   ww = gtk_widget_get_allocated_width(sd->drawarea);                      //  drawing area size
   hh = gtk_widget_get_allocated_height(sd->drawarea);
   
   if (mx < 0) mx = 0;                                                     //  limit edge excursions
   if (mx > ww) mx = ww;
   if (my < 0) my = 0;
   if (my > hh) my = hh;
   
   if (evtype == GDK_BUTTON_PRESS) Fdrag = 0;                              //  left or right click
   
   if (Fdrag)                                                              //  continuation of drag
   {
      if (sd->vert[spc]) {
         mxval = 1.0 * my / hh;                                            //  mouse position in curve space
         myval = 1.0 * mx / ww;
      }
      else {
         mxval = 1.0 * mx / ww;
         myval = 1.0 * (hh - my) / hh;
      }

      if (ap < sd->nap[spc]-1 && sd->apx[spc][ap+1] - mxval < 0.05)        //  disallow < 0.05 from next
            return 0;
      if (ap > 0 && mxval - sd->apx[spc][ap-1] < 0.05) return 0;           //    or prior anchor point 
   }
   
   else                                                                    //  mouse click or new drag begin
   {
      minspc = minap = -1;                                                 //  find closest curve/anchor point
      mindist2 = 999999;

      for (spc = 0; spc < sd->Nspc; spc++)                                 //  loop curves
      {
         if (sd->vert[spc]) {
            mxval = 1.0 * my / hh;                                         //  mouse position in curve space
            myval = 1.0 * mx / ww;
         }
         else {
            mxval = 1.0 * mx / ww;
            myval = 1.0 * (hh - my) / hh;
         }

         for (ap = 0; ap < sd->nap[spc]; ap++)                             //  loop anchor points
         {
            cxval = sd->apx[spc][ap];
            cyval = sd->apy[spc][ap];
            dist2 = (mxval-cxval)*(mxval-cxval) 
                  + (myval-cyval)*(myval-cyval);
            if (dist2 < mindist2) {
               mindist2 = dist2;                                           //  remember closest anchor point
               minspc = spc;
               minap = ap;
            }
         }
      }

      if (minspc < 0) return 0;                                            //  impossible
      spc = minspc;
      ap = minap;
   }
   
   if (evtype == GDK_BUTTON_PRESS && button == 3)                          //  right click, remove anchor point
   {
      if (sqrt(mindist2) > 0.05) return 0;                                 //  not close enough
      if (sd->nap[spc] < 3) return 0;                                      //  < 2 anchor points would remain
      sd->nap[spc]--;                                                      //  decr. before loop
      for (kk = ap; kk < sd->nap[spc]; kk++) {
         if (! kk) printf("meaningless reference %d",kk);                  //  stop gcc optimization bug   /////
         sd->apx[spc][kk] = sd->apx[spc][kk+1];
         sd->apy[spc][kk] = sd->apy[spc][kk+1];
      }
      splcurve_generate(sd,spc);                                           //  regenerate data for modified curve
      gtk_widget_queue_draw(sd->drawarea);
      sd->spcfunc(spc);                                                    //  call user function
      return 0;
   }

   if (! Fdrag)                                                            //  new drag or left click
   {
      if (sd->vert[spc]) {
         mxval = 1.0 * my / hh;                                            //  mouse position in curve space
         myval = 1.0 * mx / ww;
      }
      else {
         mxval = 1.0 * mx / ww;
         myval = 1.0 * (hh - my) / hh;
      }

      if (sqrt(mindist2) < 0.05)                                           //  existing point close enough,
      {                                                                    //    move this anchor point to mouse
         if (ap < sd->nap[spc]-1 && sd->apx[spc][ap+1] - mxval < 0.05)
               return 0;                                                   //  disallow < 0.05 from next
         if (ap > 0 && mxval - sd->apx[spc][ap-1] < 0.05) return 0;        //    or prior anchor point 
      }
   
      else                                                                 //  none close, add new anchor point
      {
         minspc = -1;                                                      //  find closest curve to mouse
         mindist2 = 999999;

         for (spc = 0; spc < sd->Nspc; spc++)                              //  loop curves
         {
            if (sd->vert[spc]) {
               mxval = 1.0 * my / hh;                                      //  mouse position in curve space
               myval = 1.0 * mx / ww;
            }
            else {
               mxval = 1.0 * mx / ww;
               myval = 1.0 * (hh - my) / hh;
            }

            cyval = splcurve_yval(sd,spc,mxval);
            dist2 = fabsf(myval - cyval);
            if (dist2 < mindist2) {
               mindist2 = dist2;                                           //  remember closest curve
               minspc = spc;
            }
         }

         if (minspc < 0) return 0;                                         //  impossible
         if (mindist2 > 0.05) return 0;                                    //  not close enough to any curve
         spc = minspc;

         if (sd->nap[spc] > 49) {
            zmessageACK(mWin,ZTX("Exceed 50 anchor points"));
            return 0;
         }

         if (sd->vert[spc]) {
            mxval = 1.0 * my / hh;                                         //  mouse position in curve space
            myval = 1.0 * mx / ww;
         }
         else {
            mxval = 1.0 * mx / ww;
            myval = 1.0 * (hh - my) / hh;
         }

         for (ap = 0; ap < sd->nap[spc]; ap++)                             //  find anchor point with next higher x
            if (mxval <= sd->apx[spc][ap]) break;                          //    (ap may come out 0 or nap)

         if (ap < sd->nap[spc] && sd->apx[spc][ap] - mxval < 0.05)         //  disallow < 0.05 from next
               return 0;                                                   //    or prior anchor point
         if (ap > 0 && mxval - sd->apx[spc][ap-1] < 0.05) return 0;

         for (kk = sd->nap[spc]; kk > ap; kk--) {                          //  make hole for new point
            sd->apx[spc][kk] = sd->apx[spc][kk-1];
            sd->apy[spc][kk] = sd->apy[spc][kk-1];
         }

         sd->nap[spc]++;                                                   //  up point count
      }
   }

   sd->apx[spc][ap] = mxval;                                               //  new or moved anchor point
   sd->apy[spc][ap] = myval;                                               //    at mouse position

   splcurve_generate(sd,spc);                                              //  regenerate data for modified curve
   gtk_widget_queue_draw(sd->drawarea);
   sd->spcfunc(spc);                                                       //  call user function
   
   if (evtype == GDK_MOTION_NOTIFY) Fdrag = 1;                             //  remember drag is underway
   return 0;
}


//  for expose event or when a curve is changed
//  draw all curves based on current anchor points

int splcurve_draw(GtkWidget *drawarea, cairo_t *cr, spldat *sd)
{
   int         ww, hh, spc, ap;
   float       xval, yval, px, py, qx, qy;
   
   ww = gtk_widget_get_allocated_width(sd->drawarea);                      //  drawing area size
   hh = gtk_widget_get_allocated_height(sd->drawarea);
   if (ww < 50 || hh < 50) return 0;

   cairo_set_line_width(cr,1);
   cairo_set_source_rgb(cr,0.7,0.7,0.7);

   for (int ii = 0; ii < sd->Nscale; ii++)                                 //  draw y-scale lines if any
   {
      px = ww * sd->xscale[0][ii];
      py = hh - hh * sd->yscale[0][ii];
      qx = ww * sd->xscale[1][ii];
      qy = hh - hh * sd->yscale[1][ii];
      cairo_move_to(cr,px,py);
      cairo_line_to(cr,qx,qy);
   }
   cairo_stroke(cr);

   cairo_set_source_rgb(cr,0,0,0);

   for (spc = 0; spc < sd->Nspc; spc++)                                    //  loop all curves
   {
      if (sd->vert[spc])                                                   //  vert. curve
      {
         for (py = 0; py < hh; py++)                                       //  generate all points for curve
         {
            xval = 1.0 * py / hh;
            yval = splcurve_yval(sd,spc,xval);
            px = ww * yval;
            if (py == 0) cairo_move_to(cr,px,py);
            cairo_line_to(cr,px,py);
         }
         cairo_stroke(cr);
         
         for (ap = 0; ap < sd->nap[spc]; ap++)                             //  draw boxes at anchor points
         {
            xval = sd->apx[spc][ap];
            yval = sd->apy[spc][ap];
            px = ww * yval;
            py = hh * xval;
            cairo_rectangle(cr,px-2,py-2,4,4);
         }
         cairo_fill(cr);
      }
      else                                                                 //  horz. curve
      {
         for (px = 0; px < ww; px++)                                       //  generate all points for curve
         {
            xval = 1.0 * px / ww;
            yval = splcurve_yval(sd,spc,xval);
            py = hh - hh * yval;
            if (px == 0) cairo_move_to(cr,px,py);
            cairo_line_to(cr,px,py);
         }
         cairo_stroke(cr);
         
         for (ap = 0; ap < sd->nap[spc]; ap++)                             //  draw boxes at anchor points
         {
            xval = sd->apx[spc][ap];
            yval = sd->apy[spc][ap];
            px = ww * xval;
            py = hh - hh * yval;
            cairo_rectangle(cr,px-2,py-2,4,4);
         }
         cairo_fill(cr);
      }
   }

   return 0;
}


//  generate all curve data points when anchor points are modified

int splcurve_generate(spldat *sd, int spc)
{
   int      kk, kklo, kkhi;
   float    xval, yvalx;

   spline1(sd->nap[spc],sd->apx[spc],sd->apy[spc]);                        //  compute curve fitting anchor points

   kklo = 1000 * sd->apx[spc][0] - 30;                                     //  xval range = anchor point range
   if (kklo < 0) kklo = 0;                                                 //    + 0.03 extra below/above
   kkhi = 1000 * sd->apx[spc][sd->nap[spc]-1] + 30;
   if (kkhi > 1000) kkhi = 1000;

   for (kk = 0; kk < 1000; kk++)                                           //  generate all points for curve
   {
      xval = 0.001 * kk;                                                   //  remove anchor point limits
      yvalx = spline2(xval);
      if (yvalx < 0) yvalx = 0;                                            //  yval < 0 not allowed, > 1 OK
      sd->yval[spc][kk] = yvalx;
   }

   return 0;
}


//  Retrieve curve data using interpolation of saved table of values

float splcurve_yval(spldat *sd, int spc, float xval)
{
   int      ii;
   float    x1, x2, y1, y2, y3;
   
   if (xval <= 0) return sd->yval[spc][0];
   if (xval >= 0.999) return sd->yval[spc][999];
   
   x2 = 1000.0 * xval;
   ii = x2;
   x1 = ii;
   y1 = sd->yval[spc][ii];
   y2 = sd->yval[spc][ii+1];
   y3 = y1 + (y2 - y1) * (x2 - x1);
   return y3;
}


//  load curve data from a file
//  returns 0 if fail (invalid file data), sd not modified
//  returns 1 if succcess, sd is initialized from file data

int splcurve_load(spldat *sd)
{
   char        *pfile;
   int         nn, ii, jj;
   FILE        *fid = 0;
   int         Nspc, vert[10], nap[10];
   float       apx[10][50], apy[10][50];

   zfuncs::F1_help_topic = "curve_edit";

   pfile = zgetfile1(ZTX("load curve from a file"),"open",saved_curves_dirk);
   if (! pfile) return 0;

   fid = fopen(pfile,"r");
   zfree(pfile);
   if (! fid) goto fail;
   
   nn = fscanf(fid,"%d ",&Nspc);                                           //  no. of curves
   if (nn != 1) goto fail;
   if (Nspc < 1 || Nspc > 10) goto fail;
   if (Nspc != sd->Nspc) goto fail2;

   for (ii = 0; ii < Nspc; ii++)                                           //  loop each curve
   {
      nn = fscanf(fid,"%d  %d ",&vert[ii],&nap[ii]);                       //  vertical flag, no. anchor points
      if (nn != 2) goto fail;
      if (vert[ii] < 0 || vert[ii] > 1) goto fail;
      if (nap[ii] < 2 || nap[ii] > 50) goto fail;

      for (jj = 0; jj < nap[ii]; jj++)                                     //  anchor point values 
      {
         nn = fscanf(fid,"%f/%f  ",&apx[ii][jj],&apy[ii][jj]);
         if (nn != 2) goto fail;
         if (apx[ii][jj] < 0 || apx[ii][jj] > 1) goto fail;
         if (apy[ii][jj] < 0 || apy[ii][jj] > 1) goto fail;
      }
   }
   
   fclose(fid);
   
   sd->Nspc = Nspc;                                                        //  copy curve data to caller's arg

   for (ii = 0; ii < Nspc; ii++)
   {
      sd->vert[ii] = vert[ii];
      sd->nap[ii] = nap[ii];
      
      for (jj = 0; jj < nap[ii]; jj++)
      {
         sd->apx[ii][jj] = apx[ii][jj];
         sd->apy[ii][jj] = apy[ii][jj];
      }
   }

   for (ii = 0; ii < Nspc; ii++)                                           //  generate curve data from anchor points
      splcurve_generate(sd,ii);
   
   if (sd->drawarea)                                                       //  regen and redraw all curves
      gtk_widget_queue_draw(sd->drawarea);

   return 1;
   
fail:
   if (fid) fclose(fid);
   zmessageACK(mWin,ZTX("curve file is invalid"));
   return 0;

fail2:
   if (fid) fclose(fid);
   zmessageACK(mWin,ZTX("curve file has different no. of curves"));
   return 0;
}


//  save curve data to a file

int splcurve_save(spldat *sd) 
{
   char        *pfile, *pp;
   int         ii, jj;
   FILE        *fid;

   zfuncs::F1_help_topic = "curve_edit";

   pp = zgetfile1(ZTX("save curve to a file"),"save",saved_curves_dirk);
   if (! pp) return 0;

   pfile = strdupz(pp,8,"splcurve");
   zfree(pp);

   pp = strrchr(pfile,'/');                                                //  force .curve extension
   if (pp) pp = strrchr(pp,'.');
   if (pp) strcpy(pp,".curve");
   else strcat(pfile,".curve");

   fid = fopen(pfile,"w");
   zfree(pfile);
   if (! fid) return 0;
   
   fprintf(fid,"%d \n",sd->Nspc);                                          //  no. of curves

   for (ii = 0; ii < sd->Nspc; ii++)                                       //  loop each curve
   {
      fprintf(fid,"%d  %d \n",sd->vert[ii],sd->nap[ii]);                   //  vertical flag, no. anchor points
      for (jj = 0; jj < sd->nap[ii]; jj++)                                 //  anchor point values 
         fprintf(fid,"%.4f/%.4f  ",sd->apx[ii][jj],sd->apy[ii][jj]);
      fprintf(fid,"\n");
   }
   
   fclose(fid);
   return 0;
}


/**************************************************************************

   edit transaction and thread support functions

   edit transaction management
      edit_setup()                     start new edit - copy E3 > E1
      edit_cancel()                    cancel edit - E1 > E3, delete E1
      edit_done()                      commit edit - add to undo stack
      edit_undo()                      undo edit - E1 > E3
      edit_redo()                      redo edit - run thread again
      edit_fullsize()                  convert preview to full-size pixmaps

   main level thread management
      start_thread(func,arg)           start thread running
      signal_thread()                  signal thread that work is pending
      wait_thread_idle()               wait for pending work complete
      wrapup_thread(command)           wait for exit or command thread exit

   thread function
      thread_idle_loop()               wait for pending work, exit if commanded
      thread_exit()                    exit thread unconditionally

   thread_status (thread ownership
      0     no thread is running
      1     thread is running and idle (no work)
      2     thread is working
      0     thread has exited

   thread_command (main program ownership)
      0     idle, no work pending
      8     exit when pending work is done
      9     exit now, unconditionally

   thread_pend       work requested counter
   thread_done       work done counter
   thread_hiwater    high water mark
   edit_action       done/cancel/undo/redo in progress
   
   bugfix v.12.08: refresh window and wait after changing E3pxm or Fpxb
                   (else draw_mousearc() -> mpaint3() may crash)

***************************************************************************/

int      thread_command = 0, thread_status = 0;
int      thread_pend = 0, thread_done = 0, thread_hiwater = 0;
int      edit_action = 0;


/**************************************************************************

  Setup for a new edit transaction
  Create E1 (edit input) and E3 (edit output) pixmaps from
  previous edit (E1) or image file (new E1).

  Fprev      0     edit full-size image
             1     edit preview image unless select area exists

  Farea      0     select_area is invalid and will be deleted (e.g. rotate)
             1     select_area not used but remains valid (e.g. red-eye)
             2     select_area can be used and remains valid (e.g. tune)

***************************************************************************/

int edit_setup(editfunc &EF)                                               //  v.12.10
{
   int      yn, rww, rhh;

   if (! curr_file) return 0;                                              //  no image file
   if (cTab != tabF) return 0;                                             //  could be Favorites menu          v.12.10
   
   if (CEF) {                                                              //  prior edit function active
      zmessageACK(mWin,ZTX("prior function still active"));                //  parallel edit removed            v.12.09
      return 0;
   }
   else if (checkmenulock()) return 0;                                     //  check no menu lock               v.12.09
   
   if (Pundo > maxedits-2) {                                               //  undo capacity reached
      zmessageACK(mWin,ZTX("Too many edits, please save image"));
      return 0;
   }
   
   sa_validate();                                                          //  delete area if not valid  v.12.12

   if (EF.Farea == 0 && sa_stat) {                                         //  select area will be lost, warn user
      yn = zmessageYN(mWin,ZTX("Select area cannot be kept.\n"
                               "Continue?"));
      if (! yn) return 0;
      sa_unselect();                                                       //  unselect area
      zdialog_free(zdsela);
   }
   
   if (EF.Farea == 2 && sa_stat && sa_stat != 3) {                         //  select area exists and can be used,
      yn = zmessageYN(mWin,ZTX("Select area not active.\n"                 //    but not active, ask user
                               "Continue?"));
      if (! yn) return 0;
   }

   if (! E0pxm) {                                                          //  first edit for this file
      E0pxm = PXM_load(curr_file,1);                                       //  get E0 image (poss. deeper color)
      if (! E0pxm) return 0;
      PXM_fixblue(E0pxm);                                                  //  blue=0 >> blue=1 for vpixel()
      curr_file_bpc = f_load_bpc;
   }

   if (! menulock(1)) return 0;                                            //  can now commit, lock menu
   
   if (Pundo == 0) save_undo("initial");                                   //  initial image >> undo stack

   mutex_lock(&Fpixmap_lock);                                              //  lock pixmaps

   Fpreview = 0;                                                           //  no preview

   if (EF.Fprev && ! Fzoom && ! (EF.Farea == 2 && sa_stat == 3))           //  use preview if wanted and no select area
      Fpreview = 1;

   if (E0pxm->ww < 1200 && E0pxm->hh < 1200)                               //  if image is small, don't use preview
      Fpreview = 0;                                                        //       v.12.10
   
   if (E0pxm->ww < 1.4 * Dww && E0pxm->hh < 1.4 * Dhh)                     //  if image slightly larger than window,
      Fpreview = 0;                                                        //       don't use preview    v.12.10
   
   if (Fpreview) {
      if (Fpxb->ww * Dhh > Fpxb->hh * Dww) {                               //  use preview image 1.4 * window size 
         rww = 1.4 * Dww;
         if (rww < 1200) rww = 1200;                                       //  at least 1200 on one side
         rhh = 1.0 * rww * Fpxb->hh / Fpxb->ww + 0.5;
      }
      else {
         rhh = 1.4 * Dhh;
         if (rhh < 1200) rhh = 1200;
         rww = 1.0 * rhh * Fpxb->ww / Fpxb->hh + 0.5;
      }
      if (rww > Fpxb->ww) Fpreview = 0;
   }
   
   if (Fpreview) {
      E1pxm = PXM_rescale(E0pxm,rww,rhh);                                  //  scale edit image to preview size
      sa_show(0);
   }
   else E1pxm = PXM_copy(E0pxm);                                           //  else use full size image

   E3pxm = PXM_copy(E1pxm);                                                //  E1 >> E3

   mutex_unlock(&Fpixmap_lock);

   Frefresh++;                                                             //  refresh window image
   gtk_widget_queue_draw(cDwin);
   zmainloop();                                                            //  must wait for completion

   CEF = &EF;                                                              //  set current edit function
   CEF->Fmod = 0;                                                          //  image not modified yet

   thread_command = thread_status = 0;                                     //  no thread running
   thread_pend = thread_done = thread_hiwater = 0;                         //  no work pending or done
   if (CEF->threadfunc) start_thread(CEF->threadfunc,0);                   //  start edit thread

   return 1;
}


/**************************************************************************/

//  process edit cancel

void edit_cancel()
{
   if (! CEF) return;                                                      //  v.12.12

   if (edit_action) return;
   edit_action++;                                                          //  stop reentry
   
   wrapup_thread(9);                                                       //  tell thread to quit now, wait
   
   if (CEF->mousefunc == mouseCBfunc)                                      //  if my mouse, free mouse
      freeMouse();
   if (CEF->zd) zdialog_free(CEF->zd);                                     //  kill dialog
   if (CEF->curves) zfree(CEF->curves);                                    //  free curves data
   CEF->zd = 0;
   CEF->curves = 0;

   mutex_lock(&Fpixmap_lock);
   PXM_free(E1pxm);                                                        //  free E1, E3, ER
   PXM_free(E3pxm);
   PXM_free(ERpxm);
   mutex_unlock(&Fpixmap_lock);
   
   Fzoom = 0;
   CEF = 0;                                                                //  no current edit func
   Fpreview = 0;                                                           //  no preview mode
   menulock(0);                                                            //  unlock menu

   Frefresh++;                                                             //  refresh window image
   gtk_widget_queue_draw(cDwin);
   zmainloop();                                                            //  must wait for completion

   edit_action = 0;
   return;
}   


/**************************************************************************/

//  process edit dialog [done]  
//  E3 >> E0

void edit_done()
{
   if (! CEF) return;                                                      //  v.12.12

   if (edit_action) return;
   edit_action++;                                                          //  stop reentry

   wait_thread_idle();                                                     //  wait for thread to finish

   if (Fpreview && CEF->Fmod) {                                            //  preview image was edited
      Fzoom = 0;
      edit_fullsize();                                                     //  update full image
   }
   wrapup_thread(8);                                                       //  tell thread to finish and exit

   if (CEF->mousefunc == mouseCBfunc)                                      //  if my mouse, free mouse
      freeMouse();
   if (CEF->zd) zdialog_free(CEF->zd);                                     //  kill dialog
   if (CEF->curves) zfree(CEF->curves);                                    //  free curves data
   CEF->zd = 0;
   CEF->curves = 0;

   mutex_lock(&Fpixmap_lock);

   if (CEF->Fmod) {                                                        //  image was modified
      Fmodified++;                                                         //  overall mod counter
      PXM_free(E0pxm);
      E0pxm = E3pxm;                                                       //  E3 updated image >> E0
      E3pxm = 0;
      PXM_free(E1pxm);                                                     //  free E1, ER
      PXM_free(ERpxm);
      Pundo++;
      Pumax = Pundo;
      save_undo(CEF->funcname);                                            //  save next undo state
   }
   
   else {                                                                  //  not modified
      PXM_free(E1pxm);                                                     //  free E1, E3, ER
      PXM_free(E3pxm);
      PXM_free(ERpxm);
   }

   mutex_unlock(&Fpixmap_lock);
   
   CEF = 0;                                                                //  no current edit func
   Fpreview = 0;                                                           //  no preview mode

   Frefresh++;                                                             //  refresh window image
   gtk_widget_queue_draw(cDwin);
   zmainloop();                                                            //  must wait for completion

   menulock(0);
   edit_action = 0;
   return;
}


/**************************************************************************/

//  edit undo, redo, reset functions
//  these apply within an active edit function

void edit_undo()
{
   if (! CEF) return;                                                      //  v.12.12

   if (thread_status == 2) return;                                         //  thread busy
   if (! CEF->Fmod) return;                                                //  not modified
   if (edit_action) return;
   edit_action++;                                                          //  stop reentry
   
   mutex_lock(&Fpixmap_lock);
   PXM_free(ERpxm);                                                        //  E3 >> redo copy
   ERpxm = E3pxm;
   E3pxm = PXM_copy(E1pxm);                                                //  E1 >> E3
   mutex_unlock(&Fpixmap_lock);

   CEF->Fmod = 0;                                                          //  reset image modified status

   Frefresh++;                                                             //  refresh window image
   gtk_widget_queue_draw(cDwin);
   zmainloop();                                                            //  must wait for completion

   edit_action = 0;
   return;
}


void edit_redo()
{
   if (! CEF) return;                                                      //  v.12.12

   if (thread_status == 2) return;                                         //  thread busy
   if (! ERpxm) return;                                                    //  no prior undo
   if (edit_action) return;
   edit_action++;                                                          //  stop reentry

   mutex_lock(&Fpixmap_lock);
   PXM_free(E3pxm);                                                        //  redo copy >> E3
   E3pxm = ERpxm;
   ERpxm = 0;
   mutex_unlock(&Fpixmap_lock);

   CEF->Fmod = 1;                                                          //  image modified

   Frefresh++;                                                             //  refresh window image
   gtk_widget_queue_draw(cDwin);
   zmainloop();                                                            //  must wait for completion

   edit_action = 0;
   return;
}


void edit_reset()
{
   if (! CEF) return;                                                      //  v.12.12

   if (thread_status == 2) return;                                         //  thread busy
   if (! CEF->Fmod) return;                                                //  not modified
   if (edit_action) return;
   edit_action++;                                                          //  stop reentry

   mutex_lock(&Fpixmap_lock);
   PXM_free(ERpxm);                                                        //  delete redo copy
   PXM_free(E3pxm);
   E3pxm = PXM_copy(E1pxm);                                                //  E1 >> E3
   mutex_unlock(&Fpixmap_lock);

   CEF->Fmod = 0;                                                          //  reset image modified status

   Frefresh++;                                                             //  refresh window image
   gtk_widget_queue_draw(cDwin);
   zmainloop();                                                            //  must wait for completion

   edit_action = 0;
   return;
}


/**************************************************************************/

//  Convert from preview mode (window-size pixmaps) to full-size pixmaps.
//  Can only be used when edit function is in an edit thread with idle_loop
//  This function is called directly from some edit functions

void edit_fullsize()
{
   if (! CEF) return;                                                      //  v.12.12
   if (! Fpreview) return;

   mutex_lock(&Fpixmap_lock);
   PXM_free(E1pxm);                                                        //  free preview pixmaps
   PXM_free(E3pxm);
   E1pxm = PXM_copy(E0pxm);                                                //  E0 >> E1, full size image
   E3pxm = PXM_copy(E1pxm);                                                //  E1 >> E3
   mutex_unlock(&Fpixmap_lock);

   Fpreview = 0;
   Fzoom = 0;

   if (! CEF->Fmod) return;                                                //  no change
   
   if (thread_status) {
      signal_thread();                                                     //  signal thread, repeat edit
      wait_thread_idle();
   }
   
   mwpaint2();                                                             //  v.12.12
   return;
}


/**************************************************************************
      undo / redo toolbar buttons
***************************************************************************/

//  [undo] menu function - reinstate previous edit in undo/redo stack
//  (toolbar button functions)

void m_undo(GtkWidget *, cchar *)
{
   if (CEF) {                                                              //  undo active edit
      edit_undo();
      return;
   }

   if (Pundo == 0) return;                                                 //  undo past edit
   if (! menulock(1)) return;
   Pundo--;
   if (KB_A_key) Pundo = 0;                                                //  if KB 'A' key, undo all
   load_undo();
   Fmodified = Pundo;
   menulock(0);
   return;
}


//  [redo] menu function - reinstate next edit in undo/redo stack

void m_redo(GtkWidget *, cchar *)
{
   if (CEF) {                                                              //  redo active edit
      edit_redo();
      return;
   }

   if (Pundo == Pumax) return;
   if (! menulock(1)) return;
   Pundo++;
   if (KB_A_key) Pundo = Pumax;                                            //  if KB 'A' key, redo all
   load_undo();
   Fmodified = Pundo;
   menulock(0);
   return;
}


//  undo all edits of the current image
//  (discard modifications)

void undo_all() 
{
   if (! menulock(1)) return;
   Pundo = 0;
   load_undo();
   Fmodified = 0;
   menulock(0);
   return;
}


//  Save E0 to undo/redo file stack
//  stack position = Pundo

void save_undo(cchar *funcname)
{
   char     *pp, buff[24];
   int      fid, cc, cc2;
   
   pp = strstr(undo_files,"_undo_");
   if (! pp) zappcrash("undo/redo stack corrupted 1");
   snprintf(pp+6,3,"%02d",Pundo);
   
   fid = open(undo_files,O_WRONLY|O_CREAT|O_TRUNC,0640);
   if (! fid) zappcrash("undo/redo stack corrupted 2");

   snprintf(buff,24," %05d %05d fotoxx ",E0pxm->ww,E0pxm->hh);
   cc = write(fid,buff,20);
   if (cc != 20) zappcrash("undo/redo stack corrupted 3");
   
   cc = E0pxm->ww * E0pxm->hh * 12;
   cc2 = write(fid,E0pxm->pixels,cc);
   if (cc2 != cc) zappcrash("undo/redo stack corrupted 4");

   close(fid);

   pvlist_replace(editlog,Pundo,funcname);                                 //  save log of edits done
   return;
}


//  Load E0 from undo/redo file stack
//  stack position = Pundo

void load_undo()
{
   char     *pp, buff[24], fotoxx[8];
   int      fid, ww, hh, cc, cc2;

   pp = strstr(undo_files,"_undo_");
   if (! pp) zappcrash("undo/redo stack corrupted 1");
   snprintf(pp+6,3,"%02d",Pundo);
   
   fid = open(undo_files,O_RDONLY);
   if (! fid) zappcrash("undo/redo stack corrupted 2");
   
   *fotoxx = 0;
   cc = read(fid,buff,20);
   sscanf(buff," %d %d %8s ",&ww, &hh, fotoxx);
   if (! strEqu(fotoxx,"fotoxx")) zappcrash("undo/redo stack corrupted 4");

   mutex_lock(&Fpixmap_lock);

   PXM_free(E0pxm);
   E0pxm = PXM_make(ww,hh);
   cc = ww * hh * 12;
   cc2 = read(fid,E0pxm->pixels,cc);
   if (cc2 != cc) zappcrash("undo/redo stack corrupted 5");
   close(fid);
   
   mutex_unlock(&Fpixmap_lock);
   
   sa_validate();                                                          //  delete area if invalid   v.12.12

   Frefresh++;                                                             //  refresh window image
   gtk_widget_queue_draw(cDwin);
   zmainloop();                                                            //  must wait for completion

   return;
}


/**************************************************************************/

//  ask user if modified image should be kept or discarded
//  returns: 0 = no mods or OK to discard,  1 = keep mods

int mod_keep()
{
   int      changes, choice;
   cchar    *title = ZTX("Discard edits?");
   cchar    *message = ZTX("This action will discard changes\n");          //  more clarity
   cchar    *keep = ZTX("Keep");
   cchar    *discard = ZTX("Discard");

   changes = 0;
   if (Fsaved < Pundo) changes = 1;                                        //  image edits not saved
   if (Fmetachanged) changes = 1;                                          //  metadata changes not saved   v.12.07
   if (! changes) return 0;                                                //  no mods
   choice = zdialog_choose(title,mWin,message,keep,discard,null);
   if (choice == 1) return 1;                                              //  keep mods
   if (Fsaved < Pundo) undo_all();                                         //  discard mods
   Fmetachanged = 0;
   return 0;
}


/**************************************************************************/

//  menu lock/unlock - some functions must not run concurrently
//  returns 1 if success, else 0

int menulock(int lock)
{
   if (! lock && ! Fmenulock) zappcrash("menu lock error");                //  unlock and not locked

   if (lock && Fmenulock) {
      zmessageACK(mWin,ZTX("prior function still active"));
      return 0;
   }

   if (lock) Fmenulock++;
   else Fmenulock--;
   return 1;
}


//  check menu lock and return 0/1 if unlocked/locked
//  if locked a user message will have been shown

int checkmenulock()
{
   if (Fmenulock) zmessageACK(mWin,ZTX("prior function still active"));
   return Fmenulock;
}

//  same but no message box

int checkmenulockquiet()
{
   return Fmenulock;
}


/**************************************************************************/

//  zdialog mouse capture and release

void takeMouse(CBfunc func, GdkCursor *cursor)                             //  capture mouse for dialog
{
   if (! gdkWin) return;                                                   //  v.12.11
   freeMouse();
   mouseCBfunc = func;
   Mcapture = 1;
   gdk_window_set_cursor(gdkWin,cursor);
   return;
}

void freeMouse()                                                           //  free mouse for main window
{
   if (! gdkWin) return;                                                   //  v.12.11
   if (! Mcapture) return;
   mouseCBfunc = 0;
   Mcapture = 0;
   draw_mousearc(-1,0,0,0);                                                //  remove mouse circle
   gdk_window_set_cursor(gdkWin,0);                                        //  set normal cursor
   return;
}


/**************************************************************************
   functions to manage working threads
***************************************************************************/

//  start thread that does the edit work

void start_thread(threadfunc func, void *arg)
{
   thread_status = 1;                                                      //  thread is running
   thread_command = thread_pend = thread_done = thread_hiwater = 0;        //  nothing pending
   start_detached_thread(func,arg);
   return;
}


//  signal thread that new work is pending

void signal_thread()
{
   if (thread_status > 0) thread_pend++;
   return;
}


//  wait for edit thread to complete pending work and become idle

void wait_thread_idle()
{
   while (thread_status && thread_pend > thread_done)
   {
      zmainloop();
      zsleep(0.01);
   }
   
   return;
}


//  wait for thread exit or command thread exit
//  command = 0    wait for normal completion
//            8    finish pending work and exit
//            9    quit, exit now

void wrapup_thread(int command)
{
   thread_command = command;                                               //  tell thread to quit or finish

   while (thread_status > 0)                                               //  wait for thread to finish
   {                                                                       //    pending work and exit
      zmainloop();
      zsleep(0.01);
   }

   return;
}


//  called only from edit threads
//  idle loop - wait for work request or exit command

void thread_idle_loop()
{
   if (thread_status == 2) Fthreadbusy--;
   thread_status = 1;                                                      //  status = idle
   thread_done = thread_hiwater;                                           //  work done = high-water mark

   while (true)
   {
      if (thread_command == 9) thread_exit();                              //  quit now command
      if (thread_command == 8)                                             //  finish work and exit
         if (thread_pend <= thread_done) thread_exit();
      if (thread_pend > thread_done) break;                                //  wait for work request
      zsleep(0.01);
   }
   
   thread_hiwater = thread_pend;                                           //  set high-water mark
   thread_status = 2;                                                      //  thread is working
   Fthreadbusy++;
   return;                                                                 //  perform edit
}


//  exit thread unconditionally, called only from edit threads

void thread_exit()
{
   if (thread_status == 2) Fthreadbusy--;
   thread_pend = thread_done = thread_hiwater = 0;
   thread_status = 0;
   pthread_exit(0);                                                        //  "return" cannot be used here
}


/**************************************************************************/

//  edit support functions for working threads (per processor core)

void start_wthread(threadfunc func, void *arg)                             //  start and increment busy count
{
   zadd_locked(Fthreadbusy,+1);
   zadd_locked(wthreads_busy,+1);
   start_detached_thread(func,arg);
   return;
}


void exit_wthread()                                                        //  decrement busy count and exit
{
   zadd_locked(Fthreadbusy,-1);
   zadd_locked(wthreads_busy,-1);
   pthread_exit(0);                                                        //  "return" cannot be used here
}


void wait_wthreads()                                                       //  wait for all working threads done
{                                                                          //  may be called from a thread
   while (wthreads_busy) {
      zsleep(0.01);
      zmainloop();
   }
   return;
}


/**************************************************************************
      other support functions
***************************************************************************/


//  favorites menu - popup graphic menu with user's favorites

void m_favorites(GtkWidget *, cchar *)                                     //  v.12.10
{
   void  favorites_callback(cchar *menu);
   char  menuconfigfile[200];
   
   zfuncs::F1_help_topic = "favorites";
   snprintf(menuconfigfile,200,"%s/menu-config",favorites_dirk);
   gmenuz(mWin,menuconfigfile,favorites_callback);
   return;
}


void favorites_callback(cchar *menu)
{
   int      ii;

   for (ii = 0; ii < Nmenus; ii++)
      if (strcasecmp(ZTX(menu),ZTX(menutab[ii].text)) == 0) break;
   
   if (ii < Nmenus) menutab[ii].func(0,menu);
   return;
}


//  help menu function

void m_help(GtkWidget *, cchar *menu)
{
   int      err;

   if (strEqu(menu,ZTX("User Guide"))) 
      showz_userguide();
   
   if (strEqu(menu,ZTX("User Guide Changes")))
      showz_userguide("changes");

   if (strEqu(menu,ZTX("Program Updates")))
      showz_userguide("updates");

   if (strEqu(menu,"README"))
      showz_doctext("README");

   if (strEqu(menu,ZTX("Edit Functions Summary")))
      showz_doctext("edit-menus");

   if (strEqu(menu,ZTX("Keyboard Shortcuts"))) {                           //  v.12.06
      sprintf(command,"xdg-open %s/KB-shortcuts",get_zuserdir());
      err = system(command);
      if (err) zmessageACK(mWin,"error: %s",wstrerror(err));
   }

   if (strEqu(menu,ZTX("Change Log")))
      showz_doctext("changelog");
   
   if (strEqu(menu,"Log File"))                                            //  v.12.04
      showz_logfile();

   if (strEqu(menu,ZTX("Translations")))
      showz_doctext("translations");
      
   if (strEqu(menu,ZTX("Home Page")))
      showz_html(fhomepage);

   if (strEqu(menu,ZTX("About")))                                          //  v.13.01
      zmessageACK(mWin,"Fotoxx",help_about);
      
   if (strEqu(menu,ZTX("Help")))                                           //  toolbar button
      showz_userguide(zfuncs::F1_help_topic);                              //  show topic if there, or page 1

   return;
}


/**************************************************************************/

//  table for loading and saving adjustable parameters between sessions

typedef struct {
   char     name[20];
   char     type[12];
   int      count;
   void     *location;
}  param;

#define Nparms 38

param paramTab[Nparms] = {
//     name                    type        count      location
{     "fotoxx version",       "char",        1,       &pversion               },
{     "first time",           "int",         1,       &Ffirsttime             },
{     "last sync_time",       "char",        1,       &last_sync_time         },
{     "file synch needed",    "int",         1,       &FsyncNeeded            },
{     "use file sync",        "int",         1,       &FsyncEnabled           },
{     "window geometry",      "int",         4,       &mwgeom                 },
{     "thumbs directory",     "char",        1,       &thumbdirk              },
{     "use thumbnails",       "int",         1,       &Fthumbnails            },
{     "thumbnail size",       "int",         1,       &navi::thumbsize        },
{     "menu style",           "char",        1,       &menu_style             },
{     "toolbar style",        "char",        1,       &tbar_style             },
{     "drag option",          "int",         1,       &Fdragopt               },
{     "warn overwrite",       "int",         1,       &Fwarnoverwrite         },
{     "zoom count",           "int",         1,       &zoomcount              },
{     "startup display",      "char",        1,       &startdisplay           },
{     "start image file",     "char",        1,       &startfile              },
{     "start directory",      "char",        1,       &startdirk              },
{     "RAW command",          "char",        1,       &RAWcommand             },
{     "RAW file types",       "char",        1,       &RAWfiletypes           },
{     "trim size",            "int",         2,       &trimsize               },
{     "trim buttons",         "char",        6,       &trimbuttons            },
{     "trim ratios",          "char",        6,       &trimratios             },
{     "edit resize",          "int",         2,       &editresize             },
{     "write text font",      "char",        1,       &writetext_font         },
{     "write text text",      "char",        1,       &writetext_text         },
{     "write text color",     "char",        3,       &writetext_color        },
{     "write text trans",     "int",         3,       &writetext_trans        },
{     "write text outline",   "int",         1,       &writetext_outline      },
{     "write text angle",     "float",       1,       &writetext_angle        },
{     "grid spacing",         "int",         2,       &gridspace              },
{     "grid counts",          "int",         2,       &gridcount              },
{     "rotate grid",          "int",         8,       &rotate_grid            },
{     "unbend grid",          "int",         8,       &unbend_grid            },
{     "warp-curved grid",     "int",         8,       &warpC_grid             },
{     "warp-linear grid",     "int",         8,       &warpL_grid             },
{     "warp-affine grid",     "int",         8,       &warpF_grid             },
{     "lens mm",              "float",       1,       &lens_mm                },
{     "lens bow",             "float",       1,       &lens_bow               }  };


//  save parameters to file /.../.fotoxx/parameters

void save_params()
{
   FILE        *fid;
   char        buff[1050], text[1000];                                     //  limit for character data cc
   char        *name, *type;
   int         count;
   void        *location;
   char        **charloc;
   int         *intloc;
   float       *floatloc;
   
   pversion = strdupz(fotoxxvers,0,"parameters");                          //  this fotoxx version
   
   snprintf(buff,199,"%s/parameters",get_zuserdir());                      //  open output file
   fid = fopen(buff,"w");
   if (! fid) return;
   
   for (int ii = 0; ii < Nparms; ii++)                                     //  write table of state data
   {
      name = paramTab[ii].name;
      type = paramTab[ii].type;
      count = paramTab[ii].count;
      location = paramTab[ii].location;
      charloc = (char **) location;
      intloc = (int *) location;
      floatloc = (float *) location;

      fprintf(fid,"%-20s  %-8s  %02d  ",name,type,count);                  //  write parm name, type, count

      for (int kk = 0; kk < count; kk++)                                   //  write "value" "value" ...
      {
         if (strEqu(type,"char")) {
            if (! *charloc) break;                                         //  missing, discontinue this parameter
            repl_1str(*charloc++,text,"\n","\\n");                         //  replace newlines with "\n"
            fprintf(fid,"  \"%s\"",text);
         }
         if (strEqu(type,"int"))
            fprintf(fid,"  \"%d\"",*intloc++);

         if (strEqu(type,"float"))
            fprintf(fid,"  \"%.2f\"",*floatloc++);
      }

      fprintf(fid,"\n");                                                   //  write EOR
   }
   
   fprintf(fid,"\n");
   fclose(fid);                                                            //  close file

   fid = fopen(recentfiles_file,"w");                                      //  recent files file
   if (! fid) return;
   
   for (int ii = 0; ii < Nrecentfiles; ii++)                               //  save list of recent files
      if (recentfiles[ii])
         fprintf(fid,"%s \n",recentfiles[ii]);
   
   fclose(fid);

   snprintf(buff,199,"%s/plugins",get_zuserdir());                         //  open file for plugins
   fid = fopen(buff,"w");
   if (! fid) return;
   
   for (int ii = 0; ii < Nplugins; ii++)                                   //  save plugins
      if (plugins[ii])
         fprintf(fid,"%s \n",plugins[ii]);
   
   fclose(fid);

   return;
}


//  load parameters from file /.../.fotoxx/parameters

void load_params()
{
   FILE        *fid;
   int         ii, err, cc, pcount;
   int         Idata;
   float       Fdata;
   char        buff[1000], text[1000], *pp;
   char        name[20], type[12], count[8], data[1000];
   void        *location;
   char        **charloc;
   int         *intloc;
   float       *floatloc;
   struct stat statb;
   
   snprintf(buff,499,"%s/parameters",get_zuserdir());                      //  open parameters file
   fid = fopen(buff,"r");
   if (! fid) return;                                                      //  none, defaults are used

   while (true)                                                            //  read parameters
   {
      pp = fgets_trim(buff,999,fid,1);
      if (! pp) break;
      if (*pp == '#') continue;                                            //  comment
      if (strlen(pp) < 40) continue;                                       //  rubbish

      err = 0;

      strncpy0(name,pp,20);                                                //  parm name
      strTrim2(name);
      
      strncpy0(type,pp+22,8);                                              //  parm type
      strTrim2(type);
      
      strncpy0(count,pp+32,4);                                             //  parm count
      strTrim2(count);
      err = convSI(count,pcount);
      
      strncpy0(data,pp+38,1000);                                           //  parm value(s)
      strTrim2(data);
      
      for (ii = 0; ii < Nparms; ii++)                                      //  match file record to param table
      {
         if (strNeq(name,paramTab[ii].name)) continue;                     //  parm name
         if (strNeq(type,paramTab[ii].type)) continue;                     //  parm type
         if (paramTab[ii].count != pcount) continue;                       //  parm count
         break;
      }
      
      if (ii == Nparms) continue;                                          //  not found, ignore file param
      
      location = paramTab[ii].location;                                    //  get parameter memory location
      charloc = (char **) location;
      intloc = (int *) location;
      floatloc = (float *) location;
      
      for (ii = 1; ii <= pcount; ii++)                                     //  get parameter value(s)
      {
         pp = (char *) strField(data,' ',ii);
         if (! pp) break;

         if (strEqu(type,"char")) {
            repl_1str(pp,text,"\\n","\n");                                 //  replace "\n" with real newlines
            *charloc++ = strdupz(text,0,"parameters");
         }

         if (strEqu(type,"int")) {
            err = convSI(pp,Idata);
            if (err) continue;
            *intloc++ = Idata;
         }

         if (strEqu(type,"float")) {
            err = convSF(pp,Fdata);
            if (err) continue;
            *floatloc++ = Fdata;
         }
      }
   }

   fclose(fid);
   
   for (ii = 0; ii < Nparms; ii++)                                         //  set null strings to "undefined"
   {
      if (strNeq(paramTab[ii].type,"char")) continue;
      charloc = (char **) paramTab[ii].location;
      pcount = paramTab[ii].count;
      for (int jj = 0; jj < pcount; jj++)
         if (! charloc[jj]) charloc[jj] = strdupz("undefined",20,"parameters");
   }
   
   if (strNeq(pversion,fotoxxvers))                                        //  fotoxx version change 
      printf("new Fotoxx version %s %s \n",pversion,fotoxxvers);

   Ntopdirk = 0;

   fid = fopen(topdirk_file,"r");                                          //  get top directories from
   if (fid)                                                                //    from top directories file
   {
      for (ii = 0; ii < 10; ii++)                                          //  multiple top directories
      {
         pp = fgets_trim(buff,499,fid,1);
         if (! pp) break;
         topdirk[ii] = strdupz(buff,0,"top_dirk");
         strTrim2(topdirk[ii]);                                            //  remove surrounding blanks
         cc = strlen(topdirk[ii]);                                         //  remove trailing '/' if present
         if (topdirk[ii][cc-1] == '/') topdirk[ii][cc-1] = 0;
         err = stat(topdirk[ii],&statb);
         if (err || ! S_ISDIR(statb.st_mode)) {
            printf("*** invalid top directory: %s \n",topdirk[ii]);        //  if invalid top directory,
            ii = ii - 1;                                                   //    remove from list
         }
      }

      fclose(fid);
      Ntopdirk = ii;
   }
   
   zoomratio = pow( 2.0, 1.0 / zoomcount);                                 //  set zoom ratio from zoom count

   for (ii = 0; ii < Nrecentfiles; ii++)                                   //  recent image file list = empty
      recentfiles[ii] = 0;

   fid = fopen(recentfiles_file,"r");                                      //  open recent files file
   if (fid) {
      for (ii = 0; ii < Nrecentfiles; ii++)                                //  read list of recent files
      {
         pp = fgets_trim(buff,499,fid,1);
         if (! pp) break;
         if (*pp == '/') 
            recentfiles[ii] = strdupz(buff,0,"recentfile");
         else ii--; 
      }

      fclose(fid);
   }

   for (ii = 0; ii < maxplugins; ii++)                                     //  plugins list = empty
      plugins[ii] = 0;
   
   plugins[0] = strdupz("Gimp = gimp",0,"plugins");                        //  if empty, default Gimp plugin
   Nplugins = 1;

   snprintf(buff,499,"%s/plugins",get_zuserdir());                         //  open plugins file
   fid = fopen(buff,"r");
   if (fid)
   {   
      for (ii = 0; ii < maxplugins; ii++)                                  //  read list of plugins
      {
         pp = fgets_trim(buff,499,fid,1);
         if (! pp) break;
         plugins[ii] = strdupz(buff,0,"plugins");
      }

      fclose(fid);
      Nplugins = ii;
   }

   return;
}


/**************************************************************************/

//  free all resources associated with the current image file

void free_resources(int fkeepundo)
{
   char        *pp;
   int         err;

   mutex_lock(&Fpixmap_lock);                                              //  lock pixmaps
   
   if (! fkeepundo) {
      strcpy(command,"rm -f ");                                            //  delete all undo files
      strcat(command,undo_files);
      pp = strstr(command,"_undo_");                                       //  /.../.fotoxx/pppppp_undo_*
      strcpy(pp + 6,"*");
      err = system(command);
      if (err) printf("*** %s \n",wstrerror(err));
      Fmodified = Pundo = Pumax = Fsaved = 0;                              //  reset undo/redo stack
      exif_editlog_cc = -1;                                                //  EXIF edit hist not loaded    v.12.12
      sa_unselect();                                                       //  unselect select area         v.12.12
   }

   if (Fshutdown) {                                                        //  stop here if shutdown mode
      mutex_unlock(&Fpixmap_lock);
      return;
   }
   
   draw_toplines(2);                                                       //  no toplines            v.12.09
   erase_topcircles();                                                     //  no topcircles          v.12.09
   freeMouse();                                                            //  free mouse
   Ffuncbusy = Fthreadbusy = 0;                                            //  just in case
   
   if (curr_file) {
      zdialog_free(zdsela);                                                //  kill select area dialog if active
      zfree(curr_file);                                                    //  free image file
      curr_file = 0;
      Fbusy_goal = 0;
      *SB_text = 0;
   }

   if (Fpxb != Wpxb) PXB_free(Fpxb);                                       //  must not be world map
   PXM_free(E0pxm);
   PXM_free(E1pxm);
   PXM_free(E3pxm);
   PXM_free(ERpxm);
   PXM_free(E8pxm);
   PXM_free(E9pxm);

   mutex_unlock(&Fpixmap_lock);
   return;
}


/**************************************************************************/

//  Get a virtual pixel at location (px,py) (real) in a PXM pixmap.
//  Get the overlapping real pixels and build a composite.

int vpixel(PXM *pxm, float px, float py, float *vpix)
{
   int            ww, hh, px0, py0;
   float          *pix0, *pix1, *pix2, *pix3;
   float          f0, f1, f2, f3;
   float          red, green, blue;
   
   ww = pxm->ww;
   hh = pxm->hh;

   px0 = px;                                                               //  integer pixel containing (px,py)
   py0 = py;

   if (px0 < 1 || py0 < 1) return 0;                                       //  void edge pixels
   if (px0 > ww-2 || py0 > hh-2) return 0;
   
   pix0 = PXMpix(pxm,px0,py0);                                             //  4 pixels based at (px0,py0)
   pix1 = pix0 + ww * 3;
   pix2 = pix0 + 3;
   pix3 = pix0 + ww * 3 + 3;

   f0 = (px0+1 - px) * (py0+1 - py);                                       //  overlap of (px,py)
   f1 = (px0+1 - px) * (py - py0);                                         //   in each of the 4 pixels
   f2 = (px - px0) * (py0+1 - py);
   f3 = (px - px0) * (py - py0);
   
   red =   f0 * pix0[0] + f1 * pix1[0] + f2 * pix2[0] + f3 * pix3[0];      //  sum the weighted inputs
   green = f0 * pix0[1] + f1 * pix1[1] + f2 * pix2[1] + f3 * pix3[1];
   blue =  f0 * pix0[2] + f1 * pix1[2] + f2 * pix2[2] + f3 * pix3[2];
   if (blue < 0.5) blue = 0;                                               //  edge or voided pixel      v.12.09
   
   vpix[0] = red;
   vpix[1] = green;
   vpix[2] = blue;
   
   if (! blue) return 0;
   return 1;
}


/**************************************************************************/

//  compare two floats for significant difference
//  return:  0  difference not significant
//          +1  d1 > d2
//          -1  d1 < d2

int sigdiff(float d1, float d2, float signf)
{
   float diff = fabsf(d1-d2);
   if (diff == 0.0) return 0;
   diff = diff / (fabsf(d1) + fabsf(d2));
   if (diff < signf) return 0;
   if (d1 > d2) return 1;
   else return -1;
}


/**************************************************************************/

//  popup a warning message if real memory is below specified threshold
//  returns margin: real memory - threshold, in megabytes 
//  (+ if enough memory, - if not)
//  the popup message only appears if real memory < threshold
//  the popup message goes away after a few seconds

int warn_memory(int req_mb)                                                //  v.12.11
{
   double   freemem, cachemem;
   int      act_mb, gap_mb;
   cchar    *message = ZTX("warning: real memory %d MB < required %d MB \n"
                           "response time may become very long \n");

   parseprocfile("/proc/meminfo","MemFree:",&freemem,null);                //  get current free memory
   parseprocfile("/proc/meminfo","Cached:",&cachemem,null);
   act_mb = (freemem + cachemem) / 1024;                                   //  megabytes
   gap_mb = act_mb - req_mb;
   if (gap_mb >= 0) return gap_mb;
   printf(message,act_mb,req_mb);
   zmessage_post(mWin,10,message,act_mb,req_mb,null);                      //  post for 10 seconds and delete
   for (int ii = 0; ii < 10; ii++) zmainloop();                            //  (returns immediately)
   return gap_mb;                                      
}


/**************************************************************************

      PXM pixmap functions - RGB float pixel map

***************************************************************************/

//  initialize PXM pixmap - allocate memory

PXM * PXM_make(int ww, int hh)
{
   if (ww < 5 || hh < 5) {                                                 //  impose reasonableness limits    v.12.10
      zmessageACK(mWin,"image too small: %d %d",ww,hh);
      return 0;
   }

   if (ww > wwhh_limit || hh > wwhh_limit) {
      zmessageACK(mWin,"image too big: %d %d",ww,hh);
      return 0;
   }

   int req_mb = ww * hh * 12 / mega;                                       //  warn if memory low              v.12.11
   warn_memory(req_mb);

   PXM *pxm = (PXM *) zmalloc(sizeof(PXM),"PXM");                          //  0.45 sec for 116 megapix image (1.4 GB)
   pxm->ww = ww;
   pxm->hh = hh;
   pxm->pixels = (float *) zmalloc(ww*hh*12,"PXM.bmp");                    //  float[3] per pixel
   strcpy(pxm->wmi,"pxmpix");
   return pxm;
}


//  free PXM pixmap

void PXM_free(PXM *&pxm)
{
   if (! pxm) return;
   if (! strEqu(pxm->wmi,"pxmpix")) 
      zappcrash("PXM_free(), bad PXM");
   strcpy(pxm->wmi,"xxxxxx");
   zfree(pxm->pixels);
   zfree(pxm);
   pxm = 0;
   return;
}


//  create a copy of a PXM pixmap

PXM * PXM_copy(PXM *pxm1)
{
   int      cc;
   PXM      *pxm2;

   pxm2 = PXM_make(pxm1->ww, pxm1->hh);
   cc = pxm1->ww * pxm1->hh * 3 * sizeof(float);
   memcpy(pxm2->pixels,pxm1->pixels,cc);
   return pxm2;
}


//  create a copy of a PXM area

PXM * PXM_copy_area(PXM *pxm1, int orgx, int orgy, int ww2, int hh2)
{
   float    *pix1, *pix2;
   PXM      *pxm2 = 0;
   int      px1, py1, px2, py2;

   pxm2 = PXM_make(ww2,hh2);
  
   for (py1 = orgy, py2 = 0; py2 < hh2; py1++, py2++) 
   {
      pix1 = PXMpix(pxm1,orgx,orgy);                                       //  v.12.10
      pix2 = PXMpix(pxm2,0,py2);

      for (px1 = orgx, px2 = 0; px2 < ww2; px1++, px2++)
      {
         pix2[0] = pix1[0];
         pix2[1] = pix1[1];
         pix2[2] = pix1[2];
         pix1 += 3;
         pix2 += 3;
      }
   }

   return pxm2;
}


//  replace blue = 0 pixels with blue = 1
//  (blue = 0 reserved for pixels voided by warp or overlay offsets)
//  needs about 0.013 seconds for 10 megapixel image and 3.3 GHz CPU

void PXM_fixblue(PXM *pxm)
{
   int      size;
   float    *pixel, *pixel0, *pixelN;

   size = pxm->ww * pxm->hh * 3;
   pixel0 = pxm->pixels + 2;
   pixelN = pixel0 + size;

   for (pixel = pixel0; pixel < pixelN; pixel += 3)                        //  +3 is +12 bytes
      if (! *pixel) *pixel = 1.0;

   return;
}


/**************************************************************************

   Rescale PXM image to new width and height.
   The scale ratios may be different for width and height.

   Method: 
   The input and output images are overlayed, stretching or shrinking the
   output pixels as needed. The contribution of each input pixel overlapping
   an output pixel is proportional to the area of the output pixel covered by
   the input pixel. The contributions of all overlaping input pixels are added.
   The work is spread among Nwt threads to reduce time on SMP processors.

   Example: if the output image is 40% of the input image, then:
     outpix[0,0] = 0.16 * inpix[0,0] + 0.16 * inpix[1,0] + 0.08 * inpix[2,0]
                 + 0.16 * inpix[0,1] + 0.16 * inpix[1,1] + 0.08 * inpix[2,1]
                 + 0.08 * inpix[0,2] + 0.08 * inpix[1,2] + 0.04 * inpix[2,2]

***************************************************************************/

namespace pxmrescale {
   float    *bmp1;
   float    *bmp2;
   int      ww1;
   int      hh1;
   int      ww2;
   int      hh2;
   int      *px1L;
   int      *py1L;
   float    *pxmap;
   float    *pymap;
   int      maxmapx;
   int      maxmapy;
   int      busy[max_threads];
}

PXM * PXM_rescale(PXM *pxm1, int ww, int hh)
{
   using namespace pxmrescale;

   void * pxm_rescale_thread(void *arg);

   PXM         *pxm2;
   int         px1, py1, px2, py2;
   int         pxl, pyl, pxm, pym, ii;
   float       scalex, scaley;
   float       px1a, py1a, px1b, py1b;
   float       fx, fy;

   ww1 = pxm1->ww;                                                         //  input PXM
   hh1 = pxm1->hh;
   bmp1 = pxm1->pixels;
   
   pxm2 = PXM_make(ww,hh);                                                 //  output PXM
   if (! pxm2) return 0;                                                   //  v.12.11
   ww2 = ww;
   hh2 = hh;
   bmp2 = pxm2->pixels;
   
   memset(bmp2, 0, ww2 * hh2 * 3 * sizeof(float));                         //  clear output pixmap

   scalex = 1.0 * ww1 / ww2;                                               //  compute x and y scales
   scaley = 1.0 * hh1 / hh2;
   
   if (scalex <= 1) maxmapx = 2;                                           //  compute max input pixels
   else maxmapx = scalex + 2;                                              //    mapping into output pixels
   maxmapx += 1;                                                           //      for both dimensions
   if (scaley <= 1) maxmapy = 2;                                           //  (pixels may not be square)
   else maxmapy = scaley + 2;
   maxmapy += 1;                                                           //  (extra entry for -1 flag)
   
   pymap = (float *) zmalloc(hh2 * maxmapy * sizeof(float),"PXMrescale");  //  maps overlap of < maxmap input
   pxmap = (float *) zmalloc(ww2 * maxmapx * sizeof(float),"PXMrescale");  //    pixels per output pixel

   py1L = (int *) zmalloc(hh2 * sizeof(int),"PXMrescale");                 //  maps first (lowest) input pixel
   px1L = (int *) zmalloc(ww2 * sizeof(int),"PXMrescale");                 //    per output pixel

   for (py2 = 0; py2 < hh2; py2++)                                         //  loop output y-pixels
   {
      py1a = py2 * scaley;                                                 //  corresponding input y-pixels
      py1b = py1a + scaley;
      if (py1b >= hh1) py1b = hh1 - 0.001;                                 //  fix precision limitation
      pyl = py1a;
      py1L[py2] = pyl;                                                     //  1st overlapping input pixel

      for (py1 = pyl, pym = 0; py1 < py1b; py1++, pym++)                   //  loop overlapping input pixels
      {
         if (py1 < py1a) {                                                 //  compute amount of overlap
            if (py1+1 < py1b) fy = py1+1 - py1a;                           //    0.0 to 1.0 
            else fy = scaley;
         }
         else if (py1+1 > py1b) fy = py1b - py1;
         else fy = 1;

         ii = py2 * maxmapy + pym;                                         //  save it
         pymap[ii] = 0.9999 * fy / scaley;
      }
      ii = py2 * maxmapy + pym;                                            //  set an end marker after
      pymap[ii] = -1;                                                      //    last overlapping pixel
   }
   
   for (px2 = 0; px2 < ww2; px2++)                                         //  do same for x-pixels
   {
      px1a = px2 * scalex;
      px1b = px1a + scalex;
      if (px1b >= ww1) px1b = ww1 - 0.001;
      pxl = px1a;
      px1L[px2] = pxl;

      for (px1 = pxl, pxm = 0; px1 < px1b; px1++, pxm++)
      {
         if (px1 < px1a) {
            if (px1+1 < px1b) fx = px1+1 - px1a;
            else fx = scalex;
         }
         else if (px1+1 > px1b) fx = px1b - px1;
         else fx = 1;

         ii = px2 * maxmapx + pxm;
         pxmap[ii] = 0.9999 * fx / scalex;
      }
      ii = px2 * maxmapx + pxm;
      pxmap[ii] = -1;
   }

   for (ii = 0; ii < Nwt; ii++) {                                          //  start working threads
      busy[ii] = 1;
      start_detached_thread(pxm_rescale_thread,&wtnx[ii]);
   }
   
   for (ii = 0; ii < Nwt; ii++)                                            //  wait for all done
      while (busy[ii]) zsleep(0.004);

   zfree(px1L);
   zfree(py1L);
   zfree(pxmap);
   zfree(pymap);

   return pxm2;
}

void * pxm_rescale_thread(void *arg)                                       //  worker thread function
{
   using namespace pxmrescale;

   int         index = *((int *) arg);
   int         px1, py1, px2, py2;
   int         pxl, pyl, pxm, pym, ii;
   float       *pixel1, *pixel2;
   float       fx, fy, ftot;
   float       red, green, blue;

   for (py2 = index; py2 < hh2; py2 += Nwt)                                //  loop output y-pixels
   {
      pyl = py1L[py2];                                                     //  corresp. 1st input y-pixel

      for (px2 = 0; px2 < ww2; px2++)                                      //  loop output x-pixels
      {
         pxl = px1L[px2];                                                  //  corresp. 1st input x-pixel

         red = green = blue = 0;                                           //  initz. output pixel

         for (py1 = pyl, pym = 0; ; py1++, pym++)                          //  loop overlapping input y-pixels
         {
            ii = py2 * maxmapy + pym;                                      //  get y-overlap
            fy = pymap[ii];
            if (fy < 0) break;                                             //  no more pixels

            for (px1 = pxl, pxm = 0; ; px1++, pxm++)                       //  loop overlapping input x-pixels
            {
               ii = px2 * maxmapx + pxm;                                   //  get x-overlap
               fx = pxmap[ii];
               if (fx < 0) break;                                          //  no more pixels

               ftot = fx * fy;                                             //  area overlap = x * y overlap
               pixel1 = bmp1 + (py1 * ww1 + px1) * 3;
               red += pixel1[0] * ftot;                                    //  add input pixel * overlap
               green += pixel1[1] * ftot;
               blue += pixel1[2] * ftot;
            }

            pixel2 = bmp2 + (py2 * ww2 + px2) * 3;                         //  save output pixel
            pixel2[0] = red;
            pixel2[1] = green;
            pixel2[2] = blue;
         }
      }
   }

   busy[index] = 0;
   return 0;
}


/**************************************************************************

      PXM *pxm2 = PXM_rotate(PXM *pxm1, float angle)

      Rotate PXM pixmap through an arbitrary angle (degrees).

      The returned image has the same size as the original, but the
      pixmap size is increased to accomodate the rotated image.
      (e.g. a 100x100 image rotated 45 deg. needs a 142x142 pixmap).

      The space added around the rotated image is black (RGB 0,0,0).
      Angle is in degrees. Positive direction is clockwise.
      Speed is about 28 million pixels/sec/thread for a 3.3 GHz CPU.
      Loss of resolution is less than 1 pixel.
      
      Work is divided among Nwt threads to gain speed.
      
***************************************************************************/

namespace pxmrotate {
   int      busy = 0;
   float    *bmp1;
   float    *bmp2;
   int      ww1;
   int      hh1;
   int      ww2;
   int      hh2;
   float    angle;
}

PXM * PXM_rotate(PXM *pxm1, float anglex)
{
   using namespace pxmrotate;

   void     *PXM_rotate_thread(void *);

   int      ii;
   PXM      *pxm2;
   
   ww1 = pxm1->ww;                                                         //  input PXM
   hh1 = pxm1->hh;
   bmp1 = pxm1->pixels;
   angle = anglex;

   while (angle < -180) angle += 360;                                      //  normalize, -180 to +180
   while (angle > 180) angle -= 360;
   angle = angle * pi / 180;                                               //  radians, -pi to +pi
   
   if (fabsf(angle) < 0.001) {                                             //  angle = 0 within my precision
      pxm2 = PXM_copy(pxm1);                                               //  return a copy of the input PXM
      return pxm2;
   }

   ww2 = ww1*fabsf(cosf(angle)) + hh1*fabsf(sinf(angle));                  //  rectangle containing rotated image
   hh2 = ww1*fabsf(sinf(angle)) + hh1*fabsf(cosf(angle));
   
   pxm2 = PXM_make(ww2,hh2);                                               //  output PXM
   bmp2 = pxm2->pixels;

   for (ii = 0; ii < Nwt; ii++)                                            //  start worker threads
      start_detached_thread(PXM_rotate_thread,&wtnx[ii]);
   zadd_locked(busy,+Nwt);

   while (busy) zsleep(0.004);                                             //  wait for completion
   return pxm2;
}

void * PXM_rotate_thread(void *arg)
{
   using namespace pxmrotate;

   int      index = *((int *) (arg));
   int      px2, py2, px0, py0;
   float    *pix0, *pix1, *pix2, *pix3;
   float    px1, py1;
   float    f0, f1, f2, f3, red, green, blue;
   float    a, b, d, e, ww15, hh15, ww25, hh25;

   ww15 = 0.5 * ww1;
   hh15 = 0.5 * hh1;
   ww25 = 0.5 * ww2;
   hh25 = 0.5 * hh2;

   a = cosf(angle);
   b = sinf(angle);
   d = - sinf(angle);
   e = cosf(angle);
   
   for (py2 = index; py2 < hh2; py2 += Nwt)                                //  loop through output pixels
   for (px2 = 0; px2 < ww2; px2++)
   {
      px1 = a * (px2 - ww25) + b * (py2 - hh25) + ww15;                    //  (px1,py1) = corresponding
      py1 = d * (px2 - ww25) + e * (py2 - hh25) + hh15;                    //    point within input pixels

      px0 = px1;                                                           //  pixel containing (px1,py1)
      py0 = py1;
      
      if (px1 < 0 || px0 >= ww1-1 || py1 < 0 || py0 >= hh1-1) {            //  if outside input pixel array
         pix2 = bmp2 + (py2 * ww2 + px2) * 3;                              //    output is black
         pix2[0] = pix2[1] = pix2[2] = 0;
         continue;
      }

      pix0 = bmp1 + (py0 * ww1 + px0) * 3;                                 //  4 input pixels based at (px0,py0)
      pix1 = pix0 + ww1 * 3;
      pix2 = pix0 + 3;
      pix3 = pix1 + 3;

      f0 = (px0+1 - px1) * (py0+1 - py1);                                  //  overlap of (px1,py1)
      f1 = (px0+1 - px1) * (py1 - py0);                                    //    in each of the 4 pixels
      f2 = (px1 - px0) * (py0+1 - py1);
      f3 = (px1 - px0) * (py1 - py0);
   
      red =   f0 * pix0[0] + f1 * pix1[0] + f2 * pix2[0] + f3 * pix3[0];   //  sum the weighted inputs
      green = f0 * pix0[1] + f1 * pix1[1] + f2 * pix2[1] + f3 * pix3[1];
      blue =  f0 * pix0[2] + f1 * pix1[2] + f2 * pix2[2] + f3 * pix3[2];
      
      pix2 = bmp2 + (py2 * ww2 + px2) * 3;                                 //  output pixel
      pix2[0] = red;
      pix2[1] = green;
      pix2[2] = blue;
   }
   
   zadd_locked(busy,-1);
   return 0;
}


/**************************************************************************

      PXB pixmap functions - RGB uint8 pixel map and GdkPixbuf wrapper

***************************************************************************/

//  Create PXB pixmap with pixels cleared to zero

PXB * PXB_make(int ww, int hh)                                             //  v.12.10
{
   int req_mb = ww * hh * 3 / mega;                                        //  warn if memory low              v.12.11
   warn_memory(req_mb);

   PXB *pxb = (PXB *) zmalloc(sizeof(PXB),"PXB");
   strcpy(pxb->wmi,"pxbpix");
   pxb->pixbuf = gdk_pixbuf_new(GDKRGB,0,8,ww,hh);
   pxb->pixels = gdk_pixbuf_get_pixels(pxb->pixbuf);
   pxb->ww = ww;
   pxb->hh = hh;
   pxb->rs = gdk_pixbuf_get_rowstride(pxb->pixbuf);
   int cc = hh * pxb->rs;
   memset(pxb->pixels,0,cc);
   return pxb;
}


//  Create a PXB pixmap from a GdkPixbuf
//  The pixbuf is used directly and is not duplicated

PXB * PXB_make(GdkPixbuf *pixbuf1)                                         //  v.12.10
{
   PXB *pxb2 = (PXB *) zmalloc(sizeof(PXB),"PXB");
   strcpy(pxb2->wmi,"pxbpix");
   pxb2->pixbuf = pixbuf1;
   pxb2->ww = gdk_pixbuf_get_width(pixbuf1);
   pxb2->hh = gdk_pixbuf_get_height(pixbuf1);
   pxb2->rs = gdk_pixbuf_get_rowstride(pixbuf1);
   pxb2->pixels = gdk_pixbuf_get_pixels(pixbuf1);
   return pxb2;
}


//  Free PXB pixmap - release memory

void PXB_free(PXB *&pxb)                                                   //  v.12.10
{
   if (! pxb) return;
   if (! strEqu(pxb->wmi,"pxbpix")) 
      zappcrash("PXB_free(), bad PXB");
   strcpy(pxb->wmi,"xxxxxx");
   g_object_unref(pxb->pixbuf);
   zfree(pxb);
   pxb = 0;
   return;
}


//  Copy a PXB pixmap to a new PXB pixmap

PXB * PXB_copy(PXB *pxb1)                                                  //  v.12.10
{
   int req_mb = pxb1->ww * pxb1->hh * 3 / mega;                            //  warn if memory low              v.12.11
   warn_memory(req_mb);

   PXB *pxb2 = (PXB *) zmalloc(sizeof(PXB),"PXB");
   strcpy(pxb2->wmi,"pxbpix");
   pxb2->pixbuf = gdk_pixbuf_copy(pxb1->pixbuf);
   pxb2->ww = pxb1->ww;
   pxb2->hh = pxb1->hh;
   pxb2->rs = pxb1->rs;
   pxb2->pixels = gdk_pixbuf_get_pixels(pxb2->pixbuf);
   return pxb2;
}


//  Rescale PXB pixmap to given width and height.

PXB * PXB_rescale(PXB *pxb1, int ww2, int hh2)                             //  v.12.10
{
   int req_mb = ww2 * hh2 * 3 / mega;                                      //  warn if memory low              v.12.11
   warn_memory(req_mb);

   GdkPixbuf *pixbuf2 = gdk_pixbuf_scale_simple(pxb1->pixbuf,ww2,hh2,BILINEAR);
   PXB *pxb2 = PXB_make(pixbuf2);
   return pxb2;
}


//  Copy a PXM image (RGB float) to a PXB image (RGB uint8).

PXB * PXM_PXB_copy(PXM *pxm1)                                              //  v.12.10
{
   float    *pix1;
   uint8    *pix2;
   PXB      *pxb2;
   int      ww, hh, px, py;

   ww = pxm1->ww;
   hh = pxm1->hh;

   pxb2 = PXB_make(ww,hh);

   for (py = 0; py < hh; py++) 
   {
      pix1 = PXMpix(pxm1,0,py); 
      pix2 = PXBpix(pxb2,0,py); 

      for (px = 0; px < ww; px++)
      {
         pix2[0] = pix1[0];
         pix2[1] = pix1[1];
         pix2[2] = pix1[2];
         pix1 += 3;
         pix2 += 3;
      }
   }

   return pxb2;
}


//  Update a PXB section from an updated PXM section.
//  The PXM and PXB must have the same dimensions.
//  px3, py3, ww3, hh3: modified section within pxm1 to propagate to pxb2;

void PXM_PXB_update(PXM *pxm1, PXB *pxb2, int px3, int py3, int ww3, int hh3)          //  v.12.10
{
   float    *pix1;
   uint8    *pix2;
   int      px, py;
   int      lox, hix, loy, hiy;
   
   lox = px3;
   hix = px3 + ww3;
   if (lox < 0) lox = 0;
   if (hix > pxb2->ww) hix = pxb2->ww;
   
   loy = py3;
   hiy = py3 + hh3;
   if (loy < 0) loy = 0;
   if (hiy > pxb2->hh) hiy = pxb2->hh;
   
   for (py = loy; py < hiy; py++)
   {
      pix1 = PXMpix(pxm1,lox,py);
      pix2 = PXBpix(pxb2,lox,py);
      
      for (px = lox; px < hix; px++)
      {
         pix2[0] = pix1[0];                                                //  0.0 - 255.99  >>  0 - 255
         pix2[1] = pix1[1];
         pix2[2] = pix1[2];
         pix1 += 3;
         pix2 += 3;
      }
   }
   
   return;
}


//  Update an output PXB section from a corresponding input PXB section.
//  The two PXBs represent the same image, scaled according to their 
//  respective width and height pixel dimensions.
//  px3, py3, ww3, hh3: modified section within pxb1 to propagate to pxb2;

namespace pxbpxbupdate {
   uint8       *bmp1, *bmp2;
   int         ww1, hh1, ww2, hh2, rs1, rs2;
   int         pww1, phh1, pww2, phh2;
   int         *px1L, *py1L;
   float       scalex, scaley;
   float       *pxmap, *pymap;
   int         maxmapx, maxmapy;
}

void PXB_PXB_update(PXB *pxb1, PXB *pxb2, int px3, int py3, int ww3, int hh3)             //  v.12.10
{
   using namespace pxbpxbupdate;

   int         px1, py1, px2, py2;
   int         pxl, pyl, pxm, pym, ii;
   float       px1a, py1a, px1b, py1b;
   float       fx, fy, ftot;
   uint8       *pixel1, *pixel2;
   float       red, green, blue;
   int         lox, hix, loy, hiy;
   

   ww1 = pxb1->ww;
   hh1 = pxb1->hh;
   rs1 = pxb1->rs;
   bmp1 = pxb1->pixels;
   
   ww2 = pxb2->ww;
   hh2 = pxb2->hh;
   rs2 = pxb2->rs;
   bmp2 = pxb2->pixels;
   
   if (ww1 == pww1 && hh1 == phh1 && ww2 == pww2 && hh2 == phh2)           //  if the sizes are the same as before,
      goto copy_pixels;                                                    //    the pixel mapping math can be avoided.
   
   pww1 = ww1;
   phh1 = hh1;
   pww2 = ww2;
   phh2 = hh2;
   
   if (px1L) {                                                             //  unless this is the first call,
      zfree(px1L);                                                         //    free prior map memory
      zfree(py1L);
      zfree(pxmap);
      zfree(pymap);
   }

   scalex = 1.0 * ww1 / ww2;                                               //  compute x and y scales
   scaley = 1.0 * hh1 / hh2;
   
   if (scalex <= 1) maxmapx = 2;                                           //  compute max input pixels
   else maxmapx = scalex + 2;                                              //    mapping into output pixels
   maxmapx += 1;                                                           //      for both dimensions
   if (scaley <= 1) maxmapy = 2;                                           //  (pixels may not be square)
   else maxmapy = scaley + 2;
   maxmapy += 1;                                                           //  (extra entry for -1 flag)
   
   pymap = (float *) zmalloc(hh2 * maxmapy * sizeof(float),"pxbpxbupd");   //  maps overlap of < maxmap input
   pxmap = (float *) zmalloc(ww2 * maxmapx * sizeof(float),"pxbpxbupd");   //    pixels per output pixel

   py1L = (int *) zmalloc(hh2 * sizeof(int),"pxbpxbupd");                  //  maps first (lowest) input pixel
   px1L = (int *) zmalloc(ww2 * sizeof(int),"pxbpxbupd");                  //    per output pixel

   for (py2 = 0; py2 < hh2; py2++)                                         //  loop output y-pixels
   {
      py1a = py2 * scaley;                                                 //  corresponding input y-pixels
      py1b = py1a + scaley;
      if (py1b >= hh1) py1b = hh1 - 0.001;                                 //  fix precision limitation
      pyl = py1a;
      py1L[py2] = pyl;                                                     //  1st overlapping input pixel

      for (py1 = pyl, pym = 0; py1 < py1b; py1++, pym++)                   //  loop overlapping input pixels
      {
         if (py1 < py1a) {                                                 //  compute amount of overlap
            if (py1+1 < py1b) fy = py1+1 - py1a;                           //    0.0 to 1.0 
            else fy = scaley;
         }
         else if (py1+1 > py1b) fy = py1b - py1;
         else fy = 1;

         ii = py2 * maxmapy + pym;                                         //  save it
         pymap[ii] = 0.9999 * fy / scaley;
      }
      ii = py2 * maxmapy + pym;                                            //  set an end marker after
      pymap[ii] = -1;                                                      //    last overlapping pixel
   }
   
   for (px2 = 0; px2 < ww2; px2++)                                         //  do same for x-pixels
   {
      px1a = px2 * scalex;
      px1b = px1a + scalex;
      if (px1b >= ww1) px1b = ww1 - 0.001;
      pxl = px1a;
      px1L[px2] = pxl;

      for (px1 = pxl, pxm = 0; px1 < px1b; px1++, pxm++)
      {
         if (px1 < px1a) {
            if (px1+1 < px1b) fx = px1+1 - px1a;
            else fx = scalex;
         }
         else if (px1+1 > px1b) fx = px1b - px1;
         else fx = 1;

         ii = px2 * maxmapx + pxm;
         pxmap[ii] = 0.9999 * fx / scalex;
      }
      ii = px2 * maxmapx + pxm;
      pxmap[ii] = -1;
   }

copy_pixels:

   px3 = px3 / scalex;                                                     //  convert input area to output area
   py3 = py3 / scaley;
   ww3 = ww3 / scalex + 2;
   hh3 = hh3 / scaley + 2;
   
   lox = px3;
   hix = px3 + ww3;
   if (lox < 0) lox = 0;
   if (hix > ww2) hix = ww2;
   
   loy = py3;
   hiy = py3 + hh3;
   if (loy < 0) loy = 0;
   if (hiy > hh2) hiy = hh2;
   
   for (py2 = loy; py2 < hiy; py2++)                                       //  loop output y-pixels
   {
      pyl = py1L[py2];                                                     //  corresp. 1st input y-pixel

      for (px2 = lox; px2 < hix; px2++)                                    //  loop output x-pixels
      {
         pxl = px1L[px2];                                                  //  corresp. 1st input x-pixel

         red = green = blue = 0;                                           //  initz. output pixel

         for (py1 = pyl, pym = 0; ; py1++, pym++)                          //  loop overlapping input y-pixels
         {
            ii = py2 * maxmapy + pym;                                      //  get y-overlap
            fy = pymap[ii];
            if (fy < 0) break;                                             //  no more pixels

            for (px1 = pxl, pxm = 0; ; px1++, pxm++)                       //  loop overlapping input x-pixels
            {
               ii = px2 * maxmapx + pxm;                                   //  get x-overlap
               fx = pxmap[ii];
               if (fx < 0) break;                                          //  no more pixels

               ftot = fx * fy;                                             //  area overlap = x * y overlap
               pixel1 = bmp1 + py1 * rs1 + px1 * 3;
               red += pixel1[0] * ftot;                                    //  add input pixel * overlap
               green += pixel1[1] * ftot;
               blue += pixel1[2] * ftot;
            }

            pixel2 = bmp2 + py2 * rs2 + px2 * 3;                           //  save output pixel
            pixel2[0] = red;                                               //  0.0 - 255.996  >>  0 - 255
            pixel2[1] = green;
            pixel2[2] = blue;
         }
      }
   }

   return;
}



