/**************************************************************************

   Fotoxx      edit photos and manage collections

   Copyright 2007 2008 2009 2010 2011 2012  Michael Cornelison
   Source URL: http://kornelix.com/fotoxx
   Contact: kornelix2@gmail.com
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see http://www.gnu.org/licenses/.

***************************************************************************/

#include <tiffio.h>
#include <png.h>
#include "zfuncs.h"

//  Fotoxx definitions

#define fotoxxvers "fotoxx v.13.01"                                        //  this fotoxx version
#define flicense "Free software - GNU General Public License v.3"
#define fhomepage "http://kornelix.com/fotoxx"
#define ftranslators                                                                \
        "Translators: \n"                                                           \
        "   Victor Ryzhikh (Виктор Рыжих), Miguel Anxo Bouzada, \n"                 \
        "   Doriano Blengino, Arthur Kalverboer, André Campos Rodovalho, \n"        \
        "   Jie Luo, Peter Landgren, Josep Antoni Miralles Puignau, \n"             \
        "   Jean-Pierre Vaissière "
#define fcredits "Software tools: GTK, libtiff, libpng, dcraw, ufraw, exiftool"
#define fcontact "Bug reports: kornelix2@gmail.com"

#define mega (1024 * 1024)                                                 //  1 million as 2**20
#define desktoptrash "Desktop/fotoxx-trash"                                //  alternate trash folder location
#define def_jpeg_quality "90"                                              //  default jpeg quality (high)
#define pi 3.141592654
#define PXMpix(PXM,px,py) (PXM->pixels+((py)*(PXM->ww)+(px))*3)            //  get PXM pixel[3] at (px,py) (float[3])
#define PXBpix(PXB,px,py) (PXB->pixels+((py)*(PXB->rs)+(px)*3))            //  get PXB pixel[3] at (px,py) (uint8[3])
#define pixbright(pix) (0.25*(pix)[0]+0.65*(pix)[1]+0.10*(pix)[2])         //  pixel brightness, 0-255.0
#define pixred(pix) (25*pix[0]/(pixbright(pix)+1))                         //  pixel redness, 0-100%
#define maxedits 100                                                       //  max. edits, undo/redo depth
#define tagdelims  ",;"                                                    //  valid tag delimiters (; obsolete)
#define tagdelim1  ','                                                     //  delimiter for new tags
#define tagdelim2  ", "                                                    //  delimiter + blank

//  compile time limits that could be increased

#define max_threads 6                                                      //  max. working threads
#define thumbfilesize 256                                                  //  thumbnail image file size
#define Nrecentfiles 100                                                   //  max. recent image files
#define search_index_max 2000                                              //  max. image files per search index file
#define indexrecl 2000                                                     //  max. search index rec. (>maxfcc >tagFcc)
#define maximages 500000                                                   //  max. image files supported      v.12.10
#define maxgeotags 10000                                                   //  max. city geotags records
#define maxtagcats 200                                                     //  max tag categories
#define tagcc  50                                                          //  max cc for one tag or category ID
#define tagFcc 1000                                                        //  max tag cc for one image file
#define tagntc 1000                                                        //  max no. tags for one category
#define tagGcc 50000                                                       //  max tag cc for one category
#define tagTcc 50000                                                       //  max tag cc for all defined tags
#define tagMcc 200                                                         //  max tag cc for batch add tags
#define tagScc 200                                                         //  max tag cc for search tags
#define tagRcc 200                                                         //  max tag cc for recent tags

//  EXIF/IPTC keys for embedded image metadata

#define iptc_keywords_key "Keywords"                                       //  keywords (tags) (IPTC)
#define iptc_rating_key "Rating"                                           //  star rating (IPTC)
#define exif_size_key "ImageSize"                                          //  image size, NNNNxNNNN
#define exif_date_key "DateTimeOriginal"                                   //  date/time (EXIF)
#define exif_width_key "ImageWidth"                                        //  width, pixels (EXIF)
#define exif_height_key "ImageHeight"                                      //  height, pixels (EXIF)
#define exif_orientation_key "Orientation"                                 //  orientation (EXIF)
#define exif_editlog_key "ImageHistory"                                    //  edit history log (EXIF)         v.12.06
#define exif_comment_key "UserComment"                                     //  user comments (EXIF)
#define iptc_caption_key "Caption-Abstract"                                //  image caption (IPTC)
#define exif_focal_length_key "FocalLengthIn35mmFormat"                    //  focal length (EXIF)
#define exif_city_key "City"                                               //  city name              geotags  v.12.05
#define exif_country_key "Country"                                         //  country name
#define exif_latitude_key "GPSLatitude"                                    //  latitude in degrees (-180 to +180)
#define exif_longitude_key "GPSLongitude"                                  //  longitude in degrees (-180 to +180)
#define exif_maxcc 2000                                                    //  max. cc for exif/iptc text      v.12.06

//  GTK/GDK shortcuts

#define TEXTWIN GTK_TEXT_WINDOW_TEXT
#define NODITHER GDK_RGB_DITHER_NONE
#define ALWAYS GTK_POLICY_ALWAYS
#define NEVER GTK_POLICY_NEVER
#define GDKRGB GDK_COLORSPACE_RGB
#define LINEATTRIBUTES GDK_LINE_SOLID, GDK_CAP_BUTT, GDK_JOIN_MITER
#define BILINEAR GDK_INTERP_BILINEAR
#define SCROLLWIN GTK_SCROLLED_WINDOW
#define MWIN GTK_WINDOW(mWin)

//  externals from zfuncs module

namespace zfuncs {
   extern char    zlang[8];                                                //  current language lc_RC
   extern cchar   *F1_help_topic;                                          //  F1 context help topic
   extern char    zicondir[200];                                           //  where application icons live
   extern int     zdialog_busy;                                            //  tracks open zdialog's
   extern int     open_popup_windows;                                      //  tracks popup windows open
}

//  externals from f.file module

namespace navi {                                                           //  gallery() data
   extern char    *galleryname;                                            //  directory path or gallery type name
   extern int     gallerytype;                                             //  1-5 = dirk/search/metadata/coll/recent
   extern int     topfileposn;                                             //  gallery window target scroll position
   extern int     nfiles, nimages;                                         //  file counts, total and images
   extern char    **flist;                                                 //  image file list
   extern char    **mdlist;                                                //  metadata list
   extern int     mdrows;                                                  //  metadata list, text rows
   extern int     xwinW, xwinH;                                            //  image gallery window size
   extern int     thumbsize;                                               //  curr. thumbnail display size
   extern int     genthumbs;                                               //  counts thumbnails generated

   int    gallery_paint(GtkWidget *, cairo_t *);                           //  gallery window paint function
   void   menufuncx(GtkWidget *win, cchar *menu);                          //  gallery toolbar buttons function
   void   changedirk(GtkWidget *widget, GdkEventButton *event);            //  gallery directory change function
   void   rootdirk(GtkWidget *widget, GdkEventButton *event);              //  gallery jump to root directory function
   void   mouse_event(GtkWidget *, GdkEvent *, void *);                    //  gallery window mouse event function
	int    KBxrelease(GtkWidget *, GdkEventKey *, void *);                  //  gallery window key release event
   int    image_fcomp(cchar *file1, cchar *file2);                         //  gallery (directory) file sequence
}

#define ccc (maxfcc*2+200)                                                 //  max. command line size
EX char        command[ccc];                                               //  (command, parameters, 2 filespecs)

EX char        PIDstring[12];                                              //  process PID as string
EX int         Nwt;                                                        //  working threads to use
EX int         wtnx[8];                                                    //  internal thread IDs
EX int         wthreads_busy;                                              //  active thread count

EX GtkWidget      *mWin, *tabWin;                                          //  main and tabbed window
EX GtkWidget      *fVbox, *fMenu, *fTbar, *fDwin, *fSbar;                  //  tab F: menu, tbar, draw.win, stbar
EX GtkWidget      *gVbox, *gHbox, *gHome, *gPath;                          //  tab G: navigation widgets
EX GtkWidget      *gTbar, *gSwin, *gDwin;                                  //  tab G: toolbar, scroll win, draw win
EX GtkAdjustment  *gSadj;                                                  //  tab G: scrollbar adjustment
EX GtkWidget      *wDwin;                                                  //  tab W: drawing window
EX GtkWidget      *cDwin;                                                  //  curr. drawing window: fDwin or wDwin
EX GtkWidget      *cSbar;                                                  //  curr. status bar: fSbar or null
EX GdkWindow      *gdkWin;                                                 //  GDK drawing window (from cDwin)
EX cairo_t        *mwcr;                                                   //  cairo context for gdkWin

#define tabF   0                                                           //  shows current image file
#define tabG   1                                                           //  shows thumbnail gallery
#define tabW   2                                                           //  shows world map
EX int         cTab;                                                       //  current tab = tabF/tabG/tabW

EX uint8       black[3], white[3], red[3], green[3];                       //  RGB values (uint8[3])
EX uint8       *fg_color;                                                  //  foreground drawing color
EX GdkRGBA     GDKdgray, GDKblack;

EX GdkCursor   *arrowcursor;                                               //  main window cursors
EX GdkCursor   *dragcursor;
EX GdkCursor   *drawcursor;

typedef void   menufunc_t(GtkWidget *,cchar *);                            //  table of menu names and functions
typedef struct {
   GtkWidget   *topmenu;                                                   //  parent menu
   cchar        *text;                                                     //  menu text
   cchar        *icon;                                                     //  menu icon
   cchar        *desc;                                                     //  menu short description (tooltip)
   menufunc_t   *func;                                                     //  menu function (GtkWidget *, cchar *)
   cchar        *arg;                                                      //  argument
}  menutab_t;

#define maxmenus 200
EX menutab_t   menutab[maxmenus];
EX int         Nmenus;

#define maxshortcuts 100
EX char        *shortcutkey[100];                                          //  KB shortcut keys (or combinations)
EX char        *shortcutmenu[100];                                         //  corresponding menu names
EX int         Nshortcuts;                                                 //  count of entries

EX int max_view_megapix, max_edit_megapix;                                 //  calculated from avail. memory   v.12.10
EX int wwhh_limit;                                                         //  initialized in main()           v.12.10

EX int         Fthumbnails;                                                //  thumbnails are used (default = 1)
EX int         Fshutdown;                                                  //  app shutdown underway
EX int         Fdebug;                                                     //  debug flag
EX int         Fkillfunc;                                                  //  signal, kill running function
EX int         Ffuncbusy;                                                  //  function is busy/working
EX int         Fthreadbusy;                                                //  thread is busy/working
EX int         Fautotrim;                                                  //  start m_trim() for each opened file
EX int         Wrepaint;                                                   //  request window paint
EX int         Wpainted;                                                   //  window was repainted
EX int         Wfullscreen;                                                //  flag, window is fullscreen
EX int         Frefresh;                                                   //  full image refresh needed
EX int         Fmodified;                                                  //  image was edited/modified
EX int         Fsaved;                                                     //  image saved since last mod
EX int         Fmetachanged;                                               //  image metadata changed and not saved
EX int         Fblowup;                                                    //  zoom small images to window size
EX int         Fmenulock;                                                  //  lock for some menu functions
EX int         Fslideshow;                                                 //  slide show mode is active
EX int         Frecent;                                                    //  start with recent files gallery
EX int         Fprev;                                                      //  start with previous file
EX int         Fblank;                                                     //  start with blank window
EX int         Fnosync;                                                    //  start without auto file sync
EX char        *topmenu;                                                   //  latest top-menu selection
EX cchar       *threadmessage;                                             //  popup message from thread process
EX char        *SS_imagefile;                                              //  command line, slide show start file
EX char        *SS_musicfile;                                              //  command line, slide show music file
EX char        *startmenu;                                                 //  command line, startup menu function
EX char        file_errmess[1000];                                         //  error message from low-level file funcs

volatile                                                                   //  stop optimizer removing code
EX double      Fbusy_goal, Fbusy_done;                                     //  BUSY message progress tracking  v.13.01
EX char        SB_text[200];                                               //  status bar application text

EX int         Ntopdirk;                                                   //  topdirk[] entries, 1-10
EX char        *topdirk[10];                                               //  user's top-level image directories
EX char        *thumbdirk;                                                 //  thumbnails directory (or null)
EX char        *curr_file, *curr_dirk;                                     //  current image file and directory
EX char        *recentfiles[Nrecentfiles];                                 //  most recent image files opened
EX char        *initial_file;                                              //  initial file on command line
EX char        curr_file_type[8];                                          //  jpg / tif / png / other
EX int         curr_file_bpc;                                              //  image file bits/color, 8/16
EX int         curr_file_size;                                             //  image file size on disk
EX int         curr_file_posn;                                             //  position in current file set, 0-last
EX int         curr_file_count;                                            //  count of images in current set
EX char        *click_file;                                                //  image file / thumbnail clicked
EX char        *rawfile;                                                   //  RAW file passed to ufraw program

EX char        f_load_type[8];                                             //  data set by f_load()
EX int         f_load_bpc, f_load_size;
EX char        *f_save_file;                                               //  data set by f_save()
EX char        f_save_type[8];
EX int         f_save_bpc, f_save_size;
EX char        jpeg_quality[8];                                            //  jpeg file save quality

//  files and directories in /.../.fotoxx/

EX char        topdirk_file[200];                                          //  top directory save file 
EX char        tags_defined_file[200];                                     //  tags defined file
EX char        search_index_dirk[200];                                     //  search index directory          v.12.10
EX char        recentfiles_file[200];                                      //  file of recent image files 
EX char        collections_dirk[200];                                      //  directory for saved image collections
EX char        saved_areas_dirk[200];                                      //  directory for saved select area files
EX char        saved_curves_dirk[200];                                     //  directory for saved curve data
EX char        writetext_dirk[200];                                        //  directory for write_text files
EX char        geotags_dirk[200];                                          //  directory for geotags data files
EX char        favorites_dirk[200];                                        //  directory for favorites menu    v.12.10
EX char        exiftool_input[200];                                        //  exiftool_server input file
EX char        KBshortcuts[200];                                           //  keyboard shortcuts file         v.12.06

//  fotoxx PXM and PXB pixmaps

typedef struct  {                                                          //  PXM pixmap, 3 x float per pixel
   char        wmi[8];                                                     //  self-identifier
   int         ww, hh;                                                     //  width, height
   float       *pixels;                                                    //  ww * hh * 3 * sizeof(float)
}  PXM;

typedef struct  {                                                          //  PXB pixmap, 3 x uint8/pixel     v.12.10
   char        wmi[8];                                                     //  self-identifier
   GdkPixbuf   *pixbuf;                                                    //  underlying GdkPixbuf
   int         ww, hh, rs;                                                 //  width, height, row-stride
   uint8       *pixels;                                                    //  ww * hh * 3 bytes
}  PXB;

EX PXB         *Fpxb;                                                      //  current file PXB, size = 1x
EX PXB         *Mpxb;                                                      //  current file PXB, size = Mscale
EX PXB         *Dpxb;                                                      //  Mpxb part within drawing window
EX PXB         *Wpxb;                                                      //  world map PXB, size = 1x

EX PXM         *E0pxm;                                                     //  edit pixmap, original image
EX PXM         *E1pxm;                                                     //  edit pixmap, base image for editing
EX PXM         *E3pxm;                                                     //  edit pixmap, edited image
EX PXM         *ERpxm;                                                     //  edit pixmap, undo/redo image
EX PXM         *E8pxm;                                                     //  scratch image for some functions
EX PXM         *E9pxm;                                                     //  scratch image for some functions

EX int         Dww, Dhh;                                                   //  main/drawing window size
EX int         dww, dhh;                                                   //  Dpxb size in draw. window, <= Dww, Dhh
EX int         Morgx, Morgy;                                               //  Dpxb origin in Mpxb image
EX int         Dorgx, Dorgy;                                               //  Dpxb origin in drawing window

EX float       Fzoom;                                                      //  image zoom scale (0 = fit window)
EX int         zoomx, zoomy;                                               //  req. zoom center of window
EX float       Mscale;                                                     //  scale factor, 1x image to window image
EX float       Pscale;                                                     //  prior scale factor

EX mutex       Fpixmap_lock;                                               //  lock for accessing PXM pixmaps
EX mutex       preload_lock;                                               //  lock for image file preload function

EX int         Mbutton;                                                    //  mouse button, 1/3 = left/right
EX int         Mwxposn, Mwyposn;                                           //  mouse position, window space
EX int         Mxposn, Myposn;                                             //  mouse position, image space
EX int         LMclick, RMclick;                                           //  mouse left, right click
EX int         Mxclick, Myclick;                                           //  mouse click position, image space
EX int         Mdrag;                                                      //  mouse drag underway
EX int         Mwdragx, Mwdragy;                                           //  drag increment, window space
EX int         Mxdown, Mydown, Mxdrag, Mydrag;                             //  mouse drag vector, image space
EX int         Fmousemain;                                                 //  mouse acts on main window not dialog
EX int         Mcapture;                                                   //  mouse captured by edit function
EX int         KBcapture;                                                  //  KB key captured by edit function
EX int         KBkey;                                                      //  active keyboard key

EX int         KBcontrolkey;                                               //  keyboard key states, 1 = down
EX int         KBshiftkey;
EX int         KB_A_key;

EX int         Ntoplines, Nptoplines;                                      //  lines drawn over image in window
EX int         toplinex1[4], topliney1[4], toplinex2[4], topliney2[4];     //  (unbend)
EX int         ptoplinex1[4], ptopliney1[4], ptoplinex2[4], ptopliney2[4];

struct toptext_t {                                                         //  text strings drawn over image in window
   int      ID, px, py;
   cchar    *text;
   cchar    *font;
};

#define maxtoptext 100                                                     //  max. text strings
EX toptext_t   toptext[100];                                               //  text strings
EX int         Ntoptext;                                                   //  current count

struct topcircle_t {                                                       //  circles drawn over image in window
   int      px, py;
   int      radius;
   int      color;                                                         //  1/2/3 = white/black/red
};

#define maxtopcircles 100                                                  //  max. circles
EX topcircle_t    topcircles[100];                                         //  circles 
EX int            Ntopcircles;                                             //  current count

//  spline curve data

typedef void t_spcfunc(int spc);                                           //  callback function, spline curve edit

typedef struct {                                                           //  spline curve data 
   GtkWidget   *drawarea;                                                  //  drawing area for spline curves
   t_spcfunc   *spcfunc;                                                   //  callback function when curve changed
   int         Nspc;                                                       //  number of active curves, 1-10
   int         vert[10];                                                   //  curve is vert. (1) or horz. (0)
   int         nap[10];                                                    //  anchor points per curve
   float       apx[10][50], apy[10][50];                                   //  up to 50 anchor points per curve
   float       yval[10][1000];                                             //  y-values for x = 0 to 1 by 0.001
   int         Nscale;                                                     //  no. of fixed scale lines, 0-10
   float       xscale[2][10];                                              //  2 x-values for end points
   float       yscale[2][10];                                              //  2 y-values for end points
} spldat;

//  select area data

#define sa_initseq 10                  //  initial sequence number (0/1/2 reserved)
#define sa_maxseq 9999                 //  ultimate limit 64K
#define mode_rect       1              /* select rectangle by drag/click */
#define mode_ellipse    2              /* select ellipse by drag */
#define mode_draw       3              /* freehand draw by drag/click */
#define mode_follow     4              /* follow edge indicated by clicks */
#define mode_replace    5              /* freehand draw and replace nearby pixels */
#define mode_mouse      6              /* select area within mouse (radius) */
#define mode_onecolor   7              /* select one matching color within mouse */
#define mode_allcolors  8              /* select all matching colors within mouse */
#define mode_image      9              /* select whole image */

EX uint16      sa_endpx[10000], sa_endpy[10000];                           //  last pixel drawn per seqence no.
EX int         sa_thresh;                                                  //  mouse pixel distance threshold
EX int         sa_mouseradius;                                             //  mouse selection radius
EX int         sa_searchrange;                                             //  search range (* mouse radius)
EX int         sa_mousex, sa_mousey;                                       //  mouse position in image
EX float       sa_colormatch;                                              //  color range to match (0.001 to 1.0)
EX int         sa_firewall;                                                //  firewall: no search thru selected pixels
EX int         sa_matchRGB[1000][3];                                       //  match colors, up to 1000
EX int         sa_Nmatch;                                                  //  count of match colors
EX int         sa_calced;                                                  //  edge calculation done
EX int         sa_blend;                                                   //  edge blend width
EX int         sa_minx, sa_maxx;                                           //  enclosing rectangle for area
EX int         sa_miny, sa_maxy;
EX char        *sa_stackdirec;                                             //  pixel search stack
EX int         *sa_stackii;    
EX int         sa_maxstack;
EX int         sa_Nstack;
EX int         sa_stat;                                                    //  0/1/2/3/4 = none/edit/xx/fin./disab.
EX int         sa_mode;                                                    //  1-7 = curr. select area edit method
EX int         sa_finOK;                                                   //  finish area - success
EX int         sa_finhole;                                                 //  finish area - area has a hole
EX int         sa_fincancel;                                               //  finish area - user cancel
EX int         sa_currseq;                                                 //  current select sequence no.
EX int         sa_Ncurrseq;                                                //  current sequence pixel count
EX char        *sa_pixselc;                                                //  maps pixels selected in current cycle
EX uint16      *sa_pixmap;                                                 //  0/1/2+ = outside/edge/inside edge dist
EX int         sa_Npixel;                                                  //  total select area pixel count
EX int         sa_fww, sa_fhh;                                             //  valid image dimensions for select area
EX int         Fshowarea;                                                  //  show select area outline
EX int         areanumber;                                                 //  increasing sequential number

//  search index file record. 
//  tags, comms, capt, gtags may have "null" (char) as a value.

struct sxrec_t {
   char        *file;                                                      //  image filespec
   char        fdate[16];                                                  //  file date, yyyymmddhhmmss
   char        pdate[12];                                                  //  EXIF (photo) date, yyyy:mm:dd
   char        rating[4];                                                  //  IPTC rating, '0' to '5' stars
   char        size[16];                                                   //  image size, "12345x9876"
   char        *tags;                                                      //  IPTC tags
   char        *capt;                                                      //  IPTC caption
   char        *comms;                                                     //  EXIF comments
   char        *gtags;                                                     //  EXIF geotags
};
EX FILE        *sxrec_fid;                                                 //  open file ID - get_search_index_rec()

EX int         edit_coll_mousecapture;                                     //  edit collection, capture gallery mouse

//  image edit function data

typedef struct {                                                           //  edit function data
   cchar       *funcname;                                                  //  function name, e.g. flatten
   int         Fprev;                                                      //  flag, use preview window
   int         Farea;                                                      //  area: 0/1/2 = delete/ignore/useable
   int         Fmod;                                                       //  flag, image modified by this func
   zdialog     *zd;                                                        //  edit dialog
   spldat      *curves;                                                    //  edit curve data
   void *      (*threadfunc)(void *);                                      //  edit thread function
   void        (*mousefunc)();                                             //  edit mouse function
} editfunc;

EX editfunc    *CEF;                                                       //  current active edit function
EX pvlist      *editlog;                                                   //  log of image edits done
EX char        *undo_files;                                                //  undo/redo stack, image files
EX int         Pundo;                                                      //  undo/redo stack position
EX int         Pumax;                                                      //  undo/redo stack depth
EX int         Fpreview;                                                   //  smaller image being used for edits
EX char        exif_editlog[exif_maxcc];                                   //  EXIF edit history log           v.12.12
EX int         exif_editlog_cc;                                            //  initial cc at file open time

#define maxplugins 100
EX int         Nplugins;                                                   //  plugin menu items
EX char        *plugins[100];

EX GtkWidget   *winhisto;                                                  //  brightness histogram window
EX GtkWidget   *winhistoH;                                                 //  histogram drawing area
EX GtkWidget   *winhistoB;                                                 //  brightness band underneath

//  The following are saved and restored across sessions.
//  They belong to different functions in different source modules.

EX char        *pversion;                                                  //  prior fotoxx version
EX char        *last_sync_time;                                            //  last file sync time
EX int         Ffirsttime;                                                 //  first time startup
EX int         FsyncEnabled;                                               //  file synch is used (not disabled)
EX int         FsyncNeeded;                                                //  file synch is needed
EX int         mwgeom[4];                                                  //  main window position and size
EX char       *menu_style;                                                 //  menu style: classic or vertical
EX char       *tbar_style;                                                 //  toolbar style: icons/text/both
EX char       *startdisplay;                                               //  display: recent/prev/blank/file/dirk
EX char       *startdirk;                                                  //  start directory (startdisplay=dirk)
EX char        *startfile;                                                 //  start image file (startdisplay=file)
EX int         Fdragopt;                                                   //  1/2 = 1x drag / magnified scroll
EX int         Fwarnoverwrite;                                             //  warn overwrite of original image
EX int         zoomcount;                                                  //  zoom count to reach 2x (1/2/3/4)
EX float       zoomratio;                                                  //  corresp. zoom ratio: 2**(1/zoomcount)
EX char       *RAWcommand;                                                 //  command line, RAW conversion to tiff
EX char       *RAWfiletypes;                                               //  list of RAW file types, .raw .rw2 ...
EX int         trimsize[2];                                                //  trim (crop) width, height
EX char        *trimbuttons[6];                                            //  trim dialog button labels
EX char        *trimratios[6];                                             //  corresponding aspect ratios
EX int         editresize[2];                                              //  edit resize width, height
EX char        *writetext_text;                                            //  write_text text
EX char        *writetext_font;                                            //  font and size, e.g. "Sans 20"
EX char        *writetext_color[3];                                        //  text, background, text outline colors
EX int         writetext_trans[3];                                         //  corresp. transparencies
EX int         writetext_outline;                                          //  text outline width, pixels
EX float       writetext_angle;                                            //  angle of text
EX int         Fgrid;                                                      //  both grid lines on / off
EX int         gridon[2];                                                  //  x/y gridlines on / off 
EX int         gridspace[2];                                               //  x/y grid line spacing
EX int         gridcount[2];                                               //  x/y grid line counts (alternative)
EX int         gridoffset[2];                                              //  x/y grid line starting offsets
EX int         rotate_grid[8];                                             //  rotate function grid settings
EX int         unbend_grid[8];                                             //  unbend function grid settings
EX int         warpC_grid[8];                                              //  warp image (curved) grid settings
EX int         warpL_grid[8];                                              //  warp image (linear) grid settings
EX int         warpF_grid[8];                                              //  warp image (affine) grid settings
EX float       lens_mm, lens_bow;                                          //  pano lens mm and bow settings

//  dialogs with global visibility

EX zdialog     *zdeditmeta;                                                //  edit metadata zdialog
EX zdialog     *zdexifview;                                                //  view EXIF/IPTC data
EX zdialog     *zdexifedit;                                                //  edit EXIF/IPTC data
EX zdialog     *zdeditcctext;                                              //  edit IPTC caption, EXIF user comments
EX zdialog     *zdaddgeotags;                                              //  add geotags data
EX zdialog     *zdrename;                                                  //  rename file zdialog
EX zdialog     *zdcopymove;                                                //  copy/move file zdialog
EX zdialog     *zddelete;                                                  //  delete file dialog
EX zdialog     *zdtrash;                                                   //  trash file dialog
EX zdialog     *zdsela;                                                    //  select area dialog
EX zdialog     *zd_editcoll;                                               //  manage collection
EX zdialog     *zd_gallery_getfiles;                                       //  gallery_getfiles

//  GTK functions (fotoxx main)

int   main(int argc, char * argv[]);                                       //  main program
int   initzfunc(void *);                                                   //  initializations
int   gtimefunc(void *arg);                                                //  periodic function
void  update_statusbar();                                                  //  update main window status bar
void  topmenufunc(GtkWidget *, cchar *menu);                               //  top level menu function
int   delete_event();                                                      //  window delete event function
int   destroy_event();                                                     //  window destroy event function
int   state_event(GtkWidget *, GdkEvent *event);                           //  window state event function
int   settab(cchar *tabname);                                              //  set visible tab F/G/W >> 0/1/2
void  tabchanged(GtkNotebook *, GtkWidget *, int tab, void *);             //  tab changed event function
int   mwpaint1(GtkWidget *, cairo_t *);                                    //  main window draw signal function
void  mwpaint2();                                                          //  window repaint (thread callable)
void  mwpaint3(int px, int py, int ww, int hh);                            //  update Mpxb area from updated E3 area
void  mwpaint4(int px, int py, int ww, int hh);                            //  update Dpxb area from updated Mpxb 
void  mouse_event(GtkWidget *, GdkEventButton *, void *);                  //  mouse event function
int   KBpress(GtkWidget *, GdkEventKey *, void *);                         //  KB key press event function
int   KBrelease(GtkWidget *, GdkEventKey *, void *);                       //  KB key release event
void  m_zoom(GtkWidget *, cchar *);                                        //  zoom image +/-
void  win_fullscreen();                                                    //  full screen without menu/tbar/st.bar
void  win_unfullscreen();                                                  //  restore to prior size with menu etc.

//  cairo drawing functions (fotoxx main)

void  draw_pixel(int px, int py);                                          //  draw pixel
void  erase_pixel(int px, int py);                                         //  erase pixel
void  draw_line(int x1, int y1, int x2, int y2);                           //  draw line
void  draw_dotline(int x1, int y1, int x2, int y2);                        //  draw a dotted line
void  erase_line(int x1, int y1, int x2, int y2);                          //  erase line
void  draw_toplines(int arg);                                              //  draw pre-set lines over image
void  draw_gridlines();                                                    //  draw grid lines on image
void  add_toptext(int ID, int px, int py, cchar *text, cchar *font);       //  add text string with ID on window
void  draw_toptext();                                                      //  draw text strings when window repainted
void  erase_toptext(int ID);                                               //  remove all text strings with ID
void  draw_text(int px, int py, cchar *text, cchar *font);                 //  draw text on window
void  add_topcircle(int px, int py, int radius, int color);                //  add circles on window
void  draw_topcircles();                                                   //  draw circles when window repainted
void  erase_topcircles();                                                  //  remove all circles
void  draw_mousearc(int cx, int cy, int ww, int hh);                       //  draw circle/ellipse around pointer

//  spline curve edit functions (fotoxx main)

spldat * splcurve_init(GtkWidget *frame, void func(int spc));              //  initialize spline curves
int      splcurve_adjust(void *, GdkEventButton *event, spldat *);         //  curve editing function
int      splcurve_draw(GtkWidget *, cairo_t *, spldat *);                  //  spline curve draw signal function
int      splcurve_generate(spldat *, int spc);                             //  generate data from anchor points
float    splcurve_yval(spldat *, int spc, float xval);                     //  get curve y-value
int      splcurve_load(spldat *sd);                                        //  load curve from a file
int      splcurve_save(spldat *sd);                                        //  save curve to a file

//  edit support functions (fotoxx main)

int   edit_setup(editfunc &EF);                                            //  start new edit transaction
void  edit_cancel();                                                       //  cancel edit
void  edit_done();                                                         //  commit edit, add undo stack
void  edit_undo();                                                         //  undo edit, back to org. image
void  edit_redo();                                                         //  redo the edit after undo
void  edit_reset();                                                        //  reset to initial status
void  edit_fullsize();                                                     //  convert to full-size pixmaps
void  m_undo(GtkWidget *, cchar *);                                        //  toolbar, undo one edit
void  m_redo(GtkWidget *, cchar *);                                        //  toolbar, redo one edit
void  undo_all();                                                          //  undo all edits of current image
void  save_undo(cchar *funcname);                                          //  undo/redo save function
void  load_undo();                                                         //  undo/redo read function
int   mod_keep();                                                          //  query keep/discard edited image
int   menulock(int lock);                                                  //  lock/unlock menu for some funcs
int   checkmenulock();                                                     //  test menu lock, message user    v.12.07
int   checkmenulockquiet();                                                //  same, no message box            v.12.09

typedef  void CBfunc();                                                    //  callback function type
EX CBfunc   *mouseCBfunc;                                                  //  current edit mouse function
void  takeMouse(CBfunc func, GdkCursor *);                                 //  capture mouse for edit dialog
void  freeMouse();                                                         //  free mouse for main window

//  thread support functions (fotoxx main)

typedef void * threadfunc(void *);                                         //  edit thread function
void  start_thread(threadfunc func, void *arg);                            //  start edit thread
void  signal_thread();                                                     //  signal thread, work is pending
void  wait_thread_idle();                                                  //  wait for work complete
void  wrapup_thread(int command);                                          //  wait for exit or command exit
void  thread_idle_loop();                                                  //  thread: wait for work or exit command
void  thread_exit();                                                       //  thread: exit unconditionally
void  start_wthread(threadfunc func, void *arg);                           //  start working thread (per proc. core)
void  exit_wthread();                                                      //  exit working thread
void  wait_wthreads();                                                     //  wait for working threads

//  other support functions (fotoxx main)

void  m_favorites(GtkWidget *, cchar *);                                   //  graphic popup menu, user favorites
void  m_help(GtkWidget *, cchar *menu);                                    //  various help menu functions
void  save_params();                                                       //  save parameters for next session
void  load_params();                                                       //  load parameters from prior session
void  free_resources(int fkeepundo = 0);                                   //  free all allocated resources
int   vpixel(PXM *pxm, float px, float py, float *vpix);                   //  get virtual pixel at (px,py)
int   sigdiff(float d1, float d2, float signf);                            //  test for significant difference
int   warn_memory(int megabytes);                                          //  popup warning if real mem < threshold

//  PXM and PXB pixmap functions (fotoxx main)

PXM * PXM_make(int ww, int hh);                                            //  create a PXM pixmap
void  PXM_free(PXM *&pxm);                                                 //  free PXM pixmap
PXM * PXM_copy(PXM *pxm);                                                  //  copy PXM pixmap
PXM * PXM_copy_area(PXM *pxm, int orgx, int orgy, int ww, int hh);         //  copy section of PXM pixmap
void  PXM_fixblue(PXM *pxm);                                               //  set blue = 0 pixels to blue = 2
PXM * PXM_rescale(PXM *pxm, int ww, int hh);                               //  rescale PXM pixmap (ww/hh)
PXM * PXM_rotate(PXM *pxm, float angle);                                   //  rotate PXM pixmap

PXB * PXB_make(int ww, int hh);                                            //  create a PXB pixmap
PXB * PXB_make(GdkPixbuf *pixbuf);                                         //  create a PXB pixmap from a pixbuf
void  PXB_free(PXB *&pxb);                                                 //  free PXB pixmap
PXB * PXB_copy(PXB *pxb);                                                  //  copy PXB
PXB * PXB_rescale(PXB *pxb, int ww, int hh);                               //  rescale PXB pixmap (ww/hh)

PXB * PXM_PXB_copy(PXM *pxm);                                              //  PXM to pixbuf, same scale
void  PXM_PXB_update(PXM *, PXB *, int px3, int py3, int ww3, int hh3);    //  update PXB area from PXM, same scale
void  PXB_PXB_update(PXB *, PXB *, int px3, int py3, int ww3, int hh3);    //  update PXB area from PXB, any scale

//  main window menu and toolbar builder (f.menus.cc)

void build_menus();

//  file menu functions (f.file.cc)

void  Rclick_Popup_setup();                                                //  setup mouse right-click popup menus
void  popup_menufunc(GtkWidget *, cchar *menu);                            //  image/thumb right-click menu func
void  image_Rclick_popup(int Nth);                                         //  popup menu for image right-click
void  gallery_Lclick_func(int Nth);                                        //  function for thumbnail left-click
void  gallery_Rclick_popup(int Nth);                                       //  popup menu for thumbnail right-click
void  copyto_clipboard(GtkWidget *, cchar *file);                          //  copy an image file to the clipboard
void  m_clone(GtkWidget *, cchar *);                                       //  start parallel fotoxx instance
void  m_open(GtkWidget *, cchar *);                                        //  open image file in same window
void  m_open_drag(int x, int y, char *file);                               //  open drag-drop file
void  m_previous(GtkWidget *, cchar *);                                    //  open previously accessed file
void  m_recent(GtkWidget *, cchar *);                                      //  open recently accessed file
void  m_prev(GtkWidget *, cchar *);                                        //  open previous file in gallery
void  m_next(GtkWidget *, cchar *);                                        //  open next file in gallery
void  m_ufraw(GtkWidget *, cchar *);                                       //  open, edit camera RAW file with ufraw
void  m_save(GtkWidget *, cchar *);                                        //  save modified image to same file
void  m_savevers(GtkWidget *, cchar *);                                    //  save modified image to new file version
void  m_saveas(GtkWidget *, cchar *);                                      //  save modified image to another file
void  m_create(GtkWidget *, cchar *);                                      //  create new blank image file
void  m_rename(GtkWidget *, cchar *);                                      //  rename an image file (same location)
void  m_copyto(GtkWidget *, cchar *);                                      //  copy an image file to a new location
void  m_moveto(GtkWidget *, cchar *);                                      //  move an image file to a new location
void  m_delete(GtkWidget *, cchar *);                                      //  delete an image file (undo not poss.)
void  m_trash(GtkWidget *, cchar *);                                       //  move an image file to trash
void  m_print(GtkWidget *, cchar *);                                       //  print an image file
void  m_quit(GtkWidget *, cchar *);                                        //  exit application with checks
void  quitxx();                                                            //  exit unconditionally
void  add_recent_file(cchar *file);                                        //  add file to recent file list
void  set_mwin_title();                                                    //  main window title = curr_file path

//  image file load and save functions (f.file.cc)

char * raw_to_tiff(cchar *rawfile);                                        //  get corresp. tiff file for a RAW file
void   f_preload(int next);                                                //  start preload of next file
int    f_open(cchar *file, int lock, int Nth = 0, int keep = 0,            //  open new current file
                                                  int Fack = 1);
void   f_open_ufraw(GtkWidget *, cchar *);                                 //  open RAW file with ufraw
int    f_save(char *outfile, cchar *type, int bpc, int Fack = 1);          //  save current image file to disk

PXB * PXB_load(cchar *filespec, int fack = 0);                             //  load file to PXB pixmap, 8 bpc
PXM * PXM_load(cchar *filespec, int fack = 0);                             //  load file to PXM pixmap, 8/16 bpc

PXB * TIFF_PXB_load(cchar *filespec);                                      //  TIFF file to PXB, 8 bpc
PXM * TIFF_PXM_load(cchar *filespec);                                      //  TIFF file to PXM, 8/16 bpc
int   PXM_TIFF_save(PXM *pxm, cchar *filespec, int bpc);                   //  PXM to TIFF file, 8/16 bpc

PXB * PNG_PXB_load(cchar *filespec);                                       //  PNG file to PXB, 8 bpc
PXM * PNG_PXM_load(cchar *filespec);                                       //  PNG file to PXM, 8/16 bpc
int   PXM_PNG_save(PXM *pxm, cchar *filespec, int bpc);                    //  PXM to PNG file, 8/16 bpc

PXB * ANY_PXB_load(cchar *filespec);                                       //  ANY file to PXB, 8 bpc
PXM * ANY_PXM_load(cchar *filespec);                                       //  ANY file to PXM, 8 bpc
int   PXM_ANY_save(PXM *pxm, cchar *filespec);                             //  PXM to ANY file, 8 bpc

//  thumbnail gallery and navigation functions (f.file.cc)

char * gallery(cchar *filez, cchar *action, int Nth = 0);                  //  display image gallery window, navigate
void   set_gwin_title();                                                   //  main window title = gallery name
int    gallery_position(cchar *file, int Nth);                             //  get rel. position of file in gallery
int    image_file_type(cchar *file);                                       //  get file type: dirk, image, RAW, other
char * thumb2imagefile(cchar *thumbfile);                                  //  get image file for thumbnail file
char * image2thumbfile(cchar *imagefile);                                  //  get thumbnail file for image file
char * image_thumbfile(char *imagefile);                                   //  get thumbnail filespec, create if nec.
GdkPixbuf * image_thumbnail(char *imagefile, int size = 0);                //  get thumbnail pixbuf, create if nec.
GdkPixbuf * get_thumbnail_pixbuf(char *imagefile, int size);               //  make thumbnail pixbuf from image file
char * gallery_find(int Nth);                                              //  return Nth file in curr. gallery
void   gallery_monitor(cchar *action);                                     //  start/stop gallery directory monitor
char ** gallery_getfiles();                                                //  select files from gallery window
void   gallery_getfiles_Lclick_func(int Nth);                              //  get files, thumbnail left-click
void   gallery_getfiles_Rclick_func(int Nth);                              //  get files, thumbnail right-click

//  plugin functions (f.file.cc)

void  m_edit_plugins(GtkWidget *, cchar *);                                //  add and remove plugin functions
void  m_run_plugin(GtkWidget *, cchar *);                                  //  perform a plugin edit function

//  tools menu functions (f.tools.cc)

void  m_fg_color(GtkWidget *, cchar *);                                    //  toggle FG color (gridlines, areas ...)
void  m_batch_rename(GtkWidget *, cchar *);                                //  rename many files, base + sequence
void  m_batch_convert(GtkWidget *, cchar *);                               //  convert/resize/export N image files
void  m_batch_conv_raw(GtkWidget *, cchar *);                              //  convert RAW files to tiff
void  m_manage_coll(GtkWidget *, cchar *);                                 //  create and edit image collections
int   edit_coll_mousefunc(int Nth);                                        //  gallery mouse click function
void  m_move_coll(GtkWidget *, cchar *);                                   //  change top directory of a collection
void  m_slideshow(GtkWidget *, cchar *);                                   //  enter or leave slideshow mode
void  slideshow_next(cchar *);                                             //  show prev/next image
void  m_syncfiles(GtkWidget *, cchar *);                                   //  rebuild thumbnails and search index
int   syncfiles_func(int mode);                                            //  sync files function
void  m_show_RGB(GtkWidget *, cchar *);                                    //  show RGB values at mouse click
void  m_gridlines(GtkWidget *, cchar *);                                   //  set up grid lines
void  load_grid(int *griddata);                                            //  load grid from saved state
void  save_grid(int *griddata);                                            //  save grid to saved state
void  toggle_grid(int action);                                             //  set grid off/on or toggle (0/1/2)
void  m_burn(GtkWidget *, cchar *);                                        //  burn selected images to CD/DVD
void  m_moncheck(GtkWidget *, cchar *);                                    //  check monitor brightness and color
void  m_mongamma(GtkWidget *, cchar *);                                    //  adjust monitor gamma
void  m_histogram(GtkWidget *, cchar *);                                   //  start brightness histogram
void  histogram_paint(GtkWidget *, cairo_t *, int);                        //  brightness hist. draw signal function
void  histogram_destroy();                                                 //  remove histogram window
void  m_lang(GtkWidget *, cchar *);                                        //  change language
void  m_translate(GtkWidget *, cchar *);                                   //  edit translation
void  m_menu_launcher(GtkWidget *, cchar *);                               //  make desktop menu entry and launcher
void  m_settings(GtkWidget *, cchar *);                                    //  user settings
void  m_resources(GtkWidget *, cchar *);                                   //  report CPU and memory usage

//  metadata menu functions (EXIF, IPTC, etc.) (f.meta.cc) 

void  m_edit_cctext(GtkWidget *, cchar *);                                 //  edit caption and user comments dialog
void  m_edit_metadata(GtkWidget *, cchar *);                               //  edit date/rating/tags dialog
void  manage_tags();                                                       //  manage tags dialog function
void  load_filemeta(cchar *file);                                          //  EXIF/IPTC >> tags_* in memory
void  save_filemeta(cchar *file);                                          //  tags_* in memory >> EXIF/IPTC
void  update_search_index(cchar *file);                                    //  update search index file 
void  delete_search_index(cchar *file);                                    //  delete entry from search index file
void  m_batchAddTags(GtkWidget *, cchar *);                                //  add tags to many files at once
void  m_batchDelTag(GtkWidget *, cchar *);                                 //  delete/replace a tag for many files
void  m_meta_view_short(GtkWidget *, cchar *);                             //  view selected EXIF/IPTC data
void  m_meta_view_long(GtkWidget *, cchar *);                              //  view all EXIF/IPTC data
void  meta_view(int length);                                               //  view EXIF/IPTC data in popup window
void  m_meta_edit_any(GtkWidget *, cchar *);                               //  add or change any EXIF/IPTC data
void  m_meta_delete(GtkWidget *, cchar *);                                 //  delete EXIF/IPTC data

int   init_geotags();                                                      //  set up world map and geolocations
void  m_download_geolocs(GtkWidget *, cchar *);                            //  download world map and geolocations
void  m_worldmap_test(GtkWidget *, cchar *);                               //  test world map coordinate conversions
void  m_edit_geotags(GtkWidget *, cchar *);                                //  edit geotags in image file EXIF
void  m_batch_add_geotags(GtkWidget *, cchar *menu);                       //  batch add geotags
void  m_geotag_groups(GtkWidget *, cchar *);                               //  report images grouped by date, geotags
void  m_geotag_worldmap(GtkWidget *, cchar *);                             //  report images by clicking world map
void  m_search_images(GtkWidget *, cchar *);                               //  search image metadata

char ** exif_get(cchar *file, cchar **keys, int nkeys);                    //  get EXIF/IPTC data for given key(s)
int     exif_put(cchar *file, cchar **keys, cchar **text, int nkeys);      //  put EXIF/IPTC data for given key(s)
int     exif_copy(cchar *f1, cchar *f2, cchar **k, cchar **t, int nk);     //  copy EXIF/IPTC data from file to file
char ** exiftool_server(char **inputs);                                    //  run exiftool as a server process

int  read_search_index_seq(sxrec_t &sxrec, int &ftf);                      //  read search index recs. 1-last  v.12.10
int  write_search_index_seq(sxrec_t *sxrec, int &ftf);                     //  write search index recs. 1-last v.12.10
int  get_search_index_rec(sxrec_t &sxrec, cchar *file);                    //  get search index rec.           v.13.01
int  get_search_index_rec_min(cchar *file, char *pdate, char *size);       //   " min. for gallery paint       v.13.01
int  put_search_index_rec(sxrec_t *sxrec, cchar *file);                    //  add/update search index rec.    v.12.10

//  select area menu functions (f.area.cc)

void  m_select(GtkWidget *, cchar *);                                      //  select area within image
void  m_select_show(GtkWidget *, cchar *);                                 //  enable area for subsequent edits
void  m_select_hide(GtkWidget *, cchar *);                                 //  show area outline
void  m_select_enable(GtkWidget *, cchar *);                               //  hide area outline
void  m_select_disable(GtkWidget *, cchar *);                              //  disable area
void  m_select_invert(GtkWidget *, cchar *);                               //  invert area 
void  m_select_unselect(GtkWidget *, cchar *);                             //  unselect area 
void  m_select_copy(GtkWidget *, cchar *);                                 //  copy area, save in memory
void  m_select_paste(GtkWidget *, cchar *);                                //  paste saved area into image
void  m_select_open(GtkWidget *, cchar *);                                 //  open an area file, save in memory
void  m_select_save(GtkWidget *, cchar *);                                 //  save area to a file

void  sa_geom_mousefunc();                                                 //  select rectangle or ellipse
void  sa_draw_mousefunc();                                                 //  line drawing functions
int   sa_nearpix(int mx, int my, int rad, int &px, int &py, int fx);       //  find nearest pixel
void  sa_draw_line(int px1, int py1, int px2, int py2);                    //  draw a connected line
void  sa_draw1pix(int px, int py);                                         //  add one pixel to select area
void  sa_mouse_select();                                                   //  select by mouse (opt. color match)
void  sa_nextseq();                                                        //  start next sequence number
void  sa_unselect_pixels();                                                //  remove current selection
void  sa_finish();                                                         //  finish - map interior pixels
void  sa_finish_auto();                                                    //  finish - interior pixels already known
void  sa_unfinish();                                                       //  set finished area back to edit mode
void  sa_map_pixels();                                                     //  map edge and interior pixels
void  sa_show(int flag);                                                   //  show or hide area outline
void  sa_show_rect(int px1, int py1, int ww, int hh);                      //  show area outline inside a rectangle
int   sa_validate();                                                       //  validate area for curr. image context
void  sa_enable();                                                         //  enable area
void  sa_disable();                                                        //  disable area
void  sa_invert();                                                         //  invert area
void  sa_unselect();                                                       //  unselect area
void  sa_edgecalc();                                                       //  calculate edge distances
   
//  image edit functions - transform menu (f.xform.cc)

void  m_rotate(GtkWidget *, cchar *);                                      //  rotate image
void  m_trim(GtkWidget *, cchar *);                                        //  trim image
void  m_autotrim(GtkWidget *, cchar *);                                    //  autotrim image after rotate, warp
void  m_resize(GtkWidget *, cchar *);                                      //  resize image
void  m_write_text(GtkWidget *, cchar *);                                  //  write text on image
void  m_flip(GtkWidget *, cchar *);                                        //  flip horizontally or vertically
void  m_colormode(GtkWidget *, cchar *);                                   //  B+W/color, negative/positive, sepia
void  m_unbend(GtkWidget *, cchar *);                                      //  fix perspective problems
void  m_keystone(GtkWidget *, cchar *);                                    //  convert tetragon image to rectangle
void  m_warp_area(GtkWidget *, cchar *);                                   //  warp image within an area
void  m_warp_curved(GtkWidget *, cchar *);                                 //  warp image, curved transform
void  m_warp_linear(GtkWidget *, cchar *);                                 //  warp image, linear transform
void  m_warp_affine(GtkWidget *, cchar *);                                 //  warp image, affine transform

//  image edit functions - retouch menu (f.retouch.cc)

void  m_voodoo(GtkWidget *, cchar *);                                      //  automatic image retouch
void  m_tune(GtkWidget *, cchar *);                                        //  brightness / color adjustments
void  m_gamma(GtkWidget *, cchar *);                                       //  gamma adjustments
void  m_tonemap(GtkWidget *, cchar *);                                     //  tone mapping
void  m_flatten(GtkWidget *, cchar *);                                     //  flatten brightness distribution
void  m_xbrange(GtkWidget *, cchar *);                                     //  clip brightness, expand rest
void  m_whitebal(GtkWidget *, cchar *);                                    //  adjust white balance
void  m_redeye(GtkWidget *, cchar *);                                      //  red-eye removal
void  m_blur(GtkWidget *, cchar *);                                        //  blur image
void  m_sharpen(GtkWidget *, cchar *);                                     //  sharpen image
void  m_denoise(GtkWidget *, cchar *);                                     //  image noise reduction
void  m_smart_erase(GtkWidget *, const char *);                            //  smart erase object
void  m_vignette(GtkWidget *, cchar *);                                    //  vignette tool
void  m_match_color(GtkWidget *, cchar *);                                 //  set image2 colors to match image1
void  m_CMYK(GtkWidget *, cchar *);                                        //  brightness / color adjust with OD units
void  m_revise_RGB(GtkWidget *, cchar *);                                  //  revise RGB values
void  m_brightramp(GtkWidget *, cchar *);                                  //  ramp brightness across image
void  m_dust(GtkWidget *, const char *);                                   //  remove dust
void  m_stuckpix(GtkWidget *, const char *);                               //  fix stuck pixels
void  m_chromatic(GtkWidget *, const char *);                              //  reduce chromatic abberation
void  m_paint_pixels(GtkWidget *, cchar *);                                //  paint pixels with the mouse
void  m_lever_edits(GtkWidget *, cchar *);                                 //  leverage edits by pixel bright/color
void  m_paint_edits(GtkWidget *, cchar *);                                 //  select and edit in parallel

//  image edit functions - art menu (f.art.cc)

void  m_colordep(GtkWidget *, cchar *);                                    //  set color depth 1-16 bits/color
void  m_draw(GtkWidget *, cchar *);                                        //  make simulated drawing
void  m_outlines(GtkWidget *, cchar *);                                    //  make an outline image
void  m_emboss(GtkWidget *, cchar *);                                      //  make simulated embossing
void  m_tiles(GtkWidget *, cchar *);                                       //  make simulated tiles (pixelate)
void  m_dots(GtkWidget *, cchar *);                                        //  convert image to dot matrix form
void  m_painting(GtkWidget *, cchar *);                                    //  make simulated painting
void  m_shiftRGB(GtkWidget *, cchar *);                                    //  shift one color into another color

//  image edit functions - combine menu (f.comp.cc)

void  m_HDR(GtkWidget *, cchar *);                                         //  make HDR composite image
void  m_HDF(GtkWidget *, cchar *);                                         //  make HDF composite image
void  m_STP(GtkWidget *, cchar *);                                         //  stack / paint image
void  m_STN(GtkWidget *, cchar *);                                         //  stack / noise reduction
void  m_pano(GtkWidget *, cchar *);                                        //  make panorama composite image
void  m_vpano(GtkWidget *, cchar *);                                       //  make vertical panorama

//  translatable strings used in multiple dialogs

#define Babsolute ZTX("absolute")
#define Badd ZTX("Add")
#define Baddall ZTX("Add All")
#define Ball ZTX("All")
#define Bamount ZTX("Amount")
#define Bapply ZTX("Apply")
#define Bautocomp ZTX("Autocomplete")
#define Bblack ZTX("Black")
#define Bblendwidth ZTX("Blend Width")
#define Bblue ZTX("Blue")
#define Bbrightness ZTX("Brightness")
#define Bbrowse ZTX("Browse")
#define Bcancel ZTX("Cancel")
#define Bchoose ZTX("Choose")
#define Bclear ZTX("Clear")
#define Bcolor ZTX("Color")
#define Bcopy ZTX("Copy")
#define Bcurvefile ZTX("Curve File:")
#define Bcut ZTX("Cut")
#define Bdarker ZTX("Darker Areas")
#define Bdelete ZTX("Delete")
#define Bdisable ZTX("Disable")
#define Bdone ZTX("Done")
#define Bedit ZTX("Edit")
#define Benable ZTX("Enable")
#define Berase ZTX("Erase")
#define Bfetch ZTX("Fetch")
#define Bfind ZTX("Find")
#define Bfinish ZTX("Finish")
#define Bfont ZTX("Font")
#define Bgreen ZTX("Green")
#define Bheight ZTX("Height")
#define Bhelp ZTX("Help")
#define Bhistogram ZTX("histogram")
#define Bhide ZTX("Hide")
#define Binsert ZTX("Insert")
#define Binvert ZTX("Invert")
#define Blighter ZTX("Lighter Areas")
#define Blimit ZTX("limit")
#define Bnew ZTX("New")
#define Bnext ZTX("Next")
#define Bnosearchindex ZTX("search index file missing")
#define BOK ZTX("OK")
#define Bopen ZTX("Open")
#define Bopenrawfile ZTX("Open RAW File")
#define Bpaste ZTX("Paste")
#define Bpause ZTX("Pause")
#define Bpercent ZTX("Percent")
#define Bpresets ZTX("Presets")
#define Bprev ZTX("Prev")
#define Bproceed ZTX("Proceed")
#define Bradius ZTX("Radius")
#define Brange ZTX("range")
#define Bred ZTX("Red")
#define Bredo ZTX("Redo")
#define Breduce ZTX("Reduce")
#define Bremove ZTX("Remove")
#define Breset ZTX("Reset")
#define Brotate ZTX("Rotate")
#define Bsave ZTX("Save")
#define Bsavetoedit ZTX("Unknown file type, save as tiff/jpeg/png to edit")
#define Bsearch ZTX("Search")
#define Bselect ZTX("Select")
#define Bselectfiles ZTX("Select Files")
#define Bshow ZTX("Show")
#define Bstart ZTX("Start")
#define Bthresh ZTX("Threshold")
#define Btoomanyfiles ZTX("exceed %d files")
#define Bundoall ZTX("Undo All")
#define Bundolast ZTX("Undo Last")
#define Bundo ZTX("Undo")
#define Bunfinish ZTX("Unfinish")
#define Bunselect ZTX("Unselect")
#define Bview ZTX("View")
#define Bweb ZTX("Web")
#define Bwhite ZTX("White")
#define Bwidth ZTX("Width")
#define Bmap ZTX("Map")


