/**************************************************************************

   Fotoxx      edit photos and manage collections

   Copyright 2007 2008 2009 2010 2011 2012 Michael Cornelison
   Source URL: http://kornelix.com/fotoxx
   Contact: kornelix2@gmail.com
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see http://www.gnu.org/licenses/.

***************************************************************************/

#define EX extern                                                          //  enable extern declarations
#include "fotoxx.h"


/**************************************************************************
   Fotoxx image editor - image metadata functions, EXIF/IPTC etc.
***************************************************************************/

int   get_mouse_tag(GtkTextView *, int px, int py, cchar *);               //  get tag selected by mouse
int   add_tag(char *tag, char *taglist, int maxcc);                        //  add tag if unique and enough space
int   del_tag(char *tag, char *taglist);                                   //  remove tag from tag list
int   add_recentag(char *tag);                                             //  add tag to recent tags, keep recent
void  load_deftags();                                                      //  tags_defined file >> tags_deftags[]
void  save_deftags();                                                      //  tags_deftags[] >> defined_tags file
int   find_deftag(char *tag);                                              //  find tag in tags_deftags[]
int   add_deftag(char *catg, char *tag);                                   //  add tag to tags_deftags[]
int   del_deftag(char *tag);                                               //  remove tag from tags_deftags[]
void  deftags_stuff(zdialog *zd);                                          //  tags_deftags[] >> dialog widget

char     meta_date[12] = "";                                               //  image date, yyyymmdd
char     meta_prdate[12] = "";                                             //  previous image date read or set
char     meta_rating[4] = "0";                                             //  image rating in stars, "0" to "5"
char     meta_size[16] = "";                                               //  image size, "NNNNxNNNN"
char    *tags_deftags[maxtagcats];                                         //  defined tags: catg: tag1, ... tagN,
char     tags_deftext[tagTcc] = "";                                        //  defined tags as flat text buffer
char     tags_currtags[tagFcc] = "";                                       //  tags for current image file
char     tags_recentags[tagRcc] = "";                                      //  recently added tags list
char     tags_batchAddTags[tagMcc] = "";                                   //  batch add tags list
char     tags_searchtags[tagScc] = "";                                     //  search tags list
char     tags_searchtext[tagScc] = "";                                     //  search comments & captions word list
char     meta_searchfiles[tagScc] = "";                                    //  search files list
char     meta_comments[exif_maxcc];                                        //  image comments            expanded
char     meta_caption[exif_maxcc];                                         //  image caption             v.12.06
char     meta_city[100], meta_country[100];                                //  geotags: city, country
char     meta_latitude[20], meta_longitude[20];                            //  geotags: lati/longitude (-123.4567)


/**************************************************************************/

//  Private function for other metadata functions.
//  Check if synchronization of search index file is enabled and
//  provide explanation for user for metadata functions that are blocked.
//  Returns 1 if sync is enabled (allow function) and 0 if not (block function).

int check_index_sync()
{
   #define blockmess    \
      "Metadata functions do not work unless file sync is done. \n"    \
      "Use Tools > Synchronize Files to perform file sync."

   if (FsyncEnabled && ! FsyncNeeded) return 1;                            //  file sync enabled, not needed   v.12.12
   zmessageACK(mWin,ZTX(blockmess));
   return 0;
}


/**************************************************************************/

//  View and edit IPTC Caption and EXIF UserComment data

void m_edit_cctext(GtkWidget *, cchar *menu)                               //  2 functions combined
{
   int      edit_cctext_dialog_event(zdialog *zd, cchar *event);
   char     text[exif_maxcc+50];
   cchar    *title = ZTX("Edit Caption and Comments");

   if (menu) zfuncs::F1_help_topic = "edit_cctext";
   if (! check_index_sync()) return;                                       //  index fle sync disabled      v.12.08
   if (! curr_file) return;

   if (! zdeditcctext)                                                     //  popup dialog if not already
   {   
      zdeditcctext = zdialog_new(title,mWin,Bapply,Bcancel,null);
      zdialog *zd = zdeditcctext;
      zdialog_add_widget(zd,"hbox","hbcap","dialog");
      zdialog_add_widget(zd,"label","labcap","hbcap",ZTX("Caption"),"space=5");
      zdialog_add_widget(zd,"frame","framecap","hbcap",0,"expand");
      zdialog_add_widget(zd,"edit","textcap","framecap",0,"space=5|wrap");    
      zdialog_add_widget(zd,"hbox","hbcom","dialog");
      zdialog_add_widget(zd,"label","labcom","hbcom",ZTX("Comments"),"space=5");
      zdialog_add_widget(zd,"frame","framecom","hbcom",0,"expand");
      zdialog_add_widget(zd,"edit","textcom","framecom",0,"space=5|wrap");    
      zdialog_resize(zd,600,0);
      zdialog_help(zd,"edit_cctext");                                      //  zdialog help topic
      zdialog_run(zd,edit_cctext_dialog_event);
   }
   
   load_filemeta(curr_file);                                               //  get current caption and comments
   repl_1str(meta_caption,text,"\\n","\n");                                //  replace "\n" with real newlines
   zdialog_stuff(zdeditcctext,"textcap",text);                             //  stuff into dialog
   repl_1str(meta_comments,text,"\\n","\n");                               //  replace "\n" with real newlines
   zdialog_stuff(zdeditcctext,"textcom",text);                             //  stuff into dialog

   return;
}


//  dialog event and completion callback function

int  edit_cctext_dialog_event(zdialog *zd, cchar *event)
{
   char     text[exif_maxcc+50];
   
   if (! zd->zstat) return 0;
   
   else if (zd->zstat == 1)                                                //  apply, save text
   {
      zd->zstat = 0;                                                       //  keep dialog active
      zdialog_fetch(zd,"textcap",text,exif_maxcc);                         //  get new cctext
      repl_1str(text,meta_caption,"\n","\\n");                             //  replace newlines with "\n"
      zdialog_fetch(zd,"textcom",text,exif_maxcc);                         //  get new cctext
      repl_1str(text,meta_comments,"\n","\\n");                            //  replace newlines with "\n"
      Fmetachanged = 1;
      save_filemeta(curr_file);                                            //  save in image file
   }

   else zdialog_free(zdeditcctext);                                        //  cancel

   return 1;
}


/**************************************************************************/

//  edit metadata menu function

void m_edit_metadata(GtkWidget *, cchar *menu)
{
   void  edit_filetags_clickfunc(GtkWidget *widget, int &line, int &pos);
   void  edit_recentags_clickfunc(GtkWidget *widget, int &line, int &pos);
   void  edit_deftags_clickfunc(GtkWidget *widget, int &line, int &pos);
   int   editmeta_dialog_event(zdialog *zd, cchar *event);

   char        *ppv, starsN[8];
   cchar       *mantags = ZTX("Manage Tags");
   GtkWidget   *widget;
   zdialog     *zd;

   if (menu) zfuncs::F1_help_topic = "edit_tags";
   if (! check_index_sync()) return;                                       //  index fle sync disabled      v.12.08
   if (! curr_file) return;

   if (! zdeditmeta)                                                       //  (re) start tag edit dialog 
   {
      zd = zdialog_new(ZTX("Edit Metadata"),mWin,mantags,Bapply,Bcancel,null);
      zdeditmeta = zd;

      zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
      zdialog_add_widget(zd,"label","labfile","hb1",ZTX("file:"),"space=5");
      zdialog_add_widget(zd,"label","file","hb1");

      zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");
      zdialog_add_widget(zd,"label","lab21","hb2",ZTX("image date (yyyymmdd)"),"space=5");
      zdialog_add_widget(zd,"entry","date","hb2",0,"scc=12");
      zdialog_add_widget(zd,"button","prdate","hb2",ZTX("use last"),"space=5");

      zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
      zdialog_add_widget(zd,"label","labstars","hb3",ZTX("image stars"),"space=5");
      zdialog_add_widget(zd,"vbox","vb3","hb3");
      zdialog_add_widget(zd,"hbox","hb31","vb3",0,"homog");
      zdialog_add_widget(zd,"hbox","hb32","vb3",0,"homog");
      zdialog_add_widget(zd,"label","lab30","hb31","0");
      zdialog_add_widget(zd,"label","lab31","hb31","1");
      zdialog_add_widget(zd,"label","lab32","hb31","2");
      zdialog_add_widget(zd,"label","lab33","hb31","3");
      zdialog_add_widget(zd,"label","lab34","hb31","4");
      zdialog_add_widget(zd,"label","lab35","hb31","5");
      zdialog_add_widget(zd,"radio","stars0","hb32",0);
      zdialog_add_widget(zd,"radio","stars1","hb32",0);
      zdialog_add_widget(zd,"radio","stars2","hb32",0);
      zdialog_add_widget(zd,"radio","stars3","hb32",0);
      zdialog_add_widget(zd,"radio","stars4","hb32",0);
      zdialog_add_widget(zd,"radio","stars5","hb32",0);

      zdialog_add_widget(zd,"hbox","hb4","dialog",0,"space=5");
      zdialog_add_widget(zd,"label","lab4","hb4",ZTX("current tags"),"space=5");
      zdialog_add_widget(zd,"frame","frame4","hb4",0,"expand");
      zdialog_add_widget(zd,"text","filetags","frame4",0,"expand|wrap");

      zdialog_add_widget(zd,"hbox","hb5","dialog",0,"space=5");
      zdialog_add_widget(zd,"label","recent","hb5",ZTX("recent tags"),"space=5");
      zdialog_add_widget(zd,"frame","frame5","hb5",0,"expand");
      zdialog_add_widget(zd,"text","recentags","frame5",0,"expand|wrap");

      zdialog_add_widget(zd,"hbox","hb8","dialog");
      zdialog_add_widget(zd,"label","labdeftags","hb8",ZTX("defined tags"),"space=5");
      zdialog_add_widget(zd,"hbox","hb9","dialog",0,"expand");
      zdialog_add_widget(zd,"frame","frame8","hb9",0,"space=5|expand");
      zdialog_add_widget(zd,"scrwin","scrwin8","frame8",0,"expand");
      zdialog_add_widget(zd,"text","deftags","scrwin8",0,"expand|wrap");

      load_deftags();                                                      //  stuff defined tags into dialog
      deftags_stuff(zd);

      widget = zdialog_widget(zd,"filetags");                              //  tag widget mouse functions   v.12.06
      textwidget_set_clickfunc(widget,edit_filetags_clickfunc);

      widget = zdialog_widget(zd,"recentags");
      textwidget_set_clickfunc(widget,edit_recentags_clickfunc);

      widget = zdialog_widget(zd,"deftags");
      textwidget_set_clickfunc(widget,edit_deftags_clickfunc);
      
      zdialog_resize(zd,500,500);                                          //  run dialog
      zdialog_help(zd,"edit_tags");                                        //  zdialog help topic 
      zdialog_run(zd,editmeta_dialog_event);
   }
   
   zd = zdeditmeta;

   load_filemeta(curr_file);                                               //  get EXIF/IPTC data

   ppv = (char *) strrchr(curr_file,'/');
   zdialog_stuff(zd,"file",ppv+1);                                         //  stuff dialog file name

   zdialog_stuff(zd,"date",meta_date);                                     //  stuff dialog data
   sprintf(starsN,"stars%c",meta_rating[0]);
   zdialog_stuff(zd,starsN,1);
   zdialog_stuff(zd,"filetags",tags_currtags);
   zdialog_stuff(zd,"recentags",tags_recentags);

   return;
}


//  mouse click functions for various text widgets for tags                //  v.12.06

void edit_filetags_clickfunc(GtkWidget *widget, int &line, int &pos)       //  existing image tag was clicked
{
   char     *txline, *txtag, end = 0;
   cchar    *delims = tagdelims":";
   
   txline = textwidget_get_line(widget,line);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,delims,end);
   if (! txtag) { zfree(txline); return; }
   
   del_tag(txtag,tags_currtags);                                           //  remove tag from image
   zdialog_stuff(zdeditmeta,"filetags",tags_currtags);
   
   zfree(txline);
   zfree(txtag);
   return;
}


void edit_recentags_clickfunc(GtkWidget *widget, int &line, int &pos)      //  recent tag was clicked
{
   char     *txline, *txtag, end = 0;
   cchar    *delims = tagdelims":";
   
   txline = textwidget_get_line(widget,line);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,delims,end);
   if (! txtag) { zfree(txline); return; }
   
   add_tag(txtag,tags_currtags,tagFcc);                                    //  add recent tag to image
   zdialog_stuff(zdeditmeta,"filetags",tags_currtags);
   
   zfree(txline);
   zfree(txtag);
   return;
}


void edit_deftags_clickfunc(GtkWidget *widget, int &line, int &pos)        //  defined tag was clicked
{
   char     *txline, *txtag, end = 0;
   cchar    *delims = tagdelims":";
   
   txline = textwidget_get_line(widget,line);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,delims,end);
   if (! txtag || end == ':') { zfree(txline); return; }                   //  tag category clicked, ignore
   
   add_tag(txtag,tags_currtags,tagFcc);                                    //  add new tag to image
   zdialog_stuff(zdeditmeta,"filetags",tags_currtags);                     //    from defined tags list

   add_recentag(txtag);                                                    //  and add to recent tags
   zdialog_stuff(zdeditmeta,"recentags",tags_recentags);
   
   zfree(txline);
   zfree(txtag);
   return;
}


//  dialog event and completion callback function

int editmeta_dialog_event(zdialog *zd, cchar *event)
{
   int         err;
   
   if (zd->zstat) {
      if (zd->zstat == 1) {                                                //  manage tags
         zd->zstat = 0;                                                    //  keep dialog active
         zdialog_show(zd,0);                                               //  hide parent dialog        v.12.11
         manage_tags();
         zdialog_show(zd,1);
         return 0;
      }

      if (zd->zstat == 2) {                                                //  [apply]
         zd->zstat = 0;                                                    //  keep dialog active
         save_filemeta(curr_file);                                         //  save tag changes
         return 0;
      }

      zdialog_free(zdeditmeta);                                            //  cancel - kill dialog
      return 0;
   }
   
   if (strEqu(event,"date")) {                                             //  image date revised
      err = zdialog_fetch(zd,"date",meta_date,11);
      if (err) return 1;
      if (strlen(meta_date) == 4) strcat(meta_date,"0101");                //  yyyy >> yyyy0101
      if (strlen(meta_date) == 6) strcat(meta_date,"01");                  //  yyyymm >> yyyymm01
      Fmetachanged++;
   }
   
   if (strEqu(event,"prdate")) {                                           //  repeat last date used
      if (*meta_prdate) {
         zdialog_stuff(zd,"date",meta_prdate);
         strcpy(meta_date,meta_prdate);
         Fmetachanged++;
      }
   }

   if (strnEqu(event,"stars",5)) {                                         //  event = stars0 to stars5
      meta_rating[0] = event[5];                                           //  "0" to "5"
      meta_rating[1] = 0;
      Fmetachanged++;
   }
   
   if (strEqu(event,"tags-changed"))                                       //  get new defined tags data
      deftags_stuff(zdeditmeta);

   return 0;
}


/**************************************************************************/

//  manage tags function - auxilliary dialog

zdialog  *zdmanagetags = 0;

void manage_tags()
{
   void  manage_deftags_clickfunc(GtkWidget *widget, int &line, int &pos);
   int   managetags_dialog_event(zdialog *zd, cchar *event);

   GtkWidget   *widget;
   zdialog     *zd;

   if (zdmanagetags) return;
   zd = zdialog_new(ZTX("Manage Tags"),mWin,Bdone,null);
   zdmanagetags = zd;

   zdialog_add_widget(zd,"hbox","hb7","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labcatg","hb7",ZTX("category"),"space=5");
   zdialog_add_widget(zd,"entry","catg","hb7",0,"scc=12");
   zdialog_add_widget(zd,"label","space","hb7",0,"space=5");
   zdialog_add_widget(zd,"label","labtag","hb7",ZTX("tag"),"space=5");
   zdialog_add_widget(zd,"entry","tag","hb7",0,"scc=20|expand");
   zdialog_add_widget(zd,"label","space","hb7",0,"space=5");
   zdialog_add_widget(zd,"button","create","hb7",ZTX("create"));
   zdialog_add_widget(zd,"button","delete","hb7",ZTX("delete"));

   zdialog_add_widget(zd,"hbox","hb8","dialog");
   zdialog_add_widget(zd,"label","labdeftags","hb8",ZTX("defined tags"),"space=5");
   zdialog_add_widget(zd,"hbox","hb9","dialog",0,"expand");
   zdialog_add_widget(zd,"frame","frame8","hb9",0,"space=5|expand");
   zdialog_add_widget(zd,"scrwin","scrwin8","frame8",0,"expand");
   zdialog_add_widget(zd,"text","deftags","scrwin8",0,"expand|wrap");

   widget = zdialog_widget(zd,"deftags");                                  //  tag widget mouse function    v.12.06
   textwidget_set_clickfunc(widget,manage_deftags_clickfunc);

   load_deftags();                                                         //  stuff defined tags into dialog
   deftags_stuff(zd);

   zdialog_resize(zd,0,400);
   zdialog_help(zd,"edit_tags");                                           //  zdialog help topic 
   zdialog_run(zd,managetags_dialog_event);                                //  run dialog
   zdialog_wait(zd);
   
   return;
}


//  mouse click functions for widget having tags                           //  v.12.06

void manage_deftags_clickfunc(GtkWidget *widget, int &line, int &pos)      //  tag or tag category was clicked
{
   char     *txline, *txtag, end = 0;
   cchar    *delims = tagdelims":";
   
   txline = textwidget_get_line(widget,line);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,delims,end);
   if (! txtag) { zfree(txline); return; }
   
   if (end == ':')                                                         //  tag category >> dialog widget
      zdialog_stuff(zdmanagetags,"catg",txtag);
   else 
      zdialog_stuff(zdmanagetags,"tag",txtag);                             //  selected tag >> dialog widget
   
   zfree(txline);
   zfree(txtag);
   return;
}


//  dialog event and completion callback function

int managetags_dialog_event(zdialog *zd, cchar *event)
{
   char        tag[tagcc], catg[tagcc];
   int         changed = 0;
   
   if (zd->zstat) {
      zdialog_free(zd);
      zdmanagetags = 0;                                                    //  bugfix     v.12.07
      return 0;
   }
   
   if (strEqu(event,"create")) {                                           //  add new tag to defined tags
      zdialog_fetch(zd,"catg",catg,tagcc);
      zdialog_fetch(zd,"tag",tag,tagcc);
      add_deftag(catg,tag);
      changed++;
   }

   if (strEqu(event,"delete")) {                                           //  remove tag from defined tags
      zdialog_fetch(zd,"tag",tag,tagcc);
      del_deftag(tag);
      changed++;
   }

   if (changed) {   
      save_deftags();                                                      //  save tag updates to file
      deftags_stuff(zd);                                                   //  update dialog "deftags" window
      if (zdeditmeta)                                                      //  and edit metadata window if active
         zdialog_send_event(zdeditmeta,"tags-changed");
   }

   return 0;
}


/**************************************************************************/

//  add input tag to output tag list if not already there and enough room
//  returns:   0 = added OK     1 = not unique (case ignored)
//             2 = overflow     3 = bad utf8 characters     4 = null tag

int add_tag(char *tag, char *taglist, int maxcc)
{
   char     *pp1, *pp2, tag1[tagcc], tag2[tagcc];
   int      cc, cc1, cc2;

   strncpy0(tag1,tag,tagcc);                                               //  remove leading and trailing blanks
   cc = strTrim2(tag2,tag1);
   if (! cc) return 4;

   if (utf8_check(tag2)) {                                                 //  check for valid utf8 encoding
      printf("*** bad utf8 characters: %s \n",tag2);
      return 3;
   }
   
   while ((pp1 = strpbrk(tag2,tagdelims":"))) *pp1 = '-';                  //  replace problem characters

   strcpy(tag,tag2);                                                       //  replace tag with sanitized version
   
   pp1 = taglist;
   cc1 = strlen(tag);

   while (true)                                                            //  check if already in tag list
   {
      while (*pp1 == ' ' || *pp1 == tagdelim1) pp1++;
      if (! *pp1) break;
      pp2 = pp1 + 1;
      while (*pp2 && *pp2 != tagdelim1) pp2++;
      cc2 = pp2 - pp1;
      if (cc2 == cc1 && strncaseEqu(tag,pp1,cc1)) return 1;
      pp1 = pp2;
   }

   cc2 = strlen(taglist);                                                  //  append to tag list if space enough
   if (cc1 + cc2 + 3 > maxcc) return 2;
   strcpy(taglist + cc2,tag);
   strcpy(taglist + cc2 + cc1, tagdelim2);                                 //  add delimiter + space

   if (taglist == tags_currtags)                                           //  image tags were changed
      Fmetachanged++;
   return 0;
}


//  remove tag from taglist, if present
//  returns: 0 if found and deleted, otherwise 1

int del_tag(char *tag, char *taglist)
{
   int         ii, ftcc, atcc, found;
   char        *temptags;
   cchar       *pp;
   
   temptags = strdupz(taglist,0,"del_tag");
   
   *taglist = 0;
   ftcc = found = 0;
   
   for (ii = 1; ; ii++)
   {
      pp = strField(temptags,tagdelims,ii);                                //  next tag
      if (! pp) {
         zfree(temptags);
         if (found && taglist == tags_currtags)                            //  image tags were changed
            Fmetachanged++;
         return 1-found;
      }
      if (*pp == ' ') continue;
      
      if (strcaseEqu(pp,tag)) {                                            //  skip matching tag
         found = 1;
         continue;
      }

      atcc = strlen(pp);                                                   //  copy non-matching tag
      strcpy(taglist + ftcc, pp);
      ftcc += atcc;
      strcpy(taglist + ftcc, tagdelim2);                                   //  + delim + blank
      ftcc += 2;
   }
}


//  add new tag to recent tags, if not already.
//  remove oldest to make space if needed.

int add_recentag(char *tag)
{
   int         err;
   char        *pp, temptags[tagRcc];

   err = add_tag(tag,tags_recentags,tagRcc);                               //  add tag to recent tags

   while (err == 2)                                                        //  overflow
   {
      strncpy0(temptags,tags_recentags,tagRcc);                            //  remove oldest to make room
      pp = strpbrk(temptags,tagdelims);
      if (! pp) return 0;
      strcpy(tags_recentags,pp+2);                                         //  delimiter + blank before tag
      err = add_tag(tag,tags_recentags,tagRcc);
   }

   return 0;
}


/**************************************************************************/

//  Load tags_defined file into tags_deftags[ii] => category: tag1, tag2, ...
//  Read search_index recs. and add unmatched tags: => nocatg: tag1, tag2, ...

void load_deftags()
{
   int tags_Ucomp(cchar *tag1, cchar *tag2);

   static int  Floaded = 0;
   FILE *      fid;
   sxrec_t     sxrec;
   int         ii, jj, ntags, err, cc, tcc, ftf;
   int         ncats, catoverflow;
   int         nocat, nocatcc;
   char        tag[tagcc], catg[tagcc];
   char        tagsbuff[tagGcc];
   char        *pp1, *pp2;
   char        ptags[tagntc][tagcc];
   
   if (Floaded) return;                                                    //  use memory tags if already there
   Floaded++;

   for (ii = 0; ii < maxtagcats; ii++)                                     //  clean memory
      tags_deftags[ii] = 0;

   ncats = catoverflow = 0;

   fid = fopen(tags_defined_file,"r");                                     //  read tags_defined file
   if (fid) {
      while (true) {
         pp1 = fgets_trim(tagsbuff,tagGcc,fid);
         if (! pp1) break;
         if (ncats == maxtagcats-1) goto toomanycats;
         pp2 = strchr(pp1,':');                                            //  isolate "category:"
         if (! pp2) continue;                                              //  reject bad data
         cc = pp2 - pp1 + 1;
         if (cc > tagcc-1) continue;
         strncpy0(catg,pp1,cc);                                            //  (for error message)
         if (strlen(pp1) > tagGcc-2) goto cattoobig;
         pp2++;
         while (*pp2 == ' ') pp2++;
         if (strlen(pp2) < 3) continue;
         while ((pp2 = strpbrk(pp2,tagdelims))) *pp2++ = tagdelim1;        //  force comma delimiter
         tags_deftags[ncats] = strdupz(pp1,0,"def_tags");                  //  tags_deftags[ii]
         ncats++;                                                          //   = category: tag1, tag2, ... tagN,
      }
      err = fclose(fid);
      if (err) goto deftagserr;
   }

   nocat = ncats;                                                          //  make last category "nocatg" for
   ncats++;                                                                //   unmatched tags in search_index recs.
   tags_deftags[nocat] = zmalloc(tagGcc,"deftags");
   strcpy(tags_deftags[nocat],"nocatg: ");
   nocatcc = 8;

   ftf = 1;                                                                //  read all search index recs.

   while (true)
   {
      err = read_search_index_seq(sxrec,ftf);                              //  v.12.10
      if (err) break;
      
      pp1 = sxrec.tags;                                                    //  may be "null,"
      
      while (true)
      {
         while (*pp1 && strchr(tagdelims" ",*pp1)) pp1++;                  //  next image tag start
         if (! *pp1) break;
         pp2 = strpbrk(pp1,tagdelims);                                     //  end
         if (! pp2) pp2 = pp1 + strlen(pp1);  
         cc = pp2 - pp1;
         if (cc > tagcc-1) {
            pp1 = pp2;
            continue;                                                      //  ignore huge tag
         }
         
         strncpy0(tag,pp1,cc+1);                                           //  look for tag in defined tags
         err = find_deftag(tag);
         if (! err) {                                                      //  found
            pp1 = pp2;
            continue;
         }

         if (nocatcc + cc + 2 > tagGcc-2) {
            catoverflow = 1;                                               //  nocatg: length limit reached
            break;
         }
         else {
            strcpy(tags_deftags[nocat] + nocatcc, tag);                    //  append tag to list
            nocatcc += cc;
            strcpy(tags_deftags[nocat] + nocatcc, tagdelim2);              //  + delim + blank
            nocatcc += 2;
         }

         pp1 = pp2;
      }

      zfree(sxrec.file);
      zfree(sxrec.tags);
      zfree(sxrec.comms);
      zfree(sxrec.capt);
      zfree(sxrec.gtags);
   }

   if (catoverflow) goto cattoobig;
   
//  parse all the tags in each category and sort in ascending order

   for (ii = 0; ii < ncats; ii++)
   {
      pp1 = tags_deftags[ii];
      pp2 = strchr(pp1,':');
      cc = pp2 - pp1 + 1;
      strncpy0(catg,pp1,cc);
      pp1 = pp2 + 1;
      while (*pp1 == ' ') pp1++;
      tcc = 0;

      for (jj = 0; jj < tagntc; jj++)
      {
         if (! *pp1) break;
         pp2 = strchr(pp1,tagdelim1);
         if (pp2) cc = pp2 - pp1;
         else cc = strlen(pp1);
         if (cc > tagcc-1) cc = tagcc-1;
         strncpy0(ptags[jj],pp1,cc+1);
         pp1 += cc + 1;
         tcc += cc;
         while (*pp1 == ' ') pp1++;
      }
      
      ntags = jj;
      if (ntags == tagntc) goto cattoobig;
      HeapSort((char *) ptags,tagcc,ntags,tags_Ucomp);

      pp1 = tags_deftags[ii];
      tcc += strlen(catg) + 2 + 2 * ntags + 2;                             //  category, all tags, delimiters
      pp2 = zmalloc(tcc,"deftags");

      tags_deftags[ii] = pp2;                                              //  swap memory
      zfree(pp1);

      strcpy(pp2,catg);
      pp2 += strlen(catg);
      strcpy(pp2,": ");                                                    //  pp2 = "category: "
      pp2 += 2;

      for (jj = 0; jj < ntags; jj++)                                       //  add the sorted tags
      {
         strcpy(pp2,ptags[jj]);                                            //  append tag + delim + blank
         pp2 += strlen(pp2);
         strcpy(pp2,tagdelim2);
         pp2 += 2;
      }
      
      *pp2 = 0;
   }
   
//  sort the categories in ascending order
//  leave "nocatg" at the end

   for (ii = 0; ii < ncats-1; ii++)
   for (jj = ii+1; jj < ncats-1; jj++) 
   {
      pp1 = tags_deftags[ii];
      pp2 = tags_deftags[jj];
      if (strcasecmp(pp1,pp2) > 0) {
         tags_deftags[ii] = pp2;
         tags_deftags[jj] = pp1;
      }
   }

   return;

toomanycats:
   zmessLogACK(mWin,"more than %d categories",maxtagcats);
   fclose(fid);
   return;
   
cattoobig:
   zmessLogACK(mWin,"category %s is too big",catg);
   fclose(fid);
   return;

deftagserr:
   zmessLogACK(mWin,"tags_defined file error: %s",strerror(errno));
   return;
}


//  compare function for tag sorting

int tags_Ucomp(cchar *tag1, cchar *tag2)
{
   return strcasecmp(tag1,tag2);
}


//  write tags_deftags[] memory data to the defined tags file if any changes were made

void save_deftags()
{
   int         ii, err;
   FILE        *fid;

   fid = fopen(tags_defined_file,"w");                                     //  write tags_defined file
   if (! fid) goto deftagserr;

   for (ii = 0; ii < maxtagcats; ii++)
   {
      if (! tags_deftags[ii+1]) break;                                     //  omit last category, "nocatg"
      err = fprintf(fid,"%s\n",tags_deftags[ii]);                          //  each record: 
      if (err < 0) goto deftagserr;                                        //    category: tag1, tag2, ... tagN,
   }

   err = fclose(fid);
   if (err) goto deftagserr;
   return;
   
deftagserr:
   zmessLogACK(mWin,"tags_defined file error: %s",strerror(errno));
   return;
}


//  find a given tag in tags_deftags[]
//  return: 0 = found, 1 = not found

int find_deftag(char *tag)
{
   int      ii, cc;
   char     tag2[tagcc+4];
   char     *pp;

   if (! tag || *tag <= ' ') return 0;                                     //  bad tag

   strncpy0(tag2,tag,tagcc);                                               //  construct tag + delim + blank
   cc = strlen(tag2);
   strcpy(tag2+cc,tagdelim2);
   cc += 2;

   for (ii = 0; ii < maxtagcats; ii++)
   {
      pp = tags_deftags[ii];                                               //  category: tag1, tag2, ... tagN, 
      if (! pp) return 1;                                                  //  not found

      while (pp)
      {
         pp = strcasestr(pp,tag2);                                         //  look for delim + blank + tag + delim
         if (! pp) break;
         if (strchr(tagdelims":", pp[-2])) return 0;
         pp += cc;
      }
   }

   return 1;
}


//  add new tag to tags_deftags[] >> category: tag1, tag2, ... newtag,
//  returns:   0 = added OK     1 = not unique (case ignored)
//             2 = overflow     3 = bad utf8 characters     4 = null tag
//  if tag present under another category, it is moved to new category

int add_deftag(char *catg, char *tag)
{
   int         ii, cc, cc1, cc2;
   char        tag1[tagcc], tag2[tagcc];
   char        *pp1, *pp2;
   
   strncpy0(tag1,tag,tagcc);                                               //  remove leading and trailing blanks
   cc = strTrim2(tag2,tag1);
   if (! cc) return 4;

   if (utf8_check(tag2)) {                                                 //  check for valid utf8 encoding
      printf("*** bad utf8 characters: %s \n",tag2);
      return 3;
   }

   while ((pp1 = strpbrk(tag2,tagdelims":"))) *pp1 = '-';                  //  replace problem characters
   
   strcpy(tag,tag2);                                                       //  replace tag with sanitized version

   del_deftag(tag);                                                        //  delete if already there

   if (! catg || *catg <= ' ') catg = (char *) "nocatg";

   cc1 = strlen(catg);

   for (ii = 0; ii < maxtagcats; ii++)                                     //  look for given category
   {
      pp1 = tags_deftags[ii];
      if (! pp1) goto newcatg;
      if (! strnEqu(catg,pp1,cc1)) continue;
      if (pp1[cc1] == ':') goto oldcatg;
   }

newcatg:
   if (ii == maxtagcats) goto toomanycats;
   cc1 = strlen(catg) + strlen(tag) + 6;
   pp1 = zmalloc(cc1,"deftags");
   *pp1 = 0;
   strncatv(pp1,cc1,catg,": ",tag,tagdelim2,null);                         //  category: + tag + delim + blank
   tags_deftags[ii] = tags_deftags[ii-1];                                  //  move "nocatg" record to next slot
   tags_deftags[ii-1] = pp1;                                               //  insert new record before
   return 0;

oldcatg:
   cc1 = strlen(tag);
   pp2 = pp1 + 2;
   while (true) {
      pp2 = strcasestr(pp2,tag);                                           //  look for delim + blank + tag + delim
      if (! pp2) break;                                                    //        or colon + blank + tag + delim
      if (strchr(tagdelims,pp2[cc]) && strchr(tagdelims":", pp2[-2]))
         return 1;                                                         //  tag not unique
      pp2 += cc1;
   }
   cc2 = strlen(pp1);                                                      //  add new tag to old record
   if (cc1 + cc2 + 4 > tagGcc) goto cattoobig;
   pp2 = strdupz(pp1,cc2+4,"def_tags");                                    //  expand string
   zfree(pp1);
   tags_deftags[ii] = pp2;
   strcpy(pp2+cc2,tag);                                                    //  old record + tag + delim + blank
   strcpy(pp2+cc2+cc1,tagdelim2);
   return 0;

toomanycats:
   zmessLogACK(mWin,"more than %d categories",maxtagcats);
   return 2;

cattoobig:
   zmessLogACK(mWin,"category is too big");
   return 2;
}


//  delete tag from defined tags list, tags_deftags[] 
//  return: 0 = found and deleted, 1 = not found

int del_deftag(char *tag)
{
   int      ii, cc;
   char     tag2[tagcc+4];
   char     *pp, *pp1, *pp2;

   if (! tag || *tag <= ' ') return 1;                                     //  bad tag

   strncpy0(tag2,tag,tagcc);                                               //  construct tag + delim + blank
   cc = strlen(tag2);
   strcpy(tag2+cc,tagdelim2);
   cc += 2;

   for (ii = 0; ii < maxtagcats; ii++)
   {
      pp = tags_deftags[ii];
      if (! pp) return 1;                                                  //  not found

      while (pp)
      {
         pp = strcasestr(pp,tag2);                                         //  look for prior delim or colon
         if (! pp) break;
         if (strchr(tagdelims":", pp[-2])) goto found;
         pp += cc;
      }
   }
   
found:
   for (pp1 = pp, pp2 = pp+cc; *pp2; pp1++, pp2++)                         //  eliminate tag, delim, blank
      *pp1 = *pp2;
   *pp1 = 0;
   
   return 0;
}


//  stuff tags_deftags[] into given text widget and format by category
//  create tags_deftext with flat list of tags for mouse clicking

void deftags_stuff(zdialog *zd)
{
   GtkWidget      *widget;
   GtkTextBuffer  *textbuff;
   GtkTextIter    iter1, iter2;
   int            ii, cc;
   char           catgname[tagcc+3];
   char           *pp1, *pp2;
   
   widget = zdialog_widget(zd,"deftags");
   wclear(widget);

   for (ii = 0; ii < maxtagcats; ii++)
   {
      pp1 = tags_deftags[ii];
      if (! pp1) break;
      pp2 = strchr(pp1,':');
      if (! pp2) continue;
      if (pp2 > pp1 + tagcc-3) continue;
      pp2 += 2;
      cc = pp2 - pp1 + 1;
      strncpy0(catgname,pp1,cc);
      wprintx(widget,0,catgname,"monospace bold 9");
      if (*pp2) wprintx(widget,0,pp2,"monospace 9");
      wprintx(widget,0,"\n");
   }
   
   textbuff = gtk_text_view_get_buffer(GTK_TEXT_VIEW(widget));
   gtk_text_buffer_get_bounds(textbuff,&iter1,&iter2);
   pp1 = gtk_text_buffer_get_text(textbuff,&iter1,&iter2,0);
   strncpy0(tags_deftext,pp1,tagTcc);

   return;
}


/**************************************************************************/

//  image file EXIF/IPTC data >> memory data:
//    meta_date, meta_rating, tags_currtags, meta_comments, meta_caption,
//    meta_city, meta_country, meta_latitude, meta_longitude

void load_filemeta(cchar *file)
{
   void exif_tagdate(cchar *exifdate, char *tagdate);

   int      ii, jj, cc;
   char     *pp;
   cchar    *exifkeys[10] = { exif_date_key, iptc_keywords_key, 
                              iptc_rating_key, exif_size_key,
                              exif_comment_key, iptc_caption_key,
                              exif_city_key, exif_country_key, 
                              exif_latitude_key, exif_longitude_key };

   char     **ppv, *imagedate, *imagekeywords, *imagestars, *imagesize;
   char     *imagecomms, *imagecapt;
   char     *imagecity, *imagecountry, *imagelatitude, *imagelongitude;

   *tags_currtags = *meta_date = *meta_comments = *meta_caption = 0;
   strcpy(meta_rating,"0");
   *meta_city = *meta_country = *meta_latitude = *meta_longitude = 0;
   
   ppv = exif_get(file,exifkeys,10);                                       //  get metadata from image file
   imagedate = ppv[0];
   imagekeywords = ppv[1];
   imagestars = ppv[2];
   imagesize = ppv[3];
   imagecomms = ppv[4];
   imagecapt = ppv[5];
   imagecity = ppv[6];
   imagecountry = ppv[7];
   imagelatitude = ppv[8];
   imagelongitude = ppv[9];
   
   if (imagedate) {
      exif_tagdate(imagedate,meta_date);                                   //  EXIF date/time >> yyyymmdd
      strcpy(meta_prdate,meta_date);
      zfree(imagedate);
   }

   if (imagekeywords)
   {
      for (ii = 1; ; ii++)
      {
         pp = (char *) strField(imagekeywords,tagdelims,ii);
         if (! pp) break;
         if (*pp == ' ') continue;
         cc = strlen(pp);
         if (cc >= tagcc) continue;                                        //  reject tags too big
         for (jj = 0; jj < cc; jj++)
            if (pp[jj] > 0 && pp[jj] < ' ') break;                         //  reject tags with control characters
         if (jj < cc) continue;
         add_tag(pp,tags_currtags,tagFcc);                                 //  add to file tags if unique
      }

      zfree(imagekeywords);
   }

   if (imagestars) {
      meta_rating[0] = *imagestars;
      if (meta_rating[0] < '0' || meta_rating[0] > '5') meta_rating[0] = '0';
      meta_rating[1] = 0;
      zfree(imagestars);
   }
   
   if (imagesize) {
      strncpy0(meta_size,imagesize,15);
      zfree(imagesize);
   }

   if (imagecomms) {
      strncpy0(meta_comments,imagecomms,exif_maxcc);
      zfree(imagecomms);
   }

   if (imagecapt) {
      strncpy0(meta_caption,imagecapt,exif_maxcc);
      zfree(imagecapt);
   }
   
   if (imagecity) {                                                        //  geotags    v.12.05
      strncpy0(meta_city,imagecity,99);
      zfree(imagecity);
   }
   else strcpy(meta_city,"null");                                          //  replace missing data with "null"

   if (imagecountry) {
      strncpy0(meta_country,imagecountry,99);
      zfree(imagecountry);
   }
   else strcpy(meta_country,"null");

   if (imagelatitude) {
      strncpy0(meta_latitude,imagelatitude,12);
      zfree(imagelatitude);
   }
   else strcpy(meta_latitude,"null");

   if (imagelongitude) {
      strncpy0(meta_longitude,imagelongitude,12);
      zfree(imagelongitude);
   }
   else strcpy(meta_longitude,"null");
   
   Fmetachanged = 0;
   return;
}


//  add metadata in memory to image file EXIF/IPTC data and search_index recs.
//  update defined tags file if any changes

void save_filemeta(cchar *file)
{
   void tag_exifdate(cchar *tagdate, char *exifdate);
   
   cchar    *exifkeys[10] = { exif_date_key, iptc_keywords_key, 
                              iptc_rating_key, exif_size_key,
                              exif_comment_key, iptc_caption_key,
                              exif_city_key, exif_country_key, 
                              exif_latitude_key, exif_longitude_key };
   cchar       *exifdata[10];
   char        imagedate[24];

   if (! Fmetachanged) return;                                             //  no changes to tags etc.
   Fmetachanged = 0;

   *imagedate = 0;
   if (*meta_date) {
      tag_exifdate(meta_date,imagedate);                                   //  yyyymmdd >> EXIF date/time
      strcpy(meta_prdate,meta_date);
   }
   
   exifdata[0] = imagedate;                                                //  update file EXIF/IPTC data
   exifdata[1] = tags_currtags;
   exifdata[2] = meta_rating;
   exifdata[3] = meta_size;                                                //  v.13.01
   exifdata[4] = meta_comments;
   exifdata[5] = meta_caption;

   if (strEqu(meta_city,"null")) exifdata[6] = "";                         //  geotags                v.12.05
   else exifdata[6] = meta_city;                                           //  if "null" erase EXIF   v.12.07
   if (strEqu(meta_country,"null")) exifdata[7] = "";
   else exifdata[7] = meta_country;
   if (strEqu(meta_latitude,"null") || strEqu(meta_longitude,"null"))
      exifdata[8] = exifdata[9] = "";
   else {
      exifdata[8] = meta_latitude;
      exifdata[9] = meta_longitude;
   }

   exif_put(file,exifkeys,exifdata,10);                                    //  write EXIF

   update_search_index(file);                                              //  update search index file

   if (zdexifview) meta_view(0);                                           //  live EXIF/IPTC update
   return;
}


//  update search index record (replace updated file data)

void update_search_index(cchar *file)                                      //  overhauled    v.12.07
{
   char           imagedate[12], filedate[16];
   char           gtags[200];
   int            err;
   sxrec_t        sxrec;
   struct stat    statb;
   struct tm      bdt;

   //  build new metadata record to insert or replace

   err = stat(file,&statb);
   if (err) {
      zmessageACK(mWin,ZTX("file not found"));
      return;
   }
   
   memset(&sxrec,0,sizeof(sxrec_t));
   
   sxrec.file = (char *) file;                                             //  image filespec

   gmtime_r(&statb.st_mtime,&bdt);
   sprintf(filedate,"%04d%02d%02d%02d%02d%02d",                            //  file date = yyyymmddhhmmss
            bdt.tm_year + 1900, bdt.tm_mon + 1, bdt.tm_mday,
            bdt.tm_hour, bdt.tm_min, bdt.tm_sec);
   
   strcpy(sxrec.fdate,filedate);

   if (*meta_date) {                                                       //  image (photo) date 
      strncpy(imagedate,meta_date,4);
      strncpy(imagedate+5,meta_date+4,2);                                  //  yyyymmdd >> yyyy:mm:dd
      strncpy(imagedate+8,meta_date+6,2);
      imagedate[4] = imagedate[7] = ':';
      imagedate[10] = 0;
      strcpy(sxrec.pdate,imagedate);
   }

   sxrec.rating[0] = meta_rating[0];                                       //  rating '0' to '5' stars
   sxrec.rating[1] = 0;
   
   strncpy0(sxrec.size,meta_size,15);                                      //  v.13.01
   
   if (*tags_currtags)                                                     //  tags
      sxrec.tags = tags_currtags;
  
   if (*meta_caption)                                                      //  user caption
      sxrec.capt = meta_caption;
   
   if (*meta_comments)                                                     //  user comments
      sxrec.comms = meta_comments;

   if (! *meta_city) strcpy(meta_city,"null");                             //  geotags                v.12.05
   if (! *meta_country) strcpy(meta_country,"null");
   if (! *meta_latitude) strcpy(meta_latitude,"null");                     //  "null" for city/country is searchable
   if (! *meta_longitude) strcpy(meta_longitude,"null");
   
   snprintf(gtags,200,"%s^ %s^ %s^ %s",meta_city, meta_country, 
                                  meta_latitude, meta_longitude);
   sxrec.gtags = gtags;

   put_search_index_rec(&sxrec,file);                                      //  v.12.10
   return;
}


//  delete given image file from search index recs.

void delete_search_index(cchar *file)
{
   put_search_index_rec(null,file);                                        //  v.12.10
   return;
}


/**************************************************************************/

//  menu function - add tags to many files at once

char        **batchAddTags_filelist = 0;
int         batchAddTags_filecount = 0;
zdialog     *zdbatchAddTags = null;                                        //  batch add tags dialog

void m_batchAddTags(GtkWidget *, cchar *)
{
   void  batch_addtags_clickfunc(GtkWidget *widget, int &line, int &pos);
   void  batch_deftags_clickfunc(GtkWidget *widget, int &line, int &pos);
   int   batchAddTags_dialog_event(zdialog *zd, cchar *event);
   
   char        *ptag, **flist, *file;
   int         ii, jj, err;
   zdialog     *zd;
   GtkWidget   *widget;

   zfuncs::F1_help_topic = "batch_add_tags";

   if (checkmenulock()) return;                                            //  menu locked               v.12.09
   if (mod_keep()) return;                                                 //  unsaved edits
   if (! check_index_sync()) return;                                       //  index fle sync disabled   v.12.08

   zd = zdialog_new(ZTX("Batch Add Tags"),mWin,Bproceed,Bcancel,null);
   zdbatchAddTags = zd;

   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","files","hb3",Bselectfiles,"space=10");
   zdialog_add_widget(zd,"label","labcount","hb3","0 files selected","space=10");

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labaddtags","hb1",ZTX("tags to add"),"space=10");
   zdialog_add_widget(zd,"frame","frame1","hb1",0,"expand");
   zdialog_add_widget(zd,"text","addtags","frame1",0,"expand|wrap");

   zdialog_add_widget(zd,"hbox","space","dialog",0,"space=3");
   zdialog_add_widget(zd,"hbox","hb5","dialog");
   zdialog_add_widget(zd,"label","labdeftags","hb5",ZTX("defined tags"),"space=10");
   zdialog_add_widget(zd,"hbox","hb6","dialog",0,"expand");
   zdialog_add_widget(zd,"frame","frame5","hb6",0,"space=5|expand");
   zdialog_add_widget(zd,"scrwin","scrwin5","frame5",0,"expand");
   zdialog_add_widget(zd,"text","deftags","scrwin5",0,"expand|wrap");

   load_deftags();                                                         //  stuff defined tags into dialog
   deftags_stuff(zd);

   batchAddTags_filelist = 0;
   batchAddTags_filecount = 0;
   *tags_batchAddTags = 0;

   widget = zdialog_widget(zd,"addtags");                                  //  tag widget mouse functions      v.12.06
   textwidget_set_clickfunc(widget,batch_addtags_clickfunc);

   widget = zdialog_widget(zd,"deftags");
   textwidget_set_clickfunc(widget,batch_deftags_clickfunc);

   zdialog_resize(zd,500,400);                                             //  run dialog
   zdialog_help(zd,"batch_add_tags");                                      //  zdialog help topic
   zdialog_run(zd,batchAddTags_dialog_event);
   
   zdialog_wait(zd);                                                       //  wait for dialog completion

   flist = batchAddTags_filelist;                                          //  selected files

   if (zd->zstat != 1) goto cleanup;
   if (! batchAddTags_filecount) goto cleanup;
   if (*tags_batchAddTags <= ' ') goto cleanup;

   write_popup_text("open","Adding Tags",500,200,mWin);                    //  status monitor popup window

   for (ii = 0; flist[ii]; ii++)                                           //  loop all selected files
   {
      file = flist[ii];                                                    //  display image
      err = f_open(file,0,0,0,0);
      if (err) continue;

      write_popup_text("write",file);                                      //  report progress
      zmainloop();

      load_filemeta(file);                                                 //  load current file tags

      for (jj = 1; ; jj++)                                                 //  add new tags unless already
      {
         ptag = (char *) strField(tags_batchAddTags,tagdelims,jj);
         if (! ptag) break;
         if (*ptag == ' ') continue;
         err = add_tag(ptag,tags_currtags,tagFcc);
         if (err == 2) 
            zmessageACK(mWin,ZTX("%s \n too many tags"),file);
      }

      save_filemeta(file);                                                 //  save tag changes
   }

   write_popup_text("write","COMPLETED");
   write_popup_text("close",0);
   
cleanup:

   zdialog_free(zdbatchAddTags);

   if (batchAddTags_filecount) {
      for (ii = 0; flist[ii]; ii++) 
         zfree(flist[ii]);
      zfree(flist);
   }

   return;
}


//  mouse click functions for widgets having tags                          //  v.12.06

void batch_addtags_clickfunc(GtkWidget *widget, int &line, int &pos)       //  a tag in the add list was clicked
{
   char     *txline, *txtag, end = 0;
   cchar    *delims = tagdelims":";
   
   txline = textwidget_get_line(widget,line);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,delims,end);
   if (! txtag) { zfree(txline); return; }
   
   del_tag(txtag,tags_batchAddTags);                                       //  remove tag from list
   zdialog_stuff(zdbatchAddTags,"addtags",tags_batchAddTags);
   
   zfree(txline);
   zfree(txtag);
   return;
}


void batch_deftags_clickfunc(GtkWidget *widget, int &line, int &pos)       //  a defined tag was clicked
{
   char     *txline, *txtag, end = 0;
   cchar    *delims = tagdelims":";
   
   txline = textwidget_get_line(widget,line);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,delims,end);
   if (! txtag) { zfree(txline); return; }
   
   add_tag(txtag,tags_batchAddTags,tagMcc);                                //  add defined tag to list
   zdialog_stuff(zdbatchAddTags,"addtags",tags_batchAddTags);
   
   zfree(txline);
   zfree(txtag);
   return;
}


//  batchAddTags dialog event function

int batchAddTags_dialog_event(zdialog *zd, cchar *event)
{
   int      ii;
   char     **flist = batchAddTags_filelist, countmess[50];
   
   if (strEqu(event,"files"))                                              //  select images to add tags
   {
      if (flist) {                                                         //  free prior list
         for (ii = 0; flist[ii]; ii++) 
            zfree(flist[ii]);
         zfree(flist);
      }

      zdialog_show(zd,0);                                                  //  hide parent dialog              v.12.11
      flist = gallery_getfiles();                                          //  get file list from user
      zdialog_show(zd,1);

      batchAddTags_filelist = flist;

      if (flist)                                                           //  count files in list
         for (ii = 0; flist[ii]; ii++);
      else ii = 0;
      batchAddTags_filecount = ii;
      
      snprintf(countmess,50,ZTX("%d files selected"),batchAddTags_filecount);
      zdialog_stuff(zd,"labcount",countmess);
   }
   
   return 1;
}
   

/**************************************************************************/

//  menu function - delete or replace a tag for many files at once

char        **batchDelTag_filelist = 0;

void m_batchDelTag(GtkWidget *, cchar *)
{
   int   batchDelTag_dialog_event(zdialog *zd, cchar *event);
   
   char           **flist, *file;
   char           deltag[tagcc], reptag[tagcc];
   int            ii, err;
   zdialog        *zd;

   zfuncs::F1_help_topic = "batch_delete_tag";

   if (checkmenulock()) return;                                            //  menu locked                  v.12.09
   if (mod_keep()) return;                                                 //  unsaved edits
   if (! check_index_sync()) return;                                       //  index fle sync disabled      v.12.08

   zd = zdialog_new(ZTX("Batch Delete Tag"),mWin,Bproceed,Bcancel,null);

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","lab1","hb1",ZTX("tag to remove"),"space=10");
   zdialog_add_widget(zd,"entry","deltag","hb1",0,"scc=20");

   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","lab2","hb2",ZTX("optional replacement"),"space=10");
   zdialog_add_widget(zd,"entry","reptag","hb2",0,"scc=20");

   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","files","hb3",Bselectfiles,"space=10");
   zdialog_add_widget(zd,"label","labcount","hb3",ZTX("0 files selected"),"space=10");
   
   zdialog_add_widget(zd,"hbox","hb4","dialog",0,"space=5");
   zdialog_add_widget(zd,"check","allfiles","hb4",ZTX("search all files"),"space=10");

   batchDelTag_filelist = flist = 0;

   zdialog_help(zd,"batch_delete_tag");                                    //  zdialog help topic
   zdialog_run(zd,batchDelTag_dialog_event);                               //  run dialog
   zdialog_wait(zd);                                                       //  wait for completion

   if (zd->zstat != 1) goto cleanup;                                       //  canceled

   flist = batchDelTag_filelist;                                           //  get selected files
   if (! flist) goto cleanup;
   
   zdialog_fetch(zd,"deltag",deltag,tagcc);                                //  get tag to delete
   strTrim2(deltag);                                                       //  remove leading and trailing blanks
   strToLower(deltag);                                                     //  use lower case for matching
   if (! *deltag) goto cleanup;

   zdialog_fetch(zd,"reptag",reptag,tagcc);                                //  optional replacement tag
   strTrim2(reptag);

   write_popup_text("open","Deleting Tags",500,200,mWin);                  //  status monitor popup window

   for (ii = 0; flist[ii]; ii++)                                           //  loop all selected files
   {
      file = flist[ii];
      err = f_open(file,0,0,0,0);                                          //  display image
      if (err) continue;

      write_popup_text("write",file);                                      //  report progress
      zmainloop();

      load_filemeta(file);                                                 //  load current file tags

      err = del_tag(deltag,tags_currtags);                                 //  remove tag, if present
      if (err) continue;                                                   //  not present

      if (*reptag) {                                                       //  add replacement tag, if defined
         err = add_tag(reptag,tags_currtags,tagFcc);
         if (err == 2) 
            zmessageACK(mWin,ZTX("%s \n too many tags"),file);
      }

      save_filemeta(file);                                                 //  save tag changes
   }
   
   write_popup_text("write","COMPLETED");
   write_popup_text("close",0);
   
cleanup:

   zdialog_free(zd);

   if (flist) {
      for (ii = 0; flist[ii]; ii++) 
         zfree(flist[ii]);
      zfree(flist);
   }

   return;
}


//  batchDelTag dialog event function

int batchDelTag_dialog_event(zdialog *zd, cchar *event)
{
   char           deltag[tagcc];
   char           **flist = batchDelTag_filelist;
   int            ftf, nfiles, ii, err, allfiles;
   char           *tags, countmess[50];
   cchar          *ppf;
   int            maxfiles = 10000;                                        //  max. files with tag to delete 
   sxrec_t        sxrec;
   
   zdialog_fetch(zd,"deltag",deltag,tagcc);                                //  get tag to delete
   strTrim2(deltag);                                                       //  remove leading and trailing blanks
   strToLower(deltag);                                                     //  use lower case for matching

   if (zd->zstat == 1)                                                     //  [proceed]
   {
      if (! flist) {
         zmessageACK(mWin,ZTX("no files selected"));
         zd->zstat = 0;                                                    //  keep dialog active
         goto finish;
      }
      
      if (! *deltag) {
         zmessageACK(mWin,ZTX("no tag specified"));
         zd->zstat = 0;
         goto finish;
      }
      
      return 0;
   }

   if (strEqu(event,"deltag"))                                             //  tag changed
   {
      if (flist) {                                                         //  clear file list
         for (ii = 0; flist[ii]; ii++) 
            zfree(flist[ii]);
         zfree(flist);
         flist = 0;
      }

      zdialog_stuff(zd,"allfiles",0);                                      //  reset "all files" checkbox
      goto finish;
   }

   if (strEqu(event,"files"))                                              //  select images to delete tags
   {
      if (! *deltag) {
         zmessageACK(mWin,ZTX("specify tag"));
         goto finish;
      }

      if (flist) {                                                         //  free prior list
         for (ii = 0; flist[ii]; ii++) 
            zfree(flist[ii]);
         zfree(flist);
         flist = 0;
      }

      zdialog_show(zd,0);                                                  //  hide parent dialog              v.12.11
      flist = gallery_getfiles();                                          //  get file list from user
      zdialog_show(zd,1);

      zdialog_stuff(zd,"allfiles",0);                                      //  reset "all files" checkbox
      goto finish;
   }

   if (strEqu(event,"allfiles"))
   {
      if (flist) {                                                         //  clear file list
         for (ii = 0; flist[ii]; ii++) 
            zfree(flist[ii]);
         zfree(flist);
         flist = 0;
      }

      zdialog_fetch(zd,"allfiles",allfiles);                               //  get checkbox "all files"
      if (! allfiles) goto finish;                                         //  was unselected
      
      if (! *deltag) {
         zmessageACK(mWin,ZTX("specify tag"));
         goto finish;
      }
      
      flist = (char **) zmalloc(maxfiles*sizeof(char *),"DeleteTag");      //  list for files found
      nfiles = 0;                                                          //  count files found
      ftf = 1;

      while (true)
      {
         err = read_search_index_seq(sxrec,ftf);                           //  v.12.10
         if (err) break;

         tags = sxrec.tags;
         strToLower(tags);                                                 //  use lower case for matching
         
         for (ii = 1; ; ii++)                                              //  step thru file tags
         {
            ppf = strField(tags,tagdelims,ii);
            if (! ppf) break;
            if (*ppf == ' ') continue;
            if (strEqu(ppf,deltag)) break;                                 //  look for my tag
         }

         if (ppf) {                                                        //  add matching file to list
            flist[nfiles] = strdupz(sxrec.file,0,"batch_del_tag");
            nfiles++;
         }

         zfree(sxrec.file);
         zfree(sxrec.tags);
         zfree(sxrec.comms);
         zfree(sxrec.capt);
         zfree(sxrec.gtags);

         if (nfiles == maxfiles-1) break;
      }

      flist[nfiles] = 0;                                                   //  EOL marker
   }

finish:

   batchDelTag_filelist = flist;                                           //  file list to process

   if (flist)                                                              //  count files in list
      for (ii = 0; flist[ii]; ii++);
   else ii = 0;
   
   snprintf(countmess,50,ZTX("%d files selected"),ii);                     //  stuff file count into dialog
   zdialog_stuff(zd,"labcount",countmess);
   
   return 0;
}
   

/**************************************************************************
   functions to view and edit ANY metadata: EXIF, IPTC, etc. 
***************************************************************************/

//  menu function and popup dialog to show EXIF/IPTC data
//  window is updated when navigating to another image

void m_meta_view_short(GtkWidget *, cchar *menu)
{
   if (menu) zfuncs::F1_help_topic = "view_meta";
   meta_view(1);
   return;
}

void m_meta_view_long(GtkWidget *, cchar *menu)
{
   if (menu) zfuncs::F1_help_topic = "view_meta";
   meta_view(2);
   return;
}

void meta_view(int arg)
{
   int   meta_view_dialog_event(zdialog *zd, cchar *event);

   static int     length = 1;
   char           *buff;
   int            contx = 0;
   GtkWidget      *widget;

   if (! curr_file) return;
   if (! menulock(1)) return;                                              //  stop hang from rapid [next]     v.12.09
   
   if (arg > 0) length = arg;                                              //  change short/long report mode

   if (! zdexifview)                                                       //  popup dialog if not already
   {
      zdexifview = zdialog_new(ZTX("View Metadata"),mWin,Bcancel,null);
      zdialog_add_widget(zdexifview,"scrwin","scroll","dialog",0,"expand");
      zdialog_add_widget(zdexifview,"text","exifdata","scroll",0,"expand|wrap");
      zdialog_resize(zdexifview,500,400);
      zdialog_help(zdexifview,"view_meta");                                //  zdialog help topic
      zdialog_run(zdexifview,meta_view_dialog_event);
   }

   widget = zdialog_widget(zdexifview,"exifdata");
   gtk_text_view_set_editable(GTK_TEXT_VIEW(widget),0);                    //  disable widget editing
   gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(widget),GTK_WRAP_NONE);       //  disable text wrap

   if (length == 1)                                                        //  short report
   {
      snprintf(command,ccc,"exiftool -s -common "                          //                            (-s v.12.06)
         "-%s -%s -%s -%s -%s -%s -%s -%s -%s -%s \"%s\" ",
         iptc_keywords_key, iptc_rating_key, exif_editlog_key, exif_comment_key, 
         iptc_caption_key, exif_focal_length_key, exif_city_key, exif_country_key, 
         exif_latitude_key, exif_longitude_key, curr_file);
   }
   else 
      snprintf(command,ccc,"exiftool -s -e \"%s\" ",curr_file);            //  long, output everything   (-s v.12.06)
      
   widget = zdialog_widget(zdexifview,"exifdata");                         //  widget for output
   wclear(widget);

   while ((buff = command_output(contx,command))) {                        //  run command, output into window
      wprintf(widget,"%s\n",buff);
      zfree(buff);
   }

   command_status(contx);                                                  //  free resources
   menulock(0);

   return;
}


//  dialog event and completion callback function

int meta_view_dialog_event(zdialog *zd, cchar *event)                      //  kill dialog
{
   if (! zd->zstat) return 0;
   zdialog_free(zdexifview);
   return 0;
}


/**************************************************************************/

//  edit EXIF/IPTC data - add or change specified EXIF/IPTC/etc. key

void m_meta_edit_any(GtkWidget *, cchar *menu)
{
   char        keyname[40], keydata[exif_maxcc];
   cchar       *pp1[1];
   char        **pp2;

   int   meta_edit_any_dialog_event(zdialog *zd, cchar *event);
   
   if (menu) zfuncs::F1_help_topic = "edit_meta";
   if (! check_index_sync()) return;                                       //  index fle sync disabled      v.12.08
   if (! curr_file) return;

   if (! zdexifedit)                                                       //  popup dialog if not already
   {   
      zdexifedit = zdialog_new(ZTX("Edit Metadata"),mWin,Bfetch,Bsave,Bcancel,null);
      zdialog_add_widget(zdexifedit,"vbox","hb1","dialog");
      zdialog_add_widget(zdexifedit,"hbox","hbkey","dialog",0,"space=5");
      zdialog_add_widget(zdexifedit,"hbox","hbdata","dialog",0,"space=5");
      zdialog_add_widget(zdexifedit,"label","labkey","hbkey","key name");
      zdialog_add_widget(zdexifedit,"entry","keyname","hbkey",0,"scc=20");
      zdialog_add_widget(zdexifedit,"label","labdata","hbdata","key value");
      zdialog_add_widget(zdexifedit,"entry","keydata","hbdata",0,"expand");
      zdialog_help(zdexifedit,"edit_meta");                                //  zdialog help topic 
      zdialog_run(zdexifedit,meta_edit_any_dialog_event);
   }

   zdialog_fetch(zdexifedit,"keyname",keyname,40);                         //  get key name from dialog
   strCompress(keyname);

   if (*keyname)                                                           //  update live dialog
   {
      pp1[0] = keyname;                                                    //  look for key data 
      pp2 = exif_get(curr_file,pp1,1);
      if (pp2[0]) {
         strncpy0(keydata,pp2[0],exif_maxcc);
         zfree(pp2[0]);
      }
      else *keydata = 0;
      zdialog_stuff(zdexifedit,"keydata",keydata);                         //  stuff into dialog
   }

   return;
}


//  dialog event and completion callback function

int  meta_edit_any_dialog_event(zdialog *zd, cchar *event)
{
   char        keyname[40], keydata[exif_maxcc];
   cchar       *pp1[1], *pp2[1];
   char        **pp3;
   int         err;
   
   if (! zd->zstat) return 0;
   
   zdialog_fetch(zd,"keyname",keyname,40);
   zdialog_fetch(zd,"keydata",keydata,exif_maxcc);
   strCompress(keyname);                                                   //  remove blanks

   if (zd->zstat == 1)                                                     //  fetch
   {
      zd->zstat = 0;
      if (! *keyname) return 0;
      pp1[0] = keyname;
      pp3 = exif_get(curr_file,pp1,1);
      if (pp3[0]) {
         strncpy0(keydata,pp3[0],exif_maxcc);
         zfree(pp3[0]);
      }
      else *keydata = 0;
      zdialog_stuff(zd,"keydata",keydata);
   }

   else if (zd->zstat == 2)                                                //  save
   {
      zd->zstat = 0;                                                       //  keep dialog active
      if (! *keyname) return 0;
      pp1[0] = keyname;
      pp2[0] = keydata;
      err = exif_put(curr_file,pp1,pp2,1);
      if (err) zmessageACK(mWin,"error: %s",strerror(err));
      load_filemeta(curr_file);                                            //  update search index in case   v.12.10
      update_search_index(curr_file);                                      //    searchable metadata item updated
      if (zdexifview) meta_view(0);                                        //  update exif view if active
   }

   else zdialog_free(zdexifedit);                                          //  other

   return 1;
}


/**************************************************************************/

//  delete EXIF/IPTC data, specific key or all data

void m_meta_delete(GtkWidget *, cchar *menu)
{
   int   meta_delete_dialog_event(zdialog *zd, cchar *event);
   
   zdialog     *zd;

   zfuncs::F1_help_topic = "delete_meta";
   if (! check_index_sync()) return;                                       //  index fle sync disabled      v.12.08
   if (! curr_file) return;

   zd = zdialog_new(ZTX("Delete Metadata"),mWin,Bapply,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"radio","kall","hb1",ZTX("All"),"space=5");
   zdialog_add_widget(zd,"radio","key1","hb1",ZTX("One Key:"));
   zdialog_add_widget(zd,"entry","keyname","hb1",0,"scc=20");
   zdialog_stuff(zd,"key1",1);

   zdialog_help(zd,"delete_meta");                                         //  zdialog help topic 
   zdialog_run(zd,meta_delete_dialog_event);
   return;
}


//  dialog event and completion callback function

int meta_delete_dialog_event(zdialog *zd, cchar *event)
{
   int         kall, key1, err;
   char        keyname[40];
   
   if (! zd->zstat) return 0;

   if (zd->zstat != 1) {                                                   //  canceled
      zdialog_free(zd);
      return 1;
   }
   
   zd->zstat = 0;                                                          //  dialog remains active

   zdialog_fetch(zd,"kall",kall);
   zdialog_fetch(zd,"key1",key1);
   zdialog_fetch(zd,"keyname",keyname,40);
   strCompress(keyname);

   if (kall)                                                               //  -P preserve date v.12.09
      snprintf(command,ccc,"exiftool -m -q -P -overwrite_original -all=  \"%s\"",curr_file);
   else if (key1)
      snprintf(command,ccc,"exiftool -m -q -P -overwrite_original -%s=  \"%s\"",keyname,curr_file);
   else return 1;
   
   gallery_monitor("stop");                                                //  stop excess gallery inits    v.12.10

   err = system(command);
   if (err) zmessageACK(mWin,"%s",wstrerror(err));
   
   load_filemeta(curr_file);                                               //  update search index in case a   v.12.10
   update_search_index(curr_file);                                         //    searchable metadata deleted

   gallery_monitor("start");                                               //  v.12.10
   
   if (zdexifview) meta_view(0);                                           //  update exif view if active

   return 1;
}


/**************************************************************************
            GEOTAG functions
***************************************************************************/

struct geolocs_t {                                                         //  geotag location data, memory DB
   char     *city, *country;
   char     *lati, *longi;
   float    flati, flongi;
};
geolocs_t   *geolocs;
int      Ngeolocs = 0;                                                     //  number of geotag location records

zdialog  *zd_geotags;                                                      //  active zdialog to get geotags data
int      worldmap_range = 20;                                              //  default geotag search range, km
char     prevcity[100] = "", prevcountry[100] = "";                        //  last geotag data used
char     prevlatitude[20] = "", prevlongitude[20] = "";                    //  (recalled with [prev] button)

typedef void worldmap_callback_t(float flati, float flongi);
worldmap_callback_t  *worldmap_callback = 0;

int geotags_choosecity(char *location[2], char *coord[2]);                 //  choose one city from multiple options
cchar * web_geocode(char *location[2], char *coord[2]);                    //  find latitude/longitude via web service
int get_worldmap_coordinates(int mx, int my, float &flat, float &flong);   //  pixel position to latitude/longitude
int get_worldmap_position(float flat, float flong, int &mx, int &my);      //  latitude/longitude to pixel position


/**************************************************************************/

//  validate and convert earth coordinates, latitude and longitude
//  return: 0  OK
//          1  both are missing ("null" or "")
//          2  invalid data
//  if status is > 0, 0.0 is returned for both values

int validate_latilongi(char *latitude, char *longitude, float &flati, float &flongi)
{
   if (! latitude || *latitude == 0 || strEqu(latitude,"null"))
      if (! longitude || *longitude == 0 || strEqu(longitude,"null"))
         goto status1;                                                     //  both missing
   
   if (! latitude || *latitude == 0 || strEqu(latitude,"null"))
      goto status2;                                                        //  one missing
   if (! longitude || *longitude == 0 || strEqu(longitude,"null")) 
      goto status2;
   
   flati = atof(latitude);                                                 //  convert strings to float
   flongi = atof(longitude);
   if (flati < -90.0 || flati > +90.0) goto status2;                       //  check range
   if (flongi < -180.0 || flongi > +180.0) goto status2;
   if (flati == 0.0 && flongi == 0.0) goto status2;                        //  reject both = 0.0 
   return 0;

status1:
   flati = flongi = 0.0;                                                   //  both missing
   return 1;

status2:                                                                   //  either missing or invalid
   flati = flongi = 0.0;
   return 2;
}


/**************************************************************************/

//  download world-map-mercator.jpg and cities-geotags files.
//  save in /.../.fotoxx/geotags/*

void m_download_geolocs(GtkWidget *, cchar *)                              //  new v.12.05
{
   int      yn, err;
   char     scriptfile[200];
   char     destfile1[200], sourcefile1[200];
   char     destfile2[200], sourcefile2[200];
   FILE     *fid;

   zfuncs::F1_help_topic = "download_geolocs";

   if (! check_index_sync()) return;                                       //  index fle sync disabled      v.13.01

   if (Ngeolocs) zfree(geolocs);                                           //  v.12.07
   Ngeolocs = 0;

   yn = zmessageYN(mWin,ZTX("Download geotags data (8 megabytes). \n"
                            "Save in %s \n"
                            "Proceed?"),geotags_dirk);
   if (! yn) return;
   
   sprintf(sourcefile1,"http://www.kornelix.com/uploads/1/3/0/3/13035936/world-map-mercator.jpg");
   sprintf(destfile1,"%s/world-map-mercator.jpg",geotags_dirk);
   sprintf(sourcefile2,"http://www.kornelix.com/uploads/1/3/0/3/13035936/cities-geotags");
   sprintf(destfile2,"%s/cities-geotags",geotags_dirk);
   sprintf(scriptfile,"%s/download.scr",geotags_dirk);
   
   fid = fopen(scriptfile,"w");
   if (! fid) { 
      zmessageACK(mWin,"cannot create script file \n %s",strerror(errno)); 
      return;
   }
   
   fprintf(fid,"#! /bin/bash \n");                                         //  build bash script
   fprintf(fid,"#  download geotags data \n");
   fprintf(fid,"# \n");
   fprintf(fid,"echo DOWNLOAD WORLD MAP \n");
   fprintf(fid,"wget --progress=dot:mega -O %s %s 2>&1 \n",destfile1,sourcefile1);
   fprintf(fid,"err=$? \n");
   fprintf(fid,"if [ $err -ne 0 ]; then exit $err; fi \n");
   fprintf(fid,"echo DOWNLOAD GEOTAGS FILE \n");
   fprintf(fid,"wget -O %s %s 2>&1 \n",destfile2,sourcefile2);
   fprintf(fid,"err=$? \n");
   fprintf(fid,"if [ $err -ne 0 ]; then exit $err; fi \n");
   fprintf(fid,"echo DONE \n");
   fprintf(fid,"exit 0 \n");
   fclose(fid);

   chmod(scriptfile,0744);
   
   sprintf(command,"bash %s",scriptfile);                                  //  run the script
   err = popup_command(command,800,600,mWin);
   if (err) zmessageACK(mWin,ZTX("download failed"));
   
   return;
}


/**************************************************************************/

//  Initialize for geotag functions.
//  Returns no. geotags in memory DB or 0 if downloads are needed.
//  (user is informed with popup message)

int init_geotags()
{
   int  init_glocs_comp(cchar *rec1, cchar *rec2);
   
   char     geotagsfile[200], mapfile[200];
   char     buff[200], city[100], country[100];
   char     latitude[20], longitude[20];
   char     *filename, *gtags, *pp;
   float    flati, flongi;
   int      ftf, cc, ii, jj;
   int      err, found, keep, Ngeolocs2;
   int      mx, my, px, py;
   uint8    *pixel;
   FILE     *fid;
   sxrec_t  sxrec;

   if (! check_index_sync()) return 0;                                     //  index fle sync disabled      v.12.08

   if (! Wpxb)                                                             //  world map image     v.12.07
   {
      snprintf(mapfile,200,"%s/world-map-mercator.jpg",geotags_dirk);      //  load world map image
      Wpxb = PXB_load(mapfile,0);                                          //  v.12.10
      if (! Wpxb) {
         zmessageACK(mWin,ZTX("please download geolocations data"));
         return 0;
      }
      gtk_widget_show_all(mWin);                                           //  show all widgets
   }

   if (Ngeolocs) return Ngeolocs;                                          //  already done

   cc = maxgeotags * sizeof(geolocs_t);                                    //  get memory for geotag locations DB 
   geolocs = (geolocs_t *) zmalloc(cc,"geolocs");                          //  v.12.07
   
   //  populate geolocs[] from search-index file (= image EXIF data)

   ftf = 1;
   while (true)
   {
      err = read_search_index_seq(sxrec,ftf);                              //  read search index recs.      v.12.10
      if (err) break;
      
      filename = sxrec.file;
      gtags = sxrec.gtags;

      strcpy(city,"null");
      strcpy(country,"null");
      strcpy(latitude,"null");
      strcpy(longitude,"null");
      
      found = 0;
      
      pp = (char *) strField(gtags,'^',1);                                 //  city name or "null"
      if (pp) {
         strncpy0(city,pp,99);
         found++;
      }
      
      pp = (char *) strField(gtags,'^',2);                                 //  country
      if (pp) { 
         strncpy0(country,pp,99);
         found++;
      }

      pp = (char *) strField(gtags,'^',3);                                 //  latitude
      if (pp) strncpy0(latitude,pp,19);

      pp = (char *) strField(gtags,'^',4);                                 //  longitude
      if (pp) strncpy0(longitude,pp,19);
      
      err = validate_latilongi(latitude,longitude,flati,flongi);           //  validate and replace bad data
      if (err) {
         strcpy(latitude,"null");                                          //  missing (1) or bad (2) data
         strcpy(longitude,"null");
         if (err == 2) {                                                   //  exclude invalid        v.13.01
            printf("*** bad geotag in search_index: %s \n",filename);
            found = 0;
         }
      }

      if (found > 0 && Ngeolocs) {                                         //  at least city or country was present
         ii = Ngeolocs - 1;
         if (strEqu(city,geolocs[ii].city))                                //  ignore same city in sequence (frequent)
            if (strEqu(country,geolocs[ii].country)) found = 0;            //    to reduce subsequent sort
      }

      if (found > 0) {  
         ii = Ngeolocs;
         geolocs[ii].city = strdupz(city,0,"geotags");
         geolocs[ii].country = strdupz(country,0,"geotags");
         geolocs[ii].lati = strdupz(latitude,0,"geotags");
         geolocs[ii].longi = strdupz(longitude,0,"geotags");
         geolocs[ii].flati = atof(latitude);
         geolocs[ii].flongi = atof(longitude);
         Ngeolocs++;
      }
   
      zfree(sxrec.file);
      zfree(sxrec.tags);
      zfree(sxrec.capt);
      zfree(sxrec.comms);
      zfree(sxrec.gtags);

      if (Ngeolocs == maxgeotags) goto exceedmax;
   }

   if (Ngeolocs > 1)                                                       //  sort by country, city
      HeapSort((char *) geolocs, sizeof(geolocs_t), Ngeolocs, init_glocs_comp);

   for (ii = 0, jj = 1; jj < Ngeolocs; jj++)                               //  eliminate duplicate cities
   {
      keep = 0;
      if (strNeq(geolocs[jj].country,geolocs[ii].country)) keep = 1;
      if (! keep && strNeq(geolocs[jj].city,geolocs[ii].city)) keep = 1;
      if (keep) {
         ii++;
         geolocs[ii] = geolocs[jj];
      }
      else {
         zfree(geolocs[jj].country);
         zfree(geolocs[jj].city);
         zfree(geolocs[jj].lati);
         zfree(geolocs[jj].longi);
      }
   }
   
   Ngeolocs = ii + 1;
   Ngeolocs2 = Ngeolocs;                                                   //  geolocs where images available  v.12.09

   //  add data from cities-geotags file to geolocs[] data
   //  (after data from image files)

   sprintf(geotagsfile,"%s/cities-geotags",geotags_dirk);                  //  read cities-geotags file
   fid = fopen(geotagsfile,"r");
   if (! fid) {
      zmessageACK(mWin,ZTX("geotags file must be downloaded"));
      return 0;
   }
   
   while (true)
   {
      pp = fgets_trim(buff,200,fid,1);                                     //  read each city record into memory
      if (! pp) break;                                                     //  (remove trailing \n)
      if (strlen(pp) < 2) continue;
      
      ii = Ngeolocs;
      
      pp = (char *) strField(buff,'^',1);                                  //  get latitude
      if (! pp) goto badrec;
      strTrim2(pp);
      geolocs[ii].lati = strdupz(pp,0,"geotags");
      geolocs[ii].flati = atof(pp);

      pp = (char *) strField(buff,'^',2);                                  //  longitude
      if (! pp) goto badrec;
      strTrim2(pp);
      geolocs[ii].longi = strdupz(pp,0,"geotags");
      geolocs[ii].flongi = atof(pp);
      
      err = validate_latilongi(geolocs[ii].lati,geolocs[ii].longi,flati,flongi);
      if (err) goto badrec;

      pp = (char *) strField(buff,'^',3);                                  //  city
      if (! pp) goto badrec;
      strTrim2(pp);
      geolocs[ii].city = strdupz(pp,0,"geotags");
         
      pp = (char *) strField(buff,'^',4);                                  //  country
      if (! pp) goto badrec;
      strTrim2(pp);
      geolocs[ii].country = strdupz(pp,0,"geotags");
      
      Ngeolocs++;                                                          //  count good recs found
      if (Ngeolocs == maxgeotags) goto exceedmax;
      continue;
   
   badrec:
      printf("*** bad cities-geotags record: %s \n",buff);
      continue;
   }
   
   fclose(fid);

   for (ii = 0; ii < Ngeolocs2; ii++)                                      //  paint red dots on world map     v.12.09
   {                                                                       //    where images are present
      flati = geolocs[ii].flati;
      flongi = geolocs[ii].flongi;
      err = get_worldmap_position(flati,flongi,mx,my);
      if (err) continue;
      for (px = -1; px <= 1; px++)                                         //  paint 3x3 block of pixels
      for (py = -1; py <= 1; py++) 
      {
         pixel = PXBpix(Wpxb,mx+px,my+py);
         pixel[0] = 255;
         pixel[1] = pixel[2] = 0;
      }
   }

   return Ngeolocs;

exceedmax:
   zmessageACK(mWin,"max. geotags exceeded");
   return 0;
}


//  Compare 2 geolocs records by country and city
//  return  <0  0  >0   for   rec1  <  ==  >  rec2.

int  init_glocs_comp(cchar *rec1, cchar *rec2)
{
   int      ii;

   char * country1 = ((geolocs_t *) rec1)->country;                        //  compare countries
   char * country2 = ((geolocs_t *) rec2)->country;
   ii = strcmp(country1,country2);
   if (ii) return ii;

   char * city1 = ((geolocs_t *) rec1)->city;                              //  compare cities
   char * city2 = ((geolocs_t *) rec2)->city;
   ii = strcmp(city1,city2);
   return ii;
}


/**************************************************************************/

//  get latitude/longitude data for a city [ country ]
//  inputs:        location[0] = city
//                 location[1] = country (opt)
//  outputs:       coord[2] = latitude -90 to +90, longitude -180 to +180
//                 matches[20][2] up to 20 matching city/country locations
//  returns:       no. matches for input city [ country ]
//                 (max. 20)
//
//  use null or "" for missing city or country input (no NULL pointer)
//  coordinates are returned for the first match only, if any
//

int get_geotags(char *location[2], char *coord[2], char *matches[20][2])   //  new v.12.05
{
   int      cc, ii, jj, Nmatch;
   int      fcity = 0, fcountry = 0;
   
   fcity = location[0] && *location[0];                                    //  city or country must be present
   fcountry = location[1] && *location[1];                                 //  v.12.07
   if (! fcity && ! fcountry) return 0;   
   
   for (ii = Nmatch = 0; ii < Ngeolocs; ii++)                              //  search for exact city [ country ]
   {
      if (fcity && strcasecmp(location[0],geolocs[ii].city) != 0) continue;
      if (fcountry && strcasecmp(location[1],geolocs[ii].country) != 0) continue;

      for (jj = 0; jj < Nmatch; jj++) {                                    //  look for duplicate match   v.12.07
         if (strEqu(geolocs[ii].city,matches[jj][0]))                      //  (EXIF and cities-geotag file)
            if (strEqu(geolocs[ii].country,matches[jj][1])) break;
      }
      if (jj < Nmatch) continue;                                           //  discard duplicates

      matches[Nmatch][0] = geolocs[ii].city;
      matches[Nmatch][1] = geolocs[ii].country;
      Nmatch++;
      if (Nmatch == 20) break;                                             //  no more than 20 matches are reported
      if (Nmatch == 1) {
         coord[0] = geolocs[ii].lati;                                      //  return geolat/long of 1st match
         coord[1] = geolocs[ii].longi;
         StripZeros(coord[0]);
         StripZeros(coord[1]);
      }
   }
   
   if (Nmatch) return Nmatch;                                              //  exact match found

   for (ii = Nmatch  = 0; ii < Ngeolocs; ii++)                             //  search for partial city [ country ]
   {
      if (fcity) {
         cc = strlen(location[0]);
         if (strncasecmp(location[0],geolocs[ii].city,cc) != 0) continue;
      }
      if (fcountry) {
         cc = strlen(location[1]);
         if (strncasecmp(location[1],geolocs[ii].country,cc) != 0) continue;
      }
      matches[Nmatch][0] = geolocs[ii].city;
      matches[Nmatch][1] = geolocs[ii].country;
      Nmatch++;
      if (Nmatch == 20) break;                                             //  no more than 20 matches are reported
      if (Nmatch == 1) {
         coord[0] = geolocs[ii].lati;                                      //  return geolat/long of 1st match
         coord[1] = geolocs[ii].longi;
         StripZeros(coord[0]);
         StripZeros(coord[1]);
      }
   }

   return Nmatch;
}


/**************************************************************************/

//  Save geotags in image file EXIF and search index file.
//  location[2] = city and country
//  coord[2] = latitude and longitude
//  return value:  0    OK, no geotag revision (incomplete data)
//                 1    OK, no geotag revision (matches existing data)
//                 2    OK, geotag lat/long updated
//                 3    OK, geotag new location added
//                -1    error, lat/long bad
//  (cities-geotags-new is no longer used)                                 //  v.12.07

int put_geotags(char *location[2], char *coord[2])                         //  new v.12.05
{
   char        acoord[2][20];
   float       flati, flongi;
   int         ii, err, Fnew = 0, Fchange = 0;
   int         px, py, mx, my;
   uint8       *pixel;
   int         retval;

   err = validate_latilongi(coord[0],coord[1],flati,flongi);
   if (err) {
      if (err == 2) goto badcoord;                                         //  reject bad data
      strcpy(acoord[0],"null");                                            //  replace missing data with "null"
      strcpy(acoord[1],"null");
      flati = flongi = 0;                                                  //  lati/longi missing value  v.13.01
   }
   else {
      sprintf(acoord[0],"%.4f",flati);                                     //  reformat with std. precision
      sprintf(acoord[1],"%.4f",flongi);
      StripZeros(acoord[0]);
      StripZeros(acoord[1]);
   }
   
   if (strNeq(location[0],"null"))                                         //  unless null,              v.12.06.2
      *location[0] = toupper(*location[0]);                                //  force capitalization      v.12.06
   if (strNeq(location[1],"null"))
      *location[1] = toupper(*location[1]);

   strncpy0(meta_city,location[0],99);                                     //  save geotags in image file EXIF
   strncpy0(meta_country,location[1],99);                                  //    and in search-index file
   strncpy0(meta_latitude,acoord[0],12);
   strncpy0(meta_longitude,acoord[1],12);
   Fmetachanged = 1;
   save_filemeta(curr_file);
   
   if (! *location[0] || strEqu(location[0],"null")) return 0;             //  quit here if city data not complete
   if (! *location[1] || strEqu(location[1],"null")) return 0;
   
   for (ii = 0; ii < Ngeolocs; ii++) {                                     //  search geotags for city, country
      if (strcasecmp(location[0],geolocs[ii].city) != 0) continue;         //  (case-insensitive compare)
      if (strcasecmp(location[1],geolocs[ii].country) == 0) break;
   }

   if (ii < Ngeolocs) {                                                    //  found, check for revised lat/long
      if (strNeq(geolocs[ii].lati,acoord[0])) Fchange = 1;
      if (strNeq(geolocs[ii].longi,acoord[1])) Fchange = 1;
      if (strNeq(geolocs[ii].city,location[0])) Fchange = 1;               //  or revised capitalization   v.12.06
      if (strNeq(geolocs[ii].country,location[1])) Fchange = 1;
   }
   else Fnew = 1;                                                          //  a new city, country
   
   if (Fnew + Fchange == 0) return 1;                                      //  no change
   
   if (Fchange) 
   {
      zfree(geolocs[ii].city);                                             //  change geotag data in memory
      geolocs[ii].city = strdupz(location[0],0,"geotags");                 //  (to be used subsequently)    v.12.07
      zfree(geolocs[ii].country);
      geolocs[ii].country = strdupz(location[1],0,"geotags");              //  presense in image EXIF will make
      zfree(geolocs[ii].lati);                                             //    this the preferred version
      geolocs[ii].lati = strdupz(acoord[0],0,"geotags");
      zfree(geolocs[ii].longi);
      geolocs[ii].longi = strdupz(acoord[1],0,"geotags");
      geolocs[ii].flati = flati;
      geolocs[ii].flongi = flongi;
      retval = 2;
   }
   
   else if (Fnew)
   {
      if (Ngeolocs == maxgeotags) {
         zmessageACK(mWin,"max. geotags exceeded");
         return -1;
      }
      for (ii = Ngeolocs; ii > 0; ii--)                                    //  shift all geotag data up
         geolocs[ii] = geolocs[ii-1];
      Ngeolocs++;

      ii = 0;
      geolocs[ii].city = strdupz(location[0],0,"geotags");                 //  new geotag is now first
      geolocs[ii].country = strdupz(location[1],0,"geotags");              //  (find again faster)
      geolocs[ii].lati = strdupz(acoord[0],0,"geotags");
      geolocs[ii].longi = strdupz(acoord[1],0,"geotags");
      geolocs[ii].flati = flati;
      geolocs[ii].flongi = flongi;
      retval = 3;
   }

   else return -1;                                                         //  should not happen

   if (flati && flongi) {                                                  //  put new red dot on map    v.13.01
      err = get_worldmap_position(flati,flongi,mx,my);
      if (! err) {
         for (px = -1; px <= 1; px++)                                      //  paint 3x3 block of pixels
         for (py = -1; py <= 1; py++) 
         {
            pixel = PXBpix(Wpxb,mx+px,my+py);
            pixel[0] = 255;
            pixel[1] = pixel[2] = 0;
         }
      }
   }

   return retval;

badcoord:
   zmessageACK(mWin,ZTX("bad latitude/longitude: %s %s"),coord[0],coord[1]);
   return -1;
}


/**************************************************************************/

//  show a world map and get geotag location from mouse click on map

int click_worldmap()                                                       //  v.12.09
{
   void click_worldmap_mousefunc();
   
   if (! init_geotags()) return 0;                                         //  initialize geotags

   if (cTab != tabW)                                                       //  open tab W if not already
      if (! settab("W")) return 0;

   takeMouse(click_worldmap_mousefunc,dragcursor);                         //  connect mouse 
   return 1;
}


//  Respond to mouse movement and left clicks on world map.
//  Set longitude and latitude, and city and country.

void click_worldmap_mousefunc()                                            //  v.12.09
{
   int         err, mx, my, ii, minii;
   char        *city, *country;
   float       flati, flongi, glati, glongi;
   float       dist, mindist;
   
   if (cTab != tabW) {                                                     //  no longer in tab W
      poptext_window(0,0,0,0);
      freeMouse();
      return;
   }
   
   if ((Mxdrag || Mydrag)) return;                                         //  handle normally (pan/scroll)
   if (RMclick) return;                                                    //  handle normally (zoom - fit window)
   
   mx = Mxposn;                                                            //  mouse position, image space
   my = Myposn;

   err = get_worldmap_coordinates(mx,my,flati,flongi);                     //  convert map location to lat/long
   if (err) return;

   dist = mindist = 999999;
   minii = 0;

   for (ii = 0; ii < Ngeolocs; ii++)                                       //  find nearest city/country
   {
      glati = geolocs[ii].flati;
      dist = (flati - glati) * (flati - glati);
      if (dist > mindist) continue;
      glongi = geolocs[ii].flongi;
      dist += (flongi - glongi) * (flongi - glongi);                       //  degrees**2
      if (dist > mindist) continue;
      mindist = dist;
      minii = ii;
   }

   mindist = 111 * sqrtf(mindist);                                         //  degrees**2 to km at earth surface

   if (mindist <= worldmap_range) {                                        //  within range of a known place
      ii = minii;
      city = geolocs[ii].city;
      country = geolocs[ii].country;
      flati = geolocs[ii].flati;
      flongi = geolocs[ii].flongi;
   }
   else  city = country = 0;

   if (LMclick)                                                            //  left mouse click
   {
      LMclick = 0;
      
      if (zd_geotags) {
         zdialog_stuff(zd_geotags,"city",city);                            //  stuff calling dialog   v.12.12
         zdialog_stuff(zd_geotags,"country",country);
         zdialog_stuff(zd_geotags,"latitude",flati);
         zdialog_stuff(zd_geotags,"longitude",flongi);
         poptext_window(0,0,0,0);                                          //  erase popup text
         zdialog_send_event(zd_geotags,"worldmap");                        //  activate calling dialog
      }
      
      else if (worldmap_callback && city) {                                //  callback function      v.12.12
         poptext_window(0,0,0,0);                                          //  erase popup text
         freeMouse();                                                      //  disconnect mouse
         worldmap_callback(flati,flongi);                                  //  do callback function
      }
   }

   else if (city)                                                          //  mouse movement, no click
      poptext_window(geolocs[ii].city,MWIN,Mwxposn,Mwyposn);               //  popup the city name at mouse

   else poptext_window(0,0,0,0);                                           //  erase popup text
   return;
}


/**************************************************************************/

//  World map functions.
//  See the Wikipedia article on Mercator Projection for the math.

namespace worldmap 
{
   int         xmargin = 112;
   int         ymargin = 90;
   int         mapw = 13242;                                               //  map width less margins
   int         maph = 8827;                                                //  map height less margins
   int         mape = 6217;                                                //  equator less margins
   double      px, py;
   double      R, rad = 180.0 / pi;
}


//  Adjust latitude based on a correction table.
//  Input is latitude ranged 0 to +90.

double worldmap_fudgelat(double flat)
{
   int      ii, jj;
   double   span;
   double   fudge[10] = { 0.0, 0.06, 0.13, 0.15, 0.20, 0.18, 0.16, 0.12, 0.07, 0.0 };

   flat = 0.1 * flat;
   for (ii = 1; ii < 10; ii++)
      if (flat < ii) break;
   if (ii > 9) ii = 9;
   jj = ii - 1;
   span = flat - jj;
   return fudge[jj] + span * (fudge[ii] - fudge[jj]);
}  


//  Convert a world map position mx/my into latitude and longitude.
//  Return 0 if OK, +N if error (off the map).

int get_worldmap_coordinates(int mx, int my, float &flat, float &flong)    //  v.12.10
{
   using namespace worldmap;
   
   flat = flong = 0;
   
   mx = mx - xmargin;                                                      //  remove margins
   my = my - ymargin;

   if (mx < 0 || mx > mapw-1) return 1;                                    //  check if on the map
   if (my < 0 || my > maph-1) return 1;
   
   px = (2.0 * mx - mapw) / mapw;                                          //  -1 to +1
   flong = 180.0 * px;                                                     //  longitude -180 to +180

   py = fabs(mape - my);                                                   //  displacement from equator
   R = 2.0 * pi * py / mapw;
   R = exp(R);
   R = 2.0 * atan(R) * rad;
   flat = R - 90.0;                                                        //  latitude 0 to +90

   flat += worldmap_fudgelat(flat);                                        //  add map correction

   if (my > mape) flat = -flat;                                            //  S. latitude 0 to -90
   
   return 0;
}


//  Convert latitude and longitude into a world map position mx/my;
//  Return 0 if OK, +N if error (off the map).

int get_worldmap_position(float flat, float flong, int &mx, int &my)       //  v.12.10
{
   using namespace worldmap;
   
   mx = my = 0;

   if (flat < -90 || flat > 90) return 1;                                  //  check if out of bounds
   if (flong < -180 || flong > 180) return 1;
   
   mx = mapw * (0.5 + flong / 360.0) + 0.5;                                //  0 to mapw

   if (flat >= 0) flat -= worldmap_fudgelat(flat);                         //  subtract map correction
   else flat += worldmap_fudgelat(-flat);

   flat = 0.5 * flat / rad + pi / 4.0;
   py = 0.5 * mapw / pi * log(tan(flat));                                  //  displacement from equator
   my = mape - py + 0.5;                                                   //  0 to maph
   
   if (mx < 0 || mx > mapw) return 1;
   if (my < 0 || my > maph) return 1;
   
   mx += xmargin;                                                          //  add margins
   my += ymargin;

   return 0;
}


//  World map test function.
//  Convert clicked position into earth coordinates and 
//  then the earth coordinates back into the clicked position.
//  Results: accuracy is about 0.02 degrees or 2 km.

void m_worldmap_test(GtkWidget *, cchar *menu)                             //  v.12.10
{
   void worldmap_test_mousefunc();
   
   if (! init_geotags()) return;                                           //  initialize geotags

   if (cTab != tabW)                                                       //  open tab W if not already
      if (! settab("W")) return;

   takeMouse(worldmap_test_mousefunc,dragcursor);                          //  connect mouse 

   return;
}


//  Respond to mouse movement and clicks on world map.
//  Set longitude and latitude, and city and country.

void worldmap_test_mousefunc()                                             //  v.12.10
{
   int         err, mx1, my1, mx2, my2;
   float       flati, flongi;
   char        text[100];
   
   if (cTab != tabW) {                                                     //  no longer in tab W
      poptext_window(0,0,0,0);
      freeMouse();
      return;
   }

   if (! LMclick) return;
   LMclick = 0;
   
   mx1 = Mxposn;                                                           //  mouse position, image space
   my1 = Myposn;

   err = get_worldmap_coordinates(mx1,my1,flati,flongi);                   //  convert map location to lat/long
   if (err) return;

   err = get_worldmap_position(flati,flongi,mx2,my2);                      //  convert lat/long to map location
   if (err) return;
   
   sprintf(text,"%d/%d  %.2f/%.2f  %d/%d",mx1,my1,flati,flongi,mx2,my2);
   poptext_window(text,MWIN,Mwxposn,Mwyposn);

   return;
}


/**************************************************************************/

//  add geotags to an image - city, country, latitude, longitude

void m_edit_geotags(GtkWidget *, cchar *menu)
{
   int  edit_geotags_dialog_event(zdialog *zd, cchar *event);

/**
                  Edit Geotags

      city [______________]  country [______________]
      latitude [_______] longitude [_______]
      
      [find] [web] [prev] [map] [apply] [clear] [cancel]

**/

   cchar    *title = ZTX("Edit Geotags");
   zdialog  *zd;

   zfuncs::F1_help_topic = "edit_geotags";

   if (checkmenulock()) return;                                            //  menu locked               v.12.09
   if (mod_keep()) return;                                                 //  unsaved edits
   if (! check_index_sync()) return;                                       //  index fle sync disabled   v.12.08
   if (! init_geotags()) return;                                           //  initialize geotags
   if (! curr_file) return;

   if (! zdaddgeotags)                                                     //  start dialog if not already
   {   
      zdaddgeotags = zdialog_new(title,mWin,Bfind,Bweb,Bprev,Bmap,Bapply,Bclear,Bcancel,null);
      zd = zdaddgeotags;
      zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
      zdialog_add_widget(zd,"label","labcity","hb1",ZTX("city"),"space=5");
      zdialog_add_widget(zd,"entry","city","hb1",0,"expand");
      zdialog_add_widget(zd,"label","space","hb1",0,"space=5");
      zdialog_add_widget(zd,"label","labcountry","hb1",ZTX("country"),"space=5");
      zdialog_add_widget(zd,"entry","country","hb1",0,"expand");
      zdialog_add_widget(zd,"hbox","hb4","dialog",0,"space=3");
      zdialog_add_widget(zd,"label","lablat","hb4","Latitude","space=3");
      zdialog_add_widget(zd,"entry","latitude","hb4",0,"scc=10");
      zdialog_add_widget(zd,"label","space","hb4",0,"space=5");
      zdialog_add_widget(zd,"label","lablong","hb4","Longitude","space=3");
      zdialog_add_widget(zd,"entry","longitude","hb4",0,"scc=10");

      zdialog_help(zd,"edit_geotags");                                     //  zdialog help topic
      zdialog_run(zd,edit_geotags_dialog_event);
   }
   
   load_filemeta(curr_file);                                               //  get current image geotags (EXIF)
   zd = zdaddgeotags;

   if (strEqu(meta_city,"null")) zdialog_stuff(zd,"city","");              //  v.13.01
   else zdialog_stuff(zd,"city",meta_city);

   if (strEqu(meta_country,"null")) zdialog_stuff(zd,"country","");
   else zdialog_stuff(zd,"country",meta_country);

   if (strEqu(meta_latitude,"null")) zdialog_stuff(zd,"latitude","");
   else zdialog_stuff(zd,"latitude",meta_latitude);

   if (strEqu(meta_longitude,"null")) zdialog_stuff(zd,"longitude","");
   else zdialog_stuff(zd,"longitude",meta_longitude);

   return;
}


//  dialog event and completion callback function

int edit_geotags_dialog_event(zdialog *zd, cchar *event)                   //  new v.12.05
{
   int          zstat, Nmatch;
   char         *location[2], *coord[2], *matches[20][2];
   char         city[100], country[100];
   char         latitude[20], longitude[20], temp[20];
   float        fcoord[2];
   cchar        *errmess;

   if (strEqu(event,"worldmap"))                                           //  have geotags data from world map
      return 1;                                                            //  do nothing                      v.12.12

   if (! zd->zstat) return 1;                                              //  wait for action button
   zstat = zd->zstat;
   zd->zstat = 0;                                                          //  keep dialog active
   settab("F");                                                            //  back to image                   v.12.12

   zdialog_fetch(zd,"city",city,99);                                       //  get city [country] from dialog
   zdialog_fetch(zd,"country",country,99);
   location[0] = city;
   location[1] = country;
   zdialog_fetch(zd,"latitude",temp,12);                                   //  and latitude, longitude
   repl_1str(temp,latitude,",",".");                                       //  (replace comma decimal)         v.13.01
   zdialog_fetch(zd,"longitude",temp,12);
   repl_1str(temp,longitude,",",".");
   coord[0] = latitude;
   coord[1] = longitude;

   if (zstat == 1)                                                         //  [find]
   {
      Nmatch = get_geotags(location,coord,matches);                        //  find in city-geotags file
      if (Nmatch == 0)                                                     //  no matches
         zmessageACK(mWin,ZTX("city not found"));

      else if (Nmatch == 1) {                                              //  one match
         zdialog_stuff(zd,"city",matches[0][0]);                           //  stuff matching city data into dialog
         zdialog_stuff(zd,"country",matches[0][1]);
         zdialog_stuff(zd,"latitude",coord[0]);
         zdialog_stuff(zd,"longitude",coord[1]);
      }

      else {                                                               //  multiple matching cities
         zstat = geotags_choosecity(location,coord);                       //  ask user to choose one
         if (zstat == 1) {                                                 //  response is available
            zdialog_stuff(zd,"city",location[0]);                          //  stuff matching city data into dialog
            zdialog_stuff(zd,"country",location[1]);
            zdialog_stuff(zd,"latitude",coord[0]);
            zdialog_stuff(zd,"longitude",coord[1]);
         }
      }
   }

   else if (zstat == 2)                                                    //  [web]
   {
      errmess = web_geocode(location,coord);                               //  look-up in web service
      if (errmess) 
         zmessageACK(mWin,errmess);                                        //  fail
      else {
         zdialog_stuff(zd,"city",location[0]);                             //  success, return all data
         zdialog_stuff(zd,"country",location[1]);                          //  (location may have been completed)
         zdialog_stuff(zd,"latitude",coord[0]);
         zdialog_stuff(zd,"longitude",coord[1]);
      }
   }

   else if (zstat == 3)                                                    //  [prev] 
   {
      zdialog_stuff(zd,"city",prevcity);                                   //  get last-used geotags
      zdialog_stuff(zd,"country",prevcountry);
      zdialog_stuff(zd,"latitude",prevlatitude);
      zdialog_stuff(zd,"longitude",prevlongitude);
   }

   else if (zstat == 4)                                                    //  [map]
   {
      zd_geotags = zd;                                                     //  open world map and get geotags
      click_worldmap();                                                    //    via mouse click
   }
   
   else if (zstat == 5)                                                    //  [apply]
   {
      if (strEqu(coord[0],"null")) *coord[0] = 0;                          //  replace "null" with ""    v.12.06.2
      if (strEqu(coord[1],"null")) *coord[1] = 0;

      if (*coord[0] || *coord[1]) {                                        //  if coordinates present, validate
         if (! *coord[0] || ! *coord[1]) goto badcoord;
         fcoord[0] = atof(coord[0]);
         if (fcoord[0] < -90 || fcoord[0] > 90) goto badcoord;
         fcoord[1] = atof(coord[1]);
         if (fcoord[1] < -180 || fcoord[1] > 180) goto badcoord;
         if (fcoord[0] == 0 && fcoord[1] == 0) goto badcoord;
      }

      put_geotags(location,coord);                                         //  update EXIF and geotags city data

      strcpy(prevcity,city);                                               //  save data for later use as [prev]
      strcpy(prevcountry,country);
      strcpy(prevlatitude,latitude);
      strcpy(prevlongitude,longitude);
   }

   else if (zstat == 6)                                                    //  [clear]
   {
      zdialog_stuff(zd,"city","");                                         //  erase dialog fields
      zdialog_stuff(zd,"country","");
      zdialog_stuff(zd,"latitude","");
      zdialog_stuff(zd,"longitude","");
   }

   else {                                                                  //  cancel
      zdialog_free(zd);
      zdaddgeotags = 0;
      zd_geotags = 0;                                                      //  deactivate world map clicks
      settab("F"); 
   }

   return 1;

badcoord:
   zmessageACK(mWin,ZTX("bad latitude/longitude: %s %s"),coord[0],coord[1]);
   return 1;
}


/**************************************************************************/

//  batch add geotags - set geotags for multiple image files

char     **batch_add_geotags_filelist = 0;
int      batch_add_geotags_filecount = 0;

void m_batch_add_geotags(GtkWidget *, cchar *menu)                         //  new v.12.05
{
   int   batch_add_geotags_dialog_event(zdialog *zd, cchar *event);

   cchar       *title = ZTX("Batch Add Geotags");
   char        *file, **flist;
   zdialog     *zd;
   int         ii, err;
   char        *location[2], *coord[2];
   char        city[100], country[100];
   char        latitude[20], longitude[20];

   zfuncs::F1_help_topic = "batch_add_geotags";

   if (checkmenulock()) return;                                            //  menu locked               v.12.09
   if (mod_keep()) return;                                                 //  unsaved edits
   if (! check_index_sync()) return;                                       //  index fle sync disabled   v.12.08
   if (! init_geotags()) return;                                           //  initialize geotags

/**
                Batch Add Geotags

      [select files]  NN files selected
      city [______________]  country [______________]
      latitude [_______] longitude [_______]
      
      [find] [web] [prev] [map] [proceed] [cancel]

**/

   zd = zdialog_new(title,mWin,Bfind,Bweb,Bprev,Bmap,Bproceed,Bcancel,null);

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","files","hb1",Bselectfiles,"space=10");
   zdialog_add_widget(zd,"label","labcount","hb1","0 files selected","space=10");
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labcity","hb2",ZTX("city"),"space=5");
   zdialog_add_widget(zd,"entry","city","hb2",0,"expand");
   zdialog_add_widget(zd,"label","space","hb2",0,"space=5");
   zdialog_add_widget(zd,"label","labcountry","hb2",ZTX("country"),"space=5");
   zdialog_add_widget(zd,"entry","country","hb2",0,"expand");
   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","lablat","hb3","Latitude","space=3");
   zdialog_add_widget(zd,"entry","latitude","hb3",0,"scc=10");
   zdialog_add_widget(zd,"label","space","hb3",0,"space=5");
   zdialog_add_widget(zd,"label","lablong","hb3","Longitude","space=3");
   zdialog_add_widget(zd,"entry","longitude","hb3",0,"scc=10");

   batch_add_geotags_filelist = 0;
   batch_add_geotags_filecount = 0;
   flist = 0;

   zdialog_help(zd,"batch_add_geotags");
   zdialog_run(zd,batch_add_geotags_dialog_event);                         //  run dialog
   zdialog_wait(zd);                                                       //  wait for dialog completion

   if (zd->zstat != 5) goto cleanup;                                       //  status not [proceed]
   if (! batch_add_geotags_filecount) goto cleanup;                        //  no files selected

   zdialog_fetch(zd,"city",city,99);                                       //  get city [country] from dialog
   zdialog_fetch(zd,"country",country,99);
   location[0] = city;
   location[1] = country;
   zdialog_fetch(zd,"latitude",latitude,12);                               //  and latitude, longitude
   zdialog_fetch(zd,"longitude",longitude,12);
   coord[0] = latitude;
   coord[1] = longitude;
   
   zdialog_free(zd);                                                       //  kill dialog
   zd_geotags = 0;                                                         //  deactivate world map clicks

   flist = batch_add_geotags_filelist;                                     //  selected files
   if (! flist) goto cleanup;

   write_popup_text("open","Adding Geotags",500,200,mWin);                 //  status monitor popup window

   for (ii = 0; flist[ii]; ii++)                                           //  loop all selected files
   {
      file = flist[ii];                                                    //  display image
      err = f_open(file,0,0,0,0);
      if (err) continue;

      write_popup_text("write",file);                                      //  report progress
      zmainloop();

      load_filemeta(file);                                                 //  load current file tags
      put_geotags(location,coord);                                         //  update geotags part
   }

   write_popup_text("write","COMPLETED");
   write_popup_text("close",0);
   
cleanup:

   if (zd) zdialog_free(zd);
   zd_geotags = 0;                                                         //  deactivate world map clicks

   if (flist) {
      for (ii = 0; flist[ii]; ii++) 
         zfree(flist[ii]);
      zfree(flist);
   }

   return;
}


//  batch_add_geotags dialog event function

int batch_add_geotags_dialog_event(zdialog *zd, cchar *event)
{
   int      ii, yn, zstat, Nmatch;
   char     **flist = batch_add_geotags_filelist;
   char     countmess[50];
   char     *location[2], *coord[2], *matches[20][2];
   char     city[100], country[100];
   char     latitude[20], longitude[20];
   cchar    *errmess;
   float    fcoord[2];

   if (strEqu(event,"files"))                                              //  select images to add tags
   {
      if (flist) {                                                         //  free prior list
         for (ii = 0; flist[ii]; ii++) 
            zfree(flist[ii]);
         zfree(flist);
      }

      zdialog_show(zd,0);                                                  //  hide parent dialog              v.12.11
      flist = gallery_getfiles();                                          //  get file list from user
      zdialog_show(zd,1);

      batch_add_geotags_filelist = flist;

      if (flist)                                                           //  count files in list
         for (ii = 0; flist[ii]; ii++);
      else ii = 0;
      batch_add_geotags_filecount = ii;
      
      snprintf(countmess,50,ZTX("%d files selected"),batch_add_geotags_filecount);
      zdialog_stuff(zd,"labcount",countmess);
   }

   if (! zd->zstat) return 1;                                              //  wait for action button
   zstat = zd->zstat;
   zd->zstat = 0;                                                          //  keep dialog active

   zdialog_fetch(zd,"city",city,99);                                       //  get city [country] from dialog
   zdialog_fetch(zd,"country",country,99);
   location[0] = city;
   location[1] = country;
   zdialog_fetch(zd,"latitude",latitude,12);                               //  and latitude, longitude
   zdialog_fetch(zd,"longitude",longitude,12);
   coord[0] = latitude;
   coord[1] = longitude;

   if (zstat == 1)                                                         //  [find]
   {
      Nmatch = get_geotags(location,coord,matches);                        //  find in city-geotags file
      if (Nmatch == 0)                                                     //  no matches
         zmessageACK(mWin,ZTX("city not found"));

      else if (Nmatch == 1) {                                              //  one match
         zdialog_stuff(zd,"city",matches[0][0]);                           //  stuff matching city data into dialog
         zdialog_stuff(zd,"country",matches[0][1]);
         zdialog_stuff(zd,"latitude",coord[0]);
         zdialog_stuff(zd,"longitude",coord[1]);
      }

      else {                                                               //  multiple matching cities
         zstat = geotags_choosecity(location,coord);                       //  ask user to choose one
         if (zstat == 1) {                                                 //  response is available
            zdialog_stuff(zd,"city",location[0]);                          //  stuff matching city data into dialog
            zdialog_stuff(zd,"country",location[1]);
            zdialog_stuff(zd,"latitude",coord[0]);
            zdialog_stuff(zd,"longitude",coord[1]);
         }
      }
   }

   else if (zstat == 2)                                                    //  [web]
   {
      errmess = web_geocode(location,coord);                               //  look-up in web service
      if (errmess) 
         zmessageACK(mWin,errmess);                                        //  fail
      else {
         zdialog_stuff(zd,"city",location[0]);                             //  success, return all data
         zdialog_stuff(zd,"country",location[1]);                          //  (location may have been completed)
         zdialog_stuff(zd,"latitude",coord[0]);
         zdialog_stuff(zd,"longitude",coord[1]);
      }
   }

   else if (zstat == 3)                                                    //  [prev] 
   {
      zdialog_stuff(zd,"city",prevcity);                                   //  get last-used geotags
      zdialog_stuff(zd,"country",prevcountry);
      zdialog_stuff(zd,"latitude",prevlatitude);
      zdialog_stuff(zd,"longitude",prevlongitude);
   }

   else if (zstat == 4)                                                    //  [map]
   {
      zd_geotags = zd;                                                     //  open world map and get geotags
      click_worldmap();                                                    //    via mouse click
   }
   
   else if (zstat == 5)                                                    //  [proceed]
   {
      if (*coord[0] || *coord[1]) {                                        //  if coordinates present, validate
         if (! *coord[0] || ! *coord[1]) goto badcoord;
         fcoord[0] = atof(coord[0]);
         if (fcoord[0] < -90 || fcoord[0] > 90) goto badcoord;
         fcoord[1] = atof(coord[1]);
         if (fcoord[1] < -180 || fcoord[1] > 180) goto badcoord;
         if (fcoord[0] == 0 && fcoord[1] == 0) goto badcoord;
      }

      if (! batch_add_geotags_filecount) goto nofiles;

      if (! *city || ! *country || ! *latitude || ! *longitude) {          //  check data is complete
         yn = zmessageYN(mWin,ZTX("data is incomplete \n proceed?"));
         if (! yn) return 1;
      }

      strcpy(prevcity,city);                                               //  save data for later use as [prev]
      strcpy(prevcountry,country);
      strcpy(prevlatitude,latitude);
      strcpy(prevlongitude,longitude);

      zd->zstat = 5;                                                       //  OK to proceed
      zdialog_destroy(zd);
   }

   else {                                                                  //  cancel                 v.12.09
      zdialog_destroy(zd);
      zd_geotags = 0;                                                      //  deactivate world map clicks
      settab("F");
   }

   return 1;

badcoord:
   zmessageACK(mWin,ZTX("bad latitude/longitude: %s %s"),coord[0],coord[1]);
   return 1;

nofiles:
   zmessageACK(mWin,ZTX("no files selected"));
   return 1;
}


//  dialog to choose one city from multiple options
//  location[2] is input city and optional country
//  (may be substrings, may have multiple matches in city geotags data)
//  location[2] is output unique city and country after user choice
//  coord[2] is output latitude, longitude

char     *geotags_chosenlocation[2];
char     *geotags_chosencoord[2];

int geotags_choosecity(char *location[2], char *coord[2])                  //  new v.12.05
{
   int  geotags_choosecity_event(zdialog *zd, cchar *event);

   char     *matches[20][2], text[200];
   int      Nmatch, ii, zstat;
   zdialog  *zd;

   Nmatch = get_geotags(location,coord,matches);                           //  get matching city geotags data

   if (Nmatch == 0) return 0;                                              //  no match

   if (Nmatch == 1) {                                                      //  one match, done
      location[0] = matches[0][0];
      location[1] = matches[0][1];
      return 1;
   }

   zd = zdialog_new(ZTX("choose city"),mWin,BOK,Bcancel,null);             //  multiple matches, start dialog
   zdialog_add_widget(zd,"comboE","cities","dialog",0,"space=5");
   for (ii = 0; ii < Nmatch; ii++) {                                       //  list matching cities to choose from
      snprintf(text,200,"%s | %s",matches[ii][0],matches[ii][1]);
      zdialog_cb_app(zd,"cities",text);                                    //  duplicates are removed
   }

   zdialog_resize(zd,300,100);
   zdialog_run(zd,geotags_choosecity_event);                               //  run dialog, wait for completion
   zstat = zdialog_wait(zd);
   zdialog_free(zd);

   if (zstat == 1) {                                                       //  valid response available
      location[0] = geotags_chosenlocation[0];
      location[1] = geotags_chosenlocation[1];
      coord[0] = geotags_chosencoord[0];
      coord[1] = geotags_chosencoord[1];
      return 1;
   }
   
   return 0;
}


//  dialog event function - get chosen city/country from multiple choices

int geotags_choosecity_event(zdialog *zd, cchar *event)
{
   char           text[200];
   static char    city[100], country[100];
   char           *location[2], *coord[2];
   char           *matches[20][2];
   cchar          *pp;
   int            nn;
   static int     ftf = 1;
   
   if (ftf) {
      zdialog_cb_popup(zd,"cities");                                       //  first time, open combo box list
      ftf = 0;
   }

   if (strEqu(event,"cities")) {                                           //  OK
      zdialog_fetch(zd,"cities",text,200);

      pp = strField(text,'|',1);
      if (pp) strncpy0(city,pp,99);
      pp = strField(text,'|',2);
      if (pp) strncpy0(country,pp,99);
      strTrim2(city);
      strTrim2(country);
      location[0] = city;
      location[1] = country;

      nn = get_geotags(location,coord,matches);                            //  find in city geotags data
      if (nn) {
         geotags_chosenlocation[0] = location[0];                          //  use 1st match if > 1
         geotags_chosenlocation[1] = location[1];
         geotags_chosencoord[0] = coord[0];
         geotags_chosencoord[1] = coord[1];
         zd->zstat = 1;
      }
      else zd->zstat = 2;                                                  //  bad status
   }

   if (zd->zstat) ftf = 1;

   return 1;
}


/**************************************************************************/

//  convert a city [country] to latitude/longitude using Yahoo web service
//  (incomplete names may be completed with a bad guess)

cchar * web_geocode(char *location[2], char *coord[2])
{
   int            err, cc;
   static char    city[100], country[100], latitude[20], longitude[20];
   char           outfile[100], URI[200], request[300];
   char           *pp1, *pp2, buffer[1000];
   FILE           *fid;
   cchar          *notfound = ZTX("not found");
   
   *coord[0] = *coord[1] = 0;                                              //  null outputs
   *city = *country = *latitude = *longitude = 0;
   
   snprintf(outfile,100,"%s/web-data",geotags_dirk);
   snprintf(URI,199,"\"http://where.yahooapis.com/geocode?q=%s,%s\"",location[0],location[1]);
   snprintf(request,299,"wget -o /dev/null -O %s %s",outfile,URI);

   err = system(request);                                                  //  send web query
   if (err) return wstrerror(err);
   
   fid = fopen(outfile,"r");                                               //  get response
   if (! fid) return notfound;
   
   while (true)
   {
      pp1 = fgets(buffer,999,fid);
      if (! pp1) break;

      pp1 = strstr(buffer,"<line2>");                                      //  parse <line2> ... </line2>
      if (pp1) {
         pp2 = strstr(pp1+7,"</line2>");
         if (pp2) {
            cc = pp2 - pp1 - 7;
            if (cc < 99) strncpy0(city,pp1+7,cc+1);
         }
      }

      pp1 = strstr(buffer,"<line4>");                                      //  parse <line4> ... </line4>
      if (pp1) {
         pp2 = strstr(pp1+7,"</line4>");
         if (pp2) {
            cc = pp2 - pp1 - 7;
            if (cc < 99) strncpy0(country,pp1+7,cc+1);
         }
      }

      pp1 = strstr(buffer,"<latitude>");                                   //  parse <latitude> ... </latitude>
      if (pp1) {
         pp2 = strstr(pp1+10,"</latitude>");
         if (pp2) {
            cc = pp2 - pp1 - 10;
            if (cc > 12) cc = 12;
            strncpy0(latitude,pp1+10,cc+1);
         }
      }

      pp1 = strstr(buffer,"<longitude>");                                  //  parse <longitude> ... </longitude>
      if (pp1) {
         pp2 = strstr(pp1+10,"</longitude>");
         if (pp2) {
            cc = pp2 - pp1 - 11;
            if (cc > 12) cc = 12;
            strncpy0(longitude,pp1+11,cc+1);
         }
      }
   }
   
   fclose(fid);
   
   if (! *city || ! *country || ! *latitude || ! *longitude)               //  check completeness of response
      return notfound;

   if (*latitude == '0' && *longitude == '0')                              //  disallow coordinate (0,0)
      return notfound;

   pp1 = strchr(latitude,'.');                                             //  keep max. 4 decimal digits
   if (pp1) *(pp1+5) = 0;
   pp1 = strchr(longitude,'.');
   if (pp1) *(pp1+5) = 0;
   
   location[0] = city;                                                     //  return all data
   location[1] = country;
   coord[0] = latitude;
   coord[1] = longitude;

   return 0;
}


/**************************************************************************/

//  Group images by location and date, with a count of images in each group. 
//  Click on a group to get a thumbnail gallery of all images in the group.

int   ggroups_comp(cchar *rec1, cchar *rec2);
void  ggroups_click(GtkWidget *widget, int &line, int &pos); 
int   ggroups_getdays(cchar *date);

struct grec_t  {                                                           //  image geotags data
   char        *city, *country;                                            //  group location
   char        pdate[12];                                                  //  nominal group date, yyyy:mm:dd
   int         lodate, hidate;                                             //  range, days since 0 CE
   int         count;                                                      //  images in group
};

grec_t   *grec = 0;
int      Ngrec = 0;
int      ggroups_groupby, ggroups_daterange;


void m_geotag_groups(GtkWidget *, cchar *)                                 //  v.12.07
{
   zdialog        *zd;
   int            zstat, ftf, err, cc, cc1, cc2;
   int            iix, iig, newgroup;
   char           country[100], city[100], buff[300];
   cchar          *pp;
   sxrec_t        sxrec;
   GtkWidget      *textwin;

   zfuncs::F1_help_topic = "geotag_groups";
   if (! check_index_sync()) return;                                       //  index fle sync disabled      v.12.08

/***
            Report Geotag Groups                                           //  v.12.07
         
         (o) Group by country
         (o) Group by country/city
         (o) Group by country/city/date
             Combine within [ xx |-|+] days
          
                        [proceed]  [cancel]
         
***/

   zd = zdialog_new(ZTX("Report Geotag Groups"),mWin,Bproceed,Bcancel,null);
   zdialog_add_widget(zd,"radio","country","dialog",ZTX("Group by country"));
   zdialog_add_widget(zd,"radio","city","dialog",ZTX("Group by country/city"));
   zdialog_add_widget(zd,"radio","date","dialog",ZTX("Group by country/city/date"));
   zdialog_add_widget(zd,"hbox","hbr","dialog");
   zdialog_add_widget(zd,"label","space","hbr",0,"space=10");
   zdialog_add_widget(zd,"label","labr1","hbr",ZTX("Combine within"),"space=10");
   zdialog_add_widget(zd,"spin","range","hbr","0|999|1|1");
   zdialog_add_widget(zd,"label","labr2","hbr",ZTX("days"),"space=10");
   
   zdialog_stuff(zd,"country",0);
   zdialog_stuff(zd,"city",1);
   zdialog_stuff(zd,"date",0);
   
   zdialog_help(zd,"geotag_groups");
   zdialog_resize(zd,300,0);
   zdialog_run(zd);
   zstat = zdialog_wait(zd);
   if (zstat != 1) {
      zdialog_free(zd);
      return;
   }
   
   zdialog_fetch(zd,"country",iix);
   if (iix) ggroups_groupby = 1;                                           //  group by country
   zdialog_fetch(zd,"city",iix);
   if (iix) ggroups_groupby = 2;                                           //  group by country/city
   zdialog_fetch(zd,"date",iix);
   if (iix) ggroups_groupby = 3;                                           //  group by country/city/date (range)
   zdialog_fetch(zd,"range",ggroups_daterange);
   zdialog_free(zd);
   
   if (Ngrec) {                                                            //  free prior memory
      for (iix = 0; iix < Ngrec; iix++) {
         if (grec[iix].city) zfree(grec[iix].city);
         if (grec[iix].country) zfree(grec[iix].country);
      }
      zfree(grec);
   }

   cc = maximages * sizeof(grec_t);                                        //  allocate memory
   grec = (grec_t *) zmalloc(cc,"geotag_groups");
   memset(grec,0,cc);

   Ngrec = 0;
   ftf = 1;

   while (true)
   {
      err = read_search_index_seq(sxrec,ftf);                              //  read search index recs.      v.12.10
      if (err) break;
      
      iix = Ngrec;

      pp = strField(sxrec.gtags,'^',1);                                    //  get city
      if (pp) grec[iix].city = strdupz(pp,0,"geotag_groups");
      else grec[iix].city = strdupz("null",0,"geotag_groups");

      pp = strField(sxrec.gtags,'^',2);                                    //  country
      if (pp) grec[iix].country = strdupz(pp,0,"geotag_groups");
      else grec[iix].country = strdupz("null",0,"geotag_groups");

      strncpy0(grec[iix].pdate,sxrec.pdate,12);
      grec[iix].lodate = ggroups_getdays(sxrec.pdate);                     //  days since 0 CE
      grec[iix].hidate = grec[iix].lodate;

      if (++Ngrec == maximages) {
         zmessageACK(mWin,"too many image files");
         return;
      }

      zfree(sxrec.file);
      zfree(sxrec.tags);
      zfree(sxrec.capt);
      zfree(sxrec.comms);
      zfree(sxrec.gtags);
   }

   if (! Ngrec) {
      zmessageACK(mWin,"no geotags data found");
      return;
   }

   if (Ngrec > 1)                                                          //  sort index by country/city/date
      HeapSort((char *) grec, sizeof(grec_t), Ngrec, ggroups_comp);

   iig = 0;                                                                //  1st group from grec[0]
   grec[iig].count = 1;                                                    //  group count = 1

   for (iix = 1; iix < Ngrec; iix++)                                       //  scan following grecs
   {
      newgroup = 0;

      if (strNeq(grec[iix].country,grec[iig].country)) 
         newgroup = 1;                                                     //  new country >> new group
      
      if (ggroups_groupby >= 2) 
         if (strNeq(grec[iix].city,grec[iig].city)) newgroup = 1;          //  new city >> new group if group by city

      if (ggroups_groupby >= 3) 
         if (grec[iix].lodate - grec[iig].hidate > ggroups_daterange)      //  new date >> new group if group by date
            newgroup = 1;                                                  //    and date out of range

      if (newgroup)
      {
         iig++;                                                            //  new group
         if (iix > iig) {
            grec[iig] = grec[iix];                                         //  copy and pack down
            grec[iix].city = grec[iix].country = 0;                        //  no zfree()
         }
         grec[iig].count = 1;                                              //  group count = 1
      }
      else
      {
         zfree(grec[iix].city);                                            //  same group
         zfree(grec[iix].country);                                         //  free memory
         grec[iix].city = grec[iix].country = 0;
         grec[iig].hidate = grec[iix].lodate;                              //  expand group date range
         grec[iig].count++;                                                //  increment group count
      }
   }

   Ngrec = iig + 1;                                                        //  unique groups count

   textwin = write_popup_text("open","geotag groups",620,400,mWin);        //  write groups to popup window
   
   if (ggroups_groupby == 1)                                               //  group by country
   {
      sprintf(buff,"%-30s  %5s ","Country","Count");
      write_popup_text("write",buff);

      for (iig = 0; iig < Ngrec; iig++)
      {
         utf8substring(country,grec[iig].country,0,30);
         cc1 = 30 + strlen(country) - utf8len(country);
         sprintf(buff,"%-*s  %5d ",cc1,country,grec[iig].count);
         write_popup_text("write",buff);
      }
   }
   
   if (ggroups_groupby == 2)                                               //  group by country/city
   {
      sprintf(buff,"%-30s  %-30s  %5s ","Country","City","Count");
      write_popup_text("write",buff);

      for (iig = 0; iig < Ngrec; iig++)
      {
         utf8substring(country,grec[iig].country,0,30);
         cc1 = 30 + strlen(country) - utf8len(country);
         utf8substring(city,grec[iig].city,0,30);
         cc2 = 30 + strlen(city) - utf8len(city);
         sprintf(buff,"%-*s  %-*s  %5d ",
                  cc1,country,cc2,city,grec[iig].count);
         write_popup_text("write",buff);
      }
   }
   
   if (ggroups_groupby == 3)                                               //  group by country/city/date (range)
   {
      sprintf(buff,"%-30s  %-30s  %-12s  %5s ","Country","City","Date","Count");
      write_popup_text("write",buff);

      for (iig = 0; iig < Ngrec; iig++)
      {
         utf8substring(country,grec[iig].country,0,30);
         cc1 = 30 + strlen(country) - utf8len(country);
         utf8substring(city,grec[iig].city,0,30);
         cc2 = 30 + strlen(city) - utf8len(city);
         sprintf(buff,"%-*s  %-*s  %-12s  %5d ",
                  cc1,country,cc2,city,grec[iig].pdate,grec[iig].count);
         write_popup_text("write",buff);
      }
   }

   write_popup_text("top");
   textwidget_set_clickfunc(textwin,ggroups_click);                        //  response function for mouse click
   return;
}


//  Compare 2 grec records by geotags and date, 
//  return < 0  = 0  > 0   for   rec1  <  =  >  rec2.

int ggroups_comp(cchar *rec1, cchar *rec2)                                 //  v.12.06 v.12.07
{
   int      ii;

   char * country1 = ((grec_t *) rec1)->country;                           //  compare countries
   char * country2 = ((grec_t *) rec2)->country;
   ii = strcmp(country1,country2);
   if (ii) return ii;

   char * city1 = ((grec_t *) rec1)->city;                                 //  compare cities
   char * city2 = ((grec_t *) rec2)->city;
   ii = strcmp(city1,city2);
   if (ii) return ii;

   int date1 = ((grec_t *) rec1)->lodate;                                  //  compare dates
   int date2 = ((grec_t *) rec2)->lodate;
   ii = date1 - date2;
   return ii;
}


//  convert yyyy:mm:dd date into days from 0 C.E.
//  "null" date returns 999999 (year 2737)

int ggroups_getdays(cchar *date)                                           //  v.12.07
{
   int   CEdays(int year, int mon, int day);

   int      year, month, day;
   cchar    *pp;

   year = month = day = 0;   

   pp = strField(date,':',1);
   if (pp) year = atoi(pp);
   if (! year) return 999999;
   pp = strField(date,':',2);
   if (pp) month = atoi(pp);
   pp = strField(date,':',3);
   if (pp) day = atoi(pp);
   return CEdays(year,month,day);
}


//   convert year/month/day into days since Jan 1, 0001 (day 0 CE)
//   year is 0001 to 9999, month is 1-12, day is 1-31
     
int CEdays(int year, int month, int day)                                   //  v.12.07
{
   int    montab[12] = { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 };
   int    elaps;
   
   elaps = 365 * (year-1) + (year-1) / 4;                                  //  elapsed days in prior years
   elaps += montab[month-1];                                               //  + elapsed days in prior months
   if (year % 4 == 0 && month > 2) elaps += 1;                             //  + 1 for Feb. 29 
   elaps += day-1;                                                         //  + elapsed days in month
   return elaps;
}


//  Receive clicks on report window and generate gallery of images
//  matching the selected country/city/date

void ggroups_click(GtkWidget *widget, int &line, int &pos)                 //  v.12.06 v.12.07
{
   int      iix, ftf, err, lodate, hidate, datex;
   cchar    *pp;
   char     city[100], country[100];
   char     resultsfile[200];
   FILE     *fid;
   sxrec_t  sxrec;

   iix = line - 1;                                                         //  clicked grec[iix]
   if (iix < 0 || iix > Ngrec-1) return;

   strncpy0(country,grec[iix].country,99);                                 //  selected country/city/date range
   strncpy0(city,grec[iix].city,99);
   lodate = grec[iix].lodate;
   hidate = grec[iix].hidate;
   
   snprintf(resultsfile,199,"%s/search_results",get_zuserdir());           //  output scratch file
   fid = fopen(resultsfile,"w");
   if (! fid) goto writerror;
   
   ftf = 1;

   while (true)                                                            //  read search index recs.
   {
      err = read_search_index_seq(sxrec,ftf);                              //  v.12.10
      if (err) break;
      
      pp = strField(sxrec.gtags,'^',2);
      if (! pp) pp = "null";                                               //  enable search for "null"

      if (strNeq(pp,country)) goto freemem;                                //  no country match
      
      if (ggroups_groupby >= 2) {
         pp = strField(sxrec.gtags,'^',1);
         if (! pp) pp = "null";
         if (strNeq(pp,city)) goto freemem;                                //  no city match
      }

      if (ggroups_groupby == 3) {
         datex = ggroups_getdays(sxrec.pdate);
         if (datex < lodate || datex > hidate) goto freemem;               //  no date match
      }

      fprintf(fid,"%s\n",sxrec.file);                                      //  output matching file
   
   freemem:
      zfree(sxrec.file);
      zfree(sxrec.tags);
      zfree(sxrec.capt);
      zfree(sxrec.comms);
      zfree(sxrec.gtags);
   }

   fclose(fid);

   navi::gallerytype = 2;                                                  //  search results
   gallery(resultsfile,"initF");                                           //  generate gallery of matching files
   gallery(0,"paint");
   return;
   
writerror:
   zmessLogACK(mWin,"output file error: %s",strerror(errno));
   return;
}


/**************************************************************************/

//  Show locations of geotagged images on a world map.
//  Click on world map to get a gallery of images at/near the location.

void m_geotag_worldmap(GtkWidget *, cchar *menu)                           //  v.12.09
{
   void geotag_worldmap(float flati, float flongi);
   
   char     trange[12], *pp;
   int      irange = 0;

   zfuncs::F1_help_topic = "geotag_worldmap";

   if (! check_index_sync()) return;                                       //  index fle sync disabled      v.12.08
   if (! init_geotags()) return;                                           //  initialize geotags

   worldmap_callback = geotag_worldmap;                                    //  set worldmap click callback function
   click_worldmap();                                                       //  wait for worldmap click

   if (menu) {                                                             //  if user menu (not tab W selection)
      while (true) {                                                       //  get a new search range if wanted
         sprintf(trange,"%d",worldmap_range);
         pp = zdialog_text(mWin,ZTX("search range (km)"),trange);
         if (pp) irange = atoi(pp);
         if (pp) zfree(pp);
         if (irange > 0 && irange < 1000) break;
      }
      worldmap_range = irange;
   }

   return;
}


//  click_worldmap callback function - receive clicked geotag location
//  and generate gallery of images geotagged with this location

void geotag_worldmap(float flati, float flongi)                            //  v.12.09
{
   int            ftf, err, nn = 0;
   char           imagefile[maxfcc], resultsfile[200];
   float          glati, glongi, grange;
   cchar          *pp;
   FILE           *fid;
   sxrec_t        sxrec;

   snprintf(resultsfile,199,"%s/search_results",get_zuserdir());           //  output scratch file

   fid = fopen(resultsfile,"w");
   if (! fid) {
      zmessLogACK(mWin,"output file error: %s",strerror(errno));
      return;
   }

   ftf = 1;

   while (true)                                                            //  read search index recs.
   {
      err = read_search_index_seq(sxrec,ftf);                              //  v.12.10
      if (err) break;

      strncpy0(imagefile,sxrec.file,maxfcc);                               //  save filespec
      
      pp = strField(sxrec.gtags,'^',3);                                    //  latitude
      if (! pp) goto freemem;
      if (strEqu(pp,"null")) goto freemem;
      glati = atof(pp);

      pp = strField(sxrec.gtags,'^',4);                                    //  longitude
      if (! pp) goto freemem;
      if (strEqu(pp,"null")) goto freemem;
      glongi = atof(pp);
      
      grange = (flati - glati) * (flati - glati);
      grange += (flongi - glongi) * (flongi - glongi);
      grange = 111.0 * sqrt(grange);

      if (grange <= worldmap_range) {                                      //  within distance limit, select
         fprintf(fid,"%s\n",imagefile);                                    //  output matching file
         nn++;
      }

   freemem:
      zfree(sxrec.file);
      zfree(sxrec.tags);
      zfree(sxrec.capt);
      zfree(sxrec.comms);
      zfree(sxrec.gtags);
   }

   fclose(fid);

   if (! nn) {
      zmessageACK(mWin,ZTX("No matching images found"));
      click_worldmap();                                                    //  try again        v.12.09
      return;
   }
   
   navi::gallerytype = 2;                                                  //  search results
   gallery(resultsfile,"initF");                                           //  generate gallery of matching files
   gallery(0,"paint");

   return;
}


/**************************************************************************/

//  Search image tags, geotags, dates, stars, comments, captions
//  to find matching images. This is very fast using the search index.
//  Search also any metadata, but this is relatively slow.

namespace search_images
{
   zdialog  *zdsearchimages = 0;                                           //  search images dialog

   char     searchDateFrom[12] = "";                                       //  search images
   char     searchDateTo[12] = "";
   char     searchStarsFrom[4] = "";
   char     searchStarsTo[4] = "";

   char     searchCity[100] = "";                                          //  search geotags
   char     searchCountry[100] = "";
   char     searchLatitude[20] = "";
   char     searchLongitude[20] = "";
   float    searchRange = 100;
   float    flatitude = 0, flongitude = 0;

   int      Nsearchkeys = 0;
   char     *searchkeys[5];                                                //  search metadata keys and match data
   char     *searchkeydata[5];
   char     searchkeyx[8], searchkeydatax[8];
}

using namespace search_images;


void m_search_images(GtkWidget *, cchar *)                                 //  overhauled    v.12.05
{
   void  search_searchtags_clickfunc(GtkWidget *widget, int &line, int &pos);
   void  search_deftags_clickfunc(GtkWidget *widget, int &line, int &pos);
   int   searchimages_dialog_event(zdialog*, cchar *event);

   zdialog     *zd;
   GtkWidget   *widget;

   zfuncs::F1_help_topic = "search_images";

   if (checkmenulock()) return;                                            //  menu locked                  v.12.09
   if (mod_keep()) return;                                                 //  unsaved edits
   if (! check_index_sync()) return;                                       //  index fle sync disabled      v.12.08

/***
                  Search Image Metadata

         date range   [_______] [_______] (yyyymmdd)
         stars range  [__] [__]                  all any
         search tags  [________________________] (o) (o)
         search text  [________________________] (o) (o)
         search files [________________________] (o) (o)
         other criteria: [geotags] [other]

         defined tags
          ----------------------------------------------
         |                                              |
         |                                              |
         |                                              |
         |                                              |
         |                                              |
          ----------------------------------------------
                                     [proceed] [cancel]
***/

   zd = zdialog_new(ZTX("Search Image Metadata"),mWin,Bproceed,Bcancel,null);
   zdsearchimages = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"vbox","vb1","hb1",0,"homog|space=3");
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"homog|space=3|expand");

   zdialog_add_widget(zd,"label","labD","vb1",ZTX("date range"));
   zdialog_add_widget(zd,"label","labS","vb1",ZTX("stars range"));
   zdialog_add_widget(zd,"label","labT","vb1",ZTX("search tags"));
   zdialog_add_widget(zd,"label","labT","vb1",ZTX("search text"));
   zdialog_add_widget(zd,"label","labF","vb1",ZTX("search files"));

   zdialog_add_widget(zd,"hbox","hbD","vb2",0,"space=1");
   zdialog_add_widget(zd,"entry","datefrom","hbD",0,"scc=10");
   zdialog_add_widget(zd,"entry","dateto","hbD",0,"scc=10");
   zdialog_add_widget(zd,"label","labD","hbD",ZTX("(yyyymmdd)"),"space=5");

   zdialog_add_widget(zd,"hbox","hbS","vb2",0,"space=1");
   zdialog_add_widget(zd,"entry","starsfrom","hbS",0,"scc=2");
   zdialog_add_widget(zd,"entry","starsto","hbS",0,"scc=2");
   zdialog_add_widget(zd,"label","space","hbS",0,"expand");
   zdialog_add_widget(zd,"label","all-any","hbS",ZTX("all/any"));

   zdialog_add_widget(zd,"hbox","hbT","vb2",0,"space=1");
   zdialog_add_widget(zd,"frame","frameT","hbT",0,"expand");
   zdialog_add_widget(zd,"text","searchtags","frameT",0,"expand|wrap");
   zdialog_add_widget(zd,"radio","alltags","hbT",0);
   zdialog_add_widget(zd,"radio","anytags","hbT",0);

   zdialog_add_widget(zd,"hbox","hbC","vb2",0,"space=1|expand");
   zdialog_add_widget(zd,"entry","searchtext","hbC",0,"expand");
   zdialog_add_widget(zd,"radio","alltext","hbC",0);
   zdialog_add_widget(zd,"radio","anytext","hbC",0);

   zdialog_add_widget(zd,"hbox","hbF","vb2",0,"space=1|expand");
   zdialog_add_widget(zd,"entry","searchfiles","hbF",0,"expand");
   zdialog_add_widget(zd,"radio","allfiles","hbF",0);
   zdialog_add_widget(zd,"radio","anyfiles","hbF",0);
   
   zdialog_add_widget(zd,"hbox","hbO","dialog","space=3");
   zdialog_add_widget(zd,"label","labO1","hbO",ZTX("other criteria"),"space=3");
   zdialog_add_widget(zd,"button","geotags","hbO","geotags","space=8");
   zdialog_add_widget(zd,"label","geotags#","hbO","( )");
   zdialog_add_widget(zd,"label","space","hbO",0,"space=5");
   zdialog_add_widget(zd,"button","other","hbO",ZTX("other"),"space=8");
   zdialog_add_widget(zd,"label","other#","hbO","( )");
   
   zdialog_add_widget(zd,"hbox","hbA1","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labdeftags","hbA1",ZTX("defined tags"),"space=5");
   zdialog_add_widget(zd,"hbox","hbA2","dialog",0,"expand");
   zdialog_add_widget(zd,"frame","frameA","hbA2",0,"space=5|expand");
   zdialog_add_widget(zd,"scrwin","scrwinA","frameA",0,"expand");
   zdialog_add_widget(zd,"text","deftags","scrwinA",0,"expand|wrap"); 

   widget = zdialog_widget(zd,"searchtags");                               //  tag widget mouse functions      v.12.06
   textwidget_set_clickfunc(widget,search_searchtags_clickfunc);

   widget = zdialog_widget(zd,"deftags");
   textwidget_set_clickfunc(widget,search_deftags_clickfunc);

   zdialog_restore_inputs(zd);                                             //  preload prior user inputs    v.12.09

   load_deftags();                                                         //  stuff defined tags into dialog
   deftags_stuff(zd);

   zdialog_resize(zd,500,550);                                             //  start dialog
   zdialog_help(zd,"search_images");                                       //  zdialog help topic 
   zdialog_run(zd,searchimages_dialog_event,"save");

   zdialog_wait(zd);                                                       //  wait for dialog completion   
   return;
}

//  mouse click functions for search tags and defined tags widgets         //  v.12.06

void search_searchtags_clickfunc(GtkWidget *widget, int &line, int &pos)   //  search tag clicked
{
   char     *txline, *txtag, end = 0;
   cchar    *delims = tagdelims":";
   
   txline = textwidget_get_line(widget,line);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,delims,end);
   if (! txtag) { zfree(txline); return; }
   
   del_tag(txtag,tags_searchtags);                                         //  remove from search list
   zdialog_stuff(zdsearchimages,"searchtags",tags_searchtags);
   
   zfree(txline);
   zfree(txtag);
   return;
}


void search_deftags_clickfunc(GtkWidget *widget, int &line, int &pos)      //  defined tag clicked
{
   char     *txline, *txtag, end = 0;
   cchar    *delims = tagdelims":";
   
   txline = textwidget_get_line(widget,line);
   if (! txline) return;

   txtag = textwidget_get_word(txline,pos,delims,end);
   if (! txtag) { zfree(txline); return; }
   
   add_tag(txtag,tags_searchtags,tagScc);                                  //  add to search tag list
   zdialog_stuff(zdsearchimages,"searchtags",tags_searchtags);
   
   zfree(txline);
   zfree(txtag);
   return;
}


//  search images dialog event and completion callback function

int searchimages_dialog_event(zdialog *zd, cchar *event)                   //  overhauled    v.12.05
{
   using namespace navi;

   int  searchimages_geotags_dialog(zdialog *zd);
   int  searchimages_metadata_dialog(zdialog *zd);
   int  searchimages_geotags_select(char *gtagsrec);
   int  searchimages_metadata_select(char *imagefile);
   int  searchimages_metadata_report();

   char     resultsfile[200];
   char     date1[12], date2[12];
   cchar    *pps, *ppf;
   int      ftf, iis, iif, err;
   int      nmatch, nfail, nfound;
   int      Fdates, Ftext, Ffiles, Ftags, Fstars, Fgtags, Fmeta;
   int      Falltags, Falltext, Fallfiles;
   sxrec_t  sxrec;
   FILE     *fid;

   if (strEqu(event,"geotags"))                                            //  get geotags search criteria   v.12.05
      searchimages_geotags_dialog(zd);

   if (strEqu(event,"other"))                                              //  get other metadata criteria   v.12.05
      searchimages_metadata_dialog(zd);

   Fgtags = 0;
   if (*searchCity || *searchCountry ||                                    //  search geotags was given
       *searchLatitude || *searchLongitude) Fgtags = 1;
   
   Fmeta = 0;
   if (Nsearchkeys) Fmeta = 1;                                             //  search other metadata was given
   
   if (Fgtags) zdialog_stuff(zd,"geotags#","(*)");                         //  add (*) visual flags to main dialog if
   else zdialog_stuff(zd,"geotags#","( )");                                //    geotags or metadata selection active
   if (Fmeta) zdialog_stuff(zd,"other#","(*)");
   else zdialog_stuff(zd,"other#","( )");
   
   if (! zd->zstat) return 1;                                              //  wait for dialog completion

   if (zd->zstat != 1) {                                                   //  cancel
      zdialog_free(zd);
      return 1;
   }

   zdialog_fetch(zd,"datefrom",searchDateFrom,10);                         //  get search date range
   zdialog_fetch(zd,"dateto",searchDateTo,10);
   zdialog_fetch(zd,"starsfrom",searchStarsFrom,2);                        //  get search stars range
   zdialog_fetch(zd,"starsto",searchStarsTo,2);
   zdialog_fetch(zd,"searchtags",tags_searchtags,tagScc);                  //  get search tags
   zdialog_fetch(zd,"searchtext",tags_searchtext,tagScc);                  //  get search text*
   zdialog_fetch(zd,"searchfiles",meta_searchfiles,tagScc);                //  get search /path*/file*

   zdialog_fetch(zd,"alltags",Falltags);                                   //  get match all/any options
   zdialog_fetch(zd,"alltext",Falltext);
   zdialog_fetch(zd,"allfiles",Fallfiles);

   zdialog_free(zd);                                                       //  cancel dialog

   strcpy(date1,"0000:01:01");                                             //  defaults for missing dates
   strcpy(date2,"9999:12:31"); 

   Fdates = 0;
   if (*searchDateFrom) {                                                  //  date from was given
      Fdates++;
      strncpy(date1,searchDateFrom,4);                                     //  convert format
      if (searchDateFrom[4] >= '0')                                        //    yyyymmdd >> yyyy:mm:dd
         strncpy(date1+5,searchDateFrom+4,2);
      if (searchDateFrom[6] >= '0')
         strncpy(date1+8,searchDateFrom+6,2);
   }
   if (*searchDateTo) {                                                    //  date to was given
      Fdates++;
      strncpy(date2,searchDateTo,4);
      if (searchDateTo[4] >= '0')
         strncpy(date2+5,searchDateTo+4,2);
      if (searchDateTo[6] >= '0')
         strncpy(date2+8,searchDateTo+6,2);
   }

   Fstars = 0;
   if (*searchStarsFrom || *searchStarsTo) Fstars = 1;                     //  stars was given
   
   Ffiles = 0;
   if (! blank_null(meta_searchfiles)) Ffiles = 1;                         //  search path / file (fragment) was given

   Ftext = 0;
   if (! blank_null(tags_searchtext)) Ftext = 1;                           //  search text was given

   Ftags = 0;
   if (! blank_null(tags_searchtags)) Ftags = 1;                           //  search tags was given
   
   if (Ffiles) strToLower(meta_searchfiles);                               //  all comparisons in lower case
   if (Ftags) strToLower(tags_searchtags);
   if (Ftext) strToLower(tags_searchtext);

   snprintf(resultsfile,199,"%s/search_results",get_zuserdir());           //  open file to save search results
   fid = fopen(resultsfile,"w");
   if (! fid) goto writerror;

   Ffuncbusy = 1;
   ftf = 1;
   nfound = 0;
   
   while (true)                                                            //  read search index recs. and test
   {                                                                       //    against select criteria
      err = read_search_index_seq(sxrec,ftf);                              //             v.12.10
      if (err) break;
      
      zmainloop(20);                                                       //  v.12.12

      if (Ffiles)                                                          //  file name match is wanted
      {
         nmatch = nfail = 0;

         for (iis = 1; ; iis++)
         {
            pps = strField(meta_searchfiles,' ',iis);                      //  step thru search file names
            if (! pps) break;
            if (strcasestr(sxrec.file,pps)) nmatch++;
            else nfail++;
         }

         if (nmatch == 0) goto nomatch;                                    //  no match to any file
         if (Fallfiles && nfail) goto nomatch;                             //  no match to all files 
      }

      if (Fdates)                                                          //  date match is wanted
      {
         if (strcmp(sxrec.pdate,date1) < 0) goto nomatch;
         if (strcmp(sxrec.pdate,date2) > 0) goto nomatch;
      }

      if (Ftags)                                                           //  tags match is wanted
      {
         strToLower(sxrec.tags);
         nmatch = nfail = 0;

         for (iis = 1; ; iis++)                                            //  step thru search tags
         {
            pps = strField(tags_searchtags,tagdelims" ",iis);              //  (delimited, wildcards)    v.12.07
            if (! pps) break;
            if (*pps == ' ') continue;

            for (iif = 1; ; iif++)                                         //  step thru file tags (delimited)
            {
               ppf = strField(sxrec.tags,tagdelims,iif);
               if (! ppf) { nfail++; break; }                              //  count matches and fails
               if (*ppf == ' ') continue;
               if (MatchWild(pps,ppf) == 0) { nmatch++; break; }           //  wildcard match 
            }
         }

         if (nmatch == 0) goto nomatch;                                    //  no match to any tag
         if (Falltags && nfail) goto nomatch;                              //  no match to all tags
      }

      if (Fstars)                                                          //  rating (stars) match is wanted
      {
         if (*searchStarsFrom && sxrec.rating[0] < *searchStarsFrom) 
            goto nomatch;
         if (*searchStarsTo && sxrec.rating[0] > *searchStarsTo) 
            goto nomatch;
      }

      if (Ftext)                                                           //  text match is wanted
      {
         nmatch = nfail = 0;

         for (iis = 1; ; iis++)                                            //  step through search words
         {
            pps = strField(tags_searchtext,' ',iis);
            if (! pps) break;
            if (*pps == ' ') continue;
            if (strcasestr(sxrec.capt,pps)) nmatch++;                      //  search captions for word
            else if (strcasestr(sxrec.comms,pps)) nmatch++;                //  search comments for word
            else nfail++;
         }
                     
         if (nmatch == 0) goto nomatch;                                    //  no match to any word
         if (Falltext && nfail) goto nomatch;                              //  no match to all words
      }

      if (Fgtags) {                                                        //  geotags match is wanted   v.12.05
         if (! searchimages_geotags_select(sxrec.gtags)) 
            goto nomatch;
      }
      
      if (Fmeta)                                                           //  other metadata match      v.12.05
         if (! searchimages_metadata_select(sxrec.file)) 
            goto nomatch;
      
      nfound++;                                                            //  count matches
      fprintf(fid,"%s\n",sxrec.file);                                      //  save matching filename

   nomatch:                                                                //  this file rejected
      zfree(sxrec.file);
      zfree(sxrec.tags);
      zfree(sxrec.capt);
      zfree(sxrec.comms);
      zfree(sxrec.gtags);
   }

   fclose(fid);

   Ffuncbusy = 0;

   if (! nfound) {
      zmessageACK(mWin,ZTX("No matching images found"));
      return 0;
   }

   navi::gallerytype = 2;                                                  //  search results
   gallery(resultsfile,"initF");                                           //  generate gallery of matching files

   if (Fmeta)                                                              //  metadata report format    v.12.05
      searchimages_metadata_report();
   
   gallery(0,"paint");
   return 0;
   
writerror:
   zmessLogACK(mWin,"output file error: %s",strerror(errno));
   return 0;
}


/**************************************************************************/

//  dialog to get geotags search criteria

int searchimages_geotags_dialog(zdialog *zdp)                              //  v.12.05
{
   int searchimages_geotags_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;
   int         zstat;

/***
            Add Geotags Search Criteria

         city [__________]  country [__________]
         latitude [_______]  longitude [_______] 
         [find]  [map]   range (km) [___]
         
                      [clear] [apply] [cancel]
***/

   zd = zdialog_new(ZTX("Add Geotags Search Criteria"),mWin,Bclear,Bapply,Bcancel,null);

   zdialog_add_widget(zd,"hbox","hbg1","dialog",0);
   zdialog_add_widget(zd,"label","labcity","hbg1",ZTX("city"),"space=3");
   zdialog_add_widget(zd,"entry","city","hbg1",0,"expand");
   zdialog_add_widget(zd,"label","space","hbg1",0,"space=5");
   zdialog_add_widget(zd,"label","labcountry","hbg1",ZTX("country"),"space=3");
   zdialog_add_widget(zd,"entry","country","hbg1",0,"expand");

   zdialog_add_widget(zd,"hbox","hbg2","dialog",0);
   zdialog_add_widget(zd,"label","lablat","hbg2","latitude","space=3");
   zdialog_add_widget(zd,"entry","latitude","hbg2",0,"scc=10");
   zdialog_add_widget(zd,"label","space","hbg2",0,"space=5");
   zdialog_add_widget(zd,"label","lablong","hbg2","longitude","space=3");
   zdialog_add_widget(zd,"entry","longitude","hbg2",0,"scc=10");

   zdialog_add_widget(zd,"hbox","hbg3","dialog",0);
   zdialog_add_widget(zd,"button","find","hbg3",Bfind,"space=10");
   zdialog_add_widget(zd,"button","map","hbg3",Bmap,"space=10");
   zdialog_add_widget(zd,"label","labkm","hbg3",ZTX("range (km)"),"space=10");
   zdialog_add_widget(zd,"entry","range","hbg3","100","scc=5");

   init_geotags();                                                         //  v.12.07
   
   zdialog_show(zdp,0);                                                    //  hide parent dialog           v.12.07
   zdialog_restore_inputs(zd);                                             //  preload prior user inputs    v.12.09
   zdialog_run(zd,searchimages_geotags_dialog_event);                      //  run dialog
   zstat = zdialog_wait(zd);                                               //  wait for completion
   zdialog_free(zd);
   zd_geotags = 0;                                                         //  deactivate world map clicks
   zdialog_show(zdp,1);                                                    //  restore parent dialog
   
   if (zstat == 1) return 1;                                               //  search criteria were entered
   else return 0;                                                          //  not
}


//  dialog event and completion function

int searchimages_geotags_dialog_event(zdialog *zd, cchar *event)           //  v.12.05
{
   int         fgtags, err, nn;
   char        city[100], country[100];
   char        *location[2], *coord[2];

   if (strEqu(event,"find")) {                                             //  get geotags data from 
      zdialog_fetch(zd,"city",city,99);                                    //    (partial) city [country] names
      zdialog_fetch(zd,"country",country,99);
      location[0] = city;
      location[1] = country;
      nn = geotags_choosecity(location,coord);                             //  find matching city geotags data
      if (nn == 1) {                                                       //    and ask user to choose
         zdialog_stuff(zd,"city",location[0]);
         zdialog_stuff(zd,"country",location[1]);
         zdialog_stuff(zd,"latitude",coord[0]);
         zdialog_stuff(zd,"longitude",coord[1]);
      }
   }

   if (strEqu(event,"map"))                                                //  get geotags data from world map
   {
      zd_geotags = zd;
      click_worldmap();
      return 0;
   }
   
   if (! zd->zstat) return 0;                                              //  wait for completion
   
   if (zd->zstat == 1) {
      zdialog_stuff(zd,"city","");                                         //  clear      v.12.12
      zdialog_stuff(zd,"country","");
      zdialog_stuff(zd,"latitude","");
      zdialog_stuff(zd,"longitude","");
      zd->zstat = 0;                                                       //  keep dialog active
      return 0;
   }

   if (zd->zstat != 2) {                                                   //  cancel
      zdialog_destroy(zd);                                                 //  kill dialog
      *searchCity = *searchCountry = 0;                                    //  clear criteria
      *searchLatitude = *searchLongitude = 0;
      zd_geotags = 0;
      return 0;
   }

   zdialog_fetch(zd,"city",searchCity,99);                                 //  apply - get search geotags
   zdialog_fetch(zd,"country",searchCountry,99);
   zdialog_fetch(zd,"latitude",searchLatitude,20);
   zdialog_fetch(zd,"longitude",searchLongitude,20);
   zdialog_fetch(zd,"range",searchRange);
   
   strToLower(searchCity);
   strToLower(searchCountry);

   err = 0;
   if (*searchLatitude || *searchLongitude) {                              //  sanity check for latitude/longitude
      if (! *searchLatitude || ! *searchLongitude) err++;
      flatitude = atof(searchLatitude);
      flongitude = atof(searchLongitude);
      if (flatitude < -90 || flatitude > 90) err++;
      if (flongitude < -180 || flongitude > 180) err++;
      if (searchRange < 1 || searchRange > 1000) err++;
      if (strEqu(searchLatitude,"null")) err = 0;                          //  search for null lat/long     v.12.06
   }

   if (err) {
      zmessageACK(mWin,ZTX("error in latitude/longitude/range"));
      zd->zstat = 0;
      return 1;
   }

   fgtags = 0;
   if (! blank_null(searchCity)) fgtags = 1;                               //  look for inputs
   if (! blank_null(searchCountry)) fgtags = 1;
   if (! blank_null(searchLatitude)) fgtags = 1;
   if (! blank_null(searchLongitude)) fgtags = 1;

   if (fgtags) zd->zstat = 1;                                              //  search criteria was given
   else zd->zstat = 2;                                                     //  no inputs were made

   zdialog_destroy(zd);                                                    //  kill dialog
   return 1;
}


//  test the image geotags data against the search criteria   

int searchimages_geotags_select(char *gtags)                               //  v.12.05
{
   float       glatitude, glongitude, distance;
   cchar       *pps;

   if (*searchLatitude) {                                                  //  if latitude/longitude present,
      pps = strField(gtags,'^',3);                                         //    ignore city/country data
      if (! pps) return 0;
      if (strEqu(pps,"null"))                                              //  no latitude/longitude data
         if (flatitude == 0 && flongitude == 0) return 1;                  //  "no data" was wanted
      glatitude = atof(pps);
      pps = strField(gtags,'^',4);
      if (! pps) return 0;
      if (strEqu(pps,"null")) return 0;
      glongitude = atof(pps);
      distance = (flatitude - glatitude) * (flatitude - glatitude);
      distance += (flongitude - glongitude) * (flongitude - glongitude);
      distance = 111.0 * sqrt(distance);
      if (distance <= searchRange) return 1;                               //  within distance limit, select
      return 0;                                                            //  outside, exclude
   }

   if (*searchCity) {                                                      //  check city and country criteria
      pps = strField(gtags,'^',1);
      if (! pps) return 0;
      if (! strcasestr(pps,searchCity)) return 0;
   }
   
   if (*searchCountry) {
      pps = strField(gtags,'^',2);
      if (! pps) return 0;
      if (! strcasestr(pps,searchCountry)) return 0;
   }
   
   return 1;
}


/**************************************************************************/

//  dialog to get metadata search criteria

int searchimages_metadata_dialog(zdialog *zdp)                             //  v.12.05
{
   int searchimages_metadata_dialog_event(zdialog *zd, cchar *event);

   cchar *metamess = ZTX("These items are always reported: \n"
                         "date, stars, tags, caption, comment");
   zdialog  *zd;
   int      zstat;

/***
         Search and Report Metadata

         These items are always reported: 
         date, stars, tags, caption, comment

            Additional Items for Report
            Keyword      Match Criteria
         [__________] [__________________]
         [__________] [__________________]
         [__________] [__________________]
         [__________] [__________________]
         [__________] [__________________]

                   [clear] [apply] [cancel]
***/

   zd = zdialog_new("Search and Report Metadata",mWin,Bclear,Bapply,Bcancel,null);
   zdialog_add_widget(zd,"label","labmeta","dialog",metamess,"space=5");
   zdialog_add_widget(zd,"label","labopts","dialog",ZTX("Additional Items for Report"));

   zdialog_add_widget(zd,"hbox","hb1","dialog");
   zdialog_add_widget(zd,"vbox","vb1","hb1",0,"space=5");
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"space=5|expand");
   
   zdialog_add_widget(zd,"label","lab1","vb1",ZTX("Keyword"));
   zdialog_add_widget(zd,"entry","key0","vb1");
   zdialog_add_widget(zd,"entry","key1","vb1");
   zdialog_add_widget(zd,"entry","key2","vb1");
   zdialog_add_widget(zd,"entry","key3","vb1");
   zdialog_add_widget(zd,"entry","key4","vb1");

   zdialog_add_widget(zd,"label","lab2","vb2",ZTX("Match Criteria"));
   zdialog_add_widget(zd,"entry","match0","vb2",0,"expand");
   zdialog_add_widget(zd,"entry","match1","vb2",0,"expand");
   zdialog_add_widget(zd,"entry","match2","vb2",0,"expand");
   zdialog_add_widget(zd,"entry","match3","vb2",0,"expand");
   zdialog_add_widget(zd,"entry","match4","vb2",0,"expand");
   
   strcpy(searchkeyx,"keyx");
   strcpy(searchkeydatax,"matchx");
   
   if (! searchkeys[0])                                                    //  first call initialization
   {
      for (int ii = 0; ii < 5; ii++) {
         searchkeys[ii] = zmalloc(40,"search_images");
         searchkeydata[ii] = zmalloc(100,"search_images");
         *searchkeys[ii] = *searchkeydata[ii] = 0;
      }
   }   

   zdialog_show(zdp,0);                                                    //  hide parent dialog        v.12.07
   zdialog_resize(zd,400,300);
   zdialog_restore_inputs(zd);                                             //  preload prior user inputs    v.12.09
   zdialog_run(zd,searchimages_metadata_dialog_event);                     //  run dialog
   zstat = zdialog_wait(zd);                                               //  wait for completion
   zdialog_free(zd);
   zdialog_show(zdp,1);                                                    //  restore parent dialog
   
   if (zstat == 1) return 1;                                               //  search criteria were entered
   else return 0;                                                          //  not
}


//  dialog event and completion callback function

int searchimages_metadata_dialog_event(zdialog *zd, cchar *event)          //  v.12.05
{
   int      ii, jj;
   char     keyx[8] = "keyx", matchx[8] = "matchx";

   if (! zd->zstat) return 1;                                              //  wait for completion

   if (zd->zstat == 1) {
      for (ii = 0; ii < 5; ii++) {                                         //  clear      v.12.12
         keyx[3] = '0' + ii;
         matchx[5] = '0' + ii;
         zdialog_stuff(zd,keyx,"");
         zdialog_stuff(zd,matchx,"");
         zd->zstat = 0;                                                    //  keep dialog active
      }
      return 0;
   }

   if (zd->zstat != 2) {
      zdialog_destroy(zd);                                                 //  cancel
      Nsearchkeys = 0;                                                     //  no search keys
      return 1;
   }
   
   Nsearchkeys = 0;                                                        //  apply

   for (ii = jj = 0; ii < 5; ii++)                                         //  get metadata keys
   {
      searchkeyx[3] = '0' + ii;
      zdialog_fetch(zd,searchkeyx,searchkeys[ii],40);
      strCompress(searchkeys[ii]);                                         //  remove all blanks from key names
      if (*searchkeys[ii] <= ' ') continue;
      memmove(searchkeys[jj],searchkeys[ii],40);                           //  repack blank keys
      searchkeydatax[5] = '0' + ii;
      zdialog_fetch(zd,searchkeydatax,searchkeydata[ii],100);              //  get corresp. match value if any
      strTrim2(searchkeydata[jj],searchkeydata[ii]);                       //  trim leading and trailing blanks
      if (ii > jj) *searchkeys[ii] = *searchkeydata[ii] = 0;
      jj++;
   }

   Nsearchkeys = jj;                                                       //  keys found, no blanks
   
   if (Nsearchkeys) zd->zstat = 1;                                         //  keys were entered
   else zd->zstat = 2;                                                     //  no keys were entered

   zdialog_destroy(zd);                                                    //  kill dialog
   return 1;
}


//  test image metadata against select criteria

int searchimages_metadata_select(char *file)                               //  v.12.05
{
   char           **kvals;
   int            ii, nth;
   cchar          *pp;

   kvals = exif_get(file,(cchar **) searchkeys,Nsearchkeys);               //  get the metadata

   for (ii = 0; ii < Nsearchkeys; ii++) {                                  //  check key values for match
      if (*searchkeydata[ii] > ' ') {                                      //  match value(s) are present
         if (! kvals[ii]) break;                                           //  key not present in metadata
         for (nth = 1; ; nth++) {
            pp = strField(searchkeydata[ii],' ',nth);                      //  get each match value
            if (! pp) break;                                               //  no more, no match found
            if (strcasestr(kvals[ii],pp)) break;                           //  found match
         }
         if (! pp) break;                                                  //  no match found
      }
   }

   if (ii < Nsearchkeys) return 0;                                         //  at least one match fail
   return 1;                                                               //  all matches passed
}


//  Report the selected metadata using a gallery window layout
//  with image thumbnails and selected metadata text.

int searchimages_metadata_report()                                         //  v.12.05
{
   using namespace navi;

   cchar          *keys1[6] = { exif_date_key, iptc_rating_key, exif_city_key, 
                                exif_country_key, iptc_caption_key, exif_comment_key };
   cchar          *keys[20];
   char           *file, **kvals;
   char           text1[2*indexrecl], text2[200];
   int            Nkeys, ii, jj, cc;
   
   if (! mdlist) {
      cc = nfiles * sizeof(char *);                                        //  allocate metadata list
      mdlist = (char **) zmalloc(cc,"search_images");                      //  nfiles = curr. gallery files
      memset(mdlist,0,cc);                                                 //  mdlist = corresp. metadata list
   }                                                                       //  (zfree() in image_navigate)
   
   for (ii = 0; ii < 6; ii++)                                              //  set first 6 key names, fixed
      keys[ii] = keys1[ii];
   
   for (ii = 0; ii < Nsearchkeys; ii++)                                    //  remaining key names from user
      keys[ii+6] = searchkeys[ii];

   Nkeys = 6 + Nsearchkeys;                                                //  total keys to extract

   for (ii = 0; ii < nfiles; ii++)                                         //  loop image gallery files
   {
      file = gallery(0,"find",ii);
      if (! file) continue;

      kvals = exif_get(file,(cchar **) keys,Nkeys);                        //  get the metadata
      
      snprintf(text2,200,"%s  %s",kvals[2],kvals[3]);                      //  combine city and country
      if (kvals[2]) zfree(kvals[2]);
      if (kvals[3]) zfree(kvals[3]);
      kvals[2] = strdupz(text2,0,"search_images");
      kvals[3] = 0;

      for (cc = jj = 0; jj < Nkeys; jj++)                                  //  add metadata to report
      {
         if (jj == 3) continue;                                            //  skip country
         snprintf(text2,200,"key: %s  value: %s \n",keys[jj], kvals[jj]);
         strcpy(text1+cc,text2);
         cc += strlen(text2);
      }

      if (mdlist[ii]) zfree(mdlist[ii]);                                   //  attach metadata for gallery paint
      mdlist[ii] = strdupz(text1,0,"search_images");
      mdrows = Nkeys - 1;                                                  //  report rows            v.13.01

      for (jj = 0; jj < Nkeys; jj++)                                       //  free memory
         if (kvals[jj]) zfree(kvals[jj]);

      zfree(file);
   }

   gallerytype = 3;                                                        //  gallery type = search results/metadata

   return 0;
}


/**************************************************************************
   Functions to read and write exif/iptc or other metadata
***************************************************************************/

//  get EXIF/IPTC metadata for given image file and EXIF/IPTC key(s)
//  returns array of pointers to corresponding key values
//  if a key is missing, corresponding pointer is null
//  returned strings belong to caller, are subject for zfree()
//  up to 20 keynames may be requested per call
//  use -fast not -fast2 which can lose geotag data                        v.12.06

char ** exif_get(cchar *file, cchar **keys, int nkeys)                     //  revised       v.12.06
{
   char           *pp, **outputs;
   char           *inputs[30];
   static char    *keyvals[20];
   int            cc, ii, jj;

   if (nkeys < 1 || nkeys > 20) zappcrash("exif_get nkeys: %d",nkeys);

   inputs[0] = (char *) "-m";                                              //  options for exiftool
   inputs[1] = (char *) "-s2";
   inputs[2] = (char *) "-n";
   inputs[3] = (char *) "-fast";                                           //  no -fast2: can miss geotag data 
   jj = 4;

   for (ii = 0; ii < nkeys; ii++)                                          //  build exiftool inputs
   {
      cc = strlen(keys[ii]);                                               //  -keyname
      inputs[jj] = zmalloc(cc+2,"exif_get");
      inputs[jj][0] = '-';
      strcpy(inputs[jj]+1,keys[ii]);
      jj++;
   }

   inputs[jj] = strdupz(file,0,"exif_get");                                //  filename last
   jj++;

   inputs[jj] = 0;                                                         //  EOL

   cc = 20 * sizeof(char *);
   memset(keyvals,0,cc);                                                   //  set all outputs to null

   outputs = exiftool_server(inputs);                                      //  get exif outputs
   
   for (ii = 4; ii < jj; ii++)                                             //  free memory
      zfree(inputs[ii]);

   for (ii = 0; outputs[ii]; ii++)                                         //  search outputs
   {
      pp = outputs[ii];                                                    //  keyname: keyvalue

      for (jj = 0; jj < nkeys; jj++)
      {
         uint cc = strlen(keys[jj]);                                       //  look for matching input keyname
         if (strncasecmp(pp,keys[jj],cc) == 0)
            if (strlen(pp) > cc+2)                                         //  if not empty, 
               keyvals[jj] = strdupz(pp+cc+2,0,"exif_get");                //    return keyvalue alone
      }
   }
   
   return keyvals;
}


/**************************************************************************/

//  create or change EXIF/IPTC metadata for given image file and key(s)
//  up to 20 keys may be processed
//  command: 
//    exiftool -m -P -overwrite_original -keyname="keyvalue" ... "file"
//
//  NOTE: exiftool replaces \n (newline) in keyvalue with . (period).

int exif_put(cchar *file, cchar **keys, cchar **text, int nkeys)           //  revised          v.12.06
{
   int      ii, jj, cc;
   char     *inputs[30];
   
   if (nkeys < 1 || nkeys > 20) zappcrash("exif_put nkeys: %d",nkeys);

   inputs[0] = (char *) "-m";                                              //  exiftool options
   inputs[1] = (char *) "-P";                                              //  -P      preserve date  v.12.09
   inputs[2] = (char *) "-overwrite_original";
   jj = 3;

   for (ii = 0; ii < nkeys; ii++)                                          //  build exiftool inputs
   {
      cc = strlen(keys[ii]) + strlen(text[ii]) + 3;
      inputs[jj] = zmalloc(cc,"exif_put");
      inputs[jj][0] = '-';                                                 //  -keyname=value
      strcpy(inputs[jj]+1,keys[ii]);
      cc = strlen(keys[ii]);
      inputs[jj][cc+1] = '=';
      strcpy(inputs[jj]+cc+2,text[ii]);
      jj++;
      
      if (strcasecmp(keys[ii],"GPSLatitude") == 0) {                       //  take care of latitude N/S 
         if (*text[ii] == '-')
            inputs[jj] = strdupz("-GPSLatitudeRef=S",0,"exif_put");
         else 
            inputs[jj] = strdupz("-GPSLatitudeRef=N",0,"exif_put");
         jj++;
      }

      if (strcasecmp(keys[ii],"GPSLongitude") == 0) {                      //  and longitude E/W
         if (*text[ii] == '-')
            inputs[jj] = strdupz("-GPSLongitudeRef=W",0,"exif_put");
         else 
            inputs[jj] = strdupz("-GPSLongitudeRef=E",0,"exif_put");
         jj++;
      }
   }

   inputs[jj] = strdupz(file,0,"exif_put");                                //  last input is filename
   jj++;
   
   inputs[jj] = 0;                                                         //  EOL

   exiftool_server(inputs);
   
   for (ii = 3; ii < jj-1; ii++)                                           //  free memory
       zfree(inputs[ii]);

   return 0;
}


/**************************************************************************/

//  copy EXIF/IPTC data from original image file to new (edited) image file
//  if nkeys > 0, up to 20 keys may be replaced with new values
//  exiftool command revised: parameter sequence and -icc_profile

int exif_copy(cchar *file1, cchar *file2, cchar **keys, cchar **text, int nkeys)
{
   char     *inputs[30];                                                   //  revised       v.12.06
   int      cc, ii, jj;
   
   if (nkeys > 20) zappcrash("exif_copy() nkeys %d",nkeys);                //  v.12.10
   
   inputs[0] = (char *) "-m";                                              //  -m
   inputs[1] = (char *) "-P";                                              //  -P            preserve date  v.12.09
   inputs[2] = (char *) "-tagsfromfile";                                   //  -tagsfromfile
   inputs[3] = strdupz(file1,0,"exif_copy");                               //  file1
   inputs[4] = (char *) "-all";                                            //  -all
   inputs[5] = (char *) "-icc_profile";                                    //  -icc_profile
   
   jj = 6;

   for (int ii = 0; ii < nkeys; ii++)                                      //  -keyname=keyvalue
   {
      cc = strlen(keys[ii]) + strlen(text[ii]) + 3;
      inputs[jj] = zmalloc(cc,"exif_copy");
      strncatv(inputs[jj],cc,"-",keys[ii],"=",text[ii],null);
      jj++;
   }
   
   inputs[jj++] = strdupz(file2,0,"exif_copy");                            //  file2
   inputs[jj++] = (char *) "-overwrite_original";                          //  -overwrite_original
   inputs[jj] = 0;                                                         //  EOL

   exiftool_server(inputs);

   zfree(inputs[3]);                                                       //  free memory
   for (ii = 6; ii < jj-1; ii++)
   zfree(inputs[ii]);

   return 0;
}


/**************************************************************************/

//  convert between EXIF and fotoxx tag date formats
//  EXIF date: yyyy:mm:dd hh:mm:ss[.ss]
//  tag date: yyyymmdd
//  

void exif_tagdate(cchar *exifdate, char *tagdate)
{
   time_t      tnow;
   struct tm   *snow;
   
   if (! exifdate || strlen(exifdate) < 10) {                              //  bad EXIF date, use current date
      tnow = time(0);
      snow = localtime(&tnow);
      snprintf(tagdate,8,"%4d%02d%02d",
               snow->tm_year+1900, snow->tm_mon+1, snow->tm_mday);
      tagdate[8] = 0;
      return;
   }

   strncpy(tagdate,exifdate,4);                                            //  convert 
   strncpy(tagdate+4,exifdate+5,2);
   strncpy(tagdate+6,exifdate+8,2);
   tagdate[8] = 0;
   return;
}

void tag_exifdate(cchar *tagdate, char *exifdate)
{
   int         cc;
   time_t      tnow;
   struct tm   *snow;
   
   if (! tagdate || strlen(tagdate) < 4) {
      tnow = time(0);
      snow = localtime(&tnow);
      snprintf(exifdate,20,"%4d:%02d:%02d %02d:%02d:%02d",
               snow->tm_year+1900, snow->tm_mon+1, snow->tm_mday,
               snow->tm_hour, snow->tm_min, snow->tm_sec);
      exifdate[19] = 0;
      return;
   }

   strncpy(exifdate,tagdate,4);
   strcpy(exifdate+4,":01:01 00:00:00");
   cc = strlen(tagdate);
   if (cc >= 6) strncpy(exifdate+5,tagdate+4,2);
   if (cc >= 8) strncpy(exifdate+8,tagdate+6,2);
   exifdate[19] = 0;
   return;
}


/**************************************************************************

   char ** exiftool_server(char **inputs)
   
   Server wrapper for exiftool for put/get exif/iptc data.
   This saves perl startup overhead for each call (0.1 >> 0.01 secs).

   Input records: -opt1
                  -opt2
                    ...
                  -keyword1=value1
                  -keyword2=value2
                    ... 
                  filename.jpg
                  (null pointer)

   Returned: list of pointers to resulting output records.
   The last output record is a null pointer.
   Returns null if there was a fatal error.
   Prior output records are freed automatically with each 
   new call, so the caller need not worry about this.

   First call: 
      Starts exiftool with pipe input and output files.
      Sends input record to exiftool and returns output records.
   Subsequent calls: 
      The existing exiftool process is re-used so that the 
      substantial startup overhead is avoided.
   To kill the exiftool process: exiftool_server(null).

***************************************************************************/

char ** exiftool_server(char **inputs)                                     //  revised API         v.12.06
{
   int            ii, cc, err;
   static int     fcf = 1;                                                 //  first call flag
   static FILE    *fid1 = 0, *fid2 = 0;                                    //  exiftool input, output files
   char           *pp, command[100];
   static char    *outrecs[100];                                           //  pointers for up to 99 output recs.
   char           outrec[exif_maxcc];                                      //  single output record
   
   if (! inputs)                                                           //  kill exiftool process
   {
      fid1 = fopen(exiftool_input,"a");                                    //  (also orphaned process)
      if (fid1) {
         fprintf(fid1,"-stay_open\nFalse\n");                              //  tell it to exit 
         fclose(fid1);                                                     //  bugfix: fflush after fclose  v.12.08.1
      }
      remove(exiftool_input);
      if (fid2) pclose(fid2);
      fid2 = 0;
      fcf = 1;                                                             //  start exiftool process if called again
      return 0;
   }

   if (fcf)                                                                //  first call only
   {
      fid1 = fopen(exiftool_input,"w");                                    //  start exiftool input file
      if (! fid1) {
         printf("*** exiftool_server, fopen error: %s \n",strerror(errno));
         return 0;
      }

      sprintf(command,"exiftool -stay_open True -@ %s",exiftool_input);    //  shell command to start exiftool

      fid2 = popen(command,"r");                                           //  start exiftool and output file
      if (! fid2) {
         printf("*** exiftool_server, popen error: %s \n",strerror(errno));
         fclose(fid1);
         return 0;
      }

      cc = 100 * sizeof(char *);                                           //  initz. no outputs
      memset(outrecs,0,cc);

      fcf = 0;                                                             //  keep last
   }

   for (ii = 0; ii < 99; ii++)                                             //  free memory from prior call
   {
      if (! outrecs[ii]) break;
      zfree(outrecs[ii]);
      outrecs[ii] = 0;
   }
   
   gallery_monitor("stop");                                                //  stop excess gallery inits    v.12.10
   
   for (ii = 0; inputs[ii]; ii++)
      fprintf(fid1,"%s\n",inputs[ii]);                                     //  write to exiftool, 1 per record

   err = fprintf(fid1,"-execute\n");                                       //  tell exiftool to process
   if (err < 0) {
      printf("*** exiftool_server error: %s \n",strerror(errno));          //  v.12.06
      return 0;
   }

   fflush(fid1);                                                           //  flush buffer

   for (ii = 0; ii < 99; ii++)                                             //  get exiftool outputs
   {
      pp = fgets_trim(outrec,exif_maxcc,fid2,1);
      if (! pp) break;
      if (strncmp(outrec,"{ready}",7) == 0) break;                         //  look for output end
      outrecs[ii] = strdupz(outrec,0,"exiftool_server");                   //  add to returned records
   }

   outrecs[ii] = 0;                                                        //  mark output end
   if (ii == 99) printf("*** exiftool_server: >99 output recs \n");

   gallery_monitor("start");                                               //  v.12.10

   return outrecs;                                                         //  return outputs to caller
}


/**************************************************************************
   Functions to read and write search index records
***************************************************************************/

//  Initialize for reading/writing to the search index.
//  Build a memory map of the first image file in each search index subfile.
//  Set refresh = 1 to initialize after an index file has been split 
//    or the first image file in a search index file has changed.
//  Returns 0 if OK, +N otherwise.
//
//    /.../.fotoxx/search_index/index_001
//    /.../.fotoxx/search_index/index_002
//    /.../.fotoxx/search_index/index_...         up to 999
//
//  Each of these files contains up to 'search_index_max' image filespecs.

char     **search_index_map = 0;                                           //  1st image file in each index file

int init_search_index_map(int refresh)                                     //  v.12.10
{
   int      ii, cc;
   char     *pp;
   char     indexfile[200];
   char     buff[indexrecl];
   FILE     *fid;

   if (! FsyncEnabled) return 0;                                           //  search index not being used
   if (search_index_map && ! refresh) return 0;                            //  already initialized

   if (search_index_map) {                                                 //  free memory
      for (ii = 0; search_index_map[ii]; ii++)
         zfree(search_index_map[ii]);
      zfree(search_index_map);
   }

   cc = 1001 * sizeof(char *);                                             //  allocate memory
   search_index_map = (char **) zmalloc(cc,"search-index");
   memset(search_index_map,0,cc);
   
   for (ii = 0; ii < 999; ii++)                                            //  read search index files 001 to 999
   {
      snprintf(indexfile,200,"%s/index_%03d",search_index_dirk,ii+1);      //  search_index_map[ii] is for
      fid = fopen(indexfile,"r");                                          //    search_index_file_(ii+1)
      if (! fid) break;
      pp = fgets_trim(buff,indexrecl,fid);
      if (pp && strnEqu(pp,"file: ",6)) 
         pp = strdupz(pp+6,0,"search-index");                              //  map first image file
      else pp = strdupz("empty",0,"search-index");                         //  if none, map "empty"
      search_index_map[ii] = pp;
      fclose(fid);
   }

   search_index_map[ii] = 0;                                               //  mark EOL

   if (ii == 999) {
      zmessLogACK(mWin,"too many search index files");
      return 1;
   }

   return 0;
}


//  Return which search index file a given image file belongs in.

int find_search_index_file(cchar *file)                                    //  v.12.10
{
   char     *pp;
   int      ii, nn;
   
   init_search_index_map(0);

   for (ii = 0; (pp = search_index_map[ii]); ii++)
   {
      if (strEqu(pp,"empty")) continue;                                    //  same sort as syncfiles_index_compare()
      nn = strcasecmp(file,pp);                                            //  bugfix v.12.11
      if (nn == 0) nn = strcmp(file,pp);
      if (nn < 0) break;                                                   //  file < 1st index file, go back 1
   }
   
   if (ii == 0) ii = 1;                                                    //  file < 1st file in 1st index file
   
   while (ii > 1 && strEqu(search_index_map[ii-1],"empty")) ii--;
   return ii;                                                              //  (map ii is for index file ii+1)
}


/**************************************************************************/

//  Split an index file when the number of entries > search_index_max
//  All following index files are renumbered +1.
//  Index_index file is refreshed.
//  Returns 0 if OK, else +N.

int split_search_index_file(int Nth)                                       //  v.12.10
{
   int            ii, err, nn, last, filecount;
   char           indexfile1[200], indexfile2[200], indexfile3[200];
   char           buff[indexrecl], *pp;
   FILE           *fid1 = 0, *fid2 = 0, *fid3 = 0;
   struct stat    statb;

   if (sxrec_fid) {                                                        //  stop get_search_index_rec()  v.13.01
      fclose(sxrec_fid);
      sxrec_fid = 0;
   }

   for (last = Nth; last < 999; last++)                                    //  find last index file
   {
      snprintf(indexfile1,200,"%s/index_%03d",search_index_dirk,last);     //  /.../.fotoxx/search_index/index_NNN
      err = stat(indexfile1,&statb);
      if (err) break;
   }

   last--;

   if (last == 999) {
      zmessLogACK(mWin,"index file limit reached");
      return 1;
   }

   for (ii = last; ii > Nth; ii--)                                         //  rename files following Nth
   {                                                                       //    index_012 >> index_013 etc.
      snprintf(indexfile1,200,"%s/index_%03d",search_index_dirk,ii);
      snprintf(indexfile2,200,"%s/index_%03d",search_index_dirk,ii+1);
      err = rename(indexfile1,indexfile2);
      if (err) goto file_err;
   }

   snprintf(indexfile1,200,"%s/index_%03d",search_index_dirk,Nth);         //  read Nth file
   snprintf(indexfile2,200,"%s/index_%03d_temp2",search_index_dirk,Nth);   //  write 2 temp. files
   snprintf(indexfile3,200,"%s/index_%03d_temp3",search_index_dirk,Nth);
   
   fid1 = fopen(indexfile1,"r");
   if (! fid1) goto file_err;
   fid2 = fopen(indexfile2,"w");
   if (! fid2) goto file_err;
   fid3 = fopen(indexfile3,"w");
   if (! fid3) goto file_err;

   filecount = 0;   

   while (true)                                                            //  copy half the entries
   {                                                                       //    to temp2 file
      pp = fgets_trim(buff,indexrecl,fid1);
      if (! pp) break;
      if (strnEqu(pp,"file: ",6)) filecount++;
      if (filecount > search_index_max/2) break;
      nn = fprintf(fid2,"%s\n",pp);
      if (! nn) goto file_err;
   }

   if (pp) {
      nn = fprintf(fid3,"%s\n",pp);                                        //  copy remaining record sets
      if (! nn) goto file_err;                                             //    to temp3 file
   }

   while (true)
   {
      pp = fgets_trim(buff,indexrecl,fid1);
      if (! pp) break;
      nn = fprintf(fid3,"%s\n",pp);
      if (! nn) goto file_err;
   }
   
   err = fclose(fid1);
   err = fclose(fid2);
   err = fclose(fid3);
   fid1 = fid2 = fid3 = 0;
   if (err) goto file_err;
   
   err = rename(indexfile2,indexfile1);                                    //  temp2 file replaces Nth index file
   if (err) goto file_err;
   
   snprintf(indexfile1,200,"%s/index_%03d",search_index_dirk,Nth+1);       //  temp3 file >> Nth+1 index file
   err = rename(indexfile3,indexfile1);
   if (err) goto file_err;
   
   err = init_search_index_map(1);                                         //  update search index map
   return err;

file_err:
   zmessLogACK(mWin,"split_search_index error \n %s",strerror(errno));
   if (fid1) fclose(fid1);
   if (fid2) fclose(fid2);
   if (fid3) fclose(fid3);
   return 3;
}


/**************************************************************************/

//  Get the search index record for given image file.
//  Returns 0 if OK, 1 if not found, 2 if error (diagnosed).
//  Returned sxrec_t data has allocated fields subject to zfree().
//
//  Index file is kept open across calls to reduce overhead for the 
//  normal case of records being accessed in image file sequence
//  (gallery paint). Random access works, but is slower.

int get_search_index_rec(sxrec_t &sxrec, cchar *file)                      //  v.13.01
{
   FILE           *fid = 0;
   char           indexfile[200];
   static char    buff[indexrecl];
   static char    pfile[maxfcc];
   int            err, Nth, nn = 0, Fcontinue;
   cchar          *pp;
   static int     pNth;
   struct stat    statb;
   struct tm      bdt;

   memset(&sxrec,0,sizeof(sxrec));                                         //  clear output record
   if (! FsyncEnabled) return 2;                                           //  search index not being used
   err = stat(file,&statb);                                                //  check image file exists
   if (err) return 1;

   Fcontinue = 1;                                                          //  test if prior search can continue
   if (! sxrec_fid) Fcontinue = 0;                                         //  no prior search still open
   Nth = find_search_index_file(file);                                     //  get index file for this image file
   if (Nth != pNth) Fcontinue = 0;                                         //  index file not the same
   if (navi::image_fcomp(file,pfile) <= 0) Fcontinue = 0;                  //  gallery file sequence

   if (Fcontinue) fid = sxrec_fid;
   else {
      if (sxrec_fid) fclose(sxrec_fid);
      snprintf(indexfile,200,"%s/index_%03d",search_index_dirk,Nth);       //  /.../.fotoxx/search_index/index_NNN
      fid = fopen(indexfile,"r");                                          //  open index_NNN file
      sxrec_fid = fid;
      if (! fid) goto file_err;
      *buff = 0;                                                           //  no prior data in buffer
   }

   if (! strnEqu(buff,"file: ",6) || strNeq(buff+6,file))                  //  file in buffer not my file
   {
      while (true)                                                         //  loop until my file found
      {
         pp = fgets_trim(buff,indexrecl,fid);                              //  read existing records
         if (! pp) break;                                                  //  EOF
         if (! strnEqu(pp,"file: ",6)) continue;                           //  to start of next file set
         nn = navi::image_fcomp(pp+6,file);
         if (nn >= 0) break;                                               //  same or after my file
      }

      if (! pp || nn > 0) {                                                //  EOF or after, file not found
         fclose(fid);
         sxrec_fid = 0;
         return 1;
      }
   }
   
   pNth = Nth;                                                             //  set up for poss. continuation
   strcpy(pfile,file);
   
   gmtime_r(&statb.st_mtime,&bdt);                                         //  image file date
   snprintf(sxrec.fdate,16,"%04d%02d%02d%02d%02d%02d",
            bdt.tm_year+1900, bdt.tm_mon+1, bdt.tm_mday,
            bdt.tm_hour, bdt.tm_min, bdt.tm_sec);

   while (true)                                                            //  get recs following "file" rec.
   {
      pp = fgets_trim(buff,indexrecl,fid);
      if (! pp) break;
      if (strnEqu(pp,"file: ",6)) break;                                   //  ran into next file rec.

      if (strnEqu(pp,"date: ",6)) {                                        //  EXIF (photo) date, yyyy:mm:dd
         pp = strField(pp,' ',2);
         if (pp) strncpy0(sxrec.pdate,pp,12);
      }
      
      else if (strnEqu(pp,"stars: ",7)) {                                  //  rating, '0' to '5' stars
         sxrec.rating[0] = *(pp+7);
         sxrec.rating[1] = 0;
      }
      
      else if (strnEqu(pp,"size: ",6))                                     //  size, "NNNNxNNNN"            v.13.01
         strncpy0(sxrec.size,pp+6,15);

      else if (strnEqu(pp,"tags: ",6))                                     //  tags
         sxrec.tags = strdupz(pp+6,0,"search-index");

      else if (strnEqu(pp,"capt: ",6))                                     //  caption
         sxrec.capt = strdupz(pp+6,0,"search-index");

      else if (strnEqu(pp,"comms: ",7))                                    //  comments
         sxrec.comms = strdupz(pp+7,0,"search-index");

      else if (strnEqu(pp,"gtags: ",7))                                    //  geotags
         sxrec.gtags = strdupz(pp+7,0,"search-index");
   }

   if (! sxrec.pdate[0])                                                   //  supply defaults for missing items
      strcpy(sxrec.pdate,"null");

   if (! sxrec.rating[0])
      strcpy(sxrec.rating,"0");

   if (! sxrec.size[0])                                                    //  v.13.01
      strcpy(sxrec.size,"null");

   if (! sxrec.tags) 
      sxrec.tags = strdupz("null"tagdelim2,0,"search-index");

   if (! sxrec.capt) 
      sxrec.capt = strdupz("null",0,"search-index");

   if (! sxrec.comms) 
      sxrec.comms = strdupz("null",0,"search-index");

   if (! sxrec.gtags) 
      sxrec.gtags = strdupz("null^ null^ null^ null",0,"search-index");

   return 0;

file_err:
   zmessLogACK(mWin,"search index error \n %s",strerror(errno));
   if (fid) fclose(fid);
   sxrec_fid = 0;
   return 3;
}


/**************************************************************************/

//  minimized version of get_search_index_rec() for use by gallery window.
//  pdate[12] and size[16] are provided by caller for returned data.

int get_search_index_rec_min(cchar *file, char *pdate, char *size)
{
   FILE           *fid = 0;
   char           indexfile[200];
   static char    buff[indexrecl];
   static char    pfile[maxfcc];
   int            Nth, nn = 0, Fcontinue;
   cchar          *pp;
   static int     pNth;

   strcpy(pdate,"undated");                                                //  outputs = missing
   strcpy(size,"");

   if (! FsyncEnabled) return 1;                                           //  search index not being used
   
   Fcontinue = 1;                                                          //  test if prior search can continue
   if (! sxrec_fid) Fcontinue = 0;                                         //  no prior search still open
   Nth = find_search_index_file(file);                                     //  get index file for this image file
   if (Nth != pNth) Fcontinue = 0;                                         //  index file not the same
   if (navi::image_fcomp(file,pfile) <= 0) Fcontinue = 0;                  //  req. file not after prior file

   if (Fcontinue) fid = sxrec_fid;
   else {
      if (sxrec_fid) fclose(sxrec_fid);
      snprintf(indexfile,200,"%s/index_%03d",search_index_dirk,Nth);       //  /.../.fotoxx/search_index/index_NNN
      fid = fopen(indexfile,"r");                                          //  open index_NNN file
      sxrec_fid = fid;
      if (! fid) goto file_err;
      *buff = 0;                                                           //  no prior data in buffer
   }

   if (! strnEqu(buff,"file: ",6) || strNeq(buff+6,file))                  //  file in buffer not my file
   {
      while (true)                                                         //  loop until my file found
      {
         pp = fgets_trim(buff,indexrecl,fid);                              //  read existing records
         if (! pp) break;                                                  //  EOF
         if (! strnEqu(pp,"file: ",6)) continue;                           //  to start of next file set
         nn = navi::image_fcomp(pp+6,file);
         if (nn >= 0) break;                                               //  same or after my file
      }

      if (! pp || nn > 0) {                                                //  EOF or after, file not found
         fclose(fid);
         sxrec_fid = 0;
         return 1;
      }
   }
   
   pNth = Nth;                                                             //  set up for poss. continuation
   strcpy(pfile,file);
   
   while (true)                                                            //  get recs following "file" rec.
   {
      pp = fgets_trim(buff,indexrecl,fid);
      if (! pp) break;
      if (strnEqu(pp,"file: ",6)) break;                                   //  ran into next file rec.

      if (strnEqu(pp,"date: ",6)) {                                        //  EXIF (photo) date, yyyy:mm:dd
         pp = strField(pp,' ',2);
         if (pp && *pp != 'n') strncpy0(pdate,pp,12);                      //  not "null"
      }
      
      else if (strnEqu(pp,"size: ",6))                                     //  size, "NNNNxNNNN"            v.13.01
         strncpy0(size,pp+6,15);
   }

   return 0;

file_err:
   zmessLogACK(mWin,"search index error \n %s",strerror(errno));
   if (fid) fclose(fid);
   sxrec_fid = 0;
   return 3;
}


/**************************************************************************/

//  Add or update search index record for given image file.
//  If sxrec is null, delete index record.
//  Return 0 if success, +N if error.

int put_search_index_rec(sxrec_t *sxrec, cchar *file)                      //  v.12.10
{
   FILE           *fid1 = 0, *fid2 = 0;
   char           indexfile[200], tempfile[200];
   char           buff[indexrecl];
   char           *pp;
   int            err, nn, Nth;
   int            filecount, Fcopy, Finsert, Finserted, Fnewfirst;
   struct stat    statb;
   struct tm      bdt;

   if (! FsyncEnabled) return 1;                                           //  search index not being used

   if (sxrec_fid) {                                                        //  stop get_search_index_rec()  v.13.01
      fclose(sxrec_fid);
      sxrec_fid = 0;
   }

   err = stat(file,&statb);                                                //  check image file exists
   if (err && sxrec) return 1;

   gmtime_r(&statb.st_mtime,&bdt);                                         //  get file date

   Nth = find_search_index_file(file);                                     //  construct index file
   snprintf(indexfile,200,"%s/index_%03d",search_index_dirk,Nth);          //  /.../.fotoxx/search_index/index_NNN
   
   err = stat(indexfile,&statb);                                           //  check if missing
   if (err) {
      nn = creat(indexfile,0640);                                          //  create if needed
      if (nn < 0) goto file_err;
      close(nn);
   }

   fid1 = fopen(indexfile,"r");                                            //  open index_NNN file
   if (! fid1) goto file_err;

   strcpy(tempfile,indexfile);                                             //  temp search index file
   strcat(tempfile,"_temp");

   fid2 = fopen(tempfile,"w");                                             //  open for output
   if (! fid2) goto file_err;

   Finsert = Finserted = Fcopy = Fnewfirst = 0;
   filecount = 0;

   while (true)                                                            //  copy input to output
   {
      pp = fgets_trim(buff,indexrecl,fid1);                                //  read existing records

      if (pp && strnEqu(pp,"file: ",6))                                    //  start of a file record set
      {
         Finsert = Fcopy = Fnewfirst = 0;

         nn = strcasecmp(file,pp+6);                                       //  compare input file to index file
         if (nn == 0) nn = strcmp(file,pp+6);                              //  bugfix v.12.11
         if (nn <= 0 && sxrec && ! Finserted) Finsert = 1;                 //  input <= index file, insert sxrec here
         if (nn != 0) Fcopy = 1;                                           //  input != index file, copy to output
         if (filecount == 0) {
            if (Finsert && nn < 0) Fnewfirst = 1;                          //  detect if first image file in index
            if (! Fcopy) Fnewfirst = 1;                                    //    file will be changed or deleted
         }
         filecount += Fcopy + Finsert;
      }

      if (! pp && sxrec && ! Finserted) {                                  //  input EOF, insert at the end
         Finsert = 1;
         if (filecount == 0) Fnewfirst = 1;
      }

      if (! pp && (Finserted || ! sxrec)) break;                           //  done

      if (Finsert && ! Finserted)
      {
         Finserted = 1;                                                    //  new index file recs. go here
         
         nn = fprintf(fid2,"file: %s\n",file);                             //  write new index file recs.
         if (! nn) goto file_err;

         sprintf(sxrec->fdate,"%04d%02d%02d%02d%02d%02d",                  //  file date, yyyymmddhhmmss
                  bdt.tm_year + 1900, bdt.tm_mon + 1, bdt.tm_mday,
                  bdt.tm_hour, bdt.tm_min, bdt.tm_sec);

         if (! sxrec->pdate[0]) strcpy(sxrec->pdate,"null");               //  EXIF (photo) date, yyyy:mm:dd

         nn = fprintf(fid2,"date: %s  %s\n",sxrec->pdate,sxrec->fdate);    //  photo and file date rec.
         if (! nn) goto file_err;

         if (sxrec->rating[0]) 
            nn = fprintf(fid2,"stars: %s\n",sxrec->rating);                //  rating rec.
         else nn = fprintf(fid2,"stars: 0\n"); 
         if (! nn) goto file_err;
         
         if (sxrec->size[0])                                               //  v.13.01
            nn = fprintf(fid2,"size: %s\n",sxrec->size);
         else nn = fprintf(fid2,"size: null\n");

         if (sxrec->tags) 
            nn = fprintf(fid2,"tags: %s\n",sxrec->tags);                   //  tags rec.
         else nn = fprintf(fid2,"tags: null"tagdelim2"\n");
         if (! nn) goto file_err;

         if (sxrec->capt) 
            nn = fprintf(fid2,"capt: %s\n",sxrec->capt);                   //  caption rec.
         else nn = fprintf(fid2,"capt: null\n");
         if (! nn) goto file_err;
         
         if (sxrec->comms) 
            nn = fprintf(fid2,"comms: %s\n",sxrec->comms);                 //  comments rec.
         else nn = fprintf(fid2,"comms: null\n");
         if (! nn) goto file_err;

         if (sxrec->gtags) 
            nn = fprintf(fid2,"gtags: %s\n",sxrec->gtags);                 //  geotags rec.
         else nn = fprintf(fid2,"gtags: null^ null^ null^ null\n");
         if (! nn) goto file_err;

         nn = fprintf(fid2,"\n");                                          //  EOL blank rec.
         if (! nn) goto file_err;
      }
      
      if (pp && Fcopy) {                                                   //  copy input to output
         nn = fprintf(fid2,"%s\n",pp);                                     //    unless replaced by sxrec
         if (! nn) goto file_err;
      }
   }

   err = fclose(fid1);                                                     //  close input file
   err = fclose(fid2);                                                     //  close output file
   fid1 = fid2 = 0;
   if (err) goto file_err;

   err = rename(tempfile,indexfile);                                       //  replace index file with temp file
   if (err) goto file_err;

   if (filecount > search_index_max)                                       //  if index file too big, split
      split_search_index_file(Nth);
   else if (Fnewfirst) 
      init_search_index_map(1);                                            //  update search index map

   return 0;

file_err:
   zmessLogACK(mWin,"search index error \n %s",strerror(errno));
   if (fid1) fclose(fid1);
   if (fid2) fclose(fid2);
   return 3;
}


/**************************************************************************/

//  Read search index files sequentially, return one index rec. per call.
//  Set ftf = 1 for first read, will be reset to 0.
//  Returns 0 if OK, 1 if EOF, 2 if error (diagnosed).
//  Returned sxrec_t data has allocated fields subject to zfree().

int read_search_index_seq(sxrec_t &sxrec, int &ftf)                        //  v.12.10
{
   char              indexfile[200];
   static FILE       *fid = 0;
   static char       buff[indexrecl];
   static int        Nth;
   cchar             *pp;
   int               err;
   struct stat       statb;
   struct tm         bdt;
   
   if (! FsyncEnabled) return 1;                                           //  search index not used     v.13.01

   if (ftf)                                                                //  initial call
   {
      ftf = 0;
      snprintf(indexfile,200,"%s/index_001",search_index_dirk);            //  first index file
      fid = fopen(indexfile,"r");
      if (! fid) return 2;                                                 //  no index file ?
      Nth = 1;
      *buff = 0;
   }
   
   while (true)
   {
      if (! strnEqu(buff,"file: ",6))                                      //  next file rec. may be already there
      {
         while (true)                                                      //  get start of next record set
         {
            pp = fgets_trim(buff,indexrecl,fid);                                 
            if (pp && strnEqu(buff,"file: ",6)) break;
            if (pp) continue;
            fclose(fid);                                                   //  EOF, start next index file
            Nth++;
            snprintf(indexfile,200,"%s/index_%03d",search_index_dirk,Nth);
            fid = fopen(indexfile,"r");
            if (! fid) return 1;                                           //  no more, final EOF
         }
      }

      *buff = 0;                                                           //  no longer match "file: "
      err = stat(buff+6,&statb);                                           //  check image file exists
      if (err) continue;
      if (S_ISREG(statb.st_mode)) break;
   }

   memset(&sxrec,0,sizeof(sxrec));                                         //  clear record
   
   sxrec.file = strdupz(buff+6,0,"search-index");                          //  image file name

   gmtime_r(&statb.st_mtime,&bdt);                                         //  refresh file date
   snprintf(sxrec.fdate,16,"%04d%02d%02d%02d%02d%02d",
            bdt.tm_year+1900, bdt.tm_mon+1, bdt.tm_mday,
            bdt.tm_hour, bdt.tm_min, bdt.tm_sec);

   while (true)                                                            //  get recs following "file" rec.
   {
      pp = fgets_trim(buff,indexrecl,fid);
      if (! pp) break;

      if (strnEqu(pp,"file: ",6)) break;                                   //  ran into next file rec.

      else if (strnEqu(pp,"date: ",6))
      {
         pp = strField(pp,' ',2);                                          //  EXIF (photo) date, yyyy:mm:dd
         if (pp) strncpy0(sxrec.pdate,pp,12);
      }
      
      else if (strnEqu(pp,"stars: ",7)) {                                  //  rating, '0' to '5' stars
         sxrec.rating[0] = *(pp+7);
         sxrec.rating[1] = 0;
      }

      else if (strnEqu(pp,"size: ",6))                                     //  size, "NNNNxNNNN"            v.13.01
         strncpy0(sxrec.size,pp+6,15);

      else if (strnEqu(pp,"tags: ",6))                                     //  tags
         sxrec.tags = strdupz(pp+6,0,"search-index");

      else if (strnEqu(pp,"capt: ",6))                                     //  caption
         sxrec.capt = strdupz(pp+6,0,"search-index");

      else if (strnEqu(pp,"comms: ",7))                                    //  comments
         sxrec.comms = strdupz(pp+7,0,"search-index");

      else if (strnEqu(pp,"gtags: ",7))                                    //  geotags
         sxrec.gtags = strdupz(pp+7,0,"search-index");
   }

   if (! sxrec.pdate[0])                                                   //  supply defaults for missing items
      strcpy(sxrec.pdate,"null");
   
   if (! sxrec.rating[0])
      strcpy(sxrec.rating,"0");

   if (! sxrec.size[0])                                                    //  v.13.01
      strcpy(sxrec.size,"null");

   if (! sxrec.tags) 
      sxrec.tags = strdupz("null"tagdelim2,0,"search-index");
   
   if (! sxrec.capt) 
      sxrec.capt = strdupz("null",0,"search-index");
   
   if (! sxrec.comms) 
      sxrec.comms = strdupz("null",0,"search-index");
   
   if (! sxrec.gtags) 
      sxrec.gtags = strdupz("null^ null^ null^ null",0,"search-index");
   
   return 0;
}


/**************************************************************************/

//  Write the search index files sequentially, 1 rec. per call
//  Set ftf = 1 for first call, will be reset to 0.
//  Set sxrec = 0 to close file after last write.
//  Returns 0 if OK, otherwise +N.
//  Used by synchronize files function.

int write_search_index_seq(sxrec_t *sxrec, int &ftf)                       //  v.12.10
{
   static int     Nth, filecount;
   static FILE    *fid = 0;
   char           indexfile[200], scratch[250];
   int            nn, err;
   struct stat    statb;
   struct tm      bdt;

   if (! FsyncEnabled) return 1;                                           //  search index not used     v.13.01

   if (sxrec_fid) {                                                        //  stop get_search_index_rec()  v.13.01
      fclose(sxrec_fid);
      sxrec_fid = 0;
   }

   if (ftf)                                                                //  first call 
   {
      ftf = 0;

      err = stat(search_index_dirk,&statb);
      if (! err && S_ISREG(statb.st_mode)) {                               //  rename old search_index
         strcpy(scratch,search_index_dirk);
         strcat(scratch,"_old");
         rename(search_index_dirk,scratch);
      }

      err = stat(search_index_dirk,&statb);                                //  create new search_index directory
      if (err) err = mkdir(search_index_dirk,0750);                        //    if not already there
      if (err) goto file_err;

      snprintf(scratch,250,"rm -f -v %s/*",search_index_dirk);             //  delete all search index files
      strcat(scratch," > /dev/null");                                      //  supress many "removed" messages
      err = system(scratch);
      Nth = 0;
      filecount = 0;
      fid = 0;
   }
   
   if (! sxrec) {                                                          //  EOF call
      if (fid) {
         err = fclose(fid);
         fid = 0;
         if (err) goto file_err;
      }
      err = init_search_index_map(1);                                      //  initz. search index map
      return err;
   }
   
   err = stat(sxrec->file,&statb);                                         //  check image file exists
   if (err || ! S_ISREG(statb.st_mode)) {
      printf("file %s not found \n",sxrec->file);
      return 0;
   }

   gmtime_r(&statb.st_mtime,&bdt);
   sprintf(sxrec->fdate,"%04d%02d%02d%02d%02d%02d",                        //  get file date
            bdt.tm_year + 1900, bdt.tm_mon + 1, bdt.tm_mday,               //  yyyymmddhhmmss
            bdt.tm_hour, bdt.tm_min, bdt.tm_sec);

   if (filecount > 0.8 * search_index_max) {                               //  if 80% max reached,
      err = fclose(fid);                                                   //    start new index file
      fid = 0;
      if (err) goto file_err;
      filecount = 0;
   }

   if (! filecount) {                                                      //  open new index file
      if (++Nth > 999) {
         zmessLogACK(mWin,"too many search index files");
         return 2;
      }
      snprintf(indexfile,200,"%s/index_%03d",search_index_dirk,Nth);
      fid = fopen(indexfile,"w");
      if (! fid) goto file_err;
   }

   filecount++;

   nn = fprintf(fid,"file: %s\n",sxrec->file);                             //  output file rec.
   if (! nn) goto file_err;

   if (! sxrec->pdate[0]) strcpy(sxrec->pdate,"null");                     //  EXIF (photo) date, yyyy:mm:dd

   nn = fprintf(fid,"date: %s  %s\n",sxrec->pdate,sxrec->fdate);           //  photo and file date rec.
   if (! nn) goto file_err;

   if (sxrec->rating[0]) 
      nn = fprintf(fid,"stars: %c\n",sxrec->rating[0]);                    //  rating rec.
   else nn = fprintf(fid,"stars: 0\n"); 
   if (! nn) goto file_err;

   if (sxrec->size[0])                                                     //  v.13.01
      nn = fprintf(fid,"size: %s\n",sxrec->size);
   else nn = fprintf(fid,"size: null\n");

   if (sxrec->tags) 
      nn = fprintf(fid,"tags: %s\n",sxrec->tags);                          //  tags rec.
   else nn = fprintf(fid,"tags: null"tagdelim2"\n");
   if (! nn) goto file_err;

   if (sxrec->capt) 
      nn = fprintf(fid,"capt: %s\n",sxrec->capt);                          //  caption rec.
   else nn = fprintf(fid,"capt: null\n");
   if (! nn) goto file_err;
   
   if (sxrec->comms) 
      nn = fprintf(fid,"comms: %s\n",sxrec->comms);                        //  comments rec.
   else nn = fprintf(fid,"comms: null\n");
   if (! nn) goto file_err;

   if (sxrec->gtags) 
      nn = fprintf(fid,"gtags: %s\n",sxrec->gtags);                        //  geotags rec.
   else nn = fprintf(fid,"gtags: null^ null^ null^ null\n");
   if (! nn) goto file_err;

   nn = fprintf(fid,"\n");                                                 //  EOL blank rec.
   if (! nn) goto file_err;

   return 0;

file_err:
   zmessLogACK(mWin,"search index file error \n %s",strerror(errno));
   if (fid) fclose(fid);
   fid = 0;
   return 1;
}


